# Timer

## Installation

```
npm install -s git+ssh://git@git.keywesttechnology.com:keywest/timer.git
```

## Usage

Import Timer from the scoped namespace


```
const Timer = require('@keywest/timer');
```

<short desc>

### Wait 5 seconds

```
await Timer.sleep(5000);
```

### Wait for some action to finish

```
const someActionReturningSomeStatus = function() { ... };
const someConditionalThatReturnsTrueForSomeStatus =  function(status) { ... };
const delayBetweenActionChecks = 5000;
const numberOfActionChecksBeforeGivingUp = 10;

await Timer.wait(
	someActionReturningSomeStatus,
	someConditionalThatReturnsTrueForSomeStatus,
	delayBetweenActionChecks,
	numberOfActionChecksBeforeGivingUp);
```

A real world usage example might be easier to follow so here is a function from the DigitalOcean API that waits for a droplet to have the desired status (waiting 6s between refreshing the droplet status and trying up to 100 times before giving up):

```
/**
 * @param {!number} dropletId
 * @param {!string} status
 * @param {!number} [delay=6000]
 * @param {!number} [attempts=100]
 * @returns {!Promise<!Action>}
 */
async waitDropletStatus(dropletId, status, delay = 6000, attempts = 100) {
    const getDroplet = this.getDropletById.bind(this, dropletId);
    const predicate = droplet => droplet.status === status;
    return Timer.wait(getDroplet, predicate, delay, attempts);
}
```
