class Timer {
	/**
	 * @param {!number} delay In milliseconds.
	 * @returns {!Promise}
	 */
	static async sleep(delay) {
		return new Promise(resolve => setTimeout(resolve, delay));
	}

	/**
	 * Waits until predicate(action()) returns true, or until a maximum number
	 * of calls to action().  Returns the result from the last call to action.
	 * @param {!Function} action
	 * @param {!Function} predicate
	 * @param {!number} [delay=5000] Delay to wait before each try.
	 * @param {!number} [attempts=3] Number of tries before giving up.
	 * @returns {!Promise}
	 */
	static async wait(action, predicate, delay = 5000, attempts = 3) {
		let tries = 0;
		let result;
		do {
			await Timer.sleep(delay);
			result = await action();
			if (++tries >= attempts) {
				break;
			}
		} while (!predicate(result));
		return result;
	}
}

module.exports = Timer;
