class FileUtils {
	/**
	 * @param {!string} name
	 * @param {!string} folder
	 * @param {!string} pathsKey
	 * @returns {!Promise<?Object>}
	 */
	static stat(name, folder, pathsKey) {
		const filepath = path.join(paths[pathsKey], folder, name);
		return FileIO.stat(filepath);
	}

	/**
	 * @param {!string} name
	 * @param {!string} oldPathsKey
	 * @param {!string} newPathsKey
	 * @returns {!Promise}
	 */
	static move(name, oldPathsKey, newPathsKey) {
		const oldPath = path.join(paths[oldPathsKey], name);
		const newPath = path.join(paths[newPathsKey], name);
		return fs.move(oldPath, newPath, { overwrite: true });
	}

	/**
	 * @param {!string} name
	 * @param {!string} folder
	 * @param {!string} pathsKey
	 * @returns {!Promise<?string>}
	 */
	static load(name, folder, pathsKey) {
		const filepath = path.join(paths[pathsKey], folder, name);
		return FileIO.read(filepath);
	}

	/**
	 * @param {!string} name
	 * @param {!string} folder
	 * @param {!string} pathsKey
	 * @param {!string} data
	 * @returns {!Promise<!string>}
	 */
	static async save(name, folder, pathsKey, data) {
		const folderpath = path.join(paths[pathsKey], folder);
		await fs.mkdirs(folderpath);

		const filepath = path.join(folderpath, name);
		const stats = await FileIO.stat(filepath);
		if (stats === null) {
			return fs.writeFile(filepath, data);
		}

		// Since the file already exists, write to a temporary directory
		// instead, and then move and overwrite the existing file.
		const tmpFolderpath = path.join(paths.TMP, folder);
		await fs.mkdirs(tmpFolderpath);

		const tmpFilepath = path.join(tmpFolderpath, name);
		await fs.writeFile(tmpFilepath, data);
		await fs.move(tmpFilepath, filepath, { overwrite: true });

		return filepath;
	}

	/**
	 * Resolves to a list of the files deleted.
	 * @param {!string} pathsKey
	 * @param {!string} ext
	 * @returns {!Promise<!string[]>}
	 */
	static async deleteFilesByExt(pathsKey, ext) {
		await fs.mkdirs(paths[pathsKey]);
		return FileIO.deleteFilesByExt(paths[pathsKey], ext);
	}

	/**
	 * Resolves to a list of the files deleted.
	 * @param {!string} pathsKey
	 * @param {!RegExp} regex
	 * @returns {!Promise<!string[]>}
	 */
	static async deleteFilesMatching(pathsKey, regex) {
		await fs.mkdirs(paths[pathsKey]);
		return FileIO.deleteFilesMatching(paths[pathsKey], regex);
	}
}
