var MEDIATYPE_IMAGE = 2;
var MEDIATYPE_VIDEO = 3;
var MEDIATYPE_AUDIO = 5;
var MEDIATYPE_MISC = 7;
var MEDIATYPE_WIDGET = 6;
var MEDIATYPE_PDF = 9;

if (typeof Node.prototype.append !== "function") {
	Node.prototype.append = function(...args) {
		var argArr = Array.prototype.slice.call(args),
			docFrag = document.createDocumentFragment();

		argArr.forEach(function (argItem) {
			var isNode = argItem instanceof Node;
			docFrag.appendChild(isNode ? argItem : document.createTextNode(String(argItem)));
		});

		this.appendChild(docFrag);
	}
}

function get(url, data = {}, abortController = new AbortController()) {
	return fetch(url + '?' + new URLSearchParams(Object.entries(data)), {
		credentials: 'same-origin',
		cache: "no-cache",
		signal: abortController.signal
	}).then(response => response.json());
}

function post(url, data = {}, abortController = new AbortController()) {
	var formData = new FormData();
	for (const key in data) {
		formData.append(key, data[key]);
	}
	return fetch(url, {
		method: "POST",
		credentials: 'same-origin',
		cache: "no-cache",
		body: formData,
		signal: abortController.signal
	}).then(response => response.json());
}

function ConfirmationBox(title, message, choice1Text, choice2Text, choice3Text) {
	if (choice1Text == undefined && choice2Text == undefined && choice3Text == undefined) {
		return new Promise(resolve => {
			let $ok;
			const $confirmationBox = document.body.appendChild($.uidialog({ config: { modal: true }, css: { width: '550px' } },
				$.header({ text: title }),
				$.div({ text: message, class: 'panel-group', css: {
					color: 'var(--font-color)',
					whiteSpace: 'pre-wrap',
					'word-break': 'break-word'
				} }),
				$.div({ class: 'panel-buttons' },
					$ok = $.button({ text: 'OK', click: function() {
						$confirmationBox.remove();
						resolve();
					} })
				)
			));
			centerOnScreen($confirmationBox);
			$ok.focus();
		});
	} else {
		return new Promise((resolve, reject) => {
			choice1Text = choice1Text == undefined ? 'OK' : choice1Text;
			choice2Text = choice2Text == undefined ? '' : choice2Text;
			choice3Text = choice3Text == undefined ? '' : choice3Text;
			let $choice1;
			const $confirmationBox = document.body.appendChild($.uidialog({ config: { modal: true }, css: { width: '550px' } },
				$.header({ text: title }),
				$.div({ text: message, class: 'panel-group', css: {
					color: 'var(--font-color)',
					whiteSpace: 'pre-wrap',
					wordBreak: 'break-word'
				} }),
				$.div({ class: 'panel-buttons' },
					$.button({ text: choice3Text, hidden: choice3Text == '', click: function() {
						$confirmationBox.remove();
						if (choice3Text == 'Cancel') {
							reject();
						} else {
							resolve(2);
						}
					} }),
					$.button({ text: choice2Text, hidden: choice2Text == '', click: function() {
						$confirmationBox.remove();
						if (choice2Text == 'Cancel' || choice2Text == 'No' && choice3Text == '') {
							reject();
						} else {
							resolve(1);
						}
					} }),
					$choice1 = $.button({ text: choice1Text, click: function() {
						$confirmationBox.remove();
						resolve(0);
					} })
				)
			));
			centerOnScreen($confirmationBox);
			$choice1.focus();
		});
	}
}

function ConfirmationBoxBool(title, message, trueText = 'OK', falseText = 'Cancel') {
	return new Promise((resolve, reject) => {
		let $true;
		const $confirmationBox = document.body.appendChild($.uidialog({ config: { modal: true }, css: { width: '550px' } },
			$.header({ text: title }),
			$.div({ text: message, class: 'panel-group', css: {
				color: 'var(--font-color)',
				whiteSpace: 'pre-wrap',
				wordBreak: 'break-word'
			} }),
			$.div({ class: 'panel-buttons' },
				$.button({ text: falseText, hidden: choice2Text == '', click: function() {
					$confirmationBox.remove();
					resolve(false);
				} }),
				$true = $.button({ text: trueText, click: function() {
					$confirmationBox.remove();
					resolve(true);
				} })
			)
		));
		centerOnScreen($confirmationBox);
		$true.focus();
	});
}

function QuestionBox(title, message) {
	return new Promise((resolve, reject) => {
		let $yes;
		const $confirmationBox = document.body.appendChild($.uidialog({ config: { modal: true }, css: { width: '550px' } },
			$.header({ text: title }),
			$.div({ text: message, class: 'panel-group', css: {
				color: 'var(--font-color)',
				whiteSpace: 'pre-wrap',
				wordBreak: 'break-word'
			} }),
			$.div({ class: 'panel-buttons' },
				$.button({ text: 'No', click: function() {
					$confirmationBox.remove();
					resolve(false);
				} }),
				$yes = $.button({ text: 'Yes', click: function() {
					$confirmationBox.remove();
					resolve(true);
				} })
			)
		));
		centerOnScreen($confirmationBox);
		$yes.focus();
	});
}

function InputBox(title, message, choice1Text, choice2Text, choice3Text, choice1Callback, choice2Callback, choice3Callback) {
	return new Promise((resolve, reject) => {
		let $input;
		const $inputBox = document.body.appendChild($.uidialog({ config: { modal: true }, css: { width: '550px' } },
			$.header({ text: title }),
			$.div({ text: message, class: 'panel-group', css: { color: 'white' } }),
			$.div({ class: 'panel-group' },
				$input = $.input({ type: 'text', class: 'inputbox-input sampleInput' })
			),
			$.div({ class: 'panel-buttons' },
				$.button({ text: choice3Text, hidden: choice3Callback == null, click: async function() {
					try {
						var close = await choice3Callback($input.value);
						if (close === true) {
							$inputBox.remove();
							resolve();
						}
					} catch (exception) {
						reject(exception);
					}
				} }),
				$.button({ text: choice2Text, hidden: choice2Callback == null, click: async function() {
					try {
						var close = await choice2Callback($input.value);
						if (close === true) {
							$inputBox.remove();
							resolve();
						}
					} catch (exception) {
						reject(exception);
					}
				} }),
				$.button({ text: choice1Text, hidden: choice1Callback == null, click: async function() {
					try {
						var close = await choice1Callback($input.value);
						if (close === true) {
							$inputBox.remove();
							resolve();
						}
					} catch (exception) {
						reject(exception);
					}
				} })
			)
		));
		centerOnScreen($inputBox);
		$input.focus();
	});
}

function PreviewContent(account, media) {
	const maxWidth = 800;
	const maxHeight = 600;
	var preview = '';
	switch (media.contenttreetype) {
		case 3:
			preview = $.img({ src: GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize), css: {
				display: 'block'
			}, load: function() {
				Object.assign(this.closest('ui-dialog').querySelector('.preview-container').style, {
					maxWidth: '800px',
					maxHeight: '600px'
				});
				const scale = Math.min(maxWidth / this.offsetWidth, maxHeight / this.offsetHeight, 1);
				Object.assign(this.style, {
					width: `${this.offsetWidth * scale}px`,
					height: `${this.offsetHeight * scale}px`
				});
				centerOnScreen(this.closest('ui-dialog'));
			} });
			break;
		case 4:
			preview = $.audio({ controls: true, css: {
				display: 'block',
				maxHeight: '500px'
			}, loadedmetadata: function() {
				centerOnScreen(this.closest('ui-dialog'));
			} },
				$.source({ src: GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize) })
			);
			break;
		case 5:
			preview = $.video({ controls: true, css: {
				display: 'block',
				maxHeight: '500px'
			}, loadedmetadata: function() {
				Object.assign(this.closest('ui-dialog').querySelector('.preview-container').style, {
					maxWidth: '800px',
					maxHeight: '600px'
				});
				const scale = Math.min(maxWidth / this.offsetWidth, maxHeight / this.offsetHeight, 1);
				Object.assign(this.style, {
					width: `${this.offsetWidth * scale}px`,
					height: `${this.offsetHeight * scale}px`
				});
				centerOnScreen(this.closest('ui-dialog'));
			} },
				$.source({ src: GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize) })
			);
			break;
		case 8:
			preview = $.object({
				data: GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize),
				type: 'application/x-shockwave-flash',
				css: {
					display: 'block',
					maxHeight: '500px'
				}
			},
				$.param({ name: 'quality', value: 'high' }),
				$.param({ name: 'bgcolor', value: '#000000' }),
				$.param({ name: 'play', value: 'true' }),
				$.param({ name: 'loop', value: 'true' }),
				$.param({ name: 'wmode', value: 'opaque' }),
				$.param({ name: 'scale', value: 'showall' }),
				$.param({ name: 'menu', value: 'true' }),
				$.param({ name: 'devicefont', value: 'false' }),
				$.param({ name: 'salign', value: '' }),
				$.param({ name: 'allowScriptAccess', value: 'sameDomain' }),
				$.p({ text: 'This player does not have Adobe Flash installed.' })
			);
			break;
		case 9:
			preview = $.div({ css: {
				display: 'inline-block',
				width: '800px',
				height: '700px',
				verticalAlign: 'top'
			} });
			RenderPDF(preview, GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize), 0, 0, true, true);
			break;
	}
	const $dialog = document.body.appendChild($.uidialog({ config: { modal: true } },
		$.header({ text: media.name }),
		$.div({ class: 'preview-container' },
			preview
		),
		$.div({ class: 'panel-buttons' },
			$.button({ text: 'OK', click: function() {
				$dialog.remove();
			} })
		)
	));
	centerOnScreen($dialog);
}

function centerOnScreen($element) {
	$element.style.left = `${document.documentElement.clientWidth / 2 - $element.offsetWidth / 2}px`;
	$element.style.top = `${document.documentElement.clientHeight / 2 - $element.offsetHeight / 2}px`;
	setTimeout(() => {
		$element.style.left = `${document.documentElement.clientWidth / 2 - $element.offsetWidth / 2}px`;
		$element.style.top = `${document.documentElement.clientHeight / 2 - $element.offsetHeight / 2}px`;
	}, 100);
}

function multiclick($element, handler) {
	const singleClick = e => {
		const secondClick = e => {
			clearTimeout(secondClickTimeout);
			const thirdClick = e => {
				clearTimeout(thirdClickTimeout);
				// Three clicks
				handler.call($element, e, 3);
				$element.addEventListener('click', singleClick, { once: true });
			};
			const thirdClickTimeout = setTimeout(() => {
				$element.removeEventListener('click', thirdClick, { once: true });
				// Two clicks
				handler.call($element, e, 2);
				$element.addEventListener('click', singleClick, { once: true });
			}, 200);
			$element.addEventListener('click', thirdClick, { once: true });
		};
		const secondClickTimeout = setTimeout(() => {
			$element.removeEventListener('click', secondClick, { once: true });
			// One click
			handler.call($element, e, 1);
			$element.addEventListener('click', singleClick, { once: true });
		}, 200);
		$element.addEventListener('click', secondClick, { once: true });
	};
	$element.addEventListener('click', singleClick, { once: true });
}

let $titleTooltip = null;
let titleTooltipFade = null;
let $lastTooltipElement = null;
let lastTooltipMessage = '';

function UndoTooltip() {
	if ($lastTooltipElement) {
		$lastTooltipElement.removeEventListener('remove', $lastTooltipRemoveHandler);
		// Don't overwrite if the title was replaced.
		// TODO: Handle the issue if the title is set to '' while it's being displayed.
		if ($lastTooltipElement.title == '') {
			$lastTooltipElement.title = lastTooltipMessage;
		}
	}
	$lastTooltipElement = null;
}

function TooltipPointerEnter(e) {
	if ((e.buttons & 0x1) == 0) {
		UndoTooltip();
		if (this.title != '') {
			// Create the tooltip if it doesn't exist
			if ($titleTooltip == null) {
				$titleTooltip = document.body.appendChild($.span({ id: 'titletooltip' }));
				titleTooltipFade = $titleTooltip.animate([
					{ opacity: 0 },
					{ opacity: 1 }
				], {
					duration: 250,
					fill: 'both'
				});
			}
			$lastTooltipElement = this;
			this.addEventListener('remove', $lastTooltipRemoveHandler = () => {
				HideTooltip();
			});
			lastTooltipMessage = this.title;
			$titleTooltip.innerText = this.title;
			this.title = '';
		
			const width = $titleTooltip.offsetWidth;
			const height = $titleTooltip.offsetHeight;
			const thisOffset = $.offset(this);
			let left = thisOffset.left + 30;
			let top = thisOffset.top + this.offsetHeight + 15;
			if (left + width > document.documentElement.clientWidth) {
				left = document.documentElement.clientWidth - width - 30;
			}
			if (top + height > document.documentElement.clientHeight) {
				top = thisOffset.top - height - 15;
			}
			$titleTooltip.style.left = `${left}px`;
			$titleTooltip.style.top = `${top}px`;
			titleTooltipFade.playbackRate = 1;
		}
	}
}

function TooltipPointerLeave(e) {
	if ((e.buttons & 0x1) != 1) {
		UndoTooltip();
		if (titleTooltipFade) {
			titleTooltipFade.playbackRate = -1;
		}
	}
}

function HideTooltip() {
	UndoTooltip();
	if (titleTooltipFade) {
		titleTooltipFade.currentTime = 0;
		titleTooltipFade.playbackRate = -1;
	}
}

function customtooltip($element) {
	$element.removeEventListener('pointerenter', TooltipPointerEnter);
	$element.removeEventListener('pointerleave', TooltipPointerLeave);
	$element.addEventListener('pointerenter', TooltipPointerEnter);
	$element.addEventListener('pointerleave', TooltipPointerLeave);
	if ($element.title == '') {
		HideTooltip();
	}
	if ($lastTooltipElement === $element && $element.title != '') {
		lastTooltipMessage = $element.title;
		$titleTooltip.innerText = $element.title;
		$element.title = '';
	}
}

function tabs($element, { before = () => {}, after = () => {}} = {}) {
	const $tabs = $element.querySelector(':scope > ul:first-child');
	$element.classList.add('tabs');
	$tabs.classList.add('tabs-tabs');
	for (const [i, $tab] of [...$tabs.querySelectorAll(':scope > li')].entries()) {
		if ($tab.title == undefined || $tab.title == '') {
			$tab.title = $tab.innerText;
		}
		$tab.addEventListener('click', () => {
			before();
			$tabs.querySelector(':scope > li.selected')?.classList.remove('selected');
			$tab.classList.add('selected');
			$element.querySelector(':scope > :not(:first-child).selected')?.classList.remove('selected');
			$element.querySelectorAll(':scope > :not(:first-child)')[i].classList.add('selected');
			after();
		});
	}
	$tabs.querySelector(':scope > li:first-child')?.classList?.add('selected');
	$element.querySelector(':scope > :not(:first-child)')?.classList?.add('selected');
}

function customTabs($tabHandles, $tabBodies) {
	for (const [i, $tabHandle] of [...$tabHandles].entries()) {
		$tabHandle.classList.add('customtab-handle');
		if ($tabHandle.title == undefined || $tabHandle.title == '') {
			$tabHandle.title = $tabHandle.innerText;
		}
		$tabHandle.addEventListener('click', () => {
			for (const $e of $tabHandles) $e.classList.toggle('selected', $e === $tabHandle);
			for (const $e of $tabBodies) {
				$e.classList.toggle('selected', $e === $tabBodies[i]);
				$e.hidden = $e !== $tabBodies[i];
			}
		});
	}
	for (const $e of $tabHandles) $e.classList.toggle('selected', $e === $tabHandles[0]);
	for (const $e of $tabBodies) {
		$e.classList.toggle('selected', $e === $tabBodies[0]);
		$e.hidden = $e !== $tabBodies[0]
	}
}

function accordion($element) {
	for (const $header of $element.querySelectorAll(':scope > li:nth-child(odd)')) {
		var headerText = $header.innerText;
		$header.innerText = '';
		$header.append(
			$.span({ text: headerText }),
			$.i({ class: 'mdi mdi-play mdi-rotate-90' })
		);
		$header.addEventListener('click', function() {
			for (const $otherHeader of $element.querySelectorAll(':scope > li:nth-child(odd)')) $otherHeader.classList.add('collapsed');
			this.classList.remove('collapsed');
		});
	}
	$element.querySelector(':scope > li:first-child')?.click();
}

function IsFocusedElementAnInput() {
	return document.activeElement && (['input', 'select', 'button', 'textarea'].indexOf(document.activeElement.tagName.toLowerCase()) != -1 || document.activeElement.isContentEditable);
}

function GetSizingContainer() {
	const sizingContainer = document.getElementById('sizing-container');
	if (sizingContainer == null) {
		return document.body.appendChild($.div({ id: 'sizing-container', css: {
			display: 'inline-block',
			position: 'absolute',
			left: 0,
			top: 0,
			width: 0,
			height: 0,
			overflow: 'hidden'
		} }));
	}
	return sizingContainer;
}

// This uses a promise because ideally all the code that uses it should add everything they can to the sizing container at once causing one reflow.
function Measure($element, $elementToMeasure = $element) {
	// We use a permanent div to limit reflows, especially body tag width overflows which on our CSS grid layout is expensive
	// This can't grow the body tag essentially
	const sizingContainer = GetSizingContainer();
	$element.style.position = 'absolute';
	if (!$element.style.display.startsWith('inline')) {
		$element.style.display = 'inline-block';
	}
	sizingContainer.appendChild($element);
	return new Promise(resolve => {
		resolve({ width: $elementToMeasure.offsetWidth, height: $elementToMeasure.offsetHeight });
		$element.remove();
	});
}

function draggable($element, {
		handle = null,
		helper = $element,
		appendTo = null,
		containment = null,
		start = () => true,
		drag = () => {},
		stop = () => {},
		threshold = null,
		thresholdX = null,
		thresholdY = null,
		sortable = false,
		movePrevious = () => {},
		moveNext = () => {},
		filterDroppable = () => true,
		scrollParent = null }) {
	const $handle = handle ? $element.querySelector(handle) : $element;
	let startX;
	let startY;
	let startOffsetX;
	let startOffsetY;
	let lastX = 0;
	let lastY = 0;
	let dragSelection = null;
	const dragTags = [];
	const activeDropTags = [];
	let sort = true;
	let lastMouseEvent;
	let dragRepeat = null;
	
	const mouseDrag = (e, skipDragRepeat) => {
		const scrollTag = scrollParent ? (scrollParent instanceof Element ? scrollParent : $element.closest(scrollParent)) : $element;
		const scrollTagOffset = $.offset(scrollTag);
		let x = Math.floor(e.pageX - scrollTagOffset.left);
		let y = Math.floor(e.pageY - scrollTagOffset.top);
		if (x < 0) {
			x = 0;
		}
		if (y < 0) {
			y = 0;
		}
		drag.call($element, e, startX, startY, startOffsetX, startOffsetY, e.pageX - lastX, e.pageY - lastY, sort);
		lastX = e.pageX;
		lastY = e.pageY;
		clearTimeout(dragRepeat);
		let scaleX = e.pageX - (scrollTagOffset.left + scrollTag.clientWidth / 2);
		if (Math.abs(scaleX) < scrollTag.clientWidth / 2) {
			scaleX = 0;
		}
		let scaleY = e.pageY - (scrollTagOffset.top + scrollTag.clientHeight / 2);
		if (Math.abs(scaleY) < scrollTag.clientHeight / 2) {
			scaleY = 0;
		}
		if (scaleX != 0) {
			scrollTag.scrollLeft = Math.max(0, Math.min(scrollTag.scrollWidth, scrollTag.scrollLeft + (scaleX - Math.sign(scaleX) * scrollTag.clientWidth / 2)));
		}
		if (scaleY != 0) {
			scrollTag.scrollTop = Math.max(0, Math.min(scrollTag.scrollHeight, scrollTag.scrollTop + (scaleY - Math.sign(scaleY) * scrollTag.clientHeight / 2)));
		}
		if (!skipDragRepeat && (scaleX != 0 || scaleY != 0)) {
			dragRepeat = setTimeout(() => mouseDrag(lastMouseEvent, false), 16);
		}
	};
	
	const documentPointerMove = async e => {
		const $containment = containment instanceof HTMLElement ? containment : (typeof containment == 'string' ? $element.closest(containment) : document.body);
		if (scrollParent) {
			mouseDrag(e, false);
		}
		lastMouseEvent = e;
		if (dragTags.length != 0) {
			drag.call($element, e, startX, startY, startOffsetX, startOffsetY, e.pageX - lastX, e.pageY - lastY, sort);
			lastX = e.pageX;
			lastY = e.pageY;
			for (const $drag of dragTags.filter($drag => !$drag.hidden)) {
				$drag.reposition(e);
			}
		} else if (threshold != null || thresholdX != null || thresholdY != null) {
			const passesThresholdX = thresholdX != null && (typeof thresholdX == 'number' && Math.abs(e.pageX - startX) > thresholdX || (typeof thresholdX == 'object' || typeof thresholdX == 'string') && (e.pageX < $.offset(thresholdX).left || e.pageX >= $.offset(thresholdX).left + thresholdX.offsetWidth));
			const passesThresholdY = thresholdY != null && (typeof thresholdY == 'number' && Math.abs(e.pageY - startY) > thresholdY || (typeof thresholdY == 'object' || typeof thresholdY == 'string') && (e.pageY < $.offset(thresholdY).top || e.pageY >= $.offset(thresholdY).top + thresholdY.offsetHeight));
			if (threshold && threshold.call($element, e, startX, startY, startOffsetX, startOffsetY, e.pageX - lastX, e.pageY - lastY, dragSelection) || sortable == 'y' && passesThresholdX || sortable == 'x' && passesThresholdY || !sortable && (passesThresholdX || passesThresholdY)) {
				sort = false;
				// Note: The order of these operations from calling helper to setting dragTags is very defined and cannot be rearranged without changing the behavior of the program. Be careful when modifying this code
				$element.removeEventListener('remove', documentPointerUp);
				
				// Drag tags (the elements that form when dragging a draggable around) are all created during the initial drag start to mostly avoid performance issues
				let defaultDragTag;
				if (typeof helper === 'function') {
					// Note, that this can select the node which is why it's called so early in this function
					defaultDragTag = await helper($element, e, '');
					if (defaultDragTag === false) {
						// The drag was cancelled
						documentPointerUp.call(this, e, false);
						return;
					}
				} else {
					defaultDragTag = $element.cloneNode(true);
					defaultDragTag.style.width = `${$element.offsetWidth}px`;
					defaultDragTag.style.height = `${$element.offsetHeight}px`;
				}
				defaultDragTag.region = '';
				dragTags.push(defaultDragTag);
				
				$element.classList.add('dragging-source');
				const uniqueRegions = new Set();
				for (const className of $element.classList) {
					// Find all the droppable tags that the dragged object can be dropped onto
					let dropTags = [...document.querySelectorAll(`.droppable-${className}`)];
					for (const $e of [$element, ...dragTags, ...(function*() { for (const $drag of dragTags) yield* $drag.querySelectorAll('*') })()]) {
						const dropTagElementIndex = dropTags.indexOf($e);
						if (dropTagElementIndex != -1) {
							dropTags.splice(dropTagElementIndex, 1);
						}
					}
					const filterResult = await Promise.all(dropTags.map($drag => filterDroppable($drag)));
					dropTags = dropTags.filter(($drag, index) => filterResult[index]);
					const onDrop = async function(e) {
						e.stopPropagation();
						documentPointerUp.call(this, e, true);
						await this[`droppable-${className}`].call(this, e, dragSelection);
					};
					const onPointerOver = function(e) {
						const dropRegion = this[`droppableregion-${className}`];
						let dragTagExists = false;
						for (const $drag of dragTags) {
							if ($drag.region == dropRegion) {
								dragTagExists = true;
							}
							$drag.hidden = $drag.region != dropRegion;
							if (!$drag.hidden) {
								$drag.reposition(e);
							}
						}
						// Show the default if a specialized drag tag doesn't exist for this drop region
						if (!dragTagExists) {
							defaultDragTag.hidden = false;
						}
					};
					const onPointerOut = function() {
						for (const $drag of dragTags) {
							$drag.hidden = $drag.region != '';
						}
					};
					for (const $drop of dropTags) {
						uniqueRegions.add($drop[`droppableregion-${className}`]);
						$drop.classList.add('active');
						$drop.drop = onDrop;
						$drop.addEventListener('pointerup', onDrop, true);
						$drop.pointerover = onPointerOver;
						$drop.addEventListener('pointerover', onPointerOver, true);
						$drop.pointerout = onPointerOut;
						$drop.addEventListener('pointerout', onPointerOut, true);
					}
					activeDropTags.push(...dropTags);
				}
				
				uniqueRegions.delete('');
				if (typeof helper === 'function') {
					// Every drop region is allowed its own unique drag helper. So the dragging items changes depending on what the user is dragging over
					for (const region of uniqueRegions) {
						const $drag = await helper($element, e, region);
						if ($drag) {
							$drag.region = region;
							dragTags.push($drag);
						}
					}
				}
				const $appendTo = appendTo instanceof HTMLElement ? appendTo : (typeof appendTo == 'string' ? $element.closest(appendTo) : document.body);
				for (const $drag of dragTags) {
					$drag.classList.add('dragging');
					$appendTo.appendChild($drag);
					$drag.reposition = e => {
						const containmentOffset = $.offset($containment);
						const rect = $drag.getBoundingClientRect();
						const maxLeft = Math.floor($containment.scrollWidth - rect.width);
						const maxTop = Math.floor($containment.scrollHeight - rect.height);
						$drag.style.left = `${Math.max(0, Math.min(e.pageX - startOffsetX - containmentOffset.left + $containment.scrollLeft, maxLeft))}px`;
						$drag.style.top = `${Math.max(0, Math.min(e.pageY - startOffsetY - containmentOffset.top + $containment.scrollTop, maxTop))}px`;
					};
					$drag.reposition(e);
					$drag.originalLeft = $drag.style.left;
					$drag.originalTop = $drag.style.top;
					// Set only the default helper to visible
					$drag.hidden = $drag.region != '';
				}
			} else if (sortable === 'y') {
				const elementOffset = $.offset($element);
				if (e.pageY < elementOffset.top && ($element.parentNode.style.display == 'grid' || $element.previousElementSibling) && movePrevious() !== false) {
					$element.previousElementSibling?.insertAdjacentElement('beforebegin', $element);
				}
				if (e.pageY > elementOffset.top + $element.offsetHeight && ($element.parentNode.style.display == 'grid' || $element.nextElementSibling) && moveNext() !== false) {
					$element.nextElementSibling?.insertAdjacentElement('afterend', $element);
				}
			} else if (sortable === 'x') {
				const elementOffset = $.offset($element);
				if (e.pageX < elementOffset.left && ($element.parentNode.style.display == 'grid' || $element.previousElementSibling) && movePrevious() !== false) {
					$element.previousElementSibling?.insertAdjacentElement('beforebegin', $element);
				}
				if (e.pageX > elementOffset.left + $element.offsetWidth && ($element.parentNode.style.display == 'grid' || $element.nextElementSibling) && moveNext() !== false) {
					$element.nextElementSibling?.insertAdjacentElement('afterend', $element);
				}
			}
		}
	};
	
	const documentPointerUp = function(e, dropped = false) {
		const $containment = containment instanceof HTMLElement ? containment : (typeof containment == 'string' ? $element.closest(containment) : document.body);
		clearTimeout(dragRepeat);
		if (!dropped) {
			e.preventDefault();
		}
		$element.removeEventListener('remove', documentPointerUp);
		$element.classList.remove('dragging-source');
		document.removeEventListener('pointermove', documentPointerMove, false);
		document.removeEventListener('pointerup', documentPointerUp, false);
		if (dragTags.length) {
			if (dropped) {
				for (const $drag of dragTags) {
					$drag.remove();
				}
			} else {
				e.stopPropagation();
				for (const $drag of dragTags.filter($drag => !$drag.hidden)) {
					const returnAnimation = $drag.animate([
					{
						left: $drag.originalLeft,
						top: $drag.originalTop
					}], {
						duration: 500
					});
					returnAnimation.onfinish = () => $drag.remove();
				}
			}
			dragTags.length = 0;
			for (const $activeDrop of activeDropTags) {
				$activeDrop.classList.remove('active');
				$activeDrop.removeEventListener('pointerup', $activeDrop.drop, true);
				delete $activeDrop.drop;
				$activeDrop.removeEventListener('pointerover', $activeDrop.pointerover, true);
				delete $activeDrop.pointerover;
				$activeDrop.removeEventListener('pointerout', $activeDrop.pointerout, true);
				delete $activeDrop.pointerout;
			}
			activeDropTags.length = 0;
			stop.call($element, e, startX, startY, startOffsetX, startOffsetY, e.pageX - lastX, e.pageY - lastY, dragSelection);
		} else {
			stop.call($element, e, startX, startY, startOffsetX, startOffsetY, e.pageX - lastX, e.pageY - lastY, dragSelection);
		}
	};
	
	if ($handle) {
		$handle.addEventListener('pointerdown', async function(e) {
			if (this.closest('.disabled')) {
				return;
			}
			// TODO: Test this code. I have no idea what it does anymore.
			const $scroll = scrollParent ? (scrollParent instanceof Element ? scrollParent : $element.closest(scrollParent)) : $element;
			const scrollOffset = $.offset($scroll);
			if (e.hasOwnProperty('scrollTagLeft') && e.scrollTagLeft != scrollOffset.left || e.hasOwnProperty('scrollTagTop') && e.scrollTagTop != scrollOffset.top) {
				return;
			}
			lastX = e.pageX;
			lastY = e.pageY;
			startX = e.pageX;
			startY = e.pageY;
			const elementOffset = $.offset($element);
			startOffsetX = e.pageX - elementOffset.left;
			startOffsetY = e.pageY - elementOffset.top;
			dragSelection = await start.call($element, e, startX, startY, startOffsetX, startOffsetY);
			if (dragSelection !== false) {
				e.preventDefault();
				document.addEventListener('pointermove', documentPointerMove, false);
				document.addEventListener('pointerup', documentPointerUp, false);
				$element.addEventListener('remove', documentPointerUp);
			}
		}, true);
	}
}

function droppable($element, {
		accept = [],
		drop = () => {},
		region = '' /* Allows the helper to be specialized when hovering over a droppable */ }) {
	for (const className of accept.map(className => className.trim().replace('.', ''))) {
		$element.classList.add(`droppable-${className}`);
		$element[`droppable-${className}`] = drop;
		$element[`droppableregion-${className}`] = region;
	}
}

function movableConstrainInWindow($element, elementOffset = $.offset($element)) {
	$element.style.left = 0;
	$element.style.top = 0;
	$element.style.left = `${Math.min(Math.floor(document.documentElement.clientWidth - $element.offsetWidth - 1), Math.max(0, elementOffset.left))}px`;
	$element.style.top = `${Math.min(Math.floor(document.documentElement.clientHeight - $element.offsetHeight - 1), Math.max(0, elementOffset.top))}px`;
}

function movable($element, { handle = null, move = () => {} } = {}) {
	const $handle = handle instanceof HTMLElement ? handle : (handle ? $element.querySelector(handle) : $element);
	
	let startOffsetX;
	let startOffsetY;
	
	$handle.classList.add('movable');
	
	const documentPointerMove = e => {
		const left = Math.min(Math.floor(document.documentElement.clientWidth - $element.offsetWidth), Math.max(0, e.pageX - startOffsetX));
		const top = Math.min(Math.floor(document.documentElement.clientHeight - $element.offsetHeight), Math.max(0, e.pageY - startOffsetY));
		$element.style.left = `${left}px`;
		$element.style.top = `${top}px`;
		move(left, top);
	};
	
	const documentPointerUp = e => {
		$element.removeEventListener('remove', documentPointerUp);
		document.removeEventListener('pointermove', documentPointerMove, false);
		document.removeEventListener('pointerup', documentPointerUp, false);
	};
	
	$handle.addEventListener('pointerdown', e => {
		const elementOffset = $.offset($element);
		startOffsetX = e.pageX - elementOffset.left;
		startOffsetY = e.pageY - elementOffset.top;
		document.addEventListener('pointermove', documentPointerMove, false);
		document.addEventListener('pointerup', documentPointerUp, false);
	}, false);
	
	window.addEventListener('resize', () => {
		movableConstrainInWindow($element);
	});
	
	$element.addEventListener('remove', documentPointerUp);
}

function toggleRow($element, toggle, setHidden = ($element, hidden) => $element.hidden = hidden) {
	if ($element.tagName == 'UI-INPUTSELECT') {
		for (const $e of [$element, $element.previousElementSibling]) {
			setHidden($e, !toggle);
		}
	} else if ($element.parentNode.tagName == 'SPAN') {
		for (const $e of [$element.parentNode, $element.parentNode.previousElementSibling]) {
			setHidden($e, !toggle);
		}
	} else {
		for (const $e of [$element, $element.previousElementSibling]) {
			setHidden($e, !toggle);
		}
	}
}

function ContentNodeToMinimalImageContent(node) {
	return {
		name: node.name,
		contentid: node.contentid,
		type: node.type,
		cdnurl: node.cdnurl,
		width: node.hasOwnProperty('width') ? node.width : 0,
		height: node.hasOwnProperty('height') ? node.height : 0,
		uploadedtimestamp: node.uploadedtimestamp,
		filesize: node.filesize,
		md5: node.hash
	};
};

function DragLeftResize(selector, storageName, resize) {
	draggable(document.querySelector(selector), {
		scrollParent: 'body',
		drag: function(e, startX, startY, startOffsetX, startOffsetY) {
			const $previous = this.previousElementSibling;
			const thisPrevOffset = $.offset($previous);
			const width = Math.ceil(e.pageX - startOffsetX - thisPrevOffset.left);
			$previous.style.width = `${width}px`;
			localStorage[storageName] = $previous.offsetWidth;
			if (resize) {
				resize();
			}
		}
	});
	const $previous = document.querySelector(selector).previousElementSibling;
	if (!localStorage.hasOwnProperty(storageName)) {
		localStorage[storageName] = $previous.offsetWidth;
	} else {
		$previous.style.width = `${localStorage[storageName]}px`;
	}
}

window.addEventListener('load', () => {
	// Triggers a 'remove' event on all DOM nodes. This event must be registered with:
	// element.addEventListener('remove', () => {});
	new MutationObserver(mutationsList => {
		for (const mutation of mutationsList) {
			for (const node of mutation.removedNodes) {
				// Moving elements in the dom fires remove events. Do not dispatch a remove event for those by checking if the element is in the dom when this runs
				if (!document.body.contains(node)) {
					node.dispatchEvent(new Event('remove'));
					if (node.nodeType != 3) { // Ignore text nodes
						for (const descendent of node.querySelectorAll('*')) {
							descendent.dispatchEvent(new Event('remove'));
						}
					}
				}
			}
		}
	}).observe(document.body, { childList: true, subtree: true });
	
	const registerCustomTooltip = $node => {
		if ($node.hasAttribute('title') && $node.title != '') {
			customtooltip($node);
		}
	};
	const observeTitleChange = $node => {
		new MutationObserver(mutationsList => {
			for (const mutation of mutationsList) {
				registerCustomTooltip(mutation.target);
			}
		}).observe($node, { subtree: true, attributeFilter: ['title'] });
	};
	const titleObserverShadowRoot = $node => {
		if ($node.shadowRoot && !$node.hasOwnProperty('titleObserved')) {
			$node.titleObserved = true;
			titleObserver($node.shadowRoot);
		}
	};
	const titleObserver = $element => {
		observeTitleChange($element);
		new MutationObserver(mutationsList => {
			for (const mutation of mutationsList) {
				for (const $node of mutation.addedNodes) {
					if ($node.nodeType == 1) {
						registerCustomTooltip($node);
						titleObserverShadowRoot($node);
						for (const $child of $node.querySelectorAll('*')) {
							registerCustomTooltip($child);
							titleObserverShadowRoot($child);
						}
					}
				}
			}
		}).observe($element, { childList: true, subtree: true });
		for (const $node of $element.querySelectorAll('*')) {
			registerCustomTooltip($node);
			titleObserverShadowRoot($node);
		}
	};
	titleObserver(document.body);
	
	document.addEventListener('keydown', e => {
		if (e.key == 'Delete' && !e.target.matches('.name, input, textarea, ui-inputselect, *[contenteditable], ui-color, ui-colorpicker')) {
			e.preventDefault();
		}
	});
});

// There can be only one networking error dialog. Certain code paths might try to generate more
var gNetworkingError = false;
async function NetworkingError() {
	if (!gNetworkingError) {
		gNetworkingError = true;
		await ConfirmationBox('Networking error', 'There was a problem communicating with the server. Wait a moment or refresh.');
		gNetworkingError = false;
	}
}

async function CheckSessionExpired(data) {
	if (!data.success && data.error == 'Session expired.') {
		await ConfirmationBox('Session expired', 'Your session expired. You must relogin.');
		window.location.replace('login.php')
	}
}

function humanFileSize(bytes, si) {
	var thresh = si ? 1000 : 1024;
	if (Math.abs(bytes) < thresh) {
		return bytes + ' B';
	}
	var units = si
		? ['kB','MB','GB','TB','PB','EB','ZB','YB']
		: ['KiB','MiB','GiB','TiB','PiB','EiB','ZiB','YiB'];
	var u = -1;
	do {
		bytes /= thresh;
		++u;
	} while (Math.abs(bytes) >= thresh && u < units.length - 1);
	return bytes.toFixed(1)+' '+units[u];
}

function HSVtoRGB(h, s, v) {
	var r, g, b, i, f, p, q, t;
	i = Math.floor(h * 6);
	f = h * 6 - i;
	p = v * (1 - s);
	q = v * (1 - f * s);
	t = v * (1 - (1 - f) * s);
	switch (i % 6)
	{
		case 0: r = v, g = t, b = p; break;
		case 1: r = q, g = v, b = p; break;
		case 2: r = p, g = v, b = t; break;
		case 3: r = p, g = q, b = v; break;
		case 4: r = t, g = p, b = v; break;
		case 5: r = v, g = p, b = q; break;
	}
	return {
		r: Math.round(r * 255),
		g: Math.round(g * 255),
		b: Math.round(b * 255)
	};
}

function TreeOutOfDate(treeName) {
	switch (treeName) {
		case 'content':
			treeName = 'Content';
			break;
		case 'csv':
			treeName = 'CSV';
			break;
		case 'customwebpage':
			treeName = 'Custom Webpage';
			break;
		case 'event':
			treeName = 'Event';
			break;
		case 'networkcommand':
			treeName = 'Network Command';
			break;
		case 'networkcommandschedule':
			treeName = 'Network Command Schedule';
			break;
		case 'player':
			treeName = 'Player';
			break;
		case 'playlist':
			treeName = 'Playlist';
			break;
		case 'rss':
			treeName = 'RSS';
			break;
		case 'schedule':
			treeName = 'Schedule';
			break;
		case 'user':
			treeName = 'User';
			break;
	}
	ConfirmationBox(`${treeName} tree out of date`, `Another user has altered the ${treeName} tree or an item existing within the ${treeName} tree and your action could not be completed. The ${treeName} tree has now been updated, please try again.`);
}

function baseName(path)
{
	return path.split(/[\\/]/).pop();
}

function LockBody() {
	document.body.classList.add('busy');
}

function UnlockBody() {
	document.body.classList.remove('busy');
}

function GetContentIcon(node) {
	if (node.error) {
		return 'error';
	}
	if (node.hasOwnProperty('file')) {
		return 'upload';
	}
	if (!node.done) {
		return 'pending';
	}
	switch (node.type) {
		case MEDIATYPE_IMAGE:
			return 'image';
		case MEDIATYPE_VIDEO:
			return 'video';
		case MEDIATYPE_AUDIO:
			return 'audio';
		case MEDIATYPE_MISC:
			return 'misc';
		case MEDIATYPE_PDF:
			return '';
	}
	return '';
}

function PlaylistTreeThumbnail(node, largeImages, overrideWidth, overrideHeight) {
	var thumbnailWidth = node.data.pagewidth;
	var thumbnailHeight = node.data.pageheight;
	var scale = 1;
	var iconSize = largeImages ? 150 : 90;
	if (Math.max(thumbnailWidth, thumbnailHeight) > iconSize) {
		scale = iconSize / Math.max(thumbnailWidth, thumbnailHeight);
	}
	thumbnailWidth *= scale;
	thumbnailHeight *= scale;
	var linearGradient = '';
	if (overrideWidth && overrideHeight) {
		if (thumbnailWidth < overrideWidth) {
			var spacing = (overrideWidth - thumbnailWidth) / 2;
			linearGradient = ', linear-gradient(90deg, var(--playlist-gradient-start-color), var(--playlist-gradient-end-color) ' + (spacing - 1) + 'px, #ccc ' + (spacing - 1) + 'px, #ccc ' + (thumbnailWidth + spacing + 1) + 'px, var(--playlist-gradient-end-color) ' + (thumbnailWidth + spacing + 1) + 'px, var(--playlist-gradient-start-color))';
		} else {
			var spacing = (overrideHeight - thumbnailHeight) / 2;
			linearGradient = ', linear-gradient(0deg, var(--playlist-gradient-start-color), var(--playlist-gradient-end-color) ' + (spacing - 1) + 'px, #ccc ' + (spacing - 1) + 'px, #ccc ' + (thumbnailHeight + spacing + 1) + 'px, var(--playlist-gradient-end-color) ' + (thumbnailHeight + spacing + 1) + 'px, var(--playlist-gradient-start-color))';
		}
	}
	var loadImageAttempt = 0;
	return $.div({ class: ['handle', node.disabled, 'disabled'], tabindex: -1 },
		// Temporary img tag to load the image or spin until loaded
		$.img({ src: `thumbnails/${account}/${node.data.playlistid}.jpg?random=${Math.floor(Math.random() * 100000)}`, css: {
			height: 0,
			opacity: 0
		}, load: function() {
			this.parentNode.querySelector(':scope > .image').style.backgroundImage = `url(${encodeURI(`thumbnails/${account}/${node.data.playlistid}.jpg?random=${Math.floor(Math.random() * 100000)}`)})${linearGradient}`;
			this.parentNode.querySelector(':scope > .image').classList.remove('busy');
			this.remove();
		}, error: function() {
			loadImageAttempt++;
			this.parentNode.querySelector(':scope > .image').classList.add('busy');
			if (loadImageAttempt < 3) {
				setTimeout(() => {
					this.src = `thumbnails/${account}/${node.data.playlistid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
				}, 30 * 1000);
			}
		} }),
		$.div({ alt: node.name, class: 'image', css: {
			backgroundSize: `${thumbnailWidth}px ${thumbnailHeight}px, 100% 100%`,
			width: `${overrideWidth ? overrideWidth : thumbnailWidth}px`,
			height: `${overrideHeight ? overrideHeight : thumbnailHeight}px`
		} }),
		node.$name = $.span({ text: node.name, class: 'name', spellcheck: false, css: {
			width: `${overrideWidth ? overrideWidth : thumbnailWidth}px`
		} })
	);
}

function PlaylistTreeThumbnailThumbnailDimensions(nodes, largeImages)
{
	var maximumWidth = 0;
	var maximumHeight = 0;
	for (const node of nodes)
	{
		var thumbnailWidth = node.data.pagewidth;
		var thumbnailHeight = node.data.pageheight;
		var scale = 1;
		var iconSize = largeImages ? 150 : 90;
		if (Math.max(thumbnailWidth, thumbnailHeight) > iconSize)
		{
			scale = iconSize / Math.max(thumbnailWidth, thumbnailHeight);
		}
		thumbnailWidth *= scale;
		thumbnailHeight *= scale;
		maximumWidth = Math.max(maximumWidth, thumbnailWidth);
		maximumHeight = Math.max(maximumHeight, thumbnailHeight);
	}
	return { width: maximumWidth, height: maximumHeight };
}

const metadataDataTypes = [
	{ name: 'Text', value: 'string', default: '' },
	{ name: 'Multiline Text', value: 'string', design: 'multiline', default: '' },
	{ name: 'True/False', value: 'bool', default: false },
	{ name: 'Integer', value: 'int', default: 0 },
	{ name: 'Number', value: 'float', default: 0 },
	{ name: 'Color', value: 'color', default: 'rgba(255, 255, 255, 255)' },
	{ name: 'Crop Rectangle', value: 'string', design: 'croprectangle', default: '0,0,50,50' }
];

function Metadata(tag, metadata, parentMetadata = [], saveHandler = null, associatedData = {}) {
	// Save is bounced meaning it can be called as much as one wants and won't be executed more frequently than 200 ms and won't overlap saves and will always save the last changes
	let bounceTimer = null;
	let saving = false;
	let saveAgain = false;
	let metadataToSave;
	const Save = () => {
		if (saveHandler) {
			metadataToSave = JSON.parse(JSON.stringify(metadata));
			const save = () => {
				bounceTimer = setTimeout(async () => {
					bounceTimer = null;
					saving = true;
					await saveHandler(metadataToSave);
					saving = false;
					if (saveAgain) {
						saveAgain = false;
						save();
					}
				}, 500);
			};
			if (saving) {
				saveAgain = true;
			} else {
				if (bounceTimer != null) {
					clearTimeout(bounceTimer);
				}
				save();
			}
		}
	};
	
	const refresh = () => {
		Metadata(tag, metadata, parentMetadata, saveHandler, associatedData);
		Save();
	};
	
	const addProperty = property => {
		const nameReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('name') || parentMetadata.find(parentProperty => parentProperty.name == property.name) != undefined;
		const typeReadonly = parentMetadata.find(parentProperty => parentProperty.name == property.name) != undefined || property.hasOwnProperty('readonly') && property.readonly.includes('type');
		const setReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('set');
		const valueReadonly = !property.set || property.hasOwnProperty('readonly') && property.readonly.includes('value');
		const optionsReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('options');
		const moveUpReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('moveup');
		const moveDownReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('movedown');
		const addReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('add');
		const deleteReadonly = property.hasOwnProperty('readonly') && property.readonly.includes('delete');
		const index = metadata.indexOf(property);
		const nameAlreadyExists = metadata.find((previousProperty, previousIndex) => previousIndex < index && previousProperty.name == property.name) != null;
		tag.append(
			$.input({ type: 'text', class: [nameAlreadyExists, 'error'], value: property.name, readonly: nameReadonly, input: function() {
				property.name = this.value;
				this.classList.toggle('error', metadata.find((previousProperty, previousIndex) => previousIndex < index && previousProperty.name == property.name) != null);
				Save();
			} }),
			$.select({ type: 'text', title: metadataDataTypes.find(type => type.value == property.type).name, value: property.hasOwnProperty('design') ? `${property.type}+${property.design}` : property.type, disabled: typeReadonly, change: function() {
				if (this.value.indexOf('+') != -1) {
					const [value, design] = this.value.split('+');
					property.type = value;
					property.design = design;
				} else {
					property.type = this.value;
					delete property.design;
				}
				refresh();
			} },
				...metadataDataTypes.map(option => $.option({ text: option.name, value: option.hasOwnProperty('design') ? `${option.value}+${option.design}` : option.value }))
			),
			$.uicheckbox({ title: 'Set', checked: property.set, change: function(e) {
				e.stopPropagation();
				property.set = this.checked;
				if (this.checked) {
					property.value = property.hasOwnProperty('default') ? property.default : metadataDataTypes.find(type => type.value == property.type).default;
				} else {
					property.value = null;
				}
				refresh();
			} }),
			(() => {
				if (property.value == null) {
					return $.input({ type: 'text', readonly: true });
				}
				switch (property.type) {
					case 'string':
						switch (property.design) {
							case 'multiline':
								return $.textarea({ rows: 2, title: 'Double click to expand editor.', value: property.value, spellcheck: false, readonly: valueReadonly, dblclick: function(e) {
									const $textArea = this;
									e.stopPropagation();
									const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '750px' } },
										$.header({ text: 'Multiline Text' }),
										$.textarea({ id: 'metadata-multilinetext', rows: 10, value: property.value }),
										$.div({ class: 'panel-buttons' },
											$.button({ text: 'Cancel', click: function() {
												$dialog.remove();
											} }),
											$.button({ text: 'OK', click: function() {
												const value = document.getElementById('metadata-multilinetext').value;
												property.value = value;
												$textArea.value = property.value;
												Save();
												$dialog.remove();
											} })
										)
									));
									centerOnScreen($dialog);
								}, input: function(e) {
									e.stopPropagation();
									property.value = this.value;
									Save();
								} });
							case 'croprectangle':
								return $.input({
									type: 'text',
									placeholder: "left,top,width,height",
									title: 'Format: "offset left,offset top,width,height". For example, "0,10,20,30" would mean 0 pixels from the left and 10 pixels down with a width and height of 20 and 30 respectively. Double click to draw crop rectangle.',
									value: property.value,
									readonly: valueReadonly,
									keydown: e => {
										// Only allow numbers and comma
										if (/^[^0-9,]$/.test(e.key)) {
											e.preventDefault();
										}
									}, dblclick: function(e) {
										const $cropRectangle = this;
										let cropValue = $cropRectangle.value;
										if (!/^([0-9]{1,4},){3}[0-9]{1,4}$/.test(cropValue)) {
											cropValue = `0,0,${associatedData.width},${associatedData.height}`;
										}
										e.stopPropagation();
										// Only images are supported
										if (!associatedData.hasOwnProperty('contenttreetype') || associatedData.contenttreetype != 3) {
											ConfirmationBox('Crop Dialog Unavailable', 'The crop editor is only available for metadata on images.');
											return;
										}
										let $imageCropBox;
										const $dialog = document.body.appendChild($.uidialog({ config: { modal: true } },
											$.header({ text: `Crop Rectangle for ${associatedData.name}` }),
											$.div({ css: {
												display: 'inline-block',
												background: "url('ui/images/transparent.png') white",
												backgroundSize: '20px 20px',
												position: 'relative'
											} },
												$.img({ src: GetCachedMediaURL(account, associatedData.cdnurl, associatedData.uploadedtimestamp, associatedData.filesize), css: {
													display: 'block',
													maxHeight: '500px',
													pointerEvents: 'none'
												}, load: function() {
													centerOnScreen($dialog);
													const $imageBox = document.getElementById('image-cropbox');
													$imageBox.width = this.width;
													$imageBox.height = this.height;
													renderCrop();
												} }),
												$imageCropBox = $.canvas({ id: 'image-cropbox', css: {
													position: 'absolute',
													left: 0,
													top: 0
												} })
											),
											$.div({ class: 'panel-buttons' },
												$.button({ text: 'Cancel', click: function() {
													$dialog.remove();
												} }),
												$.button({ text: 'OK', click: function() {
													property.value = cropValue;
													$cropRectangle.value = property.value;
													Save();
													$dialog.remove();
												} })
											)
										));
										centerOnScreen($dialog);
							
										var startCropX;
										var startCropY;
										var renderCrop = () => {
											const [cropX, cropY, cropWidth, cropHeight] = cropValue.split(',').map(value => parseInt(value, 10));
											let scale = associatedData.width / $imageCropBox.width;
											var canvas = $imageCropBox;
											var context = canvas.getContext('2d');
											context.clearRect(0, 0, canvas.width, canvas.height);
											context.save();
											context.scale(1 / scale, 1 / scale);
											context.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-fill-color');
											context.fillRect(cropX, cropY, cropWidth, cropHeight);
											context.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-stroke-color');
											context.translate(0.5, 0.5);
											context.strokeRect(cropX, cropY, cropWidth, cropHeight);
											context.restore();
										};
										draggable($imageCropBox.parentNode, {
											handle: '#image-cropbox',
											scrollParent: $imageCropBox.parentNode,
											start: (e, startX, startY, startOffsetX, startOffsetY) => {
												let scale = associatedData.width / $imageCropBox.width;
												startCropX = Math.floor(startOffsetX + $imageCropBox.parentNode.scrollLeft);
												startCropY = Math.floor(startOffsetY + $imageCropBox.parentNode.scrollTop);
												cropValue = `${Math.round(startCropX * scale)},${Math.round(startCropY * scale)},${1},${1}`;
												renderCrop();
											},
											drag: (e, startX, startY, startOffsetX, startOffsetY) => {
												const $imageCropBox = document.getElementById('image-cropbox');
												const imageCropBoxOffset = $.offset($imageCropBox);
												let relativeX = Math.max(0, Math.floor(e.pageX - imageCropBoxOffset.left + $imageCropBox.parentNode.scrollLeft));
												let relativeY = Math.max(0, Math.floor(e.pageY - imageCropBoxOffset.top + $imageCropBox.parentNode.scrollTop));
												let scale = associatedData.width / $imageCropBox.width;
												let cropX = Math.round(Math.min(startCropX, relativeX) * scale);
												let cropY = Math.round(Math.min(startCropY, relativeY) * scale);
												let cropWidth = Math.round(Math.max(1, Math.abs(relativeX - startCropX)) * scale);
												let cropHeight = Math.round(Math.max(1, Math.abs(relativeY - startCropY)) * scale);
												if (cropX + cropWidth > associatedData.width) {
													cropWidth = associatedData.width - cropX;
												}
												if (cropY + cropHeight > associatedData.height) {
													cropHeight = associatedData.height - cropY;
												}
												cropValue = `${cropX},${cropY},${cropWidth},${cropHeight}`;
												renderCrop();
											}
										});
									}, input: function(e) {
										e.stopPropagation();
										let newValue = this.value;
										if (/^([0-9]{1,4},){3}[0-9]{1,4}$/.test(newValue)) {
											this.classList.remove('error');
											property.value = newValue;
											Save();
										} else {
											this.classList.add('error');
										}
									}
								});
							default:
								return $.input({ type: 'text', value: property.value, spellcheck: false, readonly: valueReadonly, input: function(e) {
									e.stopPropagation();
									property.value = this.value;
									Save();
								} });
						}
					case 'bool':
						return $.uicheckbox({ readonly: valueReadonly, checked: property.value, change: function(e) {
							e.stopPropagation();
							property.value = this.checked;
							refresh();
						} });
					case 'int':
						return $.input({ type: 'text', placeholder: '123', value: property.value, readonly: valueReadonly, keydown: onlyInteger, input: function(e) {
							e.stopPropagation();
							property.value = this.value;
							Save();
						} });
					case 'float':
						return $.input({ type: 'text', placeholder: '1.23', value: property.value, readonly: valueReadonly, input: function(e) {
							e.stopPropagation();
							property.value = this.value;
							Save();
						} });
					case 'color':
						return $.uicolor({ type: 'text', readonly: valueReadonly, value: property.value, config: {
							compact: true,
							opacity: true,
							previewChange: function() { this.dispatchEvent(new Event('change')); },
							showPalette: true,
							localStorageKey: 'globalcolor'
						}, change: function(e) {
							e.stopPropagation();
							property.value = this.value;
							Save();
						} });
					case 'media':
						return $.input({ type: 'text', readonly: valueReadonly });
					case 'image':
						return $.input({ type: 'text', readonly: valueReadonly });
					case 'video':
						return $.input({ type: 'text', readonly: valueReadonly });
					case 'audio':
						return $.input({ type: 'text', readonly: valueReadonly });
					case 'pdf':
						return $.input({ type: 'text', readonly: valueReadonly });
				}
			})(),
			$.ul({ class: [optionsReadonly, 'disabled'], draggable: {
				handle: '.reorder',
				thresholdY: 3,
				sortable: 'y',
				scrollParent: tag,
				start: (e, startX, startY, startOffsetX, startOffsetY) => {
					if (DialogWindowOpen()) {
						BlinkDialog();
						return false;
					}
				},
				movePrevious: function() {
					if (!moveUpReadonly) {
						const moveProperty = metadata.splice(index, 1)[0];
						metadata.splice(index - 1, 0, moveProperty);
						refresh();
					}
					return false; // Stop the automatic tag movement by the draggable extension. We handled it manually
				},
				moveNext: function() {
					if (!moveDownReadonly) {
						const moveProperty = metadata.splice(index, 1)[0];
						metadata.splice(index + 1, 0, moveProperty);
						refresh();
					}
					return false; // Stop the automatic tag movement by the draggable extension. We handled it manually
				}
			} },
				$.li({ title: 'Drag Up/Down', class: ['tooltip-button mdi mdi-arrow-expand mdi-rotate-135 reorder', moveUpReadonly || moveDownReadonly, 'disabled'] }),
				$.li({ title: 'Insert Property Below', class: ['tooltip-button mdi mdi-plus-circle-outline', addReadonly, 'disabled'], click: () => {
					metadata.splice(index + 1, 0, { name: '', type: 'string', set: true, value: '' });
					refresh();
				} }),
				$.li({ title: 'Clear/Delete Property', class: ['tooltip-button mdi mdi-trash-can-outline', deleteReadonly, 'disabled'], click: () => {
					if (parentMetadata.find(parentProperty => parentProperty.name == property.name) == undefined) {
						metadata.splice(index, 1);
					} else {
						property.set = false;
						property.value = property.hasOwnProperty('default') ? property.default : metadataDataTypes.find(type => type.value == property.type).default;
					}
					refresh();
				} })
			)
		);
	};
	tag.innerHTML = '';
	tag.classList.add('metadatagrid');
	tag.append(
		$.span({ text: 'Name', title: 'A unique name that identifies this property', class: 'tooltip-help' }),
		$.span({ text: 'Type', title: 'The type of information this property stores', class: 'tooltip-help' }),
		$.span({ title: 'Set the value. If a parent set a value, then checking this will allow you to override that value.', class: 'tooltip-help mdi mdi-check-box-outline' }),
		$.span({ text: 'Value', title: 'The value assigned to this property. If this is a folder then children in the subtree can override this property in their own metadata', class: 'tooltip-help' }),
		$.span({ text: 'Options', title: 'The order of metadata is just aesthetic. If a parent defines a metadata property then it can only be cleared in children. To delete the property you must delete the property in the parent folder.', class: 'tooltip-help' })
	);
	for (const property of metadata) {
		addProperty(property);
	}
	if (metadata.length == 0) {
		metadata.push({ name: '', type: 'string', set: true, value: '' });
		refresh();
	}
}

function MetadataFilters(tag, filters) {
	const refresh = () => {
		MetadataFilters(tag, filters);
	};
	const buildFilterTree = (appendTag, node, depth = 0, parent = null) => {
		const isRoot = depth == 0;
		const groupNode = node.type == 'and' || node.type == 'or';
		const $div = $.div();
		if (!isRoot) {
			$div.appendChild($.i({ class: 'mdi mdi-cursor-move handle' }));
		}
		const index = parent ? parent.children.indexOf(node) : 0;
		switch (node.type) {
			case 'and':
			case 'or':
				$select = $div.appendChild($.select({ value: node.type, change: function() {
					node.type = this.value;
					refresh();
				} },
					$.option({ text: 'All child conditions must be satisfied (And)', value: 'and' }),
					$.option({ text: 'Any child condition can be satisfied (Or)', value: 'or' })
				));
				if (!isRoot) {
					$div.append(
						$.span({ css: { flex: '1 0 0px' } }),
						$.i({ title: 'Remove', class: 'tooltip-button mdi mdi-close', click: () => {
							parent.children.splice(index, 1);
							refresh();
						} })
					);
				}
				break;
			default:
				$div.append(
					$.input({ type: 'text', placeholder: 'Name', class: 'name', title: 'Metadata Name', value: node.name, input: function() {
						node.name = this.value;
					} }),
					$.select({ type: 'text', class: 'operator', title: 'Comparison Operator', value: node.operator, change: function() {
						node.operator = this.value;
					} },
						$.option({ text: 'Equal', value: '=' }),
						$.option({ text: 'Not Equal', value: '!=' }),
						$.option({ text: 'Less Than', value: '<', disabled: node.type != 'int' && node.type != 'float' }),
						$.option({ text: 'Less Than or Equal', value: '<=', disabled: node.type != 'int' && node.type != 'float' }),
						$.option({ text: 'Greater Than', value: '>', disabled: node.type != 'int' && node.type != 'float' }),
						$.option({ text: 'Greater Than Or Equal', value: '>=', disabled: node.type != 'int' && node.type != 'float' })
					),
					$.select({ type: 'text', title: 'Type. Note: Types like multiline text and cropping rectangle are really text.', class: 'type', value: node.type, change: function() {
						node.type = this.value;
						refresh();
					} },
						metadataDataTypes.map(option => $.option({ text: option.name, value: option.value }))
					),
					$.input({ type: 'text', placeholder: 'Value', class: 'value', title: 'Value', value: node.value, input: function() {
						node.value = this.value;
					} }),
					$.i({ title: 'Use metadata value. Enabling this turns the value to the left into the default value. If no metadata is found on this widget matching the name then the default value is used.', class: 'tooltip-button mdi mdi-code-brackets', click: () => {
						if (node.hasOwnProperty('metadata')) {
							delete node.metadata;
						} else {
							node.metadata = '';
						}
						refresh();
					} }),
					$.i({ title: 'Remove', class: 'tooltip-button mdi mdi-close', click: () => {
						parent.children.splice(index, 1);
						refresh();
					} }),
					$.input({ type: 'text', placeholder: 'Metadata Name', class: 'metadata', value: node.metadata, hidden: !node.hasOwnProperty('metadata'), input: function() {
						node.metadata = this.value;
					} })
				)
				if (node.hasOwnProperty('metadata')) {
					$div.style.gridTemplateRows = '26px 26px';
				}
				break;
		}
		const $ul = $.ul();
		if (node.hasOwnProperty('children')) {
			for (const child of node.children) {
				$children = $ul.appendChild($.li());
				buildFilterTree($children, child, depth + 1, node);
			}
		}
		if (groupNode) {
			$ul.appendChild($.li({ class: 'add' },
				$.span(),
				$.label({ title: 'Add And/Or/Condition', class: 'mdi mdi-plus tooltip-button' },
					$.select({ value: '', change: function() {
						switch (this.value) {
							case 'and':
								node.children.push({ type: 'and', children: [] });
								break;
							case 'or':
								node.children.push({ type: 'or', children: [] });
								break;
							case 'condition':
								node.children.push({ name: '', operator: '=', type: 'string', value: '' });
								break;
						}
						refresh();
					} },
						$.option({ text: 'All child conditions must be satisfied (And)', value: 'and' }),
						$.option({ text: 'Any child condition is satisfied (Or)', value: 'or' }),
						$.option({ text: 'Condition', value: 'condition' })
					)
				)
			));
		}
		appendTag.appendChild($.ul({ class: [
			'metadata-filters-nodeContainer',
			groupNode, 'metadata-filters-groupContainer',
			!groupNode, 'metadata-filters-leafContainer',
			isRoot, 'root'
		] },
			$.li({},
				$.i(),
				$.i()
			),
			$.li({ class: ['metadata-filters-node', node.type == 'and', 'and', node.type == 'or', 'or'], draggable: {
				handle: ':scope > div > .handle',
				helper: async $node => {
					const $drop = $node.cloneNode(true);
					$drop.classList.add('filter-dragging');
					$drop.style.width = `${$node.offsetWidth}px`;
					return $drop;
				},
				appendTo: 'body',
				containment: 'body',
				thresholdX: 3,
				thresholdY: 3
			} },
				$div,
				$ul
			)
		));
	};
	tag.innerHTML = '';
	tag.classList.add('metadata-filters');
	tag.append(
		$.li({ class: 'metadata-filters-operations' },
			$.i({ title: 'Erase filter', class: 'mdi mdi-eraser', click: async () => {
				// Erase all filters confirmation box
				try {
					await ConfirmationBox("Clear filter?", 'Are you sure you want to clear the filter?', 'Yes', 'No');
					filters.type = 'and';
					filters.children.length = 0;
					refresh();
				} catch (exception) {
					// User selected no
				}
			} })
		),
		$.li()
	);
	buildFilterTree(tag.querySelector(':scope > li:last-child'), filters);
}

function CreatePastels(n) {
	var colors = [];
	for (var i = 0; i < n; ++i) {
		var interpolation = (i / n) * 0.65;
		var color = HSVtoRGB(interpolation, 0.3, 0.8);
		colors.push('rgba(' + color.r + ',' + color.g + ',' + color.b + ',1)');
	}
	return colors;
}

function TreeSort(a, b) {
	const aDateMatch = a.match(/^([0-9]{1,2})-([0-9]{1,2})-([0-9]{2,4}).*$/);
	const bDateMatch = b.match(/^([0-9]{1,2})-([0-9]{1,2})-([0-9]{2,4}).*$/);
	// abc.100.mp4 is an example and we want to extract the 100. We use negative lookahead on the last piece to detect if its a mix of numbers and letters and if it is ignore it
	const aMatches = a.match(/^(.*?)([0-9]*)(?:\.(?!\d+).+)?$/);
	const bMatches = b.match(/^(.*?)([0-9]*)(?:\.(?!\d+).+)?$/);
	if (aDateMatch != null && bDateMatch != null && aDateMatch.length == 4 && bDateMatch.length == 4) {
		if (parseInt(aDateMatch[1], 10) != parseInt(bDateMatch[1], 10)) {
			return parseInt(aDateMatch[1], 10) - parseInt(bDateMatch[1], 10);
		} else if (parseInt(aDateMatch[2], 10) != parseInt(bDateMatch[2], 10)) {
			return parseInt(aDateMatch[2], 10) - parseInt(bDateMatch[2], 10);
		} else {
			return parseInt(aDateMatch[3], 10) - parseInt(bDateMatch[3], 10);
		}
	} else if (aMatches != null && bMatches != null && aMatches.length == 3 && bMatches.length == 3 && aMatches[1] == bMatches[1]) {
		return parseInt(aMatches[2], 10) - parseInt(bMatches[2], 10);
	} else {
		return a.localeCompare(b);
	}
}

function SetupCalendarTooltips() {
	document.querySelector('.fc-dayGridMonth-button').title = 'View by month';
	document.querySelector('.fc-timeGridWeek-button').title = 'View by week';
	document.querySelector('.fc-timeGridDay-button').title = 'View by day';
	
	document.querySelector('.fc-prev-button').title = 'View previous';
	document.querySelector('.fc-next-button').title = 'View next';
	
	document.querySelector('.fc-today-button').title = 'Return to the current date in calendar view';
}

// TODO: move textchanged to the custom widgets section
class $ {
	static $(tagName, {
			id = null,
			class:cls = '',
			css = {},
			html = null,
			text = null,
			hidden = false,
			click = null,
			dblclick = null,
			pointerdown = null,
			pointerup = null,
			pointermove = null,
			pointerenter = null,
			pointerleave = null,
			pointerover = null,
			pointerout = null,
			keydown = null,
			keyup = null,
			change = null,
			input = null,
			paste = null,
			blur = null,
			focus = null,
			scroll = null,
			load = null,
			error = null,
			loadedmetadata = null,
			textchanged = null,
			multiclick:multiclickHandler = null,
			draggable:draggableOptions = null,
			droppable:droppableOptions = null,
			movable:movableOptions = null,
			tabs:tabsOptions = null,
			process = null,
			...attributes } = {}, ...$children) {
		// This is an explicit list of attributes so we can find errors in our code and typos.
		let title,
			type,
			placeholder,
			src,
			width,
			height,
			alt,
			rel,
			href,
			target,
			min,
			max,
			step,
			rows,
			tabindex,
			name,
			checked,
			indeterminate,
			selected,
			disabled,
			allow,
			span,
			data,
			size,
			maxlength,
			spellcheck,
			sandbox,
			readonly,
			contenteditable,
			crossOrigin,
			multiple,
			other;
		const validAttributes = {
			title,
			type,
			placeholder,
			src,
			width,
			height,
			alt,
			rel,
			href,
			target,
			min,
			max,
			step,
			rows,
			tabindex,
			name,
			checked,
			indeterminate,
			selected,
			disabled,
			allow,
			span,
			data,
			size,
			maxlength,
			spellcheck,
			sandbox,
			readonly,
			contenteditable,
			crossOrigin,
			multiple, ...other } = attributes;
		const $element = document.createElement(tagName);
		for (const [key, value] of Object.entries(validAttributes)) {
			if (value !== false) {
				$element.setAttribute(key, value);
			}
		}
		if (other.hasOwnProperty('for')) {
			$element.setAttribute('for', other.for);
			delete other.for;
		}
		if (Object.keys(other).length != 0){
			console.trace();
			console.log('Unsupported: ', other);
		}
		
		if (id != null) {
			$element.id = id;
		}
		if (typeof cls == 'string') {
			if (cls != '') {
				$element.classList.add(...cls.split(' '));
			}
		} else if (Array.isArray(cls)) {
			// This allows a format with toggleable classes [false, 'foo', true, 'bar']
			for (let i = 0; i < cls.length; ++i) {
				if (typeof cls[i] == 'boolean') {
					if (cls[i]) {
						$element.classList.add(...cls[i + 1].split(' '));
					}
					i++;
				} else {
					$element.classList.add(...cls[i].split(' '));
				}
			}
		}
		Object.assign($element.style, css);
		
		$element.hidden = hidden;
		if (html) {
			$element.innerHTML = html;
		} else if (text) {
			$element.innerText = text;
		}
		for (const [eventName, event] of Object.entries({
			click,
			dblclick,
			pointerdown,
			pointerup,
			pointermove,
			pointerenter,
			pointerleave,
			pointerover,
			pointerout,
			keydown,
			keyup,
			change,
			input,
			paste,
			blur,
			focus,
			scroll,
			load,
			error,
			loadedmetadata,
			textchanged })) {
			if (event) {
				if (typeof event == 'object') {
					$element.addEventListener(eventName, event.listener, {
						capture: event.capture,
						once: event.once,
						passive: event.passive,
						signal: event.passive
					});
				} else {
					$element.addEventListener(eventName, event);
				}
			}
		}
		if (multiclickHandler != null) {
			multiclick($element, multiclickHandler);
		}
		$element.append(...$children);
		if (draggableOptions != null) {
			draggable($element, draggableOptions);
		}
		if (droppableOptions != null) {
			droppable($element, droppableOptions);
		}
		if (tabsOptions != null) {
			tabs($element, tabsOptions);
		}
		if (movableOptions != null) {
			movable($element, movableOptions);
		}
		if (process) {
			process.call($element, $element);
		}
		return $element;
	}
	static offset($element) {
		const box = $element.getBoundingClientRect();
		return {
			top: box.top + window.pageYOffset - document.documentElement.clientTop,
			left: box.left + window.pageXOffset - document.documentElement.clientLeft
		};
	}
	// Calculates the offsetWidth + margins
	static outerWidth($element) {
		const computedStyle = getComputedStyle($element);
		return parseFloat(computedStyle.marginLeft) + $element.offsetWidth + parseFloat(computedStyle.marginRight);
	}
	// Calculates the offsetHeight + margins
	static outerHeight($element) {
		const computedStyle = getComputedStyle($element);
		return parseFloat(computedStyle.marginTop) + $element.offsetHeight + parseFloat(computedStyle.marginBottom);
	}
	static prev($element, selector) {
		let $node = $element;
		while ($node && !$node.matches(selector)) {
			$node = $node.previousElementSibling;
		}
		return $node;
	}
	static next($element, selector) {
		let $node = $element;
		while ($node && !$node.matches(selector)) {
			$node = $node.nextElementSibling;
		}
		return $node;
	}
	static attachFadeAnimation($element) {
		if (!$element.hasOwnProperty('fadeAnimation')) {
			const duration = 500;
			$element.fadeAnimation = $element.animate([
				{ opacity: 0 },
				{ opacity: 1 }
			], {
				duration,
				fill: 'both'
			});
			$element.fadeAnimation.currentTime = duration;
			$element.fadeAnimationPromise = null;
			$element.fadeAnimationReject = null;
		}
	}
	static async fadeOut($element) {
		$.attachFadeAnimation($element);
		if ($element.fadeAnimation.playbackRate != -1) {
			$element.fadeAnimation.playbackRate = -1;
			if ($element.fadeAnimationReject) {
				// Cancels the fade in
				$element.fadeAnimationReject();
				$element.fadeAnimationReject = null;
			}
			if ($element.fadeAnimation.playState != 'finished') {
				$element.fadeAnimationPromise = new Promise((resolve, reject) => {
					$element.fadeAnimation.onfinish = resolve
					$element.fadeAnimationReject = reject;
				});
				await $element.fadeAnimationPromise;
				$element.fadeAnimationPromise = null;
				$element.fadeAnimationReject = null;
			}
		} else if ($element.fadeAnimationPromise) {
			await $element.fadeAnimationPromise;
		}
	}
	static async fadeIn($element) {
		$.attachFadeAnimation($element);
		if ($element.fadeAnimation.playbackRate != 1) {
			$element.fadeAnimation.playbackRate = 1;
			if ($element.fadeAnimationReject) {
				// Cancels the fade out
				$element.fadeAnimationReject();
				$element.fadeAnimationReject = null;
			}
			if ($element.fadeAnimation.playState != 'finished') {
				$element.fadeAnimationPromise = new Promise((resolve, reject) => {
					$element.fadeAnimation.onfinish = resolve;
					$element.fadeAnimationReject = reject;
				});
				await $element.fadeAnimationPromise;
				$element.fadeAnimationPromise = null;
				$element.fadeAnimationReject = null;
			}
		} else if ($element.fadeAnimationPromise) {
			await $element.fadeAnimationPromise;
		}
	}
	static http({
		method = '',
		url = '',
		data = {},
		dataType = ''
	}) {
		function formatParams(params) {
			return "?" + Object.keys(params).map(key => {
				return key + `=${encodeURIComponent(params[key])}`
			}).join("&")
		}

		if (method == '') {
			console.error('method required');
			console.trace();
			throw new Error('method required');
		}
		if (url == '') {
			console.error('url required');
			console.trace();
			throw new Error('url required');
		}
		const request = new XMLHttpRequest();
		return {
			xhr: request,
			promise: new Promise((resolve, reject) => {
				if (method == 'GET') { 
					url += formatParams(data);
				}
				request.open(method, url);
				request.setRequestHeader('Cache-Control', 'no-cache, no-store, max-age=0');
				request.addEventListener('load', function() {
					if (this.readyState === XMLHttpRequest.DONE && this.status === 200) {
						switch (dataType) {
							case 'json':
								// json
								if (this.responseType == 'json') {
									resolve(this.response);
								} else {
									try {
										resolve(JSON.parse(this.response));
									} catch (exception) { 
										reject(new Error(`Could not parse response json: ${exception}`));
									}
								}
								break;
							case 'xml':
								// Document
								if (this.responseType == 'document') {
									resolve(this.response);
								} else {
									const parser = new DOMParser();
									const xml = parser.parseFromString(this.response, "application/xml");
									if (xml.documentElement.tagName == 'parsererror') {
										console.warn(xml.documentElement.textContent);
										console.trace();
										reject(new Error(xml.documentElement.textContent));
									} else {
										resolve(xml);
									}
								}
								break;
							case '':
								// String
								if (this.responseType == '' || this.responseType == 'text') {
									resolve(this.response);
								} else {
									console.warn(`expected responseType to be '' or 'text', got ${this.responseType}`);
									console.trace();
									reject(new Error(`expected responseType to be '' or 'text', got ${this.responseType}`));
								}
								break;
							default:
								reject(new Error('Invalid dataType passed to $.http(...)'));
								break;
						}
					} else {
					
					}
				});
				request.addEventListener('error', function() {
					console.warn(`${this.status}: ${url}`);
					console.trace();
					reject(new Error(`${this.status}: ${url}`));
				});
				if (method == 'POST') {
					const formData = new FormData();
					for (const [key, value] of Object.entries(data)) {
						formData.append(key, value);
					}
					request.send(formData);
				} else {
					request.send(null);
				}
			})
		};
	};
}

// Tag list taken from MDN. Don't add hyphenated tags here
for (const tagName of [
	'script',
	'style',
	'link',
	'address',
	'article',
	'aside',
	'footer',
	'header',
	'h1',
	'h2',
	'h3',
	'h4',
	'h5',
	'h6',
	'hgroup',
	'main',
	'nav',
	'section',
	'blockquote',
	'dd',
	'div',
	'dl',
	'dt',
	'figcaption',
	'figure',
	'hr',
	'li',
	'ol',
	'p',
	'pre',
	'ul',
	'a',
	'abbr',
	'b',
	'bdi',
	'bdo',
	'br',
	'cite',
	'code',
	'data',
	'dfn',
	'em',
	'i',
	'kbd',
	'mark',
	'q',
	'rb',
	'rp',
	'rt',
	'rtc',
	'ruby',
	's',
	'samp',
	'small',
	'span',
	'strong',
	'sub',
	'sup',
	'time',
	'u',
	'var',
	'wbr',
	'area',
	'audio',
	'img',
	'map',
	'track',
	'video',
	'embed',
	'iframe',
	'object',
	'param',
	'picture',
	'portal',
	'source',
	'svg',
	'math',
	'canvas',
	'del',
	'ins',
	'caption',
	'col',
	'colgroup',
	'table',
	'tbody',
	'td',
	'tfoot',
	'th',
	'thead',
	'tr',
	'button',
	'datalist',
	'fieldset',
	'form',
	'input',
	'label',
	'legend',
	'meter',
	'optgroup',
	'option',
	'output',
	'progress',
	'select',
	'textarea',
	'details',
	'dialog',
	'menu',
	'summary',
	'slot']) {
	$[tagName] = function(options = {}, ...$children) {
		const { value, ...otherOptions } = options;
		const $element = $.$(tagName, otherOptions, ...$children);
		if (value !== undefined) {
			$element.value = value;
		}
		return $element;
	}
}

for (const tagName of [
	'ui-tree',
	'ui-treeselect',
	'ui-inputselect',
	'ui-slider',
	'ui-color',
	'ui-colorpicker',
	'ui-checkbox',
	'ui-dialog',
	'ui-toast' ]) {
	$[tagName.replace('-', '')] = function(options = {}, ...$children) {
		const { config = {}, process = null, value, ...otherOptions } = options;
		const $element = $.$(tagName, otherOptions, ...$children);
		$element.options(config);
		if (process) {
			process.call($element, $element);
		}
		if (value !== undefined) {
			$element.value = value;
		}
		return $element;
	};
}

function range(size, startAt = 0) {
	return [...Array(size).keys()].map(i => i + startAt);
}

if (!Array.prototype.flat) {
	Object.defineProperty(Array.prototype, 'flat', {
		configurable: true,
		value: function flat () {
			var depth = isNaN(arguments[0]) ? 1 : Number(arguments[0]);

			return depth ? Array.prototype.reduce.call(this, function (acc, cur) {
				if (Array.isArray(cur)) {
					acc.push.apply(acc, flat.call(cur, depth - 1));
				} else {
					acc.push(cur);
				}

				return acc;
			}, []) : Array.prototype.slice.call(this);
		},
		writable: true
	});
}

function CreateNavigationMenu() {
	if (!localStorage.hasOwnProperty('navigatecollapsed')) {
		localStorage.navigatecollapsed = false;
	}
	const $navigation = $.div({ id: 'navigation', class: [localStorage.navigatecollapsed == 'true', 'collapsed'] }, ...[
		$.div({ id: 'navigation-toggle', click: () => {
			localStorage.navigatecollapsed = localStorage.navigatecollapsed != 'true';
			document.getElementById('navigation').classList.toggle('collapsed', localStorage.navigatecollapsed == 'true');
		} },
			$.span({ text: 'Navigation' }),
			$.i({ class: 'tooltip-button mdi mdi-play', title: 'Collapse navigation' })
		),
		permissions['Editor'] && permissions['contenttree'].length != 0 ? 
			[$.a({ href: 'editor.php', class: 'tooltip-button mdi mdi-lead-pencil', title: 'Editor' }), $.a({ href: 'editor.php', text: 'Editor' })] : 
			[],
		permissions['Scheduler'] && permissions['scheduletree'].length != 0 ?
			[$.a({ href: 'scheduler.php', class: 'tooltip-button mdi mdi-calendar-outline', title: 'Scheduler' }), $.a({ href: 'scheduler.php', text: 'Scheduler' })] :
			[],
		(permissions['Players Admin'] || permissions['Players Operator']) && permissions['playertree'].length != 0 ?
			[$.a({ href: 'players.php', class: 'tooltip-button mdi mdi-monitor-dashboard', title: 'Players' }), $.a({ href: 'players.php', text: 'Players' })] :
			[],
		$.div({ class: 'separator' }),
		permissions['Events Tool'] && permissions['eventtree'].length != 0 ?
			[$.a({ href: 'events.php', class: 'tooltip-button mdi mdi-calendar-outline', title: 'Events' }), $.a({ href: 'events.php', text: 'Events' })] :
			[],
		permissions['RSS Creator'] && permissions['rsstree'].length != 0 ?
			[$.a({ href: 'rsscreator.php', class: 'tooltip-button mdi mdi-rss', title: 'RSS Creator' }), $.a({ href: 'rsscreator.php', text: 'RSS Creator' })] :
			[],
		permissions['Network Commands'] && permissions['networkcommandtree'].length != 0 ?
			[$.a({ href: 'networkcommands.php', class: 'tooltip-button mdi mdi-lan-connect', title: 'RS-232/UDP/TCP Commands' }), $.a({ href: 'networkcommands.php', text: 'RS-232/UDP/TCP Commands' })] :
			[],
		permissions['Network Command Scheduler'] && permissions['networkcommandscheduletree'].length != 0 ?
			[$.a({ href: 'networkcommandscheduler.php', class: 'tooltip-button mdi mdi-calendar-outline', title: 'RS-232/UDP/TCP Command Scheduler' }), $.a({ href: 'networkcommandscheduler.php', text: 'RS-232/UDP/TCP Command Scheduler' })] :
			[],
		permissions['CSV File Uploader'] && permissions['csvtree'].length ?
			[$.a({ href: 'csvrolluploader.php', class: 'tooltip-button mdi mdi-format-list-bulleted-square', title: 'CSV File Uploader' }), $.a({ href: 'csvrolluploader.php', text: 'CSV File Uploader' })] :
			[],
		$.div({ class: 'separator' }),
		permissions['Logs'] ?
			[$.a({ href: 'logs.php', class: 'tooltip-button mdi mdi-note-text-outline', title: 'Logs' }), $.a({ href: 'logs.php', text: 'Logs' })] :
			[],
		permissions['Configuration'] ?
			[$.a({ href: 'configuration.php', class: 'tooltip-button mdi mdi-cog-outline', title: 'Configuration' }), $.a({ href: 'configuration.php', text: 'Configuration' })] :
			[],
		$.a({ href: 'users.php', class: 'tooltip-button mdi mdi-account-multiple-outline', title: 'Users' }), $.a({ href: 'users.php', text: 'Users' }),
		$.a({ href: 'https://link.keywesttechnology.com/o9nl8', target: '_blank', class: 'tooltip-button mdi mdi-book-open-page-variant', title: 'Knowledge Base' }), $.a({ href: 'https://link.keywesttechnology.com/o9nl8', target: '_blank', text: 'Knowledge Base' }),
		$.a({ href: 'https://link.keywesttechnology.com/mgn59', target: '_blank', class: 'tooltip-button mdi mdi-lifebuoy', title: 'Support' }), $.a({ href: 'https://link.keywesttechnology.com/mgn59', target: '_blank', text: 'Support' }),
		$.a({ href: 'https://link.keywesttechnology.com/d49o5', target: '_blank', class: 'tooltip-button mdi mdi-content-duplicate', title: 'Templates' }), $.a({ href: 'https://link.keywesttechnology.com/d49o5', target: '_blank', text: 'Templates' }),
		$.a({ href: 'https://link.keywesttechnology.com/portal', target: '_blank', class: 'tooltip-button mdi mdi-shield-account', title: 'KWT Portal' }), $.a({ href: 'https://link.keywesttechnology.com/portal', target: '_blank', text: 'KWT Portal' }),
		$.a({ href: 'logout.php', id: 'logoutButton', class: 'tooltip-button mdi mdi-logout', title: 'Logout' }), $.a({ href: 'logout.php', id: 'logoutButton', text: 'Logout' }),
		$.span({ class: 'spacer' }),
		$.div({ class: 'breeze-logo' })
	].flat());
	const pageName = document.location.pathname.match(/[^\/]+$/)[0];
	for (const $e of $navigation.querySelectorAll(`a[href="${pageName}"]`)) $e.classList.add('selected');
	return $navigation;
}

function generateRegistrationPage() {
	return new Promise(resolve => {
		let $playername;
		let $username;
		let $password;
		let $subdomain;
		let $registerButton;
		
		let systemId = (() => {
			let result = '';
			for (let i = 0; i < 32; ++i) {
				result += 'abcdefghijklmnopqrstuvwxyz0123456789'.charAt(Math.floor(Math.random() * 36));
			}
			return result;
		})();
		
		function setError(el, msg) {
			el.style.border = '2px solid #f00';
			el.title = msg;
		}
		
		function clearError(el) {
			el.style.border = '';
			el.title = '';
		}
		
		document.body.style.backgroundColor = 'var(--background-color)';
		document.body.innerHTML = '';
		document.body.append(
			$.div({ class: 'dialog panel', css: { maxWidth: '600px', margin: '20px', alignSelf: 'start', justifySelf: 'center' } },
				$.header({ text: 'Register Player' }),
				$.div({ css: { margin: '10px 10px 0 10px', color: 'var(--font-color)' }, text: 'If you already have an existing account, fill out the form below to connect this player to your server.' }),
				$.div({ class: 'formgrid' },
					$.label({ text: 'Name' }),
					$playername = $.input({ type: 'text', id: 'playername', class: 'sampleInput', name: 'playername', size: 40, placeholder: 'Player Name' }),
					$.label({ text: 'Username' }),
					$username = $.input({ type: 'text', id: 'username', class: 'sampleInput', name: 'username', size: 40, placeholder: 'Username' }),
					$.label({ text: 'Password' }),
					$password = $.input({ type: 'password', id: 'password', class: 'sampleInput', name: 'password', size: 40, placeholder: 'Password' }),
					$.label({ text: 'Subdomain' }),
					$subdomain = $.input({ type: 'text', id: 'subdomain', class: 'sampleInput', name: 'domain', size: 40, placeholder: 'demo' })
				),
				$.div({ class: 'panel-buttons' },
					$registerButton = $.button({ id: 'registerButton', text: 'Register', click: async () => {
						LockBody();
					
						function registrationError(e) {
							console.log(`registration error: ${e.stack}`);
							if (e.message !== undefined) {
								// TODO: Verify this is an issue with connecting to the server rather than some other issue before setting it to the subdomain
								setError($subdomain, e.message);
							}
						}
					
						clearError($playername);
						clearError($username);
						clearError($password);
						clearError($subdomain);
					
						const formData = {
							playername: $playername.value,
							username: $username.value,
							password: $password.value,
							systemid: systemId,
							localip: await ContainerAPI.QueryLocalIp()
						};
						try {
							const data = await $.http({
								method: 'POST',
								url: `https://${$subdomain.value}.breezedigitalsignage.com/rest/registerplayer.php`,
								data: formData,
								dataType: 'json'
							}).promise;
							UnlockBody();
							if (data.success) {
								try {
									resolve({
										playername: formData.playername,
										account: formData.username,
										key: formData.systemid,
										subdomain: $subdomain.value
									});
								} catch (err) {
									console.log(err);
								}
							} else {
								if (data.error == 'The username supplied is invalid.' || data.error == 'Invalid Login') {
									setError($username, data.error);
								}
								if (data.error == 'The password supplied is invalid.') {
									setError($password, data.error);
								}
								if (data.error == 'The player name supplied is invalid.') {
									setError($playername, data.error);
								}
								if (data.error == 'The system id supplied is invalid.' || data.error == 'Could not connect to user database') {
									console.log(data.error);
								}
								registrationError(data.error);
							}
						} catch (exception) {
							UnlockBody();
							registrationError(exception);
						}
					} })
				),
				$.div({ css: { display: 'flex', flexFlow: 'row nowrap', color: 'white' } },
					$.img({ src: 'ui/images/registrationstart.png', width: 200, height: 200, css: { flex: '0 0 auto', margin: '5px' } }),
					$.div({ css: { flex: '1 0 0', minWidth: '200px', margin: '10px' } },
						$.div({ css: { marginBottom: '10px' }, text: 'If you do not already have an account, you will need to complete an online form first. You can do that on your computer or your phone. Please scan the QR code or go to https://link.keywesttechnology.com/start' }),
						$.div({ text: 'Once your account is created, you will receive a few emails with details to get you started. At that point, you can come back to register this player to your new server using the form above.' })
					)
				)
			)
		);
		
		function setupInput($input, $previous, $next) {
			$input.addEventListener('keydown', e => {
				if (e.keyCode == 38) {
					if ($previous) $previous.focus();
				} else if (e.keyCode == 40) {
					if ($next) $next.focus();
				} else if (e.keyCode == 65376) {
					if ($next && $next.value == '') {
						$next.focus();
					} else {
						$registerButton.click();
					}
				}
			});
		}
		
		setupInput($playername, null, $username);
		setupInput($username, $playername, $password);
		setupInput($password, $username, $subdomain);
		setupInput($subdomain, $password, null);
	});
}

function rgbaToColor(rgba) {
	const matches = rgba.match(/^rgba\(([0-9]+) *, *([0-9]+) *, *([0-9]+) *, *([0-9]+(?:\.[0-9]*)?)\)$/);
	if (matches && matches.length == 5) {
		return {
			r: Math.min(Math.max(0, parseInt(matches[1], 10)), 255),
			g: Math.min(Math.max(0, parseInt(matches[2], 10)), 255),
			b: Math.min(Math.max(0, parseInt(matches[3], 10)), 255),
			a: Math.min(Math.max(0, Math.round(parseFloat(matches[4]) * 100)), 100)
		};
	} else {
		return {
			r: 0,
			g: 0,
			b: 0,
			a: 100
		};
	}
}

async function ShowKeyboard(numbers = true) {
	let $keyboard = document.getElementById('overlaykeyboard');
	if ($keyboard == null) {
		$keyboard = document.body.appendChild($.div({ id: 'overlaykeyboard', class: 'collapsed' }));
		var capsLock = true;
		var mode = 0;
		const keys = [
			[['1', '1'], ['2', '2'], ['3', '3'], ['4', '4'], ['5', '5'], ['6', '6'], ['7', '7'], ['8', '8'], ['9', '9'], ['0', '0']],
			[['q', '!'], ['w', '@'], ['e', '#'], ['r', '$'], ['t', '%'], ['y', '^'], ['u', '&'], ['i', '*'], ['o', '('], ['p', ')']],
			[[' ', ' '], ['a', '_'], ['s', '+'], ['d', '='], ['f', ':'], ['g', ';'], ['h', '/'], ['j', '?'], ['k', '\\'], ['l', ','], [' ', ' ']],
			[['!@#', 'ABC'], ['z', '.'], ['x', '<'], ['c', '>'], ['v', '['], ['b', ']'], ['n', '{'], ['m', '}'], ['-', '-'], ['\'', '\'']]
		];
		const renderKeyboard = () => {
			for (const $row of $keyboard.querySelectorAll(':scope > div')) $row.remove();
			keys.forEach((row, rowIndex) => {
				if (rowIndex == 0 && !numbers) {
					return;
				}
				var rowTag = $.div();
				row.forEach((key, keyIndex) => {
					key = key[mode];
					var keyLetter = key;
					if (capsLock) {
						keyLetter = key.toUpperCase();
					}
					var keyTag = $.div({}, $.span({ text: keyLetter }));
					if (key == ' ') {
						keyTag.style.opacity = 0.5;
					} else if (key != 'ABC' && key != '!@#') {
						keyTag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
							e.preventDefault();
							e.stopImmediatePropagation();
							window.postMessage({ type: 'keyboard', data: keyTag.innerText }, '*');
						}, true);
					}
					if (rowIndex == 2 && (keyIndex == 0 || keyIndex == 10)) {
						keyTag.style.flex = '0.5 0 0px';
						if (keyIndex == 0) {
							if (capsLock) {
								keyTag.querySelector(':scope > span').classList.add('mdi', 'mdi-play', 'mdi-rotate-90');
							} else {
								keyTag.querySelector(':scope > span').classList.add('mdi', 'mdi-play', 'mdi-rotate-270');
							}
							keyTag.style.opacity = 1;
							keyTag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
								e.preventDefault();
								e.stopImmediatePropagation();
								capsLock = !capsLock;
								renderKeyboard();
							});
						}
						if (keyIndex == 10) {
							keyTag.querySelector(':scope > span').classList.add('mdi', 'mdi-play', 'mdi-rotate-180');
							keyTag.style.opacity = 1;
							keyTag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
								e.preventDefault();
								e.stopImmediatePropagation();
								window.postMessage({ type: 'keyboard', data: 'backspace' }, '*');
							});
						}
					} else if (rowIndex == 3) {
						if (keyIndex == 0) {
							keyTag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
								e.preventDefault();
								e.stopImmediatePropagation();
								mode = mode == 0 ? 1 : 0;
								renderKeyboard();
							});
						}
					}
					rowTag.appendChild(keyTag);
				});
				$keyboard.appendChild(rowTag);
			});
			$keyboard.appendChild($.div({},
				$.div({ process: function() {
					this.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
						e.preventDefault();
						e.stopImmediatePropagation();
						window.postMessage({ type: 'keyboard', data: ' ' }, '*');
					}, true);
				} }, $.span({ text: 'Space' })),
				$.div({ css: { flex: '0.20 0 auto' }, process: function() {
					this.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
						e.preventDefault();
						e.stopImmediatePropagation();
						window.postMessage({ type: 'keyboard', data: 'enter' }, '*');
					}, true);
				} }, $.span({ class: 'mdi mdi-keyboard-return' })),
				$.div({ css: { flex: '0.20 0 auto' }, process: function() {
					this.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', e => {
						e.preventDefault();
						e.stopImmediatePropagation();
						this.querySelector(':scope > span').innerText = this.innerText == 'Close' ? 'Keyboard' : 'Close';
						$keyboard.classList.toggle('collapsed');
					}, true);
				} }, $.span({ text: 'Keyboard' }))
			));
		};
		renderKeyboard();
	}
}

function HideKeyboard() {
	let $keyboard = document.getElementById('overlaykeyboard');
	if ($keyboard != null) {
		$keyboard.remove();
	}
}