let $providerTree;

function selectCategory(id) {
	for (const $e of document.getElementById('configuration-right').children) $e.hidden = true;
	document.getElementById(`configuration-section${id}`).hidden = false;
	
	for (const $e of document.querySelectorAll('#categories > .button-choices')) $e.classList.remove('selected');
	document.getElementById(`configuration-${id}`).classList.add('selected');
	
	switch (id) {
		case 'generaloptions':
		default:
			const $configurationSection = document.getElementById(`configuration-section${id}`);
			const $formGrid = $configurationSection.querySelector(':scope > .formgrid');
			$formGrid.innerHTML = '';
			const editedConfiguration = {};
			let foundSection = id == 'generaloptions';
			for (const setting of configurationConfig) {
				const settingId = `configuration-section${id}-${setting.name}`;
				const settingGroup = setting?.group?.replace(/[^a-z0-9]/i, '')?.toLowerCase();
				if (foundSection && setting.hasOwnProperty('group') && settingGroup != id) {
					// This setting is for another group, so stop
					break;
				} else if (setting.hasOwnProperty('group') && settingGroup == id) {
					foundSection = true;
				} else if (!foundSection) {
					continue;
				}
				if (setting.hasOwnProperty('support') && setting.support && !support) {
					continue;
				}
				if (setting.hasOwnProperty('server') && setting.server) {
					continue;
				}
				editedConfiguration[setting.name] = configuration[setting.name];
				
				$formGrid.appendChild(
					$.label({ for: settingId, text: setting.display, title: setting.title })
				);
				switch (setting.type) {
					case 'bool':
						$formGrid.appendChild(
							$.uicheckbox({ id: settingId, change: async function() {
								editedConfiguration[setting.name] = this.checked;
							} })
						);
						document.getElementById(settingId).checked = editedConfiguration[setting.name];
						break;
					case 'string':
						if (setting.hasOwnProperty('options')) {
							const $select = $formGrid.appendChild(
								$.select({ id: settingId })
							);
							if (Array.isArray(setting.options)) {
								for (const option of setting.options) {
									$select.appendChild($.option({ text: option.name, value: option.value }));
								}
							} else {
								switch (setting.options) {
									default:
										throw new Error('Not implemented');
								}
							}
							$select.value = editedConfiguration[setting.name];
							$select.addEventListener('change', function() {
								editedConfiguration[setting.name] = this.value;
							});
						} else {
							$formGrid.appendChild(
								$.input({ id: settingId, type: setting.hasOwnProperty('hidden') ? 'hidden' : (setting.hasOwnProperty('design') && setting.design == 'password' ? 'password' : 'text'), value: editedConfiguration[setting.name], input: function() { editedConfiguration[setting.name] = this.value; } })
							);
						}
						break;
					case 'int':
						$formGrid.appendChild(
							$.input({ id: settingId, type: 'text', value: editedConfiguration[setting.name], keydown: onlyInteger, input: function() {
								const value = parseInt(this.value, 10);
								editedConfiguration[setting.name] = isNaN(value) ? setting.default : value;
							} })
						);
						break;
					case 'json':
						break;
					default:
						throw new Error('Not implemented');
				}
			}
			if (id == 'microsoftpowerbi') {
				const uuidv4 = () => {
					return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
						var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
						return v.toString(16);
					});
				};
				$formGrid.appendChild(
					$.div({ class: 'formgrid-row', text: configuration.serversidesettingset.includes('powerbiaccesstoken') ? 'App Authorized' : 'App Not Authorized. You can authorize the app by clicking the button below.' })
				);
				$formGrid.appendChild(
					$.div({ class: 'formgrid-buttonrow' },
						$.button({ text: 'Authorize Keywest Technology App Integration', title: 'Click this and sign into the Power BI user that will be used to access the Power BI API.', click: async () => {
							const clientId = encodeURIComponent(powerBIClientId);
							const redirectURL = encodeURIComponent('https://powerbi.keywesttechnology.com/powerbi.php');
							const resource = encodeURIComponent('https://analysis.windows.net/powerbi/api');
							const nonce = encodeURIComponent(uuidv4());
							const state = encodeURIComponent(`${host}/post/powerbi.php`);
							window.open(`https://login.microsoftonline.com/common/oauth2/authorize?client_id=${clientId}&redirect_uri=${redirectURL}&resource=${resource}&response_type=id_token+code&scope=openid%20profile&response_mode=form_post&nonce=${nonce}&state=${state}`, 'Keywest Technology Integration')
						} })
					)
				);
				/*
				$configurationSection.appendChild($.div({ class: 'formgrid-buttonrow' },
					$.button({ text: 'Logout of current MS account and delete saved tokens', title: 'Click this to sign out and delete the saved access tokens. This can be useful if you are currently signed into the wrong account. The OAuth consent prompt might default to your current signed in user and will not let you switch until you logout.', click: async () => {
						const redirectURL = encodeURIComponent('https://powerbi.keywesttechnology.com/powerbi.php');
						window.open(`https://login.microsoftonline.com/common/oauth2/logout?post_logout_redirect_uri=${redirectURL}`, 'Keywest Technology Integration Logout');
						await post('/post/powerbilogout.php', {
						});
					} })
				));
				*/
				$formGrid.appendChild(
					$.div({ class: 'formgrid-row', html: 'Note: If you want to revoke the Power BI integration you can navigate to <a href="https://myapplications.microsoft.com">https://myapplications.microsoft.com</a>. From there you can click the dropdown for the application and select "Manage your application" and revoke permissions from that page.' })
				);
			} else {
				$formGrid.appendChild($.div({ class: 'formgrid-buttonrow' },
					$.button({ class: 'primary', text: 'Save', click: async () => {
						try {
							if (editedConfiguration.hasOwnProperty('disableaccount') && configuration['disableaccount'] != editedConfiguration['disableaccount']) {
								if (editedConfiguration['disableaccount']) {
									await ConfirmationBox("Disable account?", 'Are you sure you want to disable the account? This will turn all the players black and lock the user out of the account.', 'Yes', 'No');
								} else {
									await ConfirmationBox("Enable account?", 'Are you sure you want to enable the account? This will update the players back to their original schedule. Users will be able to use the software again.', 'Yes', 'No');
								}
							}
							
							try {
								LockBody();
								var data = await post('/post/saveuserconfig.php', {
									configuration: JSON.stringify(editedConfiguration)
								});
								UnlockBody();
								if (data.success) {
									configuration = data.configuration;
									localStorage['configuration'] = JSON.stringify(configuration);
								} else {
									ConfirmationBox('Error updating configuration', data.error);
								}
							} catch (exception) {
								UnlockBody();
								NetworkingError();
							}
						} catch (exception) {
							// Cancelled
						}
					} })
				));
				if (document.getElementById('clear-server-cache')) {
					document.getElementById('clear-server-cache').addEventListener('click', async () => {
						try {
							LockBody();
							var data = await post('/post/clearservercache.php');
							UnlockBody();
							if (data.success) {
								usage = data.usage;
								ConfirmationBox('Successfully Cleared Server Cache', `Server usage is now at ${usage}%. Please refresh the page.`);
							} else {
								ConfirmationBox('Error clearing server cache', data.error);
							}
						} catch (exception) {
							UnlockBody();
							NetworkingError();
						}
					});
				}
			}
			break;
		case 'fonts':
			LoadFonts();
			break;
		case 'ssoidentityproviders':
			LoadIdPs();
			break;
		case 'eventproviders':
			break;
		case 'capsfeeds':
			FeedControl(document.getElementById('configuration-sectioncapsfeeds'));
			break;
		case 'capsfilters':
			FilterControl(document.getElementById('configuration-sectioncapsfilters'));
			break;
		case 'api':
			ApiControl();
			break;
	}
}

async function LoadIdPs() {
	// Query for all the IdPs
	UnlockBody();
	try {
		var data = await get('/get/idps.php');
		CheckSessionExpired(data);
		if (data.success) {
			let idps = data.data;
			const render = () => {
				document.getElementById('idp-list-saml').innerHTML = '';
				document.getElementById('idp-list-saml').append(
					$.header({},
						$.label({ text: 'Name', title: 'A unique name used to identify this identity provider. Ideally no spaces should be used.', class: 'tooltip-help' }),
						$.label({ text: 'Entity ID', title: 'Identity provider entity ID', class: 'tooltip-help', css: { 'grid-column-end': 'span 3' } }),
						$.label({ text: 'IdP URL', title: 'Service provider identity provider URL', class: 'tooltip-help' }),
						$.label({ text: 'Sign-On Service URL', title: 'Service provider sign-on service URL', class: 'tooltip-help' }),
						$.label({ text: 'Logout Service URL', title: 'Service provider logout service URL', class: 'tooltip-help' }),
						$.label({ text: 'IdP Encryption', title: 'Identity provider encryption', class: 'tooltip-help' }),
						$.label({ text: 'SP Encryption', title: 'Service provider encryption', class: 'tooltip-help', css: { 'grid-column-end': 'span 2' } }),
						$.i({ class: "tooltip-button mdi mdi-plus-circle-outline", title: 'Add an Identity Provider.', click: () => {
							idps.push({
								id: -1,
								type: 'saml',
								name: '',
								entityid: '',
								idp: '',
								singlesignonservice: '',
								singlelogoutservice: '',
								certformat: 'utf8',
								cert: '',
								spcertformat: 'utf8',
								spcert: ''
							});
							render();
						} })
					)
				);
				for (const [index, idp] of idps.entries()) {
					const idpIdInputChange = function() { idp.type = this.value; };
					const idpNameInputChange = function() { idp.name = this.value; };
					const idpURLInputChange = function() { idp.idp = this.value; };
					const idpSingleSignOnServiceInputChange = function() { idp.singlesignonservice = this.value; };
					const idpSingleLogoutServiceInputChange = function() { idp.singlelogoutservice = this.value; };
					const idpCertFormatInputChange = function() { idp.certformat = this.value; render(); };
					const idpCertInputChange = function() { idp.cert = this.value; };
					const idpSPCertFormatInputChange = function() { idp.spcertformat = this.value; render(); };
					document.getElementById('idp-list-saml').append(
						$.input({ type: 'hidden', class: 'idp-id', value: idp.id, input: idpIdInputChange, change: idpIdInputChange }),
						$.input({ type: 'text', placeholder: 'Name', title: 'Name', class: 'idp-name', value: idp.name, readonly: idp.id != -1, input: idpNameInputChange, change: idpNameInputChange }),
						$.input({ type: 'text', placeholder: 'Entity ID', title: 'Entity ID', class: 'idp-type-saml-entityid', value: idp.entityid.match(/^.*\/(.*)$/) ? idp.entityid.match(/^.*\/(.*)$/)[1] : '', readonly: true }),
						$.i({ title: 'Copy Service Provider Metadata URL', class: 'mdi mdi-content-copy tooltip-button', click: () => {
							const tmp = document.body.appendChild($.input({ value: idp.entityid }));
							tmp.select();
							document.execCommand('copy');
							tmp.remove();
						} }),
						$.i({ title: 'Download Service Provider Metadata', class: 'mdi mdi-download-outline tooltip-button', click: () => {
							if (idp.entityid != '') {
								window.location.assign(idp.entityid);
							} else {
								ConfirmationBox('Metadata unavailable', 'Click Update to save first.');
							}
						} }),
						$.input({ type: 'text', placeholder: 'IdP URL', title: 'IdP URL', class: 'idp-type-saml-idp', value: idp.idp, input: idpURLInputChange, change: idpURLInputChange }),
						$.input({ type: 'text', placeholder: 'Sign-On Service URL', title: 'Sign-On Service URL', class: 'idp-type-saml-singlesignonservice', value: idp.singlesignonservice, input: idpSingleSignOnServiceInputChange, change: idpSingleSignOnServiceInputChange }),
						$.input({ type: 'text', placeholder: 'Logout Service URL', title: 'Logout Service URL', class: 'idp-type-saml-singlelogoutservice', value: idp.singlelogoutservice, input: idpSingleLogoutServiceInputChange, change: idpSingleLogoutServiceInputChange }),
						$.select({ title: 'Identity Provider encryption. If encryption is enabled on the IdP you must download the cert from there and paste it here.', class: 'idp-type-saml-certformat', value: idp.certformat, input: idpCertFormatInputChange, change: idpCertFormatInputChange },
							$.option({ text: 'No Identity Provider Certificate', value: 'none' }),
							$.option({ text: 'ASCII/UTF-8 (IdP Certificate Input Format)', value: 'utf8' }),
							$.option({ text: 'Base64 (IdP Certificate Input Format)', value: 'base64' })
						),
						$.select({ title: 'This turns on Service Provider encryption. If enabled you must copy/download the below cert and register it with the IdP when configuring the SP there.', class: 'idp-type-saml-spcertformat', value: idp.spcertformat, css: {
							gridColumnEnd: idp.spcertformat == 'none' ? 'span 2' : 'span 1'
						}, input: idpSPCertFormatInputChange, change: idpSPCertFormatInputChange },
							$.option({ text: 'No Service Provider Certificate', value: 'none' }),
							$.option({ text: 'ASCII/UTF-8 (SP Certificate Output Format)', value: 'utf8' }),
							$.option({ text: 'Base64 Encoded (SP Certificate Output Format)', value: 'base64' })
						),
						$.i({ title: 'Download Service Provider Certificate', class: 'mdi mdi-download-outline tooltip-button', hidden: idp.spcertformat == 'none', css: {
							border: '1px solid var(--input-border-color)',
							borderLeft: 'none'
						}, click: () => {
							if (idp.spcert != '') {
								var a = document.body.appendChild($.a({ href: window.URL.createObjectURL(new Blob([idp.spcert], {type: 'text/plain'})) }));
								// Use text/plain so that the browser initiates a download rather than an import into its cert store
								a.download = 'KeywestTechnologyBreezeSP.cert';
								a.click();
								document.body.removeChild(a);
							} else {
								ConfirmationBox('Cert unavailable', 'Click Update Identity Providers to save and generate a Service Provider certificate first.');
							}
						} }),
						$.i({ title: 'Delete', class: 'mdi mdi-trash-can-outline tooltip-button', click: function() {
							idps.splice(index, 1);
						} }),
						$.div({ class: 'textarea-container', hidden: idp.certformat == 'none' },
							$.textarea({ rows: 5, placeholder: 'Identity Provider Certificate', title: 'Identity Provider Certificate', class: 'idp-type-saml-cert', value: idp.cert, input: idpCertInputChange, change: idpCertInputChange }),
						),
						$.div({ class: 'textarea-container', hidden: idp.spcertformat == 'none' },
							$.textarea({ rows: 5, placeholder: 'Service Provider Certificate', title: 'Service Provider Certificate (Click to copy)', readonly: true, class: 'idp-type-saml-spcert', value: (() => {
								switch (idp.spcertformat) {
									case 'none':
										return '';
										break;
									case 'utf8':
										return idp.spcert;
										break;
									case 'base64':
										return idp.spcert.replace('-----BEGIN CERTIFICATE-----', '').replace('-----END CERTIFICATE-----', '').replace(/\r?\n|\r/g, '').trim();
										break;
								}
							})(), click: function() {
								const tmp = document.body.appendChild($.textarea({ value: this.value }));
								tmp.select();
								document.execCommand('copy');
								tmp.remove();
							} }),
						),
						$.div({ class: 'textarea-container' },
							$.textarea({ rows: 5, placeholder: 'Service Provider Metadata', title: 'Service Provider Metadata (Click to copy)', readonly: true, class: 'idp-type-saml-spmetadata', value: idp.spmetadata, click: () => {
								const tmp = document.body.appendChild($.textarea>({ value: idp.spmetadata }));
								tmp.select();
								document.execCommand('copy');
								tmp.remove();
							} })
						),
						$.uiinputselect({
							class: 'idp-attributefirstname',
							placeholder: 'Attribute First Name (Optional)',
							title: 'Attribute First Name (Optional): The attribute field (returned from the IdP) to use for the first name for the created user account.',
							value: idp.attributefirstname = idp.hasOwnProperty('attributefirstname') ? idp.attributefirstname : '',
							config: {
								options:[
									{ name: 'firstname', value: 'firstname' }
								],
								allowInput: true
							}, change: function() { idp.attributefirstname = this.value; }
						}),
						$.uiinputselect({
							class: 'idp-attributelastname',
							placeholder: 'Attribute Last Name (Optional)',
							title: 'Attribute Last Name (Optional): The attribute field (returned from the IdP) to use for the last name for the created user account.',
							value: idp.attributelastname = idp.hasOwnProperty('attributelastname') ? idp.attributelastname : '',
							config: {
								options: [
									{ name: 'lastname', value: 'lastname' }
								],
								allowInput: true
							},
							change: function() { idp.attributelastname = this.value; }
						}),
						$.uiinputselect({
							class: 'idp-attributeemailaddress',
							placeholder: 'Attribute Email Address (Optional)',
							title: 'Attribute Email Address (Optional): The attribute field (returned from the IdP) to use for the email address for the created user account.',
							value: idp.attributeemailaddress = idp.hasOwnProperty('attributeemailaddress') ? idp.attributeemailaddress : '',
							config: {
								options: [
									{ name: 'Active Directory Email Address', value: 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/emailaddress' }
								],
								allowInput: true
							},
							change: function() { idp.attributeemailaddress = this.value; }
						}),
						$.uiinputselect({
							class: 'idp-attributedisplayname',
							placeholder: 'Attribute Display Name (Optional)',
							title: 'Attribute Display Name (Optional): The name used in the user tree. If left blank the "Persistent Name ID" will be used. Many IdPs use a randomly generated ID for their Persistent Name ID (like Active Directory) so that the user can change email addresses without breaking SSO links. This option allows for a more readable name to be used in the tree.',
							value: idp.attributedisplayname = idp.hasOwnProperty('attributedisplayname') ? idp.attributedisplayname : '',
							config: {
								options: [
									{ name: 'Persistant Name ID', value: '' },
									{ name: 'First Name + Last Name', value: 'First Name + Last Name' },
									{ name: 'Email Address', value: 'Email Address' }
								],
								allowInput: true
							}, change: function() { idp.attributedisplayname = this.value; }
						})
					);
				}
			};
			render();
			
			UnlockBody();
			
			document.getElementById('updateIdPs').replaceWith($.button({ id: 'updateIdPs', class: 'primary', text: 'Update', click: async () => {
				var errors = [];
				for (const idp of idps) {
					if (!IsAlphaNumericSpacesHyphensUnderscoresNotEmptyLessThan60Characters(idp.name)) {
						errors.push('The name can only container letters, numbers, spaces, hyphens, or underscores, and must be between 1 and 60 characters.');
					}
					switch (idp.type) {
						case 'saml':
							if (idp.idp == '') {
								errors.push('An Identity Provider URL is required.');
							}
							if (idp.singlesignonservice == '') {
								errors.push('A sign-on service URL is required.');
							}
							if (idp.certformat != 'none' && idp.cert == '') {
								errors.push('You have selected to use an Identity Provider Certificate but have not supplied one. You can use the Identity Provider Certificate Format dropdown to disable the use of a certificate only if the Identity Provider does not use one.');
							}
							break;
						case 'openid':
							errors.push('OpenID is not supported');
							break;
					}
				}
				if (errors.length != 0) {
					ConfirmationBox('Invalid IdP configuration', errors.join('\n'));
				} else {
					LockBody();
					try {
						var data = await post('/post/updateidps.php',
						{
							idps: JSON.stringify(idps)
						});
						
						UnlockBody();
						CheckSessionExpired(data);
						if (data.success) {
							ConfirmationBox('Identity Providers Updated', 'Identity Providers Updated.');
							LoadIdPs();
						} else {
							ConfirmationBox('Error updating', data.error);
						}
					} catch (exception) {
						UnlockBody();
						NetworkingError();
					}
				}
			} }));
		}
	} catch (exception) {
		NetworkingError();
	}
}

async function LoadFonts() {
	try {
		LockBody();
		const data = await get('/get/fonts.php');
		UnlockBody();
		CheckSessionExpired(data);
		if (data.success) {
			document.getElementById('configuration-disabledfonts').innerHTML = '';
			document.getElementById('configuration-enabledfonts').innerHTML = '';
			document.getElementById('configuration-disabledfonts').lastSelected = null;
			document.getElementById('configuration-enabledfonts').lastSelected = null;
			for (const font of data.fonts.sort((a, b) => a.name.localeCompare(b.name))) {
				var clickItem = function(e, clicks) {
					if (clicks == 1) {
						var lastSelected = this.parentNode.lastSelected;
						if (e.ctrlKey) {
							this.classList.toggle('selected');
							lastSelected = this;
						} else if (e.shiftKey) {
							this.parentNode.querySelector('.selected').classList.remove('selected');
							if (lastSelected == null) {
								lastSelected = this.parentNode.querySelector('li:first-child');
							}
							var node = lastSelected;
							node.classList.add('selected');
							while (node !== this) {
								node = [...node.parentNode.children].indexOf(node) < [...this.parentNode.children].indexOf(this) ? node.nextElementSibling : node.previousElementSibling;
								node.classList.add('selected');
							}
						} else {
							for (const $e of this.parentNode.querySelectorAll('.selected')) $e.classList.remove('selected');
							this.classList.add('selected');
							lastSelected = this;
						}
						this.parentNode.lastSelected = lastSelected;
					} else if (clicks == 2) {
						if (this.parentNode.id == 'configuration-enabledfonts') {
							for (const $font of document.querySelectorAll('#configuration-disabledfonts li, #configuration-enabledfonts li')) {
								$font.defaultfont = false;
								$font.innerText = $font.name;
							}
							this.defaultfont = true;
							this.innerText = `${this.name} (Default)`;
						}
					}
				};
				if (font.enabled) {
					const $li = document.getElementById('configuration-enabledfonts').appendChild($.li({ text: `${font.name}${font.defaultfont ? ' (Default)' : ''}`, multiclick: clickItem, css: { fontFamily: `'${font.name}'` } }));
					$li.name = font.name;
					$li.defaultfont = font.defaultfont;
				} else {
					const $li = document.getElementById('configuration-disabledfonts').appendChild($.li({ text: font.name, multiclick: clickItem, css: { fontFamily: `'${font.name}'` } }));
					$li.name = font.name;
					$li.defaultfont = font.defaultfont;
				}
			}
			document.getElementById('configuration-customfonts').innerHTML = '';
			document.getElementById('configuration-customfonts').append(
				...data.customfonts.map(font => $.option({ text: font, value: font }))
			);
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
	}
}

function updateStorage(e) {
	if (e && e.originalEvent) {
		e = e.originalEvent;
		if (e.key == 'schedulefilterplaylistonly') {
			document.querySelector('#apiContainer .filterplaylistonly').checked = localStorage.schedulefilterplaylistonly == 'true';
		}
	}
}

async function ApiControl() {
	const response = await get('/get/apitoken.php');
	if (!response.success) {
		toastError(`Failed to load API data. ${result.error}`);
		return;
	}
	let apiTokens = response.apitokens;
	
	if (!localStorage.hasOwnProperty('schedulefilterplaylistonly')) {
		localStorage.schedulefilterplaylistonly = false;
	}
	
	window.removeEventListener('storage', updateStorage);
	window.addEventListener('storage', updateStorage);
	
	function toggleFilterPlaylistOnly() {
		localStorage.schedulefilterplaylistonly = this.checked;
	}
	
	const render = () => {
		document.getElementById('apiContainer').innerHTML = '';
		document.getElementById('apiContainer').append(
			$.header({},
				$.label({},
					$.span({ text: 'Player' }),
					$.i({ class: 'mdi tooltip-button', title: '' })
				),
				$.label({},
					$.span({ text: 'Schedule' }),
					$.uicheckbox({ class: 'filterplaylistonly', checked: localStorage.schedulefilterplaylistonly == 'true', title: 'Filter Playlist Only', change: toggleFilterPlaylistOnly },
						$.span({ text: 'Playlist Only' })
					),
					$.i({ class: 'mdi tooltip-button', title: '' })
				),
				$.label({},
					$.span({ text: 'URL' }),
					$.i({ class: 'mdi tooltip-button', title: '' })
				),
				$.label({},
					$.span({ text: 'Copy' }),
					$.i({ class: 'mdi tooltip-button', title: 'Copy to Clipboard' })
				),
				$.label({},
					$.span({ text: 'Run' }),
					$.i({ class: 'mdi tooltip-button', title: 'Run Now' })
				),
				$.i({ class: 'tooltip-button mdi mdi-plus-circle-outline', title: 'Add API Call', click: () => {
					apiTokens.push({ id: -1, data: { players: [], schedule: null, url: '', token: null, type: 'schedule' } });
					render();
				} })
			)
		);
		// TODO: Handle the try catch and error states better. Right now it just reloads which might be fine
		for (const apiToken of apiTokens) {
			const index = apiTokens.indexOf(apiToken);
			
			let $apiTokenUrl;
			const apiSave = async function(apiToken) {
				if (apiToken == null || apiToken["data"] == null || apiToken.data["schedule"] == null || apiToken.data["players"] == null) {
					return;
				}
				if (Array.isArray(apiToken.data.players) && apiToken.data.players.length > 0 && apiToken.data.schedule != null) {
					// save/update
					try {
						const result = await post('/post/apitoken.php', { action: "save", apitoken: JSON.stringify(apiToken) } );
						if (result.success) {
							const response2 = result.response;
						
							apiToken.id = response2.id;
							apiToken.data.players = response2.data.players;
							apiToken.data.schedule = response2.data.schedule;
							apiToken.data.url = response2.data.url;
							apiToken.data.token = response2.data.token;
							apiToken.data.type = response2.data.type;
						
							$apiTokenUrl.value = apiToken.data.url;
						}
					} catch (exception) {
						NetworkingError();
					}
				}
			}
			
			document.getElementById('apiContainer').append(
				$.uitreeselect({
					config: {
						tree: 'player',
						showRoot: true,
						checkbox: true,
						leafIcon: node => node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute')) ? 'playeron' : 'playeroff',
						filter: node => node.parentNode != null
					}, change:  async function() {
						apiToken.data.players = (await this.getCheckedAllChildren()).map(node => node.playerid);
						apiSave(apiToken);
					}, process: async $treeSelect => {
						await $treeSelect.setChecked(node => apiToken.data.players.includes(node.playerid));
					}
				}),
				$.uitreeselect({
					config: {
						tree: 'schedule',
						leafIcon: node => 'schedule',
						disable: node => node.group,
						filter: node => {
							if (localStorage.schedulefilterplaylistonly == 'false') {
								return true;
							} else {
								if (node.group) {
									return true;
								}
								return !/^.+\(Playlist Only\)$/.test(node.name);
							}
						}
					}, change: async function() {
						apiToken.data.schedule = (await this.getSelected())[0]?.scheduleid ?? null;
						apiSave(apiToken);
					}, process: async $treeSelect => {
						await $treeSelect.setSelected(node => apiToken.data.schedule == node.scheduleid);
					}
				}),
				$apiTokenUrl = $.input({ type: 'text', readonly: true, value: apiToken.data.url }),
				$.i({ class: 'tooltip-button mdi mdi-content-copy', title: 'Copy to Clipboard', click: () => {
					const $tmp = document.body.appendChild($.input({ value: apiToken.data.url }));
					$tmp.select();
					document.execCommand('copy');
					$tmp.remove();
					toastSuccess('URL copied to your clipboard.');
				} }),
				$.i({ class: 'tooltip-button mdi mdi-play', title: 'Run', click: async () => {
					LockBody();
					try {
						const result = await post(apiToken.data.url);
						if (result.success) {
							toastSuccess('API call successful. Your player(s) will update very soon.');
						} else {
							toastError(`API call failed. ${result.error}`);
						}
					} catch (exception) {
						NetworkingError();
						await ApiControl();
					} finally {
						UnlockBody();
					}
				} }),
				$.i({ class: 'tooltip-button mdi mdi-trash-can-outline', title: 'Delete', click: async () => {
					LockBody();
					try {
						const result = await post('/post/apitoken.php', { action: "delete", id: apiToken.id });
						if (result.success) {
							toastSuccess('API call successful deleted.');
						} else {
							toastError(`API call unsuccessful deleted. ${result.error}`);
						}
						apiTokens.splice(index, 1);
						render();
					} catch (exception) {
						NetworkingError();
						await ApiControl();
					} finally {
						UnlockBody();
					}
				} })
			);
		}
	};
	render();
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.li({ id: 'configuration-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'Categories' }),
				$.ul({ id: 'categories' }, ...[
					$.li({ id: 'configuration-generaloptions', click: () => selectCategory('generaloptions'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-cog-outline' }),
						$.span({ text: 'General Options' })
					),
					permissions['Fonts'] ? [$.li({ id: 'configuration-fonts', click: () => selectCategory('fonts'), class: 'button-choices' },
							$.i({ class: 'mdi mdi-format-font' }),
							$.span({ text: 'Fonts' })
					)] : [],
					permissions['SSO Identity Providers'] ? [$.li({ id: 'configuration-ssoidentityproviders', click: () => selectCategory('ssoidentityproviders'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-account-key-outline' }),
						$.span({ text: 'SSO Identity Providers' })
					)] : [],
					permissions['CAPS Feeds'] ? [$.li({ id: 'configuration-capsfeeds', click: () => selectCategory('capsfeeds'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-weather-hurricane' }),
						$.span({ text: 'CAPS Feeds' })
					)] : [],
					permissions['CAPS Filters'] ? [$.li({ id: 'configuration-capsfilters', click: () => selectCategory('capsfilters'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-weather-hurricane' }),
						$.span({ text: 'CAPS Filters' })
					)] : [],
					permissions['providertree'] ? [$.li({ id: 'configuration-eventproviders', click: () => selectCategory('eventproviders'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-calendar-outline' }),
						$.span({ text: 'Event Providers' })
					)] : [],
					$.li({ id: 'configuration-api', click: () => selectCategory('api'), class: 'button-choices' },
						$.i({ class: 'mdi mdi-api' }),
						$.span({ text: 'API' })
					)
				].flat())
			)
		),
		$.li({ id: 'configuration-left-resize', class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.li({ id: 'configuration-right' },
			$.div({ id: 'configuration-sectiongeneraloptions', class: 'panel' }, ...[
				$.header({ text: 'General Options' }),
				$.div({ class: 'formgrid' }),
				$.div({ id: 'configuration-about' },
					$.h3({ text: 'About' })
				),
				$.div({ id: 'configuration-version' },
					$.div({ text: `Breeze Version: ${breezeVersion}` })
				),
				isKwtAdmin ? [$.div({ id: 'configuration-hdd' },
					$.div({ text: 'Server HDD Usage:' }),
					$.div({},
						$.div({ css: { width:`${totalHddUsage}%`, backgroundColor: totalHddUsage < 80 ? 'rgba(60, 60, 60, 0.9)': (totalHddUsage < 90 ? 'rgba(200, 120, 15, 0.9)' : 'rgba(200, 15, 15, 0.9)') } },
							$.span({ text: `${Math.round(totalHddUsage)}%` })
						)
					),
					$.div({},
						$.button({ id: 'clear-server-cache', class: 'primary', text: 'Clear Server Cache' })
					)
				)] : []
			].flat()),
			$.div({ id: 'configuration-sectionfonts', class: 'panel' },
				$.header({ text: 'Fonts' }),
				$.div({ class: 'formgrid', css: { width: '450px' } },
					$.label({ text: 'Font Name' }),
					$.input({ type: 'text', id: 'configuration-fontname', class: 'sampleInput' }),
					
					$.label({ text: 'File (.ttf, .otf, .woff)' }),
					$.label({ class: 'button' },
						$.span({ id: 'configuration-filename', css: { textOverflow: 'ellipsis' }, text: 'Choose File' }),
						$.input({ type: 'file', id: 'configuration-file', css: { display: 'none' } })
					),
					
					$.button({ id: 'fonts-upload', class: 'primary formgrid-label', text: 'Upload' }),
					$.progress({ id: 'configuration-progress' })
				),
				$.div({ id: 'configuration-deletefontform' },
					$.select({ id: 'configuration-customfonts', class: 'sampleSelect' }),
					$.button({ id: 'configuration-deletefont', class: 'primary', text: 'Delete' })
				),
				$.div({ id: 'configuration-enabledisable' },
					$.div({ class: 'panel enabledisable' },
						$.header({ class: 'tooltip-help', title: `Fonts in this list are unavailable to use in the editor. Move fonts here if you don't plan to use them.`, text: 'Disabled Fonts' }),
						$.ul({ id: 'configuration-disabledfonts' })
					),
					$.ul({ class: 'fontcontrols' },
						$.li({ id: 'configuration-enable', class: 'tooltip-button mdi mdi-arrow-right', title: 'Enable the selected font' }),
						$.li({ id: 'configuration-disable', class: 'tooltip-button mdi mdi-arrow-left', title: 'Disable the selected font' })
					),
					$.div({ class: 'panel enabledisable' },
						$.header({ class: 'tooltip-help', title: 'Fonts in this list are available to use in the editor. Move fonts you need to use into this list.', text: 'Enabled Fonts (Double click to set default)' }),
						$.ul({ id: 'configuration-enabledfonts' })
					)
				),
				$.div({ class: 'panel-buttons' },
					$.button({ id: 'configuration-savefonts', class: 'primary', text: 'Save' })
				)
			),
			$.div({ id: 'configuration-sectionssoidentityproviders', class: 'panel' },
				$.header({ text: 'SSO Identity Providers' }),
				$.div({ id: 'idp-list-saml', class: 'formgrid compact' }),
				$.div({ class: 'panel-buttons' },
					$.button({ id: 'updateIdPs', class: 'primary', text: 'Update' })
				),
				$.ul({ id: 'idp-accordion' },
					$.li({ class: 'accordion-header', text: 'Single Sign-On (SSO) Basics' }),
					$.li({},
						$.ul({},
							$.li({ text: 'Single sign-on allows Breeze, the Service Provider (SP), to use an external Identity Provider (IdP) to authenticate users.' }),
							$.li({ text: 'For example, if your organization uses SAML to authenticate users for other systems then you can enter the SAML IdP metadata on this page. Your employees/faculty/etc will then be given an option on the login page to authenticate with that specified IdP.' }),
							$.li({ text: 'When a user first logs in using an IdP their account in Breeze will be created; however, they will have no permissions. It is up to the administrator account in Breeze to set their permissions on the Users page. Once that is done the user can login and access features.' }),
							$.li({ text: 'SSO cannot be configured to connect directly to an LDAP server. You must configure either a SAML or OpenID server to handle the SSO process on the IdP side.' })
						)
					),
					$.li({ class: 'accordion-header', text: 'SAML (General Configuration)' }),
					$.li({},
						$.ul({},
							$.li({ text: 'To add a SAML IdP click the plus icon. Then fill in the requested information. Note that a lot of this information can be found in the metadata on the IdP.' }),
							$.li({ html: `<b>Name</b>: The name for the IdP which will display on the Login page. This can be an organization name or anything so that users know to click it. For example, "University', will be displayed as "Login with University". This name is also used for the folder in the Users tree that new accounts will be added to.` }),
							$.li({ html: '<b>Entity ID</b>: The Entity ID is the name of the SP used on the IdP. For convenience the Entity ID is generated using the previous name field - lowercasing and removing spaces.' }),
							$.li({ html: '<b>IdP URL</b>: The IdP is a URL to the Entity ID on the IdP. Example: https://example.com/saml2/idp/metadata.php' }),
							$.li({ html: '<b>Sign-on Service URL</b>: The URL for the IdP sign-on service. Example: https://example.com/saml2/idp/SSOService.php' }),
							$.li({ html: '<b>Logout Service URL</b>: The URL for the IdP Logout Service. Example: https://example.com/saml2/idp/SingleLogoutService.php' }),
							$.li({ html: `<b>Identity Provider Encryption</b>: Whether a cert is being used on the IdP and if so the format you've entered it in. Some IdPs will give you the raw .crt file. Open it in a text editor and paste it into the field. Others will display it in Base64 format.` }),
							$.li({ html: `<b>Service Provider Encryption</b>: You can change the displayed format of the certificate below. If enabled you'll need to supply this to the IdP for its Service Provider metadata.` }),
							$.li({ html: `<b>Identity Provider Certificate</b>: The x509 certificate of the IdP. If the IdP lists a certificate it must be pasted here since it's used to decrypt communication.` }),
							$.li({ html: '<b>Service Provider Certificate</b>: You can copy this or click the download button. Note that you need to save before the certificate is generated.' }),
							$.li({ html: '<b>Service Provider Metadata</b>: You must copy this and put it into the IdP to make this SP trusted. Note you need to save before the metadata is generated.' },
								$.ul({},
									$.li({ text: 'This must be modified to define the NameID, "username", when it is copied to the IdP.' }),
									$.li({ text: 'If you are using SimpleSAMLphp there is an XML conversion tool in the IdP control panel to their format.' }),
									$.li({ text: 'As defined in the metadata, you can optionally supply firstname and lastname attributes from the IdP to fill in that information on the Users page if your authentication source has that information.' })
								)
							),
							$.li({ text: `Breeze is the SAML Service Provider and needs to be registered with the SAML Identity Provider's metadata list. (Some systems label this as SP Remote Metadata).` }),
							$.li({ text: `If it isn't clear a SAML IdP can be set up to use a database, LDAP, OpenID, or any number of data sources as its authentication source. You could run SimpleSAMLphp for instance on a publicly accessible server that is configured for say LDAP then use a SAML SP here to authenticate with it.` })
						)
					),
					$.li({ class: 'accordion-header', text: 'SAML (Azure Active Directory)' }),
					$.li({},
						$.ul({},
							$.li({ text: 'Ensure that on the Active Directory side you have correctly created the SAML Single sign-on application. If you have not created one or need to verify the process the steps are below:' },
								$.ul({},
									$.li({ html: `In the Azure Portal (<a href="https://portal.azure.com', target="_blank">https://portal.azure.com</a>) ensure you are logged in as the Global Admin account for your organization. This is generally required to create enterprise applications and also assign users to the application.` }),
									$.li({ text: `Navigate to "Enterprise applications" (use the search if you don't see it on the home screen) and select New Application then select "Non-gallery application" as this will be a custom setup.` }),
									$.li({ text: `Give the application a name like "BreezeSAML". Once created you'll be taken to the Overview page. You can get back to this page in the future by selecting "Enterprise applications" then searching for "BreezeSAML" and selecting it.` }),
									$.li({ text: `On the left of the application page you'll various sections. Select "Single sign-on" below the "Manage" group header. If you haven't already setup Single sign-on you'll be prompted to select a method. Select SAML.` }),
									$.li({ text: 'At this point you need to setup the Breeze side first, described below, to get its Metadata XML file. (This file contains all the URLs the application needs to configure SAML with Breeze).' }),
									$.li({ text: 'Once you have uploaded the Metadata file to the application everything will be configured.' }),
									$.li({ text: 'The last step is to add users to the application that can be authenticated with Single sign-on. On the left is "Users and groups". Any user or group added here can use this login system for Breeze.' }),
									$.li({ text: `Once you have added users you can test the login by logging out of Breeze and clicking the new login button on the right. You'll be redirected to Microsoft to login then taken back to Breeze.` }),
									$.li({ text: 'Note: Users created have no permissions. You must logout and login with your regular Breeze account then on the Users page set permissions for the new user.' })
								)
							),
							$.li({ text: 'To add a SAML IdP for Azure Active Directory click the plus icon. Then fill in the requested information found on the Single sign-on page.' }),
							$.li({ html: '<b>Name</b>: The name for the IdP which will display on the Login page. This can be an organization name or anything so that users know to click it. For example, "University" will be displayed as "Login with University". This name is also used for the folder in the Users tree that new accounts will be added to.' }),
							$.li({ html: '<b>Entity ID</b>: The Entity ID is the name of the SP used on the IdP. For convenience the Entity ID is generated using the previous name field - lowercasing and removing spaces.' }),
							$.li({ html: '<b>IdP URL</b>: The IdP is a URL to the Entity ID on the IdP. Example: https://sts.windows.net/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/' }),
							$.li({ html: '<b>Sign-On Service URL</b>: The URL for the IdP sign-on service. Example: https://login.microsoftonline.com/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/saml2' }),
							$.li({ html: '<b>Logout Service URL</b>: The URL for the IdP logout service. This is: https://login.microsoftonline.com/common/wsfederation?wa=wsignout1.0' }),
							$.li({ html: '<b>Identity Provider Encryption</b>: Set this to "ASCII/UTF-8".' }),
							$.li({ html: '<b>Service Provider Encryption</b>: Set this to No Service Provider Certificate. Azure does not require this.' }),
							$.li({ html: '<b>Identity Provider Certificate</b>: The x509 certificate of the IdP. In Azure application page in the Single sign-on section in the "SAML Signing Certificate" box there is a "Certificate (Base64)". Download, open it in Notepad, and then copy and paste it here.' }),
							$.li({ html: '<b>Service Provider Certificate</b>: This field is hidden when you turn off Service Provider Encryption.' }),
							$.li({ html: '<b>After filling in the above information press Update to generate the Metadata file.</b>' }),
							$.li({ html: '<b>Service Provider Metadata</b>: Use the download button in the Entity ID column to get this file. On the Azure application Single sign-on section click the "Upload metadata file". This will automatically fill in the required information in the "Basic SAML Configuration" section.' })
						)
					),
					$.li({ class: 'accordion-header', text: 'OpenID (Not Implemented)' }),
					$.li({},
						$.ul()
					)
				)
			),
			$.div({ id: 'configuration-sectioneventproviders', class: 'panel' },
				$.header({ text: 'Event Providers' }),
				$.ul({},
					$.li({},
						$.ul({ id: 'provider-toolbar', class: 'iconlist' },
							$.li({ id: 'provider-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $providerTree.deleteSelected(), title: 'Delete' }),
							$.li({ id: 'provider-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: ProviderAddCustomRoomGroup, title: 'Add room group' }),
							$.li({ id: 'provider-addcredentials', class: 'tooltip-button mdi mdi-key-plus', click: ShowNewCredentials, title: 'New credentials' }),
							$.li({ id: 'provider-adddelegate', class: 'tooltip-button mdi mdi-email-outline', click: ProviderCredentialsAddCalendar, title: 'Add Additional Calendar - Add a calendar that has delegated access to this credential.' }),
							$.li({ id: 'provider-addcampusbuilding', class: 'tooltip-button mdi mdi-home-city-outline', click: ProviderCredentialsAddCampusBuilding, title: 'Add Campuses and/or Buildings - Add a campus and all associated buildings and/or individual buildings.' }),
							$.li({ id: 'provider-addroom', class: 'tooltip-button mdi mdi-door', click: ProviderCredentialsAddRoom, title: 'New room' }),
							$.li({ id: 'provider-addrooms', class: 'tooltip-button mdi mdi-door', title: 'New Rooms from CSV. The format is the first column is the room name and the second optional column is the replacement name.' },
								$.input({ type: 'file' })
							),
						),
						$providerTree = $.uitree({ id: 'providertree', config: {
							tree: 'provider',
							dispatchChangeOnLoad: true,
							multiselect: true,
							groupTypes: [1, 2, 3, 5, 6, 8, 9, 10],
							groupIcon: node => {
								switch (node.type) {
									case 2: // Provider
										return 'provider';
									case 3: // Provider Credentials
										return 'credential';
									case 8: // Provider Delegate
										return 'delegate';
									case 10: // Ad Astra Building
										return 'building';
									default:
										return '';
								}
							},
							leafIcon: node => 'room',
							editable: node => node.type == 3 || node.type == 6,
							filterDraggable: node => node.type == 4 || node.type == 7,
							filterDroppable: (draggingNodes, node) => node.type == 6,
							deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
								switch (selectedNodes[0].type) {
									case 3: // Provider Credentials
										await ConfirmationBox("Remove credentials?", `Are you sure you want to remove the credentials for '${selectedNodes.map(node => node.name).join("', '")}'? This will also remove the associated room information.`, 'Yes', 'No');
										break;
									case 4: // Provider Credentials Room
										await ConfirmationBox("Remove room?", `Are you sure you want to remove the room '${selectedNodes.map(node => node.name).join("', '")}'?`, 'Yes', 'No');
										break;
									case 6: // Custom Room Groups Group
										await ConfirmationBox("Remove custom room group?", `Are you sure you want to remove the custom room group '${selectedNodes.map(node => node.name).join("', '")}'?`, 'Yes', 'No');
										break;
									case 7: // Custom Room Groups Group Room
										await ConfirmationBox("Remove room reference?", `Are you sure you want to remove the room '${selectedNodes.map(node => node.name).join("', '")}'?`, 'Yes', 'No');
										break;
									case 8: // Provider Credentials Delegate
										await ConfirmationBox("Remove impersonation/delegate email?", `Are you sure you want to remove the impersonation/delegate email '${selectedNodes.filter(node => node.type == 8).map(node => node.name).join("', '")}'? This will also remove the associated room information.`, 'Yes', 'No');
										break;
								}
								try {
									await deleteNodes(selectedAllChildren);
								} catch (data) {
									if (data.hasOwnProperty('roomgrouprooms')) {
										await ConfirmationBox("Remove room reference(s)?", `The following rooms have references in custom room groups (found in the bottom of the tree): '${data.roomgrouprooms.map(node => node.name).join("', '")}'. Deleting the room will also delete these custom room group rooms. Continue?`, 'Yes', 'No');
										await deleteNodes(selectedAllChildren, true);
									}
								}
							}
						}, change: async function() {
							HideProvider();
							let selectedNode = await this.getSelected();
							if (selectedNode.length == 0) {
								document.getElementById('provider-tip').hidden = false;
							} else if (selectedNode.length > 1) {
								let type = null;
								if (selectedNode.every(node => {
									if (type == null) {
										type = node.type;
									}
									return type == node.type;
								})) {
									switch (type) {
										case 3:
										case 4:
										case 6:
										case 7:
										case 8:
										case 9:
										case 10:
											document.getElementById('provider-delete').hidden = false;
											break;
									}
								}
							} else {
								selectedNode = selectedNode[0];
								try {
									CheckProviderChange();
									switch (selectedNode.type) {
										case 1: // Event Providers Root Node
											document.getElementById('provider-tip').hidden = false;
											break;
										case 2: // Provider
											document.getElementById('provider-addcredentials').hidden = false;
											document.getElementById('provider-credentialstip').hidden = false;
											break;
										case 3: // Provider Credentials
											ShowProviderCredentials(selectedNode);
											break;
										case 4: // Provider Credentials Room
											ShowProviderRoom(selectedNode);
											break;
										case 5: // Custom Room Groups
											ShowProviderCustomRoomGroupsTip();
											break;
										case 6: // Custom Room Groups Group
											document.getElementById('provider-delete').hidden = false;
											break;
										case 7: // Custom Room Groups Group Room
											document.getElementById('provider-delete').hidden = false;
											break;
										case 8: // Provider Credential Delegate
											document.getElementById('provider-addroom').hidden = false;
											document.getElementById('provider-delete').hidden = false;
											break;
										case 9: // Campus
											document.getElementById('provider-delete').hidden = false;
											break;
										case 10: // Building
											document.getElementById('provider-delete').hidden = false;
											break;
									}
								} catch (exception) {
									// Do nothing
								}
							}
						} })
					),
					$.li({ id: 'provider-resize', class: 'horizontal-resize', css: { backgroundColor: 'var(--subheader-color)' } }),
					$.li({},
						$.div({ id: 'provider-tip', class: 'panel-tip', text: 'Select a provider to add new credentials to or select the Custom Room Groups folder to create a room group.' }),
						$.div({ id: 'provider-credentialstip', class: 'panel-tip', text: 'Click new credentials above the tree to enter new login credentials for this provider.' }),
						$.div({ id: 'provider-credentials' }),
						$.div({ id: 'provider-roomdetails' },
							$.div({ class: 'formgrid' },
								$.label({ for: 'provider-roomdetails-name', text: 'Name' }),
								$.input({ type: 'text', id: 'provider-roomdetails-name' }),
								$.label({ for: 'provider-roomdetails-replacementname', text: 'Replacement Name' }),
								$.input({ type: 'text', id: 'provider-roomdetails-replacementname' }),
								$.label({ for: 'provider-roomdetails-roomid', text: 'Room ID' }),
								$.input({ type: 'text', id: 'provider-roomdetails-roomid' }),
								$.div({ class: 'formgrid-buttonrow' },
									$.button({ id: 'provider-roomdetails-save', class: 'primary', text: 'Save' })
								)
							)
						),
						$.div({ id: 'provider-customroomgrouptip', class: 'panel-tip', text: 'Click create room group above the tree to create a new room group.' })
					)
				)
			),
			$.div({ id: 'configuration-sectioncapsfeeds', class: 'panel' }),
			$.div({ id: 'configuration-sectioncapsfilters', class: 'panel' }),
			//$.div({ id: 'configuration-sectioncapsalertmanagement', class: 'panel' },
			//	$.header({ text: 'CAPS Alert Management' })
			//)
			$.div({ id: 'configuration-sectionapi', class: 'panel' },
				$.header({ text: 'API' }),
				$.div({ css: { margin: '10px' }, text: `API Instructions: Copy the generated URL below to programatically change a player's schedule. This must be sent as a POST request.` }),
				$.ul({ id: 'api-accordion' },
					$.li({ class: 'accordion-header', text: 'Player Scheduling' }),
					$.li({}, ...[
						permissions['Scheduler'] && (permissions['Players Admin'] || permissions['Players Operator']) ?
							[$.div({ id: 'apiContainer', class: 'formgrid compact' })] :
							[$.div({ css: { margin: '10px' }, text: `Unfortunately you don't have the proper permissions to interact with this API. You will need both "Scheduler" and either "Players Admin" or "Players Operator" permissions enabled.` })]
					].flat()),
					$.li({ class: 'accordion-header', text: 'Additional APIs (Not Implemented Yet)' }),
					$.li({},
						$.ul({},
							$.div({ css: { margin: '10px' }, text: 'We are evaluating what additional APIs would be most beneficial to our users at this time. Thank you for your patience.' })
						)
					)
				)
			)
		)
	);
	document.body.classList.remove('loading');
	const searchProviderFilterInputChange = function() {
		this.parentNode.previousElementSibling.hidden = this.value != '';
		this.parentNode.previousElementSibling.previousElementSibling.hidden = this.value == '';
		if (this.value == '') {
			$providerTree.removeFilterView();
		} else {
			const searchString = this.value.toLowerCase();
			$providerTree.filterView(node => {
				return node.name.toLowerCase().indexOf(searchString) != -1;
			});
		}
	};
	document.getElementById('provider-toolbar').prepend(
		$.li({ title: 'Remove filter', class: 'tooltip-button mdi mdi-close-circle-outline removefilter', hidden: true, click: function() {
			this.hidden = true;
			this.nextElementSibling.hidden = false;
			this.nextElementSibling.nextElementSibling.querySelector('input').value = '';
			this.nextElementSibling.nextElementSibling.querySelector('input').dispatchEvent(new Event('change'));
			$providerTree.removeFilterView();
		} }),
		$.li({},
			$.label({ for: 'searchProviderFilter', title: 'Search', class: 'tooltip-button mdi mdi-magnify addfilter' })
		),
		$.li({ css: {
			flex: '1 0 auto',
			opacity: 1,
			margin: '0 5px',
			cursor: 'default'
		} },
			$.input({ type: 'text', id: 'searchProviderFilter', placeholder: 'Search', class: 'sampleInput', input: searchProviderFilterInputChange, change: searchProviderFilterInputChange })
		)
	);
	
	document.getElementById('configuration-enable').addEventListener('click', () => {
		for (const $font of document.querySelectorAll('#configuration-disabledfonts li.selected')) {
			var node = document.querySelector('#configuration-enabledfonts li:first-child');
			var inserted = false;
			while (node.length != 0) {
				if ($font.innerText.localeCompare(node.innerText) < 0) {
					inserted = true;
					node.insertAdjacentElement('beforebegin', $font);
					break;
				}
				node = node.nextElementSibling;
			}
			if (!inserted) {
				document.getElementById('configuration-enabledfonts').appendChild($font);
			}
			document.getElementById('configuration-enabledfonts').lastSelected = $font;
			document.getElementById('configuration-disabledfonts').lastSelected = null;
		}
	});
	document.getElementById('configuration-disable').addEventListener('click', () => {
		for (const $font of document.querySelectorAll('#configuration-enabledfonts li.selected')) {
			let $node = document.querySelector('#configuration-disabledfonts li:first-child');
			var inserted = false;
			while ($node) {
				if ($font.innerText.localeCompare($node.innerText) < 0) {
					inserted = true;
					$node.insertAdjacentElement('beforebegin', $font);
					break;
				}
				$node = $node.nextElementSibling;
			}
			if (!inserted) {
				document.getElementById('configuration-disabledfonts').appendChild($font);
			}
			document.getElementById('configuration-disabledfonts').lastSelected = $font;
			document.getElementById('configuration-enabledfonts').lastSelected = null;
		}
	});
	document.getElementById('configuration-savefonts').addEventListener('click', async () => {
		const fonts = [];
		for (const $font of document.querySelectorAll('#configuration-enabledfonts li')) {
			fonts.push({ name: $font.name, enabled: true, defaultfont: $font.defaultfont });
		}
		for (const $font of document.querySelectorAll('#configuration-disabledfonts li')) {
			fonts.push({ name: $font.name, enabled: false, defaultfont: false });
		}
		try {
			var data = await post('/post/updatefonts.php', {
				fonts: JSON.stringify(fonts)
			});
			CheckSessionExpired(data);
			if (data.success) {
				LoadFonts();
				ConfirmationBox('Updated', 'Updated fonts. Note that certain fonts and those used in playlists cannot be disabled.');
			} else {
				ConfirmationBox('Error updating', 'Could not update font settings.');
			}
		} catch (exception) {
			NetworkingError();
		}
	});

	for (const setting of configurationConfig) {
		if (setting.hasOwnProperty('group') && (!setting.hasOwnProperty('hidegroup') || !setting.hidegroup)) {
			if (setting.hasOwnProperty('support') && setting.support && !support) {
				continue;
			}
			const categoryName = setting.group.toLowerCase().replace(/[^A-Za-z]/g, '');
			document.getElementById('categories').appendChild($.li({ id: `configuration-${categoryName}`, class: 'button-choices', click: () => selectCategory(categoryName) },
				$.i({ class: `mdi ${setting.groupicon}` }),
				$.span({ text: setting.group })
			));
			document.getElementById('configuration-right').appendChild(
				$.div({ id: `configuration-section${categoryName}`, class: 'panel' },
					$.header({ text: setting.group }),
					$.div({ class: 'formgrid' })
				)
			);
		}
	}
	
	document.getElementById('configuration-progress').value = 0;
	document.getElementById('configuration-file').addEventListener('change', function() {
		var filename = baseName(this.value);
		document.getElementById('configuration-filename').innerText = filename;
	});
	document.getElementById('fonts-upload').addEventListener('click', async () => {
		const fontName = document.getElementById('configuration-fontname').value;
		const files = document.getElementById('configuration-file').files;
		if (!IsAlphaNumericSpacesHyphensUnderscoresApostrophesNotEmptyLessThan60Characters(fontName)) {
			ConfirmationBox('Invalid name', 'The font name can only container letters, numbers, spaces, hyphens, underscores, or apostrophes and must be between 1 and 60 characters.');
		} else if (files.length == 0) {
			ConfirmationBox('Select a file', 'Select a font file.');
		} else {
			try {
				const [askToOverwrite, fileName] = await new Promise((resolve, reject) => {
					const request = new XMLHttpRequest();
					request.upload.onprogress = e => {
						if (e.lengthComputable) {
							document.getElementById('configuration-progress').setAttribute('max', e.total);
							document.getElementById('configuration-progress').value = e.loaded;
						}
					};
					request.onload = function() {
						try {
							const data = JSON.parse(this.responseText);
							CheckSessionExpired(data);
							if (data.success) {
								ConfirmationBox('Font uploaded', 'Font uploaded.');
								resolve([false, '']);
							} else {
								if (data.hasOwnProperty('overwrite')) {
									resolve([true, data.overwrite]);
								} else {
									ConfirmationBox('Error', data.error);
									reject();
								}
							}
						} catch (exception) {
							reject();
						}
					};
					const error = () => {
						NetworkingError();
						reject();
					};
					request.onerror = error;
					request.upload.onerror = error;
					request.onabort = error;
					request.upload.onabort = error;
					request.open('post', '/post/customfontupload.php');
					const data = new FormData();
					data.append('fontname', fontName);
					data.append('file', files[0]);
					request.send(data);
				});
				
				if (askToOverwrite)
				{
					if (await QuestionBox('Overwrite?', 'A custom font with that name already exist. Do you want to overwrite?'))
					{
						try
						{
							var response = await post('/post/customfontupload.php',
							{
								overwrite: fileName,
								fontname: document.getElementById('configuration-fontname').value
							});
							CheckSessionExpired(response);
							if (response.success)
							{
								ConfirmationBox('Font overwritten', 'Font overwritten.');
							}
							else
							{
								ConfirmationBox('Error', response.error);
								throw new Error();
							}
						}
						catch (exception)
						{
							NetworkingError();
							throw new Error();
						}
					}
				}
				LoadFonts();
			}
			catch (exception)
			{
				console.log(exception);
				// Do nothing, message boxes already shown
			}
		}
	});
	document.getElementById('configuration-deletefont').addEventListener('click', async () => {
		if (document.getElementById('configuration-customfonts').value != '') {
			try {
				var data = await post('/post/deletecustomfont.php', {
					name: document.getElementById('configuration-customfonts').value
				});
				CheckSessionExpired(data);
				if (data.success) {
					ConfirmationBox('Deleted', 'Custom font deleted.');
					LoadFonts();
				} else {
					ConfirmationBox('Error deleting', data.error);
				}
			} catch (exception) {
				NetworkingError();
			}
		}
	});
	
	accordion(document.getElementById('idp-accordion'));
	accordion(document.getElementById('api-accordion'));
	
	DragLeftResize('#configuration-left-resize', 'configuration-lr');
	DragLeftResize('#provider-resize', 'configuration-providerleft');
	
	for (const provider of providers) {
		const id = `providerform-${provider.name.replace(/\s/g, '')}`;
		document.getElementById('provider-credentials').appendChild(
			$.div({ id: id, class: 'formgrid' })
		);
		provider.$form = document.getElementById(id);
		provider.$form.appendChild($.h1({ text: `${provider.display} Credential`, class: 'formgrid-row' }));
		for (const property of provider.credentialproperties) {
			if (property.hasOwnProperty('server')) {
				continue;
			}
			const propertyId = `${id}-${property.name}`;
			provider.$form.appendChild($.label({ for: propertyId, text: property.display, hidden: property.hasOwnProperty('hidden') }));
			property.$label = document.querySelector(`label[for="${propertyId}"]`);
			switch (property.type) {
				case 'bool':
					provider.$form.appendChild(
						$.uicheckbox({ id: propertyId })
					);
					break;
				case 'string':
					if (property.name == 'uploadtoken') {
						provider.$form.appendChild($.select({ type: 'text', id: propertyId, placeholder: 'Upload Token' }));
					} else {
						provider.$form.appendChild($.input({ type: property.hasOwnProperty('hidden') ? 'hidden' : property.hasOwnProperty('password') && property.password ? 'password' : 'text', id: propertyId }));
					}
					break;
				case 'int':
					if (property.name == 'readerboarddatarange') {
						provider.$form.appendChild(
							$.select({ type: 'text', id: propertyId },
								$.option({ text: '1 Month', value: 1 }),
								$.option({ text: '2 Months', value: 2 }),
								$.option({ text: '3 Months', value: 3 })
							)
						);
					} else {
						provider.$form.appendChild(
							$.input({ type: property.hasOwnProperty('hidden') ? 'hidden' : property.hasOwnProperty('password') && property.password ? 'password' : 'text', id: propertyId })
						);
					}
					break;
				case 'timezone':
					provider.$form.appendChild($.select({ id: propertyId },
						...phpTimezones.sort((a, b) => a.localeCompare(b)).map(timezone => $.option({ text: timezone, value: timezone })))
					);
					break;
				case 'eventlist':
					provider.$form.appendChild($.uitreeselect({ id: propertyId, config: {
						tree: 'event',
						leafIcon: node => 'schedule',
						filter: node => !node.group
					} }));
					break;
				case 'json':
					break;
			}
			property.$input = document.getElementById(propertyId);
			if (provider.name == 'librarymarket' && property.name == 'privatecalendar') {
				property.$input.addEventListener('change', function() {
					for (const element of ['providerform-librarymarket-clientid', 'providerform-librarymarket-clientsecret', 'providerform-librarymarket-password']) {
						toggleRow(document.getElementById(element), this.checked);
					}
				});
			}
		}
		const getCredentials = async () => {
			const credentials = {};
			for (const property of provider.credentialproperties) {
				if (property.hasOwnProperty('server')) {
					continue;
				}
				switch (property.type) {
					case 'bool':
						credentials[property.name] = property.$input.checked;
						break;
					case 'string':
						credentials[property.name] = property.$input.value;
						break;
					case 'int':
						credentials[property.name] = parseInt(property.$input.value, 10);
						break;
					case 'timezone':
						credentials[property.name] = property.$input.value;
						break;
					case 'eventlist':
						credentials[property.name] = (await property.$input.getSelected())[0]?.name ?? '';
						break;
					case 'json':
						break;
				}
			}
			return credentials;
		};
		provider.$form.appendChild($.div({ class: 'formgrid-buttonrow' },
			$.button({ title: 'Verify Credentials (Note: This might take a few seconds)', class: 'tooltip-button verifycredentials', click: async function() {
				try {
					const credentials = await getCredentials();
					try {
						LockBody();
						const selectedNodes = await $providerTree.getSelected();
						let data = await post('/post/providercheck.php', { provider: provider.name, credentials: JSON.stringify(credentials), id: selectedNodes.length == 0 || selectedNodes[0].type == 2 ? null : selectedNodes[0].id });
						UnlockBody();
						CheckSessionExpired(data);
						if (data.success) {
							ConfirmationBox('Valid credentials', 'The credentials are valid.');
						} else if (provider.name == 'google calendar' && data.hasOwnProperty('authurl')) {
							window.open(data.authurl, '_blank');
							await InputBox('Enter authorization token', 'On the window that opened (you might need to allow it with your pop up blocker) authorize the app then enter the authorization token below.', 'Ok', 'Cancel', '', async authCode => {
								if (authCode == '') {
									ConfirmationBox('Authorization token blank', 'You must supply a valid authorization token to allow this calendar to work.');
								} else {
									try {
										LockBody();
										credentials.authcode = authCode;
										data = await post('/post/providercheck.php', { provider: provider.name, credentials: JSON.stringify(credentials) });
										CheckSessionExpired(data);
										if (data.success) {
											ConfirmationBox('Valid credentials', 'The credentials are valid.');
											return true;
										} else {
											ConfirmationBox('Error', data.error);
										}
									} catch (exception) {
										NetworkingError();
									} finally {
										UnlockBody();
									}
								}
							}, () => true, null);
						} else {
							ConfirmationBox('Invalid credentials', data.error);
						}
					} catch (exception) {
						UnlockBody();
						NetworkingError();
					}
				} catch (exception) {
					ConfirmationBox('Error', exception.message);
				}
			} },
				$.i({ class: 'mdi mdi-shield-check-outline' }),
				'Verify Credentials'
			),
			$.button({ title: 'Attempt to find rooms from the provider and populate the tree with those rooms. Some providers do not have room lists so event data is scanned for possible rooms. You might need to manually add rooms.', class: 'tooltip-button retrieveroomlist', css: {
				marginLeft: '5px'
			}, click: async function() {
				try {
					LockBody();
					const data = await post('/post/providerroomlist.php', { id: (await $providerTree.getSelected())[0].id });
					UnlockBody();
					CheckSessionExpired(data);
					if (data.success) {
						if (data.hasOwnProperty('message')) {
							ConfirmationBox('Retrieved Rooms', data.message);
						} else {
							ConfirmationBox('Retrieved Rooms', 'Rooms merged into tree.');
						}
						await $providerTree.reload();
					} else {
						ConfirmationBox('Error', data.error);
					}
				} catch (exception) {
					UnlockBody();
					NetworkingError();
				}
			} },
				$.i({ class: 'mdi mdi-door' }),
				'Retrieve Room List'
			),
			$.button({ text: 'Save', class: 'primary', css: { marginLeft: '5px' }, click: async function() {
				try {
					const credentials = await getCredentials();
					const selectedNode = (await $providerTree.getSelected())[0];
					LockBody();
					if (selectedNode.type == 2) {
						// New Credential
						await $providerTree.addItems([{ type: 3, name: '', parent: selectedNode.id, data: credentials }]);
					} else {
						if (await $providerTree.updateItems(
						[
							{
								id: selectedNode.id,
								data: credentials
							}
						])) {
							ConfirmationBox('Credentials updated', 'The credentials were updated. It is recommended to verify the credentials and retrieve the room list to ensure they are up to date.');
						}
					}
					UnlockBody();
				} catch (exception) {
					ConfirmationBox('Error', exception.message);
				}
			} })
		));
		if (provider.name == 'microsoft office 365 exchange') {
			const uuidv4 = () => {
				return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
					var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
					return v.toString(16);
				});
			};
			provider.$form.append(
				$.div({ class: 'formgrid-row', id: 'microsoftgraphauthorized' }),
				$.div({ class: 'formgrid-buttonrow' },
					$.button({ id: 'microsoftgraphauthorize', text: 'Authorize Keywest Technology App Integration For Microsoft Graph and Retrieve Calendars', title: 'Click this and sign into the Microsoft 365 user (or calendar account) that will be used to access the Microsoft Graph API (which is used to access calendars.', css: {
						width: '500px',
						whiteSpace: 'pre-wrap'
					}, click: async () => {
						const username = document.getElementById('providerform-microsoftoffice365exchange-username').value;
						if (username == '') {
							ConfirmationBox('Username (Email) required', 'A username (email) is required. This can be the calendar email or an email that the calendar has delegated access to.');
							return;
						}
						const clientId = encodeURIComponent(microsoftGraphClientId);
						const redirectURL = encodeURIComponent('https://microsoftgraph.keywesttechnology.com/microsoftgraph.php');
						const resource = encodeURIComponent('https://graph.microsoft.com');
						const nonce = encodeURIComponent(uuidv4());
						const state = encodeURIComponent(`${host}/post/microsoftgraph.php`);
						window.open(`https://login.microsoftonline.com/common/oauth2/authorize?client_id=${clientId}&redirect_uri=${redirectURL}&resource=${resource}&response_type=id_token+code&scope=openid%20profile&response_mode=form_post&nonce=${nonce}&state=${state}&login_hint=${username}`, 'Keywest Technology Integration Microsoft Graph')
					} })
				),
				$.div({ id: 'microsoftgraphdirectionsstep1', class: 'formgrid-row', text: 'Enter a username and then press Save.' }),
				$.div({ id: 'microsoftgraphdirectionsstep2', class: 'formgrid-row', text: 'Press the authorize button above. All calendars the account has access to will be added to the tree.' }),
				$.div({ id: 'microsoftgraphdirectionsstep3', class: 'formgrid-row', text: 'Delete any calendars that you do not want to include then press "Retrieve Room List" to automatically add rooms for the calendars based on current event data. If you add new calendars to the account press the authorize button again to retrieve them.' })
			);
			/*
			provider.$form.appendChild($.div({ class: 'formgrid-buttonrow' },
				$.button({ text: 'Logout of current MS account and delete saved tokens', title: 'Click this to sign out and delete the saved access tokens. This can be useful if you are currently signed into the wrong account. The OAuth consent prompt might default to your current signed in user and will not let you switch until you logout.', click: async () => {
					const redirectURL = encodeURIComponent('https://microsoftgraph.keywesttechnology.com/microsoftgraph.php');
					window.open(`https://login.microsoftonline.com/common/oauth2/logout?post_logout_redirect_uri=${redirectURL}`, 'Keywest Technology Integration Logout');
					await post('/post/microsoftgraphlogout.php', {
					});
				} })
			));
			*/
		}
		if (provider.name == 'micros opera hotel edition') {
			provider.$form.append(
				$.button({ text: 'Copy Uploader Token', click: () => {
					const $tmp = document.body.appendChild($.input({ value: document.getElementById('providerform-microsoperahoteledition-uploadtoken').value }));
					$tmp.select();
					document.execCommand('copy');
					$tmp.remove();
				} }),
				$.p({ text: `Note: This provider hosts local event files that must be uploaded in order to be sent to players. The uploader can be downloaded from the following link:`, class: 'formgrid-row' }),
				$.a({ href: 'https://link.keywesttechnology.com/vcl7a', target: '_blank', text: 'Download MICROS Uploader Service', class: 'formgrid-row' }),
				$.p({ text: 'Unzip the file on the computer that has network access to the API then run the "MICROS Breeze Uploader.exe" and follow the instructions.', class: 'formgrid-row' })
			);
		}
		if (provider.name == 'ems platform services on premises') {
			provider.$form.append(
				$.button({ text: 'Copy Uploader Token', click: () => {
					const $tmp = document.body.appendChild($.input({ value: document.getElementById('providerform-emsplatformservicesonpremises-uploadtoken').value }));
					$tmp.select();
					document.execCommand('copy');
					$tmp.remove();
				} }),
				$.p({ text: `Note: This provider hosts local event files that must be uploaded in order to be sent to players. The uploader can be downloaded from the following link:`, class: 'formgrid-row' }),
				$.a({ href: 'https://link.keywesttechnology.com/8umfr', target: '_blank', text: 'Download EMS Platform Services Uploader Service', class: 'formgrid-row' }),
				$.p({ text: 'Unzip the file on the computer that has network access to the API then run the "EMS Platform Services Breeze Uploader.exe" and follow the instructions.', class: 'formgrid-row' })
			);
		}
	}
	HideProvider();
	
	document.getElementById('provider-addrooms').addEventListener('click', function() {
		this.querySelector('input').click();
	});
	
	document.querySelector('#categories > li').click();
});

function HideProvider() {
	document.getElementById('provider-tip').hidden = true;
	document.getElementById('provider-credentialstip').hidden = true;
	document.getElementById('provider-credentials').hidden = true;
	document.getElementById('provider-roomdetails').hidden = true;
	document.getElementById('provider-customroomgrouptip').hidden = true;
	
	document.getElementById('provider-delete').hidden = true;
	document.getElementById('provider-addgroup').hidden = true;
	document.getElementById('provider-addcredentials').hidden = true;
	document.getElementById('provider-addroom').hidden = true;
	document.getElementById('provider-adddelegate').hidden = true;
	document.getElementById('provider-addcampusbuilding').hidden = true;
	document.getElementById('provider-addrooms').hidden = true;
	
	for (const provider of providers) {
		provider.$form.hidden = true;
	}
}

function ShowProviderCustomRoomGroupsTip() {
	document.getElementById('provider-addgroup').hidden = false;
	document.getElementById('provider-customroomgrouptip').hidden = false;
}

/**
 * @return {string}
 */
function TokenGenerator() {
	var text = '';
	var possible = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
	for (var i = 0; i < 48; ++i) {
		text += possible.charAt(Math.floor(Math.random() * possible.length));
	}
	return text;
}

async function ShowProviderCredentials(node) {
	document.getElementById('provider-credentials').hidden = false;
	const provider = providers.find(provider => provider.display == node.parentNode.name);
	provider.$form.hidden = false;
	if (node.hasOwnProperty('id')) {
		document.getElementById('provider-delete').hidden = false;
		document.getElementById('provider-addroom').hidden = !(provider.name != 'microsoft office 365 exchange' && provider.name != 'ad astra');
		document.getElementById('provider-adddelegate').hidden = provider.name != 'microsoft exchange';
		document.getElementById('provider-addcampusbuilding').hidden = provider.name != 'ad astra';
		document.getElementById('provider-addrooms').hidden = !(provider.name != 'microsoft office 365 exchange' && provider.name != 'microsoft exchange' && provider.name != 'ad astra');
		document.getElementById('provider-addrooms').querySelector('input').replaceWith(
			$.input({ type: 'file', change: async function(e) {
				let selectedNode = (await $providerTree.getSelected())[0];
				const text = await new Promise(resolve => {
					const fileReader = new FileReader();
					fileReader.onload = e => resolve(e.target.result);
					fileReader.onabort = e => resolve('');
					fileReader.readAsText(e.target.files[0]);
				});
				const elements = text.split(/\r\n|\n/).map(line => line.split(',').map(value => value.trim()));
				var rooms = [];
				for (const values of elements) {
					if (values.length > 0 && values[0] != '') {
						rooms.push({
							type: 4,
							name: values.length > 1 && values[1] != '' ? `${values[0]} (${values[1]})` : values[0],
							parent: selectedNode.id,
							data: { name: values[0], replacement: values.length > 1 ? values[1] : '' }
						});
					}
				}
				LockBody();
				await $providerTree.addItems(rooms);
				UnlockBody();
			} })
		);
		provider.$form.querySelector('.retrieveroomlist').hidden = false;
	} else {
		provider.$form.querySelector('.retrieveroomlist').hidden = true;
	}
	const credentials = node.data;
	for (const property of provider.credentialproperties) {
		if (property.hasOwnProperty('server')) {
			continue;
		}
		if (property.hasOwnProperty('readonly') && node.hasOwnProperty('id')) {
			property.$input.readOnly = true;
		} else {
			property.$input.readOnly = false;
		}
		switch (property.type) {
			case 'bool':
				property.$input.checked = credentials[property.name];
				break;
			case 'string':
				if (provider.name == 'micros opera hotel edition' || provider.name == 'ems platform services on premises') {
					try {
						LockBody();
						const data = provider.name == 'micros opera hotel edition' ? await get('/get/microsuploadtokens.php') : await get('/get/emsplatformservicesuploadtokens.php');
						CheckSessionExpired(data);
						const newToken = TokenGenerator();
						data.uploadtokens.push(newToken);
						property.$input.innerHTML = '';
						property.$input.append(
							...data.uploadtokens.map(token => $.option({ text: token, value: token }))
						);
					} catch (exception) {
						NetworkingError();
					} finally {
						UnlockBody();
					}
				}
				property.$input.value = credentials[property.name];
				break;
			case 'int':
				property.$input.value = credentials[property.name];
				break;
			case 'timezone':
				property.$input.value = credentials[property.name];
				break;
			case 'eventlist':
				await property.$input.setSelected(node => node.name == credentials[property.name]);
				break;
			case 'json':
				break;
		}
		property.$input.dispatchEvent(new Event('change'));
	}
	
	if (provider.name == 'microsoft office 365 exchange') {
		const $step1 = document.getElementById('microsoftgraphdirectionsstep1');
		const $step2 = document.getElementById('microsoftgraphdirectionsstep2');
		const $step3 = document.getElementById('microsoftgraphdirectionsstep3');
		const $authorize = document.getElementById('microsoftgraphauthorize');
		const $authorized = document.getElementById('microsoftgraphauthorized');
		$step1.hidden = true;
		$step2.hidden = true;
		$step3.hidden = true;
		provider.$form.querySelector('.retrieveroomlist').hidden = true;
		provider.$form.querySelector('.verifycredentials').hidden = true;
		if (node.hasOwnProperty('id')) {
			$authorize.hidden = false;
			$authorized.hidden = false;
			$authorized.innerText = node.securesettingset.includes('accesstoken') ? 'App Authorized' : 'App Not Authorized. You can authorize the app by clicking the button below.';
			if (node.securesettingset.includes('accesstoken')) {
				$step3.hidden = false;
				provider.$form.querySelector('.retrieveroomlist').hidden = false;
			} else {
				$step2.hidden = false;
			}
		} else {
			$authorize.hidden = true;
			$authorized.hidden = true;
			$step1.hidden = false;
		}
	}
}

function ShowProviderRoom(room) {
	document.getElementById('provider-delete').hidden = false;
	document.getElementById('provider-roomdetails').hidden = false;
	document.getElementById('provider-roomdetails-name').value = room.data.name;
	document.getElementById('provider-roomdetails-replacementname').value = room.data.replacement;
	// Ad Astra
	for (const $e of document.querySelectorAll('#provider-roomdetails-name, #provider-roomdetails-roomid')) $e.readOnly = room.parentNode.type == 10;
	toggleRow(document.getElementById('provider-roomdetails-roomid'), room.parentNode.type == 10);
	if (room.parentNode.type == 10) {
		document.getElementById('provider-roomdetails-roomid').value = room.data.data.Id;
	}
	
	document.getElementById('provider-roomdetails-save').replaceWith(
		$.button({ id: 'provider-roomdetails-save', class: 'primary', text: 'Save', click: async function() {
			if (await $providerTree.updateItems([
				{
					id: room.id,
					data: { name: document.getElementById('provider-roomdetails-name').value, replacement: document.getElementById('provider-roomdetails-replacementname').value }
				}
			])) {
				ConfirmationBox('Room updated', 'The room was updated.');
			}
		} })
	);
}

async function ShowNewCredentials() {
	const selectedNode = (await $providerTree.getSelected())[0];
	if (selectedNode) {
		const provider = providers.find(provider => provider.display == selectedNode.name);
		
		const newCredential = {
			parentNode: selectedNode,
			data: {}
		};
		for (const property of provider.credentialproperties) {
			if (property.hasOwnProperty('server')) {
				continue;
			}
			switch (property.type) {
				case 'bool':
					newCredential.data[property.name] = property.hasOwnProperty('default') ? property.default : false;
					break;
				case 'string':
					if (provider.name == 'micros opera hotel edition') {
						try {
							LockBody();
							const data = await get('/get/microsuploadtokens.php');
							CheckSessionExpired(data);
							const newToken = TokenGenerator();
							data.uploadtokens.push(newToken);
							newCredential.data[property.name] = data.uploadtokens[0];
						} catch (exception) {
							NetworkingError();
						} finally {
							UnlockBody();
						}
					} else {
						newCredential.data[property.name] = property.hasOwnProperty('default') ? JSON.parse(JSON.stringify(property.default)) : '';
					}
					break;
				case 'int':
					newCredential.data[property.name] = property.hasOwnProperty('default') ? property.default : 0;
					break;
				case 'timezone':
					newCredential.data[property.name] = 'America/Chicago';
					break;
				case 'eventlist':
					newCredential.data[property.name] = null;
					break;
				case 'json':
					newCredential.data[property.name] = property.hasOwnProperty('default') ? JSON.parse(JSON.stringify(property.default)) : null;
					break;
			}
		}
		HideProvider();
		await ShowProviderCredentials(newCredential);
	}
}

async function ProviderCredentialsAddCalendar() {
	const selectedNode = (await $providerTree.getSelected())[0];
	await InputBox('Calendar Email', 'Enter an email for the user or room that contains the calendar this credential will access. (Note the user or room must delegate access (in their settings) to the credential for this to work).', 'Ok', 'Cancel', '', async email => {
		const data = await $providerTree.addItems([
			{
				type: 8,
				name: email,
				parent: selectedNode.id,
				data: { email: email }
			}
		]);
		if (data.success) {
			return true;
		}
	}, () => true, null);
}

async function ProviderCredentialsAddCampusBuilding() {
	const selectedNode = (await $providerTree.getSelected())[0];
	const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '750px' } },
		$.header({ text: 'Ad Astra - Add Campus and/or Building Rooms' }),
		$.uitree({ id: 'adastracampusbuildings', css: { height: '500px' }, config: {
			tree: 'adastracampusbuildings',
			src: async () => {
				const data = await get('/get/adastracampusbuildings.php', { id: selectedNode.id });
				CheckSessionExpired(data);
				if (data.success) {
					return data.tree;
				} else {
					ConfirmationBox('Error', data.error);
					return [];
				}
			},
			checkbox: true,
			groupTypes: [1],
			leafIcon: node => 'building'
		} }),
		$.div({ class: 'panel-buttons' },
			$.button({ text: 'Cancel', click: function() {
				$dialog.remove();
			} }),
			$.button({ text: 'OK', click: async function() {
				const campusAndBuildings = await document.getElementById('adastracampusbuildings').getCheckedAllChildren();
				if (campusAndBuildings.length == 0) {
					ConfirmationBox('Nothing checked', 'Please check a campus or building to pull data for.');
				} else {
					// Ensure that a building's campus is included in case it isn't already in the provider tree
					for (let i = 0; i < campusAndBuildings.length; ++i) {
						if (campusAndBuildings[i].type == 2 && !campusAndBuildings.includes(campusAndBuildings[i].parentNode)) {
							campusAndBuildings.push(campusAndBuildings[i].parentNode);
						}
					}
					// Ensure campuses are processed before buildings
					campusAndBuildings.sort((a, b) => a.type - b.type);
					const data = await $providerTree.addItems(campusAndBuildings.map(campusOrBuilding => {
						if (campusOrBuilding.type == 1) {
							// campus
							const campus = campusOrBuilding;
							return {
								type: 9,
								name: campus.name,
								parent: selectedNode.id,
								data: campus.data
							};
						} else {
							// building
							const building = campusOrBuilding;
							return {
								type: 10,
								name: building.name,
								data: building.data
							};
						}
					}));
					if (data.success) {
						// Do nothing
					} else {
						// Do nothing
					}
					$dialog.remove();
				}
			} })
		)
	));
	centerOnScreen($dialog);
}

async function ProviderCredentialsAddRoom() {
	const selectedNode = (await $providerTree.getSelected())[0];
	await InputBox('Room Name', 'Enter a name for the room.', 'Ok', 'Cancel', '', async name => {
		const data = await $providerTree.addItems(
		[
			{
				type: 4,
				name: name,
				parent: selectedNode.id,
				data: { name: name, replacement: '' }
			}
		]);
		if (data.success) {
			return true;
		}
	}, () => true, null);
}

async function ProviderAddCustomRoomGroup() {
	const selectedNode = (await $providerTree.getSelected())[0];
	await InputBox('Custom Room Group Name', 'Enter a name for the custom room group.', 'Ok', 'Cancel', '', async name => {
		const data = await $providerTree.addItems(
		[
			{
				type: 6,
				name: name,
				parent: selectedNode.id,
				data: {}
			}
		]);
		if (data.success) {
			return true;
		}
	}, () => true, null);
}

function CheckProviderChange()
{
	
}