let $csvTree;

function HideCSVFile() {
	document.getElementById('csvuploader-tip').hidden = true;
	document.getElementById('csvuploader-newcsvfile').hidden = true;
}

function ShowCSVFileTip() {
	HideCSVFile();
	document.getElementById('csvuploader-header').innerText = '';
	document.getElementById('csvuploader-tip').hidden = false;
}

async function ShowNewCSVFile() {
	let selectedNode = (await $csvTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Select a group', 'Select a group to add the new CSV file to.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		$csvTree.removeEventListener('change', CSVChangeSelection);
		await $csvTree.setSelected([selectedNode.id]);
		$csvTree.addEventListener('change', CSVChangeSelection);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	HideCSVFile();
	document.getElementById('csvuploader-name').value = '';
	document.getElementById('csvuploader-file').value = null;
	document.getElementById('csvuploader-file').dispatchEvent(new Event('change'));
	document.getElementById('csvuploader-progress').setAttribute('max', 1);
	document.getElementById('csvuploader-progress').value = 0;
	document.getElementById('csvuploader-header').innerText = 'New';
	document.getElementById('csvuploader-newcsvfile').hidden = false;
}

async function CSVChangeSelection() {
	const selectedNode = (await this.getSelected())[0];
	if (selectedNode == undefined) {
		ShowCSVFileTip();
	} else {
		if (selectedNode.group) {
			ShowCSVFileTip();
		} else {
			ShowCSVFileTip();
		}
	}
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'csvuploader-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'CSV Files' }),
				$.ul({ class: 'iconlist' },
					$.li({ id: 'csvTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $csvTree.deleteSelected(), title: 'Delete selected CSV file' }),
					$.li({ class: 'spacer' }),
					$.li({ id: 'csvTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $csvTree.addGroup(), title: 'Add group' }),
					$.li({ id: 'csvTree-add', class: 'tooltip-button mdi mdi-playlist-plus', click: ShowNewCSVFile, title: 'New CSV file' }),
					$.li({ id: 'csvTree-duplicate', class: 'tooltip-button mdi mdi-content-duplicate', click: () => $csvTree.duplicateSelected(), title: 'Duplicate selected CSV file' }),
				),
				$csvTree = $.uitree({ id: 'csvTree', class: 'treeContainer', config: {
					tree: 'csv',
					dispatchChangeOnLoad: true,
					editable: node => node.group && node.parentNode != null,
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length == 0) {
							ConfirmationBox('Selection required', 'Select an CSV file to delete.');
							return;
						}
						const selectedNode = selectedNodes[0];
						if (selectedNode.group) {
							var csvCount = selectedAllChildren.filter(node => !node.group).length;
							if (csvCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${csvCount} CSV file${csvCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove CSV file?", `Are you sure you want to remove the CSV file '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				}, load: async () => {
					if ((await $csvTree.getAllNodes()).some(node => node.group)) {
						for (const $e of document.querySelectorAll('#csvTree-addgroup, #csvTree-add, #csvTree-delete, #csvTree-duplicate')) $e.hidden = false;
					}
				}, change: CSVChangeSelection })
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'csvuploader-right', class: 'panel' },
			$.header({ id: 'csvuploader-header' }),
			$.span({ id: 'csvuploader-tip', class: 'panel-tip', text: 'Select a folder and then select the icon for "New CSV file", to upload to that folder.' }),
			$.div({ id: 'csvuploader-newcsvfile' },
				$.ul({ css: { listStyleType: 'square', margin: '20px 30px', color: 'var(--tree-font-color)' } },
					$.li({ text: 'CSV files store comma-separated values. CSV files uploaded here can be used in any widget that supports CSV files, like the CSV roll widget.' }),
					$.li({ text: 'Enter a name for the uploaded CSV file and select a valid .CSV file and then click upload.' }),
					$.li({ html: 'To export a CSV file from an Excel spreadsheet open the Excel spreadsheet and then choose File -> Save As and in the drop down for the extension <em>carefully</em> choose either: <em>CSV (Comma delimited) (*.csv)</em> or <em>CSV (MS-DOS) (*.csv)</em>.  There is a third option that <strong>will NOT work</strong> called <em>CSV (Macintosh) (*.csv)</em> - again, do <strong>NOT</strong> use the Macintosh option.' })
				),
				$.div({ class: 'formgrid', css: { width: '450px' } },
					$.label({ text: 'CSV Name' }),
					$.input({ type: 'text', id: 'csvuploader-name', name: 'name', class: 'sampleInput' }),
						
					$.label({ text: 'File' }),
					$.label({ class: 'button' },
						$.span({ id: 'csvuploader-filename', css: { textOverflow: 'ellipsis' }, text: 'Choose File' }),
						$.input({ type: 'file', id: 'csvuploader-file', name: 'file', css: { display: 'none' } })
					),
						
					$.button({ id: 'csvuploader-submit', class: 'primary formgrid-label', text: 'Upload' }),
					$.progress({ id: 'csvuploader-progress' })
				)
			),
			$.div({ id: 'csvuploader-csvfile' })
		)
	);
	document.body.classList.remove('loading');
	ShowCSVFileTip();
	document.getElementById('csvuploader-progress').setAttribute('max', 1);
	document.getElementById('csvuploader-progress').value = 0;
	document.getElementById('csvuploader-file').addEventListener('change', function() {
		const filename = baseName(this.value);
		document.getElementById('csvuploader-filename').innerText = filename == '' ? 'Select file...' : filename;
	});
	document.getElementById('csvuploader-submit').addEventListener('click', async () => {
		const selectedNode = await $csvTree.getSelected();
		var name = document.getElementById('csvuploader-name').value;
		if (!IsAlphaNumericSpacesHyphensUnderscoresApostrophesPeriodsNotEmptyLessThan40Characters(name)) {
			ConfirmationBox('Invalid name', 'The csv name can only container letters, numbers, spaces, hyphens, underscores, apostrophes, and periods and must be between 1 and 40 characters.');
		} else if (document.getElementById('csvuploader-file').files.length == 0) {
			ConfirmationBox('Select a file', 'Select a CSV file to upload.');
		} else {
			const upload = async (overwrite, id = -1) => {
				document.getElementById('csvuploader-progress').setAttribute('max', 1);
				document.getElementById('csvuploader-progress').value = 0;
				const data = await $csvTree.addItems([{
					type: 2,
					id,
					name: name,
					parent: selectedNode[0].id,
					file: document.getElementById('csvuploader-file').files[0],
					overwrite: overwrite,
					progress: (node, loaded, total) => {
						document.getElementById('csvuploader-progress').setAttribute('max', total);
						document.getElementById('csvuploader-progress').value = loaded;
					}
				}]);
				if (data.success) {
					ConfirmationBox('CSV file uploading', 'CSV file uploading. This may take a moment. To use this csv file in a playlist drag in a CSV Roll widget in the editor.');
				} else if (data.errors.length != 0 && data.errors[0].hasOwnProperty('overwrite')) {
					if (await QuestionBox('Overwrite?', 'A CSV file with that name already exist. Do you want to overwrite?')) {
						upload(true, data.errors[0].id);
					}
				}
			};
			upload(false);
		}
	});
	for (const $e of document.querySelectorAll('#csvTree-addgroup, #csvTree-add, #csvTree-delete, #csvTree-duplicate')) $e.hidden = true;
	
	DragLeftResize('.horizontal-resize', 'csv-lr');
});