class DOM {
	/**
	 * @param {!string} tag
	 * @param {!Object} [options]
	 * @param {!string} [options.id]
	 * @param {!string} [options.className]
	 * @param {!Object} [options.style]
	 * @returns {!Element}
	 */
	static el(tag, options = {}) {
		const el = document.createElement(tag);
		
		// For ease of use allow .class to alias .className
		if (options.class != null && options.className == null) {
			options.className = options.class;
		}

		// Assign options.style explicitly as it is converted to a string otherwise.
		Object.assign(el, options);
		Object.assign(el.style, options.style);

		return el;
	}

	/**
	 * Returns a DOM-escaped version of text.
	 * @param {!string} text
	 * @returns {!string}
	 */
	static escape(text) {
		const el = document.createElement('div');
		el.appendChild(document.createTextNode(text));
		return el.innerHTML;
	}

	/**
	 * Tag template function for safely generating escaped HTML.
	 * @param {!string[]} strings
	 * @param {!...string} expressions
	 * @returns {!string}
	 */
	static parseHTML(strings, ...expressions) {
		if (expressions.length === 0) {
			return strings.join('');
		}
		expressions = expressions.map(DOM.escape);
		let index = 0;
		return strings
			.map(string => `${string}${index < expressions.length ? expressions[index++] : ''}`)
			.join('');
	}

	/**
	 * Tag template function for safely generating escaped HTML.
	 * @param {!string[]} strings
	 * @param {!...string} expressions
	 * @returns {!DocumentFragment}
	 */
	static html(strings, ...expressions) {
		const el = document.createElement('template');
		el.innerHTML = DOM.parseHTML(strings, ...expressions);
		return el.content;
	}
}
