let $eventTree;

function HideEventList() {
	document.getElementById('events-tip').hidden = true;
	document.getElementById('events-view').hidden = true;
}

function ShowEventListTip() {
	HideEventList();
	document.getElementById('events-tip').hidden = false;
	document.getElementById('events-header').innerText = '';
}

function ShowEventList(node) {
	HideEventList();
	document.getElementById('events-view').hidden = false;
	document.getElementById('events-header').innerText = node.name;
	
	const events = JSON.parse(node.data).events;
	let currentFilter = 'none';
	let sortColumn = 'start';
	let sortColumnOrder = 'asc';
	const toggleSort = function(column) {
		return () => {
			if (sortColumn == 'none' || sortColumn != column) {
				sortColumn = column;
				sortColumnOrder = 'asc';
			} else if (sortColumnOrder == 'asc') {
				sortColumnOrder = 'desc';
			} else {
				sortColumn = 'start';
				sortColumnOrder = 'asc';
			}
			render();
		};
	};
	const render = () => {
		document.querySelector('.filter-view').innerHTML = '';
		document.querySelector('.filter-view').append(
			...[
				{ type: 'none', display: 'Show All' },
				{ type: 'day', display: `Today's Events` },
				{ type: 'week', display: `This Week's Events` },
				{ type: 'month', display: `This Month's Events` }
			].map(({type, display}) => $.button({ class: [currentFilter == type, 'selected'], text: display, click: function() {
				document.querySelector('.filter-view .selected').classList.remove('selected');
				this.classList.add('selected');
				currentFilter = type;
				render();
			} }))
		);
		const $events = document.getElementById('eventEntryContainer');
		$events.innerHTML = '';
		$events.appendChild(
			$.header({},
				$.label({},
					$.span({ text: 'Group (Room)' }),
					$.i({ class: ['mdi', 'tooltip-button', sortColumn != 'group', 'mdi-sort', sortColumn == 'group' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'group' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('group') })
				),
				$.label({},
					$.span({ text: 'Event' }),
					$.i({ class: ['mdi', 'mdi-sort', 'tooltip-button', sortColumn != 'name', 'mdi-sort', sortColumn == 'name' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'name' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('name') })
				),
				$.label({ css: { gridColumn: 'span 2' } },
					$.span({ text: 'Start' }),
					$.i({ class: ['mdi', 'mdi-sort', 'tooltip-button', sortColumn != 'start', 'mdi-sort', sortColumn == 'start' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'start' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('start') })
				),
				$.label({ css: { gridColumn: 'span 2' } },
					$.span({ text: 'End' }),
					$.i({ class: ['mdi', 'mdi-sort', 'tooltip-button', sortColumn != 'end', 'mdi-sort', sortColumn == 'end' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'end' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('end') })
				),
				$.label({ text: 'Repeat' }),
				$.i({ class: "tooltip-button mdi mdi-plus-circle-outline", title: "Add Event Entry", click: () => {
					currentFilter = 'none';
					sortColumn = 'start';
					sortColumnOrder = 'asc';
					events.push({
						group: '',
						name: '',
						start: moment().tz(configuration.timezone).unix(),
						end: moment().tz(configuration.timezone).add(1, 'hour').unix(),
						sunday: false,
						monday: false,
						tuesday: false,
						wednesday: false,
						thursday: false,
						friday: false,
						saturday: false
					});
					render();
				} })
			)
		);
		events.sort(function(a, b) {
			switch (sortColumn) {
				case 'group':
				case 'name':
					return a[sortColumn].localeCompare(b[sortColumn]) * (sortColumnOrder == 'asc' ? 1 : -1);
				case 'start':
				case 'end':
					return (moment.utc(a[sortColumn] * 1000).tz(configuration.timezone).isBefore(moment.utc(b[sortColumn] * 1000).tz(configuration.timezone)) ? -1 : 1) * (sortColumnOrder == 'asc' ? 1 : -1);
			}
		});
		for (const event of events) {
			if (currentFilter != 'none') {
				var start = moment().startOf(currentFilter);
				var end = moment().endOf(currentFilter);
				var continuous = !event.sunday && !event.monday && !event.tuesday && !event.wednesday && !event.thursday && !event.friday && !event.saturday;
				if (!(moment.utc(event.start * 1000).tz(configuration.timezone).isSameOrBefore(end) &&
				      moment.utc(event.end * 1000).tz(configuration.timezone).isSameOrAfter(start) &&
				      (continuous || currentFilter != 'day' || event[moment().format('dddd').toLowerCase()]))) {
					continue;
				}
			}
			const index = events.indexOf(event);
			const groupInputChange = function() { event.group = this.value; };
			const nameInputChange = function() { event.name = this.value; };
			const startDateInputChange = function() { event.start = moment.tz(`${this.value} ${moment.utc(event.start * 1000).tz(configuration.timezone).format('HH:mm')}`, 'YYYY-MM-DD HH:mm', configuration.timezone).unix(); };
			const startTimeInputChange = function() { event.start = moment.tz(`${moment.utc(event.start * 1000).tz(configuration.timezone).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm', configuration.timezone).unix(); };
			const endDateInputChange = function() { event.end = moment.tz(`${this.value} ${moment.utc(event.end * 1000).tz(configuration.timezone).format('HH:mm')}`, 'YYYY-MM-DD HH:mm', configuration.timezone).unix(); };
			const endTimeInputChange = function() { event.end = moment.tz(`${moment.utc(event.end * 1000).tz(configuration.timezone).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm', configuration.timezone).unix(); };
			$events.append(
				$.input({ type: 'text', class: 'sampleInput', value: event.group, input: groupInputChange, change: groupInputChange }),
				$.input({ type: 'text', class: 'sampleInput', value: event.name, input: nameInputChange, change: nameInputChange }),
				$.input({ type: 'date', class: 'sampleInput', value: moment.utc(event.start * 1000).tz(configuration.timezone).format('YYYY-MM-DD'), input: startDateInputChange, change: startDateInputChange }),
				$.input({ type: 'time', class: 'sampleInput', value: moment.utc(event.start * 1000).tz(configuration.timezone).format('HH:mm'), input: startTimeInputChange, change: startTimeInputChange }),
				$.input({ type: 'date', class: 'sampleInput', value: moment.utc(event.end * 1000).tz(configuration.timezone).format('YYYY-MM-DD'), input: endDateInputChange, change: endDateInputChange }),
				$.input({ type: 'time', class: 'sampleInput', value: moment.utc(event.end * 1000).tz(configuration.timezone).format('HH:mm'), input: endTimeInputChange, change: endTimeInputChange }),
				$.div({ class: 'checkbox-group' },
					...['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'].map(day =>
						$.uicheckbox({ class: 'formgrid-compact-checkbox', checked: event[day.toLowerCase()], text: day.slice(0, 1), title: `${event[day.toLowerCase()] ? 'Disable' : 'Enable'} dayparting for ${day}`, change: function() { event[day.toLowerCase()] = this.checked; render(); } })
					)
				),
				$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: function() {
					events.splice(index, 1);
					render();
				} })
			);
		}
	};
	render();
	
	document.getElementById('submit').replaceWith(
		$.button({ id: 'submit', class: 'primary', text: 'Update Event List', click: async () => {
			if (await ValidateEventList(events)) {
				LockBody();
				if (await $eventTree.updateItems([
					{
						id: node.id,
						events: events
					}
				])) {
					UnlockBody();
					ConfirmationBox('Event list updated', 'The event list was updated.');
				} else {
					UnlockBody();
				}
			}
		} })
	);
}

async function ShowNewEventList() {
	let selectedNode = (await $eventTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Select a group', 'Select a group to add the new event list to.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $eventTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	await InputBox('Event List Name', 'Enter a name for the event list.', 'Ok', 'Cancel', '', async name => {
		const data = await $eventTree.addItems(
		[
			{
				type: 2,
				name: name,
				parent: selectedNode.id,
				events: [{
					group: '',
					name: '',
					start: moment().tz(configuration.timezone).unix(),
					end: moment().tz(configuration.timezone).add(1, 'hour').unix(),
					sunday: false,
					monday: false,
					tuesday: false,
					wednesday: false,
					thursday: false,
					friday: false,
					saturday: false
				}]
			}
		]);
		if (data.success) {
			await $eventTree.setSelected([data.added[0].id], false, false, true);
			return true;
		}
	}, () => true, null);
}

async function ValidateEventList(events) {
	for (var eventItr = 0; eventItr < events.length; ++eventItr) {
		var event = events[eventItr];
		if (event.name == '') {
			await ConfirmationBox('Invalid event name', 'The event fields in an Event Entry cannot be empty.');
			return false;
		}
	}
	return true;
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'events-left', class: 'panel' },
			$.header({ text: 'Events' }),
			$.ul({ class: 'iconlist' },
				$.li({ id: 'eventTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $eventTree.deleteSelected(), title: 'Delete selected' }),
				$.li({ class: 'spacer' }),
				$.li({ id: 'eventTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $eventTree.addGroup(), title: 'Add group' }),
				$.li({ id: 'eventTree-add', class: 'tooltip-button mdi mdi-playlist-plus', click: ShowNewEventList, title: 'New event list' }),
				$.li({ id: 'eventTree-duplicate', class: 'tooltip-button mdi mdi-content-duplicate', click: () => $eventTree.duplicateSelected(), title: 'Duplicate selected event list' })
			),
			$eventTree = $.uitree({ id: 'eventTree', class: 'treeContainer', config: {
				tree: 'event',
				dispatchChangeOnLoad: true,
				editable: node => node.group && node.parentNode != null,
				leafIcon: node => 'schedule',
				deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
					if (selectedNodes.length == 0) {
						ConfirmationBox('Selection required', 'Select an event list to delete.');
						return;
					}
					const selectedNode = selectedNodes[0];
					if (selectedNode.group) {
						var eventListCount = selectedAllChildren.filter(node => !node.group).length;
						if (eventListCount != 0) {
							await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${eventListCount} event list${eventListCount == 1 ? '' : 's'}?`, 'Yes', 'No');
						}
					} else {
						await ConfirmationBox("Remove event list?", `Are you sure you want to remove the event list '${selectedNode.name}'?`, 'Yes', 'No');
					}
					await deleteNodes(selectedAllChildren);
				}
			}, load: async () => {
				if ((await $eventTree.getAllNodes()).some(node => node.group)) {
					for (const $e of document.querySelectorAll('#eventTree-addgroup, #eventTree-add, #eventTree-delete, #eventTree-duplicate')) $e.hidden = false;
				}
			} })
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'events-right', class: 'panel' },
			$.header({ id: 'events-header' }),
			$.span({ id: 'events-tip', class: 'panel-tip', text: `Select an event list to edit or select a folder then "New event list', to create a new event list.` }),
			$.div({ id: 'events-view' },
				$.div({ class: 'filter-view' }),
				$.input({ type: 'hidden', id: 'eventURL', class: 'sampleInput' }),
				$.div({ id: 'eventEntryContainer', class: 'formgrid compact' }),
				$.div({ class: 'panel-buttons' },
					$.button({ id: 'submit', class: 'primary', text: 'Update Event List' })
				)
			)
		)
	);
	document.body.classList.remove('loading');
	
	ShowEventListTip();
	for (const $e of document.querySelectorAll('#eventTree-addgroup, #eventTree-add, #eventTree-delete, #eventTree-duplicate')) $e.hidden = true;
	$eventTree.addEventListener('change', async function() {
		const selectedNode = (await this.getSelected())[0];
		if (selectedNode == undefined) {
			ShowEventListTip();
		} else {
			if (selectedNode.group) {
				ShowEventListTip();
			} else {
				ShowEventList(selectedNode);
			}
		}
	});
	
	DragLeftResize('.horizontal-resize', 'events-lr');
});
