var FontLoader = 
{
	bus: Minibus.create(),
	loadedFonts: [],
	loadingFonts: [],
	loadFonts: function(fonts, done, fail)
	{
		if (fonts.length === 0)
		{
			if (done)
			{
				done(fonts);
			}
		}
		else
		{
			var _loadedFonts = [],
				_failedFonts = [];
			var finishedLoading = function()
			{
				if (_loadedFonts.length + _failedFonts.length === fonts.length)
				{
					if (_failedFonts.length > 0)
					{
						if (fail)
						{
							fail(fonts, {loadedFonts: _loadedFonts, failedFonts: _failedFonts});
						}
					}
					else
					{
						if (done)
						{
							done(fonts);
						}
					}
				}
			};
			var fontLoadSucceed = function(familyName)
			{
				_loadedFonts.push(familyName);
				finishedLoading();
			};
			var fontLoadFail = function(familyName)
			{
				_failedFonts.push(familyName);
				finishedLoading();
			};
			for (const font of fonts)
			{
				this.loadFont(font.value, fontLoadSucceed, fontLoadFail);
			}
		}
	},
	loadFont: function(font, done, fail)
	{
		//console.log("loadFont: " + font);
		if (this.loadedFonts.indexOf(font) !== -1)
		{
			if (done)
			{
				//console.log(font + " already loaded.  Calling done.");
				done(font);
			}
		}
		else if (this.loadingFonts.indexOf(font) !== -1)
		{
			//console.log(font + " already loading, waiting for ready.");
			var doneRoute = null,
				failRoute = null;
			if (done)
			{
				doneRoute = this.bus.on('FontLoadSucceed', familyName =>
				{
					//console.log("FontLoadSucceed emitted for: " + familyName);
					if (familyName === font)
					{
						//console.log("FontLoadSucceed matches font: " + font + ".  Calling done.");
						this.bus.off(doneRoute);
						if (failRoute !== null)
						{
							this.bus.off(failRoute);
						}
						done(font);
					}
				});
			}
			if (fail)
			{
				failRoute = this.bus.on('FontLoadFail', familyName =>
				{
					console.log("FontLoadFail emitted for: " + familyName);
					if (familyName === font)
					{
						console.log("FontLoadFail matches font: " + font + ".  Calling fail.");
						this.bus.off(failRoute);
						if (doneRoute !== null)
						{
							this.bus.off(doneRoute);
						}
						fail(font);
					}
				});
			}
		}
		else
		{
			//console.log("Preparing to load font " + font);
			this.loadingFonts.push(font);
			var ffo = new FontFaceObserver(font);
			ffo.load(null, 60000).then(function()
			{
				// There is a bug where rarely a font isn't ready when the load event is called but will be ready very shortly
				// thereafter.  Adding a small amount of delay should help prevent this from happening.
				return new Promise(function(resolve, reject)
				{
					setTimeout(function()
					{
						resolve();
					}, 100);
				});
			}).then(() =>
			{
				//console.log("Attempting to remove `font` (" + font + ") from `this.loadingFonts`...");
				var index = this.loadingFonts.indexOf(font);
				if (index !== -1)
				{
					this.loadingFonts.splice(index, 1);
					//console.log("Successfully removed `font` (" + font + ") from `this.loadingFonts`.");
				}
				this.loadedFonts.push(font);
				//console.log("Font load finished for: " + font + ".  Calling done.");
				if (done)
				{
					done(font);
				}
				this.bus.emit('FontLoadSucceed', font);
			})
			.catch(e =>
			{
				console.log("Font '" + font + "' timed out during load.");
				var index = this.loadingFonts.indexOf(font);
				console.log("Attempting to remove `font` (" + font + ") from `this.loadingFonts`...");
				if (index !== -1)
				{
					this.loadingFonts.splice(index, 1);
					console.log("Successfully removed `font` (" + font + ") from `this.loadingFonts`...");
				}
				console.log("Font load failed for: " + font + ".  Calling fail.");
				if (fail)
				{
					fail(font);
				}
				this.bus.emit('FontLoadFail', font);
			});
		}
	}
};
