let currentSection = 'EAS';

async function refreshEas() {
	const $eas = document.getElementById('eas');
	$eas.removeEventListener('click', refreshEas);
	currentSection = 'EAS';
	try {
		const response = await fetch('/get/emergencyevents.php', {
			credentials: 'same-origin'
		});
		const data = await response.json();
		const events = data.emergencyalerts;
		// Since network requests may take a while or timeout, ensure
		// the user is still on the EAS logs before proceeding.
		if (currentSection !== 'EAS') {
			return;
		}
		const $logsContent = document.getElementById('logs-content');
		$logsContent.innerHTML = '';
		$logsContent.append(
			$.div({ class: 'header' },
				$.span({ class: 'eas-index', text: 'Index' }),
				$.span({ class: 'eas-date', text: 'Alert Date' }),
				$.span({ class: 'eas-title', text: 'Alert Title' }),
				$.span({ class: 'eas-description', text: 'Alert Description' }),
				$.span({ class: 'eas-source', text: 'Feed' })
			),
			...events.map(alert => $.div({ class: 'row' },
				$.span({ class: 'eas-index', text: alert.emergencyalertsid }),
				$.span({ class: 'eas-date', text: moment.utc(alert.date).tz(configuration.timezone).format('YYYY-MM-DD h:mm:ssa') }),
				$.span({ class: 'eas-title', text: alert.title }),
				$.span({ class: 'eas-description', text: alert.description }),
				$.span({ class: 'eas-feed', text: alert.feed })
			))
		);
	} catch (err) {
		NetworkingError();
	} finally {
		$eas.addEventListener('click', refreshEas);
	}
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'logs-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'Logs' }),
				$.ul({ id: 'logs' },
					$.li({ id: 'eas', class: 'button-choices' },
						$.i({ class: 'mdi mdi-alarm-light-outline' }),
						$.span({ text: 'Emergency Alerts' })
					)
				)
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'logs-right' },
			$.div({ class: 'panel', css: { flex: '1 1 0', minHeight: 0 } },
				$.header({ id: 'logs-windowheader', text: 'Logs' }),
				$.div({ id: 'logs-content' })
			)
		)
	);
	document.body.classList.remove('loading');
	DragLeftResize('.horizontal-resize', 'logs-lr');
	refreshEas();
	document.getElementById('eas').addEventListener('click', refreshEas);
});
