var networkCommandPlayerTransfer = null;
var networkCommandPlayerLastSelected = null;

let $networkCommandsTree;

async function AddNetworkCommand() {
	await InputBox('Network Command Name', 'Enter a name for the network command.', 'Ok', 'Cancel', '', async name => {
		const data = await $networkCommandsTree.addItems([
			{
				type: 2,
				name: name,
				parent: 1
			}
		]);
		if (data.success) {
			await $networkCommandsTree.setSelected([data.added[0].id], false, false, true);
			return true;
		}
	}, () => true, null);
}

function GetNetworkCommandPlayerData() {
	var data = {
		protocol: document.getElementById('networkcommands-playerform-protocol').value
	};
	
	switch (data.protocol) {
		case 'none':
			break;
		case 'rs232':
			data.port = document.querySelector('.networkcommands-playerform-rs232.port').value;
			data.baudrate = document.querySelector('.networkcommands-playerform-rs232.baudrate').value;
			data.databits = document.querySelector('.networkcommands-playerform-rs232.databits').value;
			data.parity = document.querySelector('.networkcommands-playerform-rs232.parity').value;
			data.stopbits = document.querySelector('.networkcommands-playerform-rs232.stopbits').value;
			data.flowcontrol = document.querySelector('.networkcommands-playerform-rs232.flowcontrol').value;
			data.commandformat = document.querySelector('.networkcommands-playerform-rs232.commandformat').value;
			data.command = document.querySelector('.networkcommands-playerform-rs232.command').value.replaceAll('\\n', String.fromCharCode(10)).replaceAll('\\r', String.fromCharCode(13));
			break;
		case 'udp':
			data.ip = document.querySelector('.networkcommands-playerform-udp.ip').value;
			data.port = document.querySelector('.networkcommands-playerform-udp.port').value;
			data.commandformat = document.querySelector('.networkcommands-playerform-udp.commandformat').value;
			data.command = document.querySelector('.networkcommands-playerform-udp.command').value.replaceAll('\\n', String.fromCharCode(10)).replaceAll('\\r', String.fromCharCode(13));
			break;
		case 'tcp':
			data.ip = document.querySelector('.networkcommands-playerform-tcp.ip').value;
			data.port = document.querySelector('.networkcommands-playerform-tcp.port').value;
			data.commandformat = document.querySelector('.networkcommands-playerform-tcp.commandformat').value;
			data.command = document.querySelector('.networkcommands-playerform-tcp.command').value.replaceAll('\\n', String.fromCharCode(10)).replaceAll('\\r', String.fromCharCode(13));
			break;
	}
	return data;
}

async function UpdateNetworkCommandPlayer(node) {
	const data = GetNetworkCommandPlayerData();
	LockBody();
	await $networkCommandsTree.updateItems([
		{
			id: node.id,
			data: JSON.stringify(data)
		}
	]);
	UnlockBody();
	await SetWindowHeaderSaveState();
}

async function LoadNetworkCommandTemplates() {
	try {
		LockBody();
		var data = await get('/get/networkcommandtemplates.php');
		UnlockBody();
		CheckSessionExpired(data);
		if (data.success) {
			document.getElementById('templates-udp').innerHTML = '';
			document.getElementById('templates-tcp').innerHTML = '';
			for (const template of data.templates) {
				template.data = JSON.parse(template.data);
			}
			const createTemplate = template => {
				const overwriteSettings = async () => {
					var selected = (await $networkCommandsTree.getSelected())[0];
					if (selected == undefined || selected.type != 4) {
						ConfirmationBox(`Select a network command setting for a player`, `Select a network command setting for a player.`);
						return;
					}
					if (await QuestionBox(`Overwrite network command player settings?`, `Are you sure you want to overwrite the current network command player settings with this template? You must select update to save these settings for the player.`)) {
						SetNetworkCommandPlayerData(template.data);
						await SetWindowHeaderSaveState();
					}
				};
				return $.div({ id: template.id, title: 'Click to overwrite current network command player settings with this template', class: 'button-choices' },
					$.i({ class: 'mdi mdi-console-network-outline', click: overwriteSettings }),
					$.span({ text: template.name, click: overwriteSettings }),
					$.i({ title: 'Delete', class: 'mdi mdi-trash-can-outline', click: async () => {
						if (await QuestionBox("Remove network command template?", `Are you sure you want to remove the network command template '${template.name}'?`)) {
							try {
								var data = await post('/post/deletenetworkcommandtemplate.php',
								{
									id: template.id
								});
								CheckSessionExpired(data);
								if (data.success) {
									await LoadNetworkCommandTemplates();
									ConfirmationBox('Network command template deleted', 'Network command template deleted.');
								} else {
									ConfirmationBox('Error deleting network command template', 'Could not delete network command template.');
								}
							} catch (exception) {
								NetworkingError();
							}
						}
					} })
				);
			};
			document.getElementById('templates-rs232').innerHTML = '';
			document.getElementById('templates-rs232').append(...data.templates.filter(template => template.data.protocol == 'rs232').map(template => createTemplate(template)));
			
			document.getElementById('templates-udp').innerHTML = '';
			document.getElementById('templates-udp').append(...data.templates.filter(template => template.data.protocol == 'udp').map(template => createTemplate(template)));
			
			document.getElementById('templates-tcp').innerHTML = '';
			document.getElementById('templates-tcp').append(...data.templates.filter(template => template.data.protocol == 'tcp').map(template => createTemplate(template)));
		} else {
			ConfirmationBox('Error adding network command template', 'Could not add network command template.');
		}
	} catch (exception) {
		NetworkingError();
	}
}

async function AddNetworkCommandTemplate() {
	const selected = await $networkCommandsTree.getSelected();
	if (selected.length == 0 || selected[0].type != 4) {
		ConfirmationBox('Select a network command setting for a player', 'Select a network command setting for a player.');
		return;
	}
	if (document.getElementById('networkcommands-playerform-protocol').value == 'none') {
		ConfirmationBox('The protocol for this command is none', 'The selected network command setting has no protocol defined. Templates must define a protocol.');
		return;
	}
	await InputBox('Network Command Template', 'Enter a name for the network command template.', 'Ok', 'Cancel', '', async name => {
		try {
			var data = await post('/post/addnetworkcommandtemplate.php', {
				name: name,
				data: JSON.stringify(GetNetworkCommandPlayerData())
			});
			CheckSessionExpired(data);
			if (data.success) {
				ConfirmationBox('Network command template created', 'Network command template created.');
				await LoadNetworkCommandTemplates();
				return true;
			} else {
				ConfirmationBox('Error adding network command template', data.error);
			}
		} catch (exception) {
			NetworkingError();
		}
	}, () => true, null);
}

function SetNetworkCommandPlayerData(data) {
	for (const $e of document.querySelectorAll('#networkcommands-playerform > .formgrid > *')) $e.hidden = true;
	toggleRow(document.getElementById('networkcommands-playerform-protocol'), true);
	document.getElementById('networkcommands-playerform-protocol').value = data.protocol;
	switch (data.protocol) {
		case 'none':
			break;
		case 'rs232':
			for (const $element of document.querySelectorAll('label + .networkcommands-playerform-rs232')) {
				toggleRow($element, true);
			}
			document.querySelector('.networkcommands-playerform-rs232.port').value = data.port;
			document.querySelector('.networkcommands-playerform-rs232.baudrate').value = data.baudrate;
			document.querySelector('.networkcommands-playerform-rs232.databits').value = data.databits;
			document.querySelector('.networkcommands-playerform-rs232.parity').value = data.parity;
			document.querySelector('.networkcommands-playerform-rs232.stopbits').value = data.stopbits;
			document.querySelector('.networkcommands-playerform-rs232.flowcontrol').value = data.flowcontrol;
			document.querySelector('.networkcommands-playerform-rs232.commandformat').value = data.commandformat;
			document.querySelector('.networkcommands-playerform-rs232.commandformat').dispatchEvent(new Event('change'));
			document.querySelector('.networkcommands-playerform-rs232.command').value = data.command.replace(String.fromCharCode(10), '\\n').replace(String.fromCharCode(13), '\\r');
			break;
		case 'udp':
			for (const $element of document.querySelectorAll('label + .networkcommands-playerform-udp')) {
				toggleRow($element, true);
			}
			document.querySelector('.networkcommands-playerform-udp.ip').value = data.ip;
			document.querySelector('.networkcommands-playerform-udp.port').value = data.port;
			document.querySelector('.networkcommands-playerform-udp.commandformat').value = data.commandformat;
			document.querySelector('.networkcommands-playerform-udp.commandformat').dispatchEvent(new Event('change'));
			document.querySelector('.networkcommands-playerform-udp.command').value = data.command.replace(String.fromCharCode(10), '\\n').replace(String.fromCharCode(13), '\\r');
			break;
		case 'tcp':
			for (const $element of document.querySelectorAll('label + .networkcommands-playerform-tcp')) {
				toggleRow($element, true);
			}
			document.querySelector('.networkcommands-playerform-tcp.ip').value = data.ip;
			document.querySelector('.networkcommands-playerform-tcp.port').value = data.port;
			document.querySelector('.networkcommands-playerform-tcp.commandformat').value = data.commandformat;
			document.querySelector('.networkcommands-playerform-tcp.commandformat').dispatchEvent(new Event('change'));
			document.querySelector('.networkcommands-playerform-tcp.command').value = data.command.replace(String.fromCharCode(10), '\\n').replace(String.fromCharCode(13), '\\r');
			break;
	}
}

async function TransferNetworkCommandPlayer() {
	const selected = (await $networkCommandsTree.getSelected())[0];
	if (selected == undefined) {
		return;
	}
	try {
		await CheckNetworkCommandChange();
		var fromPlayerId = selected.playerid;
		var fromPlayerName = selected.name;
		var fromPlayerDeletable = selected.type == 5;
		await ConfirmationBox("Select a player to transfer to", "Select a player in the tree to transfer all the network command settings to.", 'OK', 'Cancel');
		document.getElementById('networkcommands-transfernetworkcommandplayer').hidden = true;
		document.getElementById('networkcommands-canceltransfernetworkcommandplayer').hidden = false;
		networkCommandPlayerTransfer = async (toPlayerId, toPlayerName) => {
			if (toPlayerId == fromPlayerId) {
				ConfirmationBox('Select a different player', 'You cannot transfer settings from a player to itself. Select a player to transfer the settings to or select cancel.');
				return;
			}
			const transferNetworkCommands = async deleteFrom => {
				$networkCommandsTree.lock();
				networkCommandPlayerLastSelected = null;
				try {
					const data = await post('/post/transfernetworkcommandplayer.php', {
						syncid: $networkCommandsTree.syncId,
						fromid: fromPlayerId,
						toid: toPlayerId,
						delete: deleteFrom
					});
					CheckSessionExpired(data);
					if (data.success) {
						$networkCommandsTree.syncId = data.syncid;
						ConfirmationBox('Network command settings transferred', 'Network command settings for player transferred.');
						await $networkCommandsTree.reload();
						$networkCommandsTree.unlock();
						networkCommandPlayerLastSelected = (await $networkCommandsTree.getSelected())[0];
						await ShowNetworkCommandPlayer(networkCommandPlayerLastSelected);
					} else {
						if ('treeoutdated' in data) {
							TreeOutOfDate('networkcommand');
							await $networkCommandsTree.reload();
							$networkCommandsTree.unlock()
						} else {
							ConfirmationBox('Error transferring network command settings for the player', 'Could not delete the network command settings for the player.');
							$networkCommandsTree.unlock();
						}
					}
				} catch (exception) {
					NetworkingError();
					$networkCommandsTree.unlock();
				}
			};
			if (fromPlayerDeletable) {
				switch (await ConfirmationBox("Transfer all configuration settings?", `Are you sure you want to transfer all the network command settings from '${fromPlayerName}' to '${toPlayerName}'?`, `Yes And Then Delete '${fromPlayerName}'`, 'Yes', 'No')) {
					case 0:
						await transferNetworkCommands(true);
						break;
					case 1:
						await transferNetworkCommands(false);
						break;
					case 2:
						break;
				}
			} else {
				if (await QuestionBox("Transfer all configuration settings?", `Are you sure you want to transfer all the network command settings from '${fromPlayerName}' to '${toPlayerName}'?`)) {
					await transferNetworkCommands(false);
				}
			}
			networkCommandPlayerTransfer = null;
			$networkCommandsTree.dispatchEvent(new Event('change'));
		};
	} catch (exception) {
		CancelTransferNetworkCommandPlayer();
	}
}

async function TestNetworkCommand() {
	const selectedNode = (await $networkCommandsTree.getSelected())[0];
	try {
		await CheckNetworkCommandChange();
		try {
			var data = await post('/post/testnetworkcommand.php',
			selectedNode.type == 2 ? { networkcommandid: selectedNode.networkcommandid } : { networkcommandplayerid: selectedNode.networkcommandplayerid });
			CheckSessionExpired(data);
			if (data.success) {
				ConfirmationBox('Network command sent', 'The network command was sent.');
			} else {
				ConfirmationBox('Error sending network command', 'Could not send the network command.');
			}
		} catch (exception) {
			NetworkingError();
		}
	} catch (exception) {
		// Do nothing
	}
}

function CancelTransferNetworkCommandPlayer() {
	networkCommandPlayerTransfer = null;
	$networkCommandsTree.dispatchEvent(new Event('change'));
}

async function NetworkCommandChanged() {
	const selected = (await $networkCommandsTree.getSelected())[0];
	if (selected == undefined) {
		return false;
	}
	if (selected.type == 1 || selected.type == 2 || selected.type == 3) {
		return false;
	}
	return JSON.stringify(selected.data) != JSON.stringify(GetNetworkCommandPlayerData());
}

async function SetWindowHeader(str) {
	document.querySelector('#networkcommands-windowheader > span:first-child').innerText = str;
	await SetWindowHeaderSaveState();
}

async function SetWindowHeaderSaveState() {
	document.querySelector('#networkcommands-windowheader > span:last-child').innerText = await NetworkCommandChanged() ? '*' : '';
}

async function CheckNetworkCommandChange() {
	return new Promise(async (resolve, reject) => {
		if (networkCommandPlayerLastSelected != null) {
			if (JSON.stringify(networkCommandPlayerLastSelected.data) != JSON.stringify(GetNetworkCommandPlayerData())) {
				try {
					switch (await ConfirmationBox('Save changes?', 'You have unsaved changes. Do you want to save first?', 'Yes', 'No', 'Cancel')) {
						case 0:
							await UpdateNetworkCommandPlayer(networkCommandPlayerLastSelected);
							resolve();
							break;
						case 1:
							SetNetworkCommandPlayerData(networkCommandPlayerLastSelected.data);
							resolve();
							break;
					}
				} catch (exception) {
					await $networkCommandsTree.setSelected([networkCommandPlayerLastSelected.id]);
					reject();
				}
			} else {
				resolve();
			}
		} else {
			resolve();
		}
	});
}

function HideAll() {
	document.getElementById('networkcommands-deletenetworkcommand').hidden = true;
	document.getElementById('networkcommands-deletenetworkcommandplayer').hidden = true;
	document.getElementById('networkcommands-addnetworkcommand').hidden = true;
	document.getElementById('networkcommands-transfernetworkcommandplayer').hidden = true;
	document.getElementById('networkcommands-canceltransfernetworkcommandplayer').hidden = true;
	document.getElementById('networkcommands-test').hidden = true;
	
	document.querySelector('#networkcommands-windowheader > span:first-child').innerText = '';
	document.querySelector('#networkcommands-windowheader > span:last-child').innerText = '';
	document.getElementById('networkcommands-tip').hidden = true;
	document.getElementById('networkcommands-form').hidden = true;
	document.getElementById('networkcommands-playerform').hidden = true;
	for (const $e of document.querySelectorAll('#networkcommands-playerform > .formgrid > *')) $e.hidden = true;
	document.getElementById('networkcommands-deletenetworkcommand').hidden = true;
	document.getElementById('networkcommands-deletenetworkcommandplayer').hidden = true;
}

function ShowNetworkCommandTip() {
	HideAll();
	document.getElementById('networkcommands-tip').hidden = false;
}

async function ShowNetworkCommand(node) {
	HideAll();
	await SetWindowHeader(node.name);
	document.getElementById('networkcommands-deletenetworkcommand').hidden = false;
	document.getElementById('networkcommands-test').hidden = false;
}

async function ShowNetworkCommandPlayer(node) {
	HideAll();
	for (const $e of document.querySelectorAll('#networkcommands-playerform input, #networkcommands-playerform textarea')) $e.readOnly = node.type == 5;
	for (const $e of document.querySelectorAll('#networkcommands-playerform select')) $e.disabled = node.type == 5;
	document.getElementById('networkcommands-deletenetworkcommandplayer').hidden = node.type != 5;
	document.getElementById('networkcommands-update').parentNode.hidden = node.type != 4;
	document.getElementById('networkcommands-playerform').hidden = false;
	toggleRow(document.getElementById('networkcommands-playerform-protocol'), true);
	SetNetworkCommandPlayerData(node.data);
	await SetWindowHeader('Player Command Settings - ' + node.name);
}

async function RefreshNetworkCommandsTree() {
	$networkCommandsTree.reload();
	$networkCommandsTree.dispatchEvent(new Event('change'));
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'networkcommands-left', class: 'panel' },
			$.header({ text: 'RS-232/UDP/TCP Commands' }),
			$.ul({ id: 'networkcommands-toolbar', class: 'iconlist', css: { flex: '0 0 auto' } },
				$.li({ id: 'networkcommands-deletenetworkcommand', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $networkCommandsTree.deleteSelected(), title: 'Delete selected Network Command' }),
				$.li({ id: 'networkcommands-deletenetworkcommandplayer', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $networkCommandsTree.deleteSelected(), title: 'Delete selected Network Command Player' }),
				$.li({ class: 'spacer' }),
				$.li({ id: 'networkcommands-addnetworkcommand', class: 'tooltip-button mdi mdi-plus-circle-outline', click: AddNetworkCommand, title: 'Add Network Command' }),
				$.li({ id: 'networkcommands-transfernetworkcommandplayer', class: 'tooltip-button mdi mdi-transfer-right', click: TransferNetworkCommandPlayer, title: 'Transfer all player settings to another player. This makes the two players identical' }),
				$.li({ id: 'networkcommands-canceltransfernetworkcommandplayer', class: 'tooltip-button mdi mdi-close-circle-outline', click: CancelTransferNetworkCommandPlayer, title: 'Cancel player settings transfer' }),
				$.li({ id: 'networkcommands-test', class: 'tooltip-button mdi mdi-shield-check-outline', click: TestNetworkCommand, title: 'Test the command for all players or the selected player' }),
				$.li({ class: 'tooltip-button mdi mdi-reload', click: RefreshNetworkCommandsTree, title: 'Refresh' })
			),
			$networkCommandsTree = $.uitree({ id: 'networkCommandsTree', class: 'treeContainer', config: {
				tree: 'networkcommand',
				dispatchChangeOnLoad: true,
				groupTypes: [1, 2, 3],
				groupIcon: node => node.type == 2 ? 'networkcommand' : '',
				leafIcon: node => node.type == 5 ? 'networkcommandplayerdeleted' : 'networkcommandplayer',
				editable: node => node.type == 2,
				filterDraggable: node => false,
				deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
					if (selectedNodes.length == 0) {
						ConfirmationBox('Selection required', 'Select a network command or deleted player to delete.');
						return;
					}
					var selectedNode = selectedNodes[0];
					if (selectedNode.type == 2) {
						await ConfirmationBox("Remove network command?", `Are you sure you want to remove the network command '${selectedNode.name}'? This will delete all the associated player settings for this command.`, 'Yes', 'No');
					} else {
						await ConfirmationBox("Remove network command player settings?", `Are you sure you want to remove the network command player settings for '${selectedNode.name}'?`, 'Yes', 'No');
					}
					await deleteNodes(selectedAllChildren);
				}
			}, change: async function() {
				const selectedNode = (await this.getSelected())[0];
				if (selectedNode == undefined) {
					HideAll();
					ShowNetworkCommandTip();
					networkCommandPlayerTransfer = null;
				} else {
					try {
						await CheckNetworkCommandChange();
						HideAll();
						if (networkCommandPlayerTransfer != null && selectedNode.type != 4) {
							CancelTransferNetworkCommandPlayer();
						}
						switch (selectedNode.type) {
							case 1: // Network Commands Root
								ShowNetworkCommandTip();
								document.getElementById('networkcommands-addnetworkcommand').hidden = false;
								break;
							case 2: // Network Command
								await ShowNetworkCommand(selectedNode);
								networkCommandPlayerLastSelected = null;
								break;
							case 3: // Player Folder
								networkCommandPlayerLastSelected = null;
								break;
							case 4: // Player
								networkCommandPlayerLastSelected = selectedNode;
								await ShowNetworkCommandPlayer(selectedNode);
								document.getElementById('networkcommands-test').hidden = false;
								document.getElementById('networkcommands-transfernetworkcommandplayer').hidden = networkCommandPlayerTransfer != null;
								document.getElementById('networkcommands-canceltransfernetworkcommandplayer').hidden = networkCommandPlayerTransfer == null;
								if (networkCommandPlayerTransfer != null) {
									await networkCommandPlayerTransfer(selectedNode.playerid, selectedNode.name);
								}
								break;
							case 5: // Deleted Player
								networkCommandPlayerLastSelected = selectedNode;
								await ShowNetworkCommandPlayer(selectedNode);
								document.getElementById('networkcommands-transfernetworkcommandplayer').hidden = false;
								break;
						}
					} catch (exception) {
						// Do nothing
					}
				}
			} })
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'networkcommands-center' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ id: 'networkcommands-windowheader' },
					$.span({ text: 'RS-232/UDP/TCP Commands' }), 
					$.span()
				),
				$.span({ id: 'networkcommands-tip', class: 'panel-tip', text: 'Create a network command or select an existing player to edit its network command. Players must be registered before they appear in the tree for each network command.' }),
				$.div({ id: 'networkcommands-form' }),
				$.div({ id: 'networkcommands-playerform' },
					$.div({ class: 'formgrid', css: { flex: '1 0 auto' } },
						$.label({ text: 'Protocol' }),
						$.select({ id: 'networkcommands-playerform-protocol', class: 'sampleSelect' },
							$.option({ value: 'none', text: 'None' }),
							$.option({ value: 'rs232', selected: true, text: 'RS-232' }),
							$.option({ value: 'udp', text: 'UDP' }),
							$.option({ value: 'tcp', text: 'TCP' }),
						),
						// RS-232
						$.label({ class: 'tooltip-button', title: 'The serial port of the connected device. This is OS dependent and some standard conventions are provided, but there are many serial port naming conventions.', text: 'Port' }),
						$.input({ type: 'text', class: 'networkcommands-playerform-rs232 port' }),
						
						$.label({ class: 'tooltip-button', title: `The baud rate for the device. 9600 bits per second is normal, but this could be different and defined in the documentation of the device you're connecting to.`, text: 'Baud Rate (bps)' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 baudrate' },
							$.option({ value: '1200', text: '1200' }),
							$.option({ value: '2400', text: '2400' }),
							$.option({ value: '4800', text: '4800' }),
							$.option({ value: '9600', selected: true, text: '9600' }),
							$.option({ value: '19200', text: '19200' }),
							$.option({ value: '38400', text: '38400' }),
							$.option({ value: '57600', text: '57600' }),
							$.option({ value: '115200', text: '115200' }),
						),
						
						$.label({ class: 'tooltip-button', title: 'The number of data bits. For the vast majority of applications this is 8', text: 'Data Bits' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 databits' },
							$.option({ value: '5', text: '5' }),
							$.option({ value: '6', text: '6' }),
							$.option({ value: '7', text: '7' }),
							$.option({ value: '8', selected: true, text: '8' }),
							$.option({ value: '9', text: '9' }),
						),
						
						$.label({ class: 'tooltip-button', title: 'Whether parity bits should be used and the convention. Default is no parity bits', text: 'Parity' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 parity' },
							$.option({ value: 'none', selected: true, text: 'None' }),
							$.option({ value: 'odd', text: 'Odd' }),
							$.option({ value: 'even', text: 'Even' }),
						),
						
						$.label({ class: 'tooltip-button', title: 'The number of stop bits. Default is 1', text: 'Stop Bits' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 stopbits' },
							$.option({ value: '1', selected: true, text: '1' }),
							$.option({ value: '1.5', text: '1.5' }),
							$.option({ value: '2', text: '2' }),
						),
						
						$.label({ class: 'tooltip-button', title: 'Whether RS232 should have flow control enabled. Default is false', text: 'Flow Control' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 flowcontrol' },
							$.option({ value: 'true', text: 'True' }),
							$.option({ value: 'false', selected: true, text: 'False' }),
						),
						
						$.label({ class: 'tooltip-button', title: `If you need to send regular text you'd use ASCII. Hex is useful for binary data formats`, text: 'Command Format' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-rs232 commandformat' },
							$.option({ value: 'hex', selected: true, text: 'Hex' }),
							$.option({ value: 'ascii', text: 'ASCII' }),
						),
						
						$.label({ text: 'Command' }),
						$.span({ class: 'networkcommands-playerform-rs232 command-container' },
							$.textarea({ class: 'sampleInput networkcommands-playerform-rs232 command' }),
							$.div({},
								$.button({  class: 'tooltip networkcommands-playerform-rs232 ascii-options carriagereturn', title: 'Add carriage return', text: 'Add \\r' }),
								$.button({  class: 'tooltip networkcommands-playerform-rs232 ascii-options newline', title: 'Add new line', text: 'Add \\n' }),
								$.button({  class: 'tooltip networkcommands-playerform-rs232 ascii-options carriagereturn newline', title: 'Add carriage return and new line', text: 'Add \\r\\n' }),
							)
						),
						
						$.label({ text: 'IP' }),
						$.input({ type: 'text', class: 'sampleInput networkcommands-playerform-udp ip', placeholder: '127.0.0.1' }),
						
						$.label({ text: 'Port' }),
						$.input({ type: 'text', class: 'sampleInput networkcommands-playerform-udp port', placeholder: '65535' }),
						
						$.label({ class: 'tooltip-button', title: `If you need to send regular text you'd use ASCII. Hex is useful for binary data formats`, text: 'Command Format' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-udp commandformat' },
							$.option({ value: 'hex', selected: true, text: 'Hex' }),
							$.option({ value: 'ascii', text: 'ASCII' }),
						),
						
						$.label({ text: 'Command' }),
						$.span({ class: 'networkcommands-playerform-udp command-container' },
							$.textarea({ class: 'sampleInput networkcommands-playerform-udp command' }),
							$.div({},
								$.button({  class: 'tooltip networkcommands-playerform-udp ascii-options carriagereturn', title: 'Add carriage return', text: 'Add \\r' }),
								$.button({  class: 'tooltip networkcommands-playerform-udp ascii-options newline', title: 'Add new line', text: 'Add \\n' }),
								$.button({  class: 'tooltip networkcommands-playerform-udp ascii-options carriagereturn newline', title: 'Add carriage return and new line', text: 'Add \\r\\n' })
							)
						),
						// TCP
						$.label({ text: 'IP' }),
						$.input({ type: 'text', class: 'sampleInput networkcommands-playerform-tcp ip', placeholder: '127.0.0.1' }),
						
						$.label({ text: 'Port' }),
						$.input({ type: 'text', class: 'sampleInput networkcommands-playerform-tcp port', placeholder: '65535' }),
						
						$.label({ class: 'tooltip-button', title: `If you need to send regular text you'd use ASCII. Hex is useful for binary data formats`, text: 'Command Format' }),
						$.select({ class: 'sampleSelect networkcommands-playerform-tcp commandformat' },
							$.option({ value: 'hex', selected: true, text: 'Hex' }),
							$.option({ value: 'ascii', text: 'ASCII' }),
						),
						
						$.label({ text: 'Command' }),
						$.span({ class: 'networkcommands-playerform-tcp command-container' },
							$.textarea({ class: 'sampleInput networkcommands-playerform-tcp command' }),
							$.div({},
								$.button({  class: 'tooltip networkcommands-playerform-tcp ascii-options carriagereturn', title: 'Add carriage return', text: 'Add \\r' }),
								$.button({  class: 'tooltip networkcommands-playerform-tcp ascii-options newline', title: 'Add new line', text: 'Add \\n' }),
								$.button({  class: 'tooltip networkcommands-playerform-tcp ascii-options carriagereturn newline', title: 'Add carriage return and new line', text: 'Add \\r\\n' })
							)
						)
					),
					$.div({ class: 'panel-buttons' },
						$.button({  id: 'networkcommands-update', class: 'primary wait formgrid-label', text: 'Update' })
					)
				)
			)
		),
		$.div({ id: 'networkcommands-right' },
			$.div({ css: { flex: '1 1 0', minHeight: 0, display: 'flex', flexFlow: 'column nowrap' } },
				$.div({ class: 'panel', css: { flex: '0 0 auto' } },
					$.header({ text: 'Command Templates' }),
					$.div({ id: 'networkcommands-addnetworkcommandtemplate', class: 'tooltip-button button-choices', click: AddNetworkCommandTemplate, title: 'Save the selected network command settings for this player as a template' },
						$.i({ class: 'mdi mdi-plus-circle-outline' }),
						$.span({ text: 'Save Current Command As Template' })
					)
				),
				$.div({ class: 'panel', css: { flex: '1 0 0', minHeight: 0 } },
					$.header({ class: 'tooltip-button', css: { cursor: 'help' }, title: 'Selecting an RS232 Command below will load it into the current view', text: 'RS232' }),
					$.div({ id: 'templates-rs232' })
				),
				$.div({ class: 'panel', css: { flex: '1 0 0', minHeight: 0 } },
					$.header({ class: 'tooltip-button', css: { cursor: 'help' }, title: 'Selecting a UDP Command below will load it into the current view', text: 'UDP' }),
					$.div({ id: 'templates-udp' })
				),
				$.div({ class: 'panel', css: { flex: '1 0 0', minHeight:0 } },
					$.header({ class: 'tooltip-button', css: { cursor: 'help' }, title: 'Selecting a TCP Command below will load it into the current view', text: 'TCP' }),
					$.div({ id: 'templates-tcp' })
				)
			)
		)
	);
	document.body.classList.remove('loading');
	ShowNetworkCommandTip();
	
	LoadNetworkCommandTemplates();
	
	document.querySelector('.networkcommands-playerform-rs232.port').replaceWith(
		$.uiinputselect({ class: 'networkcommands-playerform-rs232 port', config: {
			options: (() => {
				const ports = [];
				for (var i = 0; i < 10; ++i) {
					ports.push({ name: 'ttyS' + i, value: 'ttyS' + i });
				}
				for (var i = 0; i < 10; ++i) {
					ports.push({ name: 'ttyUSB' + i, value: 'ttyUSB' + i });
				}
				for (var i = 0; i < 10; ++i) {
					ports.push({ name: 'ttyACM' + i, value: 'ttyACM' + i });
				}
				for (var i = 0; i < 10; ++i) {
					ports.push({ name: 'COM' + i, value: 'COM' + i });
				}
				return ports;
			})(),
			allowInput: true
		} })
	);

	document.getElementById('networkcommands-playerform-protocol').addEventListener('change', function() {
		for (const $e of document.querySelectorAll('#networkcommands-playerform > .formgrid > *')) $e.hidden = true;
		toggleRow(this, true);
		document.getElementById('networkcommands-update').hidden = false;
		switch (this.value) {
			case 'rs232':
				for (const $element of document.querySelectorAll('label + .networkcommands-playerform-rs232')) {
					toggleRow($element, true);
				}
				document.querySelector('.networkcommands-playerform-rs232.port').value = '';
				document.querySelector('.networkcommands-playerform-rs232.baudrate').value = '9600';
				document.querySelector('.networkcommands-playerform-rs232.databits').value = '8';
				document.querySelector('.networkcommands-playerform-rs232.parity').value = 'none';
				document.querySelector('.networkcommands-playerform-rs232.stopbits').value = '1';
				document.querySelector('.networkcommands-playerform-rs232.flowcontrol').value = 'false';
				document.querySelector('.networkcommands-playerform-rs232.commandformat').value = 'hex';
				document.querySelector('.networkcommands-playerform-rs232.commandformat').dispatchEvent(new Event('change'));
				document.querySelector('.networkcommands-playerform-rs232.command').value = '';
				break;
			case 'udp':
				for (const $element of document.querySelectorAll('label + .networkcommands-playerform-udp')) {
					toggleRow($element, true);
				}
				document.querySelector('.networkcommands-playerform-udp.ip').value = '';
				document.querySelector('.networkcommands-playerform-udp.port').value = '';
				document.querySelector('.networkcommands-playerform-udp.commandformat').value = 'hex';
				document.querySelector('.networkcommands-playerform-udp.commandformat').dispatchEvent(new Event('change'));
				document.querySelector('.networkcommands-playerform-udp.command').value = '';
				break;
			case 'tcp':
				for (const $element of document.querySelectorAll('label + .networkcommands-playerform-tcp')) {
					toggleRow($element, true);
				}
				document.querySelector('.networkcommands-playerform-tcp.ip').value = '';
				document.querySelector('.networkcommands-playerform-tcp.port').value = '';
				document.querySelector('.networkcommands-playerform-tcp.commandformat').value = 'hex';
				document.querySelector('.networkcommands-playerform-tcp.commandformat').dispatchEvent(new Event('change'));
				document.querySelector('.networkcommands-playerform-tcp.command').value = '';
				break;
		}
	});
	for (const $e of document.querySelectorAll('#networkcommands-playerform-protocol, .networkcommands-playerform-rs232, .networkcommands-playerform-udp, .networkcommands-playerform-tcp')) {
		$e.addEventListener('change', () => SetWindowHeaderSaveState());
		$e.addEventListener('input', () => SetWindowHeaderSaveState());
	}
	document.querySelector('.networkcommands-playerform-rs232.commandformat').addEventListener('change', function() {
		document.querySelector('.networkcommands-playerform-rs232.command').placeholder = this.value == 'hex' ? '010203040506070809A0' : 'ASCII Text';
	});
	document.querySelector('.networkcommands-playerform-udp.commandformat').addEventListener('change', function() {
		document.querySelector('.networkcommands-playerform-udp.command').placeholder = this.value == 'hex' ? '010203040506070809A0' : 'ASCII Text';
	});
	document.querySelector('.networkcommands-playerform-tcp.commandformat').addEventListener('change', function() {
		document.querySelector('.networkcommands-playerform-tcp.command').placeholder = this.value == 'hex' ? '010203040506070809A0' : 'ASCII Text';
	});
	for (const $asciiOptions of document.getElementsByClassName('ascii-options')) {
		$asciiOptions.addEventListener('click', function() {
			const $command = $.prev(this.parentNode, 'textarea');
			const $format = this.parentNode.parentNode.previousElementSibling.previousElementSibling;
			const caretPosition = $command.selectionStart;
			const textToAdd = (this.classList.contains('carriagereturn') ? ($format.value == 'hex' ? '0D' : '\\r') : '') + (this.classList.contains('newline') ? ($format.value == 'hex' ? '0A' : '\\n') : '');
			$command.value = $command.value.substring(0, caretPosition) + textToAdd + $command.value.substring(caretPosition);
			$command.dispatchEvent(new Event('change'));
			setTimeout(() => $command.selectionStart = caretPosition + textToAdd.length, 0);
		});
	}
	// TODO: Is this necessary?
	for (const $command of document.getElementsByClassName('command')) {
		$command.addEventListener('keydown', e => {
			if (e.key == 'Enter') {
				e.preventDefault();
			}
		});
	}
	document.getElementById('networkcommands-update').addEventListener('click', async () => {
		const selectedNode = (await $networkCommandsTree.getSelected())[0];
		await UpdateNetworkCommandPlayer(selectedNode);
	});
	
	DragLeftResize('.horizontal-resize', 'networkcommand-lr');
});