let networkCommandScheduleLastView = '';
let networkCommandScheduleLastScroll = 0;
let networkCommandScheduleLastSelected = null;
let currentData = null;
let $networkCommandScheduleTree;

async function AddNetworkCommandSchedule() {
	let selectedNode = (await $networkCommandScheduleTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Select a group', 'Select a group to add the new network command scheduler to.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $networkCommandScheduleTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	await InputBox('Network Command Schedule Name', 'Enter a name for the network command schedule.', 'Ok', 'Cancel', '', async name => {
		const data = await $networkCommandScheduleTree.addItems([
			{
				type: 2,
				name: name,
				parent: selectedNode.id,
				data: { scheduledcommands: [] }
			}
		]);
		if (data.success) {
			await $networkCommandScheduleTree.setSelected([data.added[0].id], false, false, true);
			return true;
		}
	}, () => true, null);
}

async function CheckNetworkCommandScheduleChange() {
	if (networkCommandScheduleLastSelected != null) {
		if (JSON.stringify(networkCommandScheduleLastSelected.data) != JSON.stringify(currentData)) {
			try {
				switch (await ConfirmationBox('Save changes?', 'You have unsaved changes. Do you want to save first?', 'Yes', 'No', 'Cancel')) {
					case 0:
						await UpdateNetworkCommandSchedule();
						break;
					case 1:
						SetNetworkCommandScheduleData(networkCommandScheduleLastSelected.data);
						break;
				}
			} catch (exception) {
				await $networkCommandScheduleTree.setSelected([networkCommandScheduleLastSelected.id]);
				return false;
			}
		}
	}
	return true;
}

function FormatNetworkCommandScheduleCalendarData(node) {
	const dow = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'];
	const calendarItems = [];
	if (node.data.hasOwnProperty('scheduledcommands')) {
		const scheduledCommands = node.data.scheduledcommands;
		const pastels = CreatePastels(scheduledCommands.length);
		scheduledCommands.forEach((scheduledCommand, scheduledCommandItr) => {
			if (dow.some(day => scheduledCommand[day])) {
				let scheduledCommandStartDate = moment().startOf('week');
				let scheduledCommandEndDate = moment().endOf('week').add(3, 'month');
				
				if (scheduledCommand.daystartdatetime != null && scheduledCommand.dayenddatetime != null) {
					scheduledCommandStartDate = moment.utc(scheduledCommand.daystartdatetime, 'YYYY-MM-DD HH:mm:ss');
					scheduledCommandEndDate = moment.utc(scheduledCommand.dayenddatetime, 'YYYY-MM-DD HH:mm:ss');
				}
			
				dow.forEach((dow, dowItr) => {
					if (scheduledCommand[dow]) {
						const dateItr = scheduledCommandStartDate.weekday() <= dowItr ? scheduledCommandStartDate.clone().weekday(dowItr) : scheduledCommandStartDate.clone().weekday(dowItr).add(1, 'week');
						while (dateItr.isBefore(scheduledCommandEndDate)) {
							calendarItems.push({
								'id': scheduledCommand.id,
								'title': node.networkcommandtoname[scheduledCommand.command],
								'start': dateItr.clone().add(moment(scheduledCommand.daytime, 'HH:mm:ss').diff(moment('00:00:00', 'HH:mm:ss'))).format('YYYY-MM-DD HH:mm:ss'),
								'end': dateItr.clone().add(moment(scheduledCommand.daytime, 'HH:mm:ss').diff(moment('00:00:00', 'HH:mm:ss'))).add(30, 'minute').format('YYYY-MM-DD HH:mm:ss'),
								'url': "#",
								'allDay': false,
								'priority': scheduledCommandItr,
								'backgroundColor': pastels[scheduledCommandItr],
								'textColor': 'black'
							});
							dateItr.add(1, 'week');
						}
					}
				});
			} else {
				calendarItems.push(
				{
					'id': scheduledCommand.id,
					'title': node.networkcommandtoname[scheduledCommand.command],
					'start': moment.utc(scheduledCommand.datetime, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss'),
					'end': moment.utc(scheduledCommand.datetime, 'YYYY-MM-DD HH:mm:ss').add(30, 'minute').format('YYYY-MM-DD HH:mm:ss'),
					'url': "#",
					'allDay': false,
					'priority': scheduledCommandItr,
					'backgroundColor': pastels[scheduledCommandItr],
					'textColor': 'black'
				});
			}
		});
	}
	return calendarItems;
}

async function NetworkCommandScheduleChanged() {
	const selectedNode = (await $networkCommandScheduleTree.getSelected())[0];
	if (selectedNode == undefined) {
		return false;
	}
	if (selectedNode.type == 1) {
		return false;
	}
	return JSON.stringify(selectedNode.data) != JSON.stringify(currentData);
}

async function SetNetworkCommandScheduleWindowHeader(header) {
	document.querySelector('#networkcommandscheduler-windowHeader > span:first-child').innerText = header;
	await SetNetworkCommandScheduleWindowHeaderSaveState();
}

async function SetNetworkCommandScheduleWindowHeaderSaveState() {
	document.querySelector('#networkcommandscheduler-windowHeader > span:last-child').innerText = await NetworkCommandScheduleChanged() ? '*' : '';
}

async function UpdateNetworkCommandSchedule() {
	await $networkCommandScheduleTree.updateItems([
		{
			id: networkCommandScheduleLastSelected.id,
			data: currentData
		}
	]);
	await SetNetworkCommandScheduleWindowHeaderSaveState();
}

function SetNetworkCommandScheduleData(data) {
	for (const scheduledCommand of data.scheduledcommands) {
		if (!scheduledCommand.hasOwnProperty('utc')) {
			scheduledCommand.utc = true;
		}
	}
	const dow = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'];
	data = JSON.parse(JSON.stringify(data));
	currentData = data;
	let sortDailyColumn = 'time';
	let sortDailyColumnOrder = 'asc';
	let sortOneTimeColumn = 'time';
	let sortOneTimeColumnOrder = 'asc';
	const toggleDailySort = function(column) {
		return () => {
			sortDailyColumnOrder = sortDailyColumnOrder == 'asc' ? 'desc' : 'asc';
			render();
		};
	};
	const toggleOneTimeSort = function(column) {
		return () => {
			sortOneTimeColumnOrder = sortOneTimeColumnOrder == 'asc' ? 'desc' : 'asc';
			render();
		};
	};
	const render = () => {
		let $networkCommandSchedulerDailyCommands = document.getElementById('networkcommandscheduler-dailycommands');
		const dailyCommandsScrollLeft = $networkCommandSchedulerDailyCommands ? $networkCommandSchedulerDailyCommands.scrollLeft : 0;
		const dailyCommandsScrollTop = $networkCommandSchedulerDailyCommands ? $networkCommandSchedulerDailyCommands.scrollTop : 0;
		let $networkCommandSchedulerOneTimeCommands = document.getElementById('networkcommandscheduler-onetimecommands');
		const oneTimeScrollLeft = $networkCommandSchedulerOneTimeCommands ? $networkCommandSchedulerOneTimeCommands.scrollLeft : 0;
		const oneTimeScrollTop = $networkCommandSchedulerOneTimeCommands ? $networkCommandSchedulerOneTimeCommands.scrollTop : 0;
		document.getElementById('networkcommandscheduler-scheduledcommandscontainer').innerHTML = '';
		document.getElementById('networkcommandscheduler-scheduledcommandscontainer').append(
			$networkCommandSchedulerDailyCommands = $.div({ id: 'networkcommandscheduler-dailycommands', class: 'formgrid compact' },
				$.header({},
					$.label({ text: 'Daily Command', title: 'The command that should be executed at the start or end of the specified time.', class: 'tooltip-help' }),
					$.label({ text: 'UTC', title: 'Whether this scheduling is relative to UTC time. (Note that the times, on the right, are displayed using the timezone in the Configuration page for convenience when this is enabled). If the goal is to have the schedule happen in every timezone at the same time, then use UTC. If the goal is to have a schedule happen in its own timezone, then uncheck this. For example, if you need to turn off a monitor using a network command at 12 AM in each timezone, then you would uncheck UTC.' }),
					$.label({ text: 'Days' }),
					$.label({},
						$.span({ text: 'Time', title: 'The time on each of the selected days to run the command.', class: 'tooltip-help' }),
						$.i({ class: ['mdi', 'tooltip-button', sortDailyColumn != 'time', 'mdi-sort', sortDailyColumn == 'time' && sortDailyColumnOrder == 'asc', 'mdi-arrow-down', sortDailyColumn == 'time' && sortDailyColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleDailySort('time') })
					),
					$.label({ text: 'Date Range', title: 'Enabling this further restricts the command to only be ran between specific dates.', class: 'tooltip-help' }),
					$.label({ text: 'Start/End', title: 'Start/End date and time', class: 'tooltip-help', step: 1, css: { gridColumn: 'span 2' } }),
					$.i({ class: 'tooltip-button mdi mdi-plus-circle-outline', title: 'Add Daily Command', click: () => {
						data.scheduledcommands.push(
						{
							command: 0,
							utc: false,
							datetime: moment().startOf('day').utc().format('YYYY-MM-DD HH:mm:ss'),
							sunday: true,
							monday: true,
							tuesday: true,
							wednesday: true,
							thursday: true,
							friday: true,
							saturday: true,
							daystartdatetime: null,
							dayenddatetime: null,
							daytime: '12:00:00',
						});
						render();
					} })
				)
			),
			$networkCommandSchedulerOneTimeCommands = $.div({ id: 'networkcommandscheduler-onetimecommands', class: 'formgrid compact' },
				$.header({},
					$.label({ text: 'One-Time Command', title: 'The command that should be executed at the start or end of the specified time.', class: 'tooltip-help' }),
					$.label({ text: 'UTC', title: 'Whether this scheduling is relative to UTC time. (Note that the time, on the right, is displayed using the timezone in the Configuration page for convenience when this is enabled). If the goal is to have the schedule happen in every timezone at the same time, then use UTC. If the goal is to have a schedule happen in its own timezone, then uncheck this. For example, if you need to turn off a monitor using a network command at 12 AM in each timezone, then you would uncheck UTC.' }),
					$.label({ css: { gridColumn: 'span 2' } },
						$.span({ text: 'Date and Time', title: 'The date the command should run on.', class: 'tooltip-help' }),
						$.i({ class: ['mdi', 'tooltip-button', sortOneTimeColumn != 'time', 'mdi-sort', sortOneTimeColumn == 'time' && sortOneTimeColumnOrder == 'asc', 'mdi-arrow-down', sortOneTimeColumn == 'time' && sortOneTimeColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleOneTimeSort('time') })
					),
					$.i({ class: 'tooltip-button mdi mdi-plus-circle-outline', title: 'Add One-Time Command', click: () => {
						data.scheduledcommands.push({
							command: 0,
							utc: false,
							datetime: moment().startOf('day').utc().format('YYYY-MM-DD HH:mm:ss'),
							sunday: false,
							monday: false,
							tuesday: false,
							wednesday: false,
							thursday: false,
							friday: false,
							saturday: false,
							daystartdatetime: null,
							dayenddatetime: null,
							daytime: '12:00am',
						});
						render();
					} })
				)
			),
			$.div({ class: 'panel-buttons' },
				$.button({ class: 'primary', text: 'Update', click: () => UpdateNetworkCommandSchedule() })
			)
		);
		
		for (const scheduledCommand of data.scheduledcommands.filter(scheduledCommand => dow.some(day => scheduledCommand[day])).sort((a, b) => {
			switch (sortDailyColumn) {
				case 'time':
					return (moment(a.daytime, 'HH:mm:ss').isBefore(moment(b.daytime, 'HH:mm:ss')) ? -1 : 1) * (sortDailyColumnOrder == 'asc' ? 1 : -1);
			}
		})) {
			const index = data.scheduledcommands.indexOf(scheduledCommand);
			let dayDateRangeEnabled = scheduledCommand.daystartdatetime != null && scheduledCommand.dayenddatetime != null;
			let dayStartDate = '';
			let dayStartTime = '';
			let dayEndDate = '';
			let dayEndTime = '';
			
			if (dayDateRangeEnabled) {
				const startDateTime = scheduledCommand.utc ? moment.utc(scheduledCommand.daystartdatetime).tz(configuration.timezone) : moment(scheduledCommand.daystartdatetime);
				const endDateTime = scheduledCommand.utc ? moment.utc(scheduledCommand.dayenddatetime).tz(configuration.timezone) : moment(scheduledCommand.dayenddatetime);
				dayStartDate = startDateTime.format('YYYY-MM-DD');
				dayStartTime = startDateTime.format('HH:mm:ss');
				dayEndDate = endDateTime.format('YYYY-MM-DD');
				dayEndTime = endDateTime.format('HH:mm:ss');
			}
			
			const dow = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'];
			let $time;
			let $startDate;
			let $startTime;
			let $endDate;
			let $endTime;
			$networkCommandSchedulerDailyCommands.append(
				$.uitreeselect({ config: {
					tree: 'networkcommand',
					defaultOptions: [{ id: 0, name: 'None', value: 'None' }],
					groupTypes: [1, 2, 3],
					groupIcon: node => node.type == 2 ? 'networkcommand' : '',
					disable: node => node.type == 1,
					filter: node => node.type == 1 || node.type == 2
				}, css: {
					gridColumn: 1
				}, change: async function() {
					scheduledCommand.command = (await this.getSelected())[0]?.networkcommandid ?? 0;
					await SetNetworkCommandScheduleWindowHeaderSaveState();
				}, process: async $treeSelect => {
					await $treeSelect.setSelected(networkCommandScheduleNode => networkCommandScheduleNode.networkcommandid == scheduledCommand.command);
				} }),
				$.uicheckbox({ checked: scheduledCommand.utc, change: function() {
					scheduledCommand.utc = this.checked;
					if (scheduledCommand.utc) {
						scheduledCommand.daytime = moment($time.value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
						if (scheduledCommand.daystartdatetime != null && scheduledCommand.dayenddatetime != null) {
							scheduledCommand.daystartdatetime = moment.tz(`${$startDate.value} ${$startTime.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
							scheduledCommand.dayenddatetime = moment.tz(`${$endDate.value} ${$endTime.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						}
					} else {
						scheduledCommand.daytime = $time.value;
						if (scheduledCommand.daystartdatetime != null && scheduledCommand.dayenddatetime != null) {
							scheduledCommand.daystartdatetime = moment(`${$startDate.value} ${$startTime.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
							scheduledCommand.dayenddatetime = moment(`${$endDate.value} ${$endTime.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
					}
					render();
				} }),
				$.div({ class: 'dow dayoption checkbox-group' },
					...dow.map(day => [
						$.uicheckbox({ class: `${day} formgrid-compact-checkbox`, checked: scheduledCommand[day], text: day.charAt(0).toUpperCase(), title: `${scheduledCommand[day] ? 'Disable' : 'Enable'} dayparting for ${day[0].toUpperCase() + day.slice(1)}`, change: function() {
							scheduledCommand[day] = this.checked;
							// if no day checked, then check all the days
							if (dow.every(d => !scheduledCommand[d])) {
								for (const d of dow) {
									scheduledCommand[d] = true;
								}
							}
							render();
						} })
					]).flat()
				),
				$time = $.input({ type: 'time', class: 'sampleInput', step: 1, value: scheduledCommand.utc ? moment(scheduledCommand.daytime, 'HH:mm:ss').add(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss') : scheduledCommand.daytime, change: function() {
					scheduledCommand.daytime = scheduledCommand.utc ? moment(this.value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss') : this.value;
				} }),
				$.uicheckbox({ checked: dayDateRangeEnabled, change: function() {
					if (this.checked) {
						scheduledCommand.daystartdatetime = moment().startOf('day').format('YYYY-MM-DD HH:mm:ss');
						scheduledCommand.dayenddatetime = moment().endOf('day').format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.daystartdatetime = null;
						scheduledCommand.dayenddatetime = null;
					}
					render();
				} }),
				$startDate = $.input({ type: 'date', class: 'sampleInput', value: dayStartDate, readonly: !dayDateRangeEnabled, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.daystartdatetime = moment.tz(`${this.value} ${moment.utc(scheduledCommand.daystartdatetime).tz(configuration.timezone).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.daystartdatetime = moment(`${this.value} ${moment(scheduledCommand.daystartdatetime).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$startTime = $.input({ type: 'time', class: 'sampleInput', value: dayStartTime, step: 1, readonly: !dayDateRangeEnabled, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.daystartdatetime = moment.tz(`${moment.utc(scheduledCommand.daystartdatetime).tz(configuration.timezone).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.daystartdatetime = moment(`${moment(scheduledCommand.daystartdatetime).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => {
					data.scheduledcommands.splice(index, 1);
					render();
				} }),
				$.span({ css: { gridColumn: 'span 5', backgroundColor: 'var(--input-readonly-background-color)', border: '1px solid var(--input-border-color)' } }), // filler
				$endDate = $.input({ type: 'date', class: 'sampleInput', value: dayEndDate, readonly: !dayDateRangeEnabled, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.dayenddatetime = moment.tz(`${this.value} ${moment.utc(scheduledCommand.dayenddatetime).tz(configuration.timezone).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.dayenddatetime = moment(`${this.value} ${moment(scheduledCommand.dayenddatetime).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$endTime = $.input({ type: 'time', class: 'sampleInput', value: dayEndTime, step: 1, readonly: !dayDateRangeEnabled, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.dayenddatetime = moment.tz(`${moment.utc(scheduledCommand.dayenddatetime).tz(configuration.timezone).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.dayenddatetime = moment(`${moment(scheduledCommand.dayenddatetime).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$.span({ css: { backgroundColor: 'var(--input-readonly-background-color)', border: '1px solid var(--input-border-color)' } })
			);
		}
		$networkCommandSchedulerDailyCommands.scrollLeft = dailyCommandsScrollLeft;
		$networkCommandSchedulerDailyCommands.scrollTop = dailyCommandsScrollTop;
		
		for (const scheduledCommand of data.scheduledcommands.filter(scheduledCommand => !dow.some(day => scheduledCommand[day])).sort((a, b) => {
			switch (sortOneTimeColumn) {
				case 'time':
					return (moment.utc(a.datetime, 'YYYY-MM-DD HH:mm:ss').isBefore(moment.utc(b.datetime, 'YYYY-MM-DD HH:mm:ss')) ? -1 : 1) * (sortOneTimeColumnOrder == 'asc' ? 1 : -1);
			}
		})) {
			const index = data.scheduledcommands.indexOf(scheduledCommand);
			const commandDateTime = scheduledCommand.utc ? moment.utc(scheduledCommand.datetime).tz(configuration.timezone) : moment(scheduledCommand.datetime);
			const commandDate = commandDateTime.format('YYYY-MM-DD');
			const commandTime = commandDateTime.format('HH:mm:ss');
			let $date;
			let $time;
			$networkCommandSchedulerOneTimeCommands.append(
				$.uitreeselect({ config: {
					tree: 'networkcommand',
					defaultOptions: [{ id: 0, name: 'None', value: 'None' }],
					groupTypes: [1, 2, 3],
					groupIcon: node => node.type == 2 ? 'networkcommand' : '',
					disable: node => node.type == 1,
					filter: node => node.type == 1 || node.type == 2
				}, change: async function() {
					scheduledCommand.command = (await this.getSelected())[0]?.networkcommandid ?? 0;
					await SetNetworkCommandScheduleWindowHeaderSaveState();
				},
				process: async $treeSelect => {
					await $treeSelect.setSelected(networkCommandScheduleNode => networkCommandScheduleNode.networkcommandid == scheduledCommand.command);
				} }),
				$.uicheckbox({ checked: scheduledCommand.utc, change: function() {
					scheduledCommand.utc = this.checked;
					if (scheduledCommand.utc) {
						scheduledCommand.datetime = moment.tz(`${$date.value} ${$time.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.datetime = moment(`${$date.value} ${$time.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
					render();
				} }),
				$date = $.input({ type: 'date', class: 'command-date sampleInput', value: commandDate, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.datetime = moment.tz(`${this.value} ${moment.utc(scheduledCommand.datetime).tz(configuration.timezone).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.datetime = moment(`${this.value} ${moment(scheduledCommand.datetime).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$time = $.input({ type: 'time', class: 'command-time sampleInput', step: 1, value: commandTime, change: function() {
					if (scheduledCommand.utc) {
						scheduledCommand.datetime = moment.tz(`${moment.utc(scheduledCommand.datetime).tz(configuration.timezone).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
					} else {
						scheduledCommand.datetime = moment(`${moment(scheduledCommand.datetime).format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
					}
				} }),
				$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => {
					data.scheduledcommands.splice(index, 1);
					render();
				} })
			);
		}
		$networkCommandSchedulerOneTimeCommands.scrollLeft = oneTimeScrollLeft;
		$networkCommandSchedulerOneTimeCommands.scrollTop = oneTimeScrollTop;
		SetNetworkCommandScheduleWindowHeaderSaveState();
	};
	render();
}

function NetworkCommandScheduleHideAll() {
	document.querySelector('#networkcommandscheduler-windowHeader > span:first-child').innerText = '';
	document.querySelector('#networkcommandscheduler-windowHeader > span:last-child').innerText = '';
	document.getElementById('networkcommandscheduler-tip').hidden = true;
	document.getElementById('networkcommandscheduler-right').hidden = true;
	document.getElementById('networkcommandscheduler-calendar').hidden = true;
	document.getElementById('networkcommandscheduler-scheduledcommandscontainer').hidden = true;
	
	for (const $e of document.querySelectorAll('#networkcommandscheduler-shownetworkcommandschedulecalendar, #networkcommandscheduler-shownetworkcommandschedule')) $e.classList.remove('selected');
}

function ShowNetworkCommandScheduleTip() {
	networkCommandScheduleLastSelected = null;
	NetworkCommandScheduleHideAll();
	document.getElementById('networkcommandscheduler-tip').hidden = false;
}

let calendar = null;

async function ShowNetworkCommandScheduleCalendar(node) {
	if (await CheckNetworkCommandScheduleChange()) {
		networkCommandScheduleLastSelected = node;
		NetworkCommandScheduleHideAll();
		SetNetworkCommandScheduleWindowHeader(`${node.name} - Calendar`);
		document.getElementById('networkcommandscheduler-right').hidden = false;
		document.getElementById('networkcommandscheduler-calendar').hidden = false;
		document.getElementById('networkcommandscheduler-shownetworkcommandschedulecalendar').classList.add('selected');
		
		SetNetworkCommandScheduleData(node.data);
		
		setTimeout(() => {
			if (calendar) {
				calendar.destroy();
			}
			calendar = new FullCalendar.Calendar(document.getElementById('networkcommandscheduler-calendar'), {
				headerToolbar: {
					left: 'prev,next today',
					center: 'title',
					right: 'dayGridMonth,timeGridWeek,timeGridDay'
				},
				eventOrder: (a, b) => a.priority - b.priority,
				initialView: ['dayGridMonth', 'timeGridWeek', 'timeGridDay'].indexOf(networkCommandScheduleLastView) == -1 ? 'dayGridMonth' : networkCommandScheduleLastView,
				timeZone: 'UTC',
				selectable: true,
				editable: false,
				height: document.getElementById('networkcommandscheduler-calendar').offsetHeight,
				aspectRatio: document.getElementById('networkcommandscheduler-calendar').offsetWidth / document.getElementById('networkcommandscheduler-calendar').offsetHeight,
				slotDuration: '00:15:00',
				allDaySlot: false,
				buttonIcons: false,
				buttonText: {
					today: 'Today',
					month: 'Month',
					week: 'Week',
					day: 'Day',
					prev: '',
					next: ''
				},
				eventDisplay: 'block',
				eventClick: ({ event }) => {
					ShowNetworkCommandSchedule(node);
					return false;
				},
				datesSet: ({ view }) => {
					networkCommandScheduleLastView = view.type;
					if (networkCommandScheduleLastView == 'timeGridWeek') {
						setTimeout(() => {
							const $container = document.querySelector('.fc-scroller-liquid-absolute');
							if ($container.hasOwnProperty('lastScroll')) {
								$container.removeEventListener('scroll', $container.lastScroll);
							}
							$container.scrollTop = networkCommandScheduleLastScroll;
							$container.addEventListener('scroll', $container.lastScroll = () => {
								networkCommandScheduleLastScroll = $container.scrollTop;
							});
						}, 0);
					}
				}
			});
			
			calendar.addEventSource(FormatNetworkCommandScheduleCalendarData(node));
			calendar.render();
			SetupCalendarTooltips();
		}, 100);
	}
}

async function ShowNetworkCommandSchedule(node) {
	if (await CheckNetworkCommandScheduleChange()){
		networkCommandScheduleLastSelected = node;
		NetworkCommandScheduleHideAll();
		document.getElementById('networkcommandscheduler-right').hidden = false;
		SetNetworkCommandScheduleWindowHeader(node.name);
		document.getElementById('networkcommandscheduler-scheduledcommandscontainer').hidden = false;
		document.getElementById('networkcommandscheduler-shownetworkcommandschedule').classList.add('selected');
		
		SetNetworkCommandScheduleData(node.data);
	}
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'networkcommandscheduler-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'RS-232/UDP/TCP Command Schedules' }),
				$.ul({ class: 'iconlist' },
					$.li({ class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $networkCommandScheduleTree.deleteSelected(), title: 'Delete selected group or network command schedule' }),
					$.li({ class: 'spacer' }),
					$.li({ class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $networkCommandScheduleTree.addGroup(), title: 'Add group' }),
					$.li({ class: 'tooltip-button mdi mdi-plus-circle-outline', click: AddNetworkCommandSchedule, title: 'Add network command schedule' }),
					$.li({ class: 'tooltip-button mdi mdi-content-duplicate', click: () => $networkCommandScheduleTree.duplicateSelected(), title: 'Duplicate selected network command schedule' }),
				),
				$networkCommandScheduleTree = $.uitree({ id: 'networkCommandScheduleTree', class: 'treeContainer', config: {
					tree: 'networkcommandschedule',
					dispatchChangeOnLoad: true,
					editable: node => node.parentNode != null,
					leafIcon: node => 'schedule',
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length == 0) {
							ConfirmationBox('Selection required', 'Select a group or network command schedule to delete.');
							return;
						}
						const selectedNode = selectedNodes[0];
						if (selectedNode.type == 1) {
							const scheduleCount = selectedAllChildren.filter(node => !node.group).length;
							if (scheduleCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${scheduleCount} network command schedule${scheduleCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove network command schedule?", `Are you sure you want to remove the network command schedule '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				}, change: async () => {
					if (await CheckNetworkCommandScheduleChange()) {
						const selectedNode = (await $networkCommandScheduleTree.getSelected())[0];
						if (selectedNode == undefined) {
							ShowNetworkCommandScheduleTip();
						} else {
							if (selectedNode.type == 1) {
								if (selectedNode.id == 1) {
									ShowNetworkCommandScheduleTip();
								} else {
									ShowNetworkCommandScheduleTip();
								}
							} else {
								ShowNetworkCommandScheduleCalendar(selectedNode);
							}
						}
					}
				} })
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'networkcommandscheduler-center', class: 'panel' },
			$.header({ id: 'networkcommandscheduler-windowHeader' },
				$.span(),
				$.span()
			),
			$.span({ id: 'networkcommandscheduler-tip', class: 'panel-tip', text: 'Create a new network command schedule using the button in the left panel or select an existing network command schedule to begin.' }),
			$.div({ id: 'networkcommandscheduler-calendar', class: 'calendar-fullview' }),
			$.div({ id: 'networkcommandscheduler-scheduledcommandscontainer' })
		),
		$.div({ id: 'networkcommandscheduler-right', class: 'panel' },
			$.header({ text: 'Options' }),
			$.div({ id: 'networkcommandscheduler-shownetworkcommandschedulecalendar', class: 'button-choices tooltip-button', title: 'Show the current scheduled commands in a calendar view' },
				$.i({ class: 'mdi mdi-calendar-outline' }),
				$.span({ text: 'Show Calendar' })
			),
			$.div({ id: 'networkcommandscheduler-shownetworkcommandschedule', class: 'button-choices tooltip-button', title: 'Add or remove scheduled commands associated with this network command schedule' },
				$.i({ class: 'mdi mdi-plus-circle-outline' }),
				$.span({ text: 'Edit Scheduled Commands' })
			)
		)
	);
	document.body.classList.remove('loading');
	ShowNetworkCommandScheduleTip();
	
	document.getElementById('networkcommandscheduler-shownetworkcommandschedulecalendar').addEventListener('click', async () => {
		const selectedNode = (await $networkCommandScheduleTree.getSelected())[0];
		if (selectedNode == undefined) {
			return false;
		}
		if (selectedNode.type == 1) {
			return;
		}
		ShowNetworkCommandScheduleCalendar(selectedNode);
	});
	
	document.getElementById('networkcommandscheduler-shownetworkcommandschedule').addEventListener('click', async () => {
		const selectedNode = (await $networkCommandScheduleTree.getSelected())[0];
		if (selectedNode == undefined) {
			return false;
		}
		if (selectedNode.type == 1) {
			return;
		}
		ShowNetworkCommandSchedule(selectedNode);
	});
	
	window.addEventListener('resize', () => {
		if (calendar) {
			calendar.setOption('height', document.getElementById('networkcommandscheduler-calendar').offsetHeight);
			calendar.setOption('aspectRatio', document.getElementById('networkcommandscheduler-calendar').offsetWidth / document.getElementById('networkcommandscheduler-calendar').offsetHeight);
		}
	});
	
	DragLeftResize('.horizontal-resize', 'networkcommandscheduler-lr');
});