var gScreenshotTimeout = null;
var gScreenshotController = null;

var gDashboardScreenshotTimeout = null;
var gDashboardScreenshotRefreshTimeInterval = null;
var gDashboardScreenshotThumbnailAgeInterval = null;

var gServerOffsetTime = 0;

let gOverrideReplacementNames = null;

/**
 * Specialized listbox control for managing alert feeds.
 * @type {?AlertViewerControl}
 */
let gAlertViewerControl = null;

let $playersTree;
let $signwaveRooms;
let $readerboardGroups;
let $playersDashboardTree;
let $signwaveBackgroundImage;
let $readerboardBackgroundImage;

function PreviewScreenshot(node) {
	var previewScreenshotTimeout = null;
	var previewScreenshotController = null;
	
	const abortPreviewScreenshot = () => {
		clearTimeout(previewScreenshotTimeout);
		if (previewScreenshotController) {
			previewScreenshotController.abort();
		}
	};
	
	const updateScreenshot = async (screenshotRetries = 0, timestamp = 0) => {
		abortPreviewScreenshot();
		document.getElementById('previewScreenshot').hidden = true;
		document.getElementById('previewScreenshotContainer').classList.add('busy');
		document.getElementById('previewScreenshotContainer').classList.remove('error');
		previewScreenshotController = new AbortController();
		try {
			var data = await get('/get/screenshot.php', {
				playerid: node.playerid,
				requesttimestamp: timestamp,
				thumbnail: false,
				refresh: screenshotRetries == 0
			}, previewScreenshotController);
			previewScreenshotController = null;
			CheckSessionExpired(data);
			if (data.success) {
				if (data.hasOwnProperty('timestamp')) {
					timestamp = data.timestamp;
				}
				if (data.hasOwnProperty('waiting')) {
					previewScreenshotTimeout = setTimeout(() => updateScreenshot(screenshotRetries + 1, timestamp), 1000);
				} else {
					document.getElementById('previewScreenshot').src = `screenshot/${account}/${node.playerid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
				}
			} else {
				abortPreviewScreenshot();
			}
		} catch (exception) {
			previewScreenshotController = null;
			abortPreviewScreenshot();
			NetworkingError();
		}
	};
	
	const $dialog = document.body.appendChild($.uidialog({ config: { modal: true } },
		$.header({ text: 'Screenshot' }),
		$.ul({ class: 'iconlist', css: { flex: '0 0 auto' } },
			$.li({ id: 'previewScreenshotLastTime' }),
			$.li({ class: 'tooltip-button mdi mdi-camera-outline', title: 'View Last Screenshot', click: () => {
				abortPreviewScreenshot();
				var thumbnailAge = moment().unix() + gServerOffsetTime - node.screenshotlastmodified;
				document.getElementById('previewScreenshotLastTime').innerText = `Screenshot Age: ${humanizeDuration(thumbnailAge * 1000)}`;
				document.getElementById('previewScreenshot').src = `screenshot/${account}/${node.playerid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
			} }),
			$.li({ title: 'Refresh Screenshot', class: 'tooltip-button mdi mdi-camera-retake-outline', click: () => {
				document.getElementById('previewScreenshotLastTime').innerText = '';
				updateScreenshot();
			} })
		),
		$.div({ id: 'previewScreenshotContainer', css: {
			width: '800px',
			height: '600px',
			borderTop: '1px solid grey',
			borderLeft: '1px solid grey',
			overflow: 'scroll'
		} },
			$.img({ id: 'previewScreenshot', hidden: true, alt: node.name, load: function() {
				this.hidden = false;
				document.getElementById('previewScreenshotContainer').classList.remove('busy');
			}, error: function() {
				this.hidden = true;
			} })
		),
		$.div({ class: 'panel-buttons' },
			$.button({ text: 'OK', click: function() {
				abortPreviewScreenshot();
				$dialog.remove();
			} })
		)
	));
	centerOnScreen($dialog);
	
	updateScreenshot();
}

let lastScreenshotAreaClick = null;

function RefreshPlayerThumbnailScreenshot(node) {
	clearTimeout(gScreenshotTimeout);
	
	document.getElementById('expandScreenshot').hidden = true;
	const $screenshotArea = document.getElementById('screenshotArea');
	$screenshotArea.style.background = '';
	if (GetContainerOS(node) == 'browser') {
		document.getElementById('screenshotArea-screenshot').hidden = true;
		document.getElementById('screenshotArea-noscreenshot').hidden = false;
		document.getElementById('refreshScreenshot').parentNode.hidden = true;
		$screenshotArea.classList.remove('busy', 'mdi', 'mdi-alert-outline');
		$screenshotArea.title = 'Browser players do not support screenshots';
	} else {
		document.getElementById('screenshotArea-screenshot').hidden = false;
		document.getElementById('screenshotArea-noscreenshot').hidden = true;
		document.getElementById('refreshScreenshot').parentNode.hidden = false;
		$screenshotArea.classList.add('busy');
		$screenshotArea.classList.remove('mdi', 'mdi-alert-outline');
		$screenshotArea.title = 'Retrieving screenshot';
		var updateScreenshot = async (screenshotRetries = 0, timestamp = 0) => {
			gScreenshotController = new AbortController();
			try {
				var data = await get('/get/screenshot.php',
				{
					playerid: node.playerid,
					requesttimestamp: timestamp,
					thumbnail: true,
					refresh: screenshotRetries == 0
				}, gScreenshotController);
				gScreenshotController = null;
				CheckSessionExpired(data);
				if (data.success) {
					if (data.hasOwnProperty('timestamp')) {
						timestamp = data.timestamp;
					}
					if (data.hasOwnProperty('waiting')) {
						gScreenshotTimeout = setTimeout(() => {
							if (screenshotRetries > 30) {
								$screenshotArea.classList.remove('busy');
								$screenshotArea.classList.add('mdi', 'mdi-alert-outline');
								$screenshotArea.title = 'Could not get screenshot thumbnail.';
							} else {
								updateScreenshot(screenshotRetries + 1, timestamp);
							}
						}, 1000);
					} else {
						$screenshotArea.title = '';
						$screenshotArea.classList.remove('busy');
						$screenshotArea.classList.add('valid');
						$screenshotArea.removeEventListener('click', lastScreenshotAreaClick);
						lastScreenshotAreaClick = () => PreviewScreenshot(node);
						$screenshotArea.addEventListener('click', lastScreenshotAreaClick);
						$screenshotArea.style.backgroundImage = `url(../../screenshot/${account}/thumbnail${node.playerid}.jpg?random=${Math.floor(Math.random() * 100000)})`;
					}
				} else {
					AbortScreenshot();
				}
			} catch (exception) {
				gScreenshotController = null;
				AbortScreenshot();
			}
		}
		
		updateScreenshot();
	}
}

function AbortScreenshot() {
	const $screenshotArea = document.getElementById('screenshotArea');
	$screenshotArea.title = 'Could not retrieve screenshot';
	if (gScreenshotController) {
		gScreenshotController.abort();
	}
}

function RefreshThumbnailScreenshot(node) {
	if (node.hasOwnProperty('screenshotThumbnailTimeout')) {
		clearTimeout(node.screenshotThumbnailTimeout);
	}
	
	if (node.hasOwnProperty('imageTag')) {
		node.imageTag.closest('.thumbnailcontainer').classList.add('busy');
		node.imageTag.closest('.thumbnailcontainer').classList.remove('mdi', 'mdi-alert-outline');
	}
	
	const updateScreenshot = async (screenshotRetries = 0, timestamp = 0) => {
		node.screenshotThumbnailController = new AbortController();
		try {
			var data = await get('/get/screenshot.php', {
				playerid: node.playerid,
				requesttimestamp: timestamp,
				thumbnail: true,
				refresh: screenshotRetries == 0
			}, node.screenshotThumbnailController);
			node.screenshotThumbnailController = null;
			CheckSessionExpired(data);
			if (data.success) {
				if (data.hasOwnProperty('timestamp')) {
					timestamp = data.timestamp;
				}
				if (data.hasOwnProperty('waiting')) {
					node.screenshotThumbnailTimeout = setTimeout(() => {
						if (screenshotRetries > 30) {
							if (node.hasOwnProperty('imageTag')) {
								const $thumbnailContainer = node.imageTag.closest('.thumbnailcontainer');
								$thumbnailContainer.classList.remove('busy');
								$thumbnailContainer.classList.add('mdi', 'mdi-alert-outline');
								$thumbnailContainer.title = 'Could not get screenshot thumbnail.';
							}
						} else {
							updateScreenshot(screenshotRetries + 1, timestamp);
						}
					}, 1000);
				} else {
					node.thumbnailscreenshotlastmodified = data.thumbnailscreenshotlastmodified;
					// Update image
					if (node.hasOwnProperty('imageTag')) {
						node.imageTag.alt = node.name;
						node.imageTag.src = `screenshot/${account}/thumbnail${node.playerid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
						const $thumbnailContainer = node.imageTag.closest('.thumbnailcontainer');
						$thumbnailContainer.classList.add('busy');
						$thumbnailContainer.title = 'Double click to view full size.';
					}
					UpdatePlayerNodeTags(node);
				}
			} else {
				
			}
		} catch (exception) {
			node.screenshotThumbnailController = null;
			NetworkingError();
		}
	};
	
	updateScreenshot();
}

function AbortDashboardUpdate()
{
	clearTimeout(gDashboardScreenshotTimeout);
	gDashboardScreenshotTimeout = null;
	clearInterval(gDashboardScreenshotRefreshTimeInterval);
	gDashboardScreenshotRefreshTimeInterval = null;
	clearInterval(gDashboardScreenshotThumbnailAgeInterval);
	gDashboardScreenshotThumbnailAgeInterval = null;
}

async function NotOnDashboard() {
	AbortDashboardUpdate();
	
	for (const node of (await $playersDashboardTree.getAllNodes()).filter(node => !node.group)) {
		if (node.hasOwnProperty('screenshotThumbnailTimeout')) {
			clearTimeout(node.screenshotThumbnailTimeout);
		}
		node.screenshotThumbnailTimeout = null;
		
		if (node.hasOwnProperty('screenshotThumbnailController') && node.screenshotThumbnailController) {
			node.screenshotThumbnailController.abort();
		}
		node.screenshotThumbnailController = null;
	}
}

async function PlayersPromo() {
	document.querySelector('#players-dashboard > .iconlist').hidden = false;
	$playersDashboardTree.hidden = false;
	document.getElementById('players-header').innerHTML = '';
	document.getElementById('players-header').append(
		$.i({ class: 'mdi mdi-asterisk' }),
		$.i({ class: 'mdi mdi-asterisk' }),
		$.i({ class: 'mdi mdi-asterisk' })
	);
	if ((await $playersTree.getAllNodes()).filter(node => !node.group).length == 0){
		document.getElementById('players-header').append(
			$.span({ css: { cursor: 'pointer' },  click: () => window.open('https://link.keywesttechnology.com/signs','_blank'), text: 'Breeze works best with our Players. When you are ready for your first player, ' }),
			$.a({ href: 'https://link.keywesttechnology.com/signs', target: '_blank', text: 'check out our latest Players' }),
			'.',
			$.a({ href: 'https://link.keywesttechnology.com/signs', target: '_blank', css: { display: 'block' } },
				$.img({ src: '/images/player-promo.png', css: { width: '100%', marginTop: '0.5rem' }, alt: 'Check out our latest Players.' })
			)
		);
		Object.assign(document.getElementById('players-header').style, {
			height: 'fit-content'
		});
		document.querySelector('#players-dashboard > .iconlist').hidden = true;
		$playersDashboardTree.hidden = true;
	} else {
		document.getElementById('players-header').innerHTML = '';
		document.getElementById('players-header').append(
			$.span({ css: { cursor: 'pointer' },  click: () => window.open('https://link.keywesttechnology.com/signs','_blank'), text: 'Breeze works best with our Players. If you need more, ' }),
			$.a({ href: 'https://link.keywesttechnology.com/signs', target: '_blank', text: 'check out our latest Players' }),
			'.'
		);
	}
}

function HidePlayersAll() {
	AbortScreenshot();
	document.getElementById('players-header').innerText = '';
	Object.assign(document.getElementById('players-header').style, {
		paddingLeft: 10,
		paddingRight: 10
	});
	document.getElementById('players-dashboard').hidden = true;
	document.getElementById('playerForm').hidden = true;
	document.getElementById('playerTools').hidden = true;
	document.getElementById('groupForm').hidden = true;
	document.getElementById('playerControlLoading').hidden = true;
}

function UpdatePlayerNodeTags(node) {
	if (node.hasOwnProperty('serialNumberTag')) {
		const serialNumber = node.hasOwnProperty('widgetdata') && node.widgetdata.hasOwnProperty('serialnumber') ? node.widgetdata.serialnumber : '';
		node.serialNumberTag.title = `${serialNumber} (Click to copy. Double click to view player)`;
		node.serialNumberTag.innerText = serialNumber;
	}
	if (node.hasOwnProperty('lastCheckInIPTag')) {
		node.lastCheckInIPTag.title = `${node.lastcheckinip} (Click to copy. Double click to view player)`;
		node.lastCheckInIPTag.innerText = node.lastcheckinip;
	}
	if (node.hasOwnProperty('lastScheduleTag')) {
		node.lastScheduleTag.title = `${node.lastschedule} (Click to copy. Double click to view player)`;
		node.lastScheduleTag.innerText = node.lastschedule;
	}
	if (node.hasOwnProperty('lastCheckInTag')) {
		const normalizedPlayerTimezone = GetNormalizedPlayerTimezone(node);
		const time = moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').tz(normalizedPlayerTimezone).format('M/D/YYYY h:mm a');
		node.lastCheckInTag.title = `${time} (Click to copy. Double click to view player)`;
		node.lastCheckInTag.innerText = time;
	}
	if (node.hasOwnProperty('containerVersionTag')) {
		node.containerVersionTag.title = `${node.containerversion} (Click to copy. Double click to view player)`;
		node.containerVersionTag.innerText = node.containerversion;
	}
	if (node.hasOwnProperty('thumbnailAgeTag')) {
		var thumbnailAge = moment().unix() + gServerOffsetTime - node.thumbnailscreenshotlastmodified;
		node.thumbnailAgeTag.classList.toggle('old', thumbnailAge > 5 * 60 + 30);
		node.thumbnailAgeTag.innerText = humanizeDuration(thumbnailAge * 1000);
		node.thumbnailAgeTag.title = `${humanizeDuration(thumbnailAge * 1000)} (Click to copy. Double click to view player)`;
	}
	
	const online = node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute'));
	if (node.hasOwnProperty('nodeTag')) {
		node.nodeTag.classList.toggle('offline', !online);
	}
}

async function RefreshPlayersTree() {
	await $playersTree.reload();
	$playersTree.dispatchEvent(new Event('change'));
}

async function ShowPlayersDashboard() {
	HidePlayersAll();
	document.getElementById('players-dashboard').hidden = false;
	
	try {
		const data = await get('/get/tree.php', {
			tree: 'player'
		});
		CheckSessionExpired(data);
		if ($playersTree.syncId !== undefined && data.syncid != $playersTree.syncId) {
			await $playersTree.reload();
		}
		if ($playersDashboardTree.syncId !== undefined && data.syncid != $playersDashboardTree.syncId) {
			await $playersDashboardTree.reload();
		}
	} catch (exception) {
		NetworkingError();
	}
	
	PlayersPromo();
	
	if (localStorage.dashboardupdate == 'true' && gDashboardScreenshotTimeout == null && gDashboardScreenshotRefreshTimeInterval == null) {
		gDashboardScreenshotTimeout = setTimeout(() => {
			gDashboardScreenshotRefreshTimeInterval = setInterval(async () => {
				try {
					const data = await get('/get/dashboardscreenshot.php', {
						syncid: $playersDashboardTree.syncId
					});
					CheckSessionExpired(data);
					document.getElementById('playerControlLoading').hidden = true;
					const allPlayers = await $playersDashboardTree.getAllNodes();
					if (data.success) {
						gServerOffsetTime = data.time - moment().unix();
						
						for (const playerId in data.players) {
							var player = data.players[playerId];
							var playerNode = allPlayers.find(node => node.playerid == playerId);
							if (playerNode) {
								playerNode.lastcheckinip = player.lastcheckinip;
								playerNode.lastcheckin = player.lastcheckin;
								playerNode.playertime = player.playertime;
								playerNode.lastschedule = player.lastschedule;
								playerNode.containerversion = player.containerversion;
								playerNode.utctime = player.utctime;
								UpdatePlayerNodeTags(playerNode);
								if (playerNode.thumbnailscreenshotlastmodified < player.thumbnailscreenshotlastmodified) {
									playerNode.thumbnailscreenshotlastmodified = player.thumbnailscreenshotlastmodified;
									
									// Update image
									if (playerNode.hasOwnProperty('imageTag')) {
										playerNode.imageTag.alt = playerNode.name;
										playerNode.imageTag.src = `screenshot/${account}/thumbnail${playerNode.playerid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
										playerNode.imageTag.closest('.thumbnailcontainer').classList.add('busy');
										playerNode.imageTag.closest('.thumbnailcontainer').classList.remove('mdi', 'mdi-alert-outline');
									}
									UpdatePlayerNodeTags(playerNode);
								}
							}
						}
					} else if (data.hasOwnProperty('treeoutdated')) {
						await Promise.all([$playersTree.reload(), $playersDashboardTree.reload()]);
					}
				} catch (exception) {
					NetworkingError();
				}
			}, 15 * 1000); // 15 seconds
		}, 10 * 1000); // 10 seconds
	}
	
	clearInterval(gDashboardScreenshotThumbnailAgeInterval);
	gDashboardScreenshotThumbnailAgeInterval = setInterval(async () => {
		for (const node of (await $playersDashboardTree.getAllNodes()).filter(node => !node.group)) {
			UpdatePlayerNodeTags(node);
		}
	}, 5000);
}

/**
 * @param {!string} label
 * @param {!string} tooltip
 * @param {!boolean} active
 * @param {!function} clickHandler
 * @param {!HTMLElement[]} tabs
 * @returns {!HTMLElement}
 */
function createTab(label, tooltip, active, clickHandler, tabs) {
	return $.div({
		text: label,
		title: tooltip,
		css: {
			cursor: 'pointer',
			position: 'relative',
			background: active ? '#42B854' : '',
			display: 'inline-block',
			boxShadow: active ? '0px 0px 0px 4px #42B854' : '',
			padding: '0 2em'
		}, click: () => {
			for (const tab of tabs) {
				deactivateTab(tab);
			}
			activateTab(tab);
			clickHandler();
		}
	});
}

/** @param {!HTMLElement} tab */
function activateTab(tab) {
	Object.assign(tab.style, {
		background: '#42B854',
		boxShadow: '0px 0px 0px 4px #42B854'
	});
}

/** @param {!HTMLElement} tab */
function deactivateTab(tab) {
	Object.assign(tab.style, {
		background: 'var(--header-color)',
		boxShadow: '0px 0px 0px 4px var(--header-color)'
	});
}

function ShowPlayerGroup(node) {
	NotOnDashboard();
	HidePlayersAll();
	
	ShowPlayerGroupScheduling();
	
	initializePlayerGroupScheduling(node);
	initializePlayerGroupAlerts(node);
}

function HidePlayerGroupsAll() {
	document.getElementById('groupForm').hidden = true;
}

function ShowPlayerGroupScheduling(node) {
	HidePlayerGroupsAll();
	document.getElementById('groupForm').hidden = false;
}

function ShowPlayerGroupAlerts() {
	HidePlayerGroupsAll();
}

async function initializePlayerGroupScheduling(node) {
	var leaves = node.allChildren.filter(node => !node.group);

	var playerNames = leaves.map(node => node.name);
	// Stores the schedule only if every player has that schedule selected
	const scheduleId = leaves.every(node => node.scheduleid == leaves[0].scheduleid) ? leaves[0].scheduleid : null;
	
	document.getElementById('groupPlayers').value = playerNames.join(', ');
	document.getElementById('groupPlayers').title = playerNames.join(', ');
	
	const $groupSchedules = document.getElementById('groupSchedules');
	$groupSchedules.reload();
	await $groupSchedules.setSelected(node => !node.group && node.scheduleid == scheduleId);
	
	document.getElementById('updateGroup').replaceWith(
		$.button({ id: 'updateGroup', class: 'primary', text: 'Update Group', click: async () => {
			const selectedNode = (await $groupSchedules.getSelected())[0];
			const scheduleId = selectedNode ? selectedNode.scheduleid : null;
			if (scheduleId == null) {
				ConfirmationBox('Schedule required', 'A schedule must be selected to update players.');
			} else if (scheduleId == 0) {
				if (await QuestionBox('Removing schedule from group of players!', 'Are you sure you want to set a group of players schedules to "None"? The players will eventually go blank and appear offline.')) {
					await updateGroupScheduling(leaves);
				}
			} else {
				if (await QuestionBox('Are you sure?', `Are you sure you want to set a group of players schedules to '${selectedNode.name}'?`)) {
					await updateGroupScheduling(leaves);
				}
			}
		} })
	);
}

/**
 * @param {!Object} node
 * @returns {!Promise}
 */
async function initializePlayerGroupAlerts(node) {
	try {
		const allAlertFeeds = node.allChildren
			.filter(node => !node.group)
			.map(leaf => leaf.alertsettings)
			.map(feeds => AlertFeed.parse(feeds));
		for (const alertFeeds of allAlertFeeds) {
			for (const alertFeed of alertFeeds) {
				await alertFeed.hash();
			}
		}
		const digestSets = allAlertFeeds
			.map(alertFeeds => alertFeeds.map(
				alertFeed => alertFeed.digest
			)).map(digests => new Set(digests));
		const union = new Set();
		for (const set of digestSets) {
			for (const digest of set) {
				union.add(digest);
			}
		}
		const allEqual = digestSets.every(set => set.size === union.size);
		if (allEqual) {
			loadAlertFeeds(document.getElementById('group-eas-body'), allAlertFeeds.pop());
		} else {
			loadAlertFeeds(document.getElementById('group-eas-body'), []);
		}
	} catch (err) {
		console.log(err.stack || err);
	}
	
	document.getElementById('updateGroupAlertSettings').replaceWith(
		$.button({ id: 'updateGroupAlertSettings', class: 'primary', css: { marginTop: '10px' }, text: 'Update Group EAS', click: async () => {
			if (validateEAS(document.getElementById('groupEas'))) {
				const leaves = node.allChildren.filter(node => !node.group);
				if (await QuestionBox('Are you sure?', 'Are you sure you want to update the Emergency Alert Notification Settings of all players in this group?')) {
					await updateGroupAlertSettings(leaves);
				}
			}
		} })
	);
}

/**
 * Validates the EAS settings are OK, otherwise renders errors and puts the offending inputs into view.
 * @param {!HTMLElement} container
 * @returns {!boolean}
 */
function validateEAS(container) {
	const validation = gAlertViewerControl.validate();
	if (!validation.valid) {
		container.click();
		ConfirmationBox(`Invalid Input - Emergency Alert Notifications`, validation.msg);
	}
	return validation.valid;
}

/**
 * @return {string}
 */
function GetContainerOS(node) {
	return node.containerversion.split('_')[0].toLowerCase();
}

/**
 * Tries to return a timezone string that can be understood by moment.tz().
 * @param {!Object} node
 * @returns {!string}
 */
function GetNormalizedPlayerTimezone(node) {
	const os = GetContainerOS(node);
	const defaultTz = 'US/Central';
	let playerTz = node.timezone == '' ? defaultTz : node.timezone;
	let mappedTz;
	if (os === 'windows') {
		const windowsDefaultTz = 'Central Standard Time';
		if (playerTz === defaultTz) {
			playerTz = windowsDefaultTz;
		}
		mappedTz = windowsTimezoneMappings[playerTz];
	}
	return mappedTz !== undefined ? mappedTz : playerTz;
}

function updateStorage(e) {
	if (e && e.originalEvent) {
		e = e.originalEvent;
		if (e.key == 'schedulefilterplaylistonly') {
			document.querySelector('#schedulelist + .filterplaylistonly').checked = localStorage.schedulefilterplaylistonly == 'true';
		}
	}
}

async function ShowPlayer(node) {
	try {
		var data = await get('/get/player.php', {
			playerid: node.playerid
		});
		CheckSessionExpired(data);
		if (data.success) {
			for (const property in data.player) {
				node[property] = data.player[property];
			}
			
			NotOnDashboard();
			HidePlayersAll();
			document.getElementById('players-header').innerText = node.name;
			document.getElementById('playerForm').hidden = false;
			document.getElementById('playerTools').hidden = false;
			
			const normalizedPlayerTimezone = GetNormalizedPlayerTimezone(node);
			document.getElementById('lastCheckIn').value = moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').tz(normalizedPlayerTimezone).format('M/D/YYYY h:mm a');
			document.getElementById('currentSchedule').value = node.lastschedule;
			document.getElementById('lastCheckinIp').value = node.lastcheckinip;
			document.getElementById('containerVersion').value = node.containerversion;
			
			const containerOS = GetContainerOS(node);
			if (containerOS === 'windows') {
				setOptions(document.getElementById('playerTimeZone'), windowsTimezones);
				document.getElementById('playerTimeZone').value = node.timezone;
				if (!document.getElementById('playerTimeZone').value) {
					console.log(`Invalid timezone: ${node.timezone}, defaulting to CST.`);
					document.getElementById('playerTimeZone').value = 'Central Standard Time';
				}
			} else {
				setOptions(document.getElementById('playerTimeZone'), linuxTimezones);
				document.getElementById('playerTimeZone').value = node.timezone;
			}
			document.getElementById('playerTime').value = moment(node.time).format('M/D/YYYY h:mm a');
			
			if (!localStorage.hasOwnProperty('schedulefilterplaylistonly')) {
				localStorage.schedulefilterplaylistonly = false;
			}
			window.removeEventListener('storage', updateStorage);
			window.addEventListener('storage', updateStorage);
			const $scheduleList = document.getElementById('schedulelist');
			$scheduleList.reload();
			await $scheduleList.setSelected(scheduleNode => scheduleNode.scheduleid == node.scheduleid);
			document.querySelector('#schedulelist + .filterplaylistonly').checked = localStorage.schedulefilterplaylistonly == 'true';
			
			var widgetData = node.widgetdata;
			document.getElementById('serialnumber').value = widgetData.hasOwnProperty('serialnumber') ? widgetData.serialnumber : '';
			document.getElementById('serialnumber').readOnly = !(widgetData.hasOwnProperty('serialnumber') && widgetData.serialnumber == '' || (widgetData.serialnumber != 'N/A' && canChangeSerialNumber));
			document.getElementById('zipcode').value = widgetData.hasOwnProperty('zipcode') ? widgetData.zipcode : '';
			document.getElementById('audioinminvolume').value = widgetData.hasOwnProperty('audioinminvolume') ? widgetData.audioinminvolume : 0;
			document.getElementById('daily-update').value = widgetData.hasOwnProperty('dailyUpdate') ? widgetData.dailyUpdate : '';
			document.getElementById('daily-reboot').value = widgetData.hasOwnProperty('dailyReboot') ? widgetData.dailyReboot : '';
			document.getElementById('pinchtozoom').checked = widgetData.hasOwnProperty('pinchtozoom') ? widgetData.pinchtozoom : false;
			document.getElementById('audioin').checked = widgetData.hasOwnProperty('audioin') ? widgetData.audioin : false;
			document.getElementById('overridescreenshotorientation').value = widgetData.hasOwnProperty('overridescreenshotorientation') ? widgetData.overridescreenshotorientation : 'none';
			document.getElementById('keyboardoverlay').checked = widgetData.hasOwnProperty('keyboardoverlay') ? widgetData.keyboardoverlay : false;
			loadNetworkCommandSchedules(widgetData.hasOwnProperty('networkcommandschedules') ? widgetData.networkcommandschedules : []);
			loadSignwave(widgetData.hasOwnProperty('signwave') ? widgetData.signwave : {});
			loadReaderboard(widgetData.hasOwnProperty('readerboard') ? widgetData.readerboard : {});
			loadAlertFeeds(document.getElementById('eas-options-body'), node.alertsettings);
			
			// Advanced settings
			document.getElementById('ccm-provider').value = widgetData.hasOwnProperty('ccmProvider') ? widgetData.ccmProvider : '';
			
			document.getElementById('autoscale').replaceWith(
				$.uicheckbox({ id: 'autoscale', change: function() {
					toggleRow(document.getElementById('autoscaledimensions'), !this.checked);
				} })
			);
			const autoscale = document.getElementById('autoscale');
			const $playerWidth = document.getElementById('playerWidth');
			const $playerHeight = document.getElementById('playerHeight');
			if (node.screenwidth !== null && node.screenheight !== null) {
				autoscale.checked = false;
				$playerWidth.value = node.screenwidth;
				$playerHeight.value = node.screenheight;
			} else {
				autoscale.checked = true;
				$playerWidth.value = '';
				$playerHeight.value = '';
			}
			autoscale.dispatchEvent(new Event('change'));
			
			document.getElementById('refreshScreenshot').replaceWith(
				$.li({ id: 'refreshScreenshot', class: 'tooltip-button mdi mdi-camera-retake-outline', title: 'Refresh Screenshot', click: () => RefreshPlayerThumbnailScreenshot(node) })
			);
			
			RefreshPlayerThumbnailScreenshot(node);
			
			document.getElementById('updatePlayer').replaceWith(
				$.button({ id: 'updatePlayer', class: 'primary', text: 'Update Player', click: async () => {
					var online = node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute'));
					if (!online) {
						await ConfirmationBox('Warning: Player offline', 'The selected player is offline and will not change schedules until it checks back in. You might need to troubleshoot the player for network connectivity issues.');
					}
					const selectedNode = (await $scheduleList.getSelected())[0];
					const scheduleId = selectedNode ? selectedNode.scheduleid : null;
					if (scheduleId == null) {
					
					} else if (scheduleId == 0 && !await QuestionBox('Removing schedule from player', 'Do you really want to set the schedule to "None"? The player will eventually go blank and appear offline.')) {
						return;
					}
					// Advanced Settings
					var playerWidth = null;
					var playerHeight = null;
					if (!document.getElementById('autoscale').checked) {
						playerWidth = $playerWidth.value;
						playerHeight = $playerHeight.value;
						var isInteger = /^[0-9]+$/;
						if (!isInteger.test(playerWidth)) {
							ConfirmationBox('Advanced Settings: Invalid display resolution format.', 'Display Resolution Width must be a valid number.');
							return;
						}
						if (!isInteger.test(playerHeight)) {
							ConfirmationBox('Advanced Settings: Invalid display resolution format.', 'Display Resolution Height must be a valid number.');
							return;
						}
					}
				
					// Signwave
					var signwaveoverridebackgroundimage = document.getElementById('signwaveoverridebackgroundimage').checked;
					const signwavebackgroundimageNode = (await $signwaveBackgroundImage.getSelected())[0];
					const signwavebackgroundimage = signwavebackgroundimageNode ? ContentNodeToMinimalImageContent(signwavebackgroundimageNode) : null;
					const signwavebackgroundimagecoverfill = document.getElementById('signwavebackgroundimagecoverfill').checked;
				
					if (signwaveoverridebackgroundimage && signwavebackgroundimage == null) {
						ConfirmationBox('Signwave Widget Options: No override background image supplied', 'You selected to override the Signwave widget background image for this player, but you did not supply an image. Please select one.');
						return;
					}
				
					// Readerboard
					var readerboardname = document.getElementById('readerboardname').value;
					var readerboardoverridebackgroundimage = document.getElementById('readerboardoverridebackgroundimage').checked;
					const readerboardbackgroundimageNode = (await $readerboardBackgroundImage.getSelected())[0];
					const readerboardbackgroundimage = readerboardbackgroundimageNode ? ContentNodeToMinimalImageContent(readerboardbackgroundimageNode) : null;
					const readerboardbackgroundimagecoverfill = document.getElementById('readerboardbackgroundimagecoverfill').checked;
				
					if (readerboardoverridebackgroundimage && readerboardbackgroundimage == null) {
						ConfirmationBox('Readerboard Widget Options: No override background image supplied', 'You selected to override the Readerboard widget background image for this player, but you did not supply an image. Please select one.');
						return;
					}
				
					var widgetdata =
					{
						serialnumber: document.getElementById('serialnumber').value,
						serialnumbermodifiedtime: moment().unix(),
						zipcode: document.getElementById('zipcode').value,
						audioinminvolume: parseFloat(document.getElementById('audioinminvolume').value),
						dailyUpdate: document.getElementById('daily-update').value,
						dailyReboot: document.getElementById('daily-reboot').value,
						pinchtozoom: document.getElementById('pinchtozoom').checked,
						audioin: document.getElementById('audioin').checked,
						overridescreenshotorientation: document.getElementById('overridescreenshotorientation').value,
						keyboardoverlay: document.getElementById('keyboardoverlay').checked,
						signwave: {
							roomids: (await $signwaveRooms.getChecked()).map(node => node.id),
							overridebackgroundimage: signwaveoverridebackgroundimage,
							backgroundimage: signwavebackgroundimage,
							backgroundimagecoverfill: signwavebackgroundimagecoverfill,
							overridereplacementnames: gOverrideReplacementNames
						},
						readerboard: {
							name: readerboardname,
							groupids: (await $readerboardGroups.getChecked()).map(node => node.id),
							overridebackgroundimage: readerboardoverridebackgroundimage,
							backgroundimage: readerboardbackgroundimage,
							backgroundimagecoverfill: readerboardbackgroundimagecoverfill
						},
						networkcommandschedules: document.getElementById('playernetworkcommandschedules').data,
						ccmProvider: document.getElementById('ccm-provider').value
					};
				
					if (!validateEAS(document.getElementById('eas-options-header'))) {
						return;
					}
					const alertFeeds = gAlertViewerControl.toServerJSON();
					LockBody();
					if (await $playersTree.updateItems([
						{
							id: node.id,
							scheduleid: scheduleId,
							timezone: document.getElementById('playerTimeZone').value,
							widgetdata: JSON.stringify(widgetdata),
							screenwidth: playerWidth,
							screenheight: playerHeight,
							alertsettings: JSON.stringify(alertFeeds)
						}
					])) {
						UnlockBody();
						ConfirmationBox('Player updated', 'The player was updated.');
						RefreshPlayerThumbnailScreenshot(node);
					} else {
						UnlockBody();
					}
				} })
			);
		}
		else
		{
		}
	}
	catch (exception)
	{
		NetworkingError();
	}
}

async function updateGroupScheduling(nodes) {
	LockBody();
	const selectedNode = (await document.getElementById('groupSchedules').getSelected())[0];
	if (await $playersTree.updateItems(nodes.map(node => (
	{
		id: node.id,
		scheduleid: selectedNode ? selectedNode.scheduleid : null
	})))) {
		UnlockBody();
		ConfirmationBox('Players updated', 'All players were updated.');
	} else {
		UnlockBody();
		ConfirmationBox('Player update error', 'There was a problem updating the player(s).');
	}
}

async function updateGroupAlertSettings(nodes) {
	const alertFeeds = gAlertViewerControl.toServerJSON();
	LockBody();
	if (await $playersTree.updateItems(nodes.map(node => (
	{
		id: node.id,
		alertsettings: JSON.stringify(alertFeeds)
	})))) {
		UnlockBody();
		ConfirmationBox('Players updated', 'All players were updated.');
	} else {
		UnlockBody();
		ConfirmationBox('Player update error', 'There was a problem updating the player(s).');
	}
}

function loadNetworkCommandSchedules(networkCommandScheduleList) {
	const $playerNetworkCommandSchedules = document.getElementById('playernetworkcommandschedules');
	$playerNetworkCommandSchedules.data = JSON.parse(JSON.stringify(networkCommandScheduleList));
	const render = () => {
		$playerNetworkCommandSchedules.innerHTML = '';
		$playerNetworkCommandSchedules.append(
			$.header({},
				$.label({ text: 'Network Command Schedule' }),
				$.i({ class: 'mdi mdi-plus-circle-outline tooltip-button', title: 'Add a network command schedule', click: () => {
					$playerNetworkCommandSchedules.data.push(null);
					render();
				} })
			),
			...$playerNetworkCommandSchedules.data.map((id, index) => [
				$.uitreeselect({ config: {
					tree: 'networkcommandschedule',
					leafIcon: node => 'schedule',
					disable: node => node.group
				}, change: async function() {
					$playerNetworkCommandSchedules.data[index] = (await this.getSelected())[0]?.id ?? null;
				}, process: async function() {
					await this.setSelected(id == null ? [] : [id]);
				} }),
				$.i({ title: 'Delete', class: 'mdi mdi-trash-can-outline tooltip-button', click: function() {
					$playerNetworkCommandSchedules.data.splice(index, 1);
					render();
				} })
			]).flat()
		);
	};
	render();
}

async function loadSignwave(signwave) {
	toggleRow(document.getElementById('signwaveoverridereplacementnames'), false);
	gOverrideReplacementNames = signwave.hasOwnProperty('overridereplacementnames') ? signwave.overridereplacementnames : {};
	$signwaveRooms.updateNames();
	await $signwaveRooms.setChecked(node => signwave.hasOwnProperty('roomids') ? signwave.roomids.includes(node.id) : false, () => false);
	
	document.getElementById('signwaveoverridebackgroundimage').checked = signwave.hasOwnProperty('overridebackgroundimage') ? signwave.overridebackgroundimage : false;
	
	var signwaveBackgroundImage = signwave.hasOwnProperty('backgroundimage') ? signwave.backgroundimage : null;
	await $signwaveBackgroundImage.setSelected(node => signwaveBackgroundImage != null && node.type == signwaveBackgroundImage.type && node.contentid == signwaveBackgroundImage.contentid);
	
	document.getElementById('signwavebackgroundimagecoverfill').checked = signwave.hasOwnProperty('backgroundimagecoverfill') ? signwave.backgroundimagecoverfill : false;
	const signwaveOverrideBackgroundImageChange = async () => {
		for (const $element of [$signwaveBackgroundImage, document.getElementById('signwavebackgroundimagecoverfill')]) {
			toggleRow($element, document.getElementById('signwaveoverridebackgroundimage').checked);
		}
		if (!document.getElementById('signwaveoverridebackgroundimage').checked) {
			await $signwaveBackgroundImage.setSelected([]);
		}
	};
	
	document.getElementById('signwaveoverridebackgroundimage').replaceWith(
		$.uicheckbox({ id: 'signwaveoverridebackgroundimage', change: signwaveOverrideBackgroundImageChange })
	);
	await signwaveOverrideBackgroundImageChange();
}

async function loadReaderboard(readerboard) {
	document.getElementById('readerboardname').value = readerboard.hasOwnProperty('name') ? readerboard.name : '';
	await $readerboardGroups.setChecked(node => readerboard.hasOwnProperty('groupids') ? readerboard.groupids.includes(node.id) : false, () => false);
	
	document.getElementById('readerboardoverridebackgroundimage').checked = readerboard.hasOwnProperty('overridebackgroundimage') ? readerboard.overridebackgroundimage : false;
	
	var readerboardBackgroundImage = readerboard.hasOwnProperty('backgroundimage') ? readerboard.backgroundimage : null;
	await $readerboardBackgroundImage.setSelected(node => readerboardBackgroundImage != null && node.type == readerboardBackgroundImage.type && node.contentid == readerboardBackgroundImage.contentid);
	
	document.getElementById('readerboardbackgroundimagecoverfill').checked = readerboard.hasOwnProperty('backgroundimagecoverfill') ? readerboard.backgroundimagecoverfill : false;
	const readerboardOverrideBackgroundImageChange = async () => {
		for (const $element of [$readerboardBackgroundImage, document.getElementById('readerboardbackgroundimagecoverfill')]) {
			toggleRow($element, document.getElementById('readerboardoverridebackgroundimage').checked);
		}
		if (!document.getElementById('readerboardoverridebackgroundimage').checked) {
			await $readerboardBackgroundImage.setSelected([]);
		}
	};
	document.getElementById('readerboardoverridebackgroundimage').replaceWith(
		$.uicheckbox({ id: 'readerboardoverridebackgroundimage', change: readerboardOverrideBackgroundImageChange })
	);
	await readerboardOverrideBackgroundImageChange();
}

/**
 * Initializes an AlertViewerControl with the given feeds data and renders it to the given tag.
 * @param {!HTMLElement} tag
 * @param {!EASAlertSettings[]} [feeds]
 */
function loadAlertFeeds(tag, feeds = []) {
	if (gAlertViewerControl !== null) {
		gAlertViewerControl.remove();
	}
	gAlertViewerControl = new AlertViewerControl(tag, feeds);
	gAlertViewerControl.render();
}

async function resetPlayer() {
	document.getElementById('playerControlLoading').hidden = false;
	try {
		const data = await post('/post/resetplayer.php', {
			syncid: $playersTree.syncId,
			playerid: (await $playersTree.getSelected())[0].playerid
		});
		CheckSessionExpired(data);
		document.getElementById('playerControlLoading').hidden = true;
		if (data.success) {
			ConfirmationBox('Hard reboot', 'Hard Reboot sent successfully.');
		} else {
			if (data.hasOwnProperty('treeoutdated')) {
				TreeOutOfDate('player');
				await $playersTree.reload();
			}
		}
	} catch (exception) {
		NetworkingError();
	}
}

async function clearCache() {
	document.getElementById('playerControlLoading').hidden = false;
	try {
		const data = await post('/post/clearcache.php', {
			syncid: $playersTree.syncId,
			playerid: (await $playersTree.getSelected())[0].playerid
		});
		CheckSessionExpired(data);
		document.getElementById('playerControlLoading').hidden = true;
		if (data.success) {
			ConfirmationBox('Cache cleared', 'Cache cleared successfully.');
		} else {
			if (data.hasOwnProperty('treeoutdated')) {
				TreeOutOfDate('player');
				await $playersTree.reload();
			}
		}
	} catch (exception) {
		NetworkingError();
	}
}

async function softRestartPlayer() {
	document.getElementById('playerControlLoading').hidden = false;
	try {
		var data = await post('/post/softrestartplayer.php', {
			syncid: $playersTree.syncId,
			playerid: (await $playersTree.getSelected())[0].playerid
		});
		CheckSessionExpired(data);
		document.getElementById('playerControlLoading').hidden = true;
		if (data.success) {
			ConfirmationBox('Application restarted', 'Restart application sent successfully.');
		} else {
			if (data.hasOwnProperty('treeoutdated')) {
				TreeOutOfDate('player');
				await $playersTree.reload();
			}
		}
	} catch (exception) {
		NetworkingError();
	}
}

function removeBlink() {
	this.classList.remove('blink');
}

/**
* Converts an array of strings into a list of <option> elements.
* @param {string[]} arr List of strings to be converted
* @param {string} [selected] Optional value that will be set as selected, if it exists in arr.
* @return {DocumentFragment} Returns a DocumentFragment containing the list of <option> elements.
*/
function buildOptions(arr, selected) {
	var frag = document.createDocumentFragment();
	arr.forEach(function(v) {
		var el = $.option({ text: v });
		if (selected !== undefined && v === selected) {
			el.setAttribute('selected', true);
		}
		frag.appendChild(el);
	});
	return frag;
}
/**
* Sets the list of <option> elements for the given el.  This function clones el and replaces it in the DOM, which means:
* <ul>
* <li>Existing references to el will not reference the cloned element that will be inserted into the DOM
* <li>Any event listeners added using addEventListener or via the elements direct properties (e.g. node.onclick = fn) will not be cloned
* <li>Any children nodes (including text nodes) will not be cloned
* </ul>
* @param {Node} el The root element for the <option> elements.
* @param {DocumentFragment} opts A template of <option> elements to be inserted.
* @return {Node} Returns a reference to the element that replaces el in the DOM.
*/
function setOptions(el, opts) {
	let tmpEl = el.cloneNode(false);
	el.parentElement.replaceChild(tmpEl, el);
	tmpEl.appendChild(opts.cloneNode(true));
	return tmpEl;
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'players-left', class: 'panel' },
			$.header({ text: 'Players' }),
			$.ul({ id: 'players-toolbar', class: 'iconlist' }, ...[
				permissions['Players Admin'] ? [
					$.li({ class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $playersTree.deleteSelected(), title: 'Delete selected group or player' })
				]: [],
				$.li({ class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $playersTree.addGroup(), title: 'Add group' }),
				$.li({ class: 'tooltip-button mdi mdi-reload', click: RefreshPlayersTree, title: 'Refresh' })
			].flat()),
			$playersTree = $.uitree({ id: 'playersTree', class: 'treeContainer', config: {
				tree: 'player',
				dispatchChangeOnLoad: true,
				editable: node => node.parentNode != null,
				leafIcon: node => node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute')) ? 'playeron' : 'playeroff',
				nameTransform: node => {
					if (!node.group && node.hasOwnProperty('widgetdata') && node.widgetdata.hasOwnProperty('serialnumber')) {
						return `${node.name} (${node.widgetdata.serialnumber})`;
					} else {
						return node.name;
					}
				},
				deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
					if (selectedNodes.length == 0) {
						ConfirmationBox('Selection required', 'Select a group or player to delete.');
						return;
					}
					var selectedNode = selectedNodes[0];
					if (selectedNode.group) {
						var playerCount = selectedAllChildren.filter(node => !node.group).length;
						if (playerCount != 0) {
							await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${playerCount} player${playerCount == 1 ? '' : 's'}?`, 'Yes', 'No');
						}
					} else {
						await ConfirmationBox("Remove player?", `Are you sure you want to remove the player '${selectedNode.name}'?`, 'Yes', 'No');
					}
					await deleteNodes(selectedAllChildren);
				}
			}, change: async function() {
				const selectedNode = (await this.getSelected())[0];
				if (selectedNode == undefined) {
					ShowPlayersDashboard();
				} else {
					if (permissions['Metadata']) {
						document.getElementById('playersTree-metadata').classList.add('busy');
						const data = await get('/get/treemetadata.php', {
							tree: 'player',
							id: selectedNode.id
						});
						document.getElementById('playersTree-metadata').classList.remove('busy');
						if (data.success) {
							Metadata(document.getElementById('playersTree-metadata'), data.metadata, data.parentmetadata, async metadataToSave => {
								try {
									const postData = await post('/post/treemetadata.php', {
										tree: 'player',
										id: selectedNode.id,
										metadata: JSON.stringify(metadataToSave)
									});
									if (!postData.success) {
										ConfirmationBox('Save Error', postData.error);
									}
								} catch (exception) {
									NetworkingError();
								}
							});
						}
					}
					if (selectedNode.group) {
						// Is root or has no leaves
						if (selectedNode.id == 1 || selectedNode.allChildren.filter(node => !node.group).length == 0) {
							ShowPlayersDashboard();
						} else {
							ShowPlayerGroup(selectedNode);
						}
					} else {
						ShowPlayer(selectedNode);
					}
				}
			} }),
			$.div({ id: 'playersTree-metadata-header', class: 'accordion-header tooltip-button', title: 'Metadata applied to playlist media the player is playing. These key value pairs represent variables that can be used with widgets to define per player customization.', text: 'Metadata' }),
			$.div({ id: 'playersTree-metadata', class: 'metadatagrid' })
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'players-right', class: 'panel' },
			$.header({ id: 'players-header', text: 'Player' }),
			$.div({ css: { flex: '1 1 auto', minHeight: 0, display: 'flex', flexFlow: 'row nowrap', overflowX: 'auto' } },
				$.div({ id: 'players-dashboard' },
					$.div({ class: 'iconlist', css: {flex: '0 0 auto' } }),
					$playersDashboardTree = $.uitree({ id: 'players-dashboard-tree', config: {
						tree: 'player',
						showSmallLargeThumbnail: true,
						thumbnail: (node, largeImages) => {
							const showScreenshots = localStorage.dashboardshowscreenshots == 'true';
							const showDetails = localStorage.dashboardshowdetails == 'true';
							
							const nodeTag = $.div({ class: ['handle', 'dashboardplayer', node.disabled, 'disabled', largeImages, 'large'], tabindex: -1, css: { alignItems: 'stretch' } });
							node.nodeTag = nodeTag;
							if (showScreenshots) {
								nodeTag.appendChild(
									$.div({ title: GetContainerOS(node) == 'browser' ? '' : 'Double click to view full size.', class: 'thumbnailcontainer' },
										node.imageTag = $.img({ alt: node.name, src: GetContainerOS(node) == 'browser' ? 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7' : 'screenshot/' + account + '/thumbnail' + node.playerid + '.jpg?random=' + Math.floor(Math.random() * 100000), hidden: GetContainerOS(node) == 'browser', load: function() {
											var imageSize = largeImages ? 200 : 150;
											var thumbnailWidth = this.width;
											var thumbnailHeight = this.height;
											if (thumbnailWidth == 1 || thumbnailHeight == 1) {
												thumbnailWidth = imageSize;
												thumbnailHeight = imageSize / 2;
											}
											var scale = 1;
											if (Math.max(thumbnailWidth, thumbnailHeight) > imageSize) {
												scale = imageSize / Math.max(thumbnailWidth, thumbnailHeight);
											}
											thumbnailWidth *= scale;
											thumbnailHeight *= scale;
											Object.assign(this.style, {
												width: `${thumbnailWidth}px`,
												height: `${thumbnailHeight}px`
											});
											this.closest('.thumbnailcontainer').classList.remove('busy');
										}, error: function() {
											this.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
										}, dblclick: e => {
											e.stopPropagation();
											PreviewScreenshot(node);
										} }),
										$.div({ text: 'Browser Players Do Not Support Screenshots', class: 'webplayerthumbnail', hidden: GetContainerOS(node) != 'browser', css: {
											width: `${largeImages ? 200 : 150}px`,
											height: `${(largeImages ? 200 : 150) * 9 / 16}px`
										} })
									)
								);
							}
							nodeTag.appendChild(
								node.$name = $.span({ text: node.name, title: `${node.name} (Click to copy. Double click to view player)`, class: 'name', spellcheck: false, click: copyText })
							);
							if (showDetails) {
								nodeTag.append(
									$.span({ class: 'detail' },
										$.span({ text: 'SN: ' }),
										node.serialNumberTag = $.span({ click: copyText })
									),
									$.span({ class: 'detail' },
										$.span({ text: 'IP: ' }),
										node.lastCheckInIPTag = $.span({ click: copyText })
									),
									$.span({ class: 'detail' },
										$.span({ text: 'Schedule: ' }),
										node.lastScheduleTag = $.span({ click: copyText })
									),
									$.span({ class: 'detail' },
										$.span({ text: 'Check In: ' }),
										node.lastCheckInTag = $.span({ click: copyText })
									),
									$.span({ class: 'detail' },
										$.span({ text: 'Version: ' }),
										node.containerVersionTag = $.span({ click: copyText })
									)
								);
							}
							if (showScreenshots && GetContainerOS(node) != 'browser') {
								//var thumbnailAge = moment().unix() + gServerOffsetTime - node.thumbnailscreenshotlastmodified;
								nodeTag.appendChild(
									$.span({ class: 'detail' },
										$.span({ text: 'Thumbnail Age: ' }),
										node.thumbnailAgeTag = $.span({ class: 'thumbnailAge', click: copyText })
									)
								)
							}
							
							if (showScreenshots && GetContainerOS(node) != 'browser') {
								// If more options are added move this code. Basically the iconlist should only be visible if there's an action available for the player
								nodeTag.appendChild(
									$.ul({ class: 'iconlist', css: { flex: '0 0 auto' } },
										$.li({ class: 'tooltip-button mdi mdi-update', title: 'Update Player', click: async () => {
											var online = node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute'));
											if (!online) {
												await ConfirmationBox('Warning: Player offline', 'The selected player is offline and will not change schedules until it checks back in. You might need to troubleshoot the player for network connectivity issues.');
											}
							
											if (await $playersTree.updateItems([
												{
													id: node.id,
													scheduleid: node.scheduleid
												}
											])) {
												ConfirmationBox('Player updated', 'The player was updated.');
												RefreshThumbnailScreenshot(node);
											} else {
												UnlockBody();
											}
										} }),
										$.li({ class: 'tooltip-button mdi mdi-restart mdi-flip-h', title: 'Restart Application', click: async () => {
											try {
												const data = await post('/post/softrestartplayer.php', {
													playerid: node.playerid
												});
												CheckSessionExpired(data);
												if (data.success) {
													ConfirmationBox('Application restarted', 'Restart application sent successfully.');
												}
											} catch (exception) {
												NetworkingError();
											}
										} }),
										$.li({ class: 'tooltip-button mdi mdi-power', title: 'Hard Reboot', click: async () => {
											try {
												const data = await post('/post/resetplayer.php', {
													playerid: node.playerid
												});
												CheckSessionExpired(data);
												if (data.success) {
													ConfirmationBox('Hard reboot', 'Hard Reboot sent successfully.');
												}
											} catch (exception) {
												NetworkingError();
											}
										} }),
										$.li({ class: 'tooltip-button mdi mdi-delete-clock-outline', title: 'Clear Cache', click: async () => {
											try {
												const data = await post('/post/clearcache.php', {
													playerid: node.playerid
												});
												CheckSessionExpired(data);
												if (data.success) {
													ConfirmationBox('Cache cleared', 'Cache cleared successfully.');
												}
											} catch (exception) {
												NetworkingError();
											}
										} }),
										$.li({ class: 'tooltip-button mdi mdi-camera-retake-outline mdi-flip-h', title: 'Force Refresh Screenshot Thumbnail', click: () => RefreshThumbnailScreenshot(node) })
									)
								);
							}
							UpdatePlayerNodeTags(node);
							return nodeTag;
						},
						editable: node => node.parentNode != null,
						doubleClick: async (e, node) => {
							if (!node.group) {
								await $playersTree.setSelected([node.id]);
								$playersTree.dispatchEvent(new Event('change'));
								return true;
							}
						}
					} })
				),
				$.div({ id: 'playerForm' },
					$.div({ id: 'player-tabs' },
						$.ul({},
							$.li({ text: 'Basic Settings' }),
							$.li({ text: 'Advanced Settings' }),
							$.li({ text: 'Widget: Signwave Options' }),
							$.li({ text: 'Widget: Readerboard Options' })
							// $.li({ id: 'eas-options-header', text: 'Emergency Alert Notification Settings' })
						),
						$.div({ class: 'formgrid' },
							$.label({ text: 'Schedule' }),
							$.div({ id: 'schedulelist-container' },
								$.uitreeselect({ id: 'schedulelist', config: {
									tree: 'schedule',
									defaultOptions: [{ id: 0, name: 'None', value: 'None' }],
									leafIcon: node => 'schedule',
									disable: node => node.group,
									filter: node => {
										if (localStorage.schedulefilterplaylistonly == 'false') {
											return true;
										} else {
											if (node.group) {
												return true;
											}
											return !/^.+\(Playlist Only\)$/.test(node.name);
										}
									}
								} })
							),
							
							$.label({ class: 'tooltip-help', title: 'Corresponds to the timezone the player is located in', text: 'Time Zone' }),
							$.select({ id: 'playerTimeZone', class: 'sampleSelect' }),
							
							$.label({ class: 'tooltip-help', title: 'The last time the player checked into the server', text: 'Last Check In:' }),
							$.input({ type: 'text', id: 'lastCheckIn', readonly: '' }),
							
							$.label({ class: 'tooltip-help', title: 'The time according to the player', text: 'Player Date/Time:' }),
							$.input({ type: 'text', id: 'playerTime', readonly: '' }),
							
							$.label({ class: 'tooltip-help', title: 'Schedule currently showing on the player', text: 'Current Schedule:' }),
							$.input({ type: 'text', id: 'currentSchedule', readonly: '' }),
							
							$.label({ class: 'tooltip-help', title: 'The last reported IP of the player', text: 'Last Checkin IP:' }),
							$.input({ type: 'text', id: 'lastCheckinIp', readonly: '' }),
							
							$.label({ class: 'tooltip-help', title: 'The latest reported version of the container running the player.  The HTML5 Web container does not report a version.', text: 'Container Version:' }),
							$.input({ type: 'text', id: 'containerVersion', readonly: '' })
						),
						$.div({ class: 'formgrid' },
							$.label({ class: 'tooltip-help', for: 'serialnumber', title: 'Serial number of the device used for support purposes. Usual format: BRZXXXXXXXX', text: 'Serial Number' }),
							$.input({ type: 'text', id: 'serialnumber' }),
							
							$.label({ class: 'tooltip-help', title: 'When enabled the Player will stretch playlists to match the resolution of the connected display.', text: 'Stretch to fit Display' }),
							$.uicheckbox({ id: 'autoscale' }),
							
							$.label({ class: 'tooltip-help', title: `The display resolution the Player should render to, generally this is the resolution of the connected monitor.  Enable 'Stretch to fit Display' for the Player to auto-detect these values.`, text: 'Display Resolution' }),
							$.span({ id: 'autoscaledimensions' },
								$.input({ type: 'text', id: 'playerWidth', class: 'sampleInput' }),
								$.span({ css: { margin: '0 5px' }, text: 'x' }),
								$.input({ type: 'text', id: 'playerHeight', class: 'sampleInput' })
							),
							
							$.label({ class: 'tooltip-help', title: 'Assigns one or more network command schedules to the player', text: 'Network Command Schedules' }),
							$.div({ id: 'playernetworkcommandschedules', class: 'formgrid compact', css: { width: '100%', height: '150px' } }),
							
							$.label({ class: 'tooltip-help', title: 'Add a CloudCoverMusic auto login URL to enable background music provided by CloudCoverMusic.com.  A single auto login URL will only play from one player at a time, to enable background music on multiple players please use a unique auto login URL for each player.', text: 'CloudCoverMusic URL' }),
							$.input({ type: 'text', id: 'ccm-provider' }),
							
							$.label({ class: 'tooltip-help', for: 'zipcode', title: 'Feature used to replace the zipcode used in the weather widgets for individual players', text: 'Zipcode' }),
							$.input({ type: 'text', id: 'zipcode' }),
							
							$.label({ class: 'tooltip-help', for: 'audioinminvolume', title: 'The minimum volume to duck to when other media are playing over the audio in (or microphone input).', text: 'Audio In Minimum Volume' }),
							$.select({ id: 'audioinminvolume' },
								$.option({ value: '1', text: `100% (Don't duck the audio)` }),
								$.option({ value: '0.9', text: '90%' }),
								$.option({ value: '0.8', text: '80%' }),
								$.option({ value: '0.7', text: '70%' }),
								$.option({ value: '0.6', text: '60%' }),
								$.option({ value: '0.5', text: '50%' }),
								$.option({ value: '0.4', text: '40%' }),
								$.option({ value: '0.3', text: '30%' }),
								$.option({ value: '0.2', text: '20% ' }),
								$.option({ value: '0.15', text: '15%' }),
								$.option({ value: '0.10', text: '10%' }),
								$.option({ value: '0.05', text: '5%' }),
								$.option({ value: '0', text: 'Mute' })
							),
							
							$.label({ class: 'tooltip-help', for: 'daily-update', title: 'Specify a time for players to update daily', text: 'Daily Update' }),
							$.input({ type: 'time', class: 'sampleInput', id: 'daily-update', step: '1' }),
								
							$.label({ class: 'tooltip-help', for: 'daily-reboot', title: 'Specify a time for players to reboot daily', text: 'Daily Reboot' }),
							$.input({ type: 'time', class: 'sampleInput', id: 'daily-reboot', step: '1' }),
							
							$.label({ class: 'tooltip-help', title: 'Allow pinch to zoom where a user on a touch display can zoom in and out.', text: 'Allow Pinch to Zoom' }),
							$.uicheckbox({ id: 'pinchtozoom' }),
							
							$.label({ class: 'tooltip-help', title: 'This passes the microphone or audio in sound through the device and out the speakers.', text: 'Enable Audio In' }),
							$.uicheckbox({ id: 'audioin' }),
							
							$.label({ for: 'overridescreenshotorientation', class: 'tooltip-help', title: 'This is very rarely needed for Android devices that incorrectly implement the screen orientation queries. Use Device Orientation should be used by most players.', text: 'Override Screenshot Orientation' }),
							$.select({ id: 'overridescreenshotorientation' },
								$.option({ value: 'none', text: 'Use Device Orientation' }),
								$.option({ value: 'rotate0', text: 'Rotate 0 degrees' }),
								$.option({ value: 'rotate90', text: 'Rotate 90 degrees' }),
								$.option({ value: 'rotate180', text: 'Rotate 180 degrees' }),
								$.option({ value: 'rotate270', text: 'Rotate 270 degrees' })
							),
							
							$.label({ class: 'tooltip-help', title: 'Show keyboard overlay. Note this is only available on Windows and Linux players.', text: 'Keyboard Overlay', hidden: true }),
							$.uicheckbox({ id: 'keyboardoverlay', hidden: true })
						),
						$.div({ class: 'formgrid' },
							$.label({ class: 'tooltip-help', title: 'This is used by players utilizing the Signwave widget to filter and display events from single or multiple rooms', text: 'Rooms' }),
							$.div({ class: 'providerlist' },
								$.ul({ id: 'signwaverooms-toolbar', class: 'iconlist' }),
								$signwaveRooms = $.uitree({ id: 'signwaverooms', config: {
									tree: 'provider',
									fullSelect: true,
									checkbox: true,
									groupTypes: [1, 2, 3, 5, 6, 8, 9, 10],
									disable: node => node.type == 1 || node.type == 2 || node.type == 5,
									groupIcon: node => {
										switch (node.type) {
											case 2: // Provider
												return 'provider';
											case 3: // Provider Credentials
												return 'credential';
											case 8: // Provider Delegate
												return 'delegate';
											case 10: // Building
												return 'building';
											default:
												return '';
										}
									},
									nameTransform: node => {
										if (!node.group && gOverrideReplacementNames && gOverrideReplacementNames.hasOwnProperty(node.id)) {
											return `${node.name} Override: ${gOverrideReplacementNames[node.id]}`;
										} else {
											return node.name;
										}
									},
									leafIcon: node => 'room',
									editable: node => false
								} })
							),
							
							$.label({ for: 'signwaveoverridereplacementnames', class: 'tooltip-help', title: 'Override room replacement name with a player-specific name', text: 'Override Room Replacement Name' }),
							$.input({ type: 'text', id: 'signwaveoverridereplacementnames' }),
							
							$.label({ class: 'tooltip-help', title: 'Override the background image with a player defined background image', text: 'Override Background Image' }),
							$.uicheckbox({ id: 'signwaveoverridebackgroundimage' }),
							
							$.label({ class: 'tooltip-help', title: 'Background image to use for this player replacing the default background image', text: 'Background Image' }),
							$signwaveBackgroundImage = $.uitreeselect({ id: 'signwavebackgroundimage', config: {
								tree: 'content',
								filter: node => node.contenttreetype == 3
							} }),
							
							$.label({ class: 'tooltip-help', title: 'By default the image will size to fit the background. This preserves the aspect ratio covering the whole screen', text: 'Background Image Cover' }),
							$.uicheckbox({ id: 'signwavebackgroundimagecoverfill' }),
						),
						$.div({ class: 'formgrid' },
							$.label({ for: 'readerboardname', class: 'tooltip-help', title: `The readerboard header displayed at the top. For example, 'Floor 1'. This is optional`, text: 'Header' }),
							$.input({ type: 'text', id: 'readerboardname' }),
							
							$.label({ class: 'tooltip-help', title: 'This is usually the rooms that events should be grouped with, but Readerboard also allows grouping by names. If grouping by name is used then names are expected here for grouping events. Replacement names are only available when group by room or name is used in the widget.', text: 'Rooms (Groups)' }),
							$.div({ class: 'providerlist' },
								$.ul({ id: 'readerboardgroups-toolbar', class: 'iconlist' }),
								$readerboardGroups = $.uitree({ id: 'readerboardgroups', config: {
									tree: 'provider',
									fullSelect: true,
									checkbox: true,
									groupTypes: [1, 2, 3, 5, 6, 8, 9, 10],
									disable: node => node.type == 1 || node.type == 2 || node.type == 5,
									groupIcon: node => {
										switch (node.type) {
											case 2: // Provider
												return 'provider';
											case 3: // Provider Credentials
												return 'credential';
											case 8: // Provider Delegate
												return 'delegate';
											case 10: // Building
												return 'building';
											default:
												return '';
										}
									},
									leafIcon: node => 'room',
									editable: node => false
								} })
							),
								
							$.label({ for: 'readerboardoverridebackgroundimage', class: 'tooltip-help', title: 'Override the background image with a player defined background image', text: 'Override Background Image' }),
							$.uicheckbox({ id: 'readerboardoverridebackgroundimage' }),
								
							$.label({ class: 'tooltip-help', title: 'Background image to use for this player replacing the default background image', text: 'Background Image' }),
							$readerboardBackgroundImage = $.uitreeselect({ id: 'readerboardbackgroundimage', config: {
								tree: 'content',
								leafIcon: node => GetContentIcon(node),
								filter: node => node.contenttreetype == 3
							} }),
							
							$.label({ for: 'readerboardbackgroundimagecoverfill', class: 'tooltip-help', title: 'By default the image will size to fit the background. This preserves the aspect ratio covering the whole screen', text: 'Background Image Cover' }),
							$.uicheckbox({ id: 'readerboardbackgroundimagecoverfill' }),
						),
						$.div({ id: 'eas-options-body', css: { overflow: 'hidden' } })
					),
					$.div({ id: 'updatePlayerContainer' },
						$.button({ id: 'updatePlayer', class: 'primary', text: 'Update Player' })
					)
				),
				$.div({ id: 'groupForm' },
					$.div({ id: 'playergroup-tabs' },
						$.ul({},
							$.li({ title: 'Set the schedule of all players in the group at once.', text: 'Group Schedule' })
							// $.li({ title: 'Set the Emergency Event Notification Settings of all players in the group at once.', text: 'Group EAS' })
						),
						$.div({},
							$.div({ class: 'formgrid' },
								$.label({ text: 'Players:' }),
								$.input({ type: 'text', id: 'groupPlayers', readonly: true }),
									
								$.label({ text: 'Schedule' }),
								$.uitreeselect({ id: 'groupSchedules', config: {
									tree: 'schedule',
									defaultOptions: [{ id: 0, name: 'None', value: 'None' }],
									leafIcon: node => 'schedule',
									disable: node => node.group
								} })
							),
							$.div({ class: 'groupbuttoncontainer' },
								$.button({ class: 'primary', id: 'updateGroup', text: 'Update Group' })
							)
						),
						$.div({},
							$.div({ id: 'group-eas-body' }),
							$.div({ class: 'groupbuttoncontainer' },
								$.button({ id: 'updateGroupAlertSettings' })
							)
						)
					)
				),
				$.div({ id: 'playerTools', class: 'panel' },
					$.header({},
						$.span({ id: 'screenshotArea-screenshot', text: 'Player Screenshot - Click on ' }),
						$.i({ class: 'mdi mdi-reload', css: { fontSize: '15px', lineHeight: '15px', height: '15px', display: 'inline-block', verticalAlign: 'middle' } }),
						$.span({ text: ' to refresh the screenshot' }),
						$.span({ id: 'screenshotArea-noscreenshot', text: 'Browser players do not support screenshots' }),
					),
					$.div({ css: { flex: '1 0 auto' } },
						$.ul({ class: 'iconlist', css: { flex: '0 0 auto' } },
							$.li({ id: 'expandScreenshot', class: 'tooltip-button mdi mdi-arrow-expand-all', title: 'Expand Screenshot' }),
							$.li({ id: 'refreshScreenshot' })
						),
						$.div({ id: 'screenshotArea' })
					),
					$.div({ css: { flex: '0 0 auto' } },
						$.div({ id: 'playerTools-troubleshooting', class: 'panel', css: { marginBottom: '20px' } },
							$.header({ text: 'Player Troubleshooting Tools' }),
							$.div({ text: 'These buttons only function if the player is checking in (green in the tree). You may need to power cycle - unplug it and plug it back in - if a player has not checked in for a while.' }),
							$.div({ class: 'panel-buttons' },
								$.button({ id: 'softRestart',  class: 'tooltip-button primary', title: 'The Breeze Player software will exit to desktop and restart itself. This process may take up to a minute', click: softRestartPlayer, text: 'Restart App' }),
								$.button({ id: 'resetPlayer', class: 'tooltip-button primary', title: 'The Breeze Player will reboot the operating system', click: resetPlayer, text: 'Hard Reboot' }),
								$.button({ id: 'clearCache', class: 'tooltip-button primary', title: `The Breeze Player will clear the downloaded content (used for offline playing), exit to the desktop, and then restart itself. This process may take up to a minute`, click: clearCache, text: 'Clear Cache' })
							)
						)
					),
					$.div({},
						$.span({ id: 'playerControlLoading', css: { margin: '4px 4px 4px 10px' } },
							$.i({ class: 'mdi mdi-loading mdi-spin' })
						)
					)
				)
			)
		)
	);
	linuxTimezones = buildOptions(linuxTimezones, 'US/Central');
	windowsTimezones = buildOptions(windowsTimezones, 'Central Standard Time');
	
	document.body.classList.remove('loading');
	ShowPlayersDashboard();
	
	tabs(document.getElementById('player-tabs'));
	tabs(document.getElementById('playergroup-tabs'));
	
	const filterInputChange = function() {
		this.previousElementSibling.hidden = this.value != '';
		this.previousElementSibling.previousElementSibling.hidden = this.value == '';
		if (this.value == '') {
			$playersTree.removeFilterView();
		} else {
			const searchString = document.getElementById('searchPlayersFilter').value.toLowerCase();
			$playersTree.filterView(node => {
				return $playersTree.nameTransform(node).toLowerCase().indexOf(searchString) != -1 || (!node.group && node.hasOwnProperty('lastcheckinip') && node.lastcheckinip.indexOf(searchString) != -1);
			});
		}
	};
	document.getElementById('players-toolbar').prepend(
		$.i({ title: 'Remove filter', class: 'tooltip-button mdi mdi-close-circle-outline removefilter', hidden: true, click: function() {
			this.hidden = true;
			this.nextElementSibling.hidden = false;
			this.nextElementSibling.nextElementSibling.value = '';
			this.nextElementSibling.nextElementSibling.dispatchEvent(new Event('change'));
			$playersTree.removeFilterView();
		} }),
		$.label({ for: 'searchPlayersFilter', title: 'Search', class: 'tooltip-button mdi mdi-magnify addfilter' }),
		$.input({ type: 'text', id: 'searchPlayersFilter', placeholder: 'Search', class: 'sampleInput', css: {
			flex: '1 0 0px',
			opacity: 1,
			margin: '0 5px'
		}, input: filterInputChange, change: filterInputChange })
	);
	
	document.getElementById('schedulelist-container').appendChild(
		$.uicheckbox({ class: 'filterplaylistonly', title: 'Filter Playlist Only', change: function() {
			localStorage.schedulefilterplaylistonly = this.checked;
			
		} },
			$.span({ text: 'Playlist Only' })
		)
	);
	
	const copyText = function(e) {
		if (!this.isContentEditable) {
			var tmp = document.body.appendChild($.input({ value: this.innerText }));
			tmp.select();
			document.execCommand('copy');
			tmp.remove();
			if (!this.classList.contains('blink')) {
				this.removeEventListener('animationend', removeBlink);
				this.addEventListener('animationend', removeBlink);
				this.classList.add('blink');
			}
		}
	};
	
	if (!localStorage.hasOwnProperty('dashboardlarge')) {
		localStorage.dashboardlarge = false;
	}
	
	if (!localStorage.hasOwnProperty('dashboardshowscreenshots')) {
		localStorage.dashboardshowscreenshots = true;
	}
	
	if (!localStorage.hasOwnProperty('dashboardshowdetails')) {
		localStorage.dashboardshowdetails = true;
	}
	
	if (!localStorage.hasOwnProperty('dashboardupdate')) {
		localStorage.dashboardupdate = true;
	}
	
	let toggleScreenshots;
	let toggleDetails;
	document.querySelector('#players-dashboard > .iconlist').append(
		toggleScreenshots = $.uicheckbox({ id: 'players-dashboard-togglescreenshots', class: 'toolbar-checkbox', checked: localStorage.dashboardshowscreenshots == 'true', config: { icon: 'camera-outline' }, title: `${localStorage.dashboardshowscreenshots == 'true' ? 'Disable' : 'Enable'} Player Screenshots`, css: {
			marginRight: '5px'
		}, change: function() {
			localStorage.dashboardshowscreenshots = this.checked;
			this.title = `${localStorage.dashboardshowscreenshots == 'true' ? 'Disable' : 'Enable'} Player Screenshots`;
			if (!this.checked) {
				toggleDetails.checked = true;
				toggleDetails.dispatchEvent(new Event('change'));
			}
			$playersDashboardTree.render();
		} }),
		toggleDetails = $.uicheckbox({ id: 'players-dashboard-toggledetails', class: 'toolbar-checkbox', checked: localStorage.dashboardshowdetails == 'true', config: { icon: 'information-outline' }, title: `${localStorage.dashboardshowdetails == 'true' ? 'Disable' : 'Enable'} Player Details`, css: { marginRight: '20px' }, change: function() {
			localStorage.dashboardshowdetails = this.checked;
			this.title = `${localStorage.dashboardshowdetails == 'true' ? 'Disable' : 'Enable'} Player Details`;
			if (!this.checked) {
				toggleScreenshots.checked = true;
				toggleScreenshots.dispatchEvent(new Event('change'));
			}
			$playersDashboardTree.render();
		} }),
		$.uicheckbox({ id: 'players-dashboard-toggleupdate', class: 'toolbar-checkbox', checked: localStorage.dashboardupdate == 'true', config: { icon: 'camera-flip-outline flip-h' }, title: `${localStorage.dashboardupdate == 'true' ? 'Disable' : 'Enable'} Automatic Refresh`, css: { marginRight: '20px' }, change: function() {
			localStorage.dashboardupdate = this.checked;
			this.title = `${localStorage.dashboardupdate == 'true' ? 'Disable' : 'Enable'} Automatic Refresh`;
			if (this.checked) {
				ShowPlayersDashboard();
			} else {
				AbortDashboardUpdate();
			}
		} }),
		$.button({ id: 'players-dashboard-refresh', class: 'toolbar-button-icon mdi mdi-camera-retake-outline mdi-flip-h', title: 'Force Refresh Screenshots' })
	);
	
	document.getElementById('players-dashboard-refresh').addEventListener('click', async () => {
		// Since we don't get feedback from the refresh, we have to estimate how quickly to let a new batch start.
		// An online player returns in about 10-30 seconds. An offline player typically times out in about 30 seconds.
		const batchSize = 25;
		const secondsToWait = 5;
		let nodeCount = 0;
		let batchCount = 0;
		
		for (const node of (await $playersDashboardTree.getAllNodes()).filter(node => !node.group)) {
			if (GetContainerOS(node) != 'browser') {
				nodeCount++;
				
				if (nodeCount % batchSize === 0) {
					batchCount++;
				}
				
				setTimeout(() =>
				{
					RefreshThumbnailScreenshot(node);
				}, (batchCount * secondsToWait * 1000) );
			}
		}
	});
	
	for (const $accordionHeader of document.getElementsByClassName('accordion-header')) {
		var headerText = $accordionHeader.innerText;
		$accordionHeader.innerHTML = '';
		$accordionHeader.append(
			$.span({ text: headerText }),
			$.i({ class: 'mdi mdi-play mdi-rotate-90' })
		);
		$accordionHeader.addEventListener('click', function() {
			this.parentNode.querySelector('.accordion-header').classList.add('collapsed');
			this.classList.remove('collapsed');
		});
	}
	document.querySelector('.accordion-header:first-child')?.click();
	
	document.getElementById('playersTree-metadata-header').hidden = !permissions['Metadata'];
	
	let selectedRoom = null;
	$signwaveRooms.addEventListener('selectionchange', async function() {
		selectedRoom = null;
		const selectedNode = (await this.getSelected())[0];
		toggleRow(document.getElementById('signwaveoverridereplacementnames'), selectedNode && selectedNode.type == 4);
		if (selectedNode && selectedNode.type == 4) {
			selectedRoom = selectedNode;
			if (gOverrideReplacementNames.hasOwnProperty(selectedNode.id)) {
				document.getElementById('signwaveoverridereplacementnames').value = gOverrideReplacementNames[selectedNode.id];
			} else {
				document.getElementById('signwaveoverridereplacementnames').value = '';
			}
		}
	});
	const signwageOverrideReplacementNames = async function() {
		if (selectedRoom) {
			if (this.value == '') {
				delete gOverrideReplacementNames[selectedRoom.id];
			} else {
				gOverrideReplacementNames[selectedRoom.id] = this.value;
			}
			$signwaveRooms.updateName(selectedRoom);
		}
	};
	document.getElementById('signwaveoverridereplacementnames').addEventListener('change', signwageOverrideReplacementNames);
	document.getElementById('signwaveoverridereplacementnames').addEventListener('input', signwageOverrideReplacementNames);
	
	const configureProviderSearch = ($toolbar, $tree, searchId) => {
		const searchInputChange = function() {
			this.previousElementSibling.hidden = this.value != '';
			this.previousElementSibling.previousElementSibling.hidden = this.value == '';
			if (this.value == '') {
				$tree.removeFilterView();
			} else {
				const searchString = document.getElementById(searchId).value.toLowerCase();
				$tree.filterView(node => {
					return node.name.toLowerCase().indexOf(searchString) != -1;
				});
			}
		};
		$toolbar.prepend(
			$.li({ title: 'Remove filter', class: 'tooltip-button mdi mdi-close-circle-outline removefilter', hidden: true, click: function() {
				this.hidden = true;
				this.nextElementSibling.hidden = false;
				this.nextElementSibling.nextElementSibling.querySelector('input').value = '';
				this.nextElementSibling.nextElementSibling.querySelector('input').dispatchEvent(new Event('change'));
				$tree.removeFilterView();
			} }),
			$.li({},
				$.label({ for: searchId, title: 'Search', class: 'tooltip-button mdi mdi-magnify addfilter' })
			),
			$.li({ css: {
				flex: '1 0 auto',
				opacity: 1,
				margin: '0 5px',
				cursor: 'default'
			} },
				$.input({ type: 'text', id: searchId, placeholder: 'Search', class: 'sampleInput', input: searchInputChange, change: searchInputChange })
			)
		);
	};
	configureProviderSearch(document.getElementById('signwaverooms-toolbar'), $signwaveRooms, 'signwaverooms-search');
	configureProviderSearch(document.getElementById('readerboardgroups-toolbar'), $readerboardGroups, 'readerboardgroups-search');
	
	DragLeftResize('.horizontal-resize', 'players-lr');
});
