let $rssTree;

async function ValidateRSSFeed() {
	return true;
}

function HideRSS() {
	document.getElementById('rss-tip').hidden = true;
	document.getElementById('rss-view').hidden = true;
}

function ShowRSSTip() {
	HideRSS();
	document.getElementById('rss-tip').hidden = false;
	document.getElementById('rss-header').innerText = '';
}

function ShowRSSFeed(node, entries) {
	HideRSS();
	document.getElementById('rss-view').hidden = false;
	document.getElementById('rss-header').innerText = node.name;
	
	document.getElementById('rssURLContainer').hidden = false;
	
	let sortColumn = 'start';
	let sortColumnOrder = 'asc';
	const toggleSort = function(column) {
		return () => {
			if (sortColumn == 'none' || sortColumn != column) {
				sortColumn = column;
				sortColumnOrder = 'asc';
			} else if (sortColumnOrder == 'asc') {
				sortColumnOrder = 'desc';
			} else {
				sortColumn = 'date';
				sortColumnOrder = 'asc';
			}
			render();
		};
	};
	const render = () => {
		document.getElementById('rssEntryContainer').innerHTML = '';
		document.getElementById('rssEntryContainer').appendChild(
			$.header({},
				$.label({},
					$.span({ text: 'Headline' }),
					$.i({ class: ['mdi', 'tooltip-button', sortColumn != 'headline', 'mdi-sort', sortColumn == 'headline' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'headline' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('headline') })
				),
				$.label({ css: { gridColumn: 'span 2' } },
					$.span({ text: 'Publication Date' }),
					$.i({ class: ['mdi', 'tooltip-button', sortColumn != 'date', 'mdi-sort', sortColumn == 'date' && sortColumnOrder == 'asc', 'mdi-arrow-down', sortColumn == 'date' && sortColumnOrder == 'desc', 'mdi-arrow-up'], title: 'Sort', click: toggleSort('date') })
				),
				$.i({ class: 'tooltip-button mdi mdi-plus-circle-outline', title: 'Add RSS Entry', click: () => {
					sortColumn = 'date';
					sortColumnOrder = 'asc';
					entries.push({ headline: '', description: '', date: moment().format('ddd, DD MMM YYYY HH:mm:ss Z') });
					render();
				} })
			)
		);
		entries.sort((a, b) => {
			switch (sortColumn) {
				case 'headline':
					return a[sortColumn].localeCompare(b[sortColumn]) * (sortColumnOrder == 'asc' ? 1 : -1);
				case 'date':
					return (moment(a.date, 'ddd, DD MMM YYYY HH:mm:ss Z').isBefore(moment(b.date, 'ddd, DD MMM YYYY HH:mm:ss Z')) ? -1 : 1) * (sortColumnOrder == 'asc' ? 1 : -1);
			}
		});
		for (const entry of entries) {
			const index = entries.indexOf(entry);
			const headlineInputChange = function() { entry.headline = this.value; };
			const entryDateInputChange = function() { entry.date = moment(`${this.value} ${moment(entry.date, 'ddd, DD MMM YYYY HH:mm:ss Z').format('HH:mm')}`, 'YYYY-MM-DD HH:mm').format('ddd, DD MMM YYYY HH:mm:ss Z'); };
			const entryTimeInputChange = function() { entry.date = moment(`${moment(entry.date, 'ddd, DD MMM YYYY HH:mm:ss Z').format('YYYY-MM-DD')} ${this.value}`, 'YYYY-MM-DD HH:mm').format('ddd, DD MMM YYYY HH:mm:ss Z'); };
			const descriptionInputChange = function() { entry.description = this.value; };
			document.getElementById('rssEntryContainer').append(
				$.input({ type: 'text', class: 'sampleInput rssHeadline', placeholder: 'Headline', value: entry.headline, input: headlineInputChange, change: headlineInputChange }),
				$.input({ type: 'date', class: 'sampleInput', value: moment(entry.date, 'ddd, DD MMM YYYY HH:mm:ss Z').format('YYYY-MM-DD'), input: entryDateInputChange, change: entryDateInputChange }),
				$.input({ type: 'time', class: 'sampleInput', value: moment(entry.date, 'ddd, DD MMM YYYY HH:mm:ss Z').format('HH:mm'), input: entryTimeInputChange, change: entryTimeInputChange }),
				$.i({ class: 'tooltip-button mdi mdi-trash-can-outline', title: 'Delete', click: () => {
					entries.splice(index, 1);
					render();
				} }),
				$.div({ class: 'rss-description-container', css: { gridColumn: 'span 4' } },
					$.textarea({ class: 'sampleTextArea rssDescription', text: entry.description, placeholder: 'Description', input: descriptionInputChange, change: descriptionInputChange })
				)
			);
		}
	};
	render();
	
	const $submit = document.getElementById('submit');
	if ($submit.lastClickHandler) {
		$submit.removeEventListener('click', $submit.lastClickHandler);
	}
	$submit.addEventListener('click', $submit.lastClickHandler = async () => {
		const items = entries.map(entry => ({
			title: entry.headline,
			description: entry.description,
			pubDate: entry.date
		}));
		if (await ValidateRSSFeed(items)) {
			LockBody();
			if (await $rssTree.updateItems([
				{
					id: node.id,
					items: items
				}
			])) {
				UnlockBody();
				ConfirmationBox('RSS feed updated', 'The RSS feed was updated.');
			} else {
				UnlockBody();
			}
		}
	});
}

async function ShowNewRSSFeed() {
	let selectedNode = (await $rssTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Select a group', 'Select a group to add the new RSS feed to.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $rssTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	await InputBox('RSS Feed Name', 'Enter a name for the RSS feed.', 'Ok', 'Cancel', '', async name => {
		const data = await $rssTree.addItems([
			{
				type: 2,
				name: name,
				parent: selectedNode.id,
				items: [{ headline: '', description: '', pubDate: moment().format('ddd, DD MMM YYYY HH:mm:ss Z') }]
			}
		]);
		if (data.success) {
			await $rssTree.setSelected([data.added[0].id], false, false, true);
			return true;
		}
	}, () => true, null);
}

async function ReadRSSFeed(node) {
	try {
		LockBody();
		const $xml = await $.http({
			method: 'GET',
			url: `rss/${account}/${node.name}.rss`,
			dataType: 'xml'
		}).promise;
		const entries = [];
		for (const $item of $xml.querySelectorAll('item')) {
			var date = $item.querySelector("pubDate").textContent;
			entries.push({
				headline: $.div({ html: $item.querySelector('title').textContent }).textContent,
				description: $.div({ html: $item.querySelector('description').textContent }).textContent,
				date
			});
		}
		document.getElementById('rssURL').value  = `https://${document.domain}/rss/${account}/${node.name}.rss`;
		ShowRSSFeed(node, entries);
	} catch (exception) {
		NetworkingError();
	} finally {
		UnlockBody();
	}
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'rss-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'RSS Feeds' }),
				$.ul({ class: 'iconlist', css: { flex: '0 0 auto' } },
					$.li({ id: 'rssTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $rssTree.deleteSelected(), title: 'Delete selected' }),
					$.li({ class: 'spacer' }),
					$.li({ id: 'rssTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $rssTree.addGroup(), title: 'Add group' }),
					$.li({ id: 'rssTree-add', class: 'tooltip-button mdi mdi-plus-circle-outline', click: ShowNewRSSFeed, title: 'New RSS feed' }),
					$.li({ id: 'rssTree-duplicate', class: 'tooltip-button mdi mdi-content-duplicate', click: () => $rssTree.duplicateSelected(), title: 'Duplicate selected RSS feed' })
				),
				$rssTree = $.uitree({ id: 'rssTree', class: 'treeContainer', config: {
					tree: 'rss',
					dispatchChangeOnLoad: true,
					editable: node => node.type == 1 && node.parentNode != null,
					leafIcon: node => 'schedule',
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length == 0) {
							ConfirmationBox('Selection required', 'Select an RSS feed to delete.');
							return;
						}
						var selectedNode = selectedNodes[0];
						if (selectedNode.group) {
							var rssCount = selectedAllChildren.filter(node => !node.group).length;
							if (rssCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${rssCount} RSS feed${rssCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove RSS feed?", `Are you sure you want to remove the RSS feed '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				}, load: async () => {
					if ((await $rssTree.getAllNodes()).some(node => node.group)) {
						for (const $e of document.querySelectorAll('#rssTree-addgroup, #rssTree-add, #rssTree-delete, #rssTree-duplicate')) $e.hidden = false;
					}
				},
				change: async function() {
					const selectedNode = (await this.getSelected())[0];
					if (selectedNode == undefined) {
						ShowRSSTip();
					} else {
						if (selectedNode.group) {
							ShowRSSTip();
						} else {
							ReadRSSFeed(selectedNode);
						}
					}
				} })
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'rss-right' },
			$.div({ class: 'panel', css: {flex: '1 1 auto' } },
				$.header({ id: 'rss-header' }),
				$.span({ id: 'rss-tip', class: 'panel-tip', text: `Select an RSS feed to edit or select a folder then "New RSS feed', to create a new RSS feed.` }),
				$.div({ id: 'rss-view' },
					$.div({ id: 'rssURLContainer', class: 'formgrid' },
						$.label({ text: 'RSS URL:' }),
						$.input({ type: 'text', class: 'sampleInput', id: 'rssURL', readonly: true })
					),
					$.div({ id: 'rssEntryContainer', class: 'formgrid compact' }),
					$.div({ class: 'panel-buttons' },
						$.button({ id: 'submit', class: 'primary', text: 'Update RSS Feed' })
					)
				)
			)
		)
	);
	document.body.classList.remove('loading');
	ShowRSSTip();
	for (const $e of document.querySelectorAll('#rssTree-addgroup, #rssTree-add, #rssTree-delete, #rssTree-duplicate')) $e.hidden = true;
	
	DragLeftResize('.horizontal-resize', 'rss-lr');
});
