function getInnerText(node)
{
	if (typeof node.textContent != 'undefined')
	{
		return node.textContent;
	}
	else if (typeof node.innerText != 'undefined')
	{
		return node.innerText;
	}
	else if (typeof node.text != 'undefined')
	{
		return node.text;
	}
	else
	{
		switch (node.nodeType)
		{
			case 3:
			case 4:
			return node.nodeValue;
				break;
			case 1:
			case 11:
				var innerText = '';
				for (var i = 0; i < node.childNodes.length; ++i)
				{
					innerText += getInnerText(node.childNodes[i]);
				}
				return innerText;
				break;
			default:
				return '';
		}
	}
}

function loadScheduleXML(xml)
{
	xml = $.parseXML(xml);
	var root = xml.documentElement;
	var schedule = new Object();
	schedule.id = root.getAttribute('id');
	schedule.name = root.getAttribute('name');
	schedule.scheduleBlocks = [];
	for (var i = 0; i < root.childNodes.length; ++i)
	{
		if (root.childNodes[i].nodeName == 'ScheduleBlocks')
		{
			schedule.scheduleBlocks = loadScheduleBlocksXML(root.childNodes[i]);
		}
	}
	return schedule;
}

function loadScheduleBlocksXML(node)
{
	var scheduleBlocks = [];
	for (var i = 0; i < node.childNodes.length; ++i)
	{
		if (node.childNodes[i].nodeName == 'ScheduleBlock')
		{
			var scheduleBlockNode = node.childNodes[i];
			var scheduleBlock = new Object();
			scheduleBlock.id = scheduleBlockNode.getAttribute('id');
			scheduleBlock.type = scheduleBlockNode.getAttribute('type');
			scheduleBlock.ordering = scheduleBlockNode.getAttribute('ordering');
			for (var j = 0; j < scheduleBlockNode.childNodes.length; ++j)
			{
				if (scheduleBlockNode.childNodes[j].nodeName == 'ScheduleBlockDates')
				{
					scheduleBlock.dates = loadScheduleDatesXML(scheduleBlockNode.childNodes[j]);
				}
				else if (scheduleBlockNode.childNodes[j].nodeName == 'Playlists')
				{
					scheduleBlock.playlists = loadScheduleBlockPlaylistsXML(scheduleBlockNode.childNodes[j]);
					if (scheduleBlock.ordering == 'Shuffle')
					{
						scheduleBlock.playlists.sort(() => 0.5 - Math.random());
					}
				}
			}
			scheduleBlocks.push(scheduleBlock);
		}
	}
	return scheduleBlocks;
}

function loadScheduleDatesXML(node)
{
	var dates = [];
	for (var i = 0; i < node.childNodes.length; ++i)
	{
		if (node.childNodes[i].nodeName == 'ScheduleDate')
		{
			var dateNode = node.childNodes[i];
			var date = new Object();
			date.id = dateNode.getAttribute('id');
			for (var j = 0; j < dateNode.childNodes.length; ++j)
			{
				if (dateNode.childNodes[j].nodeName == 'StartDate')
				{
					date.startdate = getInnerText(dateNode.childNodes[j]);
				}
				else if (dateNode.childNodes[j].nodeName == 'EndDate')
				{
					date.enddate = getInnerText(dateNode.childNodes[j]);
				}
				else if (dateNode.childNodes[j].nodeName == 'StartTime')
				{
					date.starttime = getInnerText(dateNode.childNodes[j]);
				}
				else if (dateNode.childNodes[j].nodeName == 'EndTime')
				{
					date.endtime = getInnerText(dateNode.childNodes[j]);
				}
				else if (dateNode.childNodes[j].nodeName == 'Weeks')
				{
					date.weeks = getInnerText(dateNode.childNodes[j]);
				}
				else if (dateNode.childNodes[j].nodeName == 'Weekdays')
				{
					date.weekdays = getInnerText(dateNode.childNodes[j]);
				}
			}
			dates.push(date);
		}
	}
	return dates;
}

function loadScheduleBlockPlaylistsXML(node)
{
	var playlists = [];
	for (var i = 0; i < node.childNodes.length; ++i)
	{
		if (node.childNodes[i].nodeName == 'Playlist')
		{
			var playlistNode = node.childNodes[i];
			var playlist = new Object();
			playlist.id = playlistNode.getAttribute('id');
			playlist.name = playlistNode.getAttribute('name');
			playlist.order = playlistNode.getAttribute('order');
			for (var j = 0; j < playlistNode.childNodes.length; ++j)
			{
				if (playlistNode.childNodes[j].nodeName == 'RunLength')
				{
					playlist.runlength = playlistNode.childNodes[j].textContent;
				}
				else if (playlistNode.childNodes[j].nodeName == 'PageWidth')
				{
					playlist.pagewidth = playlistNode.childNodes[j].textContent;
				}
				else if (playlistNode.childNodes[j].nodeName == 'PageHeight')
				{
					playlist.pageheight = playlistNode.childNodes[j].textContent;
				}
			}
			playlists.push(playlist);
		}
	}
	return playlists;
}
