var gScheduleBlockId;
var gScheduleBlockDatesId;
var gScheduleBlockName;

var lastScroll = 0;

let calendar = null;

let $startDate;
let $startTime;
let $endDate;
let $endTime;

if (!localStorage.hasOwnProperty('schedulefilterplaylistonly')) {
	localStorage.schedulefilterplaylistonly = false;
}

if (!localStorage.hasOwnProperty('schedulecalendarview')) {
	localStorage.schedulecalendarview = 'month';
}

window.addEventListener('storage', function(e) {
	e = e.originalEvent;
	if (e.key == 'schedulefilterplaylistonly') {
		document.getElementById('scheduleTree-filterplaylistonly').checked = localStorage.schedulefilterplaylistonly == 'true';
		FilterPlaylistOnly();
	}
});

function FilterPlaylistOnly() {
	const $scheduleTree = document.getElementById('scheduleTree');
	const filter = localStorage.schedulefilterplaylistonly == 'true';
	if (filter) {
		$scheduleTree.filterView(node => {
			if (node.group) {
				return true;
			}
			return !/^.+\(Playlist Only\)$/.test(node.name);
		});
	} else {
		$scheduleTree.removeFilterView();
	}
}

function HideScheduler() {
	document.getElementById('scheduler-tip').hidden = true;
	document.getElementById('scheduler-right').hidden = true;
	document.getElementById('calendar').hidden = true;
	document.getElementById('scheduleForm').hidden = true;
	document.getElementById('scheduleBlockForm').hidden = true;
}

function ShowSchedulerTip()
{
	HideScheduler();
	document.getElementById('windowHeader').innerText = '';
	document.getElementById('scheduler-tip').hidden = false;
}

let renderCalendarTimer = null;

async function ShowScheduleCalendar() {
	const $scheduleTree = document.getElementById('scheduleTree');
	HideScheduler();
	document.getElementById('windowHeader').innerText = `Edit Schedule "${(await $scheduleTree.getSelected())[0].name}"`;
	document.getElementById('scheduler-right').hidden = false;
	document.getElementById('calendar').hidden = false;
	
	for (const $buttonChoice of document.getElementsByClassName('button-choices')) $buttonChoice.classList.remove('selected');
	
	requestRightScheduleBlocks();
	
	destroyCalendar();
	clearTimeout(renderCalendarTimer);
	renderCalendarTimer = setTimeout(() => renderCalendar(), 100);
}

async function ShowNewSchedule() {
	const $scheduleTree = document.getElementById('scheduleTree');
	let selectedNode = (await $scheduleTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Selection required', 'Select a group for the schedule.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $scheduleTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	HideScheduler();
	document.getElementById('windowHeader').innerText = 'New Schedule';
	document.getElementById('scheduleForm').hidden = false;
	
	document.getElementById('scheduleName').value = '';
	
	document.getElementById('defaultPlaylists').innerHTML = '';
	await requestAllPlaylists('defaultLeftPlaylists');
	
	document.getElementById('scheduleFormSubmit').replaceWith(
		$.button({ id: 'scheduleFormSubmit', class: 'primary', text: 'Submit', value: 'Create', click: submitSchedule })
	);
	
	document.getElementById('scheduleName').closest('div').hidden = false;
	
	document.getElementById('scheduleFormCancel').hidden = true;
}

async function ShowUpdateSchedule() {
	const $scheduleTree = document.getElementById('scheduleTree');
	HideScheduler();
	const scheduleName = (await $scheduleTree.getSelected())[0].name;
	document.getElementById('windowHeader').innerText = `Edit Default Schedule for "${scheduleName}" Schedule`;
	document.getElementById('scheduler-right').hidden = false;
	document.getElementById('scheduleForm').hidden = false;
	
	for (const $buttonChoice of document.getElementsByClassName('button-choices')) $buttonChoice.classList.remove('selected');
	document.getElementById('showUpdateSchedule').classList.add('selected');
	
	destroyCalendar();
	
	document.getElementById('scheduleName').value = scheduleName;
	
	document.getElementById('defaultLeftPlaylists').innerHTML = '';
	document.getElementById('defaultPlaylists').innerHTML = '';
	document.getElementById('scheduleName').closest('div').hidden = true;
	LockBody();
	await Promise.all([requestAllPlaylists('defaultLeftPlaylists'), requestDefaultPlaylists()]);
	UnlockBody();
	
	document.getElementById('scheduleFormSubmit').replaceWith(
		$.button({ id: 'scheduleFormSubmit', class: 'primary', text: 'Submit', value: 'Update', click: updateSchedule })
	);
	
	document.getElementById('scheduleFormCancel').hidden = false;
}

async function ShowNewScheduleBlock() {
	const $scheduleTree = document.getElementById('scheduleTree');
	HideScheduler();
	document.getElementById('windowHeader').innerText = `Edit Schedule "${(await $scheduleTree.getSelected())[0].name}" New Schedule Block`;
	document.getElementById('scheduler-right').hidden = false;
	document.getElementById('scheduleBlockForm').hidden = false;
	
	for (const $buttonChoice of document.getElementsByClassName('button-choices')) $buttonChoice.classList.remove('selected');
	document.getElementById('showNewScheduleBlock').classList.add('selected');
	
	document.getElementById('scheduleBlockName').value = '';
	
	destroyCalendar();
	
	setDefaultValues();
	
	document.getElementById('leftPlaylists').innerHTML = '';
	document.getElementById('rightPlaylists').innerHTML = '';
	
	await requestAllPlaylists('leftPlaylists');
	
	document.getElementById('createNewScheduleBlock').hidden = false;
	document.getElementById('cancelNewScheduleBlock').hidden = false;
	document.getElementById('clickUpdateScheduleBlock').hidden = true;
	document.getElementById('clickCancelUpdateScheduleBlock').hidden = true;
	document.getElementById('deleteScheduleBlock').hidden = true;
	document.getElementById('moveScheduleBlockUp').hidden = true;
	document.getElementById('moveScheduleBlockDown').hidden = true;
}

async function ShowUpdateScheduleBlock() {
	const $scheduleTree = document.getElementById('scheduleTree');
	HideScheduler();
	document.getElementById('windowHeader').innerText = `Edit Schedule Block "${gScheduleBlockName}" in "${(await $scheduleTree.getSelected())[0].name}" Schedule`;
	document.getElementById('scheduler-right').hidden = false;
	document.getElementById('scheduleBlockForm').hidden = false;
	
	for (const $buttonChoice of document.getElementsByClassName('button-choices')) $buttonChoice.classList.remove('selected');
	document.getElementById(`scheduleBlock${gScheduleBlockId}`)?.classList?.add('selected');
	
	destroyCalendar();
	
	document.getElementById('scheduleBlockName').value = gScheduleBlockName;
	document.getElementById('leftPlaylists').innerHTML = '';
	LockBody();
	await Promise.all([requestAllPlaylists('leftPlaylists'), requestScheduleBlockData()]);
	UnlockBody();
	
	document.getElementById('createNewScheduleBlock').hidden = true;
	document.getElementById('cancelNewScheduleBlock').hidden = true;
	document.getElementById('clickUpdateScheduleBlock').hidden = false;
	document.getElementById('clickCancelUpdateScheduleBlock').hidden = false;
	document.getElementById('deleteScheduleBlock').hidden = false;
	document.getElementById('moveScheduleBlockUp').hidden = false;
	document.getElementById('moveScheduleBlockDown').hidden = false;
}

async function renderCalendar() {
	const $scheduleTree = document.getElementById('scheduleTree');
	const $calendar = document.getElementById('calendar');
	destroyCalendar();
	const width = $calendar.offsetWidth;
	const height = $calendar.offsetHeight;
	if (calendar) {
		calendar.destroy();
	}
	let scheduleBlocks = null;
	let lastStart = moment();
	let lastEnd = moment();
	const generateCalendarEvents = (start = lastStart, end = lastEnd) => {
		if (start instanceof Date) {
			start = moment(start);
		}
		if (end instanceof Date) {
			end = moment(end);
		}
		lastStart = start;
		lastEnd = end;
		for (const eventSource of calendar.getEventSources()) {
			eventSource.remove();
		}
		if (scheduleBlocks) {
			const pastels = CreatePastels(scheduleBlocks.length);
			// Generate events
			const events = [];
			for (const scheduleBlock of scheduleBlocks) {
				if (scheduleBlock.date == null) {
					// Default
					events.push({
						id: scheduleBlock.id,
						priority: scheduleBlock.priority,
						title: 'Default',
						backgroundColor: pastels[scheduleBlock.priority],
						textColor: 'var(--button-choice-font-color)',
						url: '#',
						allDay: false,
						scheduledatesid: -1,
						start: start.format('YYYY-MM-DD HH:mm:ss'),
						end: `${end.format('YYYY-MM-DD')} ${end.clone().endOf('day').format('HH:mm:ss')}`
					});
				} else if (!(scheduleBlock.date['sunday'] || scheduleBlock.date['monday'] || scheduleBlock.date['tuesday'] || scheduleBlock.date['wednesday'] || scheduleBlock.date['thursday'] || scheduleBlock.date['friday'] || scheduleBlock.date['saturday'])) {
					events.push({
						id: scheduleBlock.id,
						priority: scheduleBlock.priority,
						title: scheduleBlock.name,
						backgroundColor: pastels[scheduleBlock.priority],
						textColor: 'var(--button-choice-font-color)',
						url: '#',
						allDay: false,
						scheduledatesid: scheduleBlock.date.scheduledatesid,
						start: `${scheduleBlock.date.startdate} ${scheduleBlock.date.starttime}`,
						end: `${scheduleBlock.date.enddate} ${scheduleBlock.date.endtime}`
					});
				} else {
					const daysOfTheWeek = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'];
					for (const dayItr = moment.max(moment(`${scheduleBlock.date.startdate}T${scheduleBlock.date.starttime}`), start.clone().startOf('day')); dayItr.isSameOrBefore(moment.min(moment(`${scheduleBlock.date.enddate}T${scheduleBlock.date.endtime}`), end.clone().startOf('day'))); dayItr.add(1, 'day')) {
						if (scheduleBlock.date[daysOfTheWeek[dayItr.day()]]) {
							events.push({
								id: scheduleBlock.id,
								priority: scheduleBlock.priority,
								title: scheduleBlock.name,
								backgroundColor: pastels[scheduleBlock.priority],
								textColor: 'var(--button-choice-font-color)',
								url: '#',
								allDay: false,
								scheduledatesid: scheduleBlock.date.scheduledatesid,
								start: `${dayItr.format('YYYY-MM-DD')} ${scheduleBlock.date.starttime}`,
								end: `${dayItr.format('YYYY-MM-DD')} ${scheduleBlock.date.endtime}`
							});
						}
					}
				}
			}
			calendar.addEventSource(events);
			calendar.render();
		}
	};
	calendar = new FullCalendar.Calendar($calendar, {
		headerToolbar: {
			left: 'prev,next today',
			center: 'title',
			right: 'dayGridMonth,timeGridWeek,timeGridDay'
		},
		eventOrder: (a, b) => a.priority - b.priority,
		initialView: ['dayGridMonth', 'timeGridWeek', 'timeGridDay'].indexOf(localStorage.schedulecalendarview) == -1 ? 'dayGridMonth' : localStorage.schedulecalendarview,
		selectable: true,
		editable: false,
		height,
		aspectRatio: width / height,
		slotDuration: '00:15:00',
		allDaySlot: false,
		buttonIcons: false,
		buttonText: {
			today: 'Today',
			month: 'Month',
			week: 'Week',
			day: 'Day',
			prev: '',
			next: ''
		},
		eventDisplay: 'block',
		eventClick: async ({ event }) => {
			if (event.title == 'Default') {
				ShowUpdateSchedule();
			} else {
				gScheduleBlockId = event.id;
				gScheduleBlockName = event.title;
				ShowUpdateScheduleBlock();
			}
			return false;
		},
		datesSet: ({ start, end, view }) => {
			localStorage.schedulecalendarview = view.type;
			const $container = document.querySelector('.fc-scroller-liquid-absolute');
			if (localStorage.schedulecalendarview == 'timeGridWeek') {
				setTimeout(() => {
					if ($container.hasOwnProperty('lastScroll')) {
						$container.removeEventListener('scroll', $container.lastScroll);
					}
					$container.scrollTop = lastScroll;
					$container.addEventListener('scroll', $container.lastScroll = () => {
						lastScroll = $container.scrollTop;
					});
				}, 0);
			}
			generateCalendarEvents(start, end);
		}
	});
	LockBody();
	try {
		const data = await get('/get/scheduleblock.php', {
			scheduleid: (await $scheduleTree.getSelected())[0].scheduleid
		});
		UnlockBody();
		if (data.success) {
			scheduleBlocks = data.scheduleblocks;
			generateCalendarEvents();
		} else {
			if ('treeoutdated' in data) {
				TreeOutOfDate('schedule');
				await $scheduleTree.reload();
			}
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
	}
	
	SetupCalendarTooltips();
}

function destroyCalendar() {
	if (calendar) {
		calendar.destroy();
		calendar = null;
	}
}

function setDefaultValues() {
	$startDate.value = moment().format('YYYY-MM-DD');
	$endDate.value = moment().format('YYYY-MM-DD');
	$startTime.value = '00:00:00';
	$endTime.value = '23:59:59';
	checkAllDays(true);
	disableMidnightCrossover(true);
}

function isAnyDayChecked()
{
	return ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'].some(day => document.getElementById(day).checked);
}

function enableMidnightCrossover(toggleHidden) {
	document.getElementById('crossoverSpan').innerText = "Continuous play is enabled";
	document.getElementById('enableCrossover').innerText = 'Disable';
	if (toggleHidden) {
		document.getElementById('daylist').hidden = true;
	}
}

function disableMidnightCrossover(toggleHidden) {
	document.getElementById('crossoverSpan').innerText = 'Continuous play is disabled';
	document.getElementById('enableCrossover').innerText = 'Enable';
	if (toggleHidden) {
		document.getElementById('daylist').hidden = false;
	}
}

function checkAllDays(value) {
	for (const day of ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday']) {
		document.getElementById(day).checked = value;
	}
}

function toggleAllDays() {
	var anyDayChecked = isAnyDayChecked();
	if (anyDayChecked) {
		checkAllDays(false);
		enableMidnightCrossover(true);
	} else {
		checkAllDays(true);
		disableMidnightCrossover(true);
	}
}

function defaultCheckCrossover() {
	var allDaysUnchecked = !isAnyDayChecked();
	if (allDaysUnchecked) {
		enableMidnightCrossover(true);
	} else {
		disableMidnightCrossover(true);
	}
}

function selectPlaylist(id) {
	for (const $playlist of this.parentNode.querySelectorAll('.playlist')) $playlist.classList.remove('selected');
	this.classList.add('selected');
}

function isStrLenValid(str,min,max) {
	strLen = str.length;
	return !((strLen < min) || (strLen > max));
}

function checkIsStrLenValid(id, min, max) {
	return isStrLenValid(document.getElementById(id).value, min, max);
}

function GetScheduleFormData() {
	var scheduleName = document.getElementById('scheduleName').value;
	var defaultPlaylistIds = [];
	for (const $playlist of document.getElementById('defaultPlaylists').children) {
		defaultPlaylistIds.push($playlist.data.playlistid);
	}
	return [
		scheduleName,
		defaultPlaylistIds
	];
}

async function submitSchedule()
{
	const $scheduleTree = document.getElementById('scheduleTree');
	const [scheduleName, defaultPlaylistIds] = GetScheduleFormData();
	const selectedNode = await $scheduleTree.getSelected();
	LockBody();
	const data = await $scheduleTree.addItems([
		{
			type: 2,
			name: scheduleName,
			parent: selectedNode[0].id,
			defaultplaylistids: defaultPlaylistIds
		}
	]);
	if (data.success) {
		await $scheduleTree.setSelected([data.added[0].id], false, false, true);
	}
	UnlockBody();
}

async function updateSchedule() {
	const $scheduleTree = document.getElementById('scheduleTree');
	const [scheduleName, defaultPlaylistIds] = GetScheduleFormData();
	const selectedNode = await $scheduleTree.getSelected();
	LockBody();
	if (await $scheduleTree.updateItems([
		{
			id: selectedNode[0].id,
			defaultplaylistids: defaultPlaylistIds
		}
	])) {
		UnlockBody();
		ShowScheduleCalendar();
	} else {
		UnlockBody();
	}
}

async function deleteScheduleBlock() {
	const $scheduleTree = document.getElementById('scheduleTree');
	if (await QuestionBox("Remove schedule block?", `Are you sure you want to delete the schedule block '${gScheduleBlockName}'?`)) {
		LockBody();
		try {
			var data = await post('/post/deleteScheduleBlock.php',
			{
				syncid: $scheduleTree.syncId,
				eventId: gScheduleBlockDatesId
			});
			UnlockBody();
			CheckSessionExpired(data);
			if (data.success) {
				$scheduleTree.syncId = data.syncid;
				ShowScheduleCalendar();
			} else {
				if ('treeoutdated' in data) {
					TreeOutOfDate('schedule');
					await $scheduleTree.reload();
				}
			}
		} catch (exception) {
			UnlockBody();
			NetworkingError();
		}
	}
}

async function moveScheduleBlockUp() {
	const $scheduleTree = document.getElementById('scheduleTree');
	LockBody();
	try {
		var data = await post('/post/movescheduleblock.php',
		{
			syncid: $scheduleTree.syncId,
			scheduleid: (await $scheduleTree.getSelected())[0].scheduleid,
			scheduleblockid: gScheduleBlockId,
			direction: 'up'
		});
		UnlockBody();
		CheckSessionExpired(data);
		if (data.success) {
			$scheduleTree.syncId = data.syncid;
			ShowScheduleCalendar();
		} else {
			if ('treeoutdated' in data) {
				TreeOutOfDate('schedule');
				await $scheduleTree.reload();
			}
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
	}
}

async function moveScheduleBlockDown() {
	const $scheduleTree = document.getElementById('scheduleTree');
	LockBody();
	try {
		var data = await post('/post/movescheduleblock.php',
		{
			syncid: $scheduleTree.syncId,
			scheduleid: (await $scheduleTree.getSelected())[0].scheduleid,
			scheduleblockid: gScheduleBlockId,
			direction: 'down'
		});
		UnlockBody();
		CheckSessionExpired(data);
		if (data.success) {
			$scheduleTree.syncId = data.syncid;
			ShowScheduleCalendar();
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
	}
}

function GetScheduleBlockFormData() {
	let startDate = $startDate.value;
	if (startDate == '') {
		startDate = moment().format('YYYY-MM-DD');
		$startDate.value = startDate;
	}
	let startTime = $startTime.value;
	if (startTime == '') {
		startTime = '00:00:00';
		$startTime.value = startTime;
	}
	let endDate = $endDate.value;
	if (endDate == '') {
		endDate = moment().format('YYYY-MM-DD');
		$endDate.value = endDate;
	}
	let endTime = $endTime.value;
	if (endTime == '') {
		endTime = '23:59:59';
		$endTime.value = endTime;
	}
	const scheduleBlockName = document.getElementById('scheduleBlockName').value;
	const sunday = document.getElementById('sunday').checked;
	const monday = document.getElementById('monday').checked;
	const tuesday = document.getElementById('tuesday').checked;
	const wednesday = document.getElementById('wednesday').checked;
	const thursday = document.getElementById('thursday').checked;
	const friday = document.getElementById('friday').checked;
	const saturday = document.getElementById('saturday').checked;
	const parsedStartDate = moment(startDate, 'YYYY-MM-DD');
	const parsedStartTime = moment(startTime, 'HH:mm:ss').diff(moment().startOf('day'));
	const parsedEndDate = moment(endDate, 'YYYY-MM-DD');
	const parsedEndTime = moment(endTime, 'HH:mm:ss').diff(moment().startOf('day'));
	
	if (scheduleBlockName == '') {
		ConfirmationBox('Invalid schedule block name', "Please enter a schedule block name.");
		throw new Error();
	}
	
	if (!/^[ A-Za-z0-9-]{1,100}$/.test(scheduleBlockName)) {
		ConfirmationBox('Invalid schedule block name', "Please enter a schedule block name using only letters, numbers, hyphens, or spaces. It must be between 1 and 100 characters.");
		throw new Error();
	}
	
	const playlists = [];
	for (const $playlist of document.getElementById('rightPlaylists').children) {
		playlists.push($playlist.data.playlistid);
	}
	
	if (playlists.length == 0) {
		ConfirmationBox('Invalid schedule block', 'Please select at least one playlist.');
		throw new Error();
	}
	
	if (parsedStartDate.isSame(parsedEndDate) && parsedStartTime == parsedEndTime) {
		ConfirmationBox('Invalid schedule block', 'The start date and time cannot be equal to the end date and time.');
		throw new Error();
	}
	
	if (isAnyDayChecked() && parsedStartTime >= parsedEndTime) {
		ConfirmationBox('Invalid schedule block', 'The start time must be before the end time.');
		throw new Error();
	}
	
	const isRecurringEvent = isAnyDayChecked();
	const scheduleBlockType = isRecurringEvent ? 2 : 3;
	
	return [
		scheduleBlockName,
		moment(parsedStartDate).format('YYYY-MM-DD'),
		moment().startOf('day').add(parsedStartTime).format('HH:mm:ss'),
		moment(parsedEndDate).format('YYYY-MM-DD'),
		moment().startOf('day').add(parsedEndTime).format('HH:mm:ss'),
		sunday,
		monday,
		tuesday,
		wednesday,
		thursday,
		friday,
		saturday,
		scheduleBlockType,
		playlists
	];
}

async function createNewScheduleBlock() {
	const $scheduleTree = document.getElementById('scheduleTree');
	try {
		const [
			scheduleBlockName,
			startDate,
			startTime,
			endDate,
			endTime,
			sunday,
			monday,
			tuesday,
			wednesday,
			thursday,
			friday,
			saturday,
			scheduleBlockType,
			playlistIds] = GetScheduleBlockFormData();
		LockBody();
		try {
			var data = await post('/post/submitNewScheduleBlock.php', {
				syncid: $scheduleTree.syncId,
				newScheduleBlockName: scheduleBlockName,
				newStartDate: startDate,
				newStartTime: startTime,
				newEndDate: endDate,
				newEndTime: endTime,
				scheduleId: (await $scheduleTree.getSelected())[0].scheduleid,
				newSunday: sunday,
				newMonday: monday,
				newTuesday: tuesday,
				newWednesday: wednesday,
				newThursday: thursday,
				newFriday: friday,
				newSaturday: saturday,
				scheduleBlockType: scheduleBlockType,
				playlistIds: JSON.stringify(playlistIds)
			});
			UnlockBody();
			CheckSessionExpired(data);
			if (data.success) {
				$scheduleTree.syncId = data.syncid;
				ShowScheduleCalendar();
			} else {
				if ('treeoutdated' in data) {
					TreeOutOfDate('schedule');
					await $scheduleTree.reload();
				} else {
					ConfirmationBox('Schedule Block Error', `There was a problem submitting the new schedule block. ${data.error}`);
				}
			}
		} catch (exception) {
			UnlockBody();
			NetworkingError();
		}
	} catch (exception) {
	}
}

async function clickUpdateScheduleBlock() {
	const $scheduleTree = document.getElementById('scheduleTree');
	try {
		const [
			scheduleBlockName,
			startDate,
			startTime,
			endDate,
			endTime,
			sunday,
			monday,
			tuesday,
			wednesday,
			thursday,
			friday,
			saturday,
			scheduleBlockType,
			playlistIds] = GetScheduleBlockFormData();
		LockBody();
		try {
			var data = await post('/post/updateScheduleBlock.php',
			{
				syncid: $scheduleTree.syncId,
				updateScheduleBlockName: scheduleBlockName,
				eventId: gScheduleBlockDatesId,
				startDate: startDate,
				startTime: startTime,
				endDate: endDate,
				endTime: endTime,
				updateSunday: sunday,
				updateMonday: monday,
				updateTuesday: tuesday,
				updateWednesday: wednesday,
				updateThursday: thursday,
				updateFriday: friday,
				updateSaturday: saturday,
				scheduleId: (await $scheduleTree.getSelected())[0].scheduleid,
				scheduleBlockType: scheduleBlockType,
				playlistIds: JSON.stringify(playlistIds)
			});
			UnlockBody();
			CheckSessionExpired(data);
			if (data.success) {
				$scheduleTree.syncId = data.syncid;
				ShowScheduleCalendar();
			} else {
				if ('treeoutdated' in data) {
					TreeOutOfDate('schedule');
					await $scheduleTree.reload();
				} else {
					ConfirmationBox('Problem updating schedule block', `There was a problem updating the schedule block. ${data.error}`);
				}
			}
		} catch (exception) {
			UnlockBody();
			NetworkingError();
		}
	} catch (exception) {
	}
}

async function requestAllPlaylists(container) {
	await document.getElementById(container).reload();
}

async function leftToRight(from, to) {
	const selected = (await from.getSelected())[0];
	if (selected && !selected.group) {
		AppendPlaylist(selected.data, to);
		const toOffset = $.offset(to);
		const toSelectedOffset = $.offset(to.querySelector('.selected'));
		to.scrollTop = toSelectedOffset.top - toOffset.top + to.scrollTop;
	}
}

function rightToLeft(from, to) {
	from.querySelector('.selected').remove();
	from.querySelector(':scope > :first-child').classList.add('selected');
}

function moveUp(from) {
	const $selectedPlaylist = from.querySelector('.selected');
	$selectedPlaylist.previousElementSibling?.insertAdjacentElement('beforebegin', $selectedPlaylist);
}

function moveDown(from) {
	const $selectedPlaylist = from.querySelector('.selected');
	$selectedPlaylist.nextElementSibling?.insertAdjacentElement('afterend', $selectedPlaylist);
}

async function requestRightScheduleBlocks() {
	const $scheduleTree = document.getElementById('scheduleTree');
	try {
		const data = await get('/get/contentscheduleblocks.php', {
			scheduleid: (await $scheduleTree.getSelected())[0].scheduleid
		});
		CheckSessionExpired(data);
		if (data.success) {
			document.getElementById('scheduleBlockContent').innerHTML = '';
			const pastels = CreatePastels(data.scheduleblocks.length);
			data.scheduleblocks.forEach((scheduleBlock, scheduleBlockIndex) => {
				const scheduleBlockName = scheduleBlock['name'];
				const scheduleBlockId = scheduleBlock['scheduleblockid'];
				if (scheduleBlockId != 1) {
					const checkIsDefault = scheduleBlockName.substr(0, 7);
					if (checkIsDefault != "default") {
						document.getElementById('scheduleBlockContent').appendChild(
							$.div({ id: 'scheduleBlock' + scheduleBlockId, title: scheduleBlockName, class: 'button-choices', css: { backgroundColor: pastels[scheduleBlockIndex] }, click: () => {
								gScheduleBlockId = scheduleBlockId;
								gScheduleBlockName = scheduleBlockName;
								ShowUpdateScheduleBlock();
							} },
								$.i({ class: 'mdi mdi-calendar-outline' }),
								$.span({ text: scheduleBlockName })
							)
						);
					}
				}
			});
		} else {
			if ('treeoutdated' in data) {
				TreeOutOfDate('schedule');
				await $scheduleTree.reload();
			}
		}
		return !data.success;
	} catch (exception) {
		NetworkingError();
		return true;
	}
}

function AppendPlaylist(playlist, to) {
	var thumbnailWidth = parseInt(playlist.pagewidth, 10);
	var thumbnailHeight = parseInt(playlist.pageheight, 10);
	var scale = 1;
	if (Math.max(thumbnailWidth, thumbnailHeight) > 170) {
		scale = 170 / Math.max(thumbnailWidth, thumbnailHeight);
	}
	thumbnailWidth *= scale;
	thumbnailHeight *= scale;
	to.querySelector('.selected')?.classList?.remove('selected');
	var loadImageAttempt = 0;
	const $playlist = to.appendChild($.div({ title: playlist.name, class: 'playlist selected', pointerdown: selectPlaylist, draggable: {
		thresholdX: to,
		sortable: 'y',
		start: function() { return this; }
	} },
		$.img({ src: `thumbnails/${account}/${playlist.playlistid}.jpg?random=${Math.floor(Math.random() * 100000)}`, css: {
			width: `${thumbnailWidth}px`,
			height: `${thumbnailHeight}px`
		}, load: function() {
			this.parentNode.classList.remove('busy');
		}, error: function() {
			loadImageAttempt++;
			this.parentNode.classList.add('busy');
			if (loadImageAttempt < 3) {
				setTimeout(() => {
					this.src = `thumbnails/${account}/${playlist.playlistid}.jpg?random=${Math.floor(Math.random() * 100000)}`;
				}, 30 * 1000);
			}
		} }),
		$.div({ text: playlist.name }),
		$.span({ text: humanizeDuration(playlist.runlength * 1000) })
	));
	$playlist.data = playlist;
}

async function requestDefaultPlaylists() {
	const $scheduleTree = document.getElementById('scheduleTree');
	try {
		const data = await get('/get/contentfailsafeplaylistsright.php', {
			syncid: $scheduleTree.syncId,
			scheduleId: (await $scheduleTree.getSelected())[0].scheduleid
		});
		CheckSessionExpired(data);
		if (data.success) {
			const $defaultPlaylists = document.getElementById('defaultPlaylists');
			for (const playlist of data.playlists) {
				AppendPlaylist(playlist, $defaultPlaylists);
			}
			const defaultPlaylistsOffset = $.offset($defaultPlaylists);
			for (const $playlist of $defaultPlaylists.children) $playlist.classList.remove('selected');
			$defaultPlaylists.children[0]?.classList?.add('selected');
			const $selectedPlaylist = $defaultPlaylists.querySelector('.selected');
			if ($selectedPlaylist) {
				const selectedPlaylistOffset = $.offset($selectedPlaylist);
				$defaultPlaylists.scrollTop = selectedPlaylistOffset.top - defaultPlaylistsOffset.top + $defaultPlaylists.scrollTop;
			}
		} else {
			if ('treeoutdated' in data) {
				TreeOutOfDate('schedule');
				await $scheduleTree.reload();
			}
		}
		return !data.success;
	} catch (exception) {
		NetworkingError();
		return true;
	}
}

async function requestScheduleBlockData() {
	const $scheduleTree = document.getElementById('scheduleTree');
	try {
		const data = await get('/get/contentscheduleblockdata.php', {
			syncid: $scheduleTree.syncId,
			scheduleblockid: gScheduleBlockId
		});
		CheckSessionExpired(data);
		if (data.success) {
			gScheduleBlockDatesId = data.scheduleblock.scheduledatesid;
			
			$startDate.value = moment(data.scheduleblock.startdate).format('YYYY-MM-DD');
			$endDate.value = moment(data.scheduleblock.enddate).format('YYYY-MM-DD');
			$startTime.value = moment(data.scheduleblock.starttime, 'HH:mm:ss').format('HH:mm:ss');
			$endTime.value = moment(data.scheduleblock.endtime, 'HH:mm:ss').format('HH:mm:ss');
			
			for (const day of ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday']) {
				document.getElementById(day).checked = data.scheduleblock[day];
			}
			const $rightPlaylists = document.getElementById('rightPlaylists');
			$rightPlaylists.innerHTML = '';
			for (const playlist of data.rightplaylists) {
				AppendPlaylist(playlist, $rightPlaylists);
			}
			// TODO: What does this do again? Does it just scroll to the item? Replace with scrollIntoView()?
			const rightPlaylistsOffset = $.offset($rightPlaylists);
			$rightPlaylists.children[0]?.classList?.add('selected');
			if ($rightPlaylists.querySelector('.selected')) {
				const selectedOffset = $.offset($rightPlaylists.querySelector('.selected'));
				$rightPlaylists.scrollTop = selectedOffset.top - rightPlaylistsOffset.top + $rightPlaylists.scrollTop;
			}
			defaultCheckCrossover();
		} else {
			if ('treeoutdated' in data) {
				TreeOutOfDate('schedule');
				await $scheduleTree.reload();
			}
		}
		return !data.success;
	} catch (exception) {
		NetworkingError();
		return true;
	}
}

function applyStartEndDateClamping($element) {
	let date = $element.value;
	if (date == '') {
		date = moment().format('YYYY-MM-DD');
		$element.value = date;
	} else {
		const parsedDate = moment(date, 'YYYY-MM-DD');
		if (parsedDate.isBefore(moment($element.getAttribute('min'), 'YYYY-MM-DD'))) {
			$element.value = $element.getAttribute('min');
		}
		if (parsedDate.isAfter(moment($element.getAttribute('max'), 'YYYY-MM-DD'))) {
			$element.value = $element.getAttribute('max');
		}
	}
}

function startDateChanged() {
	applyStartEndDateClamping($startDate);
	applyStartEndDateClamping($endDate);
	const startDate = $startDate.value;
	const endDate = $endDate.value;
	const parsedStartDate = moment(startDate, 'YYYY-MM-DD');
	const parsedEndDate = moment(endDate, 'YYYY-MM-DD');
	if (parsedStartDate.isAfter(parsedEndDate)) {
		$endDate.value = startDate;
	}
}

function endDateChanged() {
	applyStartEndDateClamping($startDate);
	applyStartEndDateClamping($endDate);
	const startDate = $startDate.value;
	const endDate = $endDate.value;
	const parsedStartDate = moment(startDate, 'YYYY-MM-DD');
	const parsedEndDate = moment(endDate, 'YYYY-MM-DD');
	if (parsedStartDate.isAfter(parsedEndDate)) {
		$startDate.value = endDate;
	}
}

function applyDwell() {
	let totalSeconds = 0;
	for (const $playlist of document.querySelectorAll('#rightPlaylists .playlist')) {
		totalSeconds += $playlist.data.runlength;
	}
	$endTime.value = moment($startTime.value, 'HH:mm:ss').add(totalSeconds, 'seconds').format('HH:mm:ss');
}

window.addEventListener('load', () => {
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'scheduler-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'Schedules' }),
				$.div({ id: 'scheduler-iconlist', class: 'iconlist' }),
				$.uitree({ id: 'scheduleTree', class: 'treeContainer', config: {
					tree: 'schedule',
					dispatchChangeOnLoad: true,
					editable: node => node.parentNode != null,
					leafIcon: node => 'schedule',
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length != 1) {
							ConfirmationBox('Selection required', 'Select a group or schedule to delete.');
							return;
						}
						var selectedNode = selectedNodes[0];
						if (selectedNode.group) {
							var scheduleCount = selectedAllChildren.filter(node => !node.group).length;
							if (scheduleCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${scheduleCount} schedule${scheduleCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove schedule?", `Are you sure you want to remove the schedule '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				} })
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'scheduler-center', class: 'panel' },
			$.header({ id: 'windowHeader', text: 'New Schedule' }),
			$.span({ id: 'scheduler-tip', class: 'panel-tip', text: 'Create a new schedule using the button in the left panel or select an existing schedule to begin.' }),
			$.div({ id: 'calendar', class: 'calendar-fullview', css: { flex: '1 1 0', minHeight: 0 } }),
			$.div({ id: 'scheduleBlockForm', css: { flex: '1 1 0', minHeight: 0, display: 'flex', flexFlow: 'column nowrap' } },
				$.div({ class: 'formgrid', css: { paddingBottom: 0 } },
					$.label({ text: 'Schedule Block Name' }),
					$.input({ type: 'text', class: 'sampleInput', id: 'scheduleBlockName' })
				),
				$.div({ class: 'selectionArea' },
					$.div({ class: 'panel' },
						$.header({ text: 'Available Playlists' }),
						$.uitree({ id: 'leftPlaylists', class: 'selectbox', config: {
							tree: 'playlist',
							showSmallLargeThumbnail: true,
							thumbnail: PlaylistTreeThumbnail,
							overrideThumbnailDimensions: PlaylistTreeThumbnailThumbnailDimensions,
							dragAppendTo: 'body'
						} })
					),
					$.ul({ class: 'arrowArea' },
						$.li({ class: 'tooltip-button mdi mdi-arrow-up', title: 'Move the selected playlist in the schedule block up', click: () => moveUp(document.getElementById('rightPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-right', title: 'Add the selected playlist on the left to the schedule block', click: () => leftToRight(document.getElementById('leftPlaylists'), document.getElementById('rightPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-left', title: 'Remove the selected playlist in the schedule block from the schedule block', click: () => rightToLeft(document.getElementById('rightPlaylists'), document.getElementById('leftPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-down', title: 'Move the selected playlist in the schedule block down', click: () => moveDown(document.getElementById('rightPlaylists')) })
					),
					$.div({ class: 'panel' },
						$.header({ text: 'Selected Playlists' }),
						$.div({ id: 'rightPlaylists', class: 'selectbox' })
					)
				),
				$.div({ class: 'panel', css: { flex: '0 0 auto' } },
					$.header({ text: 'Options' }),
					$.div({ class: 'formgrid' },
						$.div({ id: 'daylist' },
							$.uicheckbox({ id: 'sunday', value: 'sunday' }),
							$.label({ for: 'sunday', text: 'Sunday' }),
							
							$.uicheckbox({ id: 'monday', value: 'monday' }),
							$.label({ for: 'monday', text: 'Monday' }),
							
							$.uicheckbox({ id: 'tuesday', value: 'tuesday' }),
							$.label({ for: 'tuesday', text: 'Tuesday' }),
							
							$.uicheckbox({ id: 'wednesday', value: 'wednesday' }),
							$.label({ for: 'wednesday', text: 'Wednesday' }),
							
							$.uicheckbox({ id: 'thursday', value: 'thursday' }),
							$.label({ for: 'thursday', text: 'Thursday' }),
							
							$.uicheckbox({ id: 'friday', value: 'friday' }),
							$.label({ for: 'friday', text: 'Friday' }),
							
							$.uicheckbox({ id: 'saturday', value: 'saturday' }),
							$.label({ for: 'saturday', text: 'Saturday' })
						),
						$.div({ css: { margin: '5px' } },
							$.span({ id: 'crossoverSpan', text: 'Continuous play is disabled' }),
							$.button({ id: 'enableCrossover', text: 'Enable', click: toggleAllDays })
						),
						$.ul({ class: 'timeTable' },
							$.li({},
								$.label({ text: 'Start Date' }),
								$startDate = $.input({ type: 'date', class: 'sampleInput', id: 'startDate', blur: startDateChanged, min: moment().subtract(1, 'year').format('YYYY-MM-DD'), max: moment().add(10, 'year').format('YYYY-MM-DD') }),
								$.label({ text: 'Start Time' }),
								$startTime = $.input({ type: 'time', class: 'sampleInput', id: 'startTime', step: '1' }),
							),
							$.li({},
								$.label({ text: 'End Date' }),
								$endDate = $.input({ type: 'date', class: 'sampleInput', id: 'endDate', blur: endDateChanged, min: moment().subtract(1, 'year').format('YYYY-MM-DD'), max: moment().add(10, 'year').format('YYYY-MM-DD') }),
								$.label({ text: 'End Time' }),
								$endTime = $.input({ type: 'time', class: 'sampleInput', id: 'endTime', step: '1' }),
								$.button({ title: 'Sums the playlist times in this schedule block and sets the end time to that many seconds after the start time.', class: 'tooltip-button', text: 'Apply Dwell', click: applyDwell })
							)
						)
					)
				),
				$.div({ class: 'panel-buttons' },
					$.button({ id: 'createNewScheduleBlock', class: 'primary', text: 'Create', click: createNewScheduleBlock }),
					$.button({ id: 'cancelNewScheduleBlock', text: 'Cancel', click: ShowScheduleCalendar }),
					$.button({ id: 'clickUpdateScheduleBlock', class: 'primary', text: 'Update', click: clickUpdateScheduleBlock }),
					$.button({ id: 'clickCancelUpdateScheduleBlock', text: 'Cancel', click: ShowScheduleCalendar }),
					$.button({ id: 'deleteScheduleBlock', text: 'Delete Schedule Block', click: deleteScheduleBlock }),
					$.button({ id: 'moveScheduleBlockUp', text: 'Increase Schedule Block Priority', click: moveScheduleBlockUp }),
					$.button({ id: 'moveScheduleBlockDown', text: 'Decrease Schedule Block Priority', click: moveScheduleBlockDown })
				)
			),
			$.div({ id: 'scheduleForm', css: { flex: '1 1 0', minHeight: 0, display: 'flex', flexFlow: 'column nowrap' } },
				$.div({ class: 'formgrid', css: { paddingBottom: 0 } },
					$.label({ text: 'Schedule Name' }),
					$.input({ type: 'text', id: 'scheduleName', class: 'sampleInput', css: { width: '200px' } })
				),
				$.div({ class: 'selectionArea' },
					$.div({ class: 'panel' },
						$.header({ text: 'Available Playlists' }),
						$.uitree({ id: 'defaultLeftPlaylists', class: 'selectbox', config: {
							tree: 'playlist',
							showSmallLargeThumbnail: true,
							thumbnail: PlaylistTreeThumbnail,
							overrideThumbnailDimensions: PlaylistTreeThumbnailThumbnailDimensions,
							dragAppendTo: 'body'
						} })
					),
					$.ul({ class: 'arrowArea' },
						$.li({ class: 'tooltip-button mdi mdi-arrow-up', title: 'Move the selected playlist in the default schedule up', click: () => moveUp(document.getElementById('defaultPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-right', title: 'Add the selected playlist on the left to the default schedule', click: () => leftToRight(document.getElementById('defaultLeftPlaylists'), document.getElementById('defaultPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-left', title: 'Remove the selected playlist in the default schedule from the default schedule block', click: () => rightToLeft(document.getElementById('defaultPlaylists'), document.getElementById('defaultLeftPlaylists')) }),
						$.li({ class: 'tooltip-button mdi mdi-arrow-down', title: 'Move the selected playlist in the default schedule down', click: () => moveDown(document.getElementById('defaultPlaylists')) })
					),
					$.div({ class: 'panel' },
						$.header({ text: 'Selected Playlists' }),
						$.div({ id: 'defaultPlaylists', class: 'selectbox' })
					)
				),
				$.div({ class: 'panel-buttons' },
					$.button({ id: 'scheduleFormSubmit', class: 'primary', text: 'Submit' }),
					$.button({ id: 'scheduleFormCancel', click: ShowScheduleCalendar, text: 'Cancel' })
				)
			)
		),
		$.div({ id: 'scheduler-right' },
			$.div({ css: { flex: '1 1 0', minHeight: 0, display: 'flex', flexFlow: 'column nowrap' } },
				$.div({ class: 'panel', css: { flex: '0 0 auto' } },
					$.header({ text: 'Actions' }),
					$.div({ id: 'showNewScheduleBlock', class: 'button-choices tooltip-button', title: 'Add a new schedule block to play playlists between a specific date and time', click: ShowNewScheduleBlock },
						$.i({ class: 'mdi mdi-plus-circle-outline' }),
						$.span({ text: 'New Schedule Block' })
					),
					$.div({ id: 'showUpdateSchedule', class: 'button-choices tooltip-button', title: 'Edit the playlists that will play if no schedule block is configured to play at the current time', click: ShowUpdateSchedule },
						$.i({ class: 'mdi mdi-calendar-outline' }),
						$.span({ text: 'Edit Default Schedule' })
					)
				),
				$.div({ class: 'panel', css: { flex: '1 1 0', minHeight: 0 } },
					$.header({ class: 'tooltip-button', css: { cursor: 'help' }, title: 'Below are schedule blocks with playlists scheduled to show at specific dates and times', text: 'Schedule Blocks' }),
					$.div({ id: 'scheduleBlockContent' })
				)
			)
		)
	);
	document.body.classList.remove('loading');
	ShowSchedulerTip();
	for (const $e of document.querySelectorAll('#scheduleTree-addgroup, #scheduleTree-add, #scheduleTree-delete, #scheduleTree-duplicate')) $e.hidden = true;
	const $scheduleTree = document.getElementById('scheduleTree');
	$scheduleTree.addEventListener('load', async () => {
		if ((await $scheduleTree.getAllNodes()).some(node => node.group)) {
			for (const $e of document.querySelectorAll('#scheduleTree-addgroup, #scheduleTree-add, #scheduleTree-delete, #scheduleTree-duplicate')) $e.hidden = false;
		}
		FilterPlaylistOnly();
	});
	$scheduleTree.addEventListener('change', async function() {
		let selectedNode = (await this.getSelected())[0];
		if (selectedNode == undefined) {
			ShowSchedulerTip();
		} else {
			if (selectedNode.group) {
				if (selectedNode.id == 1) {
					ShowSchedulerTip();
				} else {
					ShowSchedulerTip();
				}
			} else {
				ShowScheduleCalendar();
			}
		}
	});
	$scheduleTree.addEventListener('nochange', async function() {
		// If the user clicks on a node again and it's not on the calendar, rather than editing it, the calendar is displayed
		const selectedNode = (await this.getSelected())[0];
		if (selectedNode && !selectedNode.group && document.getElementById('calendar').hidden) {
			ShowScheduleCalendar();
			return true;
		}
	});

	document.getElementById('scheduler-iconlist').append(
		$.i({ id: 'scheduleTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', title: 'Delete selected group or schedule', click: () => $scheduleTree.deleteSelected() }),
		$.span({ class: 'spacer' }),
		$.uicheckbox({ id: 'scheduleTree-filterplaylistonly', class: 'filterplaylistonly', checked: localStorage.schedulefilterplaylistonly == 'true', title: 'Filter Playlist Only', css: { marginRight: '5px' } , change: function() {
			localStorage.schedulefilterplaylistonly = this.checked;
			FilterPlaylistOnly();
		} },
			$.span({ text: 'Playlist Only' })
		),
		$.i({ id: 'scheduleTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', title: 'Add group', click: () => $scheduleTree.addGroup() }),
		$.i({ id: 'scheduleTree-add', class: 'tooltip-button mdi mdi-calendar-plus', title: 'New schedule', click: () => ShowNewSchedule() }),
		$.i({ id: 'scheduleTree-duplicate', class: 'tooltip-button mdi mdi-content-duplicate', title: 'Duplicate selected schedule', click: () => $scheduleTree.duplicateSelected() })
	);
	
	$startTime.addEventListener('change', () => {
		if ($endTime.value == '23:59:59') {
			$endTime.value = moment($startTime.value, 'HH:mm:ss').add(1, 'hour').format('HH:mm:ss');
		}
	});
	
	for (const $playlists of [document.getElementById('defaultPlaylists'), document.getElementById('rightPlaylists')]) {
		droppable($playlists,
		{
			accept: ['playlist-leaf'],
			drop: function(e, source) {
				leftToRight(source[0].$handle.closest('ui-tree'), this);
			}
		});
	}
	
	for (const $playlists of [document.getElementById('defaultLeftPlaylists'), document.getElementById('leftPlaylists')]) {
		droppable($playlists,
		{
			accept: ['playlist'],
			drop: (e, source) => {
				source.remove();
			}
		});
	}
	
	const checkCrossover = () => {
		var allDaysUnchecked = !isAnyDayChecked();
		if (allDaysUnchecked) {
			enableMidnightCrossover(false);
		} else {
			disableMidnightCrossover(false);
		}
	};
	for (const day of ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday']) {
		document.getElementById(day).addEventListener('change', checkCrossover);
	}
	
	window.addEventListener('resize', () => {
		const $calendar = document.getElementById('calendar');
		if ($calendar.innerHTML != '') {
			calendar.setOption('height', $calendar.parentNode.offsetHeight);
			calendar.setOption('aspectRatio', $calendar.parentNode.offsetWidth / $calendar.parentNode.offsetHeight);
		}
	});
	
	DragLeftResize('.horizontal-resize', 'scheduler-lr');
});