class Text {
	constructor(media) {
		this.media = media;
		this.selection = { start: 1, end: 1, leftToRight: true };
		this.defaultStyles = {
			align: 'left',
			family: defaultFont,
			size: defaultFontSize,
			lineheight: -1,
			stretch: 'normal',
			color: 'rgba(0, 0, 0, 1)',
			highlight: 'rgba(255, 255, 255, 0)',
			bold: false,
			italic: false,
			underline: false,
			strikethrough: false,
			subscript: false,
			superscript: false
		};
		this.characterStyles = this.JSONToCharacterStyles(this.media.attributes.text);
	}
	GetCaretPosition(characterStyles, x, y) {
		let index = 1;
		let error = Number.MAX_VALUE;
		let maxHeight = Number.MIN_VALUE;
		for (let i = 1; i < characterStyles.length; ++i) {
			if (Array.isArray(characterStyles[i].bounds)) {
				for (let j = 0; j < characterStyles[i].bounds.length; ++j) {
					const bounds = characterStyles[i].bounds[j];
					maxHeight = Math.max(maxHeight, bounds.top + bounds.height);
				}
			} else {
				const bounds = characterStyles[i].bounds;
				maxHeight = Math.max(maxHeight, bounds.top + bounds.height);
			}
		}
		for (let i = 1; i < characterStyles.length; ++i) {
			const bounds = characterStyles[i].bounds;
			if (Array.isArray(bounds)) {
				// A widget can stretch across multiple lines so we need to figure out the relative position of the mouse to determine if it's more on the left or right side
				let widgetWidth = 0;
				for (let j = 0; j < bounds.length; ++j) {
					widgetWidth += bounds[j].width;
				}
				let currentRelativeLeft = 0;
				for (let j = 0; j < bounds.length; ++j) {
					if ((bounds[j].top == 0 || y >= bounds[j].top) && (Math.abs(bounds[j].top + bounds[j].height - maxHeight) < 0.001 || y < bounds[j].top + bounds[j].height)) {
						if (currentRelativeLeft < widgetWidth / 2 && Math.abs(bounds[j].left - x) < error) {
							error = Math.abs(bounds[j].left - x);
							index = i;
						}
						if (currentRelativeLeft >= widgetWidth / 2 && Math.abs(bounds[j].left + bounds[j].width - x) < error) {
							error = Math.abs(bounds[j].left + bounds[j].width - x);
							index = i + 1;
						}
					}
					currentRelativeLeft += bounds[j].width;
				}
			} else if ((bounds.top == 0 || y >= bounds.top) && (Math.abs(bounds.top + bounds.height - maxHeight) < 0.001 || y < bounds.top + bounds.height)) {
				if (characterStyles[i].character != String.fromCharCode(8203)) {
					if (Math.abs(bounds.left - x) < error) {
						error = Math.abs(bounds.left - x);
						index = i;
					}
				}
				const epsilon = 0.001;
				const newError = Math.abs(bounds.left + bounds.width - x);
				if (newError - epsilon <= error) {
					error = newError;
					index = i + 1;
				}
			}
		}
		return index;
	}
	GetCharacterBounds(node, characterStyles, selection, selectionRectangles, range, currentIndex, containerBounds) {
		if (selectionRectangles === undefined) {
			const editorCanvas = document.getElementById('editor-canvas');
			const oldTransform = editorCanvas.style.transform;
			editorCanvas.style.transform = 'none';
			const selectionRectangles = [];
			containerBounds = node.getBoundingClientRect();
			const range = document.createRange();
			range.selectNodeContents(node);
			const currentIndex = { index: 0 };
			node = node.firstChild;
			while (node) {
				this.GetCharacterBounds(node, characterStyles, selection, selectionRectangles, range, currentIndex, containerBounds);
				node = node.nextSibling;
			}
			// Resize the height of each character so that each character on a line is the same height.
			let top = 0;
			let startLineWidgetCharacter = 0;
			for (let i = 0; i < characterStyles.length;) {
				// Find a paragraph's start and end
				let start = i;
				let end = i + 1;
				for (; end < characterStyles.length; ++end) {
					if (characterStyles[end].styles.hasOwnProperty('paragraph')) {
						break;
					}
				}
				if (start == end - 1) {
					// Empty paragraph. Use the height of zero-width character
					const offset = characterStyles[start].bounds.top - top;
					characterStyles[start].bounds.height += offset;
					characterStyles[start].bounds.top = top;
					top += characterStyles[start].bounds.height;
				} else {
					let maximumLeft = characterStyles[0].bounds.left;
					let startLine = start;
					for (let j = start; j <= end; ++j) {
						const processBounds = (bounds, widgetCharacter) => {
							if (bounds === undefined || Math.abs(bounds.left - maximumLeft) > 0.5) {
								// Wrap point
								let maximumLineHeight = 0;
								let endIncludeCurrentCharacter = widgetCharacter !== undefined && widgetCharacter != 0 ? j + 1 : j;
								
								for (let k = startLine; k < endIncludeCurrentCharacter; ++k) {
									if (Array.isArray(characterStyles[k].bounds)) {
										const endWidgetCharacter = widgetCharacter !== undefined ? widgetCharacter : characterStyles[k].bounds.length;
										for (let m = startLineWidgetCharacter; m < endWidgetCharacter; ++m) {
											maximumLineHeight = Math.max(maximumLineHeight, characterStyles[k].bounds[m].top + characterStyles[k].bounds[m].height - top);
										}
									} else {
										if (characterStyles[k].character != String.fromCharCode(8203)) {
											maximumLineHeight = Math.max(maximumLineHeight, characterStyles[k].bounds.top + characterStyles[k].bounds.height - top);
										}
									}
								}
								for (let k = startLine; k < endIncludeCurrentCharacter; ++k) {
									if (Array.isArray(characterStyles[k].bounds)) {
										const endWidgetCharacter = widgetCharacter !== undefined ? widgetCharacter : characterStyles[k].bounds.length;
										for (let m = startLineWidgetCharacter; m < endWidgetCharacter; ++m) {
											characterStyles[k].bounds[m].top = top;
											characterStyles[k].bounds[m].height = maximumLineHeight;
										}
									} else {
										characterStyles[k].bounds.top = top;
										characterStyles[k].bounds.height = maximumLineHeight;
									}
								}
								if (selection.start != selection.end && selection.start < endIncludeCurrentCharacter && selection.end > startLine) {
									let selectStart = Math.max(selection.start, startLine);
									let selectEnd = Math.min(selection.end, endIncludeCurrentCharacter);
									let boundsStart = characterStyles[selectStart].bounds;
									if (Array.isArray(boundsStart)) {
										boundsStart = characterStyles[selectStart].bounds[startLineWidgetCharacter];
									}
									let boundsEnd = characterStyles[selectEnd - 1].bounds;
									if (Array.isArray(boundsEnd)) {
										if (widgetCharacter === undefined) {
											boundsEnd = characterStyles[selectEnd - 1].bounds[characterStyles[selectEnd - 1].bounds.length - 1];
										} else {
											boundsEnd = characterStyles[selectEnd - 1].bounds[widgetCharacter - 1];
										}
									}
									selectionRectangles.push({
										left: boundsStart.left,
										top: boundsStart.top,
										width: Math.max(0, boundsEnd.left + boundsEnd.width - boundsStart.left),
										height: boundsStart.height
									});
								}
								if (j != end) {
									// Can bounds be undefined here?
									maximumLeft = bounds.left + bounds.width;
								}
								startLine = j;
								if (j != end && Array.isArray(characterStyles[j].bounds)) {
									if (widgetCharacter == characterStyles[j].bounds.length - 1) {
										startLineWidgetCharacter = 0;
									} else {
										startLineWidgetCharacter = widgetCharacter;
									}
								}
								top += maximumLineHeight;
							} else {
								maximumLeft = Math.max(maximumLeft, bounds.left + bounds.width);
							}
						};
						if (j == end) {
							processBounds();
						} else if (Array.isArray(characterStyles[j].bounds)) {
							for (let m = 0; m < characterStyles[j].bounds.length; ++m) {
								processBounds(characterStyles[j].bounds[m], m);
							}
						} else {
							processBounds(characterStyles[j].bounds);
						}
					}
				}
				i = end;
			}
			editorCanvas.style.transform = oldTransform;
			return selectionRectangles;
		} else {
			if (node.nodeType == 1) {
				// Element Node
				let recurse = false;
				switch (node.nodeName) {
					case 'SPAN':
						if (node.classList.contains('widget')) {
							let textNode = node.childNodes[1].childNodes[0];
							characterStyles[currentIndex.index].bounds = [];
							for (let i = 0; i < textNode.data.length; ++i) {
								range.setStart(textNode, i);
								range.setEnd(textNode, i + 1);
								const bounds = range.getBoundingClientRect();
								let left = bounds.left - containerBounds.left;
								let width = bounds.width;
								const leftSpacerBounds = node.childNodes[0].getBoundingClientRect();
								const rightSpacerBounds = node.childNodes[2].getBoundingClientRect();
								if (i == 0 && leftSpacerBounds.top >= bounds.top && leftSpacerBounds.bottom <= bounds.bottom) {
									left -= bounds.left - leftSpacerBounds.left;
									width += bounds.left - leftSpacerBounds.left;
								}
								if (i == textNode.data.length - 1 && rightSpacerBounds.top >= bounds.top && rightSpacerBounds.bottom <= bounds.bottom) {
									width += rightSpacerBounds.right - bounds.right;
								}
								characterStyles[currentIndex.index].bounds.push({ left: left, top: bounds.top - containerBounds.top, width, height: bounds.height });
							}
							currentIndex.index++;
							break;
						}
						if (node.classList.contains('caret')) {
							break;
						}
					case 'P':
					case 'B':
					case 'I':
					case 'U':
					case 'S':
					case 'SUB':
					case 'SUP':
						recurse = true;
						break;
				}
				if (recurse) {
					node = node.firstChild;
					while (node) {
						this.GetCharacterBounds(node, characterStyles, selection, selectionRectangles, range, currentIndex, containerBounds);
						node = node.nextSibling;
					}
				}
			} else if (node.nodeType == 3) {
				// Text Node
				for (let i = 0; i < node.data.length; ++i) {
					range.setStart(node, i);
					range.setEnd(node, i + 1);
					let bounds = range.getBoundingClientRect();
					let boundary = { left: bounds.left - containerBounds.left, top: bounds.top - containerBounds.top, width: bounds.width, height: bounds.height };
					if (bounds.width == 0 && bounds.height == 0) {
						bounds = node.parentNode.getBoundingClientRect();
						boundary = { left: bounds.left - containerBounds.left, top: bounds.top - containerBounds.top, width: 0, height: bounds.height };
					}
					characterStyles[currentIndex.index].bounds = boundary;
					currentIndex.index++;
				}
			}
		}
	}
	JSONToCharacterStyles(node, selection = { start: 1, end: 1, leftToRight: true }, characterStyles, currentStyle) {
		if (characterStyles == undefined) {
			characterStyles = [];
			for (const child of node.children) {
				this.JSONToCharacterStyles(child, selection, characterStyles, {});
			}
			for (let i = 0; i < characterStyles.length; ++i) {
				if (characterStyles[i].styles.hasOwnProperty('paragraph')) {
					if (i + 1 < characterStyles.length && !characterStyles[i + 1].styles.hasOwnProperty('paragraph')) {
						for (const style in characterStyles[i + 1].styles) {
							characterStyles[i].styles[style] = characterStyles[i + 1].styles[style];
						}
					}
				}
			}
			this.CharacterStylesSetDefaults(characterStyles, selection);
			return characterStyles;
		} else {
			if (node.hasOwnProperty('text')) {
				for (let i = 0; i < node.text.length; ++i) {
					characterStyles.push({ character: node.text[i], styles: JSON.parse(JSON.stringify(currentStyle)) });
				}
			} else if (node.hasOwnProperty('widget')) {
				characterStyles.push({ widget: node.widget, styles: currentStyle });
			} else if (node.hasOwnProperty('paragraph')) {
				currentStyle.align = node.paragraph;
				const newStyle = JSON.parse(JSON.stringify(currentStyle));
				newStyle.paragraph = true;
				characterStyles.push({ character: String.fromCharCode(8203), styles: newStyle });
			} else {
				for (const style in this.defaultStyles) {
					if (node.hasOwnProperty(style)) {
						currentStyle[style] = node[style];
					}
				}
			}
			if (node.hasOwnProperty('children')) {
				if (node.children.length == 0) {
					characterStyles.push({ character: String.fromCharCode(8203), styles: JSON.parse(JSON.stringify(currentStyle)) });
				}
				for (const child of node.children) {
					this.JSONToCharacterStyles(child, selection, characterStyles, JSON.parse(JSON.stringify(currentStyle)));
				}
			}
		}
	}
	CharacterStylesSetDefaults(characterStyles, selection) {
		// Remove zero width characters inside paragraphs if the caret isn't after them and it's not a paragraph character
		for (let i = 0; i < characterStyles.length; ++i) {
			if (!characterStyles[i].styles.hasOwnProperty('paragraph') && characterStyles[i].character == String.fromCharCode(8203) && (selection.start != selection.end || i + 1 != selection.start)) {
				characterStyles.splice(i, 1);
				if (selection.start > i) {
					selection.start--;
				}
				if (selection.end > i) {
					selection.end--;
				}
				if (selection.hasOwnProperty('started') && selection.started > i) {
					selection.started--;
				}
				i--;
			}
		}
		// If the cursor isn't after the paragraph then shift the style into the paragraph character
		for (let i = 0; i < characterStyles.length - 1; ++i) {
			if (characterStyles[i].styles.hasOwnProperty('paragraph') && (selection.start != selection.end || i + 1 != selection.start)) {
				if (!characterStyles[i + 1].styles.hasOwnProperty('paragraph')) {
					characterStyles[i].styles = { paragraph: true };
					for (const style in characterStyles[i + 1].styles) {
						characterStyles[i].styles[style] = characterStyles[i + 1].styles[style];
					}
				}
			}
		}
	}
	CharacterStylesToJSON(characterStyles, container, currentStyle, index, length) {
		if (container === undefined) {
			container = { children: [] };
			currentStyle = {};
			characterStyles = JSON.parse(JSON.stringify(characterStyles));
			for (let i = 0; i < characterStyles.length; ++i) {
				delete characterStyles[i].bounds;
			}
			this.CharacterStylesToJSON(characterStyles, container, currentStyle, 0, characterStyles.length);
			return container;
		} else {
			// Index and length of the longest sequence for a style
			let sequenceIndex = 0;
			let sequenceLength = 0;
			let sequenceStyle = '';
			// Find paragraphs
			if (sequenceLength == 0) {
				for (let i = index; i < index + length; ++i) {
					if (characterStyles[i].styles.hasOwnProperty('paragraph') && sequenceLength == 0) {
						sequenceIndex = i;
						sequenceLength = 1;
						for (let j = i + 1; j < index + length; ++j) {
							if (characterStyles[j].styles.hasOwnProperty('paragraph')) {
								break;
							} else {
								sequenceLength++;
							}
						}
						sequenceStyle = { style: 'paragraph', value: characterStyles[i].styles.align };
					}
				}
			}
			// If there are no paragraphs then look for the longest styles
			if (sequenceLength == 0) {
				for (let i = index; i < index + length; ++i) {
					for (const style in characterStyles[i].styles) {
						let styleLength = 1;
						for (let j = i + 1; j < index + length; ++j) {
							if (characterStyles[j].styles.hasOwnProperty(style) && characterStyles[i].styles[style] == characterStyles[j].styles[style]) {
								styleLength++;
							} else {
								break;
							}
						}
						if (styleLength > sequenceLength) {
							sequenceIndex = i;
							sequenceLength = styleLength;
							sequenceStyle = { style: style, value: characterStyles[i].styles[style] };
						}
					}
				}
			}
			if (sequenceLength == 0) {
				// Build up widget and text nodes
				let textNodeString = '';
				const InsertTextNode = () => {
					if (textNodeString != '') {
						container.children.push({ text: textNodeString });
						textNodeString = '';
					}
				};
				for (let i = index; i < index + length; ++i) {
					if (characterStyles[i].character) {
						textNodeString += characterStyles[i].character;
					} else if (characterStyles[i].widget) {
						InsertTextNode();
						container.children.push({ widget: characterStyles[i].widget });
					}
				}
				InsertTextNode();
			} else {
				const newNode = {};
				switch (sequenceStyle.style) {
					case 'paragraph':
					case 'family':
					case 'size':
					case 'lineheight':
					case 'stretch':
					case 'color':
					case 'highlight':
						newNode[sequenceStyle.style] = sequenceStyle.value;
						break;
					case 'bold':
					case 'italic':
					case 'underline':
					case 'strikethrough':
					case 'subscript':
					case 'superscript':
						newNode[sequenceStyle.style] = true;
						break;
					default:
						return;
				}
				for (let i = sequenceIndex; i < sequenceIndex + sequenceLength; ++i) {
					delete characterStyles[i].styles[sequenceStyle.style];
					if (sequenceStyle.style == 'paragraph') {
						delete characterStyles[i].styles.align;
					}
				}
				const paragraphOffset = characterStyles[sequenceIndex].character == String.fromCharCode(8203) && Object.keys(characterStyles[sequenceIndex].styles).length == 0 ? 1 : 0;
				this.CharacterStylesToJSON(characterStyles, container, currentStyle, index, sequenceIndex - index);
				newNode.children = [];
				this.CharacterStylesToJSON(characterStyles, newNode, currentStyle, sequenceIndex + paragraphOffset, sequenceLength - paragraphOffset);
				container.children.push(newNode);
				this.CharacterStylesToJSON(characterStyles, container, currentStyle, sequenceIndex + sequenceLength, index + length - (sequenceIndex + sequenceLength));
			}
		}
	}
	rgbaToColor(rgba) {
		const matches = rgba.match(/^rgba\(([0-9]+) *, *([0-9]+) *, *([0-9]+) *, *([0-9]+(?:\.[0-9]*)?)\)$/);
		if (matches && matches.length == 5) {
			return {
				r: Math.min(Math.max(0, parseInt(matches[1], 10)), 255),
				g: Math.min(Math.max(0, parseInt(matches[2], 10)), 255),
				b: Math.min(Math.max(0, parseInt(matches[3], 10)), 255),
				a: Math.min(Math.max(0, Math.round(parseFloat(matches[4]) * 100)), 100)
			};
		} else {
			return {
				r: 0,
				g: 0,
				b: 0,
				a: 100
			};
		}
	}
	CharacterStylesToDOM($container, characterStyles, selection, applySelectionStyleChanges = true) {
		if (applySelectionStyleChanges) {
			if (selection.start != selection.end) {
				characterStyles = JSON.parse(JSON.stringify(characterStyles));
			}
			// Adjust the highlight and text color for selected text
			for (let i = selection.start; i < selection.end; ++i) {
				//characterStyles[i].styles.color = 'rgba(255, 255, 255, 1)';
				if (characterStyles[i].styles.hasOwnProperty('highlight')) {
					const { r, g, b, a } = this.rgbaToColor(characterStyles[i].styles.highlight);
					characterStyles[i].styles.highlight = `rgba(${r}, ${g}, ${b}, ${Math.min(a, 0.2)})`;
				}
			}
		}
		const padding = { left: this.media.attributes.paddingleft, right: this.media.attributes.paddingright, top: this.media.attributes.paddingtop, bottom: this.media.attributes.paddingbottom };
		// TODO: $cursorCapture variable isn't used. Remove it?
		let $cursorCapture;
		if ($container.querySelector(':scope > div.cursorcapture') == null) {
			$cursorCapture = $container.appendChild($.div({ class: 'cursorcapture' }));
		} else {
			$cursorCapture = $container.querySelector(':scope > div.cursorcapture');
		}
		let $text;
		if ($container.querySelector(':scope > div.text') == null) {
			$text = document.createElement('div');
			$text.removeAttribute('contenteditable');
			$text.style.left = `${padding.left}px`;
			$text.style.right = `${padding.right}px`;
			$text.style.top = `${padding.top}px`;
			$text.style.bottom = `${padding.bottom}px`;
			$text.classList.add('text');
			$container.appendChild($text);
		} else {
			$text = $container.querySelector(':scope > div.text');
		}
		// Clear the text control
		for (const $element of $text.querySelectorAll(':scope > p')) {
			$element.remove();
		}
		
		let $background;
		if ($container.querySelector(':scope > div.background') == null) {
			$background = document.createElement('div');
			$background.classList.add('background');
			$background.style.zIndex = -2;
			$container.appendChild($background);
		} else {
			$background = $container.querySelector(':scope > div.background');
		}
		$background.style.backgroundColor = this.media.attributes.background;
		
		let $selectionCanvas;
		// Create the selection canvas if one doesn't exist
		if ($container.querySelector(':scope > canvas') == null) {
			$selectionCanvas = document.createElement('canvas');
			$selectionCanvas.style.position = 'absolute';
			$selectionCanvas.style.left = `${padding.left}px`;
			$selectionCanvas.style.top = `${padding.top}px`;
			$selectionCanvas.style.zIndex = -1;
			$selectionCanvas.width = Math.floor(Math.max(0, $container.clientWidth - (padding.left + padding.right)));
			$selectionCanvas.height = Math.floor(Math.max(0, $container.clientHeight - (padding.top + padding.bottom)));
			$container.appendChild($selectionCanvas);
		} else {
			$selectionCanvas = $container.querySelector(':scope > canvas');
		}
		
		let $caret;
		if ($container.querySelector(':scope > .caret') == null) {
			$caret = document.createElement('span');
			$caret.classList.add('caret');
			$caret.innerHTML = String.fromCharCode(8203);
			$container.appendChild($caret);
		} else {
			$caret = $container.querySelector(':scope > .caret');
		}
		
		let temporaryCaret = null;
		
		const buildDOM = ($node, characterStyles, index, length) => {
			if (length == 0) {
				return;
			}
			// Index and length of the longest sequence for a style
			let sequenceIndex = 0;
			let sequenceLength = 0;
			let sequenceStyle = '';
			// Find paragraphs
			if (sequenceLength == 0) {
				for (let i = index; i < index + length; ++i) {
					if (characterStyles[i].styles.hasOwnProperty('paragraph') && sequenceLength == 0) {
						sequenceIndex = i;
						sequenceLength = 1;
						for (let j = i + 1; j < index + length; ++j) {
							if (characterStyles[j].styles.hasOwnProperty('paragraph')) {
								break;
							} else {
								sequenceLength++;
							}
						}
						sequenceStyle = { style: 'paragraph', value: characterStyles[i].styles.align };
					}
				}
			}
			const priorityStyle = ['lineheight', 'size', 'family'];
			for (const style of priorityStyle) {
				if (sequenceLength == 0) {
					for (let i = index; i < index + length; ++i) {
						if (characterStyles[i].styles.hasOwnProperty(style)) {
							let styleLength = 1;
							for (let j = i + 1; j < index + length; ++j) {
								if (characterStyles[j].styles.hasOwnProperty(style) && characterStyles[i].styles[style] == characterStyles[j].styles[style]) {
									styleLength++;
								} else {
									break;
								}
							}
							if (styleLength > sequenceLength) {
								sequenceIndex = i;
								sequenceLength = styleLength;
								sequenceStyle = { style: style, value: characterStyles[i].styles[style] };
							}
						}
					}
				}
			}
			// If there are no paragraphs or priority styles then look for the other longest styles
			if (sequenceLength == 0) {
				for (let i = index; i < index + length; ++i) {
					for (const style in characterStyles[i].styles) {
						let styleLength = 1;
						for (let j = i + 1; j < index + length; ++j) {
							if (characterStyles[j].styles.hasOwnProperty(style) && characterStyles[i].styles[style] == characterStyles[j].styles[style]) {
								styleLength++;
							} else {
								break;
							}
						}
						if (styleLength > sequenceLength) {
							sequenceIndex = i;
							sequenceLength = styleLength;
							sequenceStyle = { style: style, value: characterStyles[i].styles[style] };
						}
					}
				}
			}
			if (sequenceLength == 0) {
				// Build up widget and text nodes
				let textNodeString = '';
				const InsertTextNode = () => {
					if (textNodeString != '') {
						const textNode = document.createTextNode(textNodeString);
						$node.appendChild(textNode);
						textNodeString = '';
					}
				};
				const InsertCaret = i => {
					if (temporaryCaret == null && (i == selection.start && !selection.leftToRight || i == selection.end && selection.leftToRight)) {
						InsertTextNode(i);
						// This span tag inherits all the paragraph properties including line height offsets and gives us the left, top, and height information that a caret would have
						temporaryCaret = document.createElement('span');
						temporaryCaret.innerHTML = String.fromCharCode(8203);
						$node.appendChild(temporaryCaret);
					}
				};
				for (var i = index; i < index + length; ++i) {
					InsertCaret(i);
					if (characterStyles[i].character) {
						textNodeString += characterStyles[i].character;
					} else if (characterStyles[i].widget) {
						InsertTextNode(i);
						const $widget = document.createElement('span');
						$widget.classList.add('widget');
						$widget.removeAttribute('contenteditable');
						
						const $widgetSpacer1 = document.createElement('span');
						$widgetSpacer1.classList.add('widget-spacer');
						$widget.appendChild($widgetSpacer1);
						
						const $widgetText = document.createElement('span');
						$widgetText.classList.add('widget-text');
						$widgetText.innerHTML = characterStyles[i].widget;
						$widget.appendChild($widgetText);
						
						const $widgetSpacer2 = document.createElement('span');
						$widgetSpacer2.classList.add('widget-spacer');
						$widget.appendChild($widgetSpacer2);
						
						$node.appendChild($widget);
					}
				}
				InsertTextNode(i);
				InsertCaret(index + length);
			} else {
				let tag;
				switch (sequenceStyle.style) {
					case 'paragraph':
						tag = document.createElement('p');
						tag.style.textAlign = sequenceStyle.value;
						break;
					case 'family':
						tag = document.createElement('span');
						tag.style.fontFamily = sequenceStyle.value;
						break;
					case 'size':
						tag = document.createElement('span');
						tag.style.fontSize = `${sequenceStyle.value}px`;
						break;
					case 'lineheight':
						tag = document.createElement('span');
						tag.style.lineHeight = `${sequenceStyle.value}px`;
						break;
					case 'stretch':
						tag = document.createElement('span');
						tag.style.fontStretch = sequenceStyle.value;
						break;
					case 'color':
						tag = document.createElement('span');
						tag.style.color = sequenceStyle.value;
						break;
					case 'highlight':
						tag = document.createElement('span');
						tag.style.backgroundColor = sequenceStyle.value;
						break;
					case 'bold':
						tag = document.createElement('b');
						break;
					case 'italic':
						tag = document.createElement('i');
						break;
					case 'underline':
						tag = document.createElement('u');
						break;
					case 'strikethrough':
						tag = document.createElement('s');
						break;
					case 'subscript':
						tag = document.createElement('sub');
						break;
					case 'superscript':
						tag = document.createElement('sup');
						break;
					default:
						return;
				}
				for (let i = sequenceIndex; i < sequenceIndex + sequenceLength; ++i) {
					delete characterStyles[i].styles[sequenceStyle.style];
					if (sequenceStyle.style == 'paragraph') {
						delete characterStyles[i].styles.align;
					}
				}
				buildDOM($node, characterStyles, index, sequenceIndex - index);
				$node.appendChild(tag);
				buildDOM(tag, characterStyles, sequenceIndex, sequenceLength);
				buildDOM($node, characterStyles, sequenceIndex + sequenceLength, index + length - (sequenceIndex + sequenceLength));
			}
		};
		
		buildDOM($text, JSON.parse(JSON.stringify(characterStyles)), 0, characterStyles.length);
		// Caret calculation
		if (temporaryCaret) {
			let color = 'rgba(0, 0, 0, 1)';
			let cursorIndex = selection.leftToRight ? selection.end : selection.start;
			if (characterStyles[cursorIndex - 1].styles.hasOwnProperty('color')) {
				color = characterStyles[cursorIndex - 1].styles.color;
			}
			if (temporaryCaret == null || selection.start != selection.end) {
				color = 'transparent';
			}
			document.documentElement.style.setProperty('--caret-color', color);
			const textBoundary = $container.getBoundingClientRect();
			const bounds = temporaryCaret.getBoundingClientRect();
			temporaryCaret.remove();
			$caret.style.left = `${toCanvasScale(bounds.left - textBoundary.left)}px`;
			$caret.style.top = `${toCanvasScale(bounds.top - textBoundary.top)}px`;
			$caret.style.height = `${toCanvasScale(bounds.height)}px`;
			// Removing the temporaryCaret separates two text nodes sometimes. Normalize will combine text nodes ensuring kerning is correct
			$text.normalize();
		}
		// Select text by rendering the selection to the canvas element
		const context = $selectionCanvas.getContext('2d');
		context.clearRect(0, 0, $selectionCanvas.width, $selectionCanvas.height);
		context.fillStyle = '#3399FF';
		for (const rectangle of this.GetCharacterBounds($text, characterStyles, selection)) {
			context.fillRect(rectangle.left, rectangle.top, rectangle.width, rectangle.height);
		}
	}
	ApplyStyle(style, value, focusToCaret = true, applySelectionStyleChanges = true, recursive = false) {
		const defaultValue = this.defaultStyles[style];
		if (!recursive) {
			// subscript and superscript toggle each other
			if (style == 'subscript') {
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				if (styles.superscript == null || styles.superscript) {
					this.ApplyStyle('superscript', false, false, applySelectionStyleChanges, true);
				}
			} else if (style == 'superscript') {
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				if (styles.subscript == null || styles.subscript) {
					this.ApplyStyle('subscript', false, false, applySelectionStyleChanges, true);
				}
			}
		}
		this.Update((selection, characterStyles) => {
			if (selection.start == selection.end) {
				if (style == 'align') {
					// Select the whole paragraph
					let start = selection.start;
					let end = selection.end;
					for (; start > 0 && (start == characterStyles.length || start == selection.start && characterStyles[start].styles.hasOwnProperty('paragraph') || !characterStyles[start].styles.hasOwnProperty('paragraph')); start--) {
					}
					for (; end < characterStyles.length && !characterStyles[end].styles.hasOwnProperty('paragraph'); end++) {
					}
					for (let i = start; i < end; ++i) {
						if (value != defaultValue) {
							characterStyles[i].styles[style] = value;
						} else {
							characterStyles[i].styles[style] = 'left';
						}
					}
				} else {
					if (value != defaultValue) {
						if (characterStyles[selection.start - 1].character == String.fromCharCode(8203)) {
							characterStyles[selection.start - 1].styles[style] = value;
						} else if (selection.start == characterStyles.length || characterStyles[selection.start].styles.hasOwnProperty('paragraph') || characterStyles[selection.start].character == ' ' || characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
							const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start - 1].styles));
							newStyle[style] = value;
							characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
							selection.start++;
							selection.end++;
						} else {
							let start = selection.start;
							let end = selection.end;
							for (; start > 0 && !characterStyles[start].styles.hasOwnProperty('paragraph') && characterStyles[start - 1].character != ' '; start--) {
							}
							for (; end < characterStyles.length && !characterStyles[end].styles.hasOwnProperty('paragraph') && characterStyles[end].character != ' '; end++) {
							}
							for (let i = start; i < end; ++i) {
								characterStyles[i].styles[style] = value;
							}
						}
					} else {
						if (characterStyles[selection.start - 1].character == String.fromCharCode(8203)) {
							delete characterStyles[selection.start - 1].styles[style];
						} else if (selection.start == characterStyles.length || characterStyles[selection.start].styles.hasOwnProperty('paragraph') || characterStyles[selection.start].character == ' ' || characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
							const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start - 1].styles));
							delete newStyle[style];
							characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
							selection.start++;
							selection.end++;
						} else {
							let start = selection.start;
							let end = selection.end;
							for (; start > 0 && !characterStyles[start].styles.hasOwnProperty('paragraph') && characterStyles[start - 1].character != ' '; start--) {
							}
							for (; end < characterStyles.length && !characterStyles[end].styles.hasOwnProperty('paragraph') && characterStyles[end].character != ' '; end++) {
							}
							for (let i = start; i < end; ++i) {
								delete characterStyles[i].styles[style];
							}
						}
					}
				}
			} else {
				if (value != defaultValue) {
					// Add the style to every selected character
					for (let i = selection.start; i < selection.end; ++i) {
						characterStyles[i].styles[style] = value;
					}
					if (style == 'align') {
						if (characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
							characterStyles[selection.start - 1].styles[style] = value;
						} else if (characterStyles[selection.start - 1].styles.align != value) {
							const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start].styles));
							newStyle.paragraph = true;
							characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
							selection.start++;
							selection.end++;
						}
						if (selection.end != characterStyles.length && !characterStyles[selection.end].styles.hasOwnProperty('paragraph')) {
							if (characterStyles[selection.end].styles.align != value) {
								const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.end].styles));
								newStyle.paragraph = true;
								characterStyles.splice(selection.end, 0, { character: String.fromCharCode(8203), styles: newStyle });
							}
						}
					}
				} else {
					// Remove the style from every selected character
					for (let i = selection.start; i < selection.end; ++i) {
						delete characterStyles[i].styles[style];
						if (style == 'align') {
							characterStyles[i].styles[style] = 'left';
						}
					}
					if (style == 'align') {
						if (characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
							characterStyles[selection.start - 1].styles[style] = value;
						} else if (characterStyles[selection.start - 1].styles.align != value) {
							const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start].styles));
							newStyle.paragraph = true;
							characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
							selection.start++;
							selection.end++;
						}
						if (selection.end != characterStyles.length && !characterStyles[selection.end].styles.hasOwnProperty('paragraph')) {
							if (characterStyles[selection.end].styles.align != value) {
								const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.end].styles));
								newStyle.paragraph = true;
								characterStyles.splice(selection.end, 0, { character: String.fromCharCode(8203), styles: newStyle });
							}
						}
					}
				}
			}
		}, focusToCaret, applySelectionStyleChanges);
	}
	AddUndoState(selection, currentJSONString) {
		if (!this.media.hasOwnProperty('undoStack')) {
			this.media.undoStack = [];
		}
		if (!this.media.hasOwnProperty('undoStackIndex')) {
			this.media.undoStackIndex = -1;
		}
		selection = JSON.parse(JSON.stringify(selection));
		if (this.media.undoStackIndex == -1 || this.media.undoStack[this.media.undoStackIndex].state != currentJSONString) {
			while (this.media.undoStack.length - 1 > this.media.undoStackIndex) {
				this.media.undoStack.pop();
			}
			// Delete any undo state that happened in the last 1500 ms
			while (this.media.undoStack.length > 0 && Date.now() - this.media.undoStack[this.media.undoStack.length - 1].timestamp < 500) {
				this.media.undoStackIndex--;
				this.media.undoStack.pop();
			}
			this.media.undoStack.push({ selection, state: currentJSONString, timestamp: Date.now() });
			this.media.undoStackIndex++;
		} else {
			this.media.undoStack[this.media.undoStackIndex].selection = selection;
		}
	}
	UpdateToolbar(characterStyles, selection) {
		const styles = this.GetSelectionStyle(characterStyles, selection);
		document.getElementById('text2bold').checked = styles.bold;
		document.getElementById('text2italic').checked = styles.italic;
		document.getElementById('text2underline').checked = styles.underline;
		document.getElementById('text2strikethrough').checked = styles.strikethrough;
		document.getElementById('text2subscript').checked = styles.subscript;
		document.getElementById('text2superscript').checked = styles.superscript;
		
		document.getElementById('text2left').checked = styles.align == 'left';
		document.getElementById('text2center').checked = styles.align == 'center';
		document.getElementById('text2right').checked = styles.align == 'right';
		
		const stretch = styles.stretch == null ? '' : styles.stretch;
		if (document.getElementById('text2stretch').value != stretch) {
			document.getElementById('text2stretch').value = stretch;
		}
		
		const family = styles.family == null ? '' : styles.family;
		document.getElementById('text2family').value = family;
		
		const size = styles.size == null ? '' : styles.size;
		document.getElementById('text2size').value = size;
		
		const lineHeight = styles.lineheight == null ? '' : styles.lineheight;
		document.getElementById('text2lineheight').value = lineHeight;
		
		const color = styles.color == null ? '' : styles.color;
		document.getElementById('text2color').value = color;
		
		const highlight = styles.highlight == null ? '' : styles.highlight;
		document.getElementById('text2highlight').value = highlight;
	}
	Update(changeCharacterStyles, focusToCaret = true, applySelectionStyleChanges = true) {
		const $container = this.media.$canvasView;
		this.CharacterStylesSetDefaults(this.characterStyles, this.selection);
		if (changeCharacterStyles) {
			this.GetCharacterBounds($container.querySelector(':scope > .text'), this.characterStyles, this.selection);
			const skip = changeCharacterStyles(this.selection, this.characterStyles);
			if (skip !== undefined && skip) {
				return;
			}
			this.CharacterStylesSetDefaults(this.characterStyles, this.selection);
		}
		// Set button toggle states based on selection
		if (applySelectionStyleChanges) {
			this.UpdateToolbar(this.characterStyles, this.selection);
		}
		this.CharacterStylesToDOM($container, this.characterStyles, this.selection, applySelectionStyleChanges);
		// Change focus back to the text and hide the caret
		if (focusToCaret && (document.activeElement == null ||
							document.activeElement.shadowRoot == null && document.activeElement.tagName != 'INPUT' && document.activeElement.contentEditable !== 'true' ||
							document.activeElement.shadowRoot && document.activeElement.shadowRoot.activeElement != null && document.activeElement.shadowRoot.activeElement.tagName != 'INPUT' && document.activeElement.shadowRoot.activeElement.contentEditable !== 'true'
							)
		) {
			$container.focus();
		}
		/*
		// Make sure the cursor is visible
		if (focusToCaret) {
			const bounds = $container.parentNode.getBoundingClientRect();
			//const caretBounds = this.selection.caretTrap.getBoundingClientRect(); // caretTrap has been removed. It used to refer to an element in the generated dom
			
			if (caretBounds.left < bounds.left) {
				textScrollTag.scrollLeft += caretBounds.left - bounds.left - 5;
			}
			if (caretBounds.right > bounds.right) {
				textScrollTag.scrollLeft += caretBounds.right - bounds.right + 5;
			}
			if (caretBounds.top < bounds.top) {
				textScrollTag.scrollTop += caretBounds.top - bounds.top - 5;
			}
			if (caretBounds.bottom > bounds.bottom) {
				textScrollTag.scrollLeft += caretBounds.bottom - bounds.bottom + 5;
			}
		}
		*/
		// Set the new JSON
		const newTextJSON = this.CharacterStylesToJSON(this.characterStyles);
		const newTextJSONString = JSON.stringify(newTextJSON);
		// Add an undo state
		this.AddUndoState(this.selection, newTextJSONString);
		//this.SetElementTextScrollbars(element);
		this.media.attributes.text = JSON.parse(newTextJSONString);
	}
	GetSelectionStyle(characterStyles, selection) {
		const styles = {};
		const unionStyle = characterStyle => {
			for (const style in this.defaultStyles) {
				if (styles.hasOwnProperty(style)) {
					if (characterStyle.hasOwnProperty(style) && styles[style] != characterStyle[style] || !characterStyle.hasOwnProperty(style) && styles[style] != this.defaultStyles[style]) {
						styles[style] = null;
					}
				} else {
					if (characterStyle.hasOwnProperty(style)) {
						styles[style] = characterStyle[style];
					} else {
						styles[style] = this.defaultStyles[style];
					}
				}
			}
		};
		if (selection.start == selection.end) {
			unionStyle(characterStyles[selection.start - 1].styles);
		} else {
			for (let i = selection.start; i < selection.end; ++i) {
				unionStyle(characterStyles[i].styles);
			}
		}
		return styles;
	}
	Redo() {
		if (this.media.undoStackIndex < this.media.undoStack.length - 1) {
			this.media.undoStackIndex++;
			const redoState = this.media.undoStack[this.media.undoStackIndex];
			this.media.attributes.text = JSON.parse(redoState.state);
			this.Update((selection, characterStyles) => {
				selection.start = redoState.selection.start;
				selection.end = redoState.selection.end;
				selection.leftToRight = redoState.selection.leftToRight;
				const redoCharacterStyles = this.JSONToCharacterStyles(this.media.attributes.text, selection);
				characterStyles.splice(0, characterStyles.length);
				for (const item of redoCharacterStyles) {
					characterStyles.push(item);
				}
			});
		}
	}
	Undo() {
		if (this.media.undoStackIndex != 0) {
			this.media.undoStackIndex--;
			const undoState = this.media.undoStack[this.media.undoStackIndex];
			this.media.attributes.text = JSON.parse(undoState.state);
			this.Update((selection, characterStyles) => {
				selection.start = undoState.selection.start;
				selection.end = undoState.selection.end;
				selection.leftToRight = undoState.selection.leftToRight;
				const undoCharacterStyles = this.JSONToCharacterStyles(this.media.attributes.text, selection);
				characterStyles.splice(0, characterStyles.length);
				for (const item of undoCharacterStyles) {
					characterStyles.push(item);
				}
			});
		}
	}
	ShiftHome() {
		this.Update((selection, characterStyles) => {
			if (selection.leftToRight) {
				selection.end = selection.start;
				selection.start = 1;
				if (selection.start < selection.end) {
					selection.leftToRight = false;
				}
			} else {
				selection.start = 1;
			}
		});
	}
	Home() {
		this.Update((selection, characterStyles) => {
			selection.start = 1;
			selection.end = 1;
		});
	}
	ShiftEnd() {
		this.Update((selection, characterStyles) => {
			if (selection.leftToRight) {
				selection.end = characterStyles.length;
			} else {
				selection.start = selection.end;
				selection.end = characterStyles.length;
				selection.leftToRight = true;
			}
		});
	}
	End() {
		this.Update((selection, characterStyles) => {
			selection.start = characterStyles.length;
			selection.end = characterStyles.length;
		});
	}
	ShiftLeft() {
		this.Update((selection, characterStyles) => {
			if (selection.leftToRight) {
				if (selection.end > selection.start) {
					selection.end--;
				} else {
					selection.leftToRight = false;
					if (selection.start > 1) {
						selection.start--;
					}
				}
			} else {
				if (selection.start > 1) {
					selection.start--;
				}
			}
		});
	}
	Left() {
		this.Update((selection, characterStyles) => {
			if (selection.start == selection.end) {
				if (selection.start > 1) {
					selection.start--;
					selection.end = selection.start;
				}
			} else {
				selection.end = selection.start;
				selection.leftToRight = true;
			}
		});
	}
	ShiftRight() {
		this.Update((selection, characterStyles) => {
			if (selection.leftToRight) {
				if (selection.end < characterStyles.length) {
					selection.end++;
				}
			} else {
				if (selection.start < selection.end - 1) {
					selection.start++;
				} else {
					selection.start++
					selection.leftToRight = true;
				}
			}
		});
	}
	Right() {
		this.Update((selection, characterStyles) => {
			if (selection.start == selection.end) {
				if (selection.end < characterStyles.length) {
					selection.end++;
					selection.start = selection.end;
				}
			} else {
				selection.start = selection.end;
				selection.leftToRight = true;
			}
		});
	}
	ShiftUp() {
		this.Update((selection, characterStyles) => {
			let left, top;
			if (selection.leftToRight) {
				if (selection.end == characterStyles.length) {
					left = characterStyles[selection.end - 1].bounds.left + characterStyles[selection.end - 1].bounds.width;
					top = characterStyles[selection.end - 1].bounds.top;
				} else {
					left = characterStyles[selection.end].bounds.left;
					top = characterStyles[selection.end].bounds.top;
				}
				let index = this.GetCaretPosition(characterStyles, left, top - 1);
				if (index == selection.end) {
					index = 1;
				}
				if (index < selection.start) {
					selection.leftToRight = false;
					selection.end = selection.start;
					selection.start = index;
				} else {
					selection.end = index;
				}
			} else {
				if (selection.start == characterStyles.length || characterStyles[selection.start].character == String.fromCharCode(8203)) {
					left = characterStyles[selection.start - 1].bounds.left + characterStyles[selection.start - 1].bounds.width;
					top = characterStyles[selection.start - 1].bounds.top;
				} else {
					left = characterStyles[selection.start].bounds.left;
					top = characterStyles[selection.start].bounds.top;
				}
				let index = this.GetCaretPosition(characterStyles, left, top - 1);
				if (index == selection.start) {
					index = 1;
				}
				if (index >= selection.end) {
					selection.leftToRight = true;
					selection.start = selection.end;
					selection.end = index;
				} else {
					selection.start = index;
				}
			}
		});
	}
	Up() {
		this.Update((selection, characterStyles) => {
			let left, top;
			if (selection.start == characterStyles.length || characterStyles[selection.start].character == String.fromCharCode(8203)) {
				left = characterStyles[selection.start - 1].bounds.left + characterStyles[selection.start - 1].bounds.width;
				top = characterStyles[selection.start - 1].bounds.top;
			} else {
				left = characterStyles[selection.start].bounds.left;
				top = characterStyles[selection.start].bounds.top;
			}
			const index = this.GetCaretPosition(characterStyles, left, top - 1);
			selection.start = index;
			selection.end = index;
		});
	}
	ShiftDown() {
		this.Update((selection, characterStyles) => {
			let left, bottom;
			if (selection.leftToRight) {
				if (selection.end == characterStyles.length || characterStyles[selection.start].character == String.fromCharCode(8203)) {
					left = characterStyles[selection.end - 1].bounds.left + characterStyles[selection.end - 1].bounds.width;
					bottom = characterStyles[selection.end - 1].bounds.top + characterStyles[selection.end - 1].bounds.height;
				} else {
					left = characterStyles[selection.end].bounds.left;
					bottom = characterStyles[selection.end].bounds.top + characterStyles[selection.start].bounds.height;
				}
				let index = this.GetCaretPosition(characterStyles, left, bottom + 1);
				if (index == selection.end) {
					index = characterStyles.length;
				}
				if (index < selection.start) {
					selection.leftToRight = false;
					selection.end = selection.start;
					selection.start = index;
				} else {
					selection.end = index;
				}
			} else {
				if (selection.start == characterStyles.length) {
					left = characterStyles[selection.start - 1].bounds.left + characterStyles[selection.start - 1].bounds.width;
					bottom = characterStyles[selection.start - 1].bounds.top + characterStyles[selection.start - 1].bounds.height;
				} else {
					left = characterStyles[selection.start].bounds.left;
					bottom = characterStyles[selection.start].bounds.top + characterStyles[selection.start].bounds.height;
				}
				let index = this.GetCaretPosition(characterStyles, left, bottom + 1);
				if (index == selection.start) {
					index = characterStyles.length;
				}
				if (index >= selection.end) {
					selection.leftToRight = true;
					selection.start = selection.end;
					selection.end = index;
				} else {
					selection.start = index;
				}
			}
		});
	}
	Down() {
		this.Update((selection, characterStyles) => {
			let left, bottom;
			if (selection.start == characterStyles.length || characterStyles[selection.start].character == String.fromCharCode(8203)) {
				left = characterStyles[selection.start - 1].bounds.left + characterStyles[selection.start - 1].bounds.width;
				bottom = characterStyles[selection.start - 1].bounds.top + characterStyles[selection.start - 1].bounds.height;
			} else {
				left = characterStyles[selection.start].bounds.left;
				bottom = characterStyles[selection.start].bounds.top + characterStyles[selection.start].bounds.height;
			}
			const index = this.GetCaretPosition(characterStyles, left, bottom + 1);
			selection.start = index;
			selection.end = index;
		});
	}
	Enter() {
		this.Update((selection, characterStyles) => {
			if (selection.start != selection.end) {
				if (selection.start == 0) {
					selection.start = 1;
				}
				characterStyles.splice(selection.start, selection.end - selection.start);
				selection.end = selection.start;
			}
			const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start - 1].styles));
			newStyle.paragraph = true;
			characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
			selection.start++;
			selection.end++;
		});
	}
	Backspace() {
		this.Update((selection, characterStyles) => {
			if (selection.start == selection.end) {
				if (characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph') && characterStyles[selection.start - 1].styles.align == 'right') {
					for (let i = selection.start - 1; i < characterStyles.length && (i == selection.start - 1 || !characterStyles[i].styles.hasOwnProperty('paragraph')); ++i) {
						characterStyles[i].styles.align = 'center';
					}
				} else if (characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph') && characterStyles[selection.start - 1].styles.align == 'center') {
					for (let i = selection.start - 1; i < characterStyles.length && (i == selection.start - 1 || !characterStyles[i].styles.hasOwnProperty('paragraph')); ++i) {
						characterStyles[i].styles.align = 'left';
					}
				} else if (selection.start > 1) {
					characterStyles.splice(selection.start - 1, 1);
					selection.start--;
					selection.end--;
					for (let i = selection.start; i < characterStyles.length && !characterStyles[i].styles.hasOwnProperty('paragraph'); ++i) {
						characterStyles[i].styles.align = characterStyles[selection.start - 1].styles.align;
					}
				}
			} else {
				if (selection.start == 0) {
					selection.start = 1;
				}
				characterStyles.splice(selection.start, selection.end - selection.start);
				selection.end = selection.start;
			}
		});
	}
	Delete() {
		this.Update((selection, characterStyles) => {
			if (selection.start != characterStyles.length) {
				if (selection.start == selection.end) {
					if (characterStyles[selection.start].styles.hasOwnProperty('paragraph')) {
						delete characterStyles[selection.start].styles.paragraph;
					} else {
						characterStyles.splice(selection.start, 1);
					}
				} else {
					if (selection.start == 0) {
						selection.start = 1;
					}
					characterStyles.splice(selection.start, selection.end - selection.start);
					selection.end = selection.start;
				}
			}
		});
	}
	SingleClick() {
		this.Update((selection, characterStyles) => {
			delete selection.started;
			return true;
		});
	}
	DoubleClick(x, y) {
		this.Update((selection, characterStyles) => {
			const index = this.GetCaretPosition(characterStyles, x, y);
			delete selection.started;
			selection.start = index;
			selection.end = index;
			while (selection.start > 1 && characterStyles[selection.start - 1].character != ' ' && !characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
				selection.start--;
			}
			while (selection.end < characterStyles.length && characterStyles[selection.end].character != ' ' && !characterStyles[selection.end].styles.hasOwnProperty('paragraph')) {
				selection.end++;
			}
			selection.leftToRight = true;
		});
	}
	TripleClick(x, y) {
		this.Update((selection, characterStyles) => {
			const index = this.GetCaretPosition(characterStyles, x, y);
			delete selection.started;
			selection.start = index;
			selection.end = index;
			while (selection.start > 1 && !characterStyles[selection.start - 1].styles.hasOwnProperty('paragraph')) {
				selection.start--;
			}
			while (selection.end < characterStyles.length && !characterStyles[selection.end].styles.hasOwnProperty('paragraph')) {
				selection.end++;
			}
			selection.leftToRight = true;
		});
	}
	InsertString(characterStyles, selection, value) {
		if (selection.start != selection.end) {
			characterStyles.splice(selection.start, selection.end - selection.start);
			selection.end = selection.start;
		}
		// Replace \r\n and \n with a paragraph character
		for (let i = 0; i < value.length; ++i) {
			const newStyle = JSON.parse(JSON.stringify(characterStyles[selection.start - 1].styles));
			newStyle.paragraph = true;
			if (value[i] == String.fromCharCode(13)) {
				characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
				selection.start++;
				selection.end++;
				if (i + 1 < value.length && value[i + 1] == String.fromCharCode(10)) {
					i++;
				}
			} else if (value[i] == String.fromCharCode(10)) {
				characterStyles.splice(selection.start, 0, { character: String.fromCharCode(8203), styles: newStyle });
				selection.start++;
				selection.end++;
			} else {
				delete newStyle.paragraph;
				characterStyles.splice(selection.start, 0, { character: value[i], styles: newStyle });
				selection.start++;
				selection.end++;
			}
		}
	}
	InsertStringAtSelection(text) {
		this.Update((selection, characterStyles) => {
			this.InsertString(characterStyles, selection, text);
		});
	}
	AddListeners() {
		const $container = this.media.$canvasView;
		const $text = this.media.$canvasView.querySelector('.text');
		const pointerMove = e => {
			this.Update((selection, characterStyles) => {
				if (selection.started) {
					const x = toCanvasScale(e.clientX - $text.getBoundingClientRect().left);
					const y = toCanvasScale(e.clientY - $text.getBoundingClientRect().top);
					let index = this.GetCaretPosition(characterStyles, x, y);
					if (index >= selection.started) {
						if (selection.leftToRight && index == selection.end) {
							return true;
						}
						selection.leftToRight = true;
						selection.start = selection.started;
						selection.end = index;
					} else {
						if (!selection.leftToRight && index == selection.start) {
							return true;
						}
						selection.leftToRight = false;
						selection.start = index;
						selection.end = selection.started;
					}
				} else {
					return true;
				}
			});
		};
		$container.addEventListener('pointerdown', e => {
			e.preventDefault();
			e.stopPropagation();
			if ((e.buttons & 0x1) == 1) {
				$container.removeEventListener('pointermove', pointerMove, true);
				$container.addEventListener('pointermove', pointerMove, true);
				this.Update((selection, characterStyles) => {
					const x = toCanvasScale(e.clientX - $text.getBoundingClientRect().left);
					const y = toCanvasScale(e.clientY - $text.getBoundingClientRect().top);
					let index = this.GetCaretPosition(characterStyles, x, y);
					if (e.shiftKey) {
						if (selection.leftToRight) {
							selection.started = selection.start;
						} else {
							selection.started = selection.end;
						}
						if (index >= selection.started) {
							if (selection.leftToRight && index == selection.end) {
								return true;
							}
							selection.leftToRight = true;
							selection.start = selection.started;
							selection.end = index;
						} else {
							if (!selection.leftToRight && index == selection.start) {
								return true;
							}
							selection.leftToRight = false;
							selection.start = index;
							selection.end = selection.started;
						}
					} else {
						selection.start = index;
						selection.end = index;
						selection.started = index;
					}
				});
			}
		}, false);
		let doubleClickTimer;
		let tripleClickTimer;
		$container.addEventListener('pointerup', e => {
			e.preventDefault();
			if ((e.buttons & 0x1) == 0) {
				$container.removeEventListener('pointermove', pointerMove, true);
				const x = toCanvasScale(e.clientX - $text.getBoundingClientRect().left);
				const y = toCanvasScale(e.clientY - $text.getBoundingClientRect().top);
				if (tripleClickTimer) {
					clearTimeout(tripleClickTimer);
					tripleClickTimer = null;
					this.TripleClick(x, y);
				} else if (doubleClickTimer) {
					this.DoubleClick(x, y);
					clearTimeout(doubleClickTimer);
					doubleClickTimer = null;
					tripleClickTimer = setTimeout(() => {
						tripleClickTimer = null;
					}, 200);
				} else {
					this.SingleClick();
					doubleClickTimer = setTimeout(() => {
						doubleClickTimer = null;
					}, 200);
				}
			}
		}, true);
		/*
		$container.addEventListener('mouseleave', e => {
			if ((e.buttons & 0x1) == 0x1) {
				this.Update((selection, characterStyles) => {
					if (selection.started) {
						selection.startedMouseLeave = selection.started;
						const x = toCanvasScale(e.clientX - $text.getBoundingClientRect().left);
						const y = toCanvasScale(e.clientY - $text.getBoundingClientRect().top);
						const index = this.GetCaretPosition(characterStyles, x, y);
						if (index >= selection.started) {
							if (selection.leftToRight && index == selection.end) {
								delete selection.started;
								return true;
							}
							selection.leftToRight = true;
							selection.start = selection.started;
							selection.end = index;
						} else {
							if (!selection.leftToRight && index == selection.start) {
								delete selection.started;
								return true;
							}
							selection.leftToRight = false;
							selection.start = index;
							selection.end = selection.started;
						}
						delete selection.started;
					} else {
						return true;
					}
				});
			}
		}, true);
		$container.addEventListener('mouseenter', e => {
			// This event fires multiple times which might cause performance issues.
			if ((e.buttons & 0x1) == 0x1) {
				this.Update((selection, characterStyles) => {
					if (selection.startedMouseLeave) {
						selection.started = selection.startedMouseLeave;
						delete selection.startedMouseLeave;
					}
					return true;
				});
			}
		}, true);
		*/
		$container.addEventListener('keydown', e => {
			if (!e.altKey && !e.ctrlKey && e.key.length == 1) {
				e.preventDefault();
				this.Update((selection, characterStyles) => {
					this.InsertString(characterStyles, selection, e.key);
				});
			} else if (e.key == 'OS') { // Windows Key
				e.preventDefault();
			} else if (e.key == 'Shift') {
				return;
			} else if (e.key == 'Control') {
				return;
			} else if (e.key == 'Alt') {
				e.preventDefault();
				this.Update((selection, characterStyles) => {
					selection.altCode = '';
					return true;
				});
			} else if (e.key == 'CapsLock') {
				return;
			} else if (e.altKey) {
				e.preventDefault();
				this.Update((selection, characterStyles) => {
					// Only 0 to 9
					if (/^[0-9]$/.test(e.key)) {
						selection.altCode += e.key;
					}
					return true;
				});
			} else if (e.key == 'Insert') {
				e.preventDefault();
				// unhandled, no insert toggle support
			} else if (e.key == 'PageUp' || e.key == 'PageDown') {
				e.preventDefault();
				// unhandled, no page down or page up support
			} else if (e.ctrlKey && e.key == 'a') {
				e.preventDefault();
				this.Update((selection, characterStyles) => {
					selection.start = 1;
					selection.end = characterStyles.length;
				});
			} else if (e.ctrlKey && e.key == 'b') {
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				this.ApplyStyle('bold', styles.bold == null || !styles.bold);
			} else if (e.ctrlKey && e.key == 'i') {
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				this.ApplyStyle('italic', styles.italic == null || !styles.italic);
			} else if (e.ctrlKey && e.key == 'u') {
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				this.ApplyStyle('underline', styles.underline == null || !styles.underline);
			} else if (e.ctrlKey && e.shiftKey && e.code == 'Equal') {
				e.preventDefault();
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				this.ApplyStyle('superscript', styles.superscript == null || !styles.superscript);
			} else if (e.ctrlKey && e.code == 'Equal') {
				e.preventDefault();
				const styles = this.GetSelectionStyle(this.characterStyles, this.selection);
				this.ApplyStyle('subscript', styles.subscript == null || !styles.subscript);
			} else if (e.ctrlKey && e.shiftKey && e.key == 'Z') {
				// Redo
				e.preventDefault();
				this.Redo();
			} else if (e.ctrlKey && e.key == 'z') {
				// Undo
				e.preventDefault();
				this.Undo();
			} else if (e.ctrlKey && (e.key == 'x' || e.key == 'c' || e.key == 'v')) { // ctrl + x, ctrl + c, ctrl + v
				// Do Nothing
			} else if (e.ctrlKey) { // Any other key with ctrl held down is invalid
				e.preventDefault();
			} else if (e.shiftKey && e.key == 'Home') {
				e.preventDefault();
				this.ShiftHome();
			} else if (e.key == 'Home') {
				e.preventDefault();
				this.Home();
			} else if (e.shiftKey && e.key == 'End') {
				e.preventDefault();
				this.ShiftEnd();
			} else if (e.key == 'End') {
				e.preventDefault();
				this.End();
			} else if (e.shiftKey && e.key == 'ArrowLeft') {
				e.preventDefault();
				this.ShiftLeft();
			} else if (e.key == 'ArrowLeft') {
				e.preventDefault();
				this.Left();
			} else if (e.shiftKey && e.key == 'ArrowRight') {
				e.preventDefault();
				this.ShiftRight();
			} else if (e.key == 'ArrowRight') {
				e.preventDefault();
				this.Right();
			} else if (e.shiftKey && e.key == 'ArrowUp') {
				e.preventDefault();
				this.ShiftUp();
			} else if (e.key == 'ArrowUp') {
				e.preventDefault();
				this.Up();
			} else if (e.shiftKey && e.key == 'ArrowDown') {
				e.preventDefault();
				this.ShiftDown();
			} else if (e.key == 'ArrowDown') {
				e.preventDefault();
				this.Down();
			} else if (e.key == 'Enter') {
				e.preventDefault();
				this.Enter();
			} else if (e.key == 'Backspace') {
				e.preventDefault();
				this.Backspace();
			} else if (e.key == 'Delete') {
				e.preventDefault();
				this.Delete();
			} else if (/^F[0-9]*$/.test(e.key)) { // F1-F12
				e.preventDefault();
			} else if (e.key == 'NumLock' || e.key == 'ScrollLock' || e.key == 'Pause') {
				e.preventDefault();
			} else {
				e.preventDefault();
			}
		}, true);
		$container.addEventListener('keyup', e => {
			if (e.key == 'Shift' || e.key == 'Control' || e.key == 'ArrowLeft' || e.key == 'ArrowRight' || e.key == 'ArrowUp' || e.key == 'ArrowDown') {
				return;
			}
			this.Update((selection, characterStyles) => {
				// Alt Key Released
				if (e.key == 'Alt' && selection.altCode != '') {
					this.InsertString(characterStyles, selection, String.fromCharCode(parseInt(selection.altCode)));
				} else {
					return true;
				}
			});
		}, true);
		
		const copySelectionToClipboard = (e, selection, characterStyles) => {
			if (selection.start != selection.end) {
				const formattedText = [];
				let text = '';
				for (let i = selection.start; i < selection.end; ++i) {
					if (characterStyles[i].styles.hasOwnProperty('paragraph')) {
						text += '\r\n';
					} else if (characterStyles[i].character != String.fromCharCode(8203)) {
						text += characterStyles[i].character;
					}
					// Add all the formatted text in the selection to an array that we'll add to the clipboard
					if (characterStyles[i].hasOwnProperty('widget') || characterStyles[i].character != String.fromCharCode(8203) || characterStyles[i].styles.hasOwnProperty('paragraph')) {
						const characterCopy = JSON.parse(JSON.stringify(characterStyles[i]));
						delete characterCopy.bounds;
						formattedText.push(characterCopy);
					}
				}
				if (e.clipboardData) {
					e.clipboardData.setData('text/plain', text);
					e.clipboardData.setData('application/breezetext', JSON.stringify(formattedText));
				} else {
					window.clipboardData.setData('text', text);
					window.clipboardData.setData('text/plain', text);
					window.clipboardData.setData('application/breezetext', JSON.stringify(formattedText));
				}
			}
		};
		
		$container.addEventListener('copy', e => {
			e.preventDefault();
			this.Update((selection, characterStyles) => {
				copySelectionToClipboard(e, selection, characterStyles);
			});
		}, true);
		$container.addEventListener('cut', e => {
			e.preventDefault();
			this.Update((selection, characterStyles) => {
				copySelectionToClipboard(e, selection, characterStyles);
				characterStyles.splice(selection.start, selection.end - selection.start);
				selection.end = selection.start;
			});
		}, true);
		$container.addEventListener('paste', e => {
			// Cleans up paste data so only text is pasted. Stops HTML from being inserted
			e.preventDefault();
			let text = '';
			let formattedText = '';
			text = e.clipboardData.getData('text/plain');
			formattedText = e.clipboardData.getData('application/breezetext');
			this.Update((selection, characterStyles) => {
				if (formattedText != '') {
					if (selection.start != selection.end) {
						characterStyles.splice(selection.start, selection.end - selection.start);
						selection.end = selection.start;
					}
					formattedText = JSON.parse(formattedText);
					for (let i = 0; i < formattedText.length; ++i) {
						characterStyles.splice(selection.start, 0, formattedText[i]);
						selection.start++;
						selection.end++;
					}
				} else if (text != '') {
					this.InsertString(characterStyles, selection, text);
				} else {
					return true;
				}
			});
		}, true);
	}
	RemoveListeners() {
	}
}