const uicheckboxStyle = document.createElement('style');
uicheckboxStyle.type = 'text/css';
uicheckboxStyle.append(`
:host {
	display: grid;
	align-items: center;
	justify-items: center;
	cursor: pointer;
	color: var(--font-color);
	user-select: none;
}
:host([disabled]) {
}
:host(:focus) {
	outline: none;
}
:host(:not(.custom)):before {
	content: '';
	grid-area: 1/1;
	width: 8px;
	height: 8px;
	background-color: transparent;
}
:host(:not(.custom)[checked]):before {
	background-color: var(--font-color);
}
:host(:not(.custom)[indeterminate][checked]):before {
	background-color: var(--checkbox-indeterminate-color);
}
:host(:not(.custom)):after {
	content: '';
	grid-area: 1/1;
	width: 10px;
	height: 10px;
	border: 1px solid var(--font-color);
}

:host(.custom[checked]) {
	color: var(--checkbox-contrast-checked-font-color);
	background-color: var(--checkbox-contrast-checked-color);
}
:host(.custom) {

	background-color: var(--checkbox-contrast-background-color);
	color: var(--checkbox-contrast-checked-color);
}
`);

customElements.define('ui-checkbox', class UICheckbox extends HTMLElement {
	constructor() {
		super();
		this.optionsSet = false;
		this.attachShadow({mode: 'open'}).append(
			uicheckboxStyle.cloneNode(true),
			this.$slot = $.slot()
		);
		this.$slot.addEventListener('slotchange', () => {
			this.classList.toggle('custom', this.$slot.assignedNodes().length != 0);
		});
		this.addEventListener('click', (e) => {
			e.stopImmediatePropagation();
			this.toggleAttribute('checked');
			this.dispatchEvent(new CustomEvent('change'));
		});
	}
	options({ icon = null }) {
		if (icon) {
			this.appendChild($.i({ class: `mdi ${icon.split(' ').map(name => `mdi-${name}`).join(' ')}` }));
		}
	}
	static get observedAttributes() { return ['checked', 'indeterminate']; }
	attributeChangedCallback(name, oldValue, newValue) {
		switch (name) {
			case 'checked':
				//this.checked = newValue == 'true';
				break;
			case 'indeterminate':
				//this.indeterminate = newValue == 'true';
				break;
		}
	}
	get checked() {
		return this.hasAttribute('checked');
	}
	set checked(checked) {
		if (checked) {
			this.setAttribute('checked', '');
		} else {
			this.removeAttribute('checked');
		}
	}
	get indeterminate() {
		return this.hasAttribute('indeterminate');
	}
	set indeterminate(indeterminate) {
		if (indeterminate) {
			this.setAttribute('indeterminate', '');
		} else {
			this.removeAttribute('indeterminate');
		}
	}
});