customElements.define('ui-toast', class extends HTMLElement {
	constructor() {
		super();
		this.attachShadow({mode: 'open'});
		this.shadowRoot.innerHTML = `<style>
@import url('/ui/css/materialdesignicons.min.css');

:host {
	flex: 0 0 auto;
	background-color: var(--background-color);
	border-radius: 5px;
	box-shadow: 5px 5px 20px rgb(0 0 0 / 50%);
	opacity: 0;
	transition: opacity 500ms;
	display: flex;
	flex-flow: row nowrap;
	align-items: center;
	margin: 20px 20px 0 20px !important;
}
:host(.fadeIn) {
	opacity: 1;
}
:host(.success) {
	background-color: #e6f4e7;
	color: #82c785;
}
:host(.warning) {
	background-color: #fff1db;
	color: #ffb64c;
}
:host(.error) {
	background-color: #ffe7e0;
	color: #ff8965;
}
:host > i {
	flex: 0 0 auto;
	font-size: 28px;
	margin: 20px;
}
:host > i:last-child {
	pointer-events: all;
	cursor: pointer;
}
:host > span {
	flex: 1 0 0;
	min-width: 0;
	font-size: 16px;
	margin: 10px 0;
}
</style>`;
		this.fadeStarted = false;
		this.shadowRoot.append(
			this.$icon = $.i({ class: 'mdi' }),
			$.span({},
				$.slot()
			),
			$.i({ class: 'mdi mdi-close', click: () => this.remove() })
		);
	}
	static *customAttributes() {
		yield* ['status'];
	}
	async options({ status = 'success' } = {}) {
		this.classList.add(status);
		switch (status) {
			case 'success':
				this.$icon.classList.add('mdi-check');
				break;
			case 'warning':
				this.$icon.classList.add('mdi-alert-outline');
				break;
			case 'error':
				this.$icon.classList.add('mdi-exclamation-thick');
				break;
		}
	}
	connectedCallback() {
		if (document.getElementById('toast-container') == null) {
			document.body.appendChild(
				$.div({ id: 'toast-container', css: {
					position: 'fixed',
					right: 0,
					top: 0,
					width: '400px',
					pointerEvents: 'none',
					display: 'flex',
					flexFlow: 'column nowrap'
				} })
			);
		}
		const $toastContainer = document.getElementById('toast-container');
		if (!$toastContainer.contains(this)) {
			$toastContainer.appendChild(this);
		}
		// Move the toast to the top
		if ($toastContainer !== document.body.querySelector(':scope > :last-child')) {
			document.body.appendChild($toastContainer);
		}
		
		this.classList.add('fadeIn');
		if (!this.fadeStarted) {
			this.fadeStarted = true;
			const delay = 3000 + this.innerText.length * 40;
			setTimeout(() => {
				this.classList.remove('fadeIn');
			}, delay - 500);
			setTimeout(() => {
				this.remove();
			}, delay);
		}
	}
	disconnectedCallback() {
	}
});

function toastSuccess(message) {
	document.body.appendChild($.uitoast({ text: message, config: { status: 'success' } }));
}

function toastWarning(message) {
	document.body.appendChild($.uitoast({ text: message, config: { status: 'warning' } }));
}

function toastError(message) {
	document.body.appendChild($.uitoast({ text: message, config: { status: 'error' } }));
}