customElements.define('ui-treeselect', class extends HTMLElement {
	constructor() {
		super();
		this.optionsSet = false;
	}
	async options(config) {
		// TODO: Restructure to allow options to be called again. For now this is easier as no tree needs this
		if (this.optionsSet) {
			throw new Error('options already called');
		}
		this.optionsSet = true;
		
		this.tabIndex = -1;
		
		this.appendChild($.span());
		
		config.saveState = false;
		config.editable = () => false;
		config.showRoot = config.hasOwnProperty('showRoot') ? config.showRoot : false;
		config.showMinimizeMaximizeAll = false;
		config.showSmallLargeThumbnail = false;
		this.$tree = $.uitree({ class: 'treeselect-picker', config });
		// TODO: Remove input from uitree.js and from here
		this.$tree.input = this;
		// Events fired on the tree are passed to this
		this.$tree.addEventListener('load', e => {
			this.dispatchEvent(new CustomEvent('load'));
		});
		this.$tree.addEventListener('change', async e => {
			await this.updateInput();
			if (!this.$tree.checkbox) {
				this.hidePicker();
			}
			this.dispatchEvent(new CustomEvent('change'));
		});
		this.$tree.addEventListener('nochange', e => {
			this.$tree.dispatchEvent(new Event('change'));
		});
		this.$tree.addEventListener('click', e => e.stopPropagation());
		this.hidePicker();
	}
	connectedCallback() {
		this.bindShowPicker();
	}
	disconnectedCallback() {
		this.hidePicker();
		this.$tree.remove();
	}
	bindShowPicker() {
		this.lastShowPicker = this.showPicker.bind(this);
		this.addEventListener('click', this.lastShowPicker, true);
	}
	hidePicker() {
		document.removeEventListener('click', this.lastClickOutsideTree, true);
		if (document.body.contains(this.$tree)) {
			this.bindShowPicker();
			this.$tree.remove();
		}
	}
	async showPicker(e) {
		e.stopPropagation();
		if (!document.body.contains(this.$tree)) {
			this.removeEventListener('click', this.lastShowPicker, true);
			document.body.appendChild(this.$tree);
			this.lastClickOutsideTree = e => {
				if (!this.$tree.contains(e.target)) {
					e.stopImmediatePropagation();
					if (this.$tree.checkbox) {
						this.$tree.dispatchEvent(new Event('change'));
					}
					this.hidePicker();
				}
			};
			document.addEventListener('click', this.lastClickOutsideTree, true);
			
			this.$tree.style.width = `${this.offsetWidth}px`;
			const thisOffset = $.offset(this);
			this.$tree.style.left = `${thisOffset.left + (thisOffset.left + this.$tree.offsetWidth < document.documentElement.clientWidth ? 0 : document.documentElement.clientWidth - (thisOffset.left + this.$tree.offsetWidth))}px`;
			this.$tree.style.top = `${thisOffset.top + (thisOffset.top + this.offsetHeight + this.$tree.offsetHeight < document.documentElement.clientHeight ? this.offsetHeight : -this.$tree.offsetHeight)}px`;
			await this.reload();
			this.$tree.style.setProperty('--tree-selection-width', `${this.$tree.scrollWidth}px`);
			const selectedNode = (await this.getSelected())[0];
			if (selectedNode) {
				if (selectedNode.hasOwnProperty('tag')) {
					const top = $.offset(selectedNode.tag).top;
					const height = selectedNode.tag.querySelector(':scope > .node > .handle').offsetHeight;
					const treePickerOffset = $.offset(this.$tree);
					const treePickerHeight = this.$tree.offsetHeight;
					if (top + height >= treePickerOffset.top + treePickerHeight) {
						this.$tree.scrollTop = top + height - treePickerOffset.top - treePickerHeight;
					}
				}
			}
		}
	}
	get syncId() {
		return this.$tree.syncId;
	}
	set syncId(value) {
		this.$tree.syncId = value;
	}
	getAllNodes() {
		return this.$tree.getAllNodes();
	}
	async setSelected(...args) {
		const result = await this.$tree.setSelected(...args);
		await this.updateInput();
		return result;
	}
	getSelected() {
		return this.$tree.getSelected();
	}
	getSelectedAllChildren() {
		return this.$tree.getSelectedAllChildren();
	}
	getSelectedFilteredChildren() {
		return this.$tree.getSelectedFilteredChildren();
	}
	async setChecked(...args) {
		await this.$tree.setChecked(...args);
		this.updateInput();
	}
	getChecked() {
		return this.$tree.getChecked();
	}
	getCheckedAllChildren() {
		return this.$tree.getCheckedAllChildren();
	}
	getCheckedFilteredChildren() {
		return this.$tree.getCheckedFilteredChildren();
	}
	lock() {
		this.$tree.lock();
	}
	unlock() {
		this.$tree.unlock();
	}
	expandAll() {
		return this.$tree.expandAll();
	}
	collapseAll() {
		return this.$tree.collapseAll();
	}
	isBusy() {
		return this.$tree.isBusy();
	}
	isBusyForUploads() {
		return this.$tree.isBusyForUploads();
	}
	addItems(...args) {
		return this.$tree.addItems(...args);
	}
	updateItems(...args) {
		return this.$tree.updateItems(...args);
	}
	addGroup(...args) {
		return this.$tree.addGroup(...args);
	}
	duplicateSelected(...args) {
		return this.$tree.duplicateSelected(...args);
	}
	deleteSelected(...args) {
		return this.$tree.deleteSelected(...args);
	}
	reload() {
		return this.$tree.reload();
	}
	filterView(...args) {
		return this.$tree.filterView(...args);
	}
	removeFilterView() {
		return this.$tree.removeFilterView();
	}
	async updateInput() {
		if (this.$tree.checkbox) {
			this.querySelector(':scope > span').innerText = (await this.getChecked()).map(node => node.name).join(', ');
		} else {
			const selectedNode = (await this.getSelected())[0];
			this.querySelector(':scope > span').innerText = selectedNode ? `${selectedNode.name}${selectedNode.permitted ? '' : ' (Outside Permissions)'}` : '';
			this.title = selectedNode == undefined || selectedNode.permitted ? '' : 'This was set by a user with more permissions. If you change this you will not be able to change it back to the current value.';
		}
	}
});