let $usersTree;

function HideUsers() {
	document.getElementById('users-tip').hidden = true;
	document.getElementById('users-view').hidden = true;
	document.getElementById('users-user').hidden = true;
	document.getElementById('users-group').hidden = true;
}

function ShowUsersTip() {
	HideUsers();
	document.getElementById('users-tip').hidden = false;
	document.getElementById('users-header').innerText = '';
}

function ShowUsersGroup(node) {
	HideUsers();
	document.getElementById('users-view').hidden = false;
	document.getElementById('users-group').hidden = false;
	document.getElementById('users-header').innerText = 'Group';
	
	const nodes = node.allChildren.filter(node => !node.group);
	const names = nodes.map(node => node.name);
	
	document.getElementById('users-groupnames').value = names.join(',');

	document.getElementById('users-groupsave').replaceWith(
		$.button({ type: 'submit', id: 'users-groupsave', class: 'primary wait formgrid-label', value: 'Save', text: 'Save', click: () => UpdateUsers(nodes) })
	);
	
	SetUserData(nodes);
}

function ShowUser(node) {
	HideUsers();
	document.getElementById('users-view').hidden = false;
	document.getElementById('users-user').hidden = false;
	document.getElementById('users-header').innerText = node.name;
	
	document.getElementById('users-save').replaceWith(
		$.button({ type: 'submit', id: 'users-save', class: 'primary wait formgrid-label', value: 'Save', text: 'Save', click: () => UpdateUsers([node]) })
	);
	
	SetUserData([node]);
}

let lastShowNewUser = null;
async function ShowNewUser() {
	let selectedNode = (await $usersTree.getSelected())[0];
	if (selectedNode == undefined) {
		ConfirmationBox('Select a group', 'Select a group to add the new user to.');
		return;
	}
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $usersTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		ConfirmationBox('Invalid permissions for parent group', 'Invalid add permissions for parent group.');
		return;
	}
	
	HideUsers();
	document.getElementById('users-view').hidden = false;
	document.getElementById('users-user').hidden = false;
	document.getElementById('users-header').innerText = 'New';
	
	document.getElementById('users-save').replaceWith(
		$.button({ type: 'submit', id: 'users-save', class: 'primary wait formgrid-label', value: 'Save', text: 'Save', click: () => UpdateUsers([]) })
	);
	
	SetUserData([]);
}

function permissionId(permission) {
	return `permission-${permission.name.replace(/\s/g, '').toLowerCase()}`;
}

async function SetUserData(nodes) {
	for (const permission of permissionConfig) {
		if (permission.type == 'bool') {
			document.getElementById(permissionId(permission)).checked = false;
		} else if (permission.type == 'json') {
			await document.getElementById(permissionId(permission)).setChecked(() => false, () => false);
		}
	}
	const $username = document.getElementById('username');
	const $firstname = document.getElementById('firstname');
	const $lastname = document.getElementById('lastname');
	const $middleInitial = document.getElementById('middleinitial');
	const $password1 = document.getElementById('password1');
	const $password2 = document.getElementById('password2');
	const $emailAddress = document.getElementById('emailaddress');

	var userPermissons = null;
	if (nodes.length == 0) {
		$username.value = '';
		$username.readonly = false;
		$firstname.value = '';
		$lastname.value = '';
		$middleInitial.value = '';
		toggleRow($middleInitial, true);
		$password1.value = '';
		toggleRow($password1, true);
		$password2.value = '';
		toggleRow($password2, true);
		$emailAddress.value = '';
		for (const node of (await $usersTree.getSelected())[0].allChildren.filter(node => !node.group)) {
			if (userPermissons == null) {
				userPermissons = JSON.parse(JSON.stringify(node.permissions));
				for (const permission of permissionConfig) {
					if (permission.type == 'json') {
						userPermissons[`${permission.name}indeterminate`] = [];
					}
				}
			} else {
				for (const permission of permissionConfig) {
					if (permission.type == 'bool') {
						if (!node.permissions[permission.name]) {
							userPermissons[permission.name] = false;
						}
					} else if (permission.type == 'json') {
						for (const id of node.permissions[permission.name]) {
							if (!userPermissons[permission.name].includes(id)) {
								if (!userPermissons[`${permission.name}indeterminate`].includes(id)) {
									userPermissons[`${permission.name}indeterminate`].push(id);
								}
							}
						}
						for (let itr = 0; itr < userPermissons[permission.name].length; ++itr) {
							if (!node.permissions[permission.name].includes(userPermissons[permission.name][itr])) {
								if (!userPermissons[`${permission.name}indeterminate`].includes(userPermissons[permission.name][itr])) {
									userPermissons[`${permission.name}indeterminate`].push(userPermissons[permission.name][itr]);
								}
								userPermissons[permission.name].splice(itr--, 1);
							}
						}
					}
				}
			}
		}
	} else {
		var node = nodes[0];
		if (nodes.length == 1) {
			$username.value = node.username;
			$username.readonly = true;
			$firstname.value = node.firstname;
			$lastname.value = node.lastname;
			toggleRow($middleInitial, node.type != 3);
			$middleInitial.value = node.middleinitial;
			$emailAddress.value = node.emailaddress;
			toggleRow($password1, node.type != 3);
			$password1.value = '';
			toggleRow($password2, node.type != 3);
			$password2.value = '';
		}
		for (const node of nodes) {
			if (userPermissons == null) {
				userPermissons = JSON.parse(JSON.stringify(node.permissions));
				for (const permission of permissionConfig) {
					if (permission.type == 'json') {
						userPermissons[`${permission.name}indeterminate`] = [];
					}
				}
			} else {
				for (const permission of permissionConfig) {
					if (userPermissons[permission.name] != null) {
						if (permission.type == 'bool') {
							if (userPermissons[permission.name] != node.permissions[permission.name]) {
								userPermissons[permission.name] = null;
							}
						} else if (permission.type == 'json') {
							for (const id of node.permissions[permission.name]) {
								if (!userPermissons[permission.name].includes(id)) {
									if (!userPermissons[`${permission.name}indeterminate`].includes(id)) {
										userPermissons[`${permission.name}indeterminate`].push(id);
									}
								}
							}
							for (let itr = 0; itr < userPermissons[permission.name].length; ++itr) {
								if (!node.permissions[permission.name].includes(userPermissons[permission.name][itr])) {
									if (!userPermissons[`${permission.name}indeterminate`].includes(userPermissons[permission.name][itr])) {
										userPermissons[`${permission.name}indeterminate`].push(userPermissons[permission.name][itr]);
									}
									userPermissons[permission.name].splice(itr--, 1);
								}
							}
						}
					}
				}
			}
		}
	}
	
	if (userPermissons != null) {
		const readonly = nodes.filter(node => node.hasOwnProperty('readonly')).length != 0;
		for (const permission of permissionConfig) {
			if (permission.type == 'bool') {
				const $checkbox = document.getElementById(permissionId(permission));
				$checkbox.disabled = readonly;
				$checkbox.checked = userPermissons[permission.name] == null ? true : userPermissons[permission.name];
				$checkbox.indeterminate = userPermissons[permission.name] == null;
			} else if (permission.type == 'json') {
				document.getElementById(permissionId(permission)).classList.toggle('readonly', readonly);
				await document.getElementById(permissionId(permission)).setChecked(node => userPermissons[permission.name].includes(node.id), node => userPermissons[`${permission.name}indeterminate`].includes(node.id));
			}
		}
	}
}

async function UpdateUsers(nodes) {
	var newUser = nodes.length == 0;
	if (newUser) {
		nodes.push({});
	}
	let selectedNode = await $usersTree.getSelected();
	if (newUser) {
		selectedNode = selectedNode[0];
	}
	const items = await Promise.all(nodes.map(async node => {
		const user = {
			id: newUser ? 'newuser' : node.id,
			type: 2,
			parent: newUser ? selectedNode.id : null,
			name: nodes.length == 1 ? document.getElementById('username').value : node.username,
			username: nodes.length == 1 ? document.getElementById('username').value : node.username,
			firstname: nodes.length == 1 ? document.getElementById('firstname').value : node.firstname,
			lastname: nodes.length == 1 ? document.getElementById('lastname').value : node.lastname,
			middleinitial: nodes.length == 1 ? document.getElementById('middleinitial').value : node.middleinitial,
			emailaddress: nodes.length == 1 ? document.getElementById('emailaddress').value : node.emailaddress,
			password: document.getElementById('password1').value,
			password2: document.getElementById('password2').value,
			sendemail: document.getElementById('sendemail').checked,
			permissions: {}
		};
		const getCheckedPermission = id => {
			return document.getElementById(id).indeterminate ? null : document.getElementById(id).checked;
		};
		for (const permission of permissionConfig) {
			if (permission.type == 'bool') {
				user.permissions[permission.name] = getCheckedPermission(permissionId(permission));
			} else if (permission.type == 'json') {
				user.permissions[permission.name] = (await document.getElementById(permissionId(permission)).getChecked()).map(node => node.id);
			}
		}
		return user;
	}));
	LockBody();
	if (newUser) {
		const data = await $usersTree.addItems(items);
		if (data.success) {
			await $usersTree.setSelected([data.added[0].id], false, false, true);
		}
	} else {
		await $usersTree.updateItems(items);
	}
	UnlockBody();
}

window.addEventListener('load', async () => {
	LockBody();
	let ssoDisplayNames = await get('/get/idpsdisplaynames.php');
	if (ssoDisplayNames.success) {
		ssoDisplayNames = ssoDisplayNames.data;
	} else {
		ssoDisplayNames = {};
	}
	UnlockBody();

	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'users-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ text: 'Users' }),
				$.ul({ class: 'iconlist', css: { flex: '0 0 auto' } },
					$.li({ id: 'usersTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $usersTree.deleteSelected(), title: 'Delete selected group or user' }),
					$.li({ class: 'spacer' }),
					$.li({ id: 'usersTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $usersTree.addGroup(), title: 'Add group' }),
					$.li({ id: 'usersTree-add', class: 'tooltip-button mdi mdi-account-plus-outline', click: ShowNewUser, title: 'New user' })
				),
				$usersTree = $.uitree({ id: 'usersTree', class: 'treeContainer', config: {
					tree: 'user',
					dispatchChangeOnLoad: true,
					editable: node => node.parentNode != null,
					leafIcon: node => node.type == 2 ? 'user' : 'usersso',
					nameTransform: node => {
						let displayName = node.name;
						const matches = node.name.match(/^(.*) - (.*)$/);
						if (!node.group && matches && matches.length == 3 && ssoDisplayNames.hasOwnProperty(matches[1]) && ssoDisplayNames[matches[1]] != 0) {
							const idpName = matches[1];
							const persistentName = matches[2];
							switch (ssoDisplayNames[idpName]) {
								case '':
									// Use persistent name
									break;
								case 'First Name + Last Name':
									displayName = `${idpName} ${node.firstname} ${node.lastname} (${persistentName})`;
									break;
								case 'Email Address':
									displayName = `${idpName} ${node.emailaddress} (${persistentName})`
								default:
									if (node.middleinitial != '') {
										displayName = node.middleinitial;
									}
							}
						}
						
						if (!node.group && node.permissions['usertree'].length == 1 && node.permissions['usertree'][0] == 1) {
							return `${displayName} (Admin)`;
						} else {
							return displayName;
						}
					},
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length == 0) {
							ConfirmationBox('Selection required', 'Select a group or user to delete.');
							return;
						}
						var selectedNode = selectedNodes[0];
						if (selectedNode.group) {
							var userCount = selectedAllChildren.filter(node => !node.group).length;
							if (userCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${userCount} user${userCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove user?", `Are you sure you want to remove the user '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				}, load: async () => {
					if ((await $usersTree.getAllNodes()).some(node => node.group)) {
						for (const $e of document.querySelectorAll('#usersTree-addgroup, #usersTree-add, #usersTree-delete')) $e.hidden = false;
					}
				}, change: async function() {
					let selectedNode = (await this.getSelected())[0];
					if (selectedNode == undefined) {
						ShowUsersTip();
					} else {
						if (selectedNode.group) {
							// Is root or has no leaves
							if (selectedNode.id == 1 || selectedNode.allChildren.filter(node => !node.group).length == 0) {
								ShowUsersTip();
							} else {
								ShowUsersGroup(selectedNode);
							}
						} else {
							ShowUser(selectedNode);
						}
					}
				} })
			)
		),
		$.div({ class: 'horizontal-resize', css: { gridRow: '2 / -2' } }),
		$.div({ id: 'users-right' },
			$.div({ class: 'panel', css: { flex: '1 1 auto' } },
				$.header({ id: 'users-header', text: 'User' }),
				$.span({ id: 'users-tip', class: 'panel-tip', text: 'Select an user to edit or select a folder then "New user", to create a new user.' }),
				$.div({ id: 'users-view' },
					$.div({ id: 'users-user', class: 'scheduleFormArea', css: { display: 'inline-block', verticalAlign: 'top' } },
						$.div({ class: 'formgrid' },
							$.label({ text: 'Username:' }),
							$.input({ type: 'text', class: 'sampleInput', css: { width: '265px' }, id: 'username' }),
							
							$.label({ text: 'Firstname' }),
							$.input({ type: 'text', class: 'sampleInput', css: { width: '265px' }, id: 'firstname' }),
							
							$.label({ text: 'Lastname:' }),
							$.input({ type: 'text', class: 'sampleInput', css: { width: '265px' }, id: 'lastname' }),
							
							$.label({ text: 'Middle Initial:' }),
							$.input({ type: 'text', class: 'sampleInput', css: { width: '265px' }, id: 'middleinitial' }),
							
							$.label({ text: 'Email Address:' }),
							$.input({ type: 'text', class: 'sampleInput', css: { width: '265px' }, id: 'emailaddress' }),
							
							$.label({ text: 'Password:' }),
							$.input({ type: 'password', class: 'sampleInput', css: { width: '265px' }, id: 'password1' }),
							
							$.label({ text: 'Confirm Password:' }),
							$.input({ type: 'password', class: 'sampleInput', css: { width: '265px' }, id: 'password2' }),
							
							$.label({ class: 'tooltip', title: 'Send the user a notification email with username and password? (Note, this can take a few minutes to send).', text: 'Email Notification' }),
							$.uicheckbox({ id: 'sendemail', value: 'sendemail' }),
							
							$.button({ id: 'users-save' })
						)
					),
					$.div({ id: 'users-group', class: 'scheduleFormArea', css: { display: 'inline-block', verticalAlign: 'top' } },
						$.div({ class: 'formgrid' },
							$.label({ text: 'Users:' }),
							$.input({ type: 'text', class: 'sampleInput', id: 'users-groupnames', css: { width: '265px' }, readonly: true }),
							$.button({ id: 'users-groupsave' })
						)
					),
					$.div({ css: { display: 'inline-block', verticalAlign: 'top' } },
						$.h3({ css: { marginTop: '10px' }, text: 'Permissions' }),
						$.div({ id: 'permissionlist', class: 'formgrid' }),
						$.h3({ text: 'Access Restrictions' }),
						$.p({ css: { width: '500px', fontSize: '14px' }, text: 'Note: Giving a user permission to edit their own user account is effectively giving them admin permissions as they can change their own permissions.' }),
						$.div({ id: 'permissiontreelist', class: 'formgrid' })
					)
				)
			)
		)
	);
	for (const permission of permissionConfig) {
		const id = `permission-${permission.name.replace(/\s/g, '').toLowerCase()}`;
		if (permission.type == 'bool') {
			document.getElementById('permissionlist').append(
				$.label({ text: permission.display, title: permission.userstitle, class: 'tooltip-help' }),
				$.uicheckbox({ id }),
			);
		} else if (permission.type == 'json') {
			document.getElementById('permissiontreelist').append(
				$.label({ for: id, text: permission.userslabel }),
				$.uitreeselect({ id, config: {
					tree: permission.tree,
					checkbox: true,
					groupTypes: permission.hasOwnProperty('grouptypes') ? permission.grouptypes : [1],
					filter: node => {
						return !permission.hasOwnProperty('filter') || !permission.filter.includes(node.type);
					},
					groupIcon: node => {
						if (permission.hasOwnProperty('groupicons')) {
							if (permission.groupicons.hasOwnProperty(node.type)) {
								return permission.groupicons[node.type];
							}
							if (permission.groupicons.hasOwnProperty('*')) {
								return permission.groupicons['*'];
							}
						}
						return '';
					},
					leafIcon: node => {
						if (permission.hasOwnProperty('leaficons')) {
							if (permission.leaficons.hasOwnProperty(node.type)) {
								return permission.leaficons[node.type];
							}
							if (permission.leaficons.hasOwnProperty('*')) {
								return permission.leaficons['*'];
							}
						}
						return '';
					},
					fullSelect: true,
					showRoot: true
				} })
			);
		}
	}
	
	document.body.classList.remove('loading');
	ShowUsersTip();
	for (const $e of document.querySelectorAll('#usersTree-addgroup, #usersTree-add, #usersTree-delete')) $e.hidden = true;
	
	document.getElementById('permission-playersoperator').addEventListener('change', function() {
		if (document.getElementById('permission-playersadmin').checked) {
			document.getElementById('permission-playersadmin').checked = false;
		}
	});
	document.getElementById('permission-playersadmin').addEventListener('change', function() {
		if (document.getElementById('permission-playersoperator').checked) {
			document.getElementById('permission-playersoperator').checked = false;
		}
	});
	
	DragLeftResize('.horizontal-resize', 'users-lr');
});
