class WidgetBase {
	constructor(options) {
		Object.assign(this, options);
		// This is kind of ugly
		const widgetsSchema = this.container ? this.container.widgetsSchema : WidgetInformation;
		this.widgetSchema = widgetsSchema.find(widget => widget.name == this.media.widgettype);
		
		this.attributes = JSON.parse(JSON.stringify(this.media.attributes));
		// Repair any errors in the widget
		for (const attribute of this.widgetSchema.attributes) {
			if (!this.attributes.hasOwnProperty(attribute.name)) {
				this.attributes[attribute.name] = attribute.default;
			}
		}
		
		const overrides = {};
		for (const attribute of this.widgetSchema.attributes) {
			Object.defineProperty(this, attribute.name, {
				get() {
					if (overrides.hasOwnProperty(attribute.name)) {
						return overrides[attribute.name];
					}
					return this.GetOriginalAttributeValue(attribute.name);
				},
				set(value) {
					overrides[attribute.name] = value;
				}
			});
			if (attribute.hasOwnProperty('generatedmetadata') && attribute.generatedmetadata) {
				const attributeGeneratedMetadata = `${attribute.name}generatedmetadata`;
				// Use this attributeName_(generatedItemMetadata) syntax for accessing generated item metadata that falls back to widget metadata and then default values.
				this[`${attribute.name}_`] = generatedItemMetadata => {
					if (this.attributes.hasOwnProperty(attributeGeneratedMetadata) && this.attributes[attributeGeneratedMetadata] != '') {
						const metadata = generatedItemMetadata.find(property => property.type == attribute.type && property.name == this.attributes[attributeGeneratedMetadata]);
						if (metadata) {
							switch (attribute.type) {
								case 'int':
									return parseInt(metadata.value, 10);
								case 'float':
									return parseFloat(metadata.value);
								default:
									return metadata.value;
							}
						}
					}
					return this[attribute.name];
				}
			}
		}
	}
	GetOriginalAttributeValue(name) {
		const attributeMetadata = `${name}metadata`;
		if (this.attributes.hasOwnProperty(attributeMetadata) && this.attributes[attributeMetadata] != '' && this.media.hasOwnProperty('metadata')) {
			const attribute = this.widgetSchema.attributes.find(attribute => attribute.name == name);
			const metadata = this.media.metadata.find(property => property.type == attribute.type && property.name == this.attributes[attributeMetadata]);
			if (metadata != null) {
				return metadata.value;
			}
		}
		return this.attributes[name];
	}
	Show() {
	}
	Hide() {
	}
	Destroy() {
	}
	TimeSync() {
	}
	TagDeleted() {
		if (!this.container && !document.body.contains(this.tag)) {
			this.Destroy();
			return true;
		} else {
			return false;
		}
	}
	SetVolume() {
		
	}
}

function CreateWidget(tag, media, width, height, editor, widgetData, container, account, key, player = { serverOffset: 0, get currentTime() { return moment(); }, applyTimezone: input => input }, ready = () => {}) {
	const options = {
		tag,
		media,
		width: width == undefined ? media.width : width,
		height: height == undefined ? media.height : height,
		editor: editor == undefined ? false : editor,
		widgetData: widgetData ? widgetData : {},
		container,
		mediaid: media.id,
		account: account,
		key: key == undefined ? '' : key,
		player,
		ready
	};
	return eval(`new Widget${media.widgettype}(options)`);
}

async function UpdateWidgetCache(cacheData, mediaid, container) {
	if (cacheData == null) {
		cacheData = {};
	}
	if (container) {
		var oldCacheData = await container.readWidget(mediaid);
		if (oldCacheData == null) {
			oldCacheData = {};
		}
		// Merge the new cache data into the old replacing existing items. Note that the old cache data can have more properties than the new cache data
		Object.assign(oldCacheData, cacheData);
		container.writeWidget(mediaid, oldCacheData);
		return oldCacheData;
	}
	return cacheData;
}

function ApplyFilter(node, filter, currentMetadata)
{
	if (!node.hasOwnProperty('data') || typeof node.data != 'object' || !node.hasOwnProperty('metadata'))
	{
		return false;
	}
	switch (filter.type)
	{
		case 'and':
			return filter.children.every(childFilter => ApplyFilter(node, childFilter, currentMetadata));
		case 'or':
			return filter.children.some(childFilter => ApplyFilter(node, childFilter, currentMetadata));
		default:
			// Get the metadata record that match the filter name
			const metadata = node.metadata.find(metadata => metadata.name == filter.name);
			if (metadata == null)
			{
				return false;
			}
			if (filter.type != metadata.type)
			{
				return false;
			}
			let filterValue = filter.value;
			if (filter.hasOwnProperty('metadata'))
			{
				let metadataExists = currentMetadata.find(metadataProperty => metadataProperty.name == filter.metadata);
				if (metadataExists)
				{
					filterValue = metadataExists.value;
				}
			}
			switch (filter.operator)
			{
				case '=':
					return metadata.value == filterValue;
				case '!=':
					return metadata.value != filterValue;
				case '<':
					return metadata.value < filterValue;
				case '<=':
					return metadata.value <= filterValue;
				case '>':
					return metadata.value > filterValue;
				case '>=':
					return metadata.value >= filterValue;
			}
	}
}

function FilterToName(filter, currentMetadata, root = true) {
	switch (filter.type) {
		case 'and':
			if (filter.children.length == 1) {
				return filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false))[0];
			} else if (root) {
				return filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false)).join(' and ');
			} else {
				return `(${filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false)).join(' and ')})`;
			}
		case 'or':
			if (filter.children.length == 1) {
				return filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false))[0];
			} else if (root) {
				return filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false)).join (' or ');
			} else {
				return `(${filter.children.map(childFilter => FilterToName(childFilter, currentMetadata, false)).join (' or ')})`;
			}
		default:
			// Get the metadata record that match the filter name
			const comparison = (() => {
				switch (filter.operator)
				{
					case '=':
						return 'equals';
					case '!=':
						return 'not equal to';
					case '<':
						return 'less than';
					case '<=':
						return 'less than or equal to';
					case '>':
						return 'greater than';
					case '>=':
						return 'greater than or equal to';
				}
			})();
			switch (filter.type) {
				case 'string':
				case 'color':
					return `${filter.name} ${comparison} "${filter.value}"`;
				default:
					return `${filter.name} ${comparison} ${filter.value}`;
			}
	}
}
