class MeetingMachines {
	/**
	 * @returns {!Promise<!boolean>}
	 */
	static async isSupported() {
		const exists = await fs.pathExists(MeetingMachines.EXEC_PATH);
		return exists;
	}

	/**
	 * @returns {!string}
	 */
	static get EXEC_PATH() {
		return 'C:/gpio.exe';
	}

	/**
	 * @returns {!Object}
	 */
	static get COLOR_CODES() {
		return {
			'OFF': 0,
			'RED': 1,
			'GREEN': 2,
			'YELLOW': 4
		};
	}

	/**
	 * @param {!number} r
	 * @param {!number} g
	 * @param {!number} b
	 * @returns {!number}
	 */
	static getColorCode(r, g, b) {
		const YELLOW_RANGE = { min: 30, max: 66 };
		const GREEN_RANGE = { min: 67, max: 189 };
		const color = tinycolor({r: r, g: g, b: b});
		if (!color.isValid()) {
			console.log(`Invalid color r: ${r}, g: ${g}, b: ${b}, assuming OFF.`);
			return MeetingMachines.COLOR_CODES.OFF;
		}
		const { h: hue, l: luminance } = color.toHsl();
		if (luminance === 0) {
			return MeetingMachines.COLOR_CODES.OFF;
		} else if (hue >= YELLOW_RANGE.min && hue <= YELLOW_RANGE.max) {
			return MeetingMachines.COLOR_CODES.YELLOW;
		} else if (hue >= GREEN_RANGE.min && hue <= GREEN_RANGE.max) {
			return MeetingMachines.COLOR_CODES.GREEN;
		} else {
			return MeetingMachines.COLOR_CODES.RED;
		}
	}

	/**
	 * @param {!number} r
	 * @param {!number} g
	 * @param {!number} b
	 * @returns {!Promise}
	 */
	static async setFrontLedColor(r, g, b) {
		try {
			const colorCode = MeetingMachines.getColorCode(r, g, b);
			await Spawned.spawn(MeetingMachines.EXEC_PATH, [colorCode]);
		} catch (err) {
			console.log(`Possible error setting front led color (although the meeting machines 'gpio.exe' program always incorrectly exits with a status code of 1, despite succeeding!): ${err.stack}`);
		}
	}
}
