class NetUtils {
	/**
	 * @param {!string} url
	 * @param {!string} name
	 * @param {!string} folder
	 * @param {!string} pathsKey
	 * @param {!Agent} agent
	 * @param {!Function} [progressCb]
	 * @returns {!Promise<!string>}
	 */
	static async download(url, name, folder, pathsKey, agent, progressCb) {
		const folderpath = path.join(paths[pathsKey], folder);
		await fs.mkdirs(folderpath);

		const filepath = path.join(folderpath, name);
		const stats = await FileIO.stat(filepath);
		if (stats === null) {
			return NetworkIO.download(url, filepath, agent, progressCb);
		}

		// Since the file already exists, download to a temporary directory
		// instead, and then move and overwrite the existing file.
		const tmpFolderpath = path.join(paths.TMP, folder);
		await fs.mkdirs(tmpFolderpath);

		const tmpFilepath = path.join(tmpFolderpath, name);
		await NetworkIO.download(url, tmpFilepath, agent, progressCb);
		await fs.move(tmpFilepath, filepath, { overwrite: true });

		return filepath;
	}

	/**
	 * @param {!string} url
	 * @returns {!Promise<response: !http.IncomingMessage, body: !string|!Buffer>}
	 */
	static async get(url) {
		const data = await NetworkIO.get(url);
		return data;
	}
}

