class NetworkIO {
	/**
	 * Makes an HTTP(s) GET request to the URL.
	 * @param {!string} url
	 * @returns {!Promise<response: !http.IncomingMessage, body: !string|!Buffer|!Object>}
	 */
	static get(url) {
		return new Promise((resolve, reject) => {
			try {
				request({
					url: url
				}, (err, response, body) => {
					if (err) {
						reject(err);
					} else {
						response.resume();
						resolve({
							response: response,
							body: body
						});
					}
				});
			} catch (exception) {
				console.log(exception);
				reject(exception);
			}
		});
	}

	/**
	 * Makes an HTTP(s) POST request to the URL with the specified form data.
	 * @param {!string} url
	 * @param {?Object} formData
	 * @returns {!Promise<response: !http.IncomingMessage, body: !string|!Buffer|!Object>}
	 */
	static post(url, formData) {
		return new Promise((resolve, reject) => {
			request({
				url: url,
				method: 'POST',
				formData: formData
			}, (err, response, body) => {
				if (err) {
					reject(err);
				} else {
					resolve({
						response: response,
						body: body
					});
				}
			});
		});
	}

	/**
	 * @param {!string} url
	 * @param {!string} filepath
	 * @param {!Agent} agent
	 * @param {!Function} [progressCb]
	 * @returns {!Promise<!string>}
	 */
	static download(url, filepath, agent, progressCb) {
		return new Promise((resolve, reject) => {
			let totalSize = null;
			let error = false;
			const writable = fs.createWriteStream(filepath);
			writable.on('finish', () => {
				if (!error) {
					resolve(filepath);
				}
			});
			writable.on('error', err => {
				error = true;
				writable.end();
				req.abort();
				reject(err);
			});
			const req = progress(request({
				proxy: http_proxy,
				uri: url,
				agent: agent,
				followAllRedirects: true
			})).on('progress', state => {
				if (progressCb) {
					// state.size.total will be null if the server does not respond with a content-length header
					totalSize = state.size.total; // there's also state.size.transferred available if interested
					progressCb(state.percent, state.size.total, url);
				}
			}).on('error', err => {
				error = true;
				writable.end();
				reject(err);
			}).on('end', () => {
				if (!error && progressCb != undefined) {
					progressCb(1, totalSize, url);
				}
			});
			req.pipe(writable);
		});
	}
}
