/* exported Checkin */
/* global moment, Network */

class Checkin {
	/**
	 * @returns {!string}
	 */
	static get format() {
		return 'Y-MM-DD HH:mm:ss';
	}

	/**
	 * @param {!string} hostUrl
	 * @param {!Player} player
	 * @param {!string} account
	 * @param {!string} key
	 * @param {!string} localIp
	 * @param {!string} containerVersion
	 * @param {!string} scheduleName
	 * @param {!string} serialNumber
	 * @param {!string} serialNumberModifiedTime
	 * @returns {!Promise<!CheckinResponse>}
	 */
	static request(hostUrl, player, account, key, localIp, containerVersion, scheduleName, serialNumber, serialNumberModifiedTime) {
		let url = `${hostUrl}post/playercheckin.php?${ 
			Network.getParam('account', account)}&${
			Network.getParam('key', key)}&${
			Network.getParam('playertime', player.currentTime.format(Checkin.format))}&${
			Network.getParam('localip', localIp)}&${
			Network.getParam('containerversion', containerVersion)}&${
			Network.getParam('schedule', scheduleName)}`;
		if (serialNumber !== null && serialNumberModifiedTime !== null) {
			url += `&${Network.getParam('serialnumber', serialNumber)}&${Network.getParam('serialnumbermodifiedtime', serialNumberModifiedTime)}`;
		}
		return Network.fetch(url, 'json');
	}

	/**
	 * @typedef {!Object} CheckinResponse
	 * @property {!boolean} success
	 * @property {!string} [error]
	 * @property {!number} playerId
	 * @property {?string} timezone
	 * @property {!string} playertime
	 * @property {!boolean} restartapp
	 * @property {!boolean} screenshot
	 * @property {!boolean} thumbnailscreenshot
	 * @property {!boolean} update
	 * @property {!boolean} reboot
	 * @property {!boolean} clearcache
	 * @property {?number} screenwidth
	 * @property {?number} screenheight
	 * @property {?Object} widgetdata
	 * @property {!Object} testnetworkcommands
	 * @property {!Object} eas
	 */

	/**
	 * @typedef {!Object} CheckinSettings
	 * @property {!boolean} success
	 * @property {!string} [error]
	 * @property {!number} playerId
	 * @property {?string} timezone
	 * @property {?number} screenwidth
	 * @property {?number} screenheight
	 * @property {?Object} widgetdata
	 */
}
