class LoadingScreen {
	/**
	 * @param {!number} width
	 * @param {!number} [targetProgress]
	 */
	constructor(width, targetProgress = 110) {
		this.width = width;
		// Setting targetProgress to a value slightly over 110 adds a small delay to the
		// end of the loading bar which feels nice visually.
		this.targetProgress = targetProgress;

		this.realProgress = 0;
		this.fakeProgress = 0;
		this.rafId = -1;
		this.lineSize = 30;
		this.tickAmount = 1;
		this.done = null;

		this.canvas = $.canvas({ css: {
			position: 'absolute',
			pointerEvents: 'none',
			width: `${width}px`,
			height: `${this.lineSize}px`,
			left: 0,
			top: 0,
			zIndex: 10000
		} });
		this.ctx = this.canvas.getContext('2d');
		this.canvas.width = this.width;
		this.canvas.height = this.lineSize;

		this.rafId = requestAnimationFrame(this.animate.bind(this));
	}

	show(el) {
		el = (el !== undefined) ? el : document.body;
		el.appendChild(this.canvas);
	}

	/**
	 * @param {!number} progress
	 */
	progress(progress) {
		this.realProgress = progress;
	}

	/**
	 * @param {!function} [callback]
	 */
	finish(callback) {
		if (callback) {
			this.done = callback;
		}
		this.realProgress = this.targetProgress;
	}

	render() {
		this.ctx.clearRect(0, 0, this.width, this.lineSize);
		this.ctx.fillStyle = 'rgba(255, 255, 255, 0.25)';
		this.ctx.fillRect(0, 0, this.width, this.lineSize);
		this.ctx.fillStyle = 'rgba(27, 117, 187, 0.25)';
		this.ctx.fillRect(0, 0, this.width * (Math.min(100, this.fakeProgress) / 100), this.lineSize);
	}

	animate() {
		this.rafId = requestAnimationFrame(this.animate.bind(this));
		if (this.realProgress >= this.fakeProgress) {
			this.fakeProgress += this.tickAmount;
		}
		this.render();
		if (this.realProgress >= this.targetProgress && this.fakeProgress >= this.realProgress && this.done !== null) {
			console.log('LoadingScreen done called');
			this.done();
		}
	}

	destroy() {
		if (this.rafId > -1) {
			cancelAnimationFrame(this.rafId);
		}
		this.done = null;
		this.canvas.remove();
	}
}
