/* exported LoadingPlaylist */
/* global Playlist */

class LoadingPlaylist {
	/**
	 * @param {!string} name
	 * @param {!boolean} loadFromCache
	 * @param {!Player} player
	 * @param {!HTMLElement} $playlists
	 */
	constructor(
		name,
		loadFromCache,
		player,
		$playlists
	) {
		this.name = name;
		this.loadFromCache = loadFromCache;
		this.player = player;
		this.$playlists = $playlists;
		this.playlist = null;
	}

	get isLoaded() {
		return this.playlist && this.playlist.isLoaded;
	}

	destroy() {
		if (this.playlist) {
			this.playlist.destroy();
		}
	}

	/**
	 * @param {!boolean} [preloadMetadata]
	 * @returns {!Promise}
	 */
	async load(preloadMetadata = true) {
		if (preloadMetadata) {
			await this.preloadMetadata();
		}
		await this.playlist.load();
		this.playlist.isLoaded = true;
	}

	/**
	 * @returns {!Promise}
	 */
	async preloadMetadata() {
		this.playlist = await this.loadPlaylist();
	}

	/**
	 * @returns {!Promise<!Playlist>}
	 */
	async loadPlaylist() {
		const data = await Playlist.get(this.player, this.name, this.loadFromCache);
		
		const playlist = new Playlist(data, this.player, this.$playlists);
		// Verify all used fonts are loaded prior to creating the Playlist, preventing Media from caching incorrect layout information.
		const fonts = playlist.getUsedFonts(this.player.container.widgetsSchema);
		await this.player.container.fonts.using(fonts);
		if (this.player.isPreview) {
			this.size = playlist.size;
			this.player.container.contenttree = playlist.contenttree;
			this.player.container.metadata = playlist.metadata;
		}
		return playlist;
	}
}
