/* exported PromiseController */
/* global Timer */
const PromiseController = (() => {
	class PromiseWrapper {
		constructor() {
			this.pending = true;
			this.p = new Promise((resolve, reject) => {
				this.fulfiller = resolve;
				this.rejector = reject;
			});
		}

		/**
		 * @param {?Object} ...args
		 */
		resolve(...args) {
			this.pending = false;
		}

		/**
		 * @param {?Object} ...args
		 */
		fulfill(...args) {
			this.fulfiller(...args);
			this.resolve(...args);
		}

		/**
		 * @param {?Object} ...args
		 */
		reject(...args) {
			this.rejector(...args);
			this.resolve(...args);
		}

		/**
		 * @returns {!boolean}
		 */
		isResolved() {
			return !this.pending;
		}

		/**
		 * @returns {!Promise}
		 */
		async wait() {
			return await this.p;
		}
	}

	const Abortable = Base => class extends Base {
		abort() {
			if (this.pending) {
				this.reject(new Error('Aborted by user'));
			}
		}
	};

	const Timeoutable = Base => class extends Base {
		constructor(...args) {
			super(...args);
			if (args.timeout > 0) {
				Timer.sleep(args.timeout).then(() => {
					if (this.pending) {
						this.reject(new Error(`Exceeded maximum timeout ${args.timeout}`));
					}
				}).catch(err => {
					this.reject(err);
				});
			}
		}
	};

	return Timeoutable(Abortable(PromiseWrapper));
})();
