/* exported RouteMap */

const RouteMap = (() => {
	return class RouteMap {
		/**
		 * @param {!Minibus} bus
		 */
		constructor(bus) {
			this.bus = bus;
			this.routes = new Map();
		}

		/**
		 * @param {!string} evt
		 * @param {!function} callback
		 */
		on(evt, callback) {
			// Disallow rebinding of a route without explicit removal
			if (this.routes.has(evt)) {
				throw new Error(`Callback is already set for event '${evt}'!`);
			}
			const route = this.bus.on(evt, callback);
			this.routes.set(evt, route);
		}

		/**
		 * @param {!string} evt
		 */
		off(evt) {
			const route = this.routes.get(evt);
			if (route !== undefined) {
				this.bus.off(route);
			}
			this.routes.delete(evt);
		}

		clear() {
			const events = this.routes.keys();
			for (const evt of events) {
				this.off(evt);
			}
			this.routes.clear();
		}
	};
})();
