/* exported ScheduleLoader */
/* global FileInfo, Utility, Network */

class ScheduleLoader {
	/**
	 * @param {!Player} player
	 */
	constructor(player) {
		this.player = player;
		this.url = `${this.player.container.hostUrl}schedule/${this.player.settings.data.account}/${this.player.settings.data.key}.json`;
		this.startedAtUrl = `${this.player.container.hostUrl}get/startedat.php?${
			Network.getParam('key', this.player.settings.data.key)
		}`;
		this.file = new FileInfo('schedules/latest.schedule', false);
		this.data = null;
	}

	/**
	 * @param {!boolean} cacheSupport
	 * @param {!boolean} loadFromCache
	 * @returns {!Promise}
	 */
	async load(cacheSupport, loadFromCache) {
		const url = `${this.url}?d=${Date.now()}`;
		const scheduleData = await Utility.load(
			cacheSupport,
			loadFromCache,
			this.file,
			url,
			'json'
		);
		if (!ScheduleLoader.validate(scheduleData)) {
			return;
		}
		this.data = scheduleData;
		if (this.data.startedAt === undefined) {
			const response = await Network.fetch(
				this.startedAtUrl,
				'json'
			);
			if (!response.success) {
				throw new Error(`Unable to fetch the schedules startedAt time: ${response.error}`);
			}
			const { startedAt } = response;
			if (!Number.isInteger(startedAt) || Number.isNaN(startedAt)) {
				throw new Error(`Invalid startedAt time returned from server: ${startedAt}.`);
			}
			this.data.startedAt = startedAt;
		}
	}

	/**
	 * @param {?Object} data
	 * @returns {!boolean}
	 */
	static validate(data) {
		return data !== undefined &&
			data !== null &&
			data.id !== undefined;
	}

	/**
	 * @returns {!Promise}
	 */
	static async cleanCache() {
		const dir = new FileInfo('schedules', true);
		await dir.delete();
	}
}
