var Type = Type || {};

function TypeEnum()
{
	this.baseId = 100;
	this.labels = [];
}

TypeEnum.prototype.createEnum = function(label, ids)
{
	var results = this.labels.filter(value => value === label);
	if (results.length > 0)
	{
		throw "Cannot create duplicate Enum, label '" + label + "' already exists.";
	}

	var enumObj = {};
	for (var itr = 0; itr < ids.length; ++itr)
	{
		if (enumObj[ids[itr]] !== undefined)
		{
			throw "Cannot create duplicate enum labels for Enum '" + label + "', enum label '" + ids[itr] + "' already exists.";
		}
		enumObj[ids[itr]] = this.baseId++;
	}

	this.labels.push(label);
	return enumObj;
};

Type.generator = new TypeEnum();

Type.isTypeOf = function(id, label) {
	return Object.keys(label).indexOf(id) !== -1;
};

Type.toString = function(type) {
	var labelString = null;
	Object.keys(Type).forEach(function(v) {
		Object.keys(Type[v]).forEach(function(label) {
			if (Type[v][label] === type) {
				labelString = v + '.' + label;
			}
		});
	});
	return labelString;
};

Type.PlaybackEvent = Type.generator.createEnum('PlaybackEvent', [
	'ENTRY', 'EXIT'
]);

Type.Resource = Type.generator.createEnum('Resource', [
	'IMAGE', 'VIDEO', 'AUDIO', 'WIDGET', 'MISC', 'VIDEO_INPUT', 'BUTTON_GROUP', 'VIDEO_STREAMING', 'PDF', 'UNSUPPORTED'
]);

Type.Transition = Type.generator.createEnum('Transition', [
	'NONE', 'FADE', 'FADE_TO_BLACK', 'FADE_TO_WHITE', 'SLIDE_L', 'SLIDE_R', 'SLIDE_T', 'SLIDE_B', 'EXPAND_L', 'EXPAND_R', 'EXPAND_T', 'EXPAND_B'
]);

// TODO: Unmangle these so that the easing functions are seperated out into their own enum
Type.FX = Type.generator.createEnum('FX', [
	'LINEAR', 'SLIDE_L', 'SLIDE_R', 'SLIDE_T', 'SLIDE_B', 'SCALE_X', 'SCALE_Y', 'EASE_IN_CUBIC', 'EASE_OUT_CUBIC', 'EASE_IN_EXPO', 'EASE_OUT_EXPO'
]);

Type.Animation = Type.generator.createEnum('Animation', [
	'NONE', 'ZOOM_PAN_STANDARD'
]);

Type.Container = Type.generator.createEnum('Container', [
	'HTML5', 'ANDROID', 'NODE'
]);

Type.Visibility = Type.generator.createEnum('Visibility', [
	'SHOWN', 'HIDDEN', 'WILL_SHOW', 'WILL_HIDE'
]);

Type.PlayState = Type.generator.createEnum('PlayState', [
	'PLAYING', 'STOPPED', 'WILL_PLAY', 'WILL_STOP'
]);

Type.AlertState = Type.generator.createEnum('AlertState', [
	'ALL_CLEAR', 'ALERT', 'NONE'
]);

/**
 * This Type relation is defined in the database under the 
 * 'breezeuser.mediatype' table. The player does not use these types except to 
 * convert from the database Media type to Type.Resource.
 */
Type.Media =
{
	TEXT:    1,
	IMAGE:   2,
	VIDEO:   3,
	WEBPAGE: 4,
	AUDIO:   5,
	WIDGET:  6,
	MISC:    7,
	PDF:     9
};
