/* exported WidgetsSchema */

class WidgetsSchema {
	/**
		* @param {!string} baseUrl
		* @returns {!Promise<!Object>}
		*/
	static async load(baseUrl) {
		const src = `${baseUrl}widgets/widgets.json`;
		return await new Promise(resolve => {
			document.body.appendChild($.iframe({ src, hidden: true, load: function() {
				const raw = this.contentWindow.document.body.textContent.trim();
				try {
					const data = JSON.parse(raw);
					resolve(data);
				} catch(e) {
					resolve([]);
				}
				this.remove();
			} }));
		});
	}

	/**
		* Returns a map of schema attributes with a matching design field.
		* @param {!Object} widgetsSchema
		* @param {!string} design
		* @returns {!Map<!string, !string>}
		*/
	static getSchemasByDesign(widgetsSchema, design) {
		const schemas = new Map();
		for (const widget of widgetsSchema) {
			const attributes = widget.attributes.filter(attribute => attribute.design === design);
			if (attributes.length > 0) {
				schemas.set(widget.name, attributes);
			}
		}
		return schemas;
	}

	/**
		* Across patches the required fields for widgets to operate may change, but
		* the existing widgets are not upgraded automatically, thus this function
		* is run on widget attributes to provide the default values for missing fields.
		* @param {!WidgetsSchema[]} schemas
		* @param {!string} widgetType
		* @param {!WidgetAttributes} attributes
		*/
	static regenerateWidgetAttributes(schemas, widgetType, attributes) {
		schemas.forEach(schema => {
			if (schema.name === widgetType) {
				schema.attributes.forEach(attribute => {
					if (!attributes.hasOwnProperty(attribute.name)) {
						attributes[attribute.name] = JSON.parse(JSON.stringify(attribute.default));
					}
				});
			}
		});
	}
}