class FileStreamedRequest {
	constructor(url, dest) {
		this.timeout = null;
		
		// This download can timeout, be aborted, or resolve
		this.promise = Promise.race([
			new Promise((resolve, reject) => this.timeoutReject = reject),
			new Promise(resolve => this.abortResolve = resolve),
			new Promise((resolve, reject) => this.createRequest(url, dest, resolve, reject))
		]);
	}
	createRequest(url, dest, resolve, reject) {
		const resetTimeout = () => {
			clearTimeout(this.timeout);
			this.timeout = setTimeout(() => {
				this.timeoutReject(new Error('timeout'));
				this.request.abort();
			}, 10000);
		};
		const httpOrHttps = url.toLowerCase().startsWith('https:') ? https : http;
		this.request = httpOrHttps.get(url, response => {
			if (response.statusCode === 200) {
				const file = fs.createWriteStream(dest, { flags: 'w' });
				stream.pipeline(
					response, // Readable
					new stream.Transform({ // This transform does nothing except reset the timeout. If the download is idle for 10 seconds it is aborted
						transform(chunk, encoding, callback) {
							resetTimeout();
							this.push(chunk);
							callback();
						}
					}),
					file, // Writable
					async error => {
						clearTimeout(this.timeout);
						if (error) {
							file.destroy();
							await fs.unlink(dest);
							reject(error);
						} else {
							file.end();
							// console.log(`${uri.pathname} downloaded to: ${path}`)
							resolve();
						}
					}
				);
			} else {
				clearTimeout(this.timeout);
				reject(new Error(`Download request failed, response status: ${response.statusCode} ${response.statusMessage}`));
			}
		}).on('error', error => {
			console.log(error);
			reject(error);
		});
		resetTimeout();
	}
	abort() {
		this.request.abort();
		this.abortResolve();
	}
}
