class Updater {
	/**
	 * @param {!string} subdomain
	 * @returns {!Promise<!Version>}
	 */
	static async getServerVersion(subdomain) {
		const url = `https://${subdomain}.breezedigitalsignage.com/internal/getversion.php`;
		try {
			const r = await NetUtils.get(url);
			return new Version(r.body);
		} catch (err) {
			console.log(`Failed to retreive server version: ${err}`);
			return new Version();
		}
	}

	/**
	 * @returns {!Promise<!Version>}
	 */
	static async getClientVersion() {
		try {
			const data = await FileUtils.load('version.json', '', 'CONFIG');
			if (data === null) {
				console.log('No "version.json" file detected.');
				return new Version();
			} else {
				const o = JSON.parse(data);
				return new Version(o);
			}
		} catch (err) {
			console.log(`Failed to retrieve client version: ${err}`);
			return new Version();
		}
	}

	/**
	 * @param {!Version} server
	 * @param {!Version} client
	 * @returns {!boolean}
	 */
	static shouldUpdate(server, client) {
		return Version.compare(server, client) !== 0;
	}
	
	/**
	 * @param {!string} subdomain
	 * @param {!Version} version
	 * @returns {!Promise<!string>}
	 */
	static async downloadUpdate(subdomain, version) {
		const url = `https://${subdomain}.breezedigitalsignage.com/internal/getnwjsupdate.php`;
		const name = `${version.format('_')}.zip`;
		console.log(`Download update ${name}`);
		await fs.remove(paths.UPDATES);
		return NetUtils.download(url, name, '', 'UPDATES', agent);
	}
	/**
	 * @returns {!Promise<!string>}
	 */
	static async getVersion() {
		const clientVersion = await Updater.getClientVersion();
		return clientVersion.format();
	}
	/**
	* @param {!string} subdomain
	* @returns {!Promise}
	*/
	static async checkForUpdates(subdomain) {
		console.log(`Checking for updates from subdomain: ${subdomain}...`);
		const serverVersion = await Updater.getServerVersion(subdomain);
		const clientVersion = await Updater.getClientVersion();
		console.log(`Server version: ${serverVersion.format()}, client version: ${clientVersion.format()}`);
		if (Updater.shouldUpdate(serverVersion, clientVersion)) {
			console.log('Should update, starting update process...');
			const filepath = await Updater.downloadUpdate(subdomain, serverVersion);
			await Shell.runUpdate(filepath, serverVersion.format('_'), serverVersion.format());
			ContainerMessenger.bus.emit('closeWindow');
		} else {
			console.log('Up to date!  Skipping remainder of updating process.');
		}
	}
}
