class Widgetaudiostreaming extends WidgetBase {
	constructor(options) {
		super(options);
		this.media.duckedVolume = this.audioducking / 100;
		this.playing = false;
		this.lastCanPlayThrough = null;
		this.lastManifestParsed = null;
		this.tryAgainTimer = null;
		switch (this.format) {
			case 'https':
				this.tag.classList.add('widget-audiostreaming');
				this.$audio = this.tag.appendChild($.audio({
					crossOrigin: 'anonymous',
					error: e => {
						switch (e.target.error.code) {
							case e.target.error.MEDIA_ERR_ABORTED:
								// Do nothing
								break;
							case e.target.error.MEDIA_ERR_NETWORK:
							case e.target.error.MEDIA_ERR_DECODE:
							case e.target.error.MEDIA_ERR_SRC_NOT_SUPPORTED:
								// Retry in 1 minute
								if (this.playing) {
									if (this.tryAgainTimer == null) {
										this.tryAgainTimer = setTimeout(() => {
											this.tryAgainTimer = null;
											this.playHttps();
										}, 60 * 1000);
									}
								}
								break;
							default:
								console.log('Unknown error in audio streaming widget: ', e.target.error.code);
								break;
						}
					}
				}));
				// Build an audio analyzer to detect no volume on the stream. This is then used to reset the stream if dead air detected.
				// One of our clients has a stream that cuts out after around 1.3 days and sends no data, but stays connected.
				const audioCtx = new AudioContext();
				this.analyser = audioCtx.createAnalyser();
				var source = audioCtx.createMediaElementSource(this.$audio);
				source.connect(this.analyser);
				this.analyser.connect(audioCtx.destination);
				this.bufferLength = this.analyser.fftSize;
				this.dataArray = new Float32Array(this.bufferLength);
				this.detectNoAudio = null;
				this.playHttps();
				break;
			case 'hls':
				this.$audio = this.tag.appendChild($.audio());
				this.initializeHls();
				break;
		}
		this.ready();
	}
	initializeHls() {
		if (Hls.isSupported()) {
			this.hls = new Hls();
			this.hls.loadSource(this.url);
			this.hls.attachMedia(this.$audio);
			this.lastManifestParsed = async () => {
				this.lastManifestParsed = null;
				if (this.playing) {
					try {
						await this.$audio.play();
					} catch (exception) {
						console.log(exception);
						// Failed to play
					}
				}
			};
			this.hls.once(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
			this.hls.on(Hls.Events.ERROR, (e, data) => {
				// Only fatal errors need to be handled. All other errors per the docs should continue playing
				if (data.fatal) {
					switch(data.type) {
						// Skip all the recommended error handling and just reset the whole thing
						case Hls.ErrorTypes.NETWORK_ERROR:
							// try to recover network error
							//console.log("audio streaming: fatal network error encountered, try to recover");
							//this.hls.startLoad();
							//break;
						case Hls.ErrorTypes.MEDIA_ERROR:
							//console.log("audio streaming: fatal media error encountered, try to recover");
							//this.hls.recoverMediaError();
							//break;
						default:
							// This indicates hls should be destroyed. Try rebuilding it and start over? Unclear if this is a good idea, but we need to always recover.
							this.hls.detachMedia();
							this.hls.destroy();
							this.initializeHls();
							// This would usually mean there's no known error and hls should be destroyed
							if (this.playing && this.tryAgainTimer == null) {
								this.tryAgainTimer = setTimeout(() => {
									this.tryAgainTimer = null;
									this.playHls();
								}, 60 * 1000);
							}
							break;
					}
				}
			});
		}
	}
	isHttpsAudioPlaying() {
		this.analyser.getFloatTimeDomainData(this.dataArray);
		for (var i = 0; i < this.bufferLength; i++) {
			if (this.dataArray[i] != 0) return true;
		}
		return false;
	}
	playHttps() {
		if (this.playing) {
			this.$audio.src = '';
			this.$audio.removeAttribute('src');
			if (this.lastCanPlayThrough) {
				this.$audio.removeEventListener('canplaythrough', this.lastCanPlayThrough, { once: true });
			}
			this.lastCanPlayThrough = async () => {
				if (this.playing) {
					try {
						await this.$audio.play();
						// Every 60 seconds detect if there is audio playing. If there isn't any audio then restart the stream
						if (this.detectNoAudio) {
							clearInterval(this.detectNoAudio);
						}
						this.detectNoAudio = setInterval(() => {
							if (this.playing && !this.isHttpsAudioPlaying()) {
								clearInterval(this.detectNoAudio);
								console.log('Audio streaming widget is not playing audio. Resetting...');
								this.playHttps();
							}
						}, 60 * 1000);
					} catch(exception) {
						console.log(exception);
						if (this.playing && this.tryAgainTimer == null) {
							this.tryAgainTimer = setTimeout(() => {
								this.tryAgainTimer = null;
								this.playHttps();
							}, 60 * 1000);
						}
					}
				}
			};
			this.$audio.addEventListener('canplaythrough', this.lastCanPlayThrough, { once: true });
			this.$audio.src = this.url;
			this.$audio.load();
		}
	}
	playHls() {
		if (this.playing) {
			if (this.hls) {
				this.hls.detachMedia();
			}
			if (this.lastManifestParsed) { 
				this.hls.off(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
			}
			this.hls.loadSource(this.url);
			this.hls.attachMedia(this.$audio);
			this.lastManifestParsed = async () => {
				this.lastManifestParsed = null;
				if (this.playing) {
					try {
						await this.$audio.play();
					} catch (exception) {
						console.log(exception);
						// Failed to play
					}
				}
			};
			this.hls.once(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
		}
	}
	async Show() {
		this.playing = true;
		switch (this.format) {
			case 'https':
				this.playHttps();
				break;
			case 'hls':
				this.playHls();
				break;
		}
	}
	async Hide() {
		this.playing = false;
		clearTimeout(this.tryAgainTimer);
		this.tryAgainTimer = null;
		switch (this.format) {
			case 'https':
				// Stop the audio
				clearInterval(this.detectNoAudio);
				this.$audio.src = '';
				this.$audio.removeAttribute('src');
				this.$audio.load();
				break;
			case 'hls':
				if (this.hls) {
					this.hls.detachMedia();
				}
				break;
		}
	}
	Destroy() {
		clearTimeout(this.tryAgainTimer);
		clearInterval(this.detectNoAudio);
		if (this.hls) {
			this.hls.detachMedia();
		}
	}
	SetVolume(volume) {
		this.$audio.volume = volume;
	}
}
