function ApplyButtonStyle(container, $button, button, focus, editor = false) {
	// Repair the button here also
	const widgetsSchema = container ? container.widgetsSchema : WidgetInformation;
	const buttonSchema = widgetsSchema.find(widget => widget.name == 'button');
	for (const attribute of buttonSchema.attributes) {
		if (!button.hasOwnProperty(attribute.name)) {
			button[attribute.name] = attribute.default;
		}
	}
	if (!button.clickable) {
		return;
	}
	$button.classList.remove('buttongroup-button');
	$button.style.color = '';
	$button.style.borderColor = '';
	$button.style.borderWidth = '';
	$button.style.borderRadius = '';
	$button.style.borderStyle = '';
	$button.style.backgroundImage = '';
	$button.style.backgroundSize = '';
	$button.style.backgroundColor = '';
	if (!button.usefocustheme) {
		focus = false;
	}
	$button.classList.toggle('focus', focus);
	switch (focus ? button.focustheme : button.theme) {
		case 'none':
			$button.querySelector('span').hidden = true;
			if (editor) {
				$button.style.border = '1px solid #0000ff';
			}
			break;
		case 'default':
			$button.classList.add('buttongroup-button');
			break;
		case 'custom':
			$button.style.color = focus ? button.focusfontcolor : button.fontcolor;
			$button.style.borderColor = focus ? button.focusbordercolor : button.bordercolor;
			$button.style.borderWidth = `${focus ? button.focusborderthickness : button.borderthickness}px`;
			$button.style.borderRadius = `${focus ? button.focusborderradius : button.borderradius}px`;
			$button.style.borderStyle = 'solid';
			if (focus && button.focususebackgroundimage && button.focusbackgroundimage != null || !focus && button.usebackgroundimage && button.backgroundimage != null) {
				$button.style.backgroundImage = `url('${focus ? button.focusbackgroundimage.cdnurl : button.backgroundimage.cdnurl}')`;
				$button.style.backgroundSize = (focus ? button.focusbackgroundimagecoverfill : button.backgroundimagecoverfill) ? 'cover' : '100% 100%';
			} else {
				$button.style.backgroundColor = focus ? button.focusbackgroundcolor : button.backgroundcolor;
			}
			break;
	}
}

class Widgetbuttongroup extends WidgetBase {
	constructor(options) {
		super(options);
		
		// Repair any errors in the widget
		const widgetsSchema = this.container ? this.container.widgetsSchema : WidgetInformation;
		const buttonSchema = widgetsSchema.find(widget => widget.name == 'button');
		for (const button of this.buttons) {
			for (const attribute of buttonSchema.attributes) {
				if (!button.hasOwnProperty(attribute.name)) {
					button[attribute.name] = attribute.default;
				}
			}
		}
		
		(async () => {
			let buttons = this.buttons;
			if (this.container) {
				let imagesToCache = [];
				for (const button of buttons) {
					if (button.clickable) {
						if (button.theme == 'custom') {
							imagesToCache.push(button.backgroundimage);
						}
						if (button.usefocustheme && button.focustheme == 'custom') {
							imagesToCache.push(button.focusbackgroundimage);
						}
					}
				}
				
				const imageCacheURLs = await Promise.all(imagesToCache.map(image => this.container.cacheMediaObject(image)));
				imageCacheURLs.forEach((cacheURL, index) => {
					if (cacheURL) {
						imagesToCache[index].cdnurl = cacheURL;
					}
				});
			}
			
			let cacheData =
			{
				buttons
			};
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			
			buttons = cacheData.buttons;
			
			if (this.editor && buttons.length == 0) {
				this.tag.appendChild(
					$.div({ text: 'Double click to enter button edit mode then and drag and drop buttons from the widget panel.', class: 'buttonhelp' })
				);
			}
			
			for (const button of buttons) {
				const $button = this.tag.appendChild($.div({
					class: 'buttongroup-defaultbutton',
					css: {
						left: `${button.x}px`,
						top: `${button.y}px`,
						width: `${button.width}px`,
						height: `${button.height}px`,
						cursor: !button.clickable || this.editor ? '' : 'pointer'
					}
				}));
				
				const $span = $button.appendChild($.span({ text: button.text, css: {
					fontFamily: `'${button.font}'`,
					fontSize: `${button.fontsize}px`
				} }));
				if (button.clickable) {
					ApplyButtonStyle(this.container, $button, button, false, this.editor);
				} else {
					$span.hidden = true;
				}
			}
			this.ready();
		})();
	}
}
