class Widgetce6news extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.newsInterval = null;
		
		this.stopPlayingPromise = null;
		this.stopPlaying = null;
		
		this.tag.classList.add('widget-ce6news');
		this.tag.style.backgroundColor = this.background;
		this.tag.append(
			$.div({ class: 'backgroundimage' }),
			$.div({ class: 'news' })
		);
		if (this.usebackgroundimage && this.backgroundimage) {
			Object.assign(this.tag.querySelector('.backgroundimage').style, {
				backgroundImage: `url('${this.backgroundimage.cdnurl}')`,
				backgroundSize: this.backgroundimagecoverfill ? 'cover' : '100% 100%'
			});
		}
		
		this.processNewsData();
		this.newsInterval = setInterval(this.processNewsData.bind(this), 10 * 60 * 1000); // 10 minutes
		
		this.ready();
	}
	async processNewsData() {
		if (this.TagDeleted()) {
			return;
		}
		
		let cacheData = await new Promise(async resolve => {
			try {
				const data = await $.http({
					method: 'GET',
					url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/ce6news.php',
					data: {
						mediaid: this.mediaid
					},
					dataType: 'json'
				}).promise;
				
				if (data.success) {
					if (!this.showoldnews) {
						data.articles = data.articles.filter(article => moment(article.date).isAfter(moment().startOf('day')));
					}
					resolve({ articles: data.articles });
				} else {
					resolve(null);
				}
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('articles')) {
			cacheData.articles = [];
		}
		
		const imagesLoaded = [];
		this.tag.querySelector('.news').innerHTML = '';
		this.tag.querySelector('.news').append(
			...cacheData.articles.map(article => {
				const $article = $.div({ class: 'article', css: {
					margin: `${this.margins}px`
				} },
					$.p({ text: article.title, class: 'title', css: {
						color: this.titlecolor,
						fontFamily: `'${this.titlefont}'`,
						fontSize: `${this.titlefontsize}px`
					} }),
					$.p({ text: moment(article.date).format('M/D/YY'), class: 'date', hidden: !this.showdate, css: {
						color: this.datecolor,
						fontFamily: `'${this.datefont}'`,
						fontSize: `${this.datefontsize}px`
					} }),
					$.p({ text: article.author, class: 'author', hidden: !this.showauthor, css: {
						color: this.authorcolor,
						fontFamily: `'${this.authorfont}'`,
						fontSize: `${this.authorfontsize}px`
					} }),
					$.p({ html: article.description, class: 'description', hidden: !this.showdescription, css: {
						color: this.descriptioncolor,
						fontFamily: `'${this.descriptionfont}'`,
						fontSize: `${this.descriptionfontsize}px`
					} })
				);
				if (this.imagelayout != 'none') {
					imagesLoaded.push(new Promise(async (resolve) => {
						// Only append the largest image. A lot of RSS feeds have social media images and links so this removes those
						const imageTags = await Promise.all(article.images.map(image => {
							return new Promise(resolve => {
								$.img({ class: 'image', src: image, load: function() {
									resolve(this);
								}, error: function() {
									resolve(this);
								} });
							});
						}));
					
						for (const imageTag of imageTags) {
							if (this.imagelayout == 'bottom') {
								imageTag.style.minHeight = `${this.imageminimumheight}px`;
								$article.appendChild(imageTag);
							} else {
								$article.querySelector('.description').prepend(imageTag);
								Object.assign(imageTag.style, {
									float: 'right',
									width: '60%'
								});
							}
						}
						resolve();
					}));
				}
				$.fadeOut($article).catch(() => {});
				$article.fadeAnimation.finish();
				return $article;
			})
		);
		
		await Promise.all(imagesLoaded);
	}
	async cycle() {
		if (this.TagDeleted()) {
			return;
		}
		try {
			const $article = this.tag.querySelector('.article');
			if ($article == null) {
				return;
			}
			await Promise.race([this.stopPlayingPromise, $.fadeIn($article)]);
			if (this.showdescription) {
				const $descriptionTag = $article.querySelector('.description');
				if (this.resizefontsize) {
					while (parseInt($descriptionTag.style.fontSize, 10) > 12 && $descriptionTag.scrollHeight > $descriptionTag.clientHeight) {
						$descriptionTag.style.fontSize = `${parseInt($descriptionTag.style.fontSize, 10) - 1}px`;
					}
				}
				if ($descriptionTag.scrollHeight > $descriptionTag.clientHeight) {
					$descriptionTag.classList.add('overflow');
				}
			}
			await Promise.race([this.stopPlayingPromise, new Promise(resolve => setTimeout(resolve, 15000))]);
			await Promise.race([this.stopPlayingPromise, $.fadeOut($article)]);
			this.tag.appendChild($article);
			this.cycle();
		} catch (exception) {
			// this.stopPlayingPromise rejected
		}
	}
	Show() {
		if (!this.editor) {
			this.stopPlayingPromise = new Promise((resolve, reject) => this.stopPlaying = reject);
			this.cycle();
		}
	}
	Hide() {
		if (!this.editor) {
			if (this.stopPlaying) {
				this.stopPlaying();
			}
		}
	}
	Destroy() {
		clearInterval(this.newsInterval);
	}
}