class Widgetcountdown extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.countdownInterval = null;
		
		this.tag.classList.add('widget-countdown');
		this.tag.style.backgroundColor = this.background;
		this.tag.append(
			this.$days = $.span({ class: 'days', hidden: true }),
			this.$hours = $.span({ class: 'hours', hidden: true }),
			this.$minutes = $.span({ class: 'minutes', hidden: true }),
			this.$seconds = $.span({ class: 'seconds', hidden: true }),
			this.$special = $.span({ class: 'special', hidden: true }),
			this.$enterText = $.span({ class: 'entertext', text: this.entertext, hidden: true })
		);
		for (const $e of this.tag.querySelectorAll('span')) {
			Object.assign($e.style, {
				fontFamily: `'${this.font}'`,
				fontSize: `${this.fontsize}px`,
				color: this.foreground
			});
		}
		
		if (this.shadow) {
			this.tag.style.textShadow = '-1px -1px 5px #000, 1px -1px 5px #000, -1px  1px 5px #000, 1px  1px 5px #000';
		}
		
		this.countdownInterval = setInterval(this.update.bind(this), 1000);
		this.update();
		
		this.ready();
	}
	update() {
		if (this.TagDeleted()) {
			return;
		}
		
		const time = moment(`${this.date} ${this.time}`, 'M/D/Y h:m a');
		let now = this.player.currentTime;
		let days = '';
		let hours = '';
		let minutes = '';
		let seconds = '';
		let special = '';
		
		if (now.isSameOrAfter(time)) {
			this.$enterText.hidden = false;
		} else {
			this.$enterText.hidden = true;
			var difference = moment.duration(time.diff(now));
			
			const plural = (amount, word) => amount + ' ' + word + (amount == 1 ? '' : 's');
			const twoDigits = value => ('0' + value).slice(-2);
			if (this.timerdisplay === "Verbose") {
				days = plural(Math.floor(difference.asDays()), 'day') + ' ';
				hours = plural(difference.hours(), 'hr') + ' ';
				minutes = plural(difference.minutes(), 'min') + ' ';
				seconds = plural(difference.seconds(), 'sec');
			} else if (this.timerdisplay === "Condensed") {
				if (Math.floor(difference.asDays()) > 0) {
					days = Math.floor(difference.asDays()) + 'd';
					hours = difference.hours() + 'h';
					minutes = difference.minutes() + 'm';
					seconds = difference.seconds() + 's';
				} else if (difference.hours() > 0) {
					hours = difference.hours() + 'h';
					minutes = difference.minutes() + 'm';
					seconds = difference.seconds() + 's';
				} else if (difference.minutes() > 0) {
					minutes = difference.minutes() + 'm';
					seconds = difference.seconds() + 's';
				} else {
					seconds = difference.seconds() + 's';
				}
			} else if (this.timerdisplay === "DaysOnly") {
				days = plural(Math.floor(difference.asDays()), 'day');
			} else if (this.timerdisplay === 'hh:mm:ss') {
				special = twoDigits(difference.hours()) + ':' + twoDigits(difference.minutes()) + ':' + twoDigits(difference.seconds());
			} else if (this.timerdisplay === 'd:hh:mm:ss') {
				special = Math.floor(difference.asDays()) + ':' + twoDigits(difference.hours()) + ':' + twoDigits(difference.minutes()) + ':' + twoDigits(difference.seconds());
			}
		}
		
		const updateText = ($update, text) => {
			if (text != '') {
				if ($update.innerText != text) {
					$update.innerText = text;
				}
				$update.hidden = false;
			} else {
				$update.hidden = true;
			}
		};
		
		updateText(this.$days, days);
		updateText(this.$hours, hours);
		updateText(this.$minutes, minutes);
		updateText(this.$seconds, seconds);
		updateText(this.$special, special);
	}
	Show() {
		clearInterval(this.countdownInterval);
		this.countdownInterval = setInterval(this.update.bind(this), 1000);
		this.update();
	}
	Hide() {
		clearInterval(this.countdownInterval);
	}
	Destroy() {
		clearInterval(this.countdownInterval);
	}
	TimeSync() {
		this.update();
	}
}
