class Widgetcsvroll extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.csvInterval = null;
		this.csvRequestAnimationFrame = null;
		this.delayAnimate = null;
		this.lastTime = Date.now();
		this.topValue = 0;
		this.originalOrder = [];
		this.fitsOnPage = true;
		this.csvInterval = setInterval(this.RequestCSV.bind(this), 15 * 60 * 1000);
		this.RequestCSV();
		
		this.ready();
	}
	Step() {
		if (this.TagDeleted()) {
			return;
		}
		var currentTime = Date.now();
		var deltaTime = Math.max(0, (currentTime - this.lastTime) / 1000); // in seconds
		this.lastTime = currentTime;
		
		// We use top value to hold the floating point value and then rount to an integer so it doesn't flicker
		this.topValue -= this.rollspeed * deltaTime;
		this.$bodyTable.style.top = `${Math.round(this.topValue)}px`;
		
		const heightOfTopTwoChildren = this.$bodyTable.querySelector('tr:nth-child(1)').offsetHeight + this.$bodyTable.querySelector('tr:nth-child(2)').offsetHeight;
		if (this.topValue < -heightOfTopTwoChildren) {
			this.topValue += heightOfTopTwoChildren;
			this.$bodyTable.style.top = `${Math.round(this.topValue)}px`;
			this.$bodyTbody.appendChild(this.$bodyTable.querySelector('tr:nth-child(1)'));
			this.$bodyTbody.appendChild(this.$bodyTable.querySelector('tr:nth-child(1)'));
		}
		
		this.csvRequestAnimationFrame = window.requestAnimationFrame(this.Step.bind(this));
	}
	ParseCSV(csvFile, ignoreList) {
		var headers = [];
		var rows = [];
		
		if (csvFile != '') {
			var csvRows = CSV.parse(csvFile);
			if (csvRows.length != 0) {
				ignoreList = ignoreList.split(',');
				var ignoreColumnIndexes = [];
				for (var columnItr = 0; columnItr < csvRows[0].length; ++columnItr) {
					for (var ignoreItr = 0; ignoreItr < ignoreList.length; ++ignoreItr) {
						if (ignoreList[ignoreItr] != '') {
							if (ignoreList[ignoreItr].toLowerCase().trim() == csvRows[0][columnItr].toLowerCase().trim()) {
								ignoreColumnIndexes.push(columnItr);
							}
						}
					}
					if (ignoreColumnIndexes.indexOf(columnItr) == -1) {
						headers.push(csvRows[0][columnItr].trim());
					}
				}
				
				for (var rowItr = 1; rowItr < csvRows.length; ++rowItr) {
					var row = [];
					for (var columnItr = 0; columnItr < csvRows[rowItr].length; ++columnItr) {
						if (ignoreColumnIndexes.indexOf(columnItr) == -1) {
							row.push(csvRows[rowItr][columnItr].trim());
						}
					}
					rows.push(row);
				}
			}
		}
		return { headers: headers, rows: rows };
	}
	async RequestCSV() {
		if (this.TagDeleted()) {
			return;
		}
		
		let csvFileUrl = '';
		if (this.csvfile != 'useurl') {
			// Use the local csv file
			csvFileUrl = `${this.container ? this.container.hostUrl.replace(/\/$/, '') : ''}/csvuploads/${this.account}/${this.csvfile.replace('.csv', '')}.csv?r=${Math.floor(Math.random() * 1000000)}`;
		} else {
			csvFileUrl = `${this.container ? this.container.hostUrl.replace(/\/$/, '') : ''}/get/getfile.php?URL=${encodeURIComponent(this.url)}&r=${Math.floor(Math.random() * 1000000)}`;
		}
		
		var cacheData = await new Promise(async resolve => {
			try {
				const data = await $.http({
					method: 'GET',
					url: csvFileUrl,
				}).promise;
				resolve(this.ParseCSV(data, this.ignorelist));
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('headers')) {
			cacheData.headers = [];
		}
		if (!cacheData.hasOwnProperty('rows')) {
			cacheData.rows = [];
		}
		this.tag.innerHTML = '';
		this.tag.classList.add('widget-csvroll');
		this.$bodyContainer = this.tag.appendChild($.div({ class: 'tableBodyContainer' },
			this.$bodyTable = $.table({},
				this.$bodyColgroup = $.colgroup(),
				this.$bodyTbody = $.tbody()
			)
		));
		
		this.$headerTable = this.tag.appendChild($.table({ class: 'tableHeader' },
			this.$headerColgroup = $.colgroup(),
			$.tbody({},
				this.$headerTr = $.tr()
			)
		));
		
		// Header
		if (cacheData.headers.length == 0) {
			this.$headerTr.appendChild($.th({ text: 'No data available.', css: {
				fontFamily: `'${this.font}'`,
				fontSize: `${this.fontsize}px`,
				color: this.headerforeground,
				backgroundColor: this.headerbackground,
				borderTop: `${this.bordersize == -1 ? 0 : this.borderSize}px solid ${this.bordercolor}`,
				borderLeft: `${this.bordersize == -1 ? 0 : this.borderSize}px solid ${this.bordercolor}`,
				borderBottom: `${this.bordersize == -1 ? 1 : 0}px solid ${this.bordercolor}`,
				borderRight: `${this.bordersize}px solid ${this.bordercolor}`,
				textAlign: this.headeralignment
			} }));
		} else {
			cacheData.headers.forEach((header, headerItemIndex) => {
				this.$headerTr.appendChild($.th({ text: header, css: {
					fontFamily: `'${this.font}'`,
					fontSize: `${this.fontsize}px`,
					color: this.headerforeground,
					backgroundColor: this.headerbackground,
					borderTop: `${this.bordersize == -1 ? 0 : this.bordersize}px solid ${this.bordercolor}`,
					borderLeft: `${this.bordersize == -1 ? 0 : this.bordersize}px solid ${this.bordercolor}`,
					borderBottom: `${this.bordersize == -1 ? 1 : 0}px solid ${this.bordercolor}`,
					borderRight: `${headerItemIndex == cacheData.headers.length - 1 ? this.bordersize : 0}px solid ${this.bordercolor}`,
					textAlign: this.headeralignment
				} }));
			});
		}
		this.fitsOnPage = true;
		let rowIndex = 0;
		this.originalOrder = [];
		if (cacheData.rows.length > 0) {
			do {
				// Append a set of rows until it fills the screen
				for (const row of cacheData.rows) {
					const $tr = $.tr();
					row.forEach((rowItem, rowItemIndex) => {
						const $td = $.td({ text: rowItem == '' ? '-' : rowItem, css: {
							fontFamily: `'${this.font}'`,
							fontSize: `${this.fontsize}px`,
							color: this.foreground,
							backgroundColor: rowIndex % 2 == 0 ? this.backone : this.backtwo,
							borderTop: `${this.bordersize == -1 ? 0 : this.bordersize}px solid ${this.bordercolor}`,
							borderLeft: `${this.bordersize == -1 ? 0 : this.bordersize}px solid ${this.bordercolor}`,
							borderRight: `${rowItemIndex == row.length - 1 ? this.bordersize : 0}px solid ${this.bordercolor}`,
							borderBottom: `${rowIndex == cacheData.rows.length - 1 ? this.bordersize : 0}px solid ${this.bordercolor}`
						} });
						if (rowItem == '') {
							$td.style.color = 'transparent';
						}
						$tr.appendChild($td);
					});
					rowIndex++;
					this.originalOrder.push($tr);
					this.$bodyTbody.appendChild($tr);
				}
				if (this.fitsOnPage && this.$bodyTable.offsetHeight < this.height - this.$headerTable.offsetHeight) {
					break;
				} else {
					this.fitsOnPage = false;
				}
			} while (this.$bodyTable.offsetHeight < this.height - this.$headerTable.offsetHeight + 500);
		}
		
		const columnWidths = [];
		cacheData.headers.forEach((header, columnItr) => {
			columnWidths.push({ index: columnItr, width: 0 });
		});
		
		for (const row of cacheData.rows) {
			const $testWidth = $.div({ css: { whiteSpace: 'nowrap' } });
			await Promise.all(cacheData.headers.map(async (header, columnItr) => {
				if (columnItr < columnWidths.length) {
					columnWidths[columnItr].width = Math.max(columnWidths[columnItr].width, (await Measure($testWidth)).width + 25);
				}
			}));
			await Promise.all(row.map(async (data, columnItr) => {
				$testWidth.innerText = data;
				if (columnItr < columnWidths.length) {
					columnWidths[columnItr].width = Math.max(columnWidths[columnItr].width, (await Measure($testWidth)).width + 25);
				}
			}));
		}
		
		// Sort based on width for a moment
		columnWidths.sort((a, b) => a.width < b.width ? -1 : 1);
		
		const tableWidth = this.$headerTable.offsetWidth - this.bordersize;
		
		var totalWidth = 0;
		var usedWidth = 0;
		var overflowWidth = 0;
		var overflowStartIndex = null;
		columnWidths.forEach((column, columnItr) => {
			totalWidth += column.width;
			if (totalWidth < tableWidth) {
				usedWidth += column.width;
				column.fixedWidth = column.width;
			} else {
				if (overflowStartIndex == null) {
					overflowStartIndex = columnItr;
				}
				overflowWidth += column.width;
			}
		});
		
		if (overflowStartIndex != null) {
			for (var columnItr = overflowStartIndex; columnItr < columnWidths.length; ++columnItr) {
				var column = columnWidths[columnItr];
				column.compressedWidth = Math.round(column.width / overflowWidth * (tableWidth - usedWidth));
			}
		}
		
		// Sort the column widths back to index order
		columnWidths.sort((a, b) => a.index < b.index ? -1 : 1);
		
		for (const column of columnWidths) {
			this.$headerColgroup.appendChild($.col({ span: 1, css: {
				width: column.hasOwnProperty('fixedWidth') ? `${column.fixedWidth}px` : `${column.compressedWidth}px`
			} }));
			
			this.$bodyColgroup.appendChild($.col({ span: 1, css: {
				width: column.hasOwnProperty('fixedWidth') ? `${column.fixedWidth}px` : `${column.compressedWidth}px`
			} }));
		}
		
		Object.assign(this.$bodyContainer.style, {
			top: `${this.$headerTable.offsetHeight}px`,
			height: `calc(100% - ${this.$headerTable.offsetHeight + this.bordersize}px)`
		});
		
		// Only animate if it doesn't fit on the screen
		if (!this.fitsOnPage && !this.editor) {
			window.cancelAnimationFrame(this.csvRequestAnimationFrame);
			clearTimeout(this.delayAnimate);
			this.delayAnimate = setTimeout(() => {
				this.topValue = 0;
				this.lastTime = Date.now();
				window.cancelAnimationFrame(this.csvRequestAnimationFrame);
				this.Step();
			}, 2000);
		}
	}
	Show() {
		this.topValue = 0;
		this.lastTime = Date.now();
		for (const $tr of this.originalOrder) {
			this.$bodyTbody.appendChild($tr);
		}
		if (!this.fitsOnPage && !this.editor) {
			window.cancelAnimationFrame(this.csvRequestAnimationFrame);
			this.Step();
			if (this.csvRequestAnimationFrame) {
				window.cancelAnimationFrame(this.csvRequestAnimationFrame);
				this.csvRequestAnimationFrame = null;
			}
			clearTimeout(this.delayAnimate);
			this.delayAnimate = setTimeout(() => {
				this.topValue = 0;
				this.lastTime = Date.now();
				this.Step();
			}, 2000);
		}
	}
	Hide() {
		clearTimeout(this.delayAnimate);
		if (this.csvRequestAnimationFrame) {
			window.cancelAnimationFrame(this.csvRequestAnimationFrame);
			this.csvRequestAnimationFrame = null;
		}
	}
	Destroy() {
		clearTimeout(this.delayAnimate);
		clearInterval(this.csvInterval);
		if (this.csvRequestAnimationFrame) {
			window.cancelAnimationFrame(this.csvRequestAnimationFrame);
			this.csvRequestAnimationFrame = null;
		}
	}
}
