class Widgetevents extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.eventsInterval = null;
		
		this.tag.classList.add('widget-events');
		Object.assign(this.tag.style, {
			backgroundColor: this.background
		});
		this.tag.append(
			$.table({ class: 'events' },
				this.$eventsColgroup = $.colgroup(),
				this.$eventsTHead = $.thead(),
				this.$eventsTBody = $.tbody({},
					$.tr({},
						$.td({ text: 'Loading Events...' })
					)
				)
			)
		);
		
		this.eventsInterval = setInterval(this.processEventData.bind(this), 5 * 60 * 1000);
		this.processEventData();
		
		this.ready();
	}
	async processEventData() {
		if (this.TagDeleted()) {
			return;
		}
		
		var cacheData = await new Promise(async resolve => {
			if (this.eventlist == '') {
				resolve({
					events: []
				});
				return;
			}
			try {
				const data = await $.http({
					method: 'GET',
					url: `${(this.container ? this.container.hostUrl.replace(/\/$/, '') : '')}/events/${this.account}/${this.eventlist}?${Math.floor(Math.random() * 1000000)}`,
					dataType: 'json'
				}).promise;
				data.sort((a, b) => {
					var aTime = moment(a.start * 1000);
					var bTime = moment(b.start * 1000);
					if (this.showgroups) {
						if (a.group == b.group) {
							// Sort by time
							return aTime.isBefore(bTime) ? -1 : 1;
						} else {
							return a.group.localeCompare(b.group);
						}
					} else {
						return aTime.isBefore(bTime) ? -1 : 1;
					}
				});
				// Naive migration code until the user hits update on the events tool
				for (const event of data) {
					if (!event.hasOwnProperty('timezone')) {
						event.timezone = 'US/Central';
					}
				}
				resolve({ events: data });
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('events')) {
			cacheData.events = [];
		}
		var events = [];
		// This code is a bit awkward, but it treats the timezone as if it's UTC
		const timezone = cacheData.events.length > 0 ? cacheData.events[0].timezone : 'US/Central';
		const currentTime = moment.tz(timezone).add(this.player.serverOffset, 'ms');
		const startOfDay = moment.tz(timezone).startOf('day');
		const endOfDay = moment.tz(timezone).endOf('day');
		
		// Add all the events that occur today to the events list that are after that current time
		for (const event of cacheData.events) {
			const eventStartInTimezone = moment.tz(event.start * 1000, timezone);
			const eventEndInTimezone = moment.tz(event.end * 1000, timezone);
			if (!event.sunday && !event.monday && !event.tuesday && !event.wednesday && !event.thursday && !event.friday && !event.saturday) {
				// Dayparting is disabled and this event spans from the start date and time to the end date and time
				var rangeEvent = {
					group: event.group,
					name: event.name,
					start: eventStartInTimezone,
					end: eventEndInTimezone
				};
				// Add the event if it occurs today and ends after the current time
				if (rangeEvent.start.isBefore(endOfDay) && rangeEvent.end.isAfter(startOfDay) && rangeEvent.end.isAfter(currentTime)) {
					events.push(rangeEvent);
				} else if (rangeEvent.start.isBefore(endOfDay) && rangeEvent.end.isAfter(startOfDay) && this.showexpiredevents) {
					events.push(rangeEvent);
				}
			} else {
				var startDate = eventStartInTimezone.clone().startOf('day');
				var endDate = eventEndInTimezone.clone().endOf('day');
				if (currentTime.isAfter(startDate) && currentTime.isBefore(endDate)) {
					// Extract the start time and end time offsets for the dayparting
					var startTime = eventStartInTimezone.diff(eventStartInTimezone.clone().startOf('day'), 'seconds');
					var endTime = eventEndInTimezone.diff(eventEndInTimezone.clone().startOf('day'), 'seconds');
					var dayOfWeekEnabled = [event.sunday, event.monday, event.tuesday, event.wednesday, event.thursday, event.friday, event.saturday];
					
					var iterateDayOfWeek = currentTime.day(); // 0 = sunday, 6 = saturday
					var dayEnabled = dayOfWeekEnabled[iterateDayOfWeek];
					if (dayEnabled) {
						var dowEvent = startTime < endTime ?
						{
							group: event.group,
							name: event.name,
							start: startOfDay.clone().add(startTime, 'seconds'),
							end: startOfDay.clone().add(endTime, 'seconds')
						} :
						{
							group: event.group,
							name: event.name,
							start: startOfDay.clone().add(startTime, 'seconds'),
							end: startOfDay.clone().add(1, 'day').add(endTime, 'seconds')
						};
						if (dowEvent.start.isBefore(moment().endOf('day')) && dowEvent.end.isAfter(moment().startOf('day')) && dowEvent.end.isAfter(currentTime)) {
							events.push(dowEvent);
						} else if (dowEvent.start.isBefore(moment().endOf('day')) && dowEvent.end.isAfter(moment().startOf('day')) && this.showexpiredevents) {
							events.push(dowEvent);
						}
					}
				}
			}
		}
		const parent = this.tag.parentNode;
		const sizingContainer = GetSizingContainer();
		sizingContainer.appendChild(this.tag);
		this.$eventsColgroup.innerHTML = '';
		
		let timeHeaderText = 'Time';
		switch (this.startendtimeformat) {
			case 'none':
				break;
			case 'startend':
				break;
			case 'duration':
				timeHeaderText = 'Duration';
				break;
			case 'startonly':
				timeHeaderText = 'Start Time';
				break;
		}
		// Find the calculated width of the group and time columns
		const $testWidth = $.div({ css: {
			whiteSpace: 'nowrap',
			fontFamily: `'${this.font}'`,
			fontSize: `${this.fontsize}px`
		} });
		$testWidth.innerText = 'Group';
		var groupWidth = (await Measure($testWidth)).width;
		$testWidth.innerText = timeHeaderText;
		var startTimeWidth = (await Measure($testWidth)).width;
		var hyphenWidth = 0;
		var endTimeWidth = startTimeWidth;
		var durationWidth = startTimeWidth;
		for (const event of events) {
			// Group
			$testWidth.innerText = event.group;
			groupWidth = Math.max(groupWidth, (await Measure($testWidth)).width);
			// Time
			$testWidth.innerText = event.start.clone().local().format('h:mm a');
			startTimeWidth = Math.max(startTimeWidth, (await Measure($testWidth)).width);
			$testWidth.innerText = '-';
			hyphenWidth = Math.max(hyphenWidth, (await Measure($testWidth)).width);
			$testWidth.innerText = event.end.clone().local().format('h:mm a');
			endTimeWidth = Math.max(endTimeWidth, (await Measure($testWidth)).width);
			var duration = moment.duration(event.end.diff(event.start));
			duration = (duration.hours() == 0 ? '' : duration.hours() + (duration.hours() == 1 ? ' hour ' : ' hours ')) + duration.minutes() + (duration.minutes() == 1 ? ' minute' : ' minutes');
			$testWidth.innerText = duration;
			durationWidth = Math.max(durationWidth, (await Measure($testWidth)).width);
		}
		
		if (this.showgroups) {
			this.$eventsColgroup.appendChild(
				$.col({ span: 1, css: {
					width: `${groupWidth + 50}px`
				} })
			);
		}
		switch (this.startendtimeformat) {
			case 'none':
				break;
			case 'startend':
				this.$eventsColgroup.append(
					$.col({ span: 1, css: {
						width: `${startTimeWidth + 10}px`
					} }),
					$.col({ span: 1, css: {
						width: `${hyphenWidth + 20}px`
					} }),
					$.col({ span: 1, css: {
						width: `${endTimeWidth + 50}px`
					} })
				);
				break;
			case 'duration':
				this.$eventsColgroup.appendChild(
					$.col({ span: 1, css: {
						width: `${durationWidth + 10}px`
					} })
				);
				break;
			case 'startonly':
				this.$eventsColgroup.appendChild(
					$.col({ span: 1, css: {
						width: `${startTimeWidth + 10}px`
					} })
				);
				break;
		}
		this.$eventsColgroup.append(
			$.col({ span: 1, css: {
				width: '100%'
			} })
		);
		
		this.$eventsTHead.innerHTML = '';
		this.$eventsTBody.innerHTML = '';
		
		const $header = $.tr();
		if (this.showgroups) {
			$header.appendChild(
				$.th({ text: 'Group' })
			);
		}
		switch (this.startendtimeformat) {
			case 'none':
				break;
			case 'startend':
				$header.append(
					$.th({ text: timeHeaderText }),
					$.th({ text: '' }),
					$.th({ text: '' })
				);
				break;
			case 'duration':
				$header.appendChild(
					$.th({ text: timeHeaderText })
				);
				break;
			case 'startonly':
				$header.appendChild(
					$.th({ text: timeHeaderText })
				);
				break;
		}
		$header.appendChild(
			$.th({ text: 'Event' })
		);
		this.$eventsTHead.appendChild($header);
		
		let previousGroup = null;
		for (const event of events) {
			const starttime = event.start.clone().local().format('h:mm a');
			const endtime = event.end.clone().local().format('h:mm a');
			
			const $row = $.tr();
			if (this.showgroups) {
				$row.appendChild(
					$.td({ class: 'group', text: event.group == previousGroup ? '' : event.group })
				);
			}
			switch (this.startendtimeformat) {
				case 'none':
					break;
				case 'startend':
					$row.append(
						$.td({ class: 'starttime', text: starttime }),
						$.td({ class: 'hyphen', text: '-', css: { textAlign: 'center' } }),
						$.td({ class: 'endtime', text: endtime })
					);
					break;
				case 'duration':
					const duration = moment.duration(event.end.diff(event.start));
					const plural = value => value > 1 ? 's' : '';
					$row.appendChild(
						$.td({ text: `${duration.hours() == 0 ? '' : `${duration.hours()} hour${plural(duration.hours())} `}${duration.minutes()} minute${plural(duration.minutes())}` })
					);
					break;
				case 'startonly':
					$row.appendChild(
						$.td({ class: 'starttime', text: starttime })
					);
					break;
			}
			$row.appendChild(
				$.td({ class: 'name', text: event.name })
			);
			this.$eventsTBody.appendChild($row);
			
			previousGroup = event.group;
		}
		
		for (const $e of this.tag.querySelectorAll('th, td')) {
			Object.assign($e.style, {
				color: this.foreground,
				fontFamily: `'${this.font}'`,
				fontSize: `${this.fontsize}px`
			});
		}
		parent.appendChild(this.tag);
	}
	Show() {
		
	}
	Hide() {
		
	}
	Destroy() {
		clearInterval(this.eventsInterval);
	}
	TimeSync() {
		this.processEventData();
	}
}
