class Widgetexchange extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.exchangeInterval = null;
		
		this.tag.classList.add('widget-exchange');
		this.tag.style.backgroundColor = this.background;
		this.tag.append(
			$.div({ class: 'container' },
				$.div({ class: 'title' }, $.h1()),
				$.div({ class: 'time' }, $.h2())
			),
			$.div({ class: 'eventdisplay' },
				$.table({},
					$.colgroup({},
						$.col({ css: { width: '100%' } }),
						$.col({ css: { width: '0%' } }),
						$.col({ css: { width: '0%' } }),
						$.col({ css: { width: '0%' } })
					),
					$.tbody({},
						$.tr({},
							$.th({ text: 'Today' }),
							$.td(),
							$.td(),
							$.td()
						)
					),
					this.$todaysEvents = $.tbody({ class: 'todayevents' }),
					$.tbody({},
						$.tr({},
							$.th({ text: 'Tomorrow' }),
							$.td(),
							$.td()
						)
					),
					this.$tomorrowEvents = $.tbody({ class: 'tomorrowevents' })
				)
			)
		);
		
		this.exchangeInterval = setInterval(this.processExchangeData.bind(this), 10 * 60 * 1000);
		this.processExchangeData();
		
		this.ready();
	}
	resize() {
		for (var i = this.fontsize; i > 8; --i) {
			this.tag.querySelector('.title h1').style.fontSize = `${i}px`;
			if (this.tag.querySelector('.title h1').offsetWidth < this.tag.querySelector('.title').offsetWidth) {
				break;
			}
		}
		
		for (var i = this.fontsize; i > 8; --i) {
			this.tag.querySelector('.time h2').style.fontSize = `${i}px`;
			if (this.tag.querySelector('.time h2').offsetWidth < this.tag.querySelector('.time').offsetWidth) {
				break;
			}
		}
	}
	async processExchangeData() {
		if (this.TagDeleted()) {
			return;
		}
		
		this.tag.querySelector('.container').hidden = !this.showcurrenteventonly;
		this.tag.querySelector('.eventdisplay').hidden = this.showcurrenteventonly;
		
		var cacheData = await new Promise(async resolve => {
			if (!('zipcode' in this)) {
				var events = [];
				var subjectIndex = 0;
				for (var locationItr = 0; locationItr < 3; ++locationItr) {
					for (var hourItr = 0; hourItr < 3; ++hourItr) {
						events.push(
						{
							subject: 'Sample Subject ' + subjectIndex++,
							location: 'Sample Location ' + locationItr,
							start: moment().add(hourItr, 'hour').minutes(0),
							end: moment().add(hourItr + 1, 'hour').minutes(0)
						})
					}
				}
				resolve({ events: events });
			} else {
				try {
					const data = await $.http({
						method: 'GET',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/exchangedata.php',
						data: {
							mediaid: this.mediaid
						}
					}).promise;
					var error = null;
					try {
						error = JSON.parse(data);
						console.log('Exchange: ' + JSON.stringify(error));
					} catch (exception) {
						// Do nothing
					}
					var events = [];
					const parser = new DOMParser();
					const $xml = parser.parseFromString(data, 'application/xml');
					if (error == null) {
						for (const $event of $xml.querySelectorAll('Event')) {
							events.push({
								subject: $event.querySelector('Subject').innerText,
								location: $event.querySelector('Location').innerText,
								start: moment($event.querySelector('UTCStartTime').innerText + '+0000', 'YYYY-MM-DD[T]HH:mm:ss Z'),
								end: moment($event.querySelector('UTCEndTime').innerText + '+0000', 'YYYY-MM-DD[T]HH:mm:ss Z')
							});
							events.sort((a, b) => a.start.isBefore(b.start) ? -1 : 1);
						}
					}
					resolve({ events: events });
				} catch (exception) {
					resolve(null);
				}
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('events')) {
			cacheData.events = [];
		}
		
		if (this.showcurrenteventonly) {
			var currentEvent = null;
			var currentTime = moment();
			
			for (const event of cacheData.events) {
				if (currentEvent == null) {
					currentEvent = event;
				}
				else if (event.start.isBefore(currentTime)) {
					currentEvent = event;
				}
			}
			
			if (currentEvent !== null) {
				this.tag.querySelector('.title h1').innerText = currentEvent.subject;
				this.tag.querySelector('.time h2').innerText = `${currentEvent.start.format('h:mm a')} - ${currentEvent.end.format('h:mm a')}`;
			} else {
				this.tag.querySelector('.title h1').innerText = 'No Events';
				this.tag.querySelector('.time h2').innerText = '';
			}
			this.resize();
		} else {
			// Today
			this.$todaysEvents.innerHTML = '';
			var timeWidth = 0;
			var hyphenWidth = 0;
			var $testWidth = $.div({ css: {
				whiteSpace: 'nowrap',
				fontFamily: `'${this.font}'`,
				fontSize: `${this.fontsize}px`
			} });
			for (const event of cacheData.events) {
				$testWidth.innerText = event.start.format('h:mm a');
				timeWidth = Math.max(timeWidth, (await Measure($testWidth)).width);
				$testWidth.innerText = event.end.format('h:mm a');
				timeWidth = Math.max(timeWidth, (await Measure($testWidth)).width);
				$testWidth.innerText = '-';
				hyphenWidth = Math.max(hyphenWidth, (await Measure($testWidth)).width);
			}
			this.tag.querySelector('.eventdisplay > table > colgroup').innerHTML = '';
			this.tag.querySelector('.eventdisplay > table > colgroup').append(
				$.col({ span: 1, css: {
					width: '100%'
				} }),
				$.col({ span: 1, css: {
					width: `${timeWidth + 50}px`
				} }),
				$.col({ span: 1, css: {
					width: `${hyphenWidth + 50}px`
				} }),
				$.col({ span: 1, css: {
					width: `${timeWidth + 50}px`
				} })
			);
			for (const event of cacheData.events) {
				if (event.start.isBefore(moment().endOf('day')) && event.end.isSameOrAfter(moment().startOf('day'))) {
					this.$todaysEvents.appendChild(
						$.tr({},
							$.td({ text: event.subject }),
							$.td({ text: event.start.format('h:mm a') }),
							$.td({ text: '-' }),
							$.td({ text: event.end.format('h:mm a') })
						)
					);
				}
			}
			// Tomorrow
			this.$tomorrowEvents.innerHTML = '';
			for (const event of cacheData.events) {
				if (event.start.isBefore(moment().add(1, 'day').endOf('day')) && event.end.isSameOrAfter(moment().add(1, 'day').startOf('day'))) {
					this.$tomorrowEvents.appendChild(
						$.tr({},
							$.td({ text: event.subject }),
							$.td({ text: event.start.format('h:mm a') }),
							$.td({ text: '-' }),
							$.td({ text: event.end.format('h:mm a') })
						)
					);
				}
			}
		}
		
		for (const $e of this.tag.querySelectorAll('.title h1, .time h2, .eventdisplay table')) {
			Object.assign($e.style, {
				color: this.foreground,
				fontFamily: `'${this.font}'`
			});
		}
		this.tag.querySelector('.eventdisplay table').style.fontSize = `${this.fontsize}px`;
	}
	Show() {
		
	}
	Hide() {
		
	}
	Destroy() {
		clearInterval(this.exchangeInterval);
	}
}
