class Widgetgalleryfrommetadata extends WidgetBase {
	constructor(options) {
		super(options);
		this.popupDialog = null;
		(async () => {
			let filteredMedia = [];
			if (this.container && this.container.hasOwnProperty('contenttree')) {
				filteredMedia = this.container.contenttree.filter(node => node.type != 1 && ApplyFilter(node, this.filters, Array.isArray(this.container.metadata) ? this.container.metadata : []));
				
				if (filteredMedia.length == 0 && this.useusedefaultimage && this.usedefaultimage) {
					filteredMedia.push(this.usedefaultimage);
				}
				
				const imagesToCache = filteredMedia;
				
				const imageCacheURLs = await Promise.all(imagesToCache.map(media => this.container.cacheMediaObject(media)));
				imageCacheURLs.forEach((cacheURL, index) => {
					if (cacheURL) {
						imagesToCache[index].cdnurl = cacheURL;
					}
				});
				const cacheData = await UpdateWidgetCache(filteredMedia, this.mediaid, this.container);
			} else {
				for (let i = 0; i < 10; ++i) {
					const canvas = $.canvas({
						width: 480,
						height: 480
					});
					const context = canvas.getContext('2d');
					context.fillStyle = 'white';
					context.fillRect(0, 0, canvas.width, canvas.height);
					context.strokeStyle = 'black';
					context.textAlign = 'center';
					context.font = '50px serif';
					context.strokeText(`Image ${i + 1}`, 240, 240);
					filteredMedia.push({
						name: `Image ${i + 1}`,
						type: 2,
						cdnurl: canvas.toDataURL()
					});
				}
			}
			
			this.tag.classList.add('widget-galleryfrommetadata');
			this.tag.innerHTML = '';
			Object.assign(this.tag.style, {
				margin: `${this.margins}px`,
				justifyContent: this.horizontalalignment
			});
			// If any of the media defines an order then sort using that
			if (filteredMedia.some(media => {
					const metadata = media.hasOwnProperty('metadata') ? media.metadata : {};
					return this.mediaorder_(metadata) != 0;
				})) {
				filteredMedia.sort((a, b) => {
					const metadataA = a.hasOwnProperty('metadata') ? a.metadata : {};
					const metadataB = b.hasOwnProperty('metadata') ? b.metadata : {};
					return this.mediaorder_(metadataA) - this.mediaorder_(metadataB);
				});
			} else {
				filteredMedia.sort((a, b) => TreeSort(a.name, b.name));
			}
			
			for (const media of filteredMedia) {
				// generated item metadata uses attributeName_(metadata) in the widget.js to supply metadata
				const metadata = media.hasOwnProperty('metadata') ? media.metadata : [];
				switch (media.type) {
					case 2: // Image
						const $image = $.div({ css: {
							width: `${this.mediawidth_(metadata)}px`,
							height: `${this.mediaheight_(metadata)}px`,
							margin: `${this.mediamargins_(metadata)}px`,
							borderWidth: `${this.mediaborderthickness_(metadata)}px`,
							borderColor: this.mediabordercolor_(metadata),
							borderRadius: `${this.mediaborderradius_(metadata)}px`,
							backgroundColor: this.mediabackgroundcolor_(metadata),
							boxShadow: this.mediadropshadow_(metadata) ? `${this.mediadropshadowleft_(metadata)}px ${this.mediadropshadowtop_(metadata)}px ${this.mediadropshadowblurradius_(metadata)}px ${this.mediadropshadowcolor_(metadata)}` : 'none'
						} });
						if (!this.editor) {
							$image.addEventListener('click', e => {
								const updateArrows = () => {
									if (this.showarrows) {
										this.$previous.hidden = this.popupDialog.scrollLeft == 0;
										this.$next.hidden = this.popupDialog.scrollLeft == this.popupDialog.scrollWidth - this.popupDialog.offsetWidth;
									}
								};
								this.popupDialog = document.body.appendChild($.div({ class: 'galleryfrommetadata-dialog', css: {
									backgroundColor: this.popupbackgroundcolor
								}, click: function() {
									this.remove();
									this.popupDialog = null;
								}, scroll: function() {
									updateArrows();
								} }));
								for (const swipeMedia of filteredMedia) {
									this.popupDialog.appendChild($.div({ class: 'image', css: {
										backgroundImage: `url('${swipeMedia.cdnurl}')`
									} }));
								}
								this.popupDialog.appendChild($.div({ class: ['navigate', !this.showarrows, 'disabled'], css: {
									color: this.popuparrowclosecolor
								} },
									this.$previous = $.i({ class: 'previous mdi mdi-chevron-left', click: e => {
										e.stopPropagation();
										this.popupDialog.scrollBy(-this.popupDialog.offsetWidth, 0);
										updateArrows();
									} }),
									$.span({ class: 'spacer' }),
									this.$next = $.i({ class: 'next mdi mdi-chevron-right', click: e => {
										e.stopPropagation();
										this.popupDialog.scrollBy(this.popupDialog.offsetWidth, 0);
										updateArrows();
									} })
								));
								// Close
								this.popupDialog.appendChild($.div({ class: 'close mdi mdi-close', css: {
									color: this.popuparrowclosecolor
								}, hidden: !this.showclose }));
								this.popupDialog.scrollBy(this.popupDialog.offsetWidth * filteredMedia.indexOf(media), 0);
								updateArrows();
							});
						}
						const img = new Image();
						img.setAttribute('crossOrigin', 'anonymous');
						img.onload = () => {
							if (this.crop_(metadata).split(',').length == 4) {
								const [left, top, width, height] = this.crop_(metadata).split(',').map(value => parseInt(value, 10));
								// Crop the image using a canvas object
								const canvas = $.canvas({ width, height });
								canvas.getContext('2d').drawImage(img, left, top, width, height, 0, 0, width, height);
								$image.style.backgroundImage = `url(${canvas.toDataURL()})`;
							} else {
								$image.style.backgroundImage = `url(${img.src})`;
							}
						};
						img.src = media.cdnurl;
						switch (this.fillstyle_(metadata)) {
							case 1:
								$image.style.backgroundSize = '100% 100%';
								break;
							case 2:
								$image.style.backgroundSize = 'cover';
								break;
							case 3:
								$image.style.backgroundSize = 'contain';
								break;
						}
						this.tag.appendChild($image);
						break;
					case 3: // Video
							
						break;
				}
			}
			this.ready();
		})();
	}
	Show() {
		
	}
	Hide() {
		if (this.popupDialog) {
			this.popupDialog.remove();
		}
	}
	Destroy() {
		if (this.popupDialog) {
			this.popupDialog.remove();
		}
	}
}