// Generates a divs for each page then on scroll determines which data to load and unload
async function RenderPDF(tag, url, scrollLeft, scrollTop, allowScroll, forceScrollbar, viewportWidth, viewportHeight, scale, width, height, cropX, cropY, cropWidth, cropHeight) {
	tag.innerHTML = '';
	
	var pdfDocument = null;
	var pdfLoading = false;
	var unloadPDFTimeout = null;
	if (scale == undefined) {
		scale = 100;
	}
	scale /= 100;
	// Pages to load one at a time
	var pagesQueue = [];
	var pageRendering = false;
	
	const TagDeleted = () => {
		if (!document.body.contains(tag)) {
			return true;
		} else {
			return false;
		}
	};
	
	function loadPDF() {
		return new Promise(async resolve => {
			if (!pdfLoading && pdfDocument == null) {
				pdfLoading = true;
				const pdfDoc = await pdfjsLib.getDocument({ url, disableRange: true }).promise;;
				pdfLoading = false;
				pdfDocument = pdfDoc;
				resolve(pdfDocument);
			} else {
				resolve(pdfDocument);
			}
			// Reset the pdf timeout. This code will unload the pdf document after 30 seconds of not being used since all references are removed
			clearTimeout(unloadPDFTimeout);
			unloadPDFTimeout = setTimeout(() => {
				unloadPDFTimeout = null;
				// Removing this reference will allow the pdf to be garbage collected
				if (pdfDocument) {
					pdfDocument.destroy();
				}
				pdfDocument = null;
			}, 15 * 1000);
		});
	}
	
	if (allowScroll) {
		var resetTimer = null;
		var startResetTimer = () => {
			clearTimeout(resetTimer);
			resetTimer = setTimeout(() => {
				if (TagDeleted()) {
					return;
				}
				this.$pdfPages.animate({ scrollLeft: scrollLeft, scrollTop: scrollTop });
			}, 90 * 1000);
		};
		let startY;
		let pointerdown = false;
		if (tag.hasOwnProperty('lastPointerDown')) {
			tag.removeEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', tag.lastPointerDown);
		}
		tag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', tag.lastPointerDown = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			pointerdown = true;
			startY = e.pageY;
			startResetTimer();
		});
		if (tag.hasOwnProperty('lastPointerMove')) {
			tag.removeEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', tag.lastPointerMove);
		}
		tag.addEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', tag.lastPointerMove = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			if (pointerdown) {
				this.$pdfPages.scrollTop = Math.round(this.$pdfPages.scrollTop) + startY - e.pageY;
				startY = e.pageY;
			}
		});
		if (tag.hasOwnProperty('lastPointerUp')) {
			tag.removeEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', tag.lastPointerUp);
		}
		tag.addEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', tag.lastPointerUp = e => {
			pointerdown = false;
			startResetTimer();
		});
	}
	
	var processQueue = async () => {
		if (!pageRendering && pagesQueue.length != 0) {
			var pageNumber = pagesQueue.shift();
			var pageTag = this.$pdfPages.children[pageNumber];
			if (pageTag.matches(':empty')) {
				pageRendering = true;
				const pdfDoc = await loadPDF();
				const page = await pdfDoc.getPage(pageNumber + 1);
				var viewport = page.getViewport({ scale });
				const canvas = $.canvas({});
				pageTag.appendChild(canvas);
				canvas.width = viewport.width;
				canvas.height = viewport.height;
				// Render PDF page into canvas context
				var renderContext = {
					canvasContext: canvas.getContext('2d'),
					viewport: viewport
				};
				var renderTask = page.render(renderContext);
				try {
					// Wait for rendering to finish
					await renderTask.promise;
					pageRendering = false;
					processQueue();
				} catch (error) {
					console.log('Render task failed: ' + error);
				}
			} else {
				// Page was already rendered
				processQueue();
			}
		}
	};
	
	const loadPages = () => {
		if (TagDeleted()) {
			return;
		}
		var currentPageNumber = null;
		var pagesScrollTop = Math.round(this.$pdfPages.scrollTop);
		var height = this.$pdfPages.offsetHeight;
		var top = 0;
		for (const [pageNumber, $page] of [...this.$pdfPages.children].entries()) {
			const pageHeight = $.outerHeight($page);
			if (top - pagesScrollTop <= height && top - pagesScrollTop + pageHeight > 0) {
				pagesQueue.push(pageNumber);
				if (currentPageNumber == null) {
					currentPageNumber = pageNumber;
				}
			} else {
				// The canvas is outside of the viewable area, delete it to save memory
				this.innerHTML = '';
			}
			top += pageHeight;
		}
		processQueue();
		return currentPageNumber == null ? 0 : currentPageNumber;
	};
	
	const pdfDoc = await loadPDF();
	var pagePromises = [];
	for (var i = 0; i < pdfDoc.numPages; ++i) {
		pagePromises.push(pdfDoc.getPage(i + 1));
	}
	const pages = await Promise.all(pagePromises);
	if (TagDeleted()) {
		return;
	}
	// Calculate a scale value to keep the pages a reasonable width and height
	var maximumWidth = 0;
	var maximumHeight = 0;
	var pageDimensions;
	var calculatePageDimensions = () => {
		pageDimensions = pages.map(page => {
			var viewport = page.getViewport({ scale });
			maximumWidth = Math.max(maximumWidth, viewport.width);
			maximumHeight = Math.max(maximumHeight, viewport.height);
			return {
				width: viewport.width,
				height: viewport.height
			};
		});
	};
	calculatePageDimensions();
	var maximumDimension = 1920;
	if (maximumWidth > maximumDimension || maximumHeight > maximumDimension) {
		scale = maximumDimension / Math.max(maximumWidth, maximumHeight);
	}
	calculatePageDimensions();
	tag.innerHTML = '';
	// Generate all the pages as divs
	tag.classList.add('widget-pdf');
	tag.appendChild(
		this.$pdfPages = $.div({ class: 'pdf-pages', css: {
			width: viewportWidth != undefined ? `${viewportWidth}px` : '100%',
			height: viewportWidth != undefined ? `${viewportHeight}px` : '100%',
		} })
	);
	this.$pdfPages.append(...pageDimensions.map(page => {
			return $.div({ class: 'pdf-page', css: {
				width: `${page.width}px`,
				height: `${page.height}px`
			} });
		})
	);
	this.$pdfPages.scrollLeft = scrollLeft;
	this.$pdfPages.scrollTop = scrollTop;
	
	if (forceScrollbar) {
		this.$pdfPages.style.overflow = 'auto';
	}
	// Used in the widget rendering to show only the crop region of the PDF
	if (width != undefined && height != undefined && cropX != undefined && cropY != undefined && cropWidth != undefined && cropHeight != undefined) {
		Object.assign(this.$pdfPages.style, {
			'transform-origin': '0 0',
			transform: `scale(${width / cropWidth},${height / cropHeight}) translate(${-cropX}px,${-cropY}px)`
		});
	}
	
	var top = 0;
	for (const [index, $page] of [...this.$pdfPages.children].entries()) {
		pageDimensions[index].top = top;
		top += $.outerHeight($page);
	}
	
	// Register the scroll event at a later point does it doesn't process scroll events
	await new Promise(resolve => {
		setTimeout(() => {
			this.$pdfPages.onscroll = loadPages;
			resolve();
		}, 0);
	});
	if (TagDeleted()) {
		return;
	}
	var currentPageNumber = loadPages();
	return [pageDimensions, currentPageNumber];
}

class Widgetpdf extends WidgetBase {
	constructor(options) {
		super(options);
		(async () => {
			let pdf = this.pdf;
			if (this.container) {
				const cacheURL = await this.container.cacheMediaObject(pdf);
				if (cacheURL) {
					pdf.cdnurl = cacheURL;
				}
			}
			var cacheData =
			{
				pdf
			};
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			pdf = cacheData.pdf;
			if (pdf) {
				await RenderPDF(this.tag, pdf.cdnurl, this.scrollleft, this.scrolltop, this.allowscroll && !this.editor, false, this.viewportwidth, this.viewportheight, this.scale, this.width, this.height, this.cropx, this.cropy, this.cropwidth, this.cropheight);
			}
			this.ready();
		})();
	}
}
