class Widgetpowerbireport extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.refreshInterval = null;
		this.pageChangeInterval = null;
		this.embedTokenTimeout = null;
		this.errorRenderReportTimeout = null;
		this.currentPage = 0;
		
		if (this.mediaid == -1) {
			this.showFatalError('The playlist must be saved before the Power BI widget can be displayed.', false);
		} else {
			this.renderReport();
			this.ready();
		}
	}
	async Update()
	{
		if (this.TagDeleted()) {
			return;
		}
		
		try {
			const report = powerbi.get(this.tag);
			await report.refresh();
		} catch (exception) {
			console.log(exception);
		}
		await this.startPageChange();
	}
	showFatalError(message, refreshReport = true) {
		clearTimeout(this.embedTokenTimeout);
		clearInterval(this.refreshInterval);
		clearInterval(this.pageChangeInterval);
		powerbi.reset(this.tag);
		this.tag.innerHTML = '';
		this.tag.appendChild(
			$.div({ class: 'buttonhelp', text: message })
		);
		
		if (refreshReport && this.errorRenderReportTimeout == null) {
			this.errorRenderReportTimeout = setTimeout(this.renderReport.bind(this), 30 * 1000);
		}
	}
	async startPageChange() {
		try {
			if (this.pagechangedelay == -1) {
				// Set a specific page number
				let pageNumber = 0;
				try {
					pageNumber = parseInt(this.pagenumber, 10) - 1;
				} catch (exception) {
				}
				const report = powerbi.get(this.tag);
				const pages = await report.getPages();
				if (pageNumber >= 0 && pageNumber < pages.length) {
					pages[pageNumber].setActive();
				}
			} else {
				clearInterval(this.pageChangeInterval);
				const report = powerbi.get(this.tag);
				const pages = await report.getPages();
				if (pages.length > 0) {
					pages[0].setActive();
				}
				clearInterval(this.pageChangeInterval);
				if (pages.length > 1) {
					this.pageChangeInterval = setInterval(async () => {
						try {
							const report = powerbi.get(this.tag);
							const pages = await report.getPages();
							if (pages.length == 0) {
								return;
							}
							this.currentPage++;
							if (this.currentPage >= pages.length) {
								this.currentPage = 0;
							}
							pages[this.currentPage].setActive();
						} catch (exception) {
							console.log(exception);
						}
					}, this.pagechangedelay * 1000);
				}
			}
		} catch (exception) {
			// Do nothing
			console.log(exception);
		}
	}
	async generateEmbedToken() {
		return await new Promise(async (resolve, reject) => {
			try {
				const data = await $.http({
					method: 'GET',
					url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/powerbiaccesstoken.php',
					data: {
						mediaid: this.mediaid
					},
					dataType: 'json'
				}).promise;
				if (data.success) {
					const currentTime = Date.now();
					const expiration = Date.parse(data.data.token.expiration);
					if (expiration < currentTime + 1 * 60 * 1000) {
						console.log('Microsoft returned a token that is basically already expired... pretend it is not expired for 1 minute so we do not request too much');
						data.data.token.expiration = (new Date(Date.now() + 1 * 60 * 1000)).toISOString();
					}
					resolve(data.data);
				} else {
					// data.error can be 'Timeout' indicating the endpoint timed out. This is rare, but complicates things as it must be handled by trying again later.
					reject(new Error(data.error));
				}
			} catch (exception) {
				reject(new Error('/get/powerbiaccesstoken.php failed'));
			}
		});
	}
	async setTokenExpirationListener(tokenExpiration) {
		// get current time
		const currentTime = Date.now();
		const expiration = Date.parse(tokenExpiration);
		
		// time until token refresh in milliseconds
		const timeout = expiration - currentTime;
		
		// if token already expired, generate new token and set the access token
		if (timeout <= 0) {
			console.log("Updating Report Embed Token");
			try {
				await this.updateToken();
			} catch (exception) {
				console.log(exception);
				this.showFatalError('Token error, trying again...');
			}
		} else {
			console.log(`Report Embed Token will be updated in ${timeout} milliseconds.`);
			clearTimeout(this.embedTokenTimeout);
			this.embedTokenTimeout = setTimeout(async () => {
				try {
					await this.updateToken();
				} catch (exception) {
					console.log(exception);
					this.showFatalError('Token error, trying again...');
				}
			}, timeout);
		}
	}
	async updateToken() {
		if (this.TagDeleted()) {
			return;
		}
		
		// Generate new EmbedToken
		const { token } = await this.generateEmbedToken();
		try {
			const report = powerbi.get(this.tag);
			await report.setAccessToken(token.token);
		} catch (exception) {
			// Do nothing
		}
		
		// Set token expiration listener
		// result.expiration is in ISO format
		await this.setTokenExpirationListener(token.expiration);
	}
	async renderReport() {
		this.loaded = Promise.race([
			new Promise((resolve, reject) => setTimeout(() => reject(new Error('timing out...')), 10000)),
			new Promise(async (resolve, reject) => {
				// Try to load at least 3 times. This handles the unlikely case of an endpoint timing out with our CURL requests (which returns http code 0)
				for (let tryAgain = 0; tryAgain < 3; ++tryAgain) {
					if (this.TagDeleted()) {
						reject(new Error('widget deleted'));
						return;
					}
					try {
						this.errorRenderReportTimeout = null;
						powerbi.reset(this.tag);
						this.tag.innerHTML = '';
						// https://microsoft.github.io/PowerBI-JavaScript/demo/v2-demo/index.html
					
						const reportId = this.reportid;
						const { embedURL, token } = await this.generateEmbedToken();
					
						// Get models. models contains enums that can be used.
						var models = window['powerbi-client'].models;
					
						// Embed configuration used to describe the what and how to embed.
						// This object is used when calling powerbi.embed.
						// This also includes settings and this such as filters.
						// You can find more information at https://github.com/Microsoft/PowerBI-JavaScript/wiki/Embed-Configuration-Details.
						var config = {
							type: 'report',
							tokenType: models.TokenType.Embed,
							accessToken: token.token,
							embedUrl: embedURL,
							id: reportId,
							permissions: models.Permissions.Read,
							settings: {
								filterPaneEnabled: false,
								navContentPaneEnabled: false,
								hideErrors: true
							}
						};
						if (this.TagDeleted()) {
							reject(new Error('widget deleted'));
							return;
						}
						// Embed the report and display it within the div container.
						const report = powerbi.embed(this.tag, config);
					
						report.off('loaded');
						// Report.on will add an event handler which prints to Log window.
						report.on('loaded', async () => {
							if (this.TagDeleted()) {
								reject(new Error('widget deleted'));
								return;
							}
						
							try {
								const filters = await report.getFilters();
								this.allFilters = { filters: filters == undefined ? [] : filters, pages: [] };
								for (const page of await report.getPages()) {
									const pageFilters = { page, filters: await page.getFilters(), visuals: [] };
									for (const visual of await page.getVisuals()) {
										pageFilters.visuals.push({ visual, filters: await visual.getFilters() });
									}
									this.allFilters.pages.push(pageFilters);
								}
							} catch (exception) {
								console.log('Could not get filter information');
								reject(new Error('Could not get report filter information'));
							}
							// Note the above filter data structures must be made before we resolve
							resolve();
						
							if (this.bookmark.trim() != '') {
								try {
									for (const bookmark of await report.bookmarksManager.getBookmarks()) {
										if (bookmark.displayName == this.bookmark.trim()) {
											report.bookmarksManager.apply(bookmark.name);
											break;
										}
									}
								} catch (exception) {
									console.log('Failed to set bookmark', exception);
								}
							}
							const checkFilter = (filter, filterString) => {
								if (filter.hasOwnProperty('displaySettings') && filter.displaySettings.hasOwnProperty('displayName') && filterString.split(',').map(name => name.trim()).includes(filter.displaySettings.displayName)) {
									if (this.reportfiltersvalues.hasOwnProperty(filter.displaySettings.displayName)) {
										const filterValues = this.reportfiltersvalues[filter.displaySettings.displayName];
										let index = 0;
										const getValue = () => {
											if (index >= filterValues.length) {
												return { override: false };
											}
											return filterValues[index];
										};
										if (filter.hasOwnProperty('conditions')) {
											for (const condition of filter.conditions) {
												if (condition.hasOwnProperty('value')) {
													if (getValue().override) {
														condition.value = getValue().value;
													}
													index++;
												}
											}
										} else if (filter.hasOwnProperty('values')) {
											if (getValue().override) {
												filter.values = getValue().value.split(',').map(x => x.trim());
											}
											index++;
										}
									}
									return true;
								}
								return false;
							};
							if (this.togglereportfilters) {
								try {
									// Remove filters and selectively add custom named ones back in
									report.removeFilters();
									await report.setFilters(JSON.parse(JSON.stringify(this.allFilters.filters)).filter(filter => checkFilter(filter, this.reportfilters)));
								} catch (exception) {
									console.log('Error adding report filters', exception);
								}
							}
							if (this.togglepagefilters) {
								try {
									for (const { page, filters } of this.allFilters.pages) {
										// Remove filters and selectively add custom named ones back in
										page.removeFilters();
										await page.setFilters(JSON.parse(JSON.stringify(filters)).filter(filter => checkFilter(filter, this.pagefilters)));
									}
								} catch (exception) {
									console.log('Error adding page filters', exception);
								}
							}
							if (this.togglevisualfilters) {
								try {
									for (const { visuals } of this.allFilters.pages) {
										for (const { visual, filters } of visuals) {
											// Remove filters and selectively add custom named ones back in
											visual.removeFilters();
											await visual.setFilters(JSON.parse(JSON.stringify(filters)).filter(filter => checkFilter(filter, this.visualfilters)));
										}
									}
								} catch (exception) {
									console.log('Error adding page filters', exception);
								}
							}
							await this.startPageChange();
							console.log('Power BI Report Loaded');
							await this.setTokenExpirationListener(token.expiration);
						
							if (this.refresh != -1) {
								clearInterval(this.refreshInterval);
								this.refreshInterval = setInterval(this.Update.bind(this), this.refresh * 1000);
							}
						});
						report.off('rendered');
						// Report.on will add an event handler which prints to Log window.
						report.on('rendered', () => {
							console.log('Power BI Report Rendered');
						});
						report.off('error');
						report.on('error', event => {
							if (this.TagDeleted()) {
								reject(new Error('widget deleted'));
								return;
							}
							reject(new Error('widget error'));
							// If error is not Fatal log the error and continue
							if (event.detail.level !== models.TraceType.Fatal) {
								console.log(JSON.stringify(event.detail));
								this.showFatalError(JSON.stringify(event.detail));
							} else if (event.detail.message === models.CommonErrorCodes.TokenExpired) {
								(async () => {
									try {
										await this.updateToken();
									} catch (exception) {
										console.log(exception);
										this.showFatalError('Token error, trying again...');
									}
								})();
							} else {
								console.log(`Power BI Report: ${event.detail.detailedMessage}`);
								this.showFatalError(JSON.stringify(event.detail.detailedMessage));
							}
						});
						return;
					} catch (exception) {
						if (exception.message == 'Timeout') {
							console.log(exception);
							// generateEmbedToken simply timedout. This isn't a fatal error, and goes away after a bit
							// Wait 2 seconds
							await new Promise(resolve => setTimeout(resolve, 2000));
						} else {
							reject(new Error('widget exception'));
							console.log(exception);
							this.showFatalError(exception);
						}
					}
				}
				reject(new Error('Failed 3 times to load report...'));
				this.showFatalError('Failed 3 times to load report...');
			})
		]);
		try {
			await this.loaded;
		} catch (exception) {
		}
	}
	/*
	-filterNameToDefaultValues: {}
	-filterValues: []
	-filters []
		-values:
		-conditions []
			-value:
	-pages
		-page
			-filterValues: []
			-filters []
				-values:
				-conditions []
					-value:
			-visuals
				-filterValues: []
				-filters []
					-values:
					-conditions []
						-value:
	*/
	async GetFilterTree() {
		if (this.loaded) {
			try {
				await this.loaded;
				const filterNameToDefaultValues = {};
				const appendCustomFilterValues = (filter, filterValues) => {
					if (filter.hasOwnProperty('displaySettings') && filter.displaySettings.hasOwnProperty('displayName')) {
						// Has custom name
						if (filter.hasOwnProperty('conditions')) {
							for (const condition of filter.conditions) {
								if (condition.hasOwnProperty('value')) {
									filterValues.push(condition.value);
								}
							}
						} else if (filter.hasOwnProperty('values')) {
							filterValues.push(JSON.parse(JSON.stringify(filter.values)));
						}
						filterNameToDefaultValues[filter.displaySettings.displayName] = filterValues;
					}
				};
				// Create a corresponding array of filterValues
				// We don't store this information in the filter since we're going to add filters to the report later and we don't want to change the object
				this.allFilters.filterValues = [];
				for (const filter of this.allFilters.filters) {
					const filterValues = [];
					this.allFilters.filterValues.push(filterValues);
					appendCustomFilterValues(filter, filterValues);
				}
				for (const page of this.allFilters.pages) {
					page.filterValues = [];
					page.hasCustomFilterWithValues = false;
					for (const filter of page.filters) {
						const pageFilterValues = [];
						page.filterValues.push(pageFilterValues);
						appendCustomFilterValues(filter, pageFilterValues);
						page.hasCustomFilterWithValues = page.hasCustomFilterWithValues || pageFilterValues.length != 0;
					}
					for (const visual of page.visuals) {
						visual.filterValues = [];
						visual.hasCustomFilterWithValues = false;
						for (const filter of visual.filters) {
							const visualFilterValues = [];
							visual.filterValues.push(visualFilterValues);
							appendCustomFilterValues(filter, visualFilterValues);
							visual.hasCustomFilterWithValues = visual.hasCustomFilterWithValues || visualFilterValues.length != 0;
						}
						page.hasCustomFilterWithValues = page.hasCustomFilterWithValues || visual.hasCustomFilterWithValues;
					}
				}
				this.allFilters.filterNameToDefaultValues = filterNameToDefaultValues;
				return this.allFilters;
			} catch (exception) {
				throw exception;
			}
		} else {
			throw new Error('widget not loaded');
		}
	}
	Destroy() {
		clearTimeout(this.embedTokenTimeout);
		clearInterval(this.refreshInterval);
		clearInterval(this.pageChangeInterval);
		clearInterval(this.errorRenderReportTimeout);
		powerbi.reset(this.tag);
		this.tag.innerHTML = '';
	}
}
