class Widgetreaderboard extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.resetTimeout = null;
		this.loadEventsInterval = null;
		this.fpsInterval = null;
		this.fpsRequestAnimationFrame = null;
		this.pageChangeInterval = null;
		this.noEventImageInterval = null;
		this.pageOffsetDifferenceRequestAnimationFrame = null;
		
		this.displayedEvents = [];
		this.newAllEvents = [];
		
		this.pageIndex = 0;
		this.pageCount = 0;
		
		this.currentFPS = 0;
		
		this.startOffset = 0;
		this.pageOffsetDifference = 0;
		
		this.absolutePositionTransition = false;
		this.pageTransitionSize = 0;
		this.slideTransitionWeight = 1;
		
		this.delay = 15;
		
		this.noEventImageTransitionIndex = 0;
		
		this.oldServerOffset = this.player.serverOffset;
		
		this.onlineModeTimer = null;
		this.offlineMode = null;
		
		// Cached data
		this.events = [];
		this.groups = [];

		// Cached player data in this.widgetData.readerboard
		this.name = '';
		this.groupids = [];
		this.overridebackgroundimage = false;
		this.backgroundimage = null;
		this.backgroundimagecoverfill = false;
		
		// If the player has not been updated yet then this field will be empty. Kind of an edge case as anyone using Readerboard will have pressed update to set room information
		if (this.widgetData.hasOwnProperty('readerboard')) {
			this.name = this.widgetData.readerboard.name;
			this.groupids = this.widgetData.readerboard.groupids;
			this.overridebackgroundimage = this.widgetData.readerboard.overridebackgroundimage;
			this.backgroundimage = JSON.parse(JSON.stringify(this.widgetData.readerboard.backgroundimage));
			this.backgroundimagecoverfill = this.widgetData.readerboard.backgroundimagecoverfill;
		}
		
		this.TestOnlineMode();
	}
	ClearTimers() {
		clearTimeout(this.resetTimeout);
		this.resetTimeout = null;
		
		clearInterval(this.loadEventsInterval);
		this.loadEventsInterval = null;
		
		clearInterval(this.fpsInterval);
		this.fpsInterval = null;
		
		cancelAnimationFrame(this.fpsRequestAnimationFrame);
		this.fpsRequestAnimationFrame = null;
		
		clearInterval(this.pageChangeInterval);
		this.pageChangeInterval = null;
		
		clearInterval(this.noEventImageInterval);
		this.noEventImageInterval = null;
		
		cancelAnimationFrame(this.pageOffsetDifferenceRequestAnimationFrame);
		this.pageOffsetDifferenceRequestAnimationFrame = null;
	}
	IncrementPage() {
		if (this.TagDeleted()) {
			return;
		}
		if (this.pageCount == 1) {
			return;
		}
		if (this.pageIndex < this.pageCount - 1) {
			this.pageIndex++;
			this.pageOffsetDifference = this.pageTransitionSize;
			this.slideTransitionWeight = 1;
		} else {
			this.pageIndex = 0;
			if (this.absolutePositionTransition) {
				this.pageOffsetDifference = this.pageTransitionSize;
			} else {
				this.pageOffsetDifference = -this.pageTransitionSize * (this.pageCount - 1);
			}
			this.slideTransitionWeight = this.pageCount - 1;
		}
		this.AnimateTransition();
	}
	StartPageTransition() {
		clearInterval(this.pageChangeInterval);
		this.pageChangeInterval = setInterval(this.IncrementPage.bind(this), 15 * 1000);
	}
	StopPageTransition() {
		clearInterval(this.pageChangeInterval);
	}
	updatePositions() {
		for (const [index, $page] of [...this.tag.querySelectorAll('.pages > div')].entries()) {
			switch (this.eventlistoverflowbehavior.slice(0, -2)) {
				case 'pagified':
					$page.style.opacity = index == this.pageIndex ? 1 : 0;
					break;
				case 'pagifiedfade':
					var position = -this.pageIndex * this.pageTransitionSize + this.pageOffsetDifference;
					if (this.pageIndex == 0 && this.pageCount != 1 &&  index == this.pageCount - 1) {
						position -= this.pageCount * this.pageTransitionSize;
					}
					const opacity = Math.max(0, Math.min(1, 1 - Math.abs(position + index * this.pageTransitionSize) / this.pageTransitionSize * 2));
					$page.style.opacity = opacity;
					break;
				case 'pagifiedslideleft':
					$page.style.transform = `translate(${-this.pageIndex * this.$pages.offsetWidth + this.pageOffsetDifference}px,0)`;
					break;
				case 'pagifiedslideup':
					$page.style.transform = `translate(0,${-this.pageIndex * this.$pages.offsetHeight + this.pageOffsetDifference}px)`;
					break;
			}
		}
	}
	AnimateTransition() {
		if (this.TagDeleted()) {
			return;
		}
		if (this.pageOffsetDifferenceRequestAnimationFrame == null) {
			if (this.pageOffsetDifference != 0) {
				this.pageOffsetDifferenceRequestAnimationFrame = requestAnimationFrame(this.AnimateTransition.bind(this));
			} else {
				this.StartPageTransition();
			}
		} else {
			// Weight and exponent and be tweaked to change the swipe behavior. Depending on tablet performance some values might feel more natural
			var weight = 10;
			var exponent = 4;
			
			var fpsCorrection = 60 / Math.min(Math.max(this.currentFPS, 10), 60);
			var modifier = -Math.sign(this.pageOffsetDifference) * fpsCorrection * weight * Math.max(0.1, (Math.abs(Math.pow(this.pageTransitionSize * this.slideTransitionWeight, exponent)) - Math.abs(Math.pow(this.pageOffsetDifference, exponent))) / Math.abs(Math.pow(this.pageTransitionSize * this.slideTransitionWeight, exponent)));
			if (this.absolutePositionTransition) {
				// Only used for the fade transition
				weight = 2;
				modifier = -Math.sign(this.pageOffsetDifference) * fpsCorrection * weight;
			}
			if (Math.sign(this.pageOffsetDifference) != Math.sign(this.pageOffsetDifference + modifier)) {
				this.pageOffsetDifference = 0;
			} else {
				this.pageOffsetDifference += modifier;
			}
			
			this.updatePositions();
			if (this.pageOffsetDifference != 0) {
				this.pageOffsetDifferenceRequestAnimationFrame = requestAnimationFrame(this.AnimateTransition.bind(this));
			} else {
				this.pageOffsetDifferenceRequestAnimationFrame = null;
				this.StartPageTransition();
			}
		}
	}
	ViewEventsChanged(newEvents) {
		return newEvents.length != this.displayedEvents.length || this.displayedEvents.some((event, index) => {
			return !event.start.isSame(newEvents[index].start) ||
					!event.end.isSame(newEvents[index].end) ||
					event.title != newEvents[index].title ||
					event.room != newEvents[index].room ||
					event.name != newEvents[index].name;
		});
	}
	CurrentState(allEvents) {
		var newState =
		{
			events: [],
			ungroupedEvents: [],
			groupedEvents: []
		};
		const currentTime = this.player.currentTime;
		var isCountLimited = false;
		var countLimit = 0;
		var isRangeLimited = true;
		// Replace every name with the most common case of that name. So if "Brandon" occurs more than "brandon" it'll change all the matching names to "Brandon"
		var nameCases = {};
		for (const event of allEvents) {
			if (!nameCases.hasOwnProperty(event.name)) {
				nameCases[event.name.toLowerCase()] =
				{
					majorityCase: '',
					majorityCount: 0,
					cases: {}
				};
			}
			var nameCase = nameCases[event.name.toLowerCase()];
			if (!nameCase.cases.hasOwnProperty(event.name)) {
				nameCase.cases[event.name] = 0;
			}
			nameCase.cases[event.name]++;
			if (nameCase.cases[event.name] > nameCase.majorityCount) {
				nameCase.majorityCase = event.name;
				nameCase.majorityCount = nameCase.cases[event.name];
			}
		}
		var nameReplaceName = {};
		for (var nameCase in nameCases) {
			nameReplaceName[nameCase] = nameCases[nameCase].majorityCase;
		}
		for (const event of allEvents) {
			event.name = nameReplaceName[event.name.toLowerCase()];
		}
		
		var ungroupedEventsExpired = [];
		var groupToEvents = {};
		var groupToEventsExpired = {};
		for (const event of allEvents) {
			var start = currentTime.clone().startOf('day');
			var end = currentTime.clone().endOf('day');
			switch (this.groupby == 'none' ? this.upcomingeventcount : this.upcomingeventcountpergroup) {
				case 'allevents':
					start = moment(0);
					end = moment('2500', 'YYYY');
					break;
				case '1event': // Show 1 to 10 events for each group
				case '2event':
				case '3event':
				case '4event':
				case '5event':
				case '6event':
				case '7event':
				case '8event':
				case '9event':
				case '10event':
					isCountLimited = true;
					isRangeLimited = false;
					countLimit = parseInt((this.groupby == 'none' ? this.upcomingeventcount : this.upcomingeventcountpergroup).slice(0, -'event'.length), 10);
					if (event.end.isSameOrAfter(currentTime.clone().startOf('day')) && event.start.isBefore(currentTime.clone().endOf('day'))) {
						if (this.groupby == 'none') {
							newState.ungroupedEvents.push(event);
						} else {
							var group = event[this.groupby].toLowerCase();
							if (!groupToEvents.hasOwnProperty(group)) {
								groupToEvents[group] = [];
							}
							groupToEvents[group].push(event);
						}
					}
					break;
				case 'oneday': // Today's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day');
					break;
				case 'twoday': // Today and Tomorrow's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(1, 'day');
					break;
				case 'threeday': // Today and Next Two Day's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(2, 'day');
					break;
				case 'fourday': // Today and Next Three Day's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(3, 'day');
					break;
				case 'fiveday': // Today and Next Four Day's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(4, 'day');
					break;
				case 'sixday': // Today and Next Five Day's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(5, 'day');
					break;
				case 'sevenday': // Today and Next Six Day's Events
					start = currentTime.clone().startOf('day');
					end = currentTime.clone().endOf('day').add(6, 'day');
					break;
				case 'oneweek': // This Week's Events
					start = currentTime.clone().startOf('week');
					end = currentTime.clone().endOf('week');
					break;
				case 'twoweek': // This Week and Next Week's Events
					start = currentTime.clone().startOf('week');
					end = currentTime.clone().endOf('week').add(1, 'week');
					break;
				case 'threeweek': // This Week and the Next Two Week's Events
					start = currentTime.clone().startOf('week');
					end = currentTime.clone().endOf('week').add(2, 'week');
					break;
				case 'fourweek': // This Week and the Next Three Week's Events
					start = currentTime.clone().startOf('week');
					end = currentTime.clone().endOf('week').add(3, 'week');
					break;
				case 'onemonth': // This Month's Events
					start = currentTime.clone().startOf('month');
					end = currentTime.clone().endOf('month');
					break;
			}
			
			if (isRangeLimited) {
				if (event.end.isAfter(start) && event.start.isBefore(end)) {
					if (this.groupby == 'none') {
						if (event.end.isAfter(currentTime)) {
							newState.ungroupedEvents.push(event);
						} else {
							ungroupedEventsExpired.push(event);
						}
					} else {
						for (const group of event[this.groupby].split(';').map(group => group.trim().toLowerCase())) {
							if (!groupToEvents.hasOwnProperty(group)) {
								groupToEvents[group] = [];
							}
							if (!groupToEventsExpired.hasOwnProperty(group)) {
								groupToEventsExpired[group] = [];
							}
							if (event.end.isAfter(currentTime)) {
								groupToEvents[group].push(event);
							} else {
								groupToEventsExpired[group].push(event);
							}
						}
					}
				}
			}
		}
		
		if (isRangeLimited) {
			switch (this.expiredevents) {
				case 'none':
					// Do nothing
					break;
				case 'all':
					if (this.groupby == 'none') {
						newState.ungroupedEvents = newState.ungroupedEvents.concat(ungroupedEventsExpired);
					} else {
						for (var group in groupToEvents) {
							groupToEvents[group] = groupToEvents[group].concat(groupToEventsExpired[group]);
						}
					}
					break;
				case '1event1hour':
					if (this.groupby == 'none') {
						if (ungroupedEventsExpired.length != 0 && ungroupedEventsExpired[ungroupedEventsExpired.length - 1].end.isAfter(currentTime.clone().add(-1, 'hour'))) {
							newState.ungroupedEvents.push(ungroupedEventsExpired[ungroupedEventsExpired.length - 1]);
						}
					} else {
						for (var group in groupToEvents) {
							if (groupToEventsExpired[group].length != 0 && groupToEventsExpired[group][groupToEventsExpired[group].length - 1].end.isAfter(currentTime.clone().add(-1, 'hour'))) {
								groupToEvents[group].push(groupToEventsExpired[group][groupToEventsExpired[group].length - 1]);
							}
						}
					}
					break;
			}
		}
		
		var groupReplaceName = {};
		for (const group of this.groups) {
			if (this.groupby == 'room' && !groupToEvents.hasOwnProperty(group.name.toLowerCase())) {
				groupToEvents[group.name.toLowerCase()] = [];
			}
			groupReplaceName[group.name.toLowerCase()] = group.replacement == '' ? group.name : group.replacement;
		}
		
		for (var group in groupToEvents) {
			if (!groupReplaceName.hasOwnProperty(group.toLowerCase())) {
				groupReplaceName[group.toLowerCase()] = group;
			}
			// Only include groups with no events if there's no event text defined
			if (this.groupnoeventstext != '' || groupToEvents[group.toLowerCase()].length != 0) {
				newState.groupedEvents.push({ group: this.groupby == 'room' ? groupReplaceName[group.toLowerCase()] : nameReplaceName[group], events: groupToEvents[group.toLowerCase()] });
			}
		}
		
		newState.groupedEvents.sort((a, b) => a.group.toLowerCase().localeCompare(b.group.toLowerCase()));
		
		// Filtering expired events isn't done until this point because there is a case where we want one previous event if there's room
		if (isCountLimited) {
			while (newState.ungroupedEvents.length > countLimit) {
				newState.ungroupedEvents.pop();
			}
			for (const group of newState.groupedEvents) {
				group.moreEvents = false;
				var expiredEvents = [];
				var timeRangeEvents = [];
				for (const event of group.events) {
					// Only check for a last event if show expired events is disabled since the last event will be included if it's enabled
					// A last event is only included if it expired in the last hour
					if (event.end.isAfter(currentTime)) {
						timeRangeEvents.push(event);
					} else if (this.expiredevents == '1event1hour' && event.end.isBefore(currentTime) && event.end.isAfter(currentTime.clone().add(-1, 'hour'))) {
						// This could add multiple events
						expiredEvents.push(event);
					} else if (this.expiredevents == 'all') {
						expiredEvents.push(event);
					}
				}
				// Truncate the events array so we only have countLimit number of events
				while (timeRangeEvents.length > countLimit) {
					group.moreEvents = true;
					timeRangeEvents.pop();
				}
				// If expried events is set to 1 hour Add expired events until we reach the count
				while (this.expiredevents == 'all' && timeRangeEvents.length < countLimit && expiredEvents.length != 0) {
					timeRangeEvents.unshift(expiredEvents.pop());
				}
				// If we have room for one more event then add the last event
				if (this.expiredevents == '1event1hour' && timeRangeEvents.length < countLimit && expiredEvents.length != 0) {
					timeRangeEvents.unshift(expiredEvents.pop());
				}
				
				group.events = timeRangeEvents;
			}
		}
		
		for (const group of newState.groupedEvents) {
			group.events.sort((a, b) => a.start.unix() - b.start.unix());
		}
		
		newState.ungroupedEvents.sort((a, b) => a.start.unix() - b.start.unix());
		
		if (this.groupby == 'none') {
			newState.events = newState.ungroupedEvents.map(event => event);
		} else {
			for (const group of newState.groupedEvents) {
				for (const event of group.events) {
					newState.events.push(event);
				}
			}
		}
		
		return newState;
	}
	async ResetView(forceUpdate) {
		if (this.TagDeleted()) {
			return;
		}
		
		var state = this.CurrentState(this.newAllEvents.map(event => event).sort((a, b) => a.start.unix() - b.start.unix()));
		
		if (forceUpdate || this.ViewEventsChanged(state.events)) {
			this.displayedEvents = state.events;
			var ungroupedEvents = state.ungroupedEvents;
			var groupedEvents = state.groupedEvents;
			
			this.StopPageTransition();
			
			this.pageIndex = 0;
			this.pageCount = 0;
			
			this.DefaultActivityDetection();
			
			var titleWidth = 0;
			var nameWidth = 0;
			var roomWidth = 0;
			var startDayWidth = 0;
			var endDayWidth = 0;
			var startDayTimeWidth = 0;
			var endDayTimeWidth = 0;
			var startDateWidth = 0;
			var endDateWidth = 0;
			var startTimeWidth = 0;
			var endTimeWidth = 0;
			var hyphenWidth = 0;
			var smartFormattingDayParts = {};
			
			var minimumDay = null;
			var maximumDay = null;
			for (const event of this.displayedEvents) {
				if (minimumDay == null || minimumDay.isAfter(event.start.clone().startOf('day'))) {
					minimumDay = event.start.clone().startOf('day');
				}
				if (maximumDay == null || maximumDay.isBefore(event.end.clone().startOf('day'))) {
					maximumDay = event.end.clone().startOf('day');
				}
			}
			const currentTime = this.player.currentTime;
			var showStartDay = minimumDay && maximumDay ? !minimumDay.isSame(currentTime.clone().startOf('day')) || !minimumDay.isSame(maximumDay) : false;
			var showStartYear = minimumDay && maximumDay ? minimumDay.year() != currentTime.clone().year() || minimumDay.year() != maximumDay.year() : false;
			var showEndDay = minimumDay && maximumDay ? !minimumDay.isSame(maximumDay) : false;
			var showEndYear = minimumDay && maximumDay ? minimumDay.year() != maximumDay.year() : false;
			
			var dynamicDateWithoutYear =
			{
				'MMM D YYYY': 'MMM D',
				'MMM D, YYYY': 'MMM D',
				'MMMM D YYYY': 'MMMM D',
				'MMMM D, YYYY': 'MMMM D'
			};
			
			var startDateFormat = this.eventlistdateformat;
			var endDateFormat = this.eventlistdateformat;
			if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
				if (!showStartDay) {
					startDateFormat = '';
				} else if (!showStartYear) {
					startDateFormat = dynamicDateWithoutYear[startDateFormat];
				}
				if (!showEndDay) {
					endDateFormat = '';
				} else if (!showEndYear) {
					endDateFormat = dynamicDateWithoutYear[endDateFormat];
				}
			}
			
			this.$pages.hidden = true;
			switch (this.eventlistoverflowbehavior.slice(0, -2)) {
				case 'pagified':
				case 'pagifiedfade':
				case 'pagifiedslideleft':
				case 'pagifiedslideup':
					this.$pages.innerHTML = '';
					this.$pages.hidden = false;
					var $testWidth = $.div({ css: {
						whiteSpace: 'pre',
						fontFamily: `'${this.eventlistfont}'`,
						fontSize: `${this.eventlistfontsize}px`
					} });
					
					var pageTag = null;
					
					var rowOdd = true;
					
					const newPage = () => {
						if (pageTag != null && pageTag.offsetHeight > 0) {
							pageTag.style.position = 'static';
							this.$pages.appendChild(pageTag);
							this.pageCount++;
						}
						pageTag = this.$pages.appendChild($.div({ class: 'page', css: {
							position: 'absolute',
							width: `${this.width}px`
						} }));
						rowOdd = true;
					};
					newPage();
					
					var maxPageHeight = this.$pages.offsetHeight;
					const addEvent = async (event) => {
						var eventTag = $.ul({ class: 'event' });
						if (this.groupby != 'none') {
							eventTag.appendChild(
								$.li({ class: 'spacer' })
							);
						}
						const addColumn = async (type) => {
							switch (type) {
								case 'room':
									$testWidth.innerText = event.room;
									roomWidth = Math.max(roomWidth, (await Measure($testWidth)).width);
									eventTag.appendChild(
										$.li({ text: event.room, class: 'room' })
									);
									break;
								case 'title':
									if (event.title.indexOf('https://') == 0) {
										titleWidth = Math.max(titleWidth, this.eventlistfontsize);
										eventTag.appendChild(
											$.li({ class: 'image' },
												$.img({ src: event.title, css: { height: this.eventlistfontsize } })
											)
										);
									} else {
										$testWidth.innerText = event.title;
										titleWidth = Math.max(titleWidth, (await Measure($testWidth)).width);
										eventTag.appendChild(
											$.li({ text: event.title, class: 'title' })
										);
									}
									break;
								case 'name':
									if (event.name.indexOf('https://') == 0) {
										nameWidth = Math.max(nameWidth, this.eventlistfontsize);
										eventTag.appendChild(
											$.li({ class: 'image' },
												$.img({ src: event.name, css: { height: this.eventlistfontsize } })
											)
										);
									} else {
										$testWidth.innerText = event.name;
										nameWidth = Math.max(nameWidth, (await Measure($testWidth)).width);
										eventTag.appendChild(
											$.li({ text: event.name, class: 'name' })
										);
									}
									break;
								case 'time':
									// Dynamic 4 formats:
									// 1 PM - 2 PM
									// Mar 5 1 PM - 2 PM
									// Mar 5 1 PM - Mar 20 2 PM
									// Mar 5 2018 1 PM - Mar 20 2019 2 PM
									if (showStartDay) {
										if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
											if (startDateFormat != '') {
												for (const datePart of startDateFormat.split(' ')) {
													$testWidth.innerText = event.start.format(datePart);
													if (!smartFormattingDayParts.hasOwnProperty('startday' + datePart)) {
														smartFormattingDayParts['startday' + datePart] = 0;
													}
													smartFormattingDayParts['startday' + datePart] = Math.max(smartFormattingDayParts[`startday${datePart}`], (await Measure($testWidth)).width);
													eventTag.appendChild($.li({ text: event.start.format(datePart), class: `startday${datePart}` }));
												}
											}
										} else {
											$testWidth.innerText = event.start.format(startDateFormat);
											startDayWidth = Math.max(startDayWidth, (await Measure($testWidth)).width);
											eventTag.appendChild($.li({ text: event.start.format(startDateFormat), class: 'startday' }));
										}
									}
									$testWidth.innerText = event.start.format(this.eventlisttimeformat);
									startDayTimeWidth = Math.max(startDayTimeWidth, (await Measure($testWidth)).width);
									$testWidth.innerText = '-';
									hyphenWidth = Math.max(hyphenWidth, (await Measure($testWidth)).width);
									eventTag.append(
										$.li({ text: event.start.format(this.eventlisttimeformat), class: 'startdaytime' }),
										$.li({ text: '-', class: 'hyphen' })
									);
									if (showEndDay) {
										if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
											if (endDateFormat != '') {
												for (const datePart of endDateFormat.split(' ')) {
													$testWidth.innerText = event.end.format(datePart);
													if (!smartFormattingDayParts.hasOwnProperty('endday' + datePart)) {
														smartFormattingDayParts['endday' + datePart] = 0;
													}
													smartFormattingDayParts['endday' + datePart] = Math.max(smartFormattingDayParts[`endday${datePart}`], (await Measure($testWidth)).width);
													eventTag.appendChild($.li({ text: event.end.format(datePart), class: `endday${datePart}` }));
												}
											}
										} else {
											$testWidth.innerText = event.end.format(endDateFormat);
											endDayWidth = Math.max(endDayWidth, (await Measure($testWidth)).width);
											eventTag.appendChild($.li({ text: event.end.format(endDateFormat), class: 'endday' }));
										}
									}
									$testWidth.innerText = event.end.format(this.eventlisttimeformat);
									endDayTimeWidth = Math.max(endDayTimeWidth, (await Measure($testWidth)).width);
									eventTag.appendChild($.li({ text: event.end.format(this.eventlisttimeformat), class: 'enddaytime' }));
									break;
								case 'startdate':
									if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
										if (startDateFormat != '') {
											for (const datePart of startDateFormat.split(' ')) {
												$testWidth.innerText = event.start.format(datePart);
												if (!smartFormattingDayParts.hasOwnProperty('startday' + datePart)) {
													smartFormattingDayParts['startday' + datePart] = 0;
												}
												smartFormattingDayParts['startday' + datePart] = Math.max(smartFormattingDayParts[`startday${datePart}`], (await Measure($testWidth)).width);
												eventTag.appendChild($.li({ text: event.start.format(datePart), class: `startday${datePart}` }));
											}
										}
									} else {
										$testWidth.innerText = event.start.format(this.eventlistdateformat);
										startDateWidth = Math.max(startDateWidth, (await Measure($testWidth)).width);
										eventTag.appendChild($.li({ text: event.start.format(this.eventlistdateformat), class: 'startdate' }));
									}
									break;
								case 'enddate':
									if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
										if (endDateFormat != '') {
											for (const datePart of endDateFormat.split(' ')) {
												$testWidth.innerText = event.end.format(datePart);
												if (!smartFormattingDayParts.hasOwnProperty('endday' + datePart)) {
													smartFormattingDayParts['endday' + datePart] = 0;
												}
												smartFormattingDayParts['endday' + datePart] = Math.max(smartFormattingDayParts[`endday${datePart}`], (await Measure($testWidth)).width);
												eventTag.appendChild($.li({ text: event.end.format(datePart), class: `endday${datePart}` }));
											}
										}
									} else {
										$testWidth.innerText = event.end.format(this.eventlistdateformat);
										endDateWidth = Math.max(endDateWidth, (await Measure($testWidth)).width);
										eventTag.appendChild($.li({ text: event.end.format(this.eventlistdateformat), class: 'enddate' }));
									}
									break;
								case 'starttime':
									$testWidth.innerText = event.start.format(this.eventlisttimeformat);
									startTimeWidth = Math.max(startTimeWidth, (await Measure($testWidth)).width);
									eventTag.appendChild($.li({ text: event.start.format(this.eventlisttimeformat), class: 'starttime' }));
									break;
								case 'endtime':
									$testWidth.innerText = event.end.format(this.eventlisttimeformat);
									endTimeWidth = Math.max(endTimeWidth, (await Measure($testWidth)).width);
									eventTag.appendChild($.li({ text: event.end.format(this.eventlisttimeformat), class: 'endtime' }));
									break;
								case 'hyphen':
									$testWidth.innerText = '-';
									hyphenWidth = Math.max(hyphenWidth, (await Measure($testWidth)).width);
									eventTag.appendChild($.li({ text: '-', class: 'hyphen' }));
									break;
							}
						};
						await Promise.all([
							addColumn(this.hasOwnProperty('eventlistcolumnfirst') ? this.eventlistcolumnfirst : this.eventlistcolumn1),
							addColumn(this.hasOwnProperty('eventlistcolumnsecondary') ? this.eventlistcolumnsecondary : this.eventlistcolumn2),
							addColumn(this.hasOwnProperty('eventlistcolumntertiary') ? this.eventlistcolumntertiary : this.eventlistcolumn3),
							addColumn(this.hasOwnProperty('eventlistcolumnquaternary') ? this.eventlistcolumnquaternary : this.eventlistcolumn4),
							addColumn(this.eventlistcolumn5),
							addColumn(this.eventlistcolumn6),
							addColumn(this.eventlistcolumn7),
							addColumn(this.eventlistcolumn8)
						]);
						
						eventTag.style.backgroundColor = rowOdd ? this.eventlistoddrowbackgroundcolor : this.eventlistevenrowbackgroundcolor;
						
						eventsTag.appendChild(
							$.li({ css: {
								fontFamily: `'${this.eventlistfont}'`,
								fontSize: `${this.eventlistfontsize}px`,
								color: this.eventlistcolor,
								lineHeight: `${this.eventlistlineheight == -1 ? this.eventlistfontsize + 8 : this.eventlistlineheight}px`
							} },
								eventTag
							)
						);
						rowOdd = !rowOdd;
					};
					
					if (this.groupby == 'none') {
						var usedEvents = [];
						
						const newPageUngroupedEvents = () => {
							return pageTag.appendChild($.ul());
						};
						
						var eventsTag = newPageUngroupedEvents();
						var iterationMax = 0; // If the height of the widget is too small it might loop forever. This limits it.
						while (ungroupedEvents.length != 0 && iterationMax < 100) {
							iterationMax++;
							var event = ungroupedEvents.shift();
							usedEvents.push(event);
							await addEvent(event);
							if (pageTag.offsetHeight > maxPageHeight) {
								eventsTag.querySelector(':scope > li:last-child').remove();
								ungroupedEvents.unshift(usedEvents.pop());
								newPage();
								eventsTag = newPageUngroupedEvents();
							}
						}
						newPage();
						pageTag.remove();
					} else {
						var groupTag = null;
						
						const newGroup = () => {
							var eventsTag = $.ul();
							groupTag = pageTag.appendChild($.div({},
								$.div({ text: group.group, class: 'group', css: {
									fontFamily: `'${this.groupfont}'`,
									fontSize: `${this.groupfontsize}px`,
									color: this.groupcolor,
									backgroundColor: this.groupbackgroundcolor,
									lineHeight: `${this.groupfontsize + 8}px`
								} }),
								eventsTag
							));
							return eventsTag;
						};
						
						// Process the groups until all of them are on a page
						var iterationMax = 0; // This is a sanity check because it's possible to make the height of the widget like 10 and thus no groups will ever fit
						while (groupedEvents.length != 0 && iterationMax < 100) {
							iterationMax++;
							var group = groupedEvents.shift();
							// If there are no events and show group when no events is disabled then skip this group
							if (!this.groupshownoevents && group.events.length == 0) {
								continue;
							}
							var eventsTag = newGroup();
							// Check if the height is too small to fit even the group header. If so remove it and start on the next page
							if (pageTag.offsetHeight > maxPageHeight) {
								groupTag.remove();
								groupedEvents.unshift(group);
								newPage();
								continue;
							}
							
							// Calculate the total height if all events are added to a group to see if they could even fit on a single page
							for (const event of group.events) {
								await addEvent(event);
							}
							var groupAllEventsFitsOnOnePage = groupTag.offsetHeight < maxPageHeight;
							groupTag.remove();
							eventsTag = newGroup();
							
							if (group.events.length == 0) {
								eventsTag.appendChild(
									$.li({ text: this.groupnoeventstext, class: 'noevents', css: {
										fontFamily: `'${this.eventlistfont}'`,
										fontSize: `${this.eventlistfontsize}px`,
										color: this.eventlistcolor,
										lineHeight: `${this.eventlistlineheight == -1 ? this.eventlistfontsize + 8 : this.eventlistlineheight}px`,
										backgroundColor: this.eventlistoddrowbackgroundcolor
									} })
								);
							} else {
								// Try to fit as many events into the group that will fit on the rest of the page
								var usedEvents = [];
								while (group.events.length != 0) {
									var event = group.events.shift();
									usedEvents.push(event);
									await addEvent(event);
									if (pageTag.offsetHeight > maxPageHeight) {
										eventsTag.querySelector(':scope > li:last-child').remove();
										group.events.unshift(usedEvents.pop());
										break;
									}
								}
								
								if (group.events.length > 0) {
									const addContinued = () => {
										eventsTag.append(
											$.li({ text: '...Continued', class: 'continued', css: {
												fontFamily: `'${this.groupfont}'`,
												fontSize: `${this.groupfontsize}px`,
												color: this.groupcolor,
												lineHeight: `${this.groupfontsize + 8}px`
											} })
										);
									};
									addContinued();
									
									if (groupAllEventsFitsOnOnePage && groupTag.offsetHeight / maxPageHeight < 0.30) {
										// The group is less than 30% of the page and the whole group could fit on its own page so move it to the next page so it's not split between two pages
										groupTag.remove();
										while (usedEvents.length != 0)
										{
											group.events.unshift(usedEvents.pop());
										}
									} else {
										while (pageTag.offsetHeight > maxPageHeight) { // Does the continued line fit?
											if (usedEvents.length > 1) {
												group.events.unshift(usedEvents.pop());
												for (const $e of eventsTag.querySelectorAll(':scope > li:nth-last-child(2)')) $e.remove();
											} else {
												// The continue doesn't fit and there's only one event shown of multiple
												groupTag.remove();
												if (usedEvents.length == 1) {
													group.events.unshift(usedEvents.pop());
												}
												break;
											}
										}
									}
									// Start a new page with the same group
									groupedEvents.unshift(group);
									newPage();
								}
							}
						}
						newPage();
						pageTag.remove();
					}
					
					this.StartPageTransition();
					
					var pointerdown = false;
					this.startOffset = 0;
					this.pageOffsetDifference = 0;
					this.pageOffsetDifferenceRequestAnimationFrame = null;
					
					if (this.tag.hasOwnProperty('lastPointerDown')) {
						this.tag.removeEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastPointerDown);
					}
					if (this.tag.hasOwnProperty('lastPointerMove')) {
						this.tag.removeEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastPointerMove);
					}
					if (this.tag.hasOwnProperty('lastPointerUp')) {
						this.tag.removeEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastPointerUp);
					}
					this.tag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastPointerDown = e => {
						if (!('onpointerdown' in window)) {
							e = e.touches[0];
						}
						this.StopPageTransition();
						pointerdown = true;
						switch (this.eventlistoverflowbehavior.slice(0, -2)) {
							case 'pagified':
							case 'pagifiedfade':
								// Do nothing
								break;
							case 'pagifiedslideleft':
								this.startOffset = e.pageX;
								this.StartPageTransition();
								cancelAnimationFrame(this.pageOffsetDifferenceRequestAnimationFrame);
								this.pageOffsetDifferenceRequestAnimationFrame = null;
								break;
							case 'pagifiedslideup':
								this.startOffset = e.pageY;
								this.StartPageTransition();
								cancelAnimationFrame(this.pageOffsetDifferenceRequestAnimationFrame);
								this.pageOffsetDifferenceRequestAnimationFrame = null;
								break;
						}
					});
					this.tag.addEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastPointerMove = e => {
						if (!('onpointerdown' in window)) {
							e = e.touches[0];
						}
						if (pointerdown) {
							switch (this.eventlistoverflowbehavior.slice(0, -2)) {
								case 'pagifiedslideleft':
									this.StartPageTransition();
									this.pageOffsetDifference += e.pageX - this.startOffset;
									if (this.pageOffsetDifference > width) {
										this.pageOffsetDifference = width;
									}
									if (this.pageOffsetDifference < -width) {
										this.pageOffsetDifference = -width;
									}
									this.startOffset = e.pageX;
									break;
								case 'pagifiedslideup':
									this.StartPageTransition();
									this.pageOffsetDifference += e.pageY - this.startOffset;
									if (this.pageOffsetDifference > this.$pages.offsetHeight) {
										this.pageOffsetDifference = this.$pages.offsetHeight;
									}
									if (this.pageOffsetDifference < -this.$pages.offsetHeight) {
										this.pageOffsetDifference = -this.$pages.offsetHeight;
									}
									this.startOffset = e.pageY;
									break;
							}
							this.updatePositions();
						}
					});
					this.tag.addEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastPointerUp = e => {
						pointerdown = false;
						// If fade is used then tapping will advance to the next page
						switch (this.eventlistoverflowbehavior.slice(0, -2)) {
							case 'pagified':
							case 'pagifiedfade':
								this.IncrementPage();
								break;
							case 'pagifiedslideleft':
								if (this.pageOffsetDifference < -100) {
									if (this.pageIndex < this.pageCount - 1) {
										this.pageOffsetDifference += width;
										this.pageIndex++;
										this.slideTransitionWeight = 1;
									}
								} else if (this.pageOffsetDifference > 100) {
									if (this.pageIndex > 0) {
										this.pageOffsetDifference -= width;
										this.pageIndex--;
										this.slideTransitionWeight = 1;
									}
								}
								this.AnimateTransition();
								break;
							case 'pagifiedslideup':
								if (this.pageOffsetDifference < -100) {
									if (this.pageIndex < this.pageCount - 1) {
										this.pageOffsetDifference += this.$pages.offsetHeight;
										this.pageIndex++;
										this.slideTransitionWeight = 1;
									}
								} else if (this.pageOffsetDifference > 100) {
									if (this.pageIndex > 0) {
										this.pageOffsetDifference -= this.$pages.offsetHeight;
										this.pageIndex--;
										this.slideTransitionWeight = 1;
									}
								}
								this.AnimateTransition();
								break;
						}
					});
					break;
				case 'scroll':
					// TODO: Implement such that it's a continuous scroll and allows manual scrolling from a user
					break;
			}
			
			roomWidth += 30;
			titleWidth += 30;
			nameWidth += 30;
			hyphenWidth += 10;
			
			const sizeColumn = type => {
				for (const $room of this.tag.querySelectorAll('.event > li.room')) $room.style.width = `${roomWidth}px`;
				for (const $name of this.tag.querySelectorAll('.event > li.name')) $name.style.width = `${nameWidth}px`;
				
				if (this.eventlistdatesmartformatting && dynamicDateWithoutYear.hasOwnProperty(this.eventlistdateformat)) {
					if (startDateFormat != '') {
						for (const datePart of startDateFormat.split(' ')) {
							for (const $e of this.tag.querySelectorAll(`.event > li.startday${datePart}`)) $e.style.width = `${smartFormattingDayParts[`startday${datePart}`]}px`;
						}
					}
					if (endDateFormat != '') {
						for (const datePart of endDateFormat.split(' ')) {
							for (const $e of this.tag.querySelectorAll(`.event > li.endday${datePart}`)) $e.style.width = `${smartFormattingDayParts[`endday${datePart}`]}px`;
						}
					}
				}
				for (const $e of this.tag.querySelectorAll('.event > li.startday')) $e.style.width = `${startDayWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.startdaytime')) $e.style.width = `${startDayTimeWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.endday')) $e.style.width = `${endDayWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.enddaytime')) $e.style.width = `${endDayTimeWidth}px`;
				
				for (const $e of this.tag.querySelectorAll('.event > li.startdate')) $e.style.width = `${startDateWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.enddate')) $e.style.width = `${endDateWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.starttime')) $e.style.width = `${startTimeWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.endtime')) $e.style.width = `${endTimeWidth}px`;
				for (const $e of this.tag.querySelectorAll('.event > li.hyphen')) $e.style.width = `${hyphenWidth}px`;
			};
			sizeColumn(this.hasOwnProperty('eventlistcolumnfirst') ? this.eventlistcolumnfirst : this.eventlistcolumn1);
			sizeColumn(this.hasOwnProperty('eventlistcolumnsecondary') ? this.eventlistcolumnsecondary : this.eventlistcolumn2);
			sizeColumn(this.hasOwnProperty('eventlistcolumntertiary') ? this.eventlistcolumntertiary : this.eventlistcolumn3);
			sizeColumn(this.hasOwnProperty('eventlistcolumnquaternary') ? this.eventlistcolumnquaternary : this.eventlistcolumn4);
			sizeColumn(this.eventlistcolumn5);
			sizeColumn(this.eventlistcolumn6);
			sizeColumn(this.eventlistcolumn7);
			sizeColumn(this.eventlistcolumn8);
			
			for (const $e of this.tag.querySelectorAll('.pages > div, .pages > ul')) $e.style.height = `${this.$pages.offsetHeight}px`;
			
			if (this.absolutePositionTransition) {
				for (const $e of this.tag.querySelectorAll('.pages > div, .pages > ul')) {
					Object.assign($e.style, {
						position: 'absolute',
						left: 0,
						top: 0
					});
				}
			}
			this.updatePositions();
			this.$noEventImages.innerHTML = '';
			if (this.displayedEvents.length == 0 && this.eventlistusenoeventsimages) {
				this.$noEventImages.append(
					...(this.eventlistusenoeventsimages ? this.eventlistnoeventimages : []).map((image, index) => {
						return $.div({ css: { backgroundImage: `url('${image.cdnurl}')`, backgroundSize: this.eventlistnoeventimagescoverfill ? 'cover' : '100% 100%', zIndex: -(index + 1) } });
					})
				);
				clearInterval(this.noEventImageInterval);
				this.noEventImageInterval = setInterval(async () => {
					if (this.TagDeleted()) {
						return;
					}
					if (this.$noEventImages.children.length > 1) {
						await $.fadeOut($event);
						this.$noEventImages.appendChild($event);
						$.fadeIn($event);
						$event.fadeAnimation.finish();
					}
				}, 10000);
			}
		}
	}
	async LoadEvents() {
		if (this.offlineMode === false) {
			let cacheData = await new Promise(async resolve => {
				try {
					const data = await $.http({
						method: 'GET',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/providerevents.php',
						data: {
							widget: 'readerboard',
							mediaid: this.mediaid,
							roomids: JSON.stringify(this.groupids)
						},
						dataType: 'json'
					}).promise;
					if (data.success) {
						resolve({ events: data.events, groups: data.rooms });
					} else {
						console.log('Error getting provider events.');
						resolve(null);
					}
				} catch (exception) {
					console.log('Readerboard LoadEvents providerEvents error');
					this.offlineMode = null;
					this.TestOnlineMode();
					resolve(null);
				}
			});
			if (this.TagDeleted()) {
				return;
			}
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			
			Object.assign(this, cacheData);
			
			this.newAllEvents = this.events.map(event => {
				if (event.hasOwnProperty('allday') && event.allday) {
					event.start = moment.utc(event.start).local().startOf('day');
					event.end = moment.utc(event.start).local().endOf('day');
				} else {
					event.start = moment(event.start);
					event.end = moment(event.end);
				}
				return event;
			}).sort((a, b) => a.start.unix() - b.start.unix());
			
			this.ResetView();
		}
	}
	StartResetTimer() {
		clearTimeout(this.resetTimeout);
		this.resetTimeout = setTimeout(this.ResetView.bind(this), 30 * 1000);
	}
	RemoveActivityDetection() {
		if (this.tag.hasOwnProperty('lastActivityDetection')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastActivityDetection);
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerup' : 'touchmove', this.tag.lastActivityDetection);
			this.tag.removeEventListener('onpointerdown' in window ? 'pointermove' : 'touchend', this.tag.lastActivityDetection);
		}
	}
	DefaultActivityDetection() {
		this.RemoveActivityDetection();
		this.tag.lastActivityDetection = this.StartResetTimer.bind(this);
		this.tag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastActivityDetection);
		this.tag.addEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastActivityDetection);
	}
	// This will test the connection and get the latest information for the widget. If it fails then it'll go into offline mode and then try to reconnect in 5 minutes
	async TestOnlineMode() {
		if (this.TagDeleted()) {
			return;
		}
		
		let cacheData = await new Promise(async resolve => {
			try {
				await $.http({
					method: 'GET',
					url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/testonline.php',
					data: {
						random: Math.random()
					},
					dataType: 'json'
				}).promise;
				
				if (this.offlineMode == null || this.offlineMode) {
					this.offlineMode = false;
					if (this.container) {
						// TODO: Data driven widgetData
						var imagesToCache =
						[
							this.usedefaultbackgroundimage ? this.GetOriginalAttributeValue('defaultbackgroundimage') : null,
							this.overridebackgroundimage ? this.widgetData.readerboard.backgroundimage : null,
							...(this.eventlistusenoeventsimages ? this.eventlistnoeventimages : [])
						];
						const imageCacheURLs = await Promise.all(imagesToCache.map(image => this.container.cacheMediaObject(image)));
						if (imageCacheURLs[0]) { this.defaultbackgroundimage = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('defaultbackgroundimage'))); this.defaultbackgroundimage.cdnurl = imageCacheURLs[0]; }
						if (imageCacheURLs[1]) { this.backgroundimage = JSON.parse(JSON.stringify(this.widgetData.readerboard.backgroundimage)); this.backgroundimage.cdnurl = imageCacheURLs[1]; }
						let originaleventlistnoeventimages = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('eventlistnoeventimages')));
						for (let i = 2; i < imageCacheURLs.length; ++i) {
							if (imageCacheURLs[i]) { originaleventlistnoeventimages[i - 2].cdnurl = imageCacheURLs[i]; }
						}
						if (imageCacheURLs.length > 2) {
							this.eventlistnoeventimages = originaleventlistnoeventimages;
						}
					}
					resolve(
					{
						defaultbackgroundimage: this.defaultbackgroundimage,
						backgroundimage: this.backgroundimage,
						eventlistnoeventimages: this.eventlistnoeventimages
					});
				} else {
					resolve({});
				}
			} catch (exception) {
				// Offline mode
				if (this.offlineMode == null || !this.offlineMode) {
					console.log('Readerboard going into offline mode');
					this.offlineMode = true;
				}
				// Try again in 5 minutes
				clearTimeout(this.onlineModeTimer);
				this.onlineModeTimer = setTimeout(this.TestOnlineMode.bind(this), 5 * 60 * 1000);
				
				resolve({});
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		
		// Cache everything that isn't an image in this.readerboard
		if (this.widgetData.hasOwnProperty('readerboard')) {
			Object.assign(cacheData,
			{
				name: this.name,
				groupids: this.groupids,
				overridebackgroundimage: this.overridebackgroundimage,
				backgroundimagecoverfill: this.backgroundimagecoverfill
			});
		}
		
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		
		Object.assign(this, cacheData);
		
		if (!this.container || this.container.type == 'Preview') {
			this.name = 'Sample Name';
			this.groups = [{ name: 'Sample Room 1', replacement: '' }];
			this.overridebackgroundimage = false;
			this.backgroundimage = null;
			this.events = [];
			var allGroups = this.groups.map(room => room.name);
			var uniqueId = 0;
			const currentTime = this.player.currentTime;
			for (var roomItr = 0; roomItr < 2; ++roomItr) {
				for (var dayItr = 0; dayItr < 5; ++dayItr) {
					for (var hourItr = 0; hourItr < 11; ++hourItr) {
						this.events.push({
							start: currentTime.clone().add(dayItr, 'day').add(hourItr * 2 + roomItr, 'hour').minutes(0).valueOf(),
							end: currentTime.clone().add(dayItr, 'day').add(hourItr * 2 + roomItr, 'hour').add(1, 'hour').minutes(0).valueOf(),
							room: allGroups.length == 0 ? 'Room Sample ' + uniqueId : allGroups[roomItr % allGroups.length],
							title: 'Sample Title ' + (uniqueId++),
							name: 'John Smith'
						});
					}
				}
			}
		}
		
		this.newAllEvents = this.events.map(event => {
			if (event.hasOwnProperty('allday') && event.allday) {
				event.start = moment.utc(event.start).local().startOf('day');
				event.end = moment.utc(event.start).local().endOf('day');
			} else {
				event.start = moment(event.start);
				event.end = moment(event.end);
			}
			return event;
		}).sort((a, b) => a.start.unix() - b.start.unix());
		
		// Clear all timers
		this.ClearTimers();
		
		this.tag.innerHTML = '';
		this.tag.classList.add('widget-readerboard');
		Object.assign(this.tag.style, {
			backgroundColor: this.background
		});
		this.tag.append(
			this.$backgroundImage = $.div({ class: 'backgroundimage' }),
			this.$noEventImages = $.div({ class: 'noeventimages' }),
			$.div({ class: 'name', text: this.name, css: {
				fontFamily: `'${this.namefont}'`,
				fontSize: `${this.namefontsize}px`, 
				color: this.namecolor,
				textAlign: this.namealignment,
				lineHeight: `${this.namefontsize + 8}px`
			}, hidden: !this.showname }),
			this.$pages = $.div({ class: 'pages' })
		);
		
		if (this.overridebackgroundimage && this.backgroundimage) {
			this.$backgroundImage.style.backgroundImage = `url('${this.backgroundimage.cdnurl}')`;
			this.$backgroundImage.style.backgroundSize = this.backgroundimagecoverfill ? 'cover' : '100% 100%';
		} else if (this.usedefaultbackgroundimage && this.defaultbackgroundimage) {
			this.$backgroundImage.style.backgroundImage = `url('${this.defaultbackgroundimage.cdnurl}')`;
			this.$backgroundImage.style.backgroundSize = this.defaultbackgroundimagecoverfill ? 'cover' : '100% 100%';
		}
		
		this.delay = parseInt(this.eventlistoverflowbehavior.slice(-2), 10);
		switch (this.eventlistoverflowbehavior.slice(0, -2)) {
			case 'pagified':
			case 'pagifiedfade':
				this.absolutePositionTransition = true;
				this.pageTransitionSize = 1000;
				break;
			case 'pagifiedslideleft':
				this.$pages.style.flexFlow = 'row nowrap';
				this.pageTransitionSize = this.$pages.offsetWidth;
				break;
			case 'pagifiedslideup':
				this.$pages.flexFlow = 'column nowrap';
				this.pageTransitionSize = this.$pages.offsetHeight;
				break;
		}
		
		var fps = 0;
		this.fpsInterval = setInterval(() => {
			if (this.TagDeleted()) {
				return;
			}
			this.currentFPS = fps;
			fps = 0;
		}, 1000);
		
		const fpsIncrement = () => {
			if (this.TagDeleted()) {
				return;
			}
			fps++;
			this.fpsRequestAnimationFrame = requestAnimationFrame(fpsIncrement);
		};
		this.fpsRequestAnimationFrame = requestAnimationFrame(fpsIncrement);
		
		if (this.widgetData.hasOwnProperty('readerboard')) {
			this.loadEventsInterval = setInterval(this.LoadEvents.bind(this), 5 * 60 * 1000);
			this.LoadEvents();
		}
		
		this.ResetView(true);
		
		this.ready();
	}
	Destroy() {
		this.ClearTimers();
	}
	TimeSync() {
		if (Math.abs(this.oldServerOffset - this.player.serverOffset) > 1000) {
			this.oldServerOffset = this.player.serverOffset;
			this.ResetView(true);
		}
	}
}
