class Widgetrss extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.rssInterval = null;
		
		this.tag.classList.add('widget-rss');
		this.tag.style.backgroundColor = this.background;
		
		this.stopPlayingPromise = null;
		this.stopPlaying = null;
		
		this.computedFilter = this.filter;
		
		this.computedStart = this.start;
		if (this.computedStart != '') {
			this.computedStart = moment(this.computedStart, 'MM/DD/YYYY');
		} else {
			this.computedFilter = false;
		}
		this.computedEnd = this.end;
		if (this.computedEnd != '') {
			this.computedEnd = moment(this.computedEnd, 'MM/DD/YYYY');
		} else {
			this.computedFilter = false;
		}
		
		this.rssInterval = setInterval(this.processRssData.bind(this), 10 * 60 * 1000); // 10 minutes
		this.processRssData();
	}
	async cycle() {
		if (this.TagDeleted()) {
			return;
		}
		try {
			const $item = this.tag.querySelector('.item');
			if ($item == null) {
				return;
			}
			await Promise.race([this.stopPlayingPromise, $.fadeIn($item)]);
			const $p = $item.querySelector(':scope > p');
			if (this.resizefontsize) {
				while (parseInt($p.style.fontSize, 10) > 12 && $p.scrollHeight > $p.clientHeight) {
					$p.style.fontSize = `${parseInt($p.style.fontSize, 10) - 1}px`;
				}
			}
			if ($p.scrollHeight > $p.clientHeight) {
				$p.classList.add('overflow');
			}
			await Promise.race([this.stopPlayingPromise, new Promise(resolve => setTimeout(resolve, 15000))]);
			await Promise.race([this.stopPlayingPromise, $.fadeOut($item)]);
			this.tag.appendChild($item);
			this.cycle();
		} catch (exception) {
			// this.stopPlayingPromise rejected
		}
	}
	async processRssData() {
		if (this.TagDeleted()) {
			return;
		}
		var cacheData = await new Promise(async resolve => {
			var url = '';
			var data = {};
			if (this.rssfeed == 'useurl') {
				url = `${this.container ? this.container.hostUrl.replace(/\/$/, '') : ''}/get/rssdata.php`;
				data = {
					t: Date.now(),
					url: this.url
				};
			} else {
				url = `${(this.container ? this.container.hostUrl.replace(/\/$/, '') : '')}/rss/${this.account}/${this.rssfeed.replace('.rss', '')}.rss?t=${Date.now()}`;
			}
			try {
				const $xml = await $.http({
					method: 'GET',
					url,
					data,
					dataType: 'xml'
				}).promise;
				var items = [];
				for (const $item of $xml.querySelectorAll('item')) {
					if ($item.querySelector('pubDate') && $item.querySelector('title') && $item.querySelector('description')) {
						var date = moment($item.querySelector('pubDate').textContent, 'ddd, DD MMM YYYY HH:mm:ss Z');
						if (this.rssfeed == 'useurl' || !this.computedFilter || date.isSameOrAfter(this.computedStart) && date.isSameOrBefore(this.computedEnd.endOf('day'))) {
							var item = {
								title: $item.querySelector('title').textContent,
								date: $item.querySelector('pubDate').textContent,
								description: $item.querySelector('description').textContent,
								images: []
							};
							if (this.imagelayout != 'default') {
								item.description = $.span({ html: item.description });
								for (const $img of item.description.querySelectorAll('img')) {
									item.images.push($img.src);
									$img.remove();
								}
								item.description = item.description.innerHTML;
							}
							for (const $enclosure of $item.querySelectorAll('enclosure')) {
								if (['image/gif', 'image/png', 'image/jpeg', 'image/bmp'].indexOf($enclosure.getAttribute('type')) != -1) {
									item.images.push($enclosure.getAttribute('url'));
								}
							}
							for (const $content of $item.querySelectorAll('content')) {
								if ($content.hasAttribute('url') && $content.hasAttribute('medium') && $content.getAttribute('medium') == 'image') {
									item.images.push($content.getAttribute('url'));
								}
							}
							if (items.length < 200) {
								items.push(item);
							}
						}
					}
				}
				resolve({ items: items });
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('items')) {
			cacheData.items = [];
		}
		
		const imagesLoaded = [];
		this.tag.innerHTML = '';
		this.tag.append(...cacheData.items.map(item => {
			const $item = $.div({ class: 'item' });
			// Use a promise here to delay the start of the animation so all the description heights are calculated
			imagesLoaded.push(new Promise(resolve => {
				$item.appendChild($.h2({ html: item.title }));
				if (this.showdate) {
					$item.appendChild($.p({ text: moment(item.date, 'ddd, DD MMM YYYY HH:mm:ss Z').format('M/D/YY') }));
				}
				$item.appendChild($.p({ html: item.description }));
				if (this.imagelayout != 'default' || this.imagelayout == 'default' && item.images.length != 0) {
					// Only append the largest image. A lot of RSS feeds have social media images and links so this removes those
					Promise.all(item.images.map(image => {
						return new Promise(resolve => {
							$.img({ src: image, load: function() {
								resolve(this);
							}, error: function() {
								if (image.indexOf('http://') == 0) {
									// Try https
									image.replace('http://', 'https://');
									$.img({ src: image, load: function() {
										resolve(this);
									}, error: function() {
										resolve(this);
									} });
								} else {
									resolve(this);
								}
							} });
						});
					})).then(imageTags => {
						var largestImageDimension = 150;
						var $largestImage = null;
						for (const imageTag of imageTags) {
							if (Math.max(imageTag.width, imageTag.height) > largestImageDimension) {
								largestImageDimension = Math.max(imageTag.width, imageTag.height);
								$largestImage = imageTag;
							}
						}
						if ($largestImage != null) {
							$largestImage.classList.add('rss-image');
							if (this.imagelayout == 'bottom' || this.imagelayout == 'default' && item.images.length != 0) {
								$largestImage.style.minHeight = `${this.imageminimumheight}px`;
								$item.appendChild($largestImage)
							} else {
								$item.querySelector(':scope > p').prepend($largestImage);
								Object.assign($largestImage.style, {
									float: 'right',
									width: '60%'
								});
							}
						}
						resolve();
					});
				} else {
					resolve();
				}
			}));
			$.fadeOut($item).catch(() => {});
			$item.fadeAnimation.finish();
			return $item;
		}));
		
		if (this.editor) {
			const $item = this.tag.querySelector('.item');
			if ($item) {
				$.fadeIn($item).catch(() => {});
				$item.fadeAnimation.finish();
			}
		}
		
		for (const $ad of this.tag.querySelectorAll('a')) {
			if ($ad.href.indexOf('adchoice') != -1) {
				$ad.hidden = true;
			}
		}
		for (const $e of this.tag.querySelectorAll('.item > h2')) {
			Object.assign($e.style, {
				color: this.headerforeground,
				fontFamily: `'${this.font}'`,
				fontSize: `${Math.round(parseInt(this.fontsize, 10) * 1.25)}px`
			});
		}
		for (const $e of this.tag.querySelectorAll('.item > p')) {
			Object.assign($e.style, {
				color: this.foreground,
				fontFamily: `'${this.font}'`,
				fontSize: `${this.fontsize}px`
			});
		}

		//Promise.all(imagesLoaded).then(() => {
		//});
		
		this.ready();
	}
	Show() {
		if (!this.editor) {
			this.stopPlayingPromise = new Promise((resolve, reject) => this.stopPlaying = reject);
			this.cycle();
		}
	}
	Hide() {
		if (!this.editor) {
			if (this.stopPlaying) {
				this.stopPlaying();
			}
		}
	}
	Destroy() {
		clearInterval(this.rssInterval);
	}
}
