class Widgetscreenfeedstockcrawl extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.tag.classList.add('widget-screenfeedstockcrawl');
		this.tag.style.backgroundColor = this.background;
		
		this.stockInterval = null;
		this.crawl = new BreezeCrawl(this.tag, this.speed, this.alternativescroll);
		this.crawl.foreground = this.foreground;
		this.crawl.stock = true;
		this.crawl.spacer = '            ';
		
		this.stockInterval = setInterval(this.stockCrawlTick.bind(this), 15 * 60 * 1000); // 15 minute interval
		this.stockCrawlTick();
		
		this.ready();
	}
	async stockCrawlTick() {
		if (this.TagDeleted()) {
			return;
		}
		
		let cacheData = await new Promise(async resolve => {
			try {
				const data = await $.http({
					method: 'GET',
					url: this.url.replace('http://', 'https://'),
					dataType: 'xml'
				}).promise;
				/*
				<FinancialData>
					<Quote>
						<company_name>Walt Disney</company_name>
						<ticker>DIS</ticker>
						<logo_small>http://fridge.knife.screenfeed.com/content/financial/logos/500/DIS.png</logo_small>
						<logo_large>http://fridge.knife.screenfeed.com/content/financial/logos/1000/DIS.png</logo_large>
						<last>112.75</last><previous>112.12</previous>
						<change>+0.63</change>
						<change_percent>+0.56%</change_percent>
						<change_icon>http://fridge.knife.screenfeed.com/content/financial/icons/standard/up.png</change_icon>
						<time_last>8/14/2018 10:54 PM</time_last>
						<day_open>112.38</day_open>
						<day_high>113.18</day_high>
						<day_low>111.94</day_low>
						<volume>4326545</volume>
					</Quote>
				</FinancialData>
				*/
				const $xml = parser.parseFromString(data, "application/xml");
				var stocks = [];
				for (const $quote of $xml.querySelectorAll('Quote')) {
					stocks.push({
						symbol: $quote.querySelector('ticker').innerText,
						company: $quote.querySelector('company_name').innerText,
						lastPrice: $quote.querySelector('last').innerText,
						change: $quote.querySelector('change').innerText
					});
				}
				stocks.sort((a, b) => a.symbol.localeCompare(b.symbol));
				
				var text = '';
				stocks.forEach((stock, stockIndex) => {
					if (stock.company !== undefined) {
						if (stockIndex != 0) {
							text += '            ';
						}
						text += stock.company + ' (' + stock.symbol + ') ' + parseFloat(stock.lastPrice).toFixed(2) + ' ' + (parseFloat(stock.change) == 0 ? '<span style="color:gray;">▶</span>' : (parseFloat(stock.change)  > 0 ? '<span style="color:green;">▲</span>' : '<span style="color:red;">▼</span>')) + Math.abs(parseFloat(stock.change)).toFixed(2);
					}
				});
				resolve({ text: text });
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		if (!cacheData.hasOwnProperty('text')) {
			cacheData.text = '';
		}
		
		this.crawl.startCrawl(this.width, this.height, cacheData.text, this.fontsize, this.font, !this.editor);
	}
	Show() {
		this.crawl.play();
	}
	Hide() {
		this.crawl.pause();
	}
	Destroy() {
		this.crawl.destroy();
		clearInterval(this.stockInterval);
	}
}
