// For android 4.4.4 aka Chrome 53 without CSS.supports support.
if (!('CSS' in window)) {
	window.CSS = {};
}

if (!('supports' in window.CSS)) {
	window.CSS._cacheSupports = {};
	window.CSS.supports = function(propertyName, value) {
		var key = [propertyName, value].toString();
		if (key in window.CSS._cacheSupports) {
			return window.CSS._cacheSupports[key];
		}
		
		function cssSupports(propertyName, value) {
			var style = document.createElement('div').style;
			
			// 1 argument
			if (typeof value == 'undefined') {
				function mergeOdd(propertyName, reg) {
					var arr = propertyName.split(reg);

					if (arr.length > 1) {
						return arr.map(function(value, index, arr) {
							return (index % 2 == 0) ? value + arr[index+1] : '';
						}).filter(Boolean);
					}
				}
				
				// The regex will do this '( a:b ) or ( c:d )' => ["( a:b ", ")", "(", " c:d )"]
				var arrOr = mergeOdd(propertyName, /([)])\s*or\s*([(])/gi);
				if (arrOr) {
					return arrOr.some(function(supportsCondition) { return window.CSS.supports(supportsCondition); });
				}
				var arrAnd = mergeOdd(propertyName, /([)])\s*and\s*([(])/gi);
				if (arrAnd) {
					return arrAnd.every(function(supportsCondition) { return window.CSS.supports(supportsCondition); });
				}
				
				// Remove the first and last parentheses
				style.cssText = propertyName.replace('(','').replace(/[)]$/, '');
			} else { // 2 arguments
				style.cssText = propertyName + ':' + value;
			}
			return !!style.length;
		}
		return window.CSS._cacheSupports[key] = cssSupports(propertyName, value);
	};
}

class Widgetsignwave extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.currentEventResetScrollTimeout = null;
		this.eventlistResetScrollTimeout = null;
		this.timeInterval = null;
		this.loadEventsInterval = null;
		this.fpsInterval = null;
		this.fpsRequestAnimationFrame = null;
		this.currentEventDifferenceXRequestAnimationFrame = null;
		this.eventlistResetScrollRequestAnimationFrame = null;
		this.eventlistDifferenceYRequestAnimationFrame = null;
		
		this.onlineModeTimer = null;
		this.offlineMode = null;
		
		this.displayedEvents = [];
		this.allEvents = [];
		this.newAllEvents = [];
		this.internalInstantEvents = [];
		
		this.currentEventIndex = 0;
		this.currentOrNextEventIndex = 0;
		this.upcomingEvent = false;
		this.currentEventOffset = 0;
		this.pin = '';
		this.currentEventDifferenceX = 0;
		this.eventlistDifferenceY = 0;
		this.resetTimer = null;
		this.instantEventStart = 0;
		this.instantEventEnd = 0;
		this.instantEventTitle = '';
		this.instantEventName = '';
		this.currentFPS = 0;
		this.lastViewBy = '';
		this.lastViewByEventListDayOffset = 0;

		this.oldServerOffset = this.player.serverOffset;
		
		this.computedWidth = this.width;
		this.computedHeight = this.height;
		if (this.orientation == '90 degrees' || this.orientation == '270 degrees') {
			[this.computedWidth, this.computedHeight] = [this.computedHeight, this.computedWidth];
		}
		
		// Cached data
		this.events = [];
		this.rooms = [];
		
		// Cached player data in this.widgetData.signwave
		this.roomids = [];
		this.overridebackgroundimage = false;
		this.backgroundimage = null;
		this.backgroundimagecoverfill = false;
		this.overridereplacementnames = {};
		// If the player has not been updated yet then this field will be empty. Kind of an edge case as anyone using Signwave will have pressed update to set room information
		if (this.widgetData.hasOwnProperty('signwave')) {
			this.roomids = this.widgetData.signwave.roomids;
			this.overridebackgroundimage = this.widgetData.signwave.overridebackgroundimage;
			this.backgroundimage = JSON.parse(JSON.stringify(this.widgetData.signwave.backgroundimage));
			this.backgroundimagecoverfill = this.widgetData.signwave.backgroundimagecoverfill;
			this.overridereplacementnames = this.widgetData.signwave.hasOwnProperty('overridereplacementnames') ? this.widgetData.signwave.overridereplacementnames : {};
		}
		
		this.TestOnlineMode();
	}
	ConfirmationBox(title, message) {
		return new Promise(resolve => {
			const confirmationBoxTag = this.tag.appendChild($.ul({ class: 'signwave-modal', css: { width: '550px' } },
				$.li({ text: title, class: 'signwave-modal-title' }),
				$.li({ text: message, class: 'signwave-modal-message' }),
				$.li({ class: 'signwave-modal-buttons' },
					$.button({ text: 'OK', click: () => {
						confirmationBoxTag.remove();
						resolve();
					} })
				)
			));
			
			const viewportWidth = this.tag.offsetWidth;
			const viewportHeight = this.tag.offsetHeight;
			confirmationBoxTag.style.left = `${viewportWidth / 2 - confirmationBoxTag.offsetWidth / 2}px`;
			confirmationBoxTag.style.top = `${viewportHeight / 2 - confirmationBoxTag.offsetHeight / 2}px`;
		});
	}
	ClearTimers() {
		clearTimeout(this.currentEventResetScrollTimeout);
		this.currentEventResetScrollTimeout = null;
		
		clearTimeout(this.eventlistResetScrollTimeout);
		this.eventlistResetScrollTimeout = null;
		
		clearInterval(this.timeInterval);
		this.timeInterval = null;
		
		clearInterval(this.loadEventsInterval);
		this.loadEventsInterval = null;
		
		clearInterval(this.fpsInterval);
		this.fpsInterval = null;
		
		cancelAnimationFrame(this.fpsRequestAnimationFrame);
		this.fpsRequestAnimationFrame = null;
		
		cancelAnimationFrame(this.currentEventDifferenceXRequestAnimationFrame);
		this.currentEventDifferenceXRequestAnimationFrame = null;
		
		cancelAnimationFrame(this.eventlistResetScrollRequestAnimationFrame);
		this.eventlistResetScrollRequestAnimationFrame = null;
		
		cancelAnimationFrame(this.eventlistDifferenceYRequestAnimationFrame);
		this.eventlistDifferenceYRequestAnimationFrame = null;
	}
	UpdatePin() {
		this.tag.querySelector('.numpaddisplay').innerHTML = '';
		var pinHidden = '';
		for (var itr = 0; itr < this.pin.length; ++itr) {
			pinHidden += '*';
		}
		this.tag.querySelector('.numpaddisplay').innerText = pinHidden;
	}
	ResetEventList() {
		for (const $scrollArrow of this.tag.querySelectorAll('.eventlistscrollarrowup, .eventlistscrollarrowdown')) $.fadeOut($scrollArrow).catch(() => {});
		this.displayedEvents = [];
		this.eventlistDifferenceY = 0;
		cancelAnimationFrame(this.eventlistDifferenceYRequestAnimationFrame);
		this.eventlistDifferenceYRequestAnimationFrame = null;
		clearTimeout(this.eventlistResetScrollTimeout);
		this.eventlistResetScrollTimeout = null;
		cancelAnimationFrame(this.eventlistResetScrollRequestAnimationFrame);
		this.eventlistResetScrollRequestAnimationFrame = null;
	}
	ResetActivity() {
		this.ButtonFadeIn();
		this.StartResetTimer();
	}
	RemoveActivityDetection() {
		if (this.tag.hasOwnProperty('lastActivityPointerDown')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastActivityPointerDown);
		}
		if (this.tag.hasOwnProperty('lastActivityPointerMove')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastActivityPointerMove);
		}
		if (this.tag.hasOwnProperty('lastActivityPointerUp')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastActivityPointerUp);
		}
	}
	DefaultActivityDetection() {
		this.RemoveActivityDetection();
		this.tag.lastActivityPointerDown = this.tag.lastActivityPointerMove = this.tag.lastActivityPointerUp = this.ResetActivity.bind(this);
		for (const event of 'onpointerdown' in window ? ['pointerdown', 'pointermove', 'pointerup'] : ['touchstart', 'touchmove', 'touchend']) {
			this.tag.addEventListener(event, this.ResetActivity.bind(this));
		}
	}
	async ShowKeyboard(value, numbers) {
		return new Promise(resolve => {
			this.$keyboard.hidden = false;
			var capsLock = true;
			var mode = 0;
			const keys =
			[
				[['1', '1'], ['2', '2'], ['3', '3'], ['4', '4'], ['5', '5'], ['6', '6'], ['7', '7'], ['8', '8'], ['9', '9'], ['0', '0']],
				[['q', '!'], ['w', '@'], ['e', '#'], ['r', '$'], ['t', '%'], ['y', '^'], ['u', '&'], ['i', '*'], ['o', '('], ['p', ')']],
				[[' ', ' '], ['a', '_'], ['s', '+'], ['d', '='], ['f', ':'], ['g', ';'], ['h', '/'], ['j', '?'], ['k', '\\'], ['l', ','], [' ', ' ']],
				[['!@#', 'ABC'], ['z', '.'], ['x', '<'], ['c', '>'], ['v', '['], ['b', ']'], ['n', '{'], ['m', '}'], ['-', '-'], ['\'', '\'']]
			];
			const renderKeyboard = () => {
				for (const $ul of this.tag.querySelectorAll('.keyboard > ul')) $ul.remove();
				keys.forEach((row, rowIndex) => {
					if (rowIndex == 0 && !numbers) {
						return;
					}
					var rowTag = $.ul();
					row.forEach((key, keyIndex) => {
						key = key[mode];
						var keyLetter = key;
						if (capsLock) {
							keyLetter = key.toUpperCase();
						}
						var keyTag = $.li({}, $.span({ text: keyLetter }));
						if (key == ' ') {
							keyTag.style.opacity = 0.5;
						} else if (key != 'ABC' && key != '!@#') {
							keyTag.addEventListener('click', () => {
								if (this.$keyboardInputValue.innerText.length < 40) {
									this.$keyboardInputValue.innerText += keyTag.querySelector('span').innerText;
								}
							});
						}
						if (rowIndex == 2 && (keyIndex == 0 || keyIndex == 10)) {
							keyTag.style.flex = '0.5 0 0px';
							if (keyIndex == 0) {
								if (capsLock) {
									keyTag.querySelector('span').classList.add('mdi', 'mdi-play', 'mdi-rotate-90');
								} else {
									keyTag.querySelector('span').classList.add('mdi', 'mdi-play', 'mdi-rotate-270');
								}
								keyTag.style.opacity = 1;
								keyTag.addEventListener('click', () => {
									capsLock = !capsLock;
									renderKeyboard();
								});
							}
							if (keyIndex == 10) {
								keyTag.querySelector('span').classList.add('mdi', 'mdi-play', 'mdi-rotate-180');
								keyTag.style.opacity = 1;
								keyTag.addEventListener('click', () => {
									this.$keyboardInputValue.innerText = this.$keyboardInputValue.innerText.slice(0, -1);
								});
							}
						} else if (rowIndex == 3) {
							if (keyIndex == 0) {
								keyTag.addEventListener('click', () => {
									mode = mode == 0 ? 1 : 0;
									renderKeyboard();
								});
							}
						}
						rowTag.appendChild(keyTag);
					});
					this.$keyboard.appendChild(rowTag);
				});
				this.$keyboard.appendChild($.ul({},
					$.li({ click: () => {
						if (this.$keyboardInputValue.innerText.length < 40) {
							this.$keyboardInputValue.innerText += ' ';
						}
					} },
						$.span({ text: 'Space' })
					),
					$.li({ css: { flex: '0.21 0 0' }, click: () => {
						resolve(this.$keyboardInputValue.innerText);
						this.$keyboard.hidden = true;
					} },
						$.span({ text: 'OK' })
					)
				));
			};
			renderKeyboard();
			this.$keyboardInputValue.innerText = value;
		});
	}
	StartResetTimer() {
		//clearTimeout(this.resetTimer);
		//this.resetTimer = setTimeout(this.ResetView.bind(this), this.resetviewdelay * 1000);
	}
	ButtonFadeOut() {
		this.$bottom.style.pointerEvents = 'none';
		$.fadeOut(this.$bottom).catch(() => {});
	}
	ButtonFadeIn() {
		this.$bottom.style.pointerEvents = 'all';
		$.fadeIn(this.$bottom).catch(() => {});
	}
	HideRoomTimeLogo() {
		// Information
		this.$room.hidden = true;
		this.$time.hidden = true;
		this.$logo.hidden = true;
	}
	RemoveLastPointerEvents() {
		if (this.tag.hasOwnProperty('lastPointerDown')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastPointerDown);
		}
		if (this.tag.hasOwnProperty('lastPointerMove')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastPointerMove);
		}
		if (this.tag.hasOwnProperty('lastPointerUp')) {
			this.tag.removeEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastPointerUp);
		}
	}
	HideAllButRoomTimeLogo() {
		this.$loading.hidden = true;
		// Views
		this.$currentEvent.hidden = true;
		this.$previousEventArrow.hidden = true;
		this.$nextEventArrow.hidden = true;
		this.$viewEventList.hidden = true;
		this.$eventListScrollArrowUp.hidden = true;
		this.$eventListScrollArrowDown.hidden = true;
		this.$eventListNoEvent.hidden = true;
		this.$viewPin.hidden = true;
		this.$viewInstantEvent.hidden = true;
		this.$viewInstantEventForm.hidden = true;
		this.$viewInstantEventConfirmation.hidden = true;
		this.$viewReportIssueConfirmation.hidden = true;
		// Buttons
		this.$buttonSpacer.hidden = false;
		this.$buttonSpacerDayOfTheWeek.hidden = true;
		this.$viewByHour.hidden = true;
		this.$viewByDay.hidden = true;
		this.$viewByWeek.hidden = true;
		this.$back.hidden = true;
		this.$next.hidden = true;
		this.$confirm.hidden = true;
		this.$cancel.hidden = true;
		for (const $day of this.$daysOfTheWeek) $day.hidden = true;
		this.$instantEvent.hidden = true;
		this.$reportIssue.hidden = true;
		for (const $e of [
			this.$viewByHour,
			this.$viewByDay,
			this.$viewByWeek,
			...this.$daysOfTheWeek,
			this.$instantEvent,
			this.$reportIssue ]) {
			$e.classList.remove('selected');
		}
		this.UpdatePin();
		this.displayedEvents = [];
		this.currentEventIndex = 0;
		this.currentEventOffset = 0;
		this.currentEventDifferenceX = 0;
		cancelAnimationFrame(this.currentEventDifferenceXRequestAnimationFrame);
		this.currentEventDifferenceXRequestAnimationFrame = null;
		this.ResetEventList();
		this.StartResetTimer();

		this.RemoveActivityDetection();
		this.RemoveLastPointerEvents();
		this.$keyboard.hidden = true;
	}
	HideAll() {
		this.HideRoomTimeLogo();
		this.HideAllButRoomTimeLogo();
	}
	async ShowRoomTimeLogo() {
		if (CSS.supports("display: grid")) {
			// These are used to ensure that the room field collapses into a trailing overflow
			const roomLeft = this.showroom && this.roomhorizontalalignment == 'left';
			const roomRight = this.showroom && this.roomhorizontalalignment == 'right';
			const timeOrRoomLeft = roomLeft || this.showtime && this.timehorizontalalignment == 'left';
			const timeOrRoomRight = roomRight || this.showtime && this.timehorizontalalignment == 'right';
			const logoCenter = this.showlogo && this.logohorizontalalignment == 'center';
			
			this.$top.style.gridTemplateColumns = `[logo-left] ${!timeOrRoomLeft ? `minmax(max-content,1fr)` : 'max-content'} [room-left time-left] ${timeOrRoomLeft ? `minmax(${roomLeft ? '0' : 'max-content'},${roomRight && !logoCenter ? 'max-content' : '1fr'})` : 'max-content'} [room-center] minmax(0,max-content) [logo-center] max-content [time-center] max-content [room-right time-right] ${timeOrRoomRight ? `minmax(${roomRight ? '0' : 'max-content'},${roomLeft && !logoCenter ? 'max-content' : '1fr'})` : 0} [logo-right] ${!timeOrRoomRight ? `minmax(max-content,1fr)` : 'max-content'}`;
			
			Object.assign(this.$room.style, {
				gridColumn: `room-${this.roomhorizontalalignment}`,
				gridRow: this.showtime && this.roomhorizontalalignment == this.timehorizontalalignment && this.roomhorizontalalignment != 'center' ? 1 : '1 / span 2', // If room and time overlap then stack them in separate rows
				textAlign: this.roomhorizontalalignment
			});
			this.$room.hidden = !this.showroom;
			Object.assign(this.$time.style, {
				gridColumn: `time-${this.timehorizontalalignment}`,
				gridRow: this.showroom && this.timehorizontalalignment == this.roomhorizontalalignment && this.timehorizontalalignment != 'center' ? 2 : '1 / span 2', // If room and time overlap then stack them in separate rows
				textAlign: this.timehorizontalalignment
			});
			this.$time.hidden = !this.showtime;
			Object.assign(this.$logo.style, {
				gridColumn: `logo-${this.logohorizontalalignment}`,
				gridRow: '1 / span 2',
				justifySelf: { left: 'flex-start', center: 'center', right: 'flex-end' }[this.logohorizontalalignment]
			});
			this.$logo.hidden = !this.showlogo;

			// Special case where both room and time are center and there's no logo between them so stack them.
			if (this.showroom && this.showtime && (!this.showlogo || this.logohorizontalalignment != 'center') && this.roomhorizontalalignment == 'center' && this.timehorizontalalignment == 'center') {
				this.$room.style.gridRow = '1 / span 1';
				Object.assign(this.$time.style, {
					gridColumn: 'room-center',
					gridRow: '2 / span 1'
				});
			}
		} else {
			this.$room.hidden = !this.showroom;
			this.$time.hidden = !this.showtime;
			this.$logo.hidden = !this.showlogo;
			//const roomWidth = $.outerWidth(this.$room);
			const roomHeight = $.outerHeight(this.$room);
			const timeWidth = $.outerWidth(this.$time);
			const timeHeight = $.outerHeight(this.$time);
			const logoWidth = $.outerWidth(this.$logo);
			const logoHeight = $.outerHeight(this.$logo);
			const roomComputedStyle = getComputedStyle(this.$room);
			const timeComputedStyle = getComputedStyle(this.$time);
			//const logoComputedStyle = getComputedStyle(this.$logo);
			var positionRoomOffsetX = 0;
			var positionRoomOffsetY = 0;
			var positionTimeOffsetX = 0;
			var positionTimeOffsetY = 0;
			if (this.showroom && this.showtime && this.showlogo) {
				if (this.roomhorizontalalignment == this.timehorizontalalignment && this.timehorizontalalignment == this.logohorizontalalignment) {
					// room, time, and logo are all position in the same spot
					positionRoomOffsetX = this.roomhorizontalalignment == 'center' ? -(1.5 * this.$room.offsetWidth / 2 + this.$logo.offsetWidth / 2) : this.$logo.offsetWidth;
					positionRoomOffsetY = this.roomhorizontalalignment == 'center' ? this.$logo.offsetHeight / 2 - this.$room.offsetHeight / 2 - parseInt(roomComputedStyle.marginBottom, 10) : (this.buttonlocation == 'top' ? this.$time.offsetHeight : 0);
					positionTimeOffsetX = this.roomhorizontalalignment == 'center' ? this.$time.offsetWidth / 2 + this.$logo.offsetWidth / 2 : this.$logo.offsetWidth;
					positionTimeOffsetY = this.timehorizontalalignment == 'center' ? this.$logo.offsetHeight / 2 - this.$time.offsetHeight / 2 - parseInt(timeComputedStyle.marginBottom, 10) : (this.buttonlocation == 'bottom' ? this.$room.offsetHeight : 0);
					this.$top.style.height = `${Math.max(roomHeight + timeHeight - parseInt(roomComputedStyle.marginBottom, 10) - parseInt(timeComputedStyle.marginTop, 10), logoHeight)}px`;
				} else if (this.roomhorizontalalignment == this.timehorizontalalignment) {
					positionRoomOffsetY = this.buttonlocation == 'top' ? this.$time.offsetHeight : 0;
					positionTimeOffsetY = this.buttonlocation == 'bottom' ? this.$room.offsetHeight : 0;
					this.$top.style.height = `${Math.max(roomHeight + timeHeight - parseInt(roomComputedStyle.marginBottom, 10) - parseInt(timeComputedStyle.marginTop, 10), logoHeight)}px`;
					this.$room.style.maxWidth = `${Math.max(100, this.computedWidth - logoWidth)}px`;
				} else if (this.roomhorizontalalignment == this.logohorizontalalignment) {
					positionRoomOffsetX = this.$logo.offsetWidth;
					this.$top.style.height = `${Math.max(roomHeight, timeHeight, logoHeight)}px`;
					this.$room.style.maxWidth = `${Math.max(100, this.computedWidth - timeWidth)}px`;
				} else if (this.timehorizontalalignment == this.logohorizontalalignment) {
					positionTimeOffsetX = this.$logo.offsetWidth;
					this.$top.style.height = `${Math.max(roomHeight, timeHeight, logoHeight)}px`;
					this.$room.style.maxWidth = `${Math.max(100, this.computedWidth - Math.max(timeWidth, logoWidth))}px`;
				} else {
					this.$top.style.height = `${Math.max(roomHeight, timeHeight, logoHeight)}px`;
				}
			} else if (this.showroom && this.showtime) {
				this.$top.style.height = `${Math.max(roomHeight, timeHeight)}px`;
				if (this.roomhorizontalalignment == this.timehorizontalalignment) {
					positionRoomOffsetY = this.buttonlocation == 'top' ? this.$time.offsetHeight : 0;
					positionTimeOffsetY = this.buttonlocation == 'bottom' ? this.$room.offsetHeight : 0;
					this.$top.style.height = `${roomHeight + timeHeight - parseInt(roomComputedStyle.marginBottom, 10) - parseInt(timeComputedStyle.marginTop, 10)}px`;
				} else {
					this.$room.style.maxWidth = `${Math.max(100, this.computedWidth - timeWidth)}px`;
				}
			} else if (this.showroom && this.showlogo) {
				this.$top.style.height = `${Math.max(roomHeight, logoHeight)}px`;
				if (this.roomhorizontalalignment == this.logohorizontalalignment) {
					positionRoomOffsetX = this.$logo.offsetWidth;
				} else {
					this.$room.style.maxWidth = `${Math.max(100, this.computedWidth - logoWidth)}px`;
				}
			} else if (this.showtime && this.showlogo) {
				this.$top.style.height = `${Math.max(timeHeight, logoHeight)}px`;
				if (this.timehorizontalalignment == this.logohorizontalalignment) {
					positionTimeOffsetX = this.$logo.offsetWidth;
				}
			} else if (this.showroom) {
				this.$top.style.height = `${roomHeight}px`;
			} else if (this.showtime) {
				this.$top.style.height = `${timeHeight}px`;
			} else if (this.showlogo) {
				this.$top.style.height = `${logoHeight}px`;
			}
			
			Object.assign(this.$room.style, {
				left: this.roomhorizontalalignment == 'left' ? `${positionRoomOffsetX}px` : (this.roomhorizontalalignment == 'center' ? `${this.computedWidth / 2 - this.$room.offsetWidth / 2 + positionRoomOffsetX}px` : 'auto'),
				right: this.roomhorizontalalignment == 'right' ? `${positionRoomOffsetX}px` : 'auto',
				top: this.buttonlocation == 'bottom' ? `${positionRoomOffsetY}px` : 'auto',
				bottom: this.buttonlocation == 'top' ? `${positionRoomOffsetY}px` : 'auto'
			});
			
			Object.assign(this.$time.style, {
				left: this.timehorizontalalignment == 'left' ? `${positionTimeOffsetX}px` : (this.timehorizontalalignment == 'center' ? `${this.computedWidth / 2 - this.$time.offsetWidth / 2 + positionTimeOffsetX}px` : 'auto'),
				right: this.timehorizontalalignment == 'right' ? `${positionTimeOffsetX}px` : 'auto',
				top: this.buttonlocation == 'bottom' ? `${positionTimeOffsetY}px` : 'auto',
				bottom: this.buttonlocation == 'top' ? `${positionTimeOffsetY}px` : 'auto'
			});
			
			Object.assign(this.$logo.style, {
				left: this.logohorizontalalignment == 'left' ? 0 : (this.logohorizontalalignment == 'center' ? `${this.computedWidth / 2 - this.$logo.offsetWidth / 2}px` : 'auto'),
				right: this.logohorizontalalignment == 'right' ? 0 : 'auto',
				top: this.buttonlocation == 'bottom' ? 0 : 'auto',
				bottom: this.buttonlocation == 'top' ? 0 : 'auto'
			});
		}
	}
	ShowViewByButtons() {
		this.$buttonSpacer.hidden = false;
		this.$viewByHour.hidden = !(this.showcurrentevent && (this.enableviewbydaybutton || this.enableviewbyweekbutton));
		this.$viewByDay.hidden = !(this.enableviewbydaybutton && (this.showcurrentevent || this.enableviewbyweekbutton));
		this.$viewByWeek.hidden = !this.enableviewbyweekbutton;
	}
	ShowInstantAndReportButtons() {
		this.$instantEvent.hidden = !this.enableinstanteventbutton;
		this.$reportIssue.hidden = !this.enablereportanissuebutton;
	}
	PostColor(available) {
		if (window.Messenger !== undefined) {
			const color = available ?
				this.roomavailablelightcolor :
				this.roomunavailablelightcolor;
			window.Messenger.bus.emit(
				'FrontLedColorChange',
				color,
				this.mediaid
			);
		} else {
			window.parent.postMessage(
			{
				type: 'currentevent',
				mediaid: this.mediaid,
				available: available,
				availablecolor: this.roomavailablelightcolor,
				unavailablecolor: this.roomunavailablelightcolor
			}, '*');
		}
	}
	ViewEventsChanged(newEvents) {
		return newEvents.length != this.displayedEvents.length || this.displayedEvents.some((event, index) =>
		{
			return !event.start.isSame(newEvents[index].start) ||
					!event.end.isSame(newEvents[index].end) ||
					event.title != newEvents[index].title ||
					event.room != newEvents[index].room ||
					event.name != newEvents[index].name;
		});
	}
	async ResetView(forceUpdate = false) {
		//console.log('ResetView');
		if (this.TagDeleted()) {
			return;
		}
		// allEvents is just an array with the newAllEvents events and the internalInstantEvents
		this.allEvents = this.newAllEvents.concat(this.internalInstantEvents).sort((a, b) => a.start.unix() - b.start.unix());
		
		if (this.showcurrentevent) {
			var currentEventState = this.CurrentEventState(0);
			//console.log('lastViewBy: ' + this.lastViewBy + ' ViewEventsChanged: ' + ViewEventsChanged(currentEventState.events) + ' this.currentEventIndex: ' + this.currentEventIndex + ' new currentEventIndex: ' + currentEventState.currentEventIndex + ' upcomingEvent: ' + this.upcomingEvent + ' new upcomingEvent: ' + currentEventState.upcomingEvent);
			if (forceUpdate || this.lastViewBy != 'hour' || this.ViewEventsChanged(currentEventState.events) || currentEventState.currentEventIndex != this.currentEventIndex || currentEventState.upcomingEvent != this.upcomingEvent) {
				//console.log('Updating');
				await this.ShowViewByHour();
			}
		} else if (this.enableviewbydaybutton) {
			var eventListState = this.EventListState(0);
			if (forceUpdate || this.lastViewBy != 'day' || this.ViewEventsChanged(eventListState.events) || eventListState.currentEventIndex != this.currentEventIndex || eventListState.currentOrNextEventIndex != this.currentOrNextEventIndex) {
				await this.ShowViewByDay();
			}
		} else if (this.enableviewbyweekbutton) {
			var eventListState = this.EventListState(0);
			if (forceUpdate || this.lastViewBy != 'week' || this.ViewEventsChanged(eventListState.events) || eventListState.currentEventIndex != this.currentEventIndex || eventListState.currentOrNextEventIndex != this.currentOrNextEventIndex) {
				await this.ShowViewByWeek();
			}
		}
		if (this.hidebuttons) {
			this.ButtonFadeOut();
		}
		this.StartResetTimer();
	}
	GetPreviousEvent() {
		// showexpiredevents
		var previousIndex = this.currentEventIndex + this.currentEventOffset - 1;
		return previousIndex >= 0 ? this.displayedEvents[previousIndex] : null;
	}
	GetCurrentEvent() {
		return this.displayedEvents[this.currentEventIndex + this.currentEventOffset];
	}
	GetNextEvent() {
		var nextIndex = this.currentEventIndex + this.currentEventOffset + 1;
		return nextIndex < this.displayedEvents.length ? this.displayedEvents[nextIndex] : null;
	}
	CurrentEventState(dayOffset) {
		const currentTime = this.player.currentTime;
		var newState = { events: [], currentEventIndex: null, upcomingEvent: false };
		newState.events = this.allEvents.filter(event => this.showexpiredevents || event.end.isAfter(currentTime));
		newState.events = newState.events.filter(event => {
			return event.end.isSameOrAfter(currentTime.clone().startOf('day').add(dayOffset, 'day')) &&
			       event.start.isBefore(currentTime.clone().endOf('day').add(dayOffset, 'day'));
		});
		newState.events.forEach((event, eventIndex) => {
			// If the current time is inside of the event then this is the current event
			if (currentTime.isSameOrAfter(event.start) && currentTime.isBefore(event.end)) {
				newState.currentEventIndex = eventIndex;
			}
		});
		if (newState.currentEventIndex == null) {
			for (var eventIndex = 0; eventIndex < newState.events.length; ++eventIndex) {
				var event = newState.events[eventIndex];
				if (currentTime.isBefore(event.start) && currentTime.clone().add(this.upcomingeventthreshold, 'minutes').isSameOrAfter(event.start)) {
					newState.currentEventIndex = eventIndex;
					newState.upcomingEvent = true;
					break;
				}
			}
		}
		if (newState.currentEventIndex == null) {
			var currentEventPlaceholder = { start: currentTime.clone().startOf('day'), end: currentTime.clone().endOf('day'), currentEventPlaceholder: true };
			newState.currentEventIndex = 0;
			newState.events.forEach((event, eventIndex) => {
				if (currentTime.isAfter(event.end)) {
					newState.currentEventIndex = eventIndex + 1;
				}
			});
			newState.events.splice(newState.currentEventIndex, 0, currentEventPlaceholder);
		}
		return newState;
	}
	ShowCurrentEvent(dayOffset, viewBy, customStart, customEnd, customTitle, customName) {
		this.$currentEvent.hidden = false;
		this.$previousEventArrow.hidden = false;
		this.$nextEventArrow.hidden = false;
		// Filter events by today and set the currentEventIndex
		if (dayOffset == -1) {
			this.displayedEvents = [{ "start": customStart, "end": customEnd, "title": customTitle, "name": customName }];
			this.currentEventIndex = 0;
			this.upcomingEvent = false;
		} else {
			var newState = this.CurrentEventState(dayOffset);
			this.displayedEvents = newState.events;
			this.currentEventIndex = newState.currentEventIndex;
			this.upcomingEvent = newState.upcomingEvent;
		}
		if (dayOffset == 0) {
			this.PostColor(this.currentEventIndex == null || this.displayedEvents[this.currentEventIndex].hasOwnProperty('currentEventPlaceholder') || this.upcomingEvent);
		}
		
		const updateTags = () => {
			var previousEvent = this.GetPreviousEvent();
			var currentEvent = this.GetCurrentEvent();
			var nextEvent = this.GetNextEvent();
			
			var previousCurrentNextEvents = [{ event: previousEvent, name: 'previousevent' }, { event: currentEvent, name: 'currentevent' }, { event: nextEvent, name: 'nextevent' }];
			var firstSecondThirdLines = [{ type: this.currenteventfirstline, name: 'firstline' }, { type: this.currenteventsecondline, name: 'secondline' }, { type: this.currenteventthirdline, name: 'thirdline' }];
			
			previousCurrentNextEvents.forEach((event, eventIndex) => {
				firstSecondThirdLines.forEach((line, lineIndex) => {
					const lineTag = this.tag.querySelector(`.${event.name}-${line.name}`);
					lineTag.style.width = `${this.computedWidth - 100}px`;
					if (event.event && !event.event.hasOwnProperty('currentEventPlaceholder')) {
						lineTag.innerHTML = '';
						if (line.type == 'title' && event.event[line.type].indexOf('https://') == 0) {
							line.type = 'image';
						}
						if (line.type == 'name' && (event.event[line.type].indexOf('http://') == 0 || event.event[line.type].indexOf('https://') == 0)) {
							line.type = 'image2';
						}
						
						if (line.type == 'none') {
							lineTag.innerText = '';
						} else if (line.type == 'time') {
							lineTag.append(
								$.span({ text: this.player.applyTimezone(moment(event.event.start)).format('h:mm ') }),
								$.span({ text: this.player.applyTimezone(moment(event.event.start)).format('A') , css: { fontSize: '0.7em' } }),
								$.span({ text: ' - ' }),
								$.span({ text: this.player.applyTimezone(moment(event.event.end)).format('h:mm ') }),
								$.span({ text: this.player.applyTimezone(moment(event.event.end)).format('A') , css: { fontSize: '0.7em' } })
							);
						} else if (line.type == 'image') {
							lineTag.appendChild(
								$.img({ src: event.event.title }).load(updatePositions)
							);
						} else if (line.type == 'image2') {
							lineTag.appendChild(
								$.img({ src: event.event.name }).load(updatePositions)
							);
						} else {
							lineTag.innerText = event.event[line.type];
						}
						lineTag.hidden = false;
					} else {
						lineTag.hidden = true;
					}
				});
			});
			
			this.$currentUpcomingEvent.style.width = `${this.computedWidth - 100}px`,
			this.$currentUpcomingEvent.hidden = !(this.upcomingEvent && this.currentEventOffset == 0);
			
			Object.assign(this.$previousEventArrow.style, {
				left: '10px',
				top: `${this.computedHeight / 2 - this.$previousEventArrow.offsetHeight / 2}px`
			});
			if (previousEvent != null && this.enableswipepreviousandnext) {
				$.fadeIn(this.$previousEventArrow).catch(() => {});
			} else {
				$.fadeOut(this.$previousEventArrow).catch(() => {});
			}
			Object.assign(this.$nextEventArrow.style, {
				right: '10px',
				top: `${this.computedHeight / 2 - this.$nextEventArrow.offsetHeight / 2}px`
			});
			if (nextEvent != null && this.enableswipepreviousandnext) {
				$.fadeIn(this.$nextEventArrow).catch(() => {});
			} else {
				$.fadeOut(this.$nextEventArrow).catch(() => {});
			}
			switch (this.nocurrenteventimagefillstyle) {
				case 'fitinsidecurrenteventregion':
				case 'fillcurrenteventregionwithaspectratio':
					Object.assign(this.$currentEventNoEvent.style, {
						width: `${this.computedWidth - 100}px`,
						top: `${Math.max(this.$top.offsetHeight, this.$bottom.offsetHeight)}px`,
						bottom: `${Math.max(this.$top.offsetHeight, this.$bottom.offsetHeight)}px`
					});
					break;
				case 'fillfullscreen':
					Object.assign(this.$currentEventNoEvent.style, {
						width: `${this.computedWidth}px`,
						top: 0,
						bottom: 0
					});
					break;
				case 'fillfullscreenwithaspectratio':
					Object.assign(this.$currentEventNoEvent.style, {
						width: `${this.computedWidth}px`,
						top: 0,
						bottom: 0
					});
					break;
				case 'fillfullscreenwithcover':
					Object.assign(this.$currentEventNoEvent.style, {
						width: `${this.computedWidth}px`,
						top: 0,
						bottom: 0
					});
					break;
			}
			this.$currentEventNoEvent.hidden = !(this.showimagefornocurrentevent && (currentEvent == null || previousEvent != null && previousEvent.hasOwnProperty('currentEventPlaceholder') || currentEvent.hasOwnProperty('currentEventPlaceholder') || nextEvent != null && nextEvent.hasOwnProperty('currentEventPlaceholder')));
		};
		
		const updatePositions = () => {
			var previousEvent = this.GetPreviousEvent();
			var currentEvent = this.GetCurrentEvent();
			var nextEvent = this.GetNextEvent();
			
			var previousCurrentNextEvents = [{ event: previousEvent, name: 'previousevent' }, { event: currentEvent, name: 'currentevent' }, { event: nextEvent, name: 'nextevent' }];
			var firstSecondThirdLines = [{ type: this.currenteventfirstline, name: 'firstline' }, { type: this.currenteventsecondline, name: 'secondline' }, { type: this.currenteventthirdline, name: 'thirdline' }];
			
			previousCurrentNextEvents.forEach((event, eventIndex) => {
				// Find the large font size line
				var centerLineIndex = 0;
				var centerLineFontSize = 0;
				firstSecondThirdLines.forEach((line, lineIndex) => {
					const $line = this.tag.querySelector(`.${event.name}-${line.name}`);
					if (line.type != 'none') {
						if (parseInt($line.style.fontSize, 10) > centerLineFontSize) {
							centerLineFontSize = parseInt($line.style.fontSize, 10);
							centerLineIndex = lineIndex;
						}
					}
				});
				// Find the top and bottom boundaries. We'll use these to move the content up or down if it's outside the bounds.
				var topOffset = 0;
				var bottomOffset = 0;
				firstSecondThirdLines.forEach((line, lineIndex) => {
					const $line = this.tag.querySelector(`.${event.name}-${line.name}`);
					if (line.type != 'none') {
						if (lineIndex < centerLineIndex) {
							topOffset -= $line.offsetHeight;
						} else if (lineIndex == centerLineIndex) {
							topOffset += this.computedHeight / 2 - $line.offsetHeight / 2;
							bottomOffset += this.computedHeight / 2 + $line.offsetHeight / 2;
						} else {
							bottomOffset += $line.offsetHeight;
						}
					}
				});
				var heightOffset = bottomOffset - topOffset;
				
				let middleTop = this.$middle.offsetTop;
				switch (this.orientation) {
					case '180 degrees':
						middleTop = this.computedHeight - (middleTop + this.$middle.offsetHeight);
						break;
					case '90 degrees':
						middleTop = this.computedHeight - (this.$middle.offsetLeft + this.$middle.offsetHeight);
						break;
					case '270 degrees':
						middleTop = this.$middle.offsetLeft;
						break;
				}
				const pushDown = () => {
					if (topOffset < middleTop) {
						topOffset += middleTop - topOffset;
					}
				};
				const pushUp = () => {
					if (topOffset + heightOffset > middleTop + this.$middle.offsetHeight) {
						topOffset += middleTop + this.$middle.offsetHeight - (topOffset + heightOffset);
					}
				};
				if (this.buttonlocation == 'top') {
					pushUp();
					pushDown();
				} else {
					pushDown();
					pushUp();
				}
				var offset = topOffset;
				firstSecondThirdLines.forEach((line, lineIndex) => {
					if (line.type != 'none') {
						const $line = this.tag.querySelector(`.${event.name}-${line.name}`);
						$line.style.transform = `translate(${50 + (eventIndex - 1) * this.computedWidth + this.currentEventDifferenceX}px,${offset}px)`;
						offset += $line.offsetHeight;
					}
				});
				// For the current event show the upcoming event below if required
				if (eventIndex == 1 && this.upcomingEvent && this.currentEventOffset == 0) {
					this.$currentUpcomingEvent.style.transform = `translate(${50 + (eventIndex - 1) * this.computedWidth + this.currentEventDifferenceX}px,${offset}px)`;
				}
			});
			
			if (this.showimagefornocurrentevent && (currentEvent == null || previousEvent != null && previousEvent.hasOwnProperty('currentEventPlaceholder') || currentEvent.hasOwnProperty('currentEventPlaceholder') || nextEvent != null && nextEvent.hasOwnProperty('currentEventPlaceholder'))) {
				var offsetX = 0;
				if (previousEvent != null && previousEvent.hasOwnProperty('currentEventPlaceholder')) {
					offsetX = -this.computedWidth;
				}
				if (nextEvent != null && nextEvent.hasOwnProperty('currentEventPlaceholder')) {
					offsetX = this.computedWidth;
				}
				var noEventOffsetX = 0;
				switch (this.nocurrenteventimagefillstyle) {
					case 'fitinsidecurrenteventregion':
					case 'fillcurrenteventregionwithaspectratio':
						noEventOffsetX = 50;
						break;
				}
				this.$currentEventNoEvent.style.transform = `translate(${noEventOffsetX + offsetX + this.currentEventDifferenceX}px,0px)`;
			}
		};
		
		const resetScrollAnimation = () => {
			if (this.TagDeleted()) {
				return;
			}
			// Weight and exponent and be tweaked to change the swipe behavior. Depending on tablet performance some values might feel more natural
			var weight = 40;
			var exponent = 4;
			var fpsCorrection = 60 / Math.min(Math.max(this.currentFPS, 10), 60);
			var modifier = -Math.sign(this.currentEventDifferenceX) * fpsCorrection * weight * Math.max(0.1, (Math.abs(Math.pow(this.computedWidth, exponent)) - Math.abs(Math.pow(this.currentEventDifferenceX, exponent))) / Math.abs(Math.pow(this.computedWidth, exponent)));
			if (Math.sign(this.currentEventDifferenceX) != Math.sign(this.currentEventDifferenceX + modifier)) {
				this.currentEventDifferenceX = 0;
			} else {
				this.currentEventDifferenceX += modifier;
			}
			
			if (this.currentEventDifferenceX != 0) {
				updatePositions();
				this.currentEventDifferenceXRequestAnimationFrame = requestAnimationFrame(resetScrollAnimation);
			} else if (this.currentEventOffset != 0) {
				if (this.currentEventOffset < 0) {
					this.currentEventDifferenceX += this.computedWidth;
					this.currentEventOffset++;
					updateTags();
					updatePositions();
					this.currentEventDifferenceXRequestAnimationFrame = requestAnimationFrame(resetScrollAnimation);
				} else if (this.currentEventOffset > 0) {
					this.currentEventDifferenceX -= this.computedWidth;
					this.currentEventOffset--;
					updateTags();
					updatePositions();
					this.currentEventDifferenceXRequestAnimationFrame = requestAnimationFrame(resetScrollAnimation);
				}
			} else {
				this.currentEventDifferenceX = 0;
				updatePositions();
			}
		};
		
		const resetScroll = () => {
			cancelAnimationFrame(this.currentEventDifferenceXRequestAnimationFrame);
			this.currentEventDifferenceXRequestAnimationFrame = null;
			resetScrollAnimation();
		};
		
		updateTags();
		updatePositions();
		
		let pointerdown = false;
		var startX = 0;
		this.RemoveLastPointerEvents();
		this.tag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastPointerDown = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			this.ButtonFadeIn();
			this.StartResetTimer();
			pointerdown = true;
			switch (this.orientation) {
				case 'landscape':
				case '180 degrees':
					startX = e.pageX;
					break;
				case '90 degrees':
				case '270 degrees':
					startX = e.pageY;
					break;
			}
			if (this.currentEventOffset != 0) {
				clearTimeout(this.currentEventResetScrollTimeout);
				this.currentEventResetScrollTimeout = setTimeout(resetScroll, 5000);
			} else {
				clearTimeout(this.currentEventResetScrollTimeout);
			}
			cancelAnimationFrame(this.currentEventDifferenceXRequestAnimationFrame);
			this.currentEventDifferenceXRequestAnimationFrame = null;
		});
		this.tag.addEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastPointerMove = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			if (pointerdown) {
				if (this.enableswipepreviousandnext) {
					switch (this.orientation) {
						case 'landscape':
							this.currentEventDifferenceX += e.pageX - startX;
							break;
						case '90 degrees':
							this.currentEventDifferenceX += e.pageY - startX;
							break;
						case '180 degrees':
							this.currentEventDifferenceX += startX - e.pageX;
							break;
						case '270 degrees':
							this.currentEventDifferenceX += startX - e.pageY;
							break;
					}
				}
				if (this.currentEventDifferenceX > this.computedWidth) {
					this.currentEventDifferenceX = this.computedWidth;
				}
				if (this.currentEventDifferenceX < -this.computedWidth) {
					this.currentEventDifferenceX = -this.computedWidth;
				}
				switch (this.orientation) {
					case 'landscape':
					case '180 degrees':
						startX = e.pageX;
						break;
					case '90 degrees':
					case '270 degrees':
						startX = e.pageY;
						break;
				}
				updatePositions();
			}
		});
		this.tag.addEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastPointerUp = e => {
			this.ButtonFadeIn();
			this.StartResetTimer();
			pointerdown = false;
			if (this.currentEventDifferenceX < -100) {
				if (this.currentEventIndex + this.currentEventOffset < this.displayedEvents.length - 1) {
					this.currentEventDifferenceX += this.computedWidth;
					this.currentEventOffset++;
					updateTags();
					updatePositions();
					if (this.currentEventOffset != 0) {
						clearTimeout(this.currentEventResetScrollTimeout);
						this.currentEventResetScrollTimeout = setTimeout(resetScroll, 5000);
					} else {
						clearTimeout(this.currentEventResetScrollTimeout);
					}
				}
			} else if (this.currentEventDifferenceX > 100) {
				if (this.currentEventIndex + this.currentEventOffset > 0) {
					this.currentEventDifferenceX -= this.computedWidth;
					this.currentEventOffset--;
					updateTags();
					updatePositions();
					if (this.currentEventOffset != 0) {
						clearTimeout(this.currentEventResetScrollTimeout);
						this.currentEventResetScrollTimeout = setTimeout(resetScroll, 5000);
					} else {
						clearTimeout(this.currentEventResetScrollTimeout);
					}
				}
			}
			if (this.currentEventDifferenceX != 0 && this.currentEventDifferenceXRequestAnimationFrame == null) {
				const animationSlide = () => {
					if (this.TagDeleted()) {
						return;
					}
					// Weight and exponent and be tweaked to change the swipe behavior. Depending on tablet performance some values might feel more natural
					var weight = 40;
					var exponent = 4;
					var fpsCorrection = 60 / Math.min(Math.max(this.currentFPS, 10), 60);
					var modifier = -Math.sign(this.currentEventDifferenceX) * fpsCorrection * weight * Math.max(0.1, (Math.abs(Math.pow(this.computedWidth, exponent)) - Math.abs(Math.pow(this.currentEventDifferenceX, exponent))) / Math.abs(Math.pow(this.computedWidth, exponent)));
					if (Math.sign(this.currentEventDifferenceX) != Math.sign(this.currentEventDifferenceX + modifier)) {
						this.currentEventDifferenceX = 0;
					} else {
						this.currentEventDifferenceX += modifier;
					}
					
					if (this.currentEventDifferenceX != 0) {
						updatePositions();
						this.currentEventDifferenceXRequestAnimationFrame = requestAnimationFrame(animationSlide);
					} else {
						this.currentEventDifferenceX = 0;
						updatePositions();
					}
				};
				this.currentEventDifferenceXRequestAnimationFrame = requestAnimationFrame(animationSlide);
			}
		});
	}
	async ShowViewByHour() {
		this.HideAll();
		await this.ShowRoomTimeLogo();
		this.ShowViewByButtons();
		this.ShowInstantAndReportButtons();
		this.$viewByHour.classList.add('selected');
		
		this.ShowCurrentEvent(0, 'hour');
		this.lastViewBy = 'hour';
	}
	EventListState(dayOffset) {
		const currentTime = this.player.currentTime;
		var newState = { events: [], currentEventIndex: null, currentOrNextEventIndex: null };
		newState.events = this.allEvents.filter(event => this.showexpiredevents || event.end.isAfter(currentTime));
		// Filter and include only the events that happened on the specified day
		newState.events = newState.events.filter(event => {
			return event.end.isSameOrAfter(currentTime.clone().startOf('day').add(dayOffset, 'day')) &&
			       event.start.isBefore(currentTime.clone().endOf('day').add(dayOffset, 'day'));
		});
		newState.events.forEach((event, eventIndex) => {
			// If the current time is inside of the event then this is the current event
			if (currentTime.isSameOrAfter(event.start) && currentTime.isBefore(event.end)) {
				newState.currentEventIndex = eventIndex;
			}
			if (newState.currentOrNextEventIndex == null && currentTime.isBefore(event.end)) {
				newState.currentOrNextEventIndex = eventIndex;
			}
		});
		if (newState.currentOrNextEventIndex == null && newState.events.length != 0) {
			newState.currentOrNextEventIndex = newState.events.length - 1;
		}
		return newState;
	}
	ShowEventList(dayOffset, viewBy, customStart, customEnd, customTitle, customName) {
		this.$viewEventList.hidden = false;
		this.$eventListScrollArrowUp.hidden = false;
		this.$eventListScrollArrowDown.hidden = false;
		this.ResetEventList();
		$.fadeOut(this.$eventListNoEvent).catch(() => {});
		this.$eventListNoEvent.fadeAnimation.finish();
		this.$eventListNoEvent.hidden = true;
		$.fadeIn(this.$viewEventList).catch(() => {});
		this.$viewEventList.fadeAnimation.finish();
		
		// Filter events by today and set the currentEventIndex
		if (dayOffset == -1) {
			// Show the instant event preview in event list view
			this.displayedEvents = [{ "start": customStart, "end": customEnd, "title": customTitle, "name": customName }];
			this.currentEventIndex = 0;
			this.currentOrNextEventIndex = 0;
		} else {
			var newState = this.EventListState(dayOffset);
			this.displayedEvents = newState.events;
			this.currentEventIndex = newState.currentEventIndex;
			this.currentOrNextEventIndex = newState.currentOrNextEventIndex;
		}
		
		if (dayOffset == 0) {
			this.PostColor(this.currentEventIndex == null);
		}
		
		var columnCount = (this.eventlistcolumnfirst == 'none' ? 0 : 1) + (this.eventlistcolumnsecondary == 'none' ? 0 : 1) + (this.eventlistcolumntertiary == 'none' ? 0 : 1);
		this.$viewEventList.innerHTML = '';
		
		if (this.displayedEvents.length == 0) {
			if (this.eventlistshowimagesfornoevents) {
				Object.assign(this.$eventListNoEvent.style, {
					width: `${this.computedWidth - 100}px`,
					top: `${Math.max(this.$top.offsetHeight, this.$bottom.offsetHeight)}px`,
					bottom: `${Math.max(this.$top.offsetHeight, this.$bottom.offsetHeight)}px`
				});
				this.$eventListNoEvent.hidden = false;
				if (viewBy == '' || this.lastViewBy != viewBy || this.lastViewByEventListDayOffset != dayOffset) {
					$.fadeIn(this.$eventListNoEvent).catch(() => {});
				}
			}
		} else {
			this.displayedEvents.forEach((event, index) => {
				var row = $.li();
				row.appendChild($.i({ class: 'eventlistcurrenteventarrow mdi mdi-play', css: { color: this.eventlistcurrenteventarrowcolor } }));
				for (const column of [this.eventlistcolumnfirst, this.eventlistcolumnsecondary, this.eventlistcolumntertiary]) {
					switch (column) {
						case 'title':
							if (event.title.indexOf('https://') == 0) {
								row.appendChild($.img({ src: event.title, class: 'title-image' }));
							} else {
								row.appendChild($.span({ text: event.title, class: 'title' }));
							}
							break;
						case 'room':
							row.appendChild($.span({ text: event.room, class: 'name' }));
							break;
						case 'name':
							if (event.name.indexOf('https://') == 0) {
								row.appendChild($.img({ src: event.name, class: 'title-image' }));
							} else {
								row.appendChild($.span({ text: event.name, class: 'name' }));
							}
							break;
						case 'time':
							row.appendChild(
								$.span({ class: 'time' },
									$.span({ text: this.player.applyTimezone(event.start).format('h:mm ') }),
									$.span({ text: this.player.applyTimezone(event.start).format('A'), css: { fontSize: '0.7em' } }),
									$.span({ text: ' - ' }),
									$.span({ text: this.player.applyTimezone(event.end).format('h:mm ') }),
									$.span({ text: this.player.applyTimezone(event.end).format('A'), css: { fontSize: '0.7em' } })
								)
							);
							break;
					}
				}
				if (this.currentEventIndex != null && index == this.currentEventIndex) {
					row.classList.add('eventlist-currentevent');
					var viewEventListLeft = this.$viewEventList.offsetLeft;
					switch (this.orientation) {
						case '180 degrees':
							viewEventListLeft = this.computedWidth - (viewEventListLeft + this.$viewEventList.offsetWidth);
							break;
						case '90 degrees':
							viewEventListLeft = this.$viewEventList.offsetTop;
							break;
						case '270 degrees':
							viewEventListLeft = this.computedWidth - (this.$viewEventList.offsetTop + this.$viewEventList.offsetWidth);
							break;
					}
				}
				const style = getComputedStyle(this.$viewEventList);
				Object.assign(row.style, {
					left: `${parseInt(style.paddingLeft, 10)}px`,
					right: `${parseInt(style.paddingRight, 10)}px`,
					top: `${parseInt(style.paddingTop, 10)}px`
				});
				this.$viewEventList.appendChild(row);
			});
		}
		const columnsWidths = [];
		for (var itr = 0; itr < columnCount; ++itr) {
			columnsWidths.push(0);
		}
		for (const $event of this.$viewEventList.querySelectorAll(':scope > li')) {
			for (const [index, $span] of [...$event.querySelectorAll(':scope > span')].entries()) {
				columnsWidths[index] = Math.max(columnsWidths[index], $span.offsetWidth);
			}
		}
		
		for (const $event of this.$viewEventList.querySelectorAll(':scope > li')) {
			for (const [index, $span] of [...$event.querySelectorAll(':scope > span')].entries()) {
				if (!$span.classList.contains('title')) {
					$span.style.width = `${columnsWidths[index] + 30}px`;
				}
			}
		}
		
		let middleTop = this.$middle.offsetTop;
		switch (this.orientation) {
			case '180 degrees':
				middleTop = this.computedHeight - (middleTop + this.$middle.offsetHeight);
				break;
			case '90 degrees':
				middleTop = this.computedHeight - (this.$middle.offsetLeft + this.$middle.offsetHeight);
				break;
			case '270 degrees':
				middleTop = this.$middle.offsetLeft;
				break;
		}
		
		const getTotalHeight = () => {
			const $events = this.$viewEventList.querySelectorAll(':scope > li');
			if ($events.length == 0) {
				return 0;
			} else {
				return $events[$events.length - 1].getBoundingClientRect().bottom - $events[0].getBoundingClientRect().top;
			}
		}
		
		const updatePositions = () => {
			var offsetY = 0;
			const eventsHeight = getTotalHeight();
			var viewHeight = this.$viewEventList.clientHeight;
			for (const $event of this.$viewEventList.querySelectorAll(':scope > li')) {
				const eventListItemTop = offsetY + this.eventlistDifferenceY;
				const opacity = eventListItemTop < 0 ? Math.max(0, 1 - Math.abs(eventListItemTop) / 20) : (eventListItemTop > viewHeight - $event.offsetHeight ? 1 - Math.min(1, (eventListItemTop - (viewHeight - $event.offsetHeight)) / 20) : 1);
				Object.assign($event.style, {
					transform: `translate(0px,${eventListItemTop}px)`,
					opacity: opacity
				});
				offsetY += $event.offsetHeight;
			}
			if (this.eventlistviewenablescrollarrows) {
				if (this.eventlistDifferenceY < 0) {
					$.fadeIn(this.$eventListScrollArrowUp).catch(() => {});
				} else {
					$.fadeOut(this.$eventListScrollArrowUp).catch(() => {});
				}
				if (eventsHeight + this.eventlistDifferenceY > viewHeight) {
					$.fadeIn(this.$eventListScrollArrowDown).catch(() => {});
				} else {
					$.fadeOut(this.$eventListScrollArrowDown).catch(() => {});
				}
			}
		};
		updatePositions();
		
		const resetScroll = () => {
			if (this.TagDeleted()) {
				return;
			}
			var viewHeight = this.$viewEventList.offsetHeight;
			var totalHeight = getTotalHeight();
			var targetY = 0;
			if (this.currentOrNextEventIndex != null) {
				// Find the top of the event text at currentOrNextEventIndex
				for (const [index, $event] of [...this.$viewEventList.querySelectorAll(':scope > li')].entries()) {
					if (this.currentOrNextEventIndex >= 2 && index < this.currentOrNextEventIndex - 2) {
						targetY -= $event.offsetHeight; // + marginBetweenEventListEvents
					}
				}
			}
			targetY = Math.min(0, Math.max(targetY, Math.min(0, -(totalHeight - viewHeight))));
			
			const animationResetEventListView = () => {
				if (this.TagDeleted()) {
					return;
				}
				// Weight and exponent and be tweaked to change the swipe behavior. Depending on tablet performance some values might feel more natural
				var weight = 2;
				var exponent = 4;
				var fpsCorrection = 60 / Math.min(Math.max(this.currentFPS, 10), 60);
				var differenceY = targetY - this.eventlistDifferenceY;
				var modifier = Math.sign(differenceY) * fpsCorrection * weight * Math.max(1, Math.abs(Math.pow(differenceY, exponent)) / Math.abs(Math.pow(viewHeight, exponent)));
				if (differenceY == 0 || Math.sign(differenceY) != Math.sign(differenceY - modifier)) {
					this.eventlistDifferenceY = targetY;
					updatePositions();
				} else {
					this.eventlistDifferenceY += modifier;
					updatePositions();
					this.eventlistResetScrollRequestAnimationFrame = requestAnimationFrame(animationResetEventListView);
				}
			};
			this.eventlistResetScrollRequestAnimationFrame = requestAnimationFrame(animationResetEventListView);
			// Set this to null since this function is called in the timeout
			this.eventlistResetScrollTimeout = null;
		};
		var pointerdown = false;
		var startY = 0;
		this.RemoveLastPointerEvents();
		this.tag.addEventListener('onpointerdown' in window ? 'pointerdown' : 'touchstart', this.tag.lastPointerDown = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			this.ButtonFadeIn();
			this.StartResetTimer();
			pointerdown = true;
			switch (this.orientation) {
				case 'landscape':
				case '180 degrees':
					startY = e.pageY;
					break;
				case '90 degrees':
				case '270 degrees':
					startY = e.pageX;
					break;
			}
			cancelAnimationFrame(this.eventlistDifferenceYRequestAnimationFrame);
			this.eventlistDifferenceYRequestAnimationFrame = null;
			clearTimeout(this.eventlistResetScrollTimeout);
			this.eventlistResetScrollTimeout = null;
			cancelAnimationFrame(this.eventlistResetScrollRequestAnimationFrame);
			this.eventlistResetScrollRequestAnimationFrame = null;
		});
		this.tag.addEventListener('onpointerdown' in window ? 'pointermove' : 'touchmove', this.tag.lastPointerMove = e => {
			if (!('onpointerdown' in window)) {
				e = e.touches[0];
			}
			if (pointerdown) {
				switch (this.orientation) {
					case 'landscape':
						this.eventlistDifferenceY += e.pageY - startY;
						break;
					case '90 degrees':
						this.eventlistDifferenceY += startY - e.pageX;
						break;
					case '180 degrees':
						this.eventlistDifferenceY += startY - e.pageY;
						break;
					case '270 degrees':
						this.eventlistDifferenceY +=  e.pageX - startY;
						break;
				}
				var viewHeight = this.$viewEventList.offsetHeight;
				var totalHeight = getTotalHeight();
				// This might be confusing. The user is allowed to scroll up or down a total of viewHeight beyond the scroll region and then it snaps back within view
				if (this.eventlistDifferenceY > viewHeight) {
					this.eventlistDifferenceY = viewHeight;
				}
				if (this.eventlistDifferenceY < -totalHeight) {
					this.eventlistDifferenceY = -totalHeight;
				}
				switch (this.orientation) {
					case 'landscape':
					case '180 degrees':
						startY = e.pageY;
						break;
					case '90 degrees':
					case '270 degrees':
						startY = e.pageX;
						break;
				}
				updatePositions();
			}
		});
		this.tag.addEventListener('onpointerdown' in window ? 'pointerup' : 'touchend', this.tag.lastPointerUp = e => {
			this.ButtonFadeIn();
			this.StartResetTimer();
			pointerdown = false;
			if (this.eventlistDifferenceY != 0 && this.eventlistDifferenceYRequestAnimationFrame == null) {
				const animationScroll = () => {
					if (this.TagDeleted()) {
						return;
					}
					var totalHeight = getTotalHeight();
					var differenceY = 0;
					var viewHeight = this.$viewEventList.offsetHeight;
					var differenceYBottom = Math.max(0, totalHeight - viewHeight) + this.eventlistDifferenceY;
					if (differenceYBottom < 0) {
						differenceY = differenceYBottom;
					} else if (this.eventlistDifferenceY > 0) {
						differenceY = this.eventlistDifferenceY;
					}
					// Weight and exponent and be tweaked to change the swipe behavior. Depending on tablet performance some values might feel more natural
					var weight = 20;
					var exponent = 4;
					var fpsCorrection = 60 / Math.min(Math.max(this.currentFPS, 10), 60);
					var modifier = -Math.sign(differenceY) * fpsCorrection * weight * Math.max(0.1, (Math.abs(Math.pow(this.computedHeight, exponent)) - Math.abs(Math.pow(differenceY, exponent))) / Math.abs(Math.pow(this.computedHeight, exponent)));
					if (Math.sign(differenceY) != Math.sign(differenceY + modifier)) {
						this.eventlistDifferenceY -= differenceY;
						differenceY = 0;
					} else {
						differenceY += modifier;
						this.eventlistDifferenceY += modifier;
					}
					if (differenceY != 0) {
						updatePositions();
						this.eventlistDifferenceYRequestAnimationFrame = requestAnimationFrame(animationScroll);
					} else {
						updatePositions();
					}
				};
				this.eventlistDifferenceYRequestAnimationFrame = requestAnimationFrame(animationScroll);
			}
			this.eventlistResetScrollTimeout = setTimeout(resetScroll, 5000);
		});
		resetScroll();
		if (viewBy == '' || this.lastViewBy != viewBy || this.lastViewByEventListDayOffset != dayOffset) {
			$.fadeOut(this.$viewEventList).catch(() => {});
			this.$viewEventList.fadeAnimation.finish();
			$.fadeIn(this.$viewEventList).catch(() => {});
		} else {
			$.fadeIn(this.$viewEventList).catch(() => {});
			this.$viewEventList.fadeAnimation.finish();
		}
		this.lastViewByEventListDayOffset = dayOffset;
	}
	async ShowViewByDay() {
		this.HideRoomTimeLogo();
		await this.ShowRoomTimeLogo();
		// Naive delay to allow the height to update
		await new Promise(resolve => setTimeout(resolve, 200));
		this.HideAllButRoomTimeLogo();
		this.ShowViewByButtons();
		this.ShowInstantAndReportButtons();
		this.$viewByDay.classList.add('selected');
		
		this.ShowEventList(0, 'day');
		this.lastViewBy = 'day';
	}
	async ShowViewByWeek() {
		this.HideRoomTimeLogo();
		await this.ShowRoomTimeLogo();
		// Naive delay to allow the height to update
		await new Promise(resolve => setTimeout(resolve, 200));
		this.HideAllButRoomTimeLogo();
		this.ShowInstantAndReportButtons();
		if (this.showcurrentevent || this.enableviewbydaybutton) {
			this.$buttonSpacer.hidden = false;
			this.$instantEvent.hidden = true;
			this.$reportIssue.hidden = true;
			
			this.$back.removeEventListener('click', this.$back.lastClick);
			this.$back.addEventListener('click', this.$back.lastClick = () => {
				this.lastViewBy = '';
				this.ResetView();
			});
			this.$back.hidden = false;
		} else {
			this.$buttonSpacerDayOfTheWeek.hidden = false;
		}
		const currentTime = this.player.currentTime;
		for (const [index, $day] of [...this.$daysOfTheWeek].entries()) {
			if (index == 0) {
				$day.classList.add('selected');
			}
			$day.innerText = currentTime.clone().add(index, 'day').format('dddd');
			$day.removeEventListener('click', $day.lastClick);
			$day.addEventListener('click', $day.lastClick = () => {
				for (const $e of this.$daysOfTheWeek) $e.classList.remove('selected');
				$day.classList.add('selected');
				this.ShowEventList(index, 'week');
			});
		}
		this.ShowEventList(0, 'week');
		if (this.lastViewBy == 'week') {
			for (const $day of this.$daysOfTheWeek) $day.hidden = false;
		} else {
			// TODO: In the future when we drop Android 4.4 use $.fadeIn
			for (const $day of this.$daysOfTheWeek) $day.hidden = false;
		}
		this.lastViewBy = 'week';
	}
	UpdateInstantEventTimes() {
		this.tag.querySelector('.instanteventstarttimehour > div').innerText = this.instantEventStart.format('h');
		this.tag.querySelector('.instanteventstarttimeminute > div').innerText = this.instantEventStart.format('mm');
		this.tag.querySelector('.instanteventstarttimeperiod > div').innerText = this.instantEventStart.format('A');
		this.tag.querySelector('.instanteventendtimehour > div').innerText = this.instantEventEnd.format('h');
		this.tag.querySelector('.instanteventendtimeminute > div').innerText = this.instantEventEnd.format('mm');
		this.tag.querySelector('.instanteventendtimeperiod > div').innerText = this.instantEventEnd.format('A');
	}
	ResetInstantEventEditor() {
		this.instantEventTitle = '';
		this.$viewInstantEventFormNameInput.innerHTML = '';
		const currentTime = this.player.currentTime;
		this.instantEventStart = currentTime.minutes(currentTime.minutes() - currentTime.minutes() % 15).startOf('minute');
		this.instantEventEnd = currentTime.clone().add(1, 'hour');
		this.UpdateInstantEventTimes();
	}
	async ShowInstantEventPIN() {
		this.HideAll();
		this.DefaultActivityDetection();
		await this.ShowRoomTimeLogo();
		this.$viewPin.hidden = false;
		this.$instantEvent.classList.add('selected');
		this.$buttonSpacer.hidden = false;
		this.ShowInstantAndReportButtons();
		this.$back.removeEventListener('click', this.$back.lastClick);
		this.$back.addEventListener('click', this.$back.lastClick = () => {
			this.lastViewBy = '';
			this.ResetView();
		});
		this.$back.hidden = false;
		
		this.$next.removeEventListener('click', this.$next.lastClick);
		this.$next.addEventListener('click', this.$next.lastClick = async () => {
			if (!this.offlineMode) {
				this.$loading.hidden = false;
				try {
					const data = await $.http({
						method: 'POST',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/post/signwaveverifypin.php',
						data: {
							mediaid: this.mediaid,
							pin: this.pin
						},
						dataType: 'json'
					}).promise;
					
					if (this.TagDeleted()) {
						return;
					}
					
					this.$loading.hidden = true;
					if (data.success) {
						this.ResetInstantEventEditor();
						await this.ShowInstantEventEditor();
					} else {
						this.pin = '';
						this.ShowInstantEventPIN();
						this.ConfirmationBox('Error', 'Invalid PIN');
					}
				} catch (exception) {
					if (this.TagDeleted()) {
						return;
					}
					this.offlineMode = null;
					this.TestOnlineMode();
				}
			}
		});
		this.$next.hidden = false;
		this.lastViewBy = '';
	}
	async ShowInstantEventEditor() {
		this.HideAll();
		this.DefaultActivityDetection();
		await this.ShowRoomTimeLogo();
		this.$instantEvent.classList.add('selected');
		this.ShowInstantAndReportButtons();
		this.$buttonSpacer.hidden = false;
		this.$viewInstantEventForm.hidden = false;
		this.$back.removeEventListener('click', this.$back.lastClick);
		this.$back.addEventListener('click', this.$back.lastClick = () => {
			this.lastViewBy = '';
			this.ResetView();
		});
		this.$back.hidden = false;
		this.$next.removeEventListener('click', this.$next.lastClick);
		this.$next.addEventListener('click', this.$next.lastClick = async () => await this.ShowInstantEventConfirmation());
		this.$next.hidden = false;
		for (const $date of this.tag.querySelectorAll('.instanteventstartdate, .instanteventenddate')) {
			var rerender = () => {
				this.tag.querySelector('.instanteventstartdatetext > div').innerText = this.instantEventStart.format('MMM D');
				this.tag.querySelector('.instanteventenddatetext > div').innerText = this.instantEventEnd.format('MMM D');
				if ($date.hasOwnProperty('calendar')) {
					$date.calendar.destroy();
				}
				$date.calendar = new FullCalendar.Calendar($date, {
					headerToolbar: {
						left: 'prev customToday',
						center: 'title',
						right: 'next'
					},
					selectable: false,
					editable: false,
					height: $date.offsetHeight,
					aspectRatio: $date.offsetWidth / $date.offsetHeight,
					customButtons: {
						customToday: {
							text: 'Today',
							click: () => {
								const currentTime = this.player.currentTime;
								if ($date.classList.contains('instanteventstartdate')) {
									this.instantEventStart.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
									if (!this.instanteventcustomenddate) {
										this.instantEventEnd.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
									}
									rerender();
									$date.calendar.gotoDate(this.instantEventStart);
								} else {
									this.instantEventEnd.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
									rerender();
									$date.calendar.gotoDate(this.instantEventEnd);
								}
							}
						}
					},
					dateClick: ({ date }) => {
						date = FullCalendarMoment.toMoment(date, $date.calendar);
						var currentTime = date.clone();
						if ($date.classList.contains('instanteventstartdate')) {
							this.instantEventStart.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
							if (!this.instanteventcustomenddate) {
								this.instantEventEnd.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
							}
							rerender();
							$date.calendar.gotoDate(this.instantEventStart);
						} else {
							this.instantEventEnd.year(currentTime.year()).month(currentTime.month()).date(currentTime.date());
							rerender();
							$date.calendar.gotoDate(this.instantEventEnd);
						}
					}/*,
					dayRender: (date, cell) => {
						date = date.clone().add(1, 'day');
						if ($date.classList.contains('instanteventstartdate')) {
							if (this.instantEventStart.isSame(date, 'day')) {
								cell.style.background = 'radial-gradient(circle farthest-corner at 50% 50%, #7273b4 60%, #575aa6)';
							}
						} else {
							if (this.instantEventEnd.isSame(date, 'day')) {
								cell.style.background = 'radial-gradient(circle farthest-corner at 50% 50%, #7273b4 60%, #575aa6)';
							}
						}
					}*/
				});
				$date.calendar.render();
			};
			
			this.tag.querySelector('.instanteventstartdatetext > div').innerText = this.instantEventStart.format('MMM D');
			this.tag.querySelector('.instanteventenddatetext > div').innerText = this.instantEventEnd.format('MMM D');
			
			if ($date.classList.contains('instanteventstartdate') && this.instanteventcustomstartdate ||
				$date.classList.contains('instanteventenddate') && this.instanteventcustomenddate) {
				rerender();
			} else if ($date.classList.contains('instanteventstartdate') && !this.instanteventcustomstartdate) {
				this.tag.querySelector('.instanteventstartdate').innerHTML = '';
				this.tag.querySelector('.instanteventstartdate').classList.add('today');
				this.tag.querySelector('.instanteventstartdate').append(
					$.div({ text: 'Today' }),
					$.div({ text: 'Custom start date has been disabled.' })
				);
			} else if ($date.classList.contains('instanteventenddate') && !this.instanteventcustomenddate) {
				this.tag.querySelector('.instanteventenddate').innerHTML = '';
				this.tag.querySelector('.instanteventenddate').classList.add('today');
				this.tag.querySelector('.instanteventenddate').append(
					$.div({ text: !this.instanteventcustomstartdate ? 'Today' : '' }),
					$.div({ text: 'Custom end date has been disabled.' })
				);
			}
		}
		this.$viewInstantEventFormTitles.innerHTML = '';
		for (const title of this.instanteventpredefinedtitles) {
			if (title != '') {
				this.$viewInstantEventFormTitles.appendChild(
					$.div({ text: title, class: 'predefinedtitle', click: ({ target: $title }) => {
						for (const $predefinedTitle of $title.parentNode.children) $predefinedTitle.classList.remove('selected');
						$title.classList.add('selected');
						this.instantEventTitle = $title.innerText;
					} })
				);
			}
		}
		if (this.instanteventcustomtitle) {
			this.$viewInstantEventFormTitles.appendChild(
				$.div({ text: this.instantEventTitle == '' ? 'Custom' : this.instantEventTitle, class: 'predefinedtitle', click: async ({ target: $title }) => {
					for (const $predefinedTitle of $title.parentNode.children) $predefinedTitle.classList.remove('selected');
					$title.classList.add('selected');
					var customButton = $title;
					var value = await this.ShowKeyboard(this.instantEventTitle, true);
					this.instantEventTitle = value;
					customButton.innerText = this.instantEventTitle == '' ? 'Custom' : this.instantEventTitle;
				} })
			);
		}
		const $firstPredefinedTitle = this.$viewInstantEventFormTitles.querySelector(':scope > .predefinedtitle');
		if ($firstPredefinedTitle) {
			$firstPredefinedTitle.classList.add('selected');
			this.instantEventTitle = $firstPredefinedTitle.innerText;
		}
		this.$viewInstantEventFormRooms.innerHTML = '';
		for (const room of this.rooms) {
			let $room;
			this.$viewInstantEventFormRooms.appendChild(
				$room = $.div({ text: this.GetRoomName(room), class: 'view-instantevent-form-room', click: ({ target: $room }) => {
					$room.classList.toggle('selected');
				} })
			);
			$room.room = room;
		}
		this.$viewInstantEventFormRooms.hidden = this.rooms.length <= 1;
		if (this.rooms.length == 1) {
			this.tag.querySelector('.view-instantevent-form-room').classList.add('selected');
		}
		this.$viewInstantEventFormTitles.style.margin = this.rooms.length > 1 ? '20px 20px 0px 0' : '20px 20px 20px 0';
		this.lastViewBy = '';
	}
	async ShowInstantEventConfirmation() {
		this.HideAll();
		this.DefaultActivityDetection();
		await this.ShowRoomTimeLogo();
		this.$instantEvent.classList.add('selected');
		this.ShowInstantAndReportButtons();
		this.$buttonSpacer.hidden = false;
		this.$back.removeEventListener('click', this.$back.lastClick);
		this.$back.addEventListener('click', this.$back.lastClick = async () => await this.ShowInstantEventEditor());
		this.$back.hidden = false;
		this.$confirm.removeEventListener('click', this.$confirm.lastClick);
		this.$confirm.addEventListener('click', this.$confirm.lastClick = async () => {
			const instantEventRooms = [];
			for (const $room of this.tag.querySelectorAll('.view-instantevent-form-room')) {
				if ($room.classList.contains('selected')) {
					instantEventRooms.push($room.room);
				}
			}
			if (!this.offlineMode) {
				this.$loading.hidden = false;
				if (!this.container || this.container.type == 'Preview') {
					this.ConfirmationBox('Instant Event', 'Events are not created in editor preview.');
					this.ResetView();
				} else {
					try {
						const data = await $.http({
							method: 'POST',
							url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/post/signwaveinstantevent.php',
							data: {
								mediaid: this.mediaid,
								pin: this.pin,
								instanteventstart: this.instantEventStart.unix(),
								instanteventend: this.instantEventEnd.unix(),
								instanteventtitle: this.instantEventTitle,
								instanteventname: this.instantEventName,
								rooms: JSON.stringify(instantEventRooms.map(room => room.id))
							},
							dataType: 'json'
						}).promise;
					
						if (this.TagDeleted()) {
							return;
						}
					
						this.$loading.hidden = true;
						let results = '';
						if (data.success) {
							for (const roomResult of data.results) {
								const room = instantEventRooms.find(room => room.id == roomResult.roomid);
								if (roomResult.hasOwnProperty('internal')) {
									this.internalInstantEvents.push({ 'start': this.instantEventStart, 'end': this.instantEventEnd, 'title': this.instantEventTitle, 'name': this.instantEventName, 'room': room.name, internal: roomResult.internal });
									if (roomResult.internal) {
										results += `The event was added into room '${room.name}' internally on the device because the provider does not support adding events from this interface. Restarting this device will remove the event. `;
									} else {
										results += `The event was added into room '${room.name}'. `;
									}
								}
								if (roomResult.hasOwnProperty('error')) {
									if (roomResult.hasOwnProperty('internal')) {
										// Event was still added internally
										results += `The event was added into room '${room.name}' internally on the device due to error: "${roomResult.error}". Restarting this device will remove the event. `;
									} else {
										// Event was not added to the doorcard
										results += `The room "${room.name}" returned an error: "${roomResult.error}". No event was created. `;
									}
								}
							}
							const currentTimeStartOfDay = this.player.currentTime.startOf('day');
							this.internalInstantEvents = this.internalInstantEvents.filter(event => currentTimeStartOfDay.isBefore(event.end));
						} else {
							results += `Error creating event(s): ${data.error}`;
						}
						this.ConfirmationBox('Instant Event', results);
						this.ResetView();
					} catch (exception) {
						if (this.TagDeleted()) {
							return;
						}
						this.offlineMode = null;
						this.TestOnlineMode();
					}
				}
			}
		});
		this.$confirm.hidden = false;
		this.$cancel.removeEventListener('click', this.$cancel.lastClick);
		this.$cancel.addEventListener('click', this.$cancel.lastClick = () => this.ResetView());
		this.$cancel.hidden = false;
		if (this.showcurrentevent) {
			this.ShowCurrentEvent(-1, '', this.instantEventStart, this.instantEventEnd, this.instantEventTitle, this.instantEventName);
		} else {
			this.ShowEventList(-1, '', this.instantEventStart, this.instantEventEnd, this.instantEventTitle, this.instantEventName);
		}
		this.lastViewBy = '';
	}
	async ShowReportIssuePIN() {
		this.HideAll();
		this.DefaultActivityDetection();
		await this.ShowRoomTimeLogo();
		this.$reportIssue.classList.add('selected');
		this.ShowInstantAndReportButtons();
		this.$buttonSpacer.hidden = false;
		this.$viewPin.hidden = false;
		this.$back.removeEventListener('click', this.$back.lastClick);
		this.$back.addEventListener('click', this.$back.lastClick = () => {
			this.lastViewBy = '';
			this.ResetView();
		});
		this.$back.hidden = false;
		this.$next.removeEventListener('click', this.$next.lastClick);
		this.$next.addEventListener('click', this.$next.lastClick = async () => {
			if (!this.offlineMode) {
				this.$loading.hidden = false;
				try {
					const data = await $.http({
						method: 'POST',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/post/signwaveverifypin.php',
						data: {
							mediaid: this.mediaid,
							pin: this.pin
						},
						dataType: 'json'
					}).promise;
					
					if (this.TagDeleted()) {
						return;
					}
					this.$loading.hidden = true;
					if (data.success) {
						await this.ShowReportIssueConfirmation();
					} else {
						this.pin = '';
						await this.ShowReportIssuePIN();
						this.ConfirmationBox('Error', 'Invalid PIN');
					}
				} catch (exception) {
					if (this.TagDeleted()) {
						return;
					}
					this.offlineMode = null;
					this.TestOnlineMode();
				}
			}
		});
		this.$next.hidden = false;
		
		this.lastViewBy = '';
	}
	async ShowReportIssueConfirmation(){
		this.HideAll();
		this.DefaultActivityDetection();
		await this.ShowRoomTimeLogo();
		this.$reportIssue.classList.add('selected');
		this.ShowInstantAndReportButtons();
		this.$buttonSpacer.hidden = false;
		this.$viewReportIssueConfirmation.hidden = false;
		this.$confirm.removeEventListener('click', this.$confirm.lastClick);
		this.$confirm.addEventListener('click', this.$confirm.lastClick = async () => {
			if (!this.offlineMode) {
				this.$loading.hidden = false;
				try {
					const data = await $.http({
						method: 'POST',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/post/signwavereportanissue.php',
						data: {
							key: this.player.settings.data.key,
							mediaid: this.mediaid,
							pin: this.pin,
							rooms: JSON.stringify(this.rooms)
						},
						dataType: 'json'
					}).promise;
					if (this.TagDeleted()) {
						return;
					}
					this.$loading.hidden = true;
					if (data.success) {
						this.ResetView();
					} else {
						this.ResetView();
					}
				} catch (exception) {
					if (this.TagDeleted()) {
						return;
					}
					this.offlineMode = null;
					this.TestOnlineMode();
				}
			}
		});
		this.$confirm.hidden = false;
		this.$cancel.removeEventListener('click', this.$cancel.lastClick);
		this.$cancel.addEventListener('click', this.$cancel.lastClick = () => this.ResetView());
		this.$cancel.hidden = false;
		this.lastViewBy = '';
	}
	GetRoomName(room) {
		if (this.overridereplacementnames.hasOwnProperty(room.id)) {
			return this.overridereplacementnames[room.id];
		}
		return room.replacement == '' ? room.name : room.replacement;
	}
	async SetRoomText() {
		var roomText = '';
		var roomNames = {};
		this.rooms.forEach((room, roomItr) => {
			var roomName = this.GetRoomName(room);
			if (!roomNames.hasOwnProperty(roomName)) {
				if (roomText.length != '') {
					roomText += ', ';
				}
				roomText += roomName;
				roomNames[roomName] = true;
			}
		});
		this.$room.innerText = roomText;
		await this.ShowRoomTimeLogo();
	}
	async LoadEvents() {
		if (this.TagDeleted()) {
			return;
		}
		if (this.offlineMode === false) {
			var cacheData = await new Promise(async resolve => {
				try {
					const data = await $.http({
						method: 'GET',
						url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/providerevents.php',
						data: {
							widget: 'signwave',
							mediaid: this.mediaid,
							roomids: JSON.stringify(this.hasOwnProperty('roomids') ? this.roomids : [])
						},
						dataType: 'json'
					}).promise;
					if (data.success) {
						resolve({ events: data.events, rooms: data.rooms });
					} else {
						console.log('Error getting provider events.');
						resolve(null);
					}
				} catch (exception) {
					console.log('Signwave LoadEvents providerEvents error');
					this.offlineMode = null;
					this.TestOnlineMode();
					resolve(null);
				}
			});
			if (this.TagDeleted()) {
				return;
			}
			
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			
			Object.assign(this, cacheData);
			
			this.newAllEvents = this.events.map(event => {
				if (event.hasOwnProperty('allday') && event.allday) {
					event.start = moment.utc(event.start).local().startOf('day');
					event.end = moment.utc(event.start).local().endOf('day');
				} else {
					event.start = moment(event.start);
					event.end = moment(event.end);
				}
				return event;
			}).sort((a, b) => a.start.unix() - b.start.unix());
			// At this point the events we got include any non-internal events so we can remove the internal events
			const currentTimeStartOfDay = this.player.currentTime.startOf('day');
			this.internalInstantEvents = this.internalInstantEvents.filter(event => event.internal && currentTimeStartOfDay.isBefore(event.end));
			
			await this.SetRoomText();
			
			this.ResetView();
		}
	}
	// This will test the connection and get the latest information for the widget. If it fails then it'll go into offline mode and then try to reconnect in 5 minutes
	async TestOnlineMode() {
		if (this.TagDeleted()) {
			return;
		}
		
		var cacheData = await new Promise(async resolve => {
			try {
				await $.http({
					method: 'GET',
					url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/testonline.php',
					data: {
						random: Math.random()
					},
					dataType: 'json'
				}).promise;
				if (this.TagDeleted()) {
					return;
				}
				if (this.offlineMode == null || this.offlineMode) {
					this.offlineMode = false;
					
					if (this.container) {
						// TODO: Data driven widgetData
						var imagesToCache = [
							this.showlogo ? this.GetOriginalAttributeValue('logoimage') : null,
							this.usedefaultbackgroundimage ? this.GetOriginalAttributeValue('defaultbackgroundimage') : null,
							this.showcurrentevent && this.showimagefornocurrentevent ? this.GetOriginalAttributeValue('nocurrenteventimage') : null,
							(this.enableviewbydaybutton || this.enableviewbyweekbutton) && this.eventlistshowimagesfornoevents ? this.GetOriginalAttributeValue('eventlistnoeventsimage') : null,
							this.overridebackgroundimage ? this.widgetData.signwave.backgroundimage : null,
						];
						const imageCacheURLs = await Promise.all(imagesToCache.map(image => this.container.cacheMediaObject(image)));
						if (imageCacheURLs[0]) { this.logoimage = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('logoimage'))); this.logoimage.cdnurl = imageCacheURLs[0]; }
						if (imageCacheURLs[1]) { this.defaultbackgroundimage = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('defaultbackgroundimage'))); this.defaultbackgroundimage.cdnurl = imageCacheURLs[1]; }
						if (imageCacheURLs[2]) { this.nocurrenteventimage = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('nocurrenteventimage'))); this.nocurrenteventimage.cdnurl = imageCacheURLs[2]; }
						if (imageCacheURLs[3]) { this.eventlistnoeventsimage = JSON.parse(JSON.stringify(this.GetOriginalAttributeValue('eventlistnoeventsimage'))); this.eventlistnoeventsimage.cdnurl = imageCacheURLs[3]; }
						if (imageCacheURLs[4]) { this.backgroundimage = JSON.parse(JSON.stringify(this.widgetData.signwave.backgroundimage)); this.backgroundimage.cdnurl = imageCacheURLs[4]; }
					}
					resolve(
					{
						logoimage: this.logoimage,
						defaultbackgroundimage: this.defaultbackgroundimage,
						nocurrenteventimage: this.nocurrenteventimage,
						eventlistnoeventsimage: this.eventlistnoeventsimage,
						backgroundimage: this.backgroundimage
					});
				} else {
					resolve({});
				}
			} catch (exception) {
				// Offline mode
				if (this.offlineMode == null || !this.offlineMode) {
					console.log('Signwave going into offline mode');
					this.offlineMode = true;
				}
				// Try again in 5 minutes
				clearTimeout(this.onlineModeTimer);
				this.onlineModeTimer = setTimeout(this.TestOnlineMode.bind(this), 5 * 60 * 1000);
				resolve({});
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		
		if (this.widgetData.hasOwnProperty('signwave')) {
			Object.assign(cacheData,
			{
				roomids: this.roomids,
				overridebackgroundimage: this.overridebackgroundimage,
				backgroundimagecoverfill: this.backgroundimagecoverfill,
				overridereplacementnames: this.overridereplacementnames
			});
		}
		
		cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
		
		Object.assign(this, cacheData);
		
		if (!this.container || this.container.type == 'Preview') {
			this.rooms = [{ name: 'Sample Room 1', replacement: '' }];
			this.events = [];
			var allRooms = this.rooms.map(room => room.name);
			var uniqueId = 0;
			const currentTime = this.player.currentTime;
			for (var roomItr = 0; roomItr < 2; ++roomItr) {
				for (var dayItr = 0; dayItr < 5; ++dayItr) {
					for (var hourItr = 0; hourItr < 11; ++hourItr) {
						this.events.push(
						{
							start: currentTime.clone().add(dayItr, 'day').add(hourItr * 2 + roomItr, 'hour').valueOf(),
							end: currentTime.clone().add(dayItr, 'day').add(hourItr * 2 + roomItr, 'hour').add(1, 'hour').valueOf(),
							room: allRooms.length == 0 ? 'Room Sample ' + uniqueId : allRooms[roomItr % allRooms.length],
							title: 'Sample Title ' + (uniqueId++),
							name: 'John Smith'
						});
					}
				}
			}
		}
		
		this.newAllEvents = this.events.map(event => {
			if (event.hasOwnProperty('allday') && event.allday) {
				event.start = moment.utc(event.start).startOf('day');
				event.end = moment.utc(event.end).endOf('day');
			} else {
				event.start = moment(event.start);
				event.end = moment(event.end);
			}
			return event;
		}).sort((a, b) => a.start.unix() - b.start.unix());
		
		// Clear all timers
		this.ClearTimers();
		
		this.tag.innerHTML = '';
		
		this.tag.classList.add('widget-signwave');
		this.tag.style.backgroundColor = this.background;
		this.tag.appendChild(
			$.div({ css: {
				width: `${this.computedWidth}px`,
				height: `${this.computedHeight}px`
			} },
				$.div({ class: 'backgroundimage' }),
				this.$currentEvent = $.div({ class: 'currentevent' },
					$.div({ class: 'previousevent-firstline' }),
					$.div({ class: 'previousevent-secondline' }),
					$.div({ class: 'previousevent-thirdline' }),
					this.$currentEventNoEvent = $.div({ class: 'currentevent-noevent' }),
					$.div({ class: 'currentevent-firstline' }),
					$.div({ class: 'currentevent-secondline' }),
					$.div({ class: 'currentevent-thirdline' }),
					$.div({ class: 'nextevent-firstline' }),
					$.div({ class: 'nextevent-secondline' }),
					$.div({ class: 'nextevent-thirdline' }),
					this.$currentUpcomingEvent = $.div({ text: '(Upcoming Event)', class: 'currentevent-upcomingevent', css: {
						fontFamily: `'${this.currenteventfirstlinefont}'`,
						color: this.currenteventfirstlinecolor,
						textAlign: this.currenteventalignment
					} })
				),
				this.$previousEventArrow = $.i({ class: 'previouseventarrow mdi mdi-play mdi-rotate-180', css: { color: this.swipepreviousandnextarrowcolor } }),
				this.$nextEventArrow = $.i({ class: 'nexteventarrow mdi mdi-play', css: { color: this.swipepreviousandnextarrowcolor } }),
				this.$eventListNoEvent = $.div({ class: 'eventlist-noevent' }),
				this.$loading = $.div({ class: 'loading' },
					$.i({ class: 'loading-spinner' })
				),
				this.$top = $.div({ class: 'top' },
					this.$room = $.div({ class: 'top-room', css: {
						fontFamily: `'${this.roomfont}'`,
						fontSize: `${this.roomfontsize}px`,
						color: this.roomcolor
					} }),
					this.$time = $.div({ class: 'top-time', css: {
						fontFamily: `'${this.timefont}'`,
						fontSize: `${this.timefontsize}px`,
						color: this.timecolor
					} }),
					this.$logo = $.div({ class: 'top-logo' })
				),
				this.$middle = $.div({ class: 'middle' },
					this.$eventListScrollArrowUp = $.i({ class: 'eventlistscrollarrowup mdi mdi-play mdi-rotate-270', css: { color: this.eventlistscrollarrowcolor } }),
					this.$viewEventList = $.ul({ class: 'view-eventlist', css: {
						fontFamily: `'${this.eventlistfont}'`,
						fontSize: `${this.eventlistfontsize}px`,
						color: this.eventlistcolor
					}, hidden: true }),
					this.$eventListScrollArrowDown = $.i({ class: 'eventlistscrollarrowdown mdi mdi-play mdi-rotate-90', css: { color: this.eventlistscrollarrowcolor } }),
					this.$viewPin = $.div({ class: 'view-pin', hidden: true },
						$.div({},
							$.div({ class: 'numpaddisplay' }),
							$.div({},
								$.button({ text: '1', class: 'numpad' }),
								$.button({ text: '2', class: 'numpad' }),
								$.button({ text: '3', class: 'numpad' })
							),
							$.div({},
								$.button({ text: '4', class: 'numpad' }),
								$.button({ text: '5', class: 'numpad' }),
								$.button({ text: '6', class: 'numpad' })
							),
							$.div({},
								$.button({ text: '7', class: 'numpad' }),
								$.button({ text: '8', class: 'numpad' }),
								$.button({ text: '9', class: 'numpad' })
							),
							$.div({},
								$.button({ class: 'numpadempty' }),
								$.button({ text: '0', class: 'numpad' }),
								$.button({ class: 'numpadback' },
									$.i({ class: 'mdi mdi-keyboard-backspace' })
								)
							)
						)
					),
					this.$viewInstantEvent = $.div({ class: 'view-instantevent', hidden: true }),
					this.$viewInstantEventForm = $.div({ class: 'view-instantevent-form', hidden: true },
						$.div({},
							$.div({},
								$.div({},
									$.div({ text: 'Start', class: 'calendarheader', css: { fontFamily: `'${this.instanteventfont}'` } }),
									$.div({ class: 'instanteventstartdate', css: { fontFamily: `'${this.instanteventfont}'` } }),
									$.div({ class: 'instanteventstarttime', css: { fontFamily: `'${this.instanteventfont}'` } },
										$.div({ class: 'instanteventstartdatetext' },
											$.i(),
											$.div(),
											$.i()
										),
										$.div({ class: 'instanteventstarttimehour' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: '12' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										),
										$.div({ class: 'instanteventseparator' },
											$.i(),
											$.div({ text: ':' }),
											$.i()
										),
										$.div({ class: 'instanteventstarttimeminute' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: '00' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										),
										$.div({ class: 'instanteventstarttimeperiod' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: 'AM' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										)
									)
								),
								$.div({},
									$.div({ text: 'End', class: 'calendarheader', css: { fontFamily: `'${this.instanteventfont}'` } }),
									$.div({ class: 'instanteventenddate', css: { fontFamily: `'${this.instanteventfont}'` } }),
									$.div({ class: 'instanteventendtime', css: { fontFamily: `'${this.instanteventfont}'` } },
										$.div({ class: 'instanteventenddatetext' },
											$.i(),
											$.div(),
											$.i()
										),
										$.div({ class: 'instanteventendtimehour' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: '12' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										),
										$.div({ class: 'instanteventseparator' },
											$.i(),
											$.div({ text: ':' }),
											$.i()
										),
										$.div({ class: 'instanteventendtimeminute' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: '00' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										),
										$.div({ class: 'instanteventendtimeperiod' },
											$.i({ class: 'mdi mdi-play mdi-rotate-270 up' }),
											$.div({ text: 'AM' }),
											$.i({ class: 'mdi mdi-play mdi-rotate-90 down' })
										)
									)
								)
							),
							$.div({ class: 'view-instantevent-form-name', hidden: !this.intanteventrequirename },
								$.div({ text: 'Name', css: { fontFamily: `'${this.instanteventfont}'` } }),
								this.$viewInstantEventFormNameInput = $.div({ class: 'view-instantevent-form-name-input', css: { fontFamily: `'${this.instanteventfont}'` } })
							)
						),
						$.div({ css: {
							display: 'flex',
							flexFlow: 'column nowrap'
						} },
							this.$viewInstantEventFormTitles = $.div({ class: 'view-instantevent-form-titles' },
								$.div({ text: 'Title', css: { fontFamily: `'${this.instanteventfont}'` } })
							),
							this.$viewInstantEventFormRooms = $.div({ class: 'view-instantevent-form-rooms' },
								$.div({ text: 'Rooms', css: { fontFamily: `'${this.instanteventfont}'` } })
							)
						)
					),
					this.$viewInstantEventConfirmation = $.div({ class: 'view-instantevent-confirmation', hidden: true }),
					this.$viewReportIssueConfirmation = $.div({ class: 'view-reportissue-confirmation', hidden: true },
						this.$viewReportIssueConfirmationText = $.div({ class: 'view-reportissue-confirmation-text', css: {
							fontFamily: `'${this.reportanissueconfirmationfont}'`,
							fontSize: `${this.reportanissueconfirmationfontsize}px`,
							color: this.reportanissueconfirmationcolor
						} })
					),
					this.$keyboard = $.div({ class: 'keyboard', hidden: true },
						$.div({ class: 'keyboard-input' },
							this.$keyboardInputValue = $.span({ class: 'keyboard-input-value' }),
							$.span({ text: '|', class: 'keyboard-input-cursor' })
						)
					)
				),
				this.$bottom = $.div({ class: 'bottom' },
					this.$viewByHour = $.button({ text: 'View By Hour', class: 'viewbyhour' }),
					this.$viewByDay = $.button({ text: 'View By Day', class: 'viewbyday' }),
					this.$viewByWeek = $.button({ text: 'View By Week', class: 'viewbyweek' }),
					this.$back = $.button({ text: 'Back', class: 'back' }),
					this.$next = $.button({ text: 'Next', class: 'next' }),
					this.$confirm = $.button({ text: 'Confirm', class: 'confirm' }),
					this.$cancel = $.button({ text: 'Cancel', class: 'cancel' }),
					this.$buttonSpacer = $.span({ class: 'button-spacer' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					$.button({ class: 'dayoftheweek' }),
					this.$buttonSpacerDayOfTheWeek = $.span({ class: 'button-spacer-dayoftheweek' }),
					this.$instantEvent = $.button({ text: 'Instant Event', class: 'instantevent' }),
					this.$reportIssue = $.button({ text: 'Report an Issue', class: 'reportissue' })
				)
			)
		);
		this.$daysOfTheWeek = this.tag.querySelectorAll('.dayoftheweek');
		for (const $day of this.$daysOfTheWeek) $day.lastClick = null;
		this.$back.lastClick = null;
		this.$next.lastClick = null;
		this.$confirm.lastClick = null;
		this.$cancel.lastClick = null;
		
		for (const $e of this.tag.querySelectorAll('.previousevent-firstline, .currentevent-firstline, .nextevent-firstline')) {
			Object.assign($e.style, {
				fontFamily: `'${this.currenteventfirstlinefont}'`,
				fontSize: `${this.currenteventfirstlinefontsize}px`,
				color: this.currenteventfirstlinecolor,
				textAlign: this.currenteventalignment,
			});
		}
		
		for (const $e of this.tag.querySelectorAll('.previousevent-secondline, .currentevent-secondline, .nextevent-secondline')) {
			Object.assign($e.style, {
				fontFamily: `'${this.currenteventsecondlinefont}'`,
				fontSize: `${this.currenteventsecondlinefontsize}px`,
				color: this.currenteventsecondlinecolor,
				textAlign: this.currenteventalignment,
			});
		}
		
		for (const $e of this.tag.querySelectorAll('.previousevent-thirdline, .currentevent-thirdline, .nextevent-thirdline')) {
			Object.assign($e.style, {
				fontFamily: `'${this.currenteventthirdlinefont}'`,
				fontSize: `${this.currenteventthirdlinefontsize}px`,
				color: this.currenteventthirdlinecolor,
				textAlign: this.currenteventalignment,
			});
		}
		
		for (const $e of this.tag.querySelectorAll('.instanteventstarttime > div > i, .instanteventendtime > div > i')) {
			$e.style.color = this.instanteventarrowcolor;
		}
		
		if (this.orientation == '90 degrees') {
			this.tag.querySelector(':scope > div').style.transform = `translate(${this.computedHeight}px, 0px) rotate(90deg)`;
		} else if (this.orientation == '180 degrees') {
			this.tag.querySelector(':scope > div').style.transform = `translate(${this.computedWidth}px,${this.computedHeight}px) rotate(180deg)`;
		} else if (this.orientation == '270 degrees') {
			this.tag.querySelector(':scope > div').style.transform = `translate(0,${this.computedWidth}px) rotate(270deg)`;
		}
		
		// Set all the layout properties
		if (this.overridebackgroundimage && this.backgroundimage) {
			this.tag.querySelector('.backgroundimage').style.backgroundImage = `url('${this.backgroundimage.cdnurl}')`;
			this.tag.querySelector('.backgroundimage').style.backgroundSize = this.backgroundimagecoverfill ? 'cover' : '100% 100%';
		} else if (this.usedefaultbackgroundimage && this.defaultbackgroundimage) {
			this.tag.querySelector('.backgroundimage').style.backgroundImage = `url('${this.defaultbackgroundimage.cdnurl}')`;
			this.tag.querySelector('.backgroundimage').style.backgroundSize = this.defaultbackgroundimagecoverfill ? 'cover' : '100% 100%';
		}
		
		for (const $numpad of this.tag.querySelectorAll('.numpad')) {
			$numpad.addEventListener('click', ({ target: $button }) => {
				if (this.pin.length < 10) {
					this.pin += $button.innerText;
					this.UpdatePin();
				}
			});
		}
		this.tag.querySelector('.numpadback').addEventListener('click', () => {
			this.pin = this.pin.slice(0, -1);
			this.UpdatePin();
		});
		
		this.$viewByHour.addEventListener('click', async () => await this.ShowViewByHour());
		this.$viewByDay.addEventListener('click', async () => await this.ShowViewByDay());
		this.$viewByWeek.addEventListener('click', async () => await this.ShowViewByWeek());
		
		this.$instantEvent.addEventListener('click', async () => {
			if (this.instanteventrequirepin) {
				this.pin = '';
				this.ShowInstantEventPIN();
			} else {
				this.ResetInstantEventEditor();
				await await this.ShowInstantEventEditor();
			}
		});
		this.$reportIssue.addEventListener('click', async () => {
			if (this.reportanissuerequirepin) {
				this.pin = '';
				await this.ShowReportIssuePIN();
			} else {
				this.pin = '';
				await this.ShowReportIssueConfirmation();
			}
		});
		var blinkState = true;
		const updateTime = () => {
			if (this.TagDeleted()) {
				return;
			}
			blinkState = !blinkState;
			this.$time.innerHTML = '';
			const currentTime = this.player.currentTime;
			this.$time.append(
				$.span({ text: currentTime.format('h') }),
				$.span({ text: ':', css: { opacity: blinkState ? 1 : 0 } }),
				$.span({ text: currentTime.format('mm') }),
				$.span({ text: ' ' }),
				$.span({ text: currentTime.format('A'), css: { fontSize: '0.7em' } })
			);
		};
		updateTime();
		this.timeInterval = setInterval(updateTime, 800);
		
		// Set whether the buttons are on the top or bottom
		if (this.buttonlocation == 'top') {
			this.tag.querySelector(':scope > div').style.flexDirection = 'column-reverse';
			this.$bottom.style.background = 'linear-gradient(to top, rgba(0,0,0,0) 20%, rgba(0,0,0,0.3) 70%, rgba(0,0,0,0.8))';
		} else {
			this.tag.style.flexDirection = 'column';
			this.$bottom.style.background = 'linear-gradient(to bottom, rgba(0,0,0,0) 20%, rgba(0,0,0,0.3) 70%, rgba(0,0,0,0.8))';
		}
		
		if (this.showcurrentevent && this.showimagefornocurrentevent && this.nocurrenteventimage) {
			this.$currentEventNoEvent.style.backgroundImage = `url('${this.nocurrenteventimage.cdnurl}')`;
			switch (this.nocurrenteventimagefillstyle) {
				case 'fitinsidecurrenteventregion':
					// Do nothing
					break;
				case 'fillcurrenteventregionwithaspectratio':
					this.$currentEventNoEvent.style.backgroundSize = 'contain';
					break;
				case 'fillfullscreen':
					this.$currentEventNoEvent.style.backgroundSize = '100% 100%';
					break;
				case 'fillfullscreenwithaspectratio':
					this.$currentEventNoEvent.style.backgroundSize = 'contain';
					break;
				case 'fillfullscreenwithcover':
					this.$currentEventNoEvent.style.backgroundSize = 'cover';
					break;
			}
		}
		
		if ((this.enableviewbydaybutton || this.enableviewbyweekbutton) && this.eventlistshowimagesfornoevents && this.eventlistnoeventsimage) {
			this.$eventListNoEvent.style.backgroundImage = `url('${this.eventlistnoeventsimage.cdnurl}')`;
			this.$eventListNoEvent.style.backgroundSize = this.eventlistnoeventsimagefill ? 'contain' : '';
		}
		
		// Increment and decrement start and end hour
		for (const $updown of this.tag.querySelectorAll('.instanteventstarttimehour, .instanteventendtimehour')) {
			var up = $updown.querySelector(':scope > .up');
			var down = $updown.querySelector(':scope > .down');
			up.addEventListener('click', () => {
				const time = $updown.matches('.instanteventstarttimehour') ? this.instantEventStart : this.instantEventEnd;
				time.hours(time.hours() + 1 == 12 ? 0 : (time.hours() + 1 == 24 ? 12 : time.hours() + 1));
				this.UpdateInstantEventTimes();
			});
			down.addEventListener('click', () => {
				const time = $updown.matches('.instanteventstarttimehour') ? this.instantEventStart : this.instantEventEnd;
				time.hours(time.hours() - 1 == -1 ? 11 : (time.hours() - 1 == 11 ? 23 : time.hours() - 1));
				this.UpdateInstantEventTimes();
			});
		}
		// Increment and decrement start and end minute
		for (const [index, $upDown] of [...this.tag.querySelectorAll('.instanteventstarttimeminute, .instanteventendtimeminute')].entries()) {
			const up = $upDown.querySelector(':scope > .up');
			const down = $upDown.querySelector(':scope > .down');
			up.addEventListener('click', () => {
				var time = index == 0 ? this.instantEventStart : this.instantEventEnd;
				time.minutes(time.minutes() + 15 == 60 ? 0 : time.minutes() + 15);
				this.UpdateInstantEventTimes();
			});
			down.addEventListener('click', () => {
				var time = index == 0 ? this.instantEventStart : this.instantEventEnd;
				time.minutes(time.minutes() - 15 == -15 ? 45 : time.minutes() - 15);
				this.UpdateInstantEventTimes();
			});
		}
		// Increment and decrement start and end period
		for (const [index, $upDown] of [...this.tag.querySelectorAll('.instanteventstarttimeperiod, .instanteventendtimeperiod')].entries()) {
			for (const $element of $upDown.querySelectorAll(':scope > .up, :scope > .down')) {
				$element.addEventListener('click', () => {
					var time = index == 0 ? this.instantEventStart : this.instantEventEnd;
					time.add(time.hours() <= 12 ? 12 : -12, 'hours');
					this.UpdateInstantEventTimes();
				});
			}
		}
		
		this.$viewInstantEventFormNameInput.addEventListener('click', async ({ target: $input }) => {
			const value = await this.ShowKeyboard($input.innerText, false);
			this.instantEventName = value;
			$input.innerText = value;
		});
		
		this.$reportIssue.innerText = this.reportanissuebuttontext;
		this.$viewReportIssueConfirmationText.innerText = this.reportanissueconfirmationtext;
		
		if (this.showlogo && this.logoimage) {
			var image = new Image();
			image.onload = () => {
				Object.assign(this.$logo.style, {
					backgroundImage: `url('${this.logoimage.cdnurl}')`,
					backgroundSize: `auto ${this.logoheight}px`,
					width: `${this.logoheight * image.width / image.height}px`,
					height: `${this.logoheight}px`
				});
				this.ResetView(true);
			};
			image.src = this.logoimage.cdnurl;
		}
		
		var fps = 0;
		this.fpsInterval = setInterval(() => {
			if (this.TagDeleted()) {
				return;
			}
			this.currentFPS = fps;
			fps = 0;
		}, 1000);
		var fpsIncrement = () => {
			if (this.TagDeleted()) {
				return;
			}
			fps++;
			this.fpsRequestAnimationFrame = requestAnimationFrame(fpsIncrement);
		};
		this.fpsRequestAnimationFrame = requestAnimationFrame(fpsIncrement);
		
		await this.SetRoomText();
		
		if (this.widgetData.hasOwnProperty('signwave')) {
			this.LoadEvents();
			this.loadEventsInterval = setInterval(this.LoadEvents.bind(this), 5 * 60 * 1000);
		}
		this.ResetView(true);
		
		this.ready();
	}
	Destroy() {
		this.ClearTimers();
	}
	TimeSync() {
		if (Math.abs(this.oldServerOffset - this.player.serverOffset) > 1000) {
			this.oldServerOffset = this.player.serverOffset;
			this.ResetView(true);
		}
	}
};
