class Widgettext extends WidgetBase {
	constructor(options) {
		super(options);
		(async () => {
			let inlineimages = this.inlineimages
			if (this.container) {
				var imagesToCache = [
					...inlineimages
				];
				const imageCacheURLs = await Promise.all(imagesToCache.map(image => this.container.cacheMediaObject(image)));
				imageCacheURLs.forEach((cacheURL, index) => {
					if (cacheURL)
					{
						imagesToCache[index].cdnurl = cacheURL;
					}
				});
			}
			var cacheData = {
				inlineimages
			};
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			this.inlineimages = cacheData.inlineimages;
			
			this.updateText();
			
			this.ready();
		})();
	}
	updateText() {
		let text = this.text;
		
		text = text.replace(/\[metadata="[^"]*?"[^\]]*?\]|./g, metadataCode => {
			if (metadataCode.length == 1) {
				return metadataCode;
			} else {
				let metadataName = metadataCode.match(/metadata="([^"]*)"/);
				metadataName = metadataName.length == 2 ? metadataName[1] : '';
				let defaultValue = metadataCode.match(/default="([^"]*)"/);
				defaultValue = defaultValue && defaultValue.length == 2 ? defaultValue[1] : '';
				if (metadataName != '' && this.media.hasOwnProperty('metadata') && this.media.metadata.find(property => property.name == metadataName)) {
					const property = this.media.metadata.find(property => property.name == metadataName);
					return property.value;
				} else {
					return defaultValue;
				}
			}
		});
		
		var textWithInlineImages = text.replace(/\[image="[^"]*?"[^\]]*?\]|./g, characterOrImageCode => {
			if (characterOrImageCode.length == 1) {
				return $.span({ text: characterOrImageCode }).innerHTML;
			} else {
				var imageName = characterOrImageCode.match(/image="([^"]*)"/);
				imageName = imageName.length == 2 ? imageName[1] : '';
				var height = characterOrImageCode.match(/height="([^"]*)"/);
				height = height.length == 2 ? height[1] : 100;
				var align = characterOrImageCode.match(/align="([^"]*)"/);
				align = align.length == 2 ? align[1] : 'middle';
				
				var image = this.inlineimages.find(image => image.name == imageName);
				if (image != undefined) {
					var dropShadows = [];
					if (this.filters.hasOwnProperty('inlineimages')) {
						for (const filter of this.filters.inlineimages) {
							switch (filter.type) {
								case 'dropshadow':
									dropShadows.push('drop-shadow(' + filter.offset + 'px ' + filter.offset + 'px ' + filter.blurradius + 'px ' + filter.color + ')');
									break;
							}
						}
					}
					dropShadows = dropShadows.join(' ');
					return $.img({ src: image.cdnurl, css: {
						height: `${height}px`,
						verticalAlign: align,
						filter: dropShadows
					} }).outerHTML;
				} else {
					return '';
				}
			}
		});
		
		var textShadows = [];
		var textOutline = '';
		if (this.filters.hasOwnProperty('text')) {
			for (const filter of this.filters.text) {
				switch (filter.type) {
					case 'dropshadow':
						textShadows.push(`${filter.color} ${filter.offset}px ${filter.offset}px ${filter.blurradius}px`);
						break;
					case 'outline':
						textOutline = `${filter.width}px ${filter.color}`;
						break;
				}
			}
		}
		textShadows = textShadows.join(',');
		Object.assign(this.tag.style, {
			fontFamily: `'${this.fontfamily}'`,
			fontSize: `${this.fontsize}px`,
			lineHeight: `${this.lineheight == -1 ? this.fontsize : this.lineheight}px`,
			color: this.foreground,
			backgroundColor: this.background,
			fontWeight: this.bold ? 'bold' : 'normal',
			fontStyle: this.italic ? 'italic' : 'normal',
			textAlign: this.horizontal,
			whiteSpace: 'pre-wrap',
			wordWrap: 'break-word',
			display: 'flex',
			flexFlow: 'column nowrap',
			justifyContent: this.vertical == 'top' ? 'flex-start' : (this.vertical == 'middle' ? 'center' : 'flex-end'),
			textShadow: textShadows,
			WebkitTextStroke: textOutline
		});
		this.tag.innerHTML = '';
		this.tag.appendChild(
			$.div({ html: textWithInlineImages })
		);
	}
	Show() {
		this.updateText();
	}
	Hide() {
		
	}
	Destroy() {
		
	}
}
