function TextJSONToCharacterStyles(node, characterStyles, currentStyle) {
	if (characterStyles == undefined) {
		characterStyles = [];
		for (const child of node.children) {
			TextJSONToCharacterStyles(child, characterStyles, {});
		}
		for (var i = 0; i < characterStyles.length; ++i) {
			if (characterStyles[i].styles.hasOwnProperty('paragraph')) {
				if (i + 1 < characterStyles.length && !characterStyles[i + 1].styles.hasOwnProperty('paragraph')) {
					for (var style in characterStyles[i + 1].styles) {
						characterStyles[i].styles[style] = characterStyles[i + 1].styles[style];
					}
				}
			}
		}
		return characterStyles;
	} else {
		if (node.hasOwnProperty('text')) {
			for (var i = 0; i < node.text.length; ++i) {
				characterStyles.push({ character: node.text[i], styles: JSON.parse(JSON.stringify(currentStyle)) });
			}
		} else if (node.hasOwnProperty('widget')) {
			characterStyles.push({ widget: node.widget, styles: currentStyle });
		} else if (node.hasOwnProperty('paragraph')) {
			currentStyle.align = node.paragraph;
			var newStyle = JSON.parse(JSON.stringify(currentStyle));
			newStyle.paragraph = true;
			characterStyles.push({ character: String.fromCharCode(8203), styles: newStyle });
		} else {
			for (const style of [
				'family',
				'size',
				'lineheight',
				'stretch',
				'color',
				'highlight',
				'bold',
				'italic',
				'underline',
				'strikethrough',
				'subscript',
				'superscript'
			]) {
				if (node.hasOwnProperty(style)) {
					currentStyle[style] = node[style];
				}
			}
		}
		if (node.children) {
			if (node.children.length == 0) {
				characterStyles.push({ character: String.fromCharCode(8203), styles: JSON.parse(JSON.stringify(currentStyle)) });
			}
			for (const child of node.children) {
				TextJSONToCharacterStyles(child, characterStyles, JSON.parse(JSON.stringify(currentStyle)));
			}
		}
	}
}

function TextCharacterStylesToDOM($container, characterStyles, padding, index, length, originalCharacterStyles) {
	if (index === undefined) {
		originalCharacterStyles = characterStyles;
		characterStyles = JSON.parse(JSON.stringify(characterStyles));
		const $text = $container.appendChild($.div({ class: 'text', css: {
			left: `${padding.left}px`,
			right: `${padding.right}px`,
			top: `${padding.top}px`,
			bottom: `${padding.bottom}px`
		} }));
		TextCharacterStylesToDOM($text, characterStyles, padding, 0, characterStyles.length, originalCharacterStyles);
	} else {
		if (length == 0) {
			return;
		}
		// Index and length of the longest sequence for a style
		var sequenceIndex = 0;
		var sequenceLength = 0;
		var sequenceStyle = '';
		// Find paragraphs
		if (sequenceLength == 0) {
			for (var i = index; i < index + length; ++i) {
				if (characterStyles[i].styles.hasOwnProperty('paragraph') && sequenceLength == 0) {
					sequenceIndex = i;
					sequenceLength = 1;
					for (var j = i + 1; j < index + length; ++j) {
						if (characterStyles[j].styles.hasOwnProperty('paragraph')) {
							break;
						} else {
							sequenceLength++;
						}
					}
					sequenceStyle = { style: 'paragraph', value: characterStyles[i].styles.align };
				}
			}
		}
		var priorityStyle = ['lineheight', 'size', 'family'];
		for (const style of priorityStyle) {
			if (sequenceLength == 0) {
				for (var i = index; i < index + length; ++i) {
					if (characterStyles[i].styles.hasOwnProperty(style)) {
						var styleLength = 1;
						for (var j = i + 1; j < index + length; ++j) {
							if (characterStyles[j].styles.hasOwnProperty(style) && characterStyles[i].styles[style] == characterStyles[j].styles[style]) {
								styleLength++;
							} else {
								break;
							}
						}
						if (styleLength > sequenceLength) {
							sequenceIndex = i;
							sequenceLength = styleLength;
							sequenceStyle = { style: style, value: characterStyles[i].styles[style] };
						}
					}
				}
			}
		}
		// If there are no paragraphs or priority styles then look for the other longest styles
		if (sequenceLength == 0) {
			for (var i = index; i < index + length; ++i) {
				for (var style in characterStyles[i].styles) {
					var styleLength = 1;
					for (var j = i + 1; j < index + length; ++j) {
						if (characterStyles[j].styles.hasOwnProperty(style) && characterStyles[i].styles[style] == characterStyles[j].styles[style]) {
							styleLength++;
						} else {
							break;
						}
					}
					if (styleLength > sequenceLength) {
						sequenceIndex = i;
						sequenceLength = styleLength;
						sequenceStyle = { style: style, value: characterStyles[i].styles[style] };
					}
				}
			}
		}
		if (sequenceLength == 0) {
			// Build up widget and text nodes
			let textNodeString = '';
			const InsertTextNode = function(i) {
				if (textNodeString != '') {
					var textNode = document.createTextNode(textNodeString);
					$container.appendChild(textNode);
					textNodeString = '';
				}
			}
			for (var i = index; i < index + length; ++i) {
				if (characterStyles[i].character) {
					textNodeString += characterStyles[i].character;
				} else if (characterStyles[i].widget) {
					InsertTextNode(i);
					const tag = $.span();
					var currentDate = new Date();
					switch (characterStyles[i].widget) {
						case '12-Hour (HH:MM:SS AM/PM)':
							var hours = currentDate.getUTCHours() % 12;
							if (hours == 0) {
								hours = 12;
							}
							tag.innerHTML = ('0' + hours).slice(-2) + ':' + ('0' + (currentDate.getUTCMinutes())).slice(-2) + ':' + ('0' + (currentDate.getUTCSeconds())).slice(-2) + ' ' + (currentDate.getUTCHours() < 12 ? 'AM' : 'PM');
							break;
						case '24-Hour (HH:MM:SS)':
							tag.innerHTML = ('0' + (currentDate.getUTCHours())).slice(-2) + ':' + ('0' + (currentDate.getUTCMinutes())).slice(-2) + ':' + ('0' + (currentDate.getUTCSeconds())).slice(-2);
							break;
						case '12-Hour (H)':
							var hours = currentDate.getUTCHours() % 12;
							if (hours == 0) {
								hours = 12;
							}
							tag.innerHTML = hours;
							break;
						case '12-Hour (HH)':
							var hours = currentDate.getUTCHours() % 12;
							if (hours == 0) {
								hours = 12;
							}
							tag.innerHTML = ('0' + hours).slice(-2);
							break;
						case '24-Hour (H)':
							tag.innerHTML = currentDate.getUTCHours();
							break;
						case '24-Hour (HH)':
							tag.innerHTML = ('0' + currentDate.getUTCHours()).slice(-2);
							break;
						case 'Minutes (M)':
							tag.innerHTML = currentDate.getUTCMinutes();
							break;
						case 'Minutes (MM)':
							tag.innerHTML = ('0' + currentDate.getUTCMinutes()).slice(-2);
							break;
						case 'Seconds (S)':
							tag.innerHTML = currentDate.getUTCSeconds();
							break;
						case 'Seconds (SS)':
							tag.innerHTML = ('0' + currentDate.getUTCSeconds()).slice(-2);
							break;
						case 'am/pm':
							tag.innerHTML = currentDate.getUTCHours() < 12 ? 'am' : 'pm';
							break;
						case 'AM/PM':
							tag.innerHTML = currentDate.getUTCHours() < 12 ? 'AM' : 'PM';
							break;
						case 'Date (MM/DD/YY)':
							tag.innerHTML = ('0' + (currentDate.getMonth() + 1)).slice(-2) + '/' + ('0' + currentDate.getDate()).slice(-2) + '/' + ('0' + currentDate.getFullYear()).slice(-2);
							break;
						case 'Date (MM/DD/YYYY)':
							tag.innerHTML = ('0' + (currentDate.getMonth() + 1)).slice(-2) + '/' + ('0' + currentDate.getDate()).slice(-2) + '/' + currentDate.getFullYear();
							break;
						case 'Day Number (D)':
							tag.innerHTML = currentDate.getDate()
							break;
						case 'Day Number (DD)':
							tag.innerHTML = ('0' + currentDate.getDate()).slice(-2);
							break;
						case 'Month Number (M)':
							tag.innerHTML = currentDate.getMonth() + 1;
							break;
						case 'Month Number (MM)':
							tag.innerHTML = ('0' + (currentDate.getMonth() + 1)).slice(-2);
							break;
						case 'Year Number (YY)':
							tag.innerHTML = ('0' + currentDate.getFullYear()).slice(-2);
							break;
						case 'Year Number (YYYY)':
							tag.innerHTML = currentDate.getFullYear();
							break;
						case 'Day Short (Mon)':
							var days = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
							tag.innerHTML = days[currentDate.getDay()];
							break;
						case 'Day Long (Monday)':
							var days = ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'];
							tag.innerHTML = days[currentDate.getDay()];
							break;
						case 'Month Short (Jan)':
							var months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
							tag.innerHTML = months[currentDate.getMonth()];
							break;
						case 'Month Long (January)':
							var months = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
							tag.innerHTML = months[currentDate.getMonth()];
							break;
						case 'Countdown Days (D)':
							break;
						case 'Countdown Days (DD)':
							break;
						case 'Countdown Minutes (M)':
							break;
						case 'Countdown Minutes (MM)':
							break;
						case 'Countdown Seconds (S)':
							break;
						case 'Countdown Seconds (SS)':
							break;
						case 'Countdown Milliseconds (MS)':
							break;
						case "Today's Temperature (C)":
							break;
						case "Today's Temperature (F)":
							break;
						case "Tomorrow's Temperature (C)":
							break;
						case "Tomorrow's Temperature (F)":
							break;
						case "Nth-Day's Temperature (C)":
							break;
						case "Nth-Day's Temperature (F)":
							break;
						case 'Wind Direction (Compass)':
							break;
						case 'Wind Speed (Knots)':
							break;
						case 'Wind Speed (MPH)':
							break;
						case 'Wind Speed (m/s)':
							break;
						case "Today's Weather (Description)":
							break;
						case 'RSS':
							break;
						case 'Stocks':
							break;
						case 'Stocks with Change':
							break;
						case 'Stocks with Change (Arrows)':
							break;
						case 'Image':
							break;
					}
					$container.appendChild(tag);
				}
			}
			InsertTextNode(i);
		} else {
			var tag;
			switch (sequenceStyle.style) {
				case 'paragraph':
					tag = document.createElement('p');
					tag.style.textAlign = sequenceStyle.value;
					break;
				case 'family':
					tag = document.createElement('span');
					tag.style.fontFamily = sequenceStyle.value;
					break;
				case 'size':
					tag = document.createElement('span');
					tag.style.fontSize = `${sequenceStyle.value}px`;
					break;
				case 'lineheight':
					tag = document.createElement('span');
					tag.style.lineHeight = `${sequenceStyle.value}px`;
					break;
				case 'stretch':
					tag = document.createElement('span');
					tag.style.fontStretch = sequenceStyle.value;
					break;
				case 'color':
					tag = document.createElement('span');
					tag.style.color = sequenceStyle.value;
					break;
				case 'highlight':
					tag = document.createElement('span');
					tag.style.backgroundColor = sequenceStyle.value;
					break;
				case 'bold':
					tag = document.createElement('b');
					break;
				case 'italic':
					tag = document.createElement('i');
					break;
				case 'underline':
					tag = document.createElement('u');
					break;
				case 'strikethrough':
					tag = document.createElement('s');
					break;
				case 'subscript':
					tag = document.createElement('sub');
					break;
				case 'superscript':
					tag = document.createElement('sup');
					break;
				default:
					return;
			}
			for (var i = sequenceIndex; i < sequenceIndex + sequenceLength; ++i) {
				delete characterStyles[i].styles[sequenceStyle.style];
				if (sequenceStyle.style == 'paragraph') {
					delete characterStyles[i].styles.align;
				}
			}
			TextCharacterStylesToDOM($container, characterStyles, padding, index, sequenceIndex - index, originalCharacterStyles);
			TextCharacterStylesToDOM(tag, characterStyles, padding, sequenceIndex, sequenceLength, originalCharacterStyles);
			$container.appendChild(tag);
			TextCharacterStylesToDOM($container, characterStyles, padding, sequenceIndex + sequenceLength, index + length - (sequenceIndex + sequenceLength), originalCharacterStyles);
		}
	}
}

class Widgettext2 extends WidgetBase {
	constructor(options) {
		super(options);
		(async () => {
			let inlineimages = this.inlineimages;
			if (this.container) {
				/*
				var imagesToCache = 
				[
					...inlineimages
				];
				const imageCacheURLs = await Promise.all(imagesToCache.map(image => this.container.cacheMediaObject(image)));
				imageCacheURLs.forEach((cacheURL, index) =>
				{
					if (cacheURL)
					{
						imagesToCache[index].cdnurl = cacheURL;
					}
				});
				*/
			}
			var cacheData = {
				inlineimages
			};
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			inlineimages = cacheData.inlineimages;
			
			let characterStyles = TextJSONToCharacterStyles(this.text);
			
			let text = characterStyles.reduce((text, characterStyle) => text + characterStyle.character, '');
			
			var index = 0;
			text.replace(/\[metadata="[^"]*?"[^\]]*?\]|./g, metadataCode => {
				if (metadataCode.length == 1) {
					index++;
				} else {
					let metadataName = metadataCode.match(/metadata="([^"]*)"/);
					metadataName = metadataName.length == 2 ? metadataName[1] : '';
					let defaultValue = metadataCode.match(/default="([^"]*)"/);
					defaultValue = defaultValue && defaultValue.length == 2 ? defaultValue[1] : '';
					let value = defaultValue;
					if (metadataName != '' && this.media.hasOwnProperty('metadata') && this.media.metadata.find(property => property.name == metadataName)) {
						const property = this.media.metadata.find(property => property.name == metadataName);
						value = property.value;
					}
					const replacement = [];
					// Use the style of the first open bracket
					const styles = characterStyles[index].styles;
					for (const char of value) {
						replacement.push({
							character: value.charAt(replacement.length),
							styles: JSON.parse(JSON.stringify(styles))
						});
					}
					
					characterStyles.splice(index, metadataCode.length, ...replacement);
					index += replacement.length;
				}
			});
			
			this.tag.classList.add('widget-text2');
			Object.assign(this.tag.style, {
				backgroundColor: this.background
			});
			this.tag.innerHTML = '';
			TextCharacterStylesToDOM(this.tag, characterStyles, { left: this.paddingleft, right: this.paddingright, top: this.paddingtop, bottom: this.paddingbottom });
			
			this.ready();
		})();
	}
}