class Widgetvideostreaming extends WidgetBase {
	constructor(options) {
		super(options);
		this.media.duckedVolume = this.audioducking / 100;
		this.playing = false;
		this.lastCanPlayThrough = null;
		this.lastManifestParsed = null;
		this.tryAgainTimer = null;
		switch (this.format) {
			case 'other':
				this.tag.classList.add('widget-videostreaming');
				this.tag.appendChild($.span({ text: this.url }));
				break;
			case 'hls':
				this.$video = this.tag.appendChild($.video({ css: {
					width: '100%',
					height: '100%'
				} }));
				this.initializeHls();
				break;
		}
		this.ready();
	}
	initializeHls() {
		if (Hls.isSupported()) {
			this.hls = new Hls({ liveSyncDurationCount: this.hlslivesyncdurationcount });
			this.hls.loadSource(this.url);
			this.hls.attachMedia(this.$video);
			this.hls.subtitleDisplay = this.subtitles;
			this.lastManifestParsed = async () => {
				this.lastManifestParsed = null;
				// Make sure to set the subtitle track if it exists
				if (this.hls.subtitleTracks.length > 0) {
					this.hls.subtitleTrack = 0;
				}
				if (this.playing) {
					try {
						await this.$video.play();
					} catch (exception) {
						// Failed to play
					}
				}
			};
			this.hls.once(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
			this.hls.on(Hls.Events.ERROR, (e, data) => {
				// Only fatal errors need to be handled. All other errors per the docs should continue playing
				if (data.fatal) {
					switch(data.type) {
						// Skip all the recommended error handling and just reset the whole thing
						case Hls.ErrorTypes.NETWORK_ERROR:
							// try to recover network error
							//console.log("video streaming: fatal network error encountered, try to recover");
							//this.hls.startLoad();
							//break;
						case Hls.ErrorTypes.MEDIA_ERROR:
							//console.log("video streaming: fatal media error encountered, try to recover");
							//this.hls.recoverMediaError();
							//break;
						default:
							// This indicates hls should be destroyed. Try rebuilding it and start over? Unclear if this is a good idea, but we need to always recover.
							this.hls.detachMedia();
							this.hls.destroy();
							this.initializeHls();
							// This would usually mean there's no known error and hls should be destroyed
							if (this.playing && this.tryAgainTimer == null) {
								this.tryAgainTimer = setTimeout(() => {
									this.tryAgainTimer = null;
									this.playHls();
								}, 60 * 1000);
							}
							break;
					}
				}
			});
		}
	}
	playHls() {
		if (this.playing) {
			if (this.hls) {
				this.hls.detachMedia();
			}
			if (this.lastManifestParsed) { 
				this.hls.off(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
			}
			this.hls.loadSource(this.url);
			this.hls.attachMedia(this.$video);
			this.lastManifestParsed = async () => {
				this.lastManifestParsed = null;
				if (this.playing) {
					try {
						await this.$video.play();
					} catch (exception) {
						// Failed to play
					}
				}
			};
			this.hls.once(Hls.Events.MANIFEST_PARSED, this.lastManifestParsed);
		}
	}
	async Show() {
		this.playing = true;
		switch (this.format) {
			case 'other':
				break;
			case 'hls':
				this.playHls();
				break;
		}
	}
	async Hide() {
		this.playing = false;
		clearTimeout(this.tryAgainTimer);
		this.tryAgainTimer = null;
		switch (this.format) {
			case 'other':
				
			case 'hls':
				if (this.hls) {
					this.hls.detachMedia();
				}
				break;
		}
	}
	Destroy() {
		if (this.hls) {
			this.hls.detachMedia();
		}
	}
	SetVolume(volume) {
		this.$video.volume = volume;
	}
}
