class Widgetweathercrawl extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.tag.classList.add('widget-weathercrawl');
		this.tag.style.backgroundColor = this.background;
		
		this.weatherInterval = null;
		this.crawl = new BreezeCrawl(this.tag, this.speed, this.alternativescroll);
		this.crawl.foreground = this.foreground;
		this.crawl.spacer = ' ';
		
		this.weatherInterval = setInterval(this.getWeather.bind(this), 15 * 60 * 1000); // update every 15 minutes
		this.getWeather();
		
		this.ready();
	}
	fahrenheitToCelsius(temp) {
		return this.celsius ? Math.round((temp - 32) / 1.8) : temp;
	}
	printDegreeMark() {
		return this.celsius ? 'C' : 'F';
	}
	getFullDay(day) {
		switch (day) {
			case 'Mon':
				return 'Monday';
			case 'Tue':
				return 'Tuesday';
			case 'Wed':
				return 'Wednesday';
			case 'Thu':
				return 'Thursday';
			case 'Fri':
				return 'Friday';
			case 'Sat':
				return 'Saturday';
			case 'Sun':
				return 'Sunday';
			default:
				return '_';
		}
	}
	async getWeather() {
		if (this.TagDeleted()) {
			return;
		}
		
		const location = this.widgetData.hasOwnProperty('zipcode') && this.widgetData.zipcode != '' ? this.widgetData.zipcode : this.location;
		
		var cacheData = await new Promise(async resolve => {
			try {
				let $xml = await $.http({
					method: 'GET',
					url: (this.container ? this.container.hostUrl.replace(/\/$/, '') : '') + '/get/weatherdata.php',
					data: {
						location: location,
						celsius: this.celsius
					},
					dataType: 'xml'
				}).promise;
				$xml = $xml.documentElement;
				if ($xml.querySelector('forecast_conditions') == null && ($xml.querySelector('current_conditions') == null || $xml.querySelector('current_conditions').querySelector('temp_f').getAttribute('data') == null)) {
					resolve('nodata');
					return;
				}
				const currentConditions = $xml.querySelector('current_conditions');
				const city = $xml.querySelector('forecast_information').querySelector('city').getAttribute('data').replace('United States of America,', '');
				const todayTemp = this.fahrenheitToCelsius(parseFloat(currentConditions.querySelector('temp_f').getAttribute('data')));
				const todayCondition = currentConditions.querySelector('condition').getAttribute('data');
				const forecast = [];
				for (const $day of $xml.querySelectorAll('forecast_conditions')) {
					forecast.push({
						high: this.fahrenheitToCelsius(parseFloat($day.querySelector('high').getAttribute('data'))),
						low: this.fahrenheitToCelsius(parseFloat($day.querySelector('low').getAttribute('data'))),
						condition: $day.querySelector('condition').getAttribute('data'),
						dayOfWeek: $day.querySelector('day_of_week').getAttribute('data'),
						dayOfWeekLong: this.getFullDay($day.querySelector('day_of_week').getAttribute('data'))
					});
				}
				resolve({
					city: city,
					todaytemp: todayTemp,
					todaycondition: todayCondition,
					forecast: forecast
				});
			} catch (exception) {
				resolve(null);
			}
		});
		if (this.TagDeleted()) {
			return;
		}
		var weatherString = '';
		if (cacheData === 'nodata') {
			weatherString = 'Weather information is currently not available.         ';
		} else {
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			if (!cacheData.hasOwnProperty('city')) {
				cacheData.city = 'No Data for Location or No Connection';
			}
			if (!cacheData.hasOwnProperty('todaytemp')) {
				cacheData.todaytemp = 'N/A';
			}
			if (!cacheData.hasOwnProperty('todaycondition')) {
				cacheData.todaycondition = 'N/A';
			}
			if (!cacheData.hasOwnProperty('forecast') || cacheData.forecast.length < 4) {
				cacheData.forecast =
				[
					{ low: 'N/A', high: 'N/A', condition: 'N/A', dayOfWeek: 'N/A', dayOfWeekLong: 'N/A', icon: 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///ywAAAAAAQABAAACAUwAOw==' },
					{ low: 'N/A', high: 'N/A', condition: 'N/A', dayOfWeek: 'N/A', dayOfWeekLong: 'N/A', icon: 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///ywAAAAAAQABAAACAUwAOw==' },
					{ low: 'N/A', high: 'N/A', condition: 'N/A', dayOfWeek: 'N/A', dayOfWeekLong: 'N/A', icon: 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///ywAAAAAAQABAAACAUwAOw==' },
					{ low: 'N/A', high: 'N/A', condition: 'N/A', dayOfWeek: 'N/A', dayOfWeekLong: 'N/A', icon: 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///ywAAAAAAQABAAACAUwAOw==' }
				];
			}
			
			cacheData.forecast.forEach((forecastDay, forecastDayItr) => {
				if (forecastDayItr == 0) {
					if (this.shortform) {
						weatherString = `${cacheData.city}            Currently: ${cacheData.todaytemp}${this.printDegreeMark()}         `;
					} else {
						weatherString = `${cacheData.city}   Currently: ${cacheData.todaytemp}${this.printDegreeMark()}         Forecast: `;
					}
				} else {
					if (this.shortform) {
						weatherString += `${forecastDay.dayOfWeek}: ${forecastDay.condition} ${forecastDay.high}${this.printDegreeMark()} | ${forecastDay.low}${this.printDegreeMark()}             `;
					} else {
						weatherString += `${forecastDay.dayOfWeekLong} ${forecastDay.high}${this.printDegreeMark()}|${forecastDay.low}${this.printDegreeMark()} Conditions: ${forecastDay.condition}            `;
					}
				}
			});
		}
		this.crawl.startCrawl(this.width, this.height, weatherString, this.fontsize, this.font, !this.editor);
	}
	Show() {
		this.crawl.play();
	}
	Hide() {
		this.crawl.pause();
	}
	Destroy() {
		this.crawl.destroy();
		clearInterval(this.weatherInterval);
	}
}
