class Widgetwebpage extends WidgetBase {
	constructor(options) {
		super(options);
		
		this.refreshInterval = null;
		
		const baseURL = this.container ? this.container.hostUrl + 'screenshots' : '/screenshots';
		this.staticURL = `${baseURL}?url=${encodeURIComponent(this.url)}&width=${this.renderwidth}&height=${this.renderheight}&scrollleft=${this.scrollleft}&scrolltop=${this.scrolltop}&key=${this.key}`;
		
		if (this.refresh != -1) {
			this.refreshInterval = setInterval(this.Update.bind(this), this.refresh * 1000);
		}
		this.Update();
		
		this.ready();
	}
	async Update() {
		if (this.TagDeleted()) {
			return;
		}
		this.tag.innerHTML = '';
		if (this.live) {
			// iframe sandboxing will be disabled for the following origins:
			const ORIGIN_WHITELIST = [
				'https://display.tagboard.com'
			];
			const iframeUri = new URL(this.url);
			const isWhitelisted = ORIGIN_WHITELIST.some(origin => origin === iframeUri.origin);
			this.tag.style.overflow = 'hidden';
			this.tag.appendChild($.iframe({
				src: this.url,
				sandbox: !isWhitelisted ? 'allow-scripts allow-same-origin' : undefined,
				css: {
					width: `${this.renderwidth}px`,
					height: `${this.renderheight}px`,
					transformOrigin: '0 0',
					transform: `scale(${this.width / this.cropwidth},${this.height / this.cropheight}) translate(${-this.cropx}px,${-this.cropy}px)`,
					border: 'none'
				}, load: function() {
					try {
						this.contentWindow.document.head.appendChild($.style({ type: 'text/css', html: `::-webkit-scrollbar {
-webkit-appearance: none;
}
::-webkit-scrollbar:vertical {
width: 12px;
}
::-webkit-scrollbar:horizontal {
height: 12px;
}
::-webkit-scrollbar-thumb {
background-color: rgba(0, 0, 0, .5);
border-radius: 10px;
border: 2px solid #ffffff;
}
::-webkit-scrollbar-track {
border-radius: 10px;
background-color: #ffffff;
}` }));
				} catch (exception) {
					// Do nothing, only works for src on the same domain
				}
			} }));
		} else {
			let cacheData = {};
			if (this.container) {
				const fakeMedia =
				{
					id: Math.floor(Math.random() * 99999 + 100000),
					cdnurl: `${this.staticURL}&t=${Math.floor(Math.random() * 10000)}`,
					type: Type.Media.IMAGE,
					name: this.staticURL,
					filesize: null,
					uploadedtimestamp: moment().unix(),
					md5: null
				};
				const imageCacheURLs = await this.container.cacheMediaObject(fakeMedia);
				if (imageCacheURLs) {
					Object.assign(cacheData,
					{
						url: imageCacheURLs
					});
				}
			}
			Object.assign(cacheData,
			{
				url: this.staticURL
			}, JSON.parse(JSON.stringify(cacheData)));
			cacheData = await UpdateWidgetCache(cacheData, this.mediaid, this.container);
			this.tag.style.background = `url('${cacheData.url}') ${(-this.cropx * this.width / this.cropwidth)}px ${(-this.cropy * this.height / this.cropheight)}px / ${(this.renderwidth / this.cropwidth * 100)}% ${(this.renderheight / this.cropheight * 100)}% no-repeat`;
		}
	}
	Show() {
		if (this.muteonhideandreloadonshow) {
			this.Update();
			if (this.refresh != -1) {
				clearInterval(this.refreshInterval);
				this.refreshInterval = setInterval(this.Update.bind(this), this.refresh * 1000);
			}
		}
	}
	Hide() {
		if (this.muteonhideandreloadonshow) {
			clearInterval(this.refreshInterval);
			if (this.TagDeleted()) {
				return;
			}
			this.tag.innerHTML = '';
		}
	}
	Destroy() {
		clearInterval(this.refreshInterval);
		this.tag.innerHTML = '';
	}
}