let youtubeAPIReadyResolve = null;
const youtubeAPIReady = new Promise(resolve => youtubeAPIReadyResolve = resolve);
function onYouTubeIframeAPIReady() {
	youtubeAPIReadyResolve();
}

class Widgetyoutube extends WidgetBase {
	constructor(options) {
		super(options);
		
		// Only load the youtube API if it's needed, most customers don't use the Youtube widget
		if (document.querySelector('script[src="https://www.youtube.com/iframe_api"]') == null) {
			document.head.appendChild($.script({ src: 'https://www.youtube.com/iframe_api' }));
		}
		this.youtubePlayerId = `youtubeplayer-${Math.floor(Math.random() * Number.MAX_SAFE_INTEGER)}`;
		while (document.getElementById(this.youtubePlayerId) != null) {
			this.youtubePlayerId = `youtubeplayer-${Math.floor(Math.random() * Number.MAX_SAFE_INTEGER)}`;
		}
		this.playing = false;
		this.ensurePlayingTimer = null;
		this.seconds = 0;
		if (this.starttime != '') {
			const secondsMatches = this.starttime.match(/^([0-9]*)$/);
			const startTimeMatches = this.starttime.match(/^([0-9]*h)?([0-9]*m)?([0-9]*s)?$/);
			if (secondsMatches && secondsMatches.length == 2) {
				this.seconds = parseInt(secondsMatches[1], 10);
			} else if (startTimeMatches) {
				if (startTimeMatches[1]) {
					this.seconds += 60 * 60 * parseInt(startTimeMatches[1], 10);
				}
				if (startTimeMatches[2]) {
					this.seconds += 60 * parseInt(startTimeMatches[2], 10);
				}
				if (startTimeMatches[3]) {
					this.seconds += parseInt(startTimeMatches[3], 10);
				}
			}
		}
		
		this.Load();
		this.ready();
	}
	async Load() {
		if (this.TagDeleted()) {
			return;
		}
		// This div is replaced by the player
		this.tag.innerHTML = '';
		this.tag.appendChild($.div({ id: this.youtubePlayerId }));
		
		// https://youtu.be/dQw4w9WgXcQ
		// https://www.youtube.com/watch?v=dQw4w9WgXcQ
		let videoIdMatches = this.url.match(/^.*?(?:v=|youtu\.be\/)([A-Za-z0-9-_]*)$/);
		// https://www.youtube.com/watch?v=2nXYbGmF3_Q&list=PLyqf6gJt7KuHBmeVzZteZUlNUQAVLwrZS
		let playlistIdMatches = this.url.match(/^.*?(?:v=)?[A-Za-z0-9-_]*?&?list=([A-Za-z0-9-_]*)$/);
		let videoId = '';
		let playlistId = '';
		this.isVideo = false;
		if (videoIdMatches && videoIdMatches.length == 2) {
			videoId = videoIdMatches[1];
			this.isVideo = true;
		}
		if (playlistIdMatches && playlistIdMatches.length == 2) {
			playlistId = playlistIdMatches[1];
		}
		
		this.youtubePlayerReadyReject = null;
		this.youtubePlayerReady = new Promise(async (resolve, reject) => {
			this.youtubePlayerReadyReject = reject;
			
			await youtubeAPIReady;
			
			let playerOptions =  {
				width: this.width,
				height: this.height,
				// https://developers.google.com/youtube/iframe_api_reference#Events
				events: {
					'onReady': e => {
						resolve();
						if (!this.editor) {
							if (this.isVideo && !this.livestream) {
								this.youtubePlayer.seekTo(this.seconds);
							}
							if (this.mute || this.editor) {
								try {
									e.target.mute();
								} catch (exception) {
								}
							}
							if (this.playing) {
								try {
									this.youtubePlayer.playVideo();
								} catch (exception) {
								}
							} else {
								try {
									this.youtubePlayer.pauseVideo();
								} catch (exception) {
								}
							}
						}
					},
					'onStateChange': e => {
						if (e.data == YT.PlayerState.PLAYING) {
						} else if (e.data === YT.PlayerState.ENDED) {
							if (this.loop) {
								if (this.isVideo && !this.livestream) {
									this.youtubePlayer.seekTo(this.seconds);
								}
								if (!this.editor) {
									this.youtubePlayer.playVideo();
								}
							}
						}
					},
					'onError': e => reject(new Error(e.data))
				}
			};
			
			if (videoId != '') {
				Object.assign(playerOptions, {
					videoId: videoId,
					playerVars: {
						enablejsapi: 1,
						controls: 0,
						origin: window.location.origin,
						cc_load_policy: this.captions ? 1 : 0,
						cc_lang_pref: 'en',
						autoplay: 0
					}
				});
			} else if (playlistId != '') {
				Object.assign(playerOptions, {
					playerVars: {
						enablejsapi: 1,
						controls: 0,
						origin: window.location.origin,
						cc_load_policy: this.captions ? 1 : 0,
						cc_lang_pref: 'en',
						autoplay: 0,
						listType: 'playlist',
						list: playlistId
					}
				});
			} else {
				console.log('Youtube Widget: No videoId or playlistId');
				reject(new Error(0));
				return;
			}
			try {
				this.youtubePlayer = new YT.Player(this.youtubePlayerId, playerOptions);
				
			} catch (exception) {
				reject(new Error(0));
			}
		}).catch(exception => {});
	}
	async Show() {
		this.playing = true;
		try {
			await this.youtubePlayerReady;
			if (!this.playing) {
				// Hide was called interrupting this
				return;
			}
			if (this.seektobeginning && this.isVideo && !this.livestream) {
				this.youtubePlayer.seekTo(this.seconds);
				if (this.editor) {
					this.youtubePlayer.pauseVideo();
				}
			}
			if (!this.editor) {
				this.youtubePlayer.playVideo();
				if (this.livestream) {
					(async () => {
						try {
							for (let getDurationItr = 0; getDurationItr < 3; ++getDurationItr) {
								await new Promise(resolve => setTimeout(resolve, 1000));
								if (!this.playing || this.TagDeleted()) {
									return;
								}
								const duration = this.youtubePlayer.getDuration();
								if (duration != 0) {
									this.youtubePlayer.seekTo(duration);
									return;
								}
							}
						} catch (exception) {
							// Do nothing
						}
					})();
				}
				clearInterval(this.ensurePlayingTimer);
				this.ensurePlayingTimer = setInterval(() => {
					const state = this.youtubePlayer.getPlayerState();
					if (state == -1 || state == 2) { // unstarted or paused
						this.youtubePlayer.playVideo();
					}
				}, 1000);
			}
		} catch (exception) {
			// Do nothing
		}
	}
	async Hide() {
		this.playing = false;
		try {
			clearInterval(this.ensurePlayingTimer);
			await this.youtubePlayerReady;
			if (this.playing) {
				// Show was called interrupting this
				return;
			}
			clearInterval(this.ensurePlayingTimer);
			this.youtubePlayer.pauseVideo();
		} catch (exception) {
			// Do nothing
		}
	}
	Destroy() {
		if (this.youtubePlayer) {
			this.youtubePlayer.destroy();
		}
		this.tag.innerHTML = '';
		this.youtubePlayerReadyReject(new Error(0));
		clearInterval(this.ensurePlayingTimer);
	}
}