$Wcl = new-object System.Net.WebClient
$Wcl.Headers.Add("user-agent", "PowerShell Script")
$Wcl.Proxy.Credentials = [System.Net.CredentialCache]::DefaultNetworkCredentials

function FormatPath ([String]$path) {
	return [IO.Path]::GetFullPath($path)
}

function DeleteFolder ([String]$path) {
	try {
		Remove-Item $path -Recurse -ErrorAction Stop
	} catch {
		$e = $_.Exception.GetType().Name
		if ($e -ne 'ItemNotFoundException') {
			throw $_
		}
	}
}

function DeleteFile ([String]$path) {
	try {
		Remove-Item $path -ErrorAction Stop
	} catch {
		$e = $_.Exception.GetType().Name
		if ($e -ne 'ItemNotFoundException') {
			throw $_
		}
	}
}

function SaveFile ([String]$path, [String]$text) {
	$text = $text -replace "`n","`r`n"
	Out-File -Filepath $path -InputObject $text -Encoding ASCII
}

function DownloadRemoteFile ([String]$fetchUrl, [String]$destinationFilepath) {
	$startTime = Get-Date
	Write-Output "Starting download from '$fetchUrl', saving to '$destinationFilepath'..."
	Write-Output "This may take a few minutes... do not close!"
	$client = New-Object System.Net.WebClient
	$client.DownloadFile($fetchUrl, $destinationFilepath)
	$endTime = Get-Date
	$totalSeconds = $(New-Timespan $startTime $endTime).TotalSeconds
	Write-Output "Download finished in $totalSeconds seconds.."
}

function ExtractZipFile ([String]$filePath, [String]$destinationPath) {
	# Convert the filepaths into "full" filepaths
	$file = [IO.Path]::GetFullPath($filePath)
	$dest = [IO.Path]::GetFullPath($destinationPath)
	Write-Output "Extracting Zip File $file to $dest"
	$Shell = New-Object -ComObject Shell.Application
	$fileHandle = $Shell.NameSpace($file)
	$destHandle = $Shell.NameSpace($dest)
	foreach ($item in $fileHandle.Items()) {
		$destHandle.CopyHere($item, 1044)
		$filename = $item.Name
		Write-Output "../$filename moved to $dest/$filename.."
	}
}

function CopyFiles {
	Write-Output "Copying Breeze container files..."
	# Copy-Item has really weird behavior when copying to a directory that exists. Don't use it for folders
	cmd.exe /c "robocopy $updateDirectory/container/src/ $containerPath/src/ /E /NS /NC /NFL /NDL /NP /NJH /NJS"
	cmd.exe /c "robocopy $updateDirectory/container/node_modules/ $containerPath/node_modules/ /E /NS /NC /NFL /NDL /NP /NJH /NJS"
	Copy-Item "$updateDirectory/container/package.json" $packagePath
	Copy-Item "$updateDirectory/*.ico" $rootPath
	cmd.exe /c "robocopy $updateDirectory/scripts/ $rootPath/scripts/ /E /NS /NC /NFL /NDL /NP /NJH /NJS"
}

function CreateShortcut ([String]$shortcutPath, $targetPath, $arguments, $iconPath) {
	$WshShell = New-Object -ComObject WScript.Shell
	$desktopShortcut = $WshShell.CreateShortcut($shortcutPath)
	$desktopShortcut.TargetPath = $targetPath
	$desktopShortcut.Arguments = $arguments
	$desktopShortcut.IconLocation = $iconPath
	$desktopShortcut.Save()
}

function OptimizingPowershell () {
	Write-Output "Optimizing powershell"
	$env:PATH = [Runtime.InteropServices.RuntimeEnvironment]::GetRuntimeDirectory()
	[AppDomain]::CurrentDomain.GetAssemblies() | ForEach-Object {
		$path = $_.Location
		if ($path) {
			$name = Split-Path $path -Leaf
			Write-Host -ForegroundColor Yellow "`r`nRunning ngen.exe on '$name'"
			ngen.exe install $path /nologo
		}
	}
}

function CreateWatchdog {
	DeleteFile $watchdogPath
	$watchdogScript = @"
if (Test-Path $updateLockPath) {
	exit
}
# Very important. Ensure that Start-Process (and thus the nw process) is started in this directory
cd $containerPath
# If the nw.exe is not running or is using less than 15MB of memory kill it and restart
# Starting in 0.49.0 the nwjs has a bug that leaves orphaned nw.exe instances running so this 15MB check fixes that
`$nw = Get-Process nw -ErrorAction SilentlyContinue
if (`$nw -eq `$null -Or (Get-Process | Sort-Object -Descending WS | Where {`$_.Name -eq "nw"} | Select -First 1 -ExpandProperty WS) -lt 15000000) {
	Stop-Process -Name "nw" -ErrorAction SilentlyContinue
	Start-Sleep -s 1
	("`$("[{0:MM/dd/yy} {0:HH:mm:ss}]" -f (Get-Date)) | Detected container was not running... launching container...") >> $runlogPath
	Start-Process -FilePath $containerPath/nw.exe -RedirectStandardError "$errorlogPath"
}
"@
	SaveFile $watchdogPath $watchdogScript

	# This must use WShell. PowerShell is a console and there is no way to hide it without it flashing for a second. (This is a known issue and MS isn't going to fix it).
	DeleteFile $watchdogRunnerPath
	$watchdogRunnerScript = @"
Dim WShell
Set WShell = CreateObject("WScript.Shell")
WShell.Run "powershell.exe -nologo -command $watchdogPath", 0
Set WShell = Nothing
"@
	SaveFile $watchdogRunnerPath $watchdogRunnerScript
	
	# Start the scheduled task
	$action = New-ScheduledTaskAction -Execute "$watchdogRunnerPath"
	$principal = New-ScheduledTaskPrincipal -RunLevel Highest -UserId "$env:USERDOMAIN\$env:USERNAME"
	$trigger = New-ScheduledTaskTrigger -Once -At (Get-Date) -RepetitionInterval (New-TimeSpan -Minutes 1)
	$task = New-ScheduledTask -Action $action -Trigger $trigger -Principal $principal
	Register-ScheduledTask -TaskName "$taskName" -InputObject $task -Force

	DeleteFile $stopMonitorPath
	$stopMonitorScript = @"
Stop-Process -Name "nw" -ErrorAction SilentlyContinue
Start-Sleep -s 1
Unregister-ScheduledTask -TaskName "$taskName" -Confirm:`$false
"@
	SaveFile $stopMonitorPath $stopMonitorScript

	DeleteFile $startMonitorPath
	$startMonitorScript = @"
Stop-Process -Name "nw" -ErrorAction SilentlyContinue
Start-Sleep -s 1

# Schedule the watchdog to run every minute
`$action = New-ScheduledTaskAction -Execute "$watchdogRunnerPath"
`$principal = New-ScheduledTaskPrincipal -RunLevel Highest -UserId "$env:USERDOMAIN\$env:USERNAME"
`$trigger = New-ScheduledTaskTrigger -Once -At (Get-Date) -RepetitionInterval (New-TimeSpan -Minutes 1)
`$task = New-ScheduledTask -Action `$action -Trigger `$trigger -Principal `$principal
Register-ScheduledTask -TaskName "$taskName" -InputObject `$task -Force

# Run the watchdog once
Start-Process -WindowStyle Hidden PowerShell.exe -Args "-WindowStyle Hidden $watchdogPath"
"@
	SaveFile $startMonitorPath $startMonitorScript

	# Create a task that starts the watchdog monitoring on reboot (in case someone stopped it and forgot about it.)
	$action = New-ScheduledTaskAction -Execute "PowerShell.exe" -Argument "-windowstyle hidden $startMonitorScript"
	$principal = New-ScheduledTaskPrincipal -RunLevel Highest -UserId "$env:USERDOMAIN\$env:USERNAME"
	$trigger = New-ScheduledTaskTrigger -AtLogon
	$task = New-ScheduledTask -Action $action -Trigger $trigger -Principal $principal
	Register-ScheduledTask -TaskName "$rebootTaskName" -InputObject $task -Force
}

function CreateShortcuts {
	# This lets us run these powershell scripts in a powershell window to debug
	Set-ExecutionPolicy Unrestricted -force -Scope CurrentUser
	# Double clicking the shortcut opens notepad by default. Switch Windows to execute them instead
	Set-Itemproperty -path "Registry::HKEY_CLASSES_ROOT\Microsoft.PowerShellScript.1\Shell" -Name "(Default)" -value "0"

	DeleteFile $desktopShortcutPath
	
	$stopBreezeIconPath = FormatPath "$rootPath/breeze_monitor_stop_icon.ico"
	CreateShortcut $stopBreezeShortcutPath "$env:SystemRoot\System32\WindowsPowerShell\v1.0\powershell.exe" "-NoProfile -ExecutionPolicy Bypass -File $stopMonitorPath" $stopBreezeIconPath
	#ForceRequireAdmin $stopBreezeShortcutPath

	$startBreezeIconPath = FormatPath "$rootPath/breeze_monitor_start_icon.ico"
	CreateShortcut $startBreezeShortcutPath "$env:SystemRoot\System32\WindowsPowerShell\v1.0\powershell.exe" "-NoProfile -ExecutionPolicy Bypass -File $startMonitorPath" $startBreezeIconPath
	#ForceRequireAdmin $startBreezeShortcutPath
}

# This won't work for Powershell scripts that we have above! Powershell scripts are not executable.
function ForceRequireAdmin ([String]$shortcutPath) {
	# Credit to the answer by Jan Chrbolka: http://stackoverflow.com/questions/28997799/how-to-create-a-run-as-administrator-shortcut-using-powershell
	$bytes = [System.IO.File]::ReadAllBytes($shortcutPath)
	$bytes[0x15] = $bytes[0x15] -bor 0x20 #set byte 21 (0x15) bit 6 (0x20) ON
	[System.IO.File]::WriteAllBytes($shortcutPath, $bytes)
}

function ensureAdmin {
	if (-NOT $(([Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole] "Administrator"))) {
		Write-Output "Error: must be run as administrator."
		exit
	}
}

function installNode {
	# Install nodejs if its not already installed
	#Write-Output "Checking if there is a satisfactory nodejs installation... (this may produce a command not found error, you can ignore this!)"
	#$nodeVersion = node -v
	#if (!$nodeVersion) {
	#	Write-Output "Detected nodejs is not installed, downloading nodejs installer..."
	#	installNodeJs
	#}
}

function installNodejs {
	$arch = "x86";
	if ($env:PROCESSOR_ARCHITECTURE -eq "AMD64") {
		$arch = "x64"
	}
	$nodejsVersion = "v6.11.2"
	$nodejsMsiUrl = "https://nodejs.org/dist/$nodejsVersion/node-$nodejsVersion-$arch.msi"
	$nodejsMsiFilepath = FormatPath "$env:TEMP/nodejsv6_11_2.msi"
	DownloadRemoteFile $nodejsMsiUrl $nodejsMsiFilepath

	Write-Output "Installing nodejs..."
	Start-Process msiexec.exe -Wait -ArgumentList "/I $nodejsMsiFilepath /quiet"
	Write-Output "Finished installing nodejs..."
	Remove-Item $nodejsMsiFilepath
}

function installNwjs {
	$nwjsVersion = ""
	if (Test-Path $nwjsVersionPath) {
		$nwjsVersion = Get-Content $nwjsVersionPath | Out-String
		$nwjsVersion = $nwjsVersion.Replace('"', '')
	}
	
	# Identify platform information and run any platform-specific logic
	$platform = "win-ia32"
	$arch = "ia32"
	#$version = "0.39.2"
	#$version = "0.48.4"
	$version = "0.49.0"

	if ($env:PROCESSOR_ARCHITECTURE -eq "AMD64") {
		$platform = "win-x64"
		$arch = "x64"
	}

	if ($version -eq $nwjsVersion) {
		return
	}

	# Download the correct versions of nwjs.zip
	$nwjsFile = "nwjs-v$version-$platform"
	$proprietaryDllFile = "$version-$platform.zip"
	$nwjsFilepath = FormatPath "$rootPath/$nwjsFile.zip"
	$proprietaryFilepath = FormatPath "$rootPath/$proprietaryDllFile"
	#$nwjsUrl = "https://dl.nwjs.io/v$version/$nwjsFile.zip"
	#$proprietaryUrl = "https://github.com/iteufel/nwjs-ffmpeg-prebuilt/releases/download/$version/$proprietaryDllFile"
	$nwjsUrl = "https://internal.keywesttechnology.com/breeze2/mirror/$nwjsFile.zip"
	$proprietaryUrl = "https://internal.keywesttechnology.com/breeze2/mirror/$proprietaryDllFile"

	Write-Output "Downloading nwjs $version for $platform..."

	DownloadRemoteFile $nwjsUrl $nwjsFilepath
	DownloadRemoteFile $proprietaryUrl $proprietaryFilepath

	ExtractZipFile $nwjsFilepath $containerPath
	Move-Item "$containerPath/$nwjsFile/*" $containerPath -Force
	ExtractZipFile $proprietaryFilepath $containerPath

	Remove-Item $nwjsFilepath
	Remove-Item $proprietaryFilepath

	# Create chromium user data dir
	$userDataDir = FormatPath "$containerPath/chromium-config"
	DeleteFolder $userDataDir
	New-Item -Force -ItemType directory -Path $userDataDir
	# Save the version of nwjs so we know next time whether to install it
	New-Item -Force -ItemType directory -Path $configDirectory
	SaveFile $nwjsVersionPath $version
}

# Compares two version strings against one another.
# Returns:
# 	-1 when versionA represents a lower version number than versionB
# 	 0 when versionA represents the same version number as versionB
# 	 1 when versionA represents a higher version number than versionB
function CompareVersions ([String]$versionA, [String]$versionB) {
	$tokensA = $versionA.Split(".")
	if ($tokensA.length -eq 3) {
		$tokensA += "0"
	}
	for ($itr = 0; $itr -lt $tokensA.length; ++$itr) {
		$tokensA[$itr] = [int]$tokensA[$itr];
	}
	$tokensB = $versionB.Split(".")
	if ($tokensB.length -eq 3) {
		$tokensB += "0"
	}
	for ($itr = 0; $itr -lt $tokensB.length; ++$itr) {
		$tokensB[$itr] = [int]$tokensB[$itr];
	}
	if ($tokensA[0] -gt $tokensB[0]) {
		return 1;
	} elseif ($tokensA[0] -eq $tokensB[0]) {
		if ($tokensA[1] -gt $tokensB[1]) {
			return 1;
		} elseif ($tokensA[1] -eq $tokensB[1]) {
			if ($tokensA[2] -gt $tokensB[2]) {
				return 1;
			} elseif ($tokensA[2] -eq $tokensB[2]) {
				if ($tokensA[3] -gt $tokensB[3]) {
					return 1;
				} elseif ($tokensA[3] -eq $tokensB[3]) {
					return 0;
				}
			}
		}
	}
	return -1;
}

# Required directories - do not modify these path values
$rootPath = FormatPath "$env:SystemDrive/breeze"

$watchdogPath = FormatPath "$rootPath/watchdog.ps1"
$watchdogRunnerPath = FormatPath "$rootPath/watchdog_runner.vbs"
$startMonitorPath = FormatPath "$rootPath/start_monitor.ps1"
$stopMonitorPath = FormatPath "$rootPath/stop_monitor.ps1"

$desktopPath = FormatPath "$Home/Desktop"
$desktopShortcutPath = FormatPath "$desktopPath/BreezePlayer.lnk"
$stopBreezeShortcutPath = FormatPath "$desktopPath/StopBreeze.lnk"
$startBreezeShortcutPath = FormatPath "$desktopPath/StartBreeze.lnk"

$taskName = "Breeze2WatchDog"
$rebootTaskName = "Breeze2AutoWatchDog"

$containerPath = FormatPath "$rootPath/nwjscontainer"
$packagePath = FormatPath "$containerPath/package.json"

$updateDirectory = FormatPath "$scriptDirectory/.."
$configDirectory = "$containerPath/config"
$versionPath = "$configDirectory/version.json"
$updateLockPath = "$containerPath/update.lock"
$nwjsVersionPath = "$configDirectory/nwjsversion.txt"

# Logs can be obtained on Windows by running program with this format: 'nw.exe > $logPath 2>&1'
$logPath = FormatPath "$rootPath/log"
$runlogPath = FormatPath "$logPath/run.log"
$errorlogPath = FormatPath "$logPath/err.log"
$updatelogPath = FormatPath "$logPath/update.log"
