#!/bin/bash
scriptPath=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
. $scriptPath/common.sh

platform="linux-ia32"
arch="ia32"
#version="0.35.4"
#version="0.39.2"
#version="0.48.4"
version="0.49.0"
user="breeze"
if [ -f /etc/os-release ]; then
	. /etc/os-release
else
	print_error "Unable to detect file '/etc/os-release', assuming generic installation."
fi
if [ `getconf LONG_BIT` == "64" ]; then
	platform="linux-x64"
	arch="x64"
fi
# The latest nwjs build that is supported on 12.04, 12.10 and 13.10 is 22.3, which runs
# Chromium v58.0.3029.110 and Node.js v7.10.0 (https://nwjs.io/blog/v0.22.3/)
if [ $VERSION_ID == "12.04" ]; then
	version="0.22.3"
fi
if [ $VERSION_ID == "12.10" ]; then
	version="0.22.3"
fi
if [ $VERSION_ID == "13.10" ]; then
	version="0.22.3"
fi

nwjsVersion=""
if [ -e $nwjsVersionPath ]; then
	nwjsVersion="$(<$nwjsVersionPath)"
	nwjsVersion="${nwjsVersion//\"}"
fi

# If you ever change the nw.js version update the version here to the latest to force the player to reinstall nwjs. Basically we don't want to install it every time if it hasn't changed
if [ "$version" == "$nwjsVersion" ]; then
	print_info "Already at $version for $platform..."
	exit;
fi

# Download and install nwjs
(
print_info "Downloading nwjs $version for $platform..."
nwjsFile="nwjs-v$version-$platform"
proprietaryDllFile="$version-$platform"
cd $rootPath
wget "https://internal.keywesttechnology.com/breeze2/mirror/$nwjsFile.tar.gz"
wget "https://internal.keywesttechnology.com/breeze2/mirror/$proprietaryDllFile.zip"
# wget "https://dl.nwjs.io/v$version/$nwjsFile.tar.gz"
# wget "https://github.com/iteufel/nwjs-ffmpeg-prebuilt/releases/download/$version/$proprietaryDllFile.zip"
tar -xzf $rootPath/$nwjsFile.tar.gz
mkdir -p $rootPath/codecs
unzip -o $rootPath/$proprietaryDllFile.zip -d $rootPath/codecs/
cp -rf $rootPath/codecs/* $rootPath/$nwjsFile/lib/
rm -rf $rootPath/$nwjsFile.tar.gz
rm -rf $rootPath/$proprietaryDllFile.zip
rm -rf $rootPath/codecs
cp -rf $rootPath/$nwjsFile/* $containerPath
rm -rf $rootPath/$nwjsFile

# Clear the old chromium cache directory (breaks upgrades of nwjs sometimes)
chromiumCache="$containerPath/chromium-config"
rm -rf $chromiumCache
mkdir -p $chromiumCache
# Save the version of nwjs so we know next time whether to install it
echo $version > $nwjsVersionPath
)
