#!/bin/bash
function print_error {
	tput setab 0;
	tput setaf 1;
	echo "$@"
	tput sgr 0;
}

function print_info {
	tput setaf 2;
	echo "$@"
	tput sgr 0;
}

function deleteExistingConfigLines {
	# Remove any lines that contain the settings `set-default-sink` or `set-card-profile` first
	sed -i '/set-default-sink/d' /etc/pulse/default.pa
	sed -i '/set-card-profile/d' /etc/pulse/default.pa
}

if [[ $EUID -ne 0 ]]; then
	print_error "This script must be run as root" 1>&2
	exit 1
fi

configPath=/etc/pulse/default.pa
backupPath=$configPath.backup

print_info "This script attempts to set the pulse audio configuration defaults to known values.  Currently, it does not query for the local machines valid values.  This MAY NOT WORK (as in, upon reboot there may be NO sound or output devices.  If this happens run 'sudo cp -f $backupPath $configPath' and reboot the machine.)"
print_info "At the time of this writing ONLY the following hardware configurations are compatible (still, they may not work!):"
print_info "------------"
print_info "GIADA 12.10"
print_info "ASRock 13.10"
print_info "Aeon 12.04.5"
print_info "------------"
print_info ""

print_info "Creating backup of default configuration..."
cp -f $configPath $backupPath

if dmidecode | grep -q GIADA; then
	print_info "Auto-detected a GIADA motherboard..."
	deleteExistingConfigLines
	echo "line 1" >> greetings.txt
	echo "set-default-sink alsa_output.pci-0000_00_1b.0.analog-stereo" >> $configPath
	echo "set-card-profile 0 output:analog-stereo+input:analog-stereo" >> $configPath
elif dmidecode | grep -q ASRock; then
	print_info "Auto-detected an ASRock motherboard..."
	deleteExistingConfigLines
	echo "set-default-sink alsa_output.pci-0000_00_1b.0.hdmi-stereo" >> $configPath
	echo "set-card-profile 0 output:analog-stereo+input:analog-stereo" >> $configPath
else
	print_info "Could not detect GIADA or ASRock motherboards, assuming it's an Aeon motherboard..."
	deleteExistingConfigLines
	echo "set-default-sink alsa_card.pci-0000_00_1b.0" >> $configPath
	echo "set-card-profile 0 output:analog-stereo+input:analog-stereo" >> $configPath
fi
