#!/bin/bash
# If this script is run by the Breeze 2.x container, then the cwd will be set such that the directories look like this:
# - ..
# - breeze2nwjscontainer
# 	- container
# 		- package.json
#		- node_modules
#		- src
#	- scripts
updateDirectory="./breeze2nwjscontainer"
. $updateDirectory/scripts/common.sh

# Old versions of Breeze do not generate an update.lock file
if [[ ! -e $updateLockPath ]]; then
	# Infer the current patch version and create and update.lock file
	canonicalUpdateDirectory="$(readlink -f $updateDirectory)"

	# The version number is in the $canonicalUpdateDirectory as ../config/updates/xx_xx_xx_xx/..
	versionString=$(echo "$canonicalUpdateDirectory" | grep -o -E 'updates\/(_?[0-9]+_?){3,4}')
	versionString=${versionString#updates\/}
	versionString=${versionString//_/.}
	versionString=\"$versionString\"
	echo $versionString > $updateLockPath

	# Try to kill breeze
	pkill nw
fi

# Try to wait for NWJS to end on its own
# After a minute give up and try to kill it
killNwjs=true
itr=0
while [ $itr -lt 10 ]; do
	if [ "\$(pidof nw)" ]; then
		sleep 6s
	else
		killNwjs=
		break
	fi
	itr=$((itr + 1))
done
if [[ "$killNwjs" = true ]]; then
	pkill nw
	sleep 1s
fi

# Preform logic that is specific to certain updates if applicable
version="00.00.00"
if [ -e $versionPath ] ; then
	version="$(<$versionPath)"
	version="${version//\"}"
fi
freshInstall=
if [[ $(CompareVersions $version "00.00.00") -eq 0 ]] ; then
	freshInstall=true
	echo "Assuming this is a fresh install.."
fi

InstallNode
CopyFiles
InstallNwjs
CreateLogs
CreateWatchdog
CreateShortcuts

# Update the version.json file to the latest version (stored in the update.lock file)
touch $versionPath # Prevent an error from being logged if this is the first update and $versionPath doesn't exist.
rm $versionPath
cp $updateLockPath $versionPath

# Remove the update.lock file and re-launch breeze
rm $updateLockPath

nohup "$rootPath/watchdog.sh" 1>>$outlogPath 2>>$errlogPath &
