const FILLSTYLE_NONE = 1;
const FILLSTYLE_COVER = 2;
const FILLSTYLE_ASPECT_RATIO = 3;

const MOVETYPE_NONE = 0;
const MOVETYPE_MOVE_X = 1;
const MOVETYPE_MOVE_Y = 1 << 1;
const MOVETYPE_RESIZE_LEFT = 1 << 2;
const MOVETYPE_RESIZE_RIGHT = 1 << 3;
const MOVETYPE_RESIZE_TOP = 1 << 4;
const MOVETYPE_RESIZE_BOTTOM = 1 << 5;

var gCurrentPlaylist = null;
var gSelectedZone = null;
var gTimeBarSelection = 0;

var gSaveFlag = false;
var gShrink = false;
var gTimelineScale = 5;
var gExpand = false;
var gSelectMediaLocation = '';

var gCurrentWidget = null;
var gCurrentWidgetComponent = null;

var convertingFilesTimer = null;

let $contentTree;
let $editorCanvasContainer;
let $editorLayout;
let $editorPreviewControls;
let $editorTimelineZonesToolbar;
let $canvasSlider;
let $playStop;
let $editorCanvas;
let $lineLayer;
let $collapseButtonGroups;
let $buttonTab;
let $timelineView;
let $zoneView;
let $timelineOptions;
let $editorCanvasShadow;
let $timelineNameResize;
let $nameFilter;
let $imageFilter;
let $videoFilter;
let $audioFilter;
let $pdfFilter;
let $timeBarButtonZoneHighlight;
let $zoneSelect;
let $mediaSelect;

// The program can be in either 'zone' or 'media' view.
var gPropertiesView = 'zone';

function SetSaveFlag(value) {
	gSaveFlag = value;
	setWindowHeader();
}

function DialogWindowOpen() {
	return document.querySelector('ui-dialog:not([hidden])') != null;
}

function drawResizeBox(context, x, y, width, height, color, moveType, boundingMinX, boundingMinY, boundingMaxX, boundingMaxY) {
	x = Math.round(x);
	y = Math.round(y);
	width = Math.round(width);
	height = Math.round(height);
	if (boundingMinX != undefined) {
		boundingMinX = Math.round(boundingMinX);
		boundingMinY = Math.round(boundingMinY);
		boundingMaxX = Math.round(boundingMaxX);
		boundingMaxY = Math.round(boundingMaxY);
	}
	const scrollLeft = Math.round($editorCanvasContainer.scrollLeft);
	const scrollTop = Math.round($editorCanvasContainer.scrollTop);
	context.save();
	context.fillStyle = color;
	context.strokeStyle = color;
	context.translate(-scrollLeft, -scrollTop);
	const drawSelector = (x, y, focus) => {
		const thickness = focus ? 3 : 2;
		context.fillRect(x - thickness, y - thickness, thickness * 2 + 1, thickness * 2 + 1);
	};
	drawSelector(x, y, moveType == (MOVETYPE_RESIZE_TOP | MOVETYPE_RESIZE_LEFT));
	drawSelector(x + width, y, moveType == (MOVETYPE_RESIZE_TOP | MOVETYPE_RESIZE_RIGHT));
	drawSelector(x, y + height, moveType == (MOVETYPE_RESIZE_BOTTOM | MOVETYPE_RESIZE_LEFT));
	drawSelector(x + width, y + height, moveType == (MOVETYPE_RESIZE_BOTTOM | MOVETYPE_RESIZE_RIGHT));
	drawSelector(x + (width / 2), y, moveType == MOVETYPE_RESIZE_TOP);
	drawSelector(x, y + (height / 2), moveType == MOVETYPE_RESIZE_LEFT);
	drawSelector(x + (width / 2), y + height, moveType == MOVETYPE_RESIZE_BOTTOM);
	drawSelector(x + width, y + (height / 2), moveType == MOVETYPE_RESIZE_RIGHT);
	context.restore();
	drawBox(context, x, y, width, height, color, boundingMinX, boundingMinY, boundingMaxX, boundingMaxY);
}

function drawBox(context, x, y, width, height, color, boundingMinX, boundingMinY, boundingMaxX, boundingMaxY) {
	x = Math.round(x);
	y = Math.round(y);
	width = Math.round(width);
	height = Math.round(height);
	if (boundingMinX != undefined) {
		boundingMinX = Math.round(boundingMinX);
		boundingMinY = Math.round(boundingMinY);
		boundingMaxX = Math.round(boundingMaxX);
		boundingMaxY = Math.round(boundingMaxY);
	}
	var scrollLeft = Math.round($editorCanvasContainer.scrollLeft);
	var scrollTop = Math.round($editorCanvasContainer.scrollTop);
	context.save();
	context.translate(-scrollLeft + 0.5, -scrollTop + 0.5);
	context.strokeStyle = color;
	if (boundingMinX == undefined || x >= boundingMinX && x + width <= boundingMaxX && y >= boundingMinY && y + height <= boundingMaxY) {
		context.strokeRect(x, y, width, height);
	} else {
		const drawLine = (x1, y1, x2, y2) => {
			context.beginPath();
			context.moveTo(x1, y1);
			context.lineTo(x2, y2);
			context.stroke();
		};
		// Draw dashed lines outside of the bounding rectangle and solid inside
		const drawLineWithBoundingRect = (x1, y1, x2, y2) => {
			// Could simplify this by mirroring along y = x, but it might confuse future maintainers
			if (y1 == y2) {
				// Horizontal
				if (y1 < boundingMinY || y1 > boundingMaxY || x2 < boundingMinX || x1 > boundingMaxX) {
					context.save();
					context.setLineDash([2, 2]);
					drawLine(x1, y1, x2, y2);
					context.restore();
				} else {
					context.save();
					context.setLineDash([2, 2]);
					if (x1 < boundingMinX) {
						drawLine(x1, y1, boundingMinX, y2);
					}
					if (x2 > boundingMaxX) {
						drawLine(boundingMaxX, y1, x2, y2);
					}
					context.restore();
					drawLine(Math.max(x1, boundingMinX), y1, Math.min(x2, boundingMaxX), y2);
				}
			} else {
				// Vertical
				if (x1 < boundingMinX || x1 > boundingMaxX || y2 < boundingMinY || y1 > boundingMaxY) {
					context.save();
					context.setLineDash([2, 2]);
					drawLine(x1, y1, x2, y2);
					context.restore();
				} else {
					context.save();
					context.setLineDash([2, 2]);
					if (y1 < boundingMinY)
					{
						drawLine(x1, y1, x2, boundingMinY);
					}
					if (y2 > boundingMaxY)
					{
						drawLine(x1, boundingMaxY, x2, y2);
					}
					context.restore();
					drawLine(x1, Math.max(y1, boundingMinY), x2, Math.min(y2, boundingMaxY));
				}
			}
		};
		drawLineWithBoundingRect(x, y, x + width, y); // Top
		drawLineWithBoundingRect(x, y + height, x + width, y + height); // Bottom
		drawLineWithBoundingRect(x, y, x, y + height); // Left
		drawLineWithBoundingRect(x + width, y, x + width, y + height); // Right
	}
	context.restore();
}

function GetLineLayer() {
	const context = $lineLayer.getContext('2d');
	return { canvas: $lineLayer, context };
}

function LineLayerClear() {
	const { canvas, context } = GetLineLayer();
	context.clearRect(0, 0, canvas.width, canvas.height);
}

function LineLayerDrawSnapLines(snapX, snapY) {
	if (gCurrentPlaylist) {
		const { canvas, context } = GetLineLayer();
		for (const zone of gCurrentPlaylist.zones) {
			if (zone !== gSelectedZone) {
				LineLayerHighlightZone(zone, false);
			}
		}
		
		var SetStrokeByType = type => {
			switch (type) {
				case 'monitoredge':
					context.strokeStyle = 'rgba(139, 197, 63, 0.8)';
					context.shadowColor = 'rgba(139, 197, 63, 0.8)';
					context.lineWidth = 2;
					break;
				case 'playlistedge':
					context.strokeStyle = 'rgba(139, 197, 63, 1)';
					context.shadowColor = 'rgba(139, 197, 63, 1)';
					context.lineWidth = 5;
					break;
				case 'zoneedge':
					context.strokeStyle = 'rgba(139, 197, 63, 0.8)';
					context.shadowColor = 'rgba(139, 197, 63, 0.8)';
					break;
				case 'zonemiddle':
					context.strokeStyle = 'rgba(139, 197, 63, 0.6)';
					context.shadowColor = 'rgba(139, 197, 63, 0.6)';
					break;
				case 'mediaedge':
					context.strokeStyle = 'rgba(27, 117, 187, 0.7)';
					context.shadowColor = 'rgba(27, 117, 187, 0.7)';
					break;
				case 'mediamiddle':
					context.strokeStyle = 'rgba(27, 117, 187, 0.5)';
					context.shadowColor = 'rgba(27, 117, 187, 0.5)';
					break;
				case 'mediacomponentedge':
					context.strokeStyle = 'rgba(181, 222, 255, 0.7)';
					context.shadowColor = 'rgba(181, 222, 255, 0.7)';
					break;
				case 'mediacomponentmiddle':
					context.strokeStyle = 'rgba(181, 222, 255, 0.5)';
					context.shadowColor = 'rgba(181, 222, 255, 0.5)';
					break;
			}
		};
		
		context.save();
		var scrollLeft = Math.round($editorCanvasContainer.scrollLeft);
		var scrollTop = Math.round($editorCanvasContainer.scrollTop);
		context.translate(-scrollLeft + 0.5, -scrollTop + 0.5);
		for (const { value, type, highlight } of snapX) {
			if (highlight) {
				context.save();
				context.lineWidth = 1;
				context.shadowBlur = 2;
				context.setLineDash([5, 5]);
				SetStrokeByType(type);
				context.beginPath();
				context.moveTo(Math.round(toCanvasWidth(value)), 0);
				context.lineTo(Math.round(toCanvasWidth(value)), canvas.height);
				context.stroke();
				context.restore();
			}
		}
		for (const { value, type, highlight } of snapY) {
			if (highlight) {
				context.save();
				context.lineWidth = 1;
				context.shadowBlur = 2;
				context.setLineDash([5, 5]);
				SetStrokeByType(type);
				context.beginPath();
				context.moveTo(0, Math.round(toCanvasHeight(value)));
				context.lineTo(canvas.width, Math.round(toCanvasHeight(value)));
				context.stroke();
				context.restore();
			}
		}
		context.restore();
	}
}

function LineLayerHighlightZone(zone, focus) {
	if (zone) {
		const { context } = GetLineLayer();
		drawBox(context, toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.width), toCanvasHeight(zone.height), focus ? 'rgba(139, 197, 63, 1)' : 'rgba(139, 197, 63, 0.8)');
		var media = zone.GetSelectedMedia();
		LineLayerHighlightMedia(zone, media, zone === gSelectedZone);
		if (media) {
			if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
				for (const button of media.attributes.buttons) {
					LineLayerHighlightMediaComponent(zone, media, button, button === gCurrentWidgetComponent);
				}
			}
		}
	}
}

function LineLayerHighlightMedia(zone, media, focus) {
	if (media && (media.x != 0 || media.y != 0 || media.width != zone.width || media.height != zone.height)) {
		const { context } = GetLineLayer();
		drawBox(context, toCanvasWidth(zone.x + media.x), toCanvasHeight(zone.y + media.y), toCanvasWidth(media.width), toCanvasHeight(media.height), focus ? 'rgba(27, 117, 187, 1)' : 'rgba(27, 117, 187, 0.8)', toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.x + zone.width), toCanvasHeight(zone.y + zone.height));
	}
}

function LineLayerHighlightMediaComponent(zone, media, mediaComponent, focus) {
	const { context } = GetLineLayer();
	drawBox(context,
		toCanvasWidth(zone.x + media.x + mediaComponent.x),
		toCanvasHeight(zone.y + media.y + mediaComponent.y),
		toCanvasWidth(mediaComponent.width),
		toCanvasHeight(mediaComponent.height),
		focus ? 'rgba(181, 222, 255, 1)' : 'rgba(181, 222, 255, 0.8)',
		Math.max(toCanvasWidth(zone.x),toCanvasWidth(zone.x + media.x)),
		Math.max(toCanvasHeight(zone.y), toCanvasHeight(zone.y + media.y)),
		Math.min(toCanvasWidth(zone.x + zone.width),toCanvasWidth(zone.x + media.x + media.width)),
		Math.min(toCanvasHeight(zone.y + zone.height), toCanvasHeight(zone.y + media.y + media.height)));
}

function LineLayerResizeBoxZone(zone, moveType) {
	if (zone) {
		const { context } = GetLineLayer();
		drawResizeBox(context, toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.width), toCanvasHeight(zone.height), 'rgba(139, 197, 63, 1)', moveType);
		var media = zone.GetSelectedMedia();
		LineLayerHighlightMedia(zone, media, zone === gSelectedZone);
		if (media) {
			if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
				for (const button of media.attributes.buttons) {
					LineLayerHighlightMediaComponent(zone, media, button, button === gCurrentWidgetComponent);
				}
			}
		}
	}
}

function LineLayerResizeBoxMedia(zone, media, moveType) {
	if (media) {
		const { context } = GetLineLayer();
		drawResizeBox(context, toCanvasWidth(zone.x + media.x), toCanvasHeight(zone.y + media.y), toCanvasWidth(media.width), toCanvasHeight(media.height), 'rgba(27, 117, 187, 1)', moveType, toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.x + zone.width), toCanvasHeight(zone.y + zone.height));
		drawBox(context, toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.width), toCanvasHeight(zone.height), zone === gSelectedZone ? 'rgba(139, 197, 63, 1)' : 'rgba(139, 197, 63, 0.8)');
		if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
			for (const button of media.attributes.buttons) {
				LineLayerHighlightMediaComponent(zone, media, button, button === gCurrentWidgetComponent);
			}
		}
	}
}

function LineLayerResizeBoxMediaComponent(zone, media, mediaComponent, moveType) {
	if (media) {
		const { context } = GetLineLayer();
		drawResizeBox(context,
			toCanvasWidth(zone.x + media.x + mediaComponent.x),
			toCanvasHeight(zone.y + media.y + mediaComponent.y),
			toCanvasWidth(mediaComponent.width),
			toCanvasHeight(mediaComponent.height),
			'rgba(181, 222, 255, 1)',
			moveType,
			Math.max(toCanvasWidth(zone.x),toCanvasWidth(zone.x + media.x)),
			Math.max(toCanvasHeight(zone.y), toCanvasHeight(zone.y + media.y)),
			Math.min(toCanvasWidth(zone.x + zone.width),toCanvasWidth(zone.x + media.x + media.width)),
			Math.min(toCanvasHeight(zone.y + zone.height), toCanvasHeight(zone.y + media.y + media.height)));
		LineLayerHighlightMedia(zone, media, true);
		drawBox(context, toCanvasWidth(zone.x), toCanvasHeight(zone.y), toCanvasWidth(zone.width), toCanvasHeight(zone.height), zone === gSelectedZone ? 'rgba(139, 197, 63, 1)' : 'rgba(139, 197, 63, 0.8)');
	}
}

function isPointInRect(x, y, rectX, rectY, rectWidth, rectHeight) {
	return x >= rectX && y >= rectY && x < rectX + rectWidth && y < rectY + rectHeight;
}

function ResizeTimelineBars() {
	if (gCurrentPlaylist == null) {
		return;
	}
	const maxWidth = Math.max(0, ...gCurrentPlaylist.zones.map(zone => zone.media.reduce((duration, media) => { return duration + media.runlength; }, 0) * gTimelineScale));
	const width = `${maxWidth + Math.max($timelineView.offsetWidth, $zoneView.offsetWidth)}px`;
	for (const zone of gCurrentPlaylist.zones) {
		zone.$timelineBar.style.width = width;
	}
}

function GetPlaylistSnapLines(mediaMode, mediaComponentMode) {
	var snapX = [];
	var snapY = [];
	if (gCurrentPlaylist) {
		for (const zone of gCurrentPlaylist.zones.filter(zone => {
			if (!zone.visible) {
				return false;
			}
			return mediaMode || mediaComponentMode || zone !== gSelectedZone;
		})) {
			const media = zone.GetSelectedMedia();
			if ((!mediaMode || mediaComponentMode || zone !== gSelectedZone) && media) {
				if (media.x != 0) {
					snapX.push({ value: zone.x + media.x, type: 'mediaedge' });
				}
				if (media.x != 0 && media.x + media.width != zone.width) {
					snapX.push({ value: Math.round(zone.x + media.x + media.width / 2), type: 'mediamiddle' });
				}
				if (media.x + media.width != zone.width) {
					snapX.push({ value: zone.x + media.x + media.width, type: 'mediaedge' });
				}
				if (media.y != 0) {
					snapY.push({ value: zone.y + media.y, type: 'mediaedge' });
				}
				if (media.y != 0 && media.y + media.height != zone.height) {
					snapY.push({ value: Math.round(zone.y + media.y + media.height / 2), type: 'mediamiddle' });
				}
				if (media.y + media.height != zone.height) {
					snapY.push({ value: zone.y + media.y + media.height, type: 'mediaedge' });
				}
				// Buttons can be snapped to individually
				if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
					for (const button of media.attributes.buttons) {
						if (button !== gCurrentWidgetComponent) {
							snapX.push({ value: zone.x + media.x + button.x, type: 'mediacomponentedge' });
							snapX.push({ value: Math.round(zone.x + media.x + button.x + button.width / 2), type: 'mediacomponentmiddle' });
							snapX.push({ value: zone.x + media.x + button.x + button.width, type: 'mediacomponentedge' });
							
							snapY.push({ value: zone.y + media.y + button.y, type: 'mediacomponentedge' });
							snapY.push({ value: Math.round(zone.y + media.y + button.y + button.height / 2), type: 'mediacomponentmiddle' });
							snapY.push({ value: zone.y + media.y + button.y + button.height, type: 'mediacomponentedge' });
						}
					}
				}
			}
			snapX.push({ value: zone.x, type: 'zoneedge' });
			snapX.push({ value: Math.round(zone.x + zone.width / 2), type: 'zonemiddle' });
			snapX.push({ value: zone.x + zone.width, type: 'zoneedge' });
			
			snapY.push({ value: zone.y, type: 'zoneedge' });
			snapY.push({ value: Math.round(zone.y + zone.height / 2), type: 'zonemiddle' });
			snapY.push({ value: zone.y + zone.height, type: 'zoneedge' });
		}
		// Multiple monitor
		const commonMonitorResolutions =
		[
			// Landscape
			{ width: 1280, height: 720 },
			{ width: 1920, height: 1080 },
			{ width: 3840, height: 2160 },
			// Portrait
			{ width: 720, height: 1280 },
			{ width: 1080, height: 1920 },
			{ width: 2160, height: 3840 }
		];
		for (const resolution of commonMonitorResolutions) {
			// Monitor layout horizontal width x 1
			if (gCurrentPlaylist.pagewidth != resolution.width && gCurrentPlaylist.pageheight == resolution.height && gCurrentPlaylist.pagewidth % resolution.width == 0) {
				for (var x = resolution.width; x < gCurrentPlaylist.pagewidth; x += resolution.width) {
					snapX.push({ value: x, type: 'monitoredge' });
				}
			}
			// Monitor layout vertical 1 x height
			if (gCurrentPlaylist.pageheight != resolution.height && gCurrentPlaylist.pagewidth == resolution.width && gCurrentPlaylist.pageheight % resolution.height == 0) {
				for (var y = resolution.height; y < gCurrentPlaylist.pageheight; y += resolution.height) {
					snapY.push({ value: y, type: 'monitoredge' });
				}
			}
			// Monitors in a width x height configuration
			for (var width = 2; width < 6; ++width) {
				for (var height = 2; height < 6; ++height) {
					if (gCurrentPlaylist.pagewidth == resolution.width * width && gCurrentPlaylist.pageheight == resolution.height * height) {
						for (var x = 1; x < width; ++x) {
							snapX.push({ value: x * resolution.width, type: 'monitoredge' });
						}
						for (var y = 1; y < height; ++y) {
							snapY.push({ value: y * resolution.height, type: 'monitoredge' });
						}
					}
				}
			}
		}
		// Sort and cleanup snaps
		snapX.sort((a, b) => a.value - b.value);
		// Remove duplicates
		for (var itr = 0; itr < snapX.length - 1; ++itr) {
			if (snapX[itr].value == 0 || snapX[itr].value == gCurrentPlaylist.pagewidth || snapX[itr].value == snapX[itr + 1].value) {
				snapX.splice(itr--, 1);
			}
		}
		snapX.push({ value: 0, type: 'playlistedge' });
		snapX.push({ value: gCurrentPlaylist.pagewidth, type: 'playlistedge' });
		snapX.sort((a, b) => a.value - b.value);
		
		snapY.sort((a, b) => a.value - b.value);
		for (var itr = 0; itr < snapY.length - 1; ++itr) {
			if (snapY[itr].value == 0 || snapY[itr].value == gCurrentPlaylist.pageheight || snapY[itr].value == snapY[itr + 1].value) {
				snapY.splice(itr--, 1);
			}
		}
		snapY.push({ value: 0, type: 'playlistedge' });
		snapY.push({ value: gCurrentPlaylist.pageheight, type: 'playlistedge' });
		snapY.sort((a, b) => a.value - b.value);
		// Set highlight state to false
		for (const snap of snapX) {
			snap.highlight = false;
		}
		for (const snap of snapY) {
			snap.highlight = false;
		}
	}
	return [snapX, snapY];
}

function expandCurrentMedia() {
	for (const media of gSelectedZone.GetAllSelectedMedia()) {
		media.Expand();
	}
}

function clearEditorCanvas() {
	$editorCanvas.innerHTML = '';
}

function createPlaylistObject(id, name, runLength, pageWidth, pageHeight) {
	var playlist = {};
	playlist.id = id;
	playlist.name = name;
	playlist.runlength = runLength;
	playlist.pagewidth = pageWidth;
	playlist.pageheight = pageHeight;
	playlist.zones = [];
	
	playlist.IsInteractive = function() {
		return this.zones.some(zone => {
			return zone.media.some(media => {
				return media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup';
			});
		});
	};
	
	return playlist;
}

class Zone {
	constructor(id, name, x, y, z, width, height, volume, locked) {
		this.id = id;
		this.name = name;
		this.x = x;
		this.y = y;
		this.z = z;
		this.width = width;
		this.height = height;
		this.volume = volume;
		this.locked = locked;
		this.media = [];
		this.mediaSelected = [];
		this.visible = true;
		
		this.$canvasView = $editorCanvas.appendChild(
			$.div({ class: 'editor-canvas-zone', css: {
				display: 'inline-block',
				position: 'absolute',
				overflow: 'hidden'
			} })
		);
		
		this.$propertyView = $.option({ text: this.name, value: this.name });
		
		this.$timelineName = $.span({ class: 'timelineName', tabindex: 0, multiclick: (e, clicks) => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			const $input = this.$timelineName.querySelector('input');
			if (!$input.classList.contains('edit')) {
				if (clicks == 1) {
					SelectZone(this, true, false);
				} else if (clicks == 2) {
					if (this.name != 'Audio') {
						$input.classList.add('edit');
						const saveZoneName = e => {
							if (e.type == 'blur' || e.type == 'keydown' && e.key == 'Enter') {
								$input.removeEventListener('blur', saveZoneName);
								$input.removeEventListener('keydown', saveZoneName);
								$input.classList.remove('edit');
								if ($input.value != 'Audio') {
									this.name = $input.value;
									updateZoneLayers();
								}
								$input.blur();
								if ('preventDefault' in e) e.preventDefault(); // This is here due to insane design choices in SelectZone that force blur changes. I forgot why.
								UpdateProperties();
								UpdateButtonList();
							}
						};
						$input.addEventListener('blur', saveZoneName);
						$input.addEventListener('keydown', saveZoneName);
						$input.focus();
					}
				} else if (clicks == 3) {
					SelectMedia(this, this.media.map((media, mediaItr) => mediaItr), false, false, false, 'timeline', true, true);
				}
			}
		}, keyup: e => {
			if (DialogWindowOpen()) {
				return;
			}
			stopPreviewPressed();
			switch (e.key) {
				case 'Delete':
					deleteCurrentZonePressed();
					e.preventDefault();
					break;
			}
		}, draggable: {
			thresholdY: 3,
			sortable: 'y',
			scrollParent: '#timeline-view',
			start: (e, startX, startY, startOffsetX, startOffsetY) => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return false;
				}
				if (this.name == 'Audio') {
					return false;
				}
				if (e.target.classList.contains('edit')) {
					return false;
				}
			},
			movePrevious: this.MovePrevious.bind(this),
			moveNext: this.MoveNext.bind(this)
		} },
			$.input({ type: 'text', class: 'name', value: this.name, keydown: e => e.stopPropagation(), keyup: e => e.stopPropagation() }),
			$.i({ class: 'tooltip-button hide-button mdi', click: e => {
				e.stopPropagation();
				if (DialogWindowOpen()) {
					BlinkDialog();
					return;
				}
				ToggleZoneVisibility(this);
				if (this.visible) {
					SelectZone(this, true, false);
				}
			} }),
			$.span({ text: this.z, class: 'order', css: { visibility: this.name == 'Audio' ? 'hidden' : 'visible' } })
		);
		if (this.name == 'Audio') {
			this.$timelineName.style.cursor = 'pointer';
		} else {
			this.$timelineName.title = 'Double click to rename. Drag to change zone order';
		}
		
		this.$timelineBar = $.div({ class: 'timelineBar', dblclick: e => {
			e.stopPropagation();
			// Double click on the empty area of a zone selects all
			SelectMedia(this, this.media.map((media, mediaItr) => mediaItr), false, false, false, null, true, true);
		}, pointerover: () => {
			LineLayerClear();
			LineLayerHighlightZone(this, false);
			LineLayerHighlightZone(gSelectedZone, true);
		}, pointerout: () => {
			LineLayerClear();
			LineLayerHighlightZone(gSelectedZone, true);
		}, droppable: {
			accept: ['content-group', 'content-leaf', 'widget', 'moved-media'],
			drop: ContentDrop,
			region: 'content'
		} });
		
		this.$zoneViewName = $.header({ multiclick: (e, clicks) => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			if (!this.$zoneViewNameInput.classList.contains('edit')) {
				if (clicks == 1) {
					SelectZone(this, true, false);
				} else if (clicks == 2) {
					if (this.name != 'Audio') {
						this.$zoneViewNameInput.classList.add('edit');
						const saveZoneName = e => {
							if (e.type == 'blur' || e.type == 'keydown' && e.key == 'Enter') {
								this.$zoneViewNameInput.removeEventListener('blur', saveZoneName);
								this.$zoneViewNameInput.removeEventListener('keydown', saveZoneName);
								this.$zoneViewNameInput.classList.remove('edit');
								if (this.$zoneViewNameInput.value != 'Audio') {
									this.name = this.$zoneViewNameInput.value;
									updateZoneLayers();
								}
								this.$zoneViewNameInput.blur();
								if ('preventDefault' in e) e.preventDefault(); // This is here due to insane design choices in SelectZone that force blur changes. I forgot why.
								UpdateProperties();
								UpdateButtonList();
							}
						};
						this.$zoneViewNameInput.addEventListener('blur', saveZoneName);
						this.$zoneViewNameInput.addEventListener('keydown', saveZoneName);
						this.$zoneViewNameInput.focus();
					}
				}
			}
		} },
			this.$zoneViewNameInput = $.input({ type: 'text', class: 'name', value: this.name }),
			$.i({ class: 'tooltip-button hide-button mdi', click: () => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return;
				}
				ToggleZoneVisibility(this);
				if (this.visible) {
					SelectZone(this, true, false);
				}
			} }),
			$.i({ title: 'Move selected media earlier', class: 'tooltip-button mdi mdi-arrow-up', click: e => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return;
				}
				SelectZone(this, true, false);
				if (moveCurrentMediaBefore(e)) {
					updateCurrentZoneMediaOrder();
					UpdateProperties();
				}
			} }),
			$.i({ title: 'Move selected media later', class: 'tooltip-button mdi mdi-arrow-down', click: e => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return;
				}
				SelectZone(this, true, false);
				if (moveCurrentMediaAfter(e)) {
					updateCurrentZoneMediaOrder();
					UpdateProperties();
				}
			} }),
			$.i({ title: 'Delete selected media', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return;
				}
				SelectZone(this, true, false);
				if (gSelectedZone.media.length == 0) {
					deleteCurrentZone();
				} else {
					deleteCurrentMedia(true);
				}
			} })
		);

		if (this.name == 'Audio') {
			this.$zoneViewName.style.cursor = 'pointer';
		} else {
			this.$zoneViewName.title = 'Double click to rename. Drag to change zone order';
		}
		
		this.$zoneViewMedia = $.ul({ title: 'You can select multiple items here, just like on the timeline, by holding ctrl for individual items or shift for adjacent items', class: 'zone-view-item-media tooltip', keyup: e => {
			if (e.keyCode == 46) {
				deleteCurrentMedia(true);
				e.preventDefault();
			}
		}, droppable: {
			accept: ['content-group', 'content-leaf', 'widget', 'moved-media'],
			drop: ContentDrop,
			region: 'content'
		} });
		
		this.$zoneView = $.div({ class: 'zone-view-item', draggable: {
			scrollParent: '#zone-view > div',
			handle: ':scope > .panel > header',
			thresholdX: 3,
			sortable: 'x',
			start: (e, startX, startY, startOffsetX, startOffsetY) => {
				if (DialogWindowOpen()) {
					BlinkDialog();
					return false;
				}
				if (this.name == 'Audio') {
					return false;
				}
				if (e.target.classList.contains('edit')) {
					return false;
				}
			},
			movePrevious: this.MovePrevious.bind(this),
			moveNext: this.MoveNext.bind(this)
		}, pointerover: () => {
			LineLayerClear();
			LineLayerHighlightZone(this, false);
			LineLayerHighlightZone(gSelectedZone, true);
		}, pointerout: () => {
			LineLayerClear();
			LineLayerHighlightZone(gSelectedZone, true);
		} },
			$.div({ class: 'panel', css: {
				width: this.hasOwnProperty('zoneViewWidth') && this.zoneViewWidth != 0 ? `${this.zoneViewWidth}px` : ''
			} },
				this.$zoneViewName,
				this.$zoneViewMedia
			),
			this.$zoneViewResize = $.div({ class: 'resize', draggable: {
				scrollParent: '#zone-view > div',
				drag: (e, startX, startY, startOffsetX, startOffsetY) => {
					var width = Math.ceil(e.pageX - startOffsetX - $.offset(this.$zoneViewResize.previousElementSibling).left);
					this.$zoneViewResize.previousElementSibling.style.width = `${width}px`;
					this.zoneViewWidth = width;
					SaveZoneResize();
				}
			} })
		);
	}
	Remove() {
		this.$canvasView.remove();
		this.$propertyView.remove();
		this.$timelineName.remove();
		this.$timelineBar.remove();
		this.$zoneView.remove();
	}
	MovePrevious() {
		var index = gCurrentPlaylist.zones.indexOf(this);
		if (gCurrentPlaylist.zones[0].name == 'Audio' && (localStorage.swapzoneorder == 'false' && index == 1 || localStorage.swapzoneorder == 'true' && index == gCurrentPlaylist.zones.length - 1)) {
			return false;
		}
		if (localStorage.swapzoneorder == 'false') {
			MoveZoneLayer(this, -1);
		} else {
			MoveZoneLayer(this, 1);
		}
		return false; // Stop the automatic tag movement by the draggable extension. We handled it manually
	}
	MoveNext() {
		if (localStorage.swapzoneorder == 'false') {
			MoveZoneLayer(this, 1);
		} else {
			MoveZoneLayer(this, -1);
		}
		return false; // Stop the automatic tag movement by the draggable extension. We handled it manually
	}
	GetSelectedMediaIndex() {
		if (this.mediaSelected.length == 0) {
			return null;
		}
		return this.media.indexOf(this.mediaSelected[this.mediaSelected.length - 1]);
	}
	GetAllSelectedMedia() {
		var selectedMedia = [];
		for (const index of this.GetSortedMediaSelectedIndexes()) {
			selectedMedia.push(this.media[index]);
		}
		return selectedMedia;
	}
	GetSelectedMedia() {
		if (this.mediaSelected.length == 0) {
			return null;
		}
		return this.mediaSelected[this.mediaSelected.length - 1];
	}
	GetMediaType(allMedia) {
		if (allMedia.length == 0) {
			return null;
		}
		var type = null;
		for (const media of allMedia) {
			if (type == null || media.type == type) {
				type = media.type;
			} else {
				type = null;
				break;
			}
		}
		return type;
	}
	GetSelectedMediaType() {
		if (this.mediaSelected.length == 0) {
			return null;
		}
		var type = null;
		for (const media of this.mediaSelected) {
			if (type == null || media.type == type) {
				type = media.type;
			} else {
				type = null;
				break;
			}
		}
		return type;
	}
	GetSelectedWidgetType() {
		if (this.mediaSelected.length == 0) {
			return null;
		}
		var widgetType = null;
		for (const media of this.mediaSelected) {
			if (media.type == MEDIATYPE_WIDGET && (widgetType == null || media.widgettype == widgetType)) {
				widgetType = media.widgettype;
			} else {
				widgetType = null;
				break;
			}
		}
		return widgetType;
	}
	GetSelectedMediaIndexes() {
		return this.mediaSelected.map(mediaSelected => {
			return this.media.indexOf(mediaSelected);
		});
	}
	GetSortedMediaSelectedIndexes() {
		return this.mediaSelected.map(mediaSelected => {
			return this.media.indexOf(mediaSelected);
		}).sort((a, b) => a - b);
	}
	IsMediaSelectedSequential() {
		var sequential = true;
		var mediaSelectedIndexes = this.GetSortedMediaSelectedIndexes();
		var lastMediaSelectedIndex = mediaSelectedIndexes[0];
		for (var itr = 1; itr < mediaSelectedIndexes.length; ++itr) {
			if (lastMediaSelectedIndex + 1 != mediaSelectedIndexes[itr]) {
				sequential = false;
			}
			lastMediaSelectedIndex = mediaSelectedIndexes[itr];
		}
		return sequential;
	}
	HasRuntime() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		for (const media of this.mediaSelected) {
			switch (media.type) {
				case MEDIATYPE_VIDEO:
					return false;
			}
		}
		return true;
	}
	HasTransitions() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		for (const media of this.mediaSelected) {
			if (media.type == MEDIATYPE_AUDIO) {
				return false;
			}
		}
		return true;
	}
	HasLayout() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		for (const media of this.mediaSelected) {
			if (media.type == MEDIATYPE_AUDIO) {
				return false;
			}
		}
		return true;
	}
	HasVolume() {
		return this.mediaSelected.length != 0 && this.mediaSelected.every(media => {
			switch (media.type) {
				case MEDIATYPE_WIDGET:
				case MEDIATYPE_MISC:
					return false;
			}
			return true;
		});
	}
	HasDucking() {
		return this.mediaSelected.length != 0 && this.mediaSelected.every(media => {
			switch (media.type) {
				case MEDIATYPE_WIDGET:
				case MEDIATYPE_MISC:
					return false;
			}
			return true;
		});
	}
	GetMediaValues(allMedia, property) {
		if (allMedia.length == 0) {
			return null;
		}
		var value = null;
		if (property.startsWith('button-')) {
			switch (property) {
				case 'button-x':
					return gCurrentWidgetComponent.x;
				case 'button-y':
					return gCurrentWidgetComponent.y;
				case 'button-width':
					return gCurrentWidgetComponent.width;
				case 'button-height':
					return gCurrentWidgetComponent.height;
				case 'button-name':
					return gCurrentWidgetComponent.name;
				case 'button-metadata':
					if (!gCurrentWidgetComponent.hasOwnProperty('metadata')) {
						gCurrentWidgetComponent.metadata = [];
					}
					const currentMetadata = JSON.parse(JSON.stringify(gCurrentWidgetComponent.metadata));
					// currentMetadata.push({ name: 'name', type: 'string', value: gCurrentWidgetComponent.name });
					return currentMetadata;
			}
		} else {
			var value = null;
			for (const media of allMedia) {
				var mediaPropertyValue = media[property];
				if (property == 'scheduledaterange') {
					mediaPropertyValue = media.startdate != null && media.enddate != null;
				} else if (property == 'schedulestartdate') {
					// media.startdate is in the format 'YYYY-MM-DD HH:mm:ss'
					if (media.startdate) {
						mediaPropertyValue = media.utc ? moment.utc(media.startdate).tz(configuration.timezone).format('YYYY-MM-DD') : moment(media.startdate).format('YYYY-MM-DD');
					} else {
						mediaPropertyValue = '';
					}
				} else if (property == 'schedulestarttime') {
					if (media.startdate) {
						mediaPropertyValue = media.utc ? moment.utc(media.startdate).tz(configuration.timezone).format('HH:mm:ss') : moment(media.startdate).format('HH:mm:ss');
					} else {
						mediaPropertyValue = '';
					}
				} else if (property == 'scheduleenddate') {
					if (media.enddate) {
						mediaPropertyValue = media.utc ? moment.utc(media.enddate).tz(configuration.timezone).format('YYYY-MM-DD') : moment(media.enddate).format('YYYY-MM-DD');
					} else {
						mediaPropertyValue = '';
					}
				} else if (property == 'scheduleendtime') {
					if (media.enddate) {
						mediaPropertyValue = media.utc ? moment.utc(media.enddate).tz(configuration.timezone).format('HH:mm:ss') : moment(media.enddate).format('HH:mm:ss');
					} else {
						mediaPropertyValue = '';
					}
				} else if (property == 'scheduledayparting') {
					mediaPropertyValue = media.sunday || media.monday || media.tuesday || media.wednesday || media.thursday || media.friday || media.saturday;
				} else if (property == 'daypartstarttime') {
					if (media.utc) {
						mediaPropertyValue = moment(media.daypartstarttime, 'HH:mm:ss').add(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
					} else {
						mediaPropertyValue = media.daypartstarttime;
					}
				} else if (property == 'daypartendtime') {
					if (media.utc) {
						mediaPropertyValue = moment(media.daypartendtime, 'HH:mm:ss').add(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
					} else {
						mediaPropertyValue = media.daypartendtime;
					}
				}
				
				if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'blank' && ['x', 'y', 'width', 'height', 'entrytransitiontype', 'entrytransitionduration', 'exittransitiontype', 'exittransitionduration'].indexOf(property) != -1) {
					// Blank has no transiton
					continue;
				}
				if (value == null || mediaPropertyValue == value) {
					value = mediaPropertyValue;
				} else {
					value = null;
					break;
				}
			}
		}
		return value;
	}
	GetSelectedMediaXReadOnly() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		var value = false;
		for (const media of this.mediaSelected) {
			if (((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) && media.width == this.width) {
				value = true;
			}
		}
		return value;
	}
	GetSelectedMediaYReadOnly() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		var value = false;
		for (const media of this.mediaSelected) {
			if (((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) && media.height == this.height) {
				value = true;
			}
		}
		return value;
	}
	GetSelectedMediaWidthOrHeightReadOnly() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		var value = false;
		for (const media of this.mediaSelected) {
			if ((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) {
				value = true;
			}
		}
		return value;
	}
	GetSelectedMediaRunlengthReadOnly() {
		if (this.mediaSelected.length == 0) {
			return false;
		}
		for (const media of this.mediaSelected) {
			if (media.type == MEDIATYPE_AUDIO || media.type == MEDIATYPE_VIDEO) {
				return true;
			}
		}
		return false;
	}
	// This handles both widget attributes and media properties
	SetMediaValues(allMedia, property, value) {
		if (['button-x', 'button-y', 'button-width', 'button-height', 'button-metadata'].indexOf(property) != -1) {
			switch (property) {
				case 'button-x':
					gCurrentWidgetComponent.x = value;
					return;
				case 'button-y':
					gCurrentWidgetComponent.y = value;
					return;
				case 'button-width':
					gCurrentWidgetComponent.width = value;
					return;
				case 'button-height':
					gCurrentWidgetComponent.height = value;
					return;
				case 'button-metadata':
					gCurrentWidgetComponent.metadata = value;
					return;
			}
		} else {
			for (const media of allMedia) {
				if (property == 'runlength') {
					value = Math.max(1, Math.min(24 * 60 * 60, value));
				} else if (property == 'scheduled') {
					if (!value) {
						media.utc = false;
						media.startdate = null;
						media.enddate = null;
						media.sunday = false;
						media.monday = false;
						media.tuesday = false;
						media.wednesday = false;
						media.thursday = false;
						media.friday = false;
						media.saturday = false;
						media.daypartstarttime = '00:00:00';
						media.daypartendtime = '23:59:59';
					} else if (media.startdate == null && media.enddate == null && !media.sunday && !media.monday && !media.tuesday && !media.wednesday && !media.thursday && !media.friday && !media.saturday) {
						media.utc = false;
						media.startdate = moment.tz(configuration.timezone).startOf('day').format('YYYY-MM-DD HH:mm:ss');
						media.enddate = moment.tz(configuration.timezone).endOf('day').format('YYYY-MM-DD HH:mm:ss');
					}
				} else if (property == 'utc') {
					const $daypartStartTime = document.getElementById('media-daypartstarttime');
					const $daypartEndTime = document.getElementById('media-daypartendtime');
					const $startDate = document.getElementById('media-schedulestartdate');
					const $startTime = document.getElementById('media-schedulestarttime');
					const $endDate = document.getElementById('media-scheduleenddate');
					const $endTime = document.getElementById('media-scheduleendtime');
					if (value) {
						media.daypartstarttime = moment($daypartStartTime.value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
						media.daypartendtime = moment($daypartEndTime.value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
						if (media.startdate != null && media.enddate != null) {
							media.startdate = moment.tz(`${$startDate.value} ${$startTime.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
							media.enddate = moment.tz(`${$endDate.value} ${$endTime.value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						}
					} else {
						media.daypartstarttime = $daypartStartTime.value;
						media.daypartendtime = $daypartEndTime.value;
						if (media.startdate != null && media.enddate != null) {
							media.startdate = moment(`${$startDate.value} ${$startTime.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
							media.enddate = moment(`${$endDate.value} ${$endTime.value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
					}
				} else if (property == 'scheduledaterange') {
					if (!value) {
						media.startdate = null;
						media.enddate = null;
						if (!media.sunday && !media.monday && !media.tuesday && !media.wednesday && !media.thursday && !media.friday && !media.saturday) {
							media.sunday = true;
							media.monday = true;
							media.tuesday = true;
							media.wednesday = true;
							media.thursday = true;
							media.friday = true;
							media.saturday = true;
						}
					} else {
						media.startdate = moment().startOf('day').utc().format('YYYY-MM-DD HH:mm:ss');
						media.enddate = moment().endOf('day').utc().format('YYYY-MM-DD HH:mm:ss');
					}
					continue;
				} else if (property == 'schedulestartdate') {
					if (media.startdate) {
						if (media.utc) {
							media.startdate = moment.tz(`${value} ${moment.utc(media.startdate).tz(configuration.timezone).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						} else {
							media.startdate = moment(`${value} ${moment(media.startdate).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
						media.startdate = applyStartEndDateClamping(property, media.startdate);
						if (moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').isSameOrAfter(moment(media.enddate, 'YYYY-MM-DD HH:mm:ss'))) {
							media.enddate = moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').add(1, 'minute').format('YYYY-MM-DD HH:mm:ss');
						}
					}
					continue;
				} else if (property == 'schedulestarttime') {
					if (media.startdate) {
						if (media.utc) {
							media.startdate = moment.tz(`${moment.utc(media.startdate).tz(configuration.timezone).format('YYYY-MM-DD')} ${value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						} else {
							media.startdate = moment(`${moment(media.startdate).format('YYYY-MM-DD')} ${value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
						media.startdate = applyStartEndDateClamping(property, media.startdate);
						if (moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').isSameOrAfter(moment(media.enddate, 'YYYY-MM-DD HH:mm:ss'))) {
							media.enddate = moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').add(1, 'minute').format('YYYY-MM-DD HH:mm:ss');
						}
					}
					continue;
				} else if (property == 'scheduleenddate') {
					if (media.enddate) {
						if (media.utc) {
							media.enddate = moment.tz(`${value} ${moment.utc(media.enddate).tz(configuration.timezone).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						} else {
							media.enddate = moment(`${value} ${moment(media.enddate).format('HH:mm:ss')}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
						media.enddate = applyStartEndDateClamping(property, media.enddate);
						if (moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').isSameOrAfter(moment(media.enddate, 'YYYY-MM-DD HH:mm:ss'))) {
							media.startdate = moment(media.enddate, 'YYYY-MM-DD HH:mm:ss').subtract(1, 'minute').format('YYYY-MM-DD HH:mm:ss');
						}
					}
					continue;
				} else if (property == 'scheduleendtime') {
					if (media.enddate) {
						if (media.utc) {
							media.enddate = moment.tz(`${moment.utc(media.enddate).tz(configuration.timezone).format('YYYY-MM-DD')} ${value}`, 'YYYY-MM-DD HH:mm:ss', configuration.timezone).utc().format('YYYY-MM-DD HH:mm:ss');
						} else {
							media.enddate = moment(`${moment(media.enddate).format('YYYY-MM-DD')} ${value}`, 'YYYY-MM-DD HH:mm:ss').format('YYYY-MM-DD HH:mm:ss');
						}
						media.enddate = applyStartEndDateClamping(property, media.enddate);
						if (moment(media.startdate, 'YYYY-MM-DD HH:mm:ss').isSameOrAfter(moment(media.enddate, 'YYYY-MM-DD HH:mm:ss'))) {
							media.startdate = moment(media.enddate, 'YYYY-MM-DD HH:mm:ss').subtract(1, 'minute').format('YYYY-MM-DD HH:mm:ss');
						}
					}
					continue;
				} else if (property == 'scheduledayparting') {
					media.sunday = value;
					media.monday = value;
					media.tuesday = value;
					media.wednesday = value;
					media.thursday = value;
					media.friday = value;
					media.saturday = value;
					if (!value && media.startdate == null && media.enddate == null) {
						media.startdate = moment().startOf('day').utc().format('YYYY-MM-DD HH:mm:ss');
						media.enddate = moment().endOf('day').utc().format('YYYY-MM-DD HH:mm:ss');
					}
					continue;
				} else if (property == 'daypartstarttime') {
					if (media.utc) {
						media.daypartstarttime = moment(value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
					} else {
						media.daypartstarttime = value;
					}
					continue;
				} else if (property == 'daypartendtime') {
					if (media.utc) {
						media.daypartendtime = moment(value, 'HH:mm:ss').subtract(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss');
					} else {
						media.daypartendtime = value;
					}
					continue;
				} else if (['sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday'].indexOf(property) != -1) {
					media[property] = value;
					if (!media.sunday && !media.monday && !media.tuesday && !media.wednesday && !media.thursday && !media.friday && !media.saturday) {
						media.scheduledaterange = false;
					}
					continue;
				} else if (property == 'fillstyleid' && media.fillstyleid != value) {
					media.fillstyleid = value;
					media.ResizeMedia();
					if (value == FILLSTYLE_NONE) {
						media.Expand();
					}
				} else if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'blank' && (
					property == 'x' ||
					property == 'y' ||
					property == 'width' ||
					property == 'height' ||
					property == 'entrytransitiontype' ||
					property == 'entrytransitionduration' ||
					property == 'exittransitiontype' ||
					property == 'exittransitionduration'
				)) {
					// Blank has no transition
					continue;
				}
				media[property] = value;
			}
		}
	}
	GetWidgetValues(allMedia, attribute) {
		let value = null;
		for (const media of allMedia) {
			if (value == null || media.attributes[attribute.name] == value) {
				value = media.attributes[attribute.name];
			} else {
				value = null;
				break;
			}
		}
		return value;
	}
	SetWidgetValues(allMedia, attribute, value) {
		for (const media of allMedia.filter(media => media.widgettype != 'blank')) {
			if (attribute.name == 'fillstyle' && media.attributes.fillstyle != value) {
				media.attributes.fillstyle = value;
				media.ResizeMedia();
				if (value == FILLSTYLE_NONE) {
					media.Expand();
				}
			}
			media.attributes[attribute.name] = value;
		}
	}
	HasWidgetAttribute(allMedia, property) {
		// Every selected widget must have this property
		return this.mediaSelected.length != 0 && allMedia.every(media => media.HasAttribute(property));
	}
	IsMediaLayoutReadonly() {
		var media = this.GetSelectedMedia();
		if (media) {
			return (media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE;
		}
		return false;
	}
	IsMediaLayoutXReadonly() {
		var media = this.GetSelectedMedia();
		if (media) {
			if (((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) && media.width == this.width) {
				return true;
			}
		}
		return false;
	}
	IsMediaLayoutYReadonly() {
		var media = this.GetSelectedMedia();
		if (media) {
			if (((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) && media.height == this.height) {
				return true;
			}
		}
		return false;
	}
	IsMediaLayoutWidthAndHeightReadonly() {
		var media = this.GetSelectedMedia();
		if (media) {
			if ((media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_VIDEO) && media.fillstyleid != FILLSTYLE_NONE || media.type == MEDIATYPE_WIDGET && media.HasAttribute('fillstyle') && media.attributes.fillstyle != FILLSTYLE_NONE) {
				return true;
			}
		}
		return false;
	}
	IsSignwaveZone() {
		if (this.media.length == 1 && this.media[0].type == MEDIATYPE_WIDGET && this.media[0].widgettype == 'signwave') {
			toastError('The Signwave widget cannot have other content in its zone since it does not transition.');
			return true;
		}
		return false;
	}
	UpdateView() {
		this.$canvasView.innerHTML = '';
		this.$canvasView.hidden = !this.visible;
		// Remove references to the previous canvasView tags so we don't hold onto memory
		for (const media of this.media) {
			media.$canvasView = null;
		}
		if (this.GetSelectedMedia()) {
			const media = this.GetSelectedMedia();
			switch(media.type) {
				case MEDIATYPE_IMAGE:
				case MEDIATYPE_VIDEO:
					if (this.$canvasView.querySelector('.media') == null || this.$canvasView.querySelector('.media').tagName != 'CANVAS') {
						this.$canvasView.querySelector('.media')?.remove();
						media.$canvasView = this.$canvasView.appendChild($.canvas({ class: 'media', css: {
							position: 'relative', pointerEvents: 'none'
						} }));
					}
					break;
				case MEDIATYPE_AUDIO:
					// This just makes the code easier having a .media
					this.$canvasView.querySelector('.media')?.remove();
					media.$canvasView = this.$canvasView.appendChild($.div({ class: 'media', css: {
						position: 'relative',
						overflow: 'hidden',
						pointerEvents: 'none'
					} }));
					break;
				case MEDIATYPE_WIDGET:
				case MEDIATYPE_MISC:
					this.$canvasView.querySelector('.media')?.remove();
					media.$canvasView = this.$canvasView.appendChild($.div({ class: 'media', css: {
						position: 'relative',
						overflow: 'hidden',
						pointerEvents: 'none'
					} }));
					break;
			}
			const widgetMode = gCurrentWidget == media && (gCurrentWidget.widgettype == 'buttongroup' || gCurrentWidget.widgettype == 'text2');
			this.$canvasView.classList.toggle('widgetmode', widgetMode);
			// Audio doesn't have .media
			this.$canvasView.querySelector('.media')?.classList?.toggle('widgetmode', widgetMode);
		} else {
			if (this.$canvasView.querySelector('.media') == null || this.$canvasView.querySelector('.media').tagName != 'CANVAS') {
				this.$canvasView.querySelector('.media')?.remove();
				this.$canvasView.appendChild($.canvas({ class: 'media', css: {
					position: 'relative',
					pointerEvents: 'none'
				} }));
			}
		}
		const mediaContainer = this.$canvasView.querySelector('.media');
		const media = this.GetSelectedMedia();
		Object.assign(this.$canvasView.style, {
			left: `${this.x}px`,
			top: `${this.y}px`,
			width: `${this.width}px`,
			height: `${this.height}px`
		});
		if (media) {
			Object.assign(mediaContainer.style, {
				left: `${media.x}px`,
				top: `${media.y}px`,
				width: `${media.width}px`,
				height: `${media.height}px`
			});
			mediaContainer.width = media.width,
			mediaContainer.height = media.height
			switch (media.type) {
				case MEDIATYPE_IMAGE:
					progressiveLoadImage(mediaContainer, this, media);
					break;
				case MEDIATYPE_VIDEO:
					progressiveLoadVideoPoster(mediaContainer, this, media);
					break;
				case MEDIATYPE_MISC:
					mediaContainer.innerHTML = '';
					mediaContainer.appendChild(
						$.object({
							data: GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize),
							type: 'application/x-shockwave-flash',
							width: media.width,
							height: media.height
						},
							$.param({ name: 'quality', value: 'high' }),
							$.param({ name: 'bgcolor', value: '#000000' }),
							$.param({ name: 'play', value: 'true' }),
							$.param({ name: 'loop', value: 'true' }),
							$.param({ name: 'wmode', value: 'opaque' }),
							$.param({ name: 'scale', value: 'showall' }),
							$.param({ name: 'menu', value: 'true' }),
							$.param({ name: 'devicefont', value: 'false' }),
							$.param({ name: 'salign', value: '' }),
							$.param({ name: 'allowScriptAccess', value: 'sameDomain' })
						)
					);
					break;
				case MEDIATYPE_WIDGET:
					if (gCurrentWidget && gCurrentWidget.widgettype == 'text2') {
						// In widget mode the text2 widget renders using a custom renderer, not the widget's renderer
						var text = media.attributes.text;
						media.$canvasView.classList.add('widget-text2');
						media.$canvasView.tabIndex = -1;
						media.$canvasView.contentEditable = 'true';
						const characterStyles = media.text.JSONToCharacterStyles(text);
						media.text.CharacterStylesToDOM(media.$canvasView, characterStyles, { start: 1, end: 1, leftToRight: true });
						media.text.AddListeners();
						media.text.Update((selection, characterStyles) => {
							selection.start = 1;
							selection.end = 1;
							selection.leftToRight = true;
						}, false);
					} else {
						if (media.widget) {
							media.widget.Destroy();
						}
						media.widget = CreateWidget(mediaContainer, media, undefined, undefined, true, {}, null, account, null);
					}
					break;
			}
		} else {
			Object.assign(mediaContainer.style, {
				left: 0,
				top: 0,
				width: `${this.width}px`,
				height: `${this.height}px`
			});
			mediaContainer.width = this.width;
			mediaContainer.height = this.height;
			DrawEmptyZone(mediaContainer.getContext('2d'), 0, 0, mediaContainer.width, mediaContainer.height);
		}
	}
	UpdateMediaStartTimes() {
		var startTime = 0;
		for (const media of this.media) {
			media.starttime = startTime;
			startTime += media.runlength;
		}
		if (this.GetSelectedMedia()) {
			PopulateMediaStartRunlengthProperties(this);
		}
		for (const media of this.media) {
			media.UpdateNameStartTimeAndRunLength();
		}
	}
	ReloadView() {
		this.UpdateView();
		gCurrentPlaylist.zones.forEach((zone, zoneItr) => {
			if (zone.$canvasView) {
				zone.$canvasView.style.zIndex = zoneItr;
			}
		});
	}
	ConvertSelectedButtonZoneTrimToReferences() {
		// Convert all selected zone trimleft and trimright to media references
		for (const [button] of GetAllButtons()) {
			var buttonZone = button.zones[gCurrentPlaylist.zones.indexOf(this)];
			for (const mediaSelected of this.media) {
				var mediaIndex = this.media.indexOf(mediaSelected);
				if (buttonZone.trimleft == mediaIndex) {
					buttonZone.trimleft = mediaSelected;
				}
				if (buttonZone.trimright == mediaIndex) {
					buttonZone.trimright = mediaSelected;
				}
			}
		}
	}
	ConvertSelectedButtonZoneTrimReferencesToIndexes() {
		// Convert all select zone trimleft and trimright media references back to indexes
		var requiresUpdateButtonList = false;
		for (const [button] of GetAllButtons()) {
			var buttonZone = button.zones[gCurrentPlaylist.zones.indexOf(this)];
			for (const mediaSelected of this.media) {
				var mediaIndex = this.media.indexOf(mediaSelected);
				if (buttonZone.trimleft === mediaSelected) {
					buttonZone.trimleft = mediaIndex;
					requiresUpdateButtonList = true;
				}
				if (buttonZone.trimright === mediaSelected) {
					buttonZone.trimright = mediaIndex;
					requiresUpdateButtonList = true;
				}
			}
			if (buttonZone.trimleft > buttonZone.trimright) {
				var temp = buttonZone.trimleft;
				buttonZone.trimleft = buttonZone.trimright;
				buttonZone.trimright = temp;
			}
		}
		if (requiresUpdateButtonList) {
			UpdateButtonList();
		}
	}
}

class Media {
	constructor(zone, { id, name, x, y, width, height, starttime, runlength, scheduled, utc, startdate, enddate, daypartstarttime, daypartendtime, sunday, monday, tuesday, wednesday, thursday, friday, saturday }) {
		this.zone = zone;
		this.id = id == undefined ? -1 : id;
		this.name = name;
		this.x = x == undefined ? 0 : x;
		this.y = y == undefined ? 0 : y;
		this.width = width == undefined ? zone.width : width;
		this.height = height == undefined ? zone.height : height;
		this.starttime = starttime == undefined ? 0 : starttime;
		this.runlength = runlength;
		this.scheduled = scheduled == undefined ? false : scheduled;
		this.utc = utc == undefined ? true: utc;
		this.startdate = startdate == undefined ? null : startdate;
		this.enddate = enddate == undefined ? null : enddate;
		this.daypartstarttime = daypartstarttime == undefined ? '00:00:00' : moment(daypartstarttime, /^[0-9]{2}:[0-9]{2}(am|pm)$/.test(daypartstarttime) ? 'hh:mma' : 'HH:mm:ss').format('HH:mm:ss');
		this.daypartendtime = daypartendtime == undefined ? '23:59:59' : moment(daypartendtime, /^[0-9]{2}:[0-9]{2}(am|pm)$/.test(daypartendtime) ? 'hh:mma' : 'HH:mm:ss').format('HH:mm:ss');
		this.sunday = sunday == undefined ? false : sunday;
		this.monday = monday == undefined ? false : monday;
		this.tuesday = tuesday == undefined ? false : tuesday;
		this.wednesday = wednesday == undefined ? false : wednesday;
		this.thursday = thursday == undefined ? false : thursday;
		this.friday = friday == undefined ? false : friday;
		this.saturday = saturday == undefined ? false : saturday;
		
		this.selected = false;
		
		let mediaItemPointerDownSelected = false;
		let mediaItemPointerDownShiftSelected = false;
		let mediaItemPointerDownX = 0;
		let mediaItemPointerDownY = 0;
		let mediaItemPointerDownTime = 0;
		this.$propertyView = $.li({ class: 'mediaSelect-item', tabindex: 0, pointerdown: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			stopPreviewPressed();
			const propertyViewParentOffset = $.offset(this.$propertyView.parentNode);
			e.scrollTagLeft = propertyViewParentOffset.left;
			e.scrollTagTop = propertyViewParentOffset.top;
			mediaItemPointerDownX = e.pageX;
			mediaItemPointerDownY = e.pageY;
			mediaItemPointerDownTime = Date.now();
			mediaItemPointerDownSelected = false;
			mediaItemPointerDownShiftSelected = false;
			if (!this.selected) {
				// Pointerdown on a zone media item not selected then select the media item based on if ctrl or shift is pressed
				SelectMedia(this.zone, [this.zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'zone', true, true);
				mediaItemPointerDownSelected = true;
				mediaItemPointerDownShiftSelected = e.shiftKey;
			} else {
				// Move the item to the last of the selected items. If dragged then items will move relative to this
				SelectMedia(this.zone, [this.zone.media.indexOf(this)], false, false, true, 'zone', true, true);
			}
		}, pointerup: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			const currentTime = Date.now();
			if (Math.hypot(e.pageX - mediaItemPointerDownX, e.pageY - mediaItemPointerDownY) < gTimelineScale && Math.abs(currentTime - mediaItemPointerDownTime) < 500) {
				if (!mediaItemPointerDownSelected) {
					SelectMedia(this.zone, [this.zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'zone', true, true);
				}
			}
		}, dblclick: () => {
			if (this.type == MEDIATYPE_WIDGET) {
				EditWidget(zone, this);
			}
		}, draggable: {
			scrollParent: '#mediaSelect',
			drag: (e, startX, startY, startOffsetX, startOffsetY) => {
				if (!mediaItemPointerDownShiftSelected) {
					const propertyViewParentOffset = $.offset(this.$propertyView.parentNode);
					var changed = false;
					while (e.pageY - propertyViewParentOffset.top + this.$propertyView.parentNode.scrollTop > (this.zone.media.indexOf(this) + 1) * this.$propertyView.offsetHeight && moveCurrentMediaAfter(e)) {
						changed = true;
					}
					while (e.pageY - propertyViewParentOffset.top + this.$propertyView.parentNode.scrollTop < this.zone.media.indexOf(this) * this.$propertyView.offsetHeight && moveCurrentMediaBefore(e)) {
						changed = true;
					}
					if (changed) {
						updateCurrentZoneMediaOrder();
					}
				}
			}
		} });
		
		// Stores whether the pointerdown action resulted in a selection. If so stop a rapid click from selecting only one item and removing the current selection on mouse up
		let timelineItemPointerDownSelected = false;
		let timelineItemPointerDownShiftSelected = false;
		let timelineItemPointerDownX = 0;
		let timelineItemPointerDownY = 0;
		let timelineItemPointerDownTime = 0;
		this.$timelineView = $.span({ class: 'timelineItem moved-media', tabindex: 0, pointerdown: { listener: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			stopPreviewPressed();
			timelineItemPointerDownX = e.pageX;
			timelineItemPointerDownY = e.pageY;
			timelineItemPointerDownTime = Date.now();
			timelineItemPointerDownSelected = false;
			timelineItemPointerDownShiftSelected = false;
			if (!this.selected) {
				// Pointerdown on a timeline item not selected then select the media item based on if ctrl or shift is pressed
				SelectMedia(this.zone, [this.zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'timeline', true, true);
				timelineItemPointerDownSelected = true;
				timelineItemPointerDownShiftSelected = e.shiftKey;
			} else {
				// Move the item to the last of the selected items. If dragged then items will move relative to this
				SelectMedia(this.zone, [this.zone.media.indexOf(this)], false, false, true, 'timeline', true, true);
			}
		}, capture: true }, pointerup: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			var currentTime = Date.now();
			if (Math.hypot(e.pageX - timelineItemPointerDownX, e.pageY - timelineItemPointerDownY) < gTimelineScale && Math.abs(currentTime - timelineItemPointerDownTime) < 500) {
				if (!timelineItemPointerDownSelected) {
					SelectMedia(this.zone, [this.zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'timeline', true, true);
				}
			}
		}, keyup: e => {
			if (DialogWindowOpen()) {
				return;
			}
			var mediaIndex = this.zone.media.indexOf(this);
			stopPreviewPressed();
			const moveFeatureOnce = (dx, dy) => {
				if (gCurrentWidgetComponent) {
					gCurrentWidgetComponent.x += dx;
					gCurrentWidgetComponent.y += dy;
					PopulateButtonLayoutProperties(this.zone);
					this.zone.UpdateView();
				} else {
					if (e.ctrlKey) {
						this.x += dx;
						this.y += dy;
						PopulateMediaLayoutProperties(this.zone);
						var mediaContainer = this.zone.$canvasView.querySelector('.media');
						Object.assign(mediaContainer.style, {
							left: `${this.x}px`,
							top: `${this.y}px`
						});
					} else if (!zone.locked) {
						this.zone.x += dx;
						this.zone.y += dy;
						Object.assign(this.zone.$canvasView.style, {
							left: `${this.zone.x}px`,
							top: `${this.zone.y}px`,
						});
						PopulateZoneLayoutProperties(this.zone);
					} else {
						return;
					}
				}
				LineLayerClear();
				LineLayerHighlightZone(this.zone, true);
				SetSaveFlag(true);
			};
			var selectMediaOneZoneUpOrDown = offset => {
				if (offset == -1 && zoneIndex > 0 || offset == 1 && zoneIndex < gCurrentPlaylist.zones.length - 1) {
					var otherZone = gCurrentPlaylist.zones[zoneIndex + offset];
					var mediaIndex = null;
					for (var mediaItr2 = 0; mediaItr2 < otherZone.media.length; ++mediaItr2) {
						if (this.starttime >= otherZone.media[mediaItr2].starttime) {
							mediaIndex = mediaItr2;
						}
					}
					if (mediaIndex != null) {
						SelectMedia(otherZone, [mediaIndex], false, false, false, 'timeline', true, true);
					}
				}
			};
			switch (e.key) {
				case 'ArrowLeft':
					switch (gSelectMediaLocation) {
						case 'timeline':
							if (mediaIndex > 0) {
								SelectMedia(this.zone, [mediaIndex - 1], e.ctrlKey, e.shiftKey, false, 'timeline', true, true);
							}
							break;
						case 'canvas':
							moveFeatureOnce(-1, 0);
							break;
						case 'zone':
						case 'media':
							break;
					}
					break;
				case 'ArrowRight':
					switch (gSelectMediaLocation) {
						case 'timeline':
							if (mediaIndex < this.zone.media.length - 1) {
								SelectMedia(this.zone, [mediaIndex + 1], e.ctrlKey, e.shiftKey, false, 'timeline', true, true);
							}
							break;
						case 'canvas':
							moveFeatureOnce(1, 0);
							break;
						case 'zone':
						case 'media':
							break;
					}
					break;
				case 'ArrowUp':
					var zoneIndex = gCurrentPlaylist.zones.indexOf(this.zone);
					switch (gSelectMediaLocation) {
						case 'timeline':
							selectMediaOneZoneUpOrDown(localStorage.swapzoneorder == 'false' ? -1 : 1);
							break;
						case 'canvas':
							moveFeatureOnce(0, -1);
							break;
						case 'zone':
							break;
					}
					break;
				case 'ArrowDown':
					var zoneIndex = gCurrentPlaylist.zones.indexOf(this.zone);
					switch (gSelectMediaLocation) {
						case 'timeline':
							selectMediaOneZoneUpOrDown(localStorage.swapzoneorder == 'false' ? 1 : -1);
							break;
						case 'canvas':
							moveFeatureOnce(0, 1);
							break;
						case 'zone':
							break;
					}
					break;
				case 'Delete':
					deleteCurrentMedia(true);
					e.preventDefault();
					break;
			}
		}, droppable: {
			accept: ['content-group', 'content-leaf', 'widget', 'moved-media'],
			drop: ContentDrop,
			region: 'content'
		},
		draggable: {
			scrollParent: '#timeline-view',
			start: (e, startX, startY, startOffsetX, startOffsetY) => {
				if (DialogWindowOpen()) {
					// Do not blink the dialog here
					return false;
				}
				if (e.target.classList.contains('resize')) {
					return false;
				}
				// Audio cannot be moved outside of the audio zone
				if (gSelectedZone == null || gSelectedZone.name == 'Audio') {
					return false;
				}
				return gSelectedZone.GetAllSelectedMedia();
			},
			threshold: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, source) => {
				const $timelineBar = this.$timelineView.parentNode;
				const timelineViewParentOffset = $.offset($timelineBar);
				const tearOutMargin = 2;
				// Begin a ContentDrop action to move the dragged media to another zone
				const tearOut = e.pageY < timelineViewParentOffset.top - tearOutMargin || e.pageY > timelineViewParentOffset.top + $timelineBar.offsetHeight + tearOutMargin;
				if (tearOut) {
					for (const media of source) {
						media.$propertyView.classList.add('moving');
						media.$timelineView.classList.add('moving');
						media.$zoneView.classList.add('moving');
					}
				}
				return tearOut;
			},
			drag: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, sort) => {
				if (!timelineItemPointerDownShiftSelected && sort) {
					const $timelineBar = this.$timelineView.parentNode;
					const timelineViewParentOffset = $.offset($timelineBar);
					
					var changed = false;
					while (deltaX > 0 && e.pageX - timelineViewParentOffset.left + $timelineBar.scrollLeft > (this.starttime + this.runlength) * gTimelineScale && moveCurrentMediaAfter(e)) {
						changed = true;
					}
					while (deltaX < 0 && e.pageX - timelineViewParentOffset.left + $timelineBar.scrollLeft < this.starttime * gTimelineScale && moveCurrentMediaBefore(e)) {
						changed = true;
					}
					if (changed) {
						updateCurrentZoneMediaOrder();
					}
				}
			},
			helper: ($node, e, dropRegion) => {
				return $.div({ class: 'drag-timelineitem' },
					...gSelectedZone.GetAllSelectedMedia().map(media => $.span({ text: media.name }))
				);
			},
			stop: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, source) => {
				if (source) {
					for (const media of source) {
						media.$propertyView.classList.remove('moving');
						media.$timelineView.classList.remove('moving');
						media.$zoneView.classList.remove('moving');
					}
				}
			}
		} },
			$.i({ class: 'image' }),
			this.$timelineViewName = $.span({ class: 'name' }),
			this.$timelineViewResize = $.span({ class: 'resize', css: { width: `${gShrink ? 2 : 3}px` }, draggable: {
				scrollParent: '#timeline-view',
				start: (e, startX, startY, startOffsetX, startOffsetY) => {
					if (DialogWindowOpen()) {
						BlinkDialog();
						return false;
					}
					// Video and Audio can't have their runlength changed
					if (this.type == MEDIATYPE_VIDEO || this.type == MEDIATYPE_AUDIO) {
						return false;
					}
					stopPreviewPressed();
					var mediaIndex = zone.media.indexOf(this);
					SelectMedia(zone, [mediaIndex], e.ctrlKey, e.shiftKey, false, 'timeline', true, true);
				},
				drag: (e, startX, startY, startOffsetX, startOffsetY) => {
					var changed = false;
					var diffOver = e.pageX - startOffsetX - ($.offset(this.$timelineViewResize).left + this.$timelineViewResize.offsetWidth);
					if (diffOver > 0) {
						this.runlength += Math.ceil(diffOver / gTimelineScale);
						changed = true;
					}
					// The grabber is 2 or 3 pixels and the gTimelineScale is 3 or 5. Without the correction below you'd see oscilations
					var diffUnder = $.offset(this.$timelineViewResize).left - (gTimelineScale - this.$timelineViewResize.offsetWidth) - (e.pageX - startOffsetX);
					if (diffUnder > 0 && this.runlength > 1) {
						this.runlength = Math.max(1, this.runlength - Math.ceil(diffUnder / gTimelineScale));
						changed = true;
					}
					if (changed) {
						zone.UpdateMediaStartTimes();
						ResizeTimelineBars();
						SetSaveFlag(true);
					}
				}
			} }),
			$.div({ class: 'dropbefore', droppable: {
				accept: ['content-group', 'content-leaf', 'widget', 'widgetbutton', 'moved-media'],
				drop: ContentDrop,
				region: 'content'
			} }),
			$.div({ class: 'dropafter', droppable: {
				accept: ['content-group', 'content-leaf', 'widget', 'widgetbutton', 'moved-media'],
				drop: ContentDrop,
				region: 'content'
			} })
		);
		
		let zoneItemPointerDownSelected = false;
		let zoneItemPointerDownShiftSelected = false;
		var zoneItemPointerDownX = 0;
		var zoneItemPointerDownY = 0;
		var zoneItemPointerDownTime = 0;
		this.$zoneView = $.li({ class: 'zone-view-item-media-item moved-media', tabindex: 0, pointerdown: { listener: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			stopPreviewPressed();
			zoneItemPointerDownX = e.pageX;
			zoneItemPointerDownY = e.pageY;
			zoneItemPointerDownTime = Date.now();
			zoneItemPointerDownSelected = false;
			zoneItemPointerDownShiftSelected = false;
			if (!this.selected) {
				// Pointerdown on a zone media item not selected then select the media item based on if ctrl or shift is pressed
				SelectMedia(zone, [zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'zone', true, true);
				zoneItemPointerDownSelected = true;
				zoneItemPointerDownShiftSelected = e.shiftKey;
			} else {
				// Move the item to the last of the selected items. If dragged then items will move relative to this
				SelectMedia(zone, [zone.media.indexOf(this)], false, false, true, 'zone', true, true);
			}
		}, capture: true }, pointerup: e => {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			var currentTime = Date.now();
			if (Math.hypot(e.pageX - zoneItemPointerDownX, e.pageY - zoneItemPointerDownY) < gTimelineScale && Math.abs(currentTime - zoneItemPointerDownTime) < 500) {
				if (!zoneItemPointerDownSelected) {
					SelectMedia(zone, [zone.media.indexOf(this)], e.ctrlKey, e.shiftKey, false, 'zone', true, true);
				}
			}
		}, draggable: {
			scrollParent: '.zone-view-item-media',
			start: (e, startX, startY, startOffsetX, startOffsetY) => {
				if (DialogWindowOpen()) {
					// Do not blink the dialog here
					return false;
				}
				if (gSelectedZone == null || gSelectedZone.name == 'Audio') {
					return false;
				}
				return gSelectedZone.GetAllSelectedMedia();
			},
			threshold: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, source) => {
				const $zoneViewParent = this.$zoneView.parentNode;
				const zoneViewParentOffset = $.offset($zoneViewParent);
				const tearOutMargin = 2;
				// Begin a ContentDrop action to move the dragged media to another zone
				const tearOut = e.pageX < zoneViewParentOffset.left - tearOutMargin || e.pageX > zoneViewParentOffset.left + $zoneViewParent.offsetWidth + tearOutMargin;
				if (tearOut) {
					for (const media of source) {
						media.$propertyView.classList.add('moving');
						media.$timelineView.classList.add('moving');
						media.$zoneView.classList.add('moving');
					}
				}
				return tearOut;
			},
			drag: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, sort) => {
				if (!zoneItemPointerDownShiftSelected && sort) {
					const zoneViewParentOffset = $.offset(this.$zoneView.parentNode);
					var changed = false;
					while (e.pageY - zoneViewParentOffset.top + this.$zoneView.parentNode.scrollTop > (this.zone.media.indexOf(this) + 1) * this.$zoneView.offsetHeight && moveCurrentMediaAfter(e)) {
						changed = true;
					}
					while (e.pageY - zoneViewParentOffset.top + this.$zoneView.parentNode.scrollTop < this.zone.media.indexOf(this) * this.$zoneView.offsetHeight && moveCurrentMediaBefore(e)) {
						changed = true;
					}
					if (changed) {
						updateCurrentZoneMediaOrder();
					}
				}
			},
			helper: ($node, e, dropRegion) => {
				return $.div({ class: 'drag-timelineitem' },
					...gSelectedZone.GetAllSelectedMedia().map(media => $.span({ text: media.name }))
				);
			},
			stop: (e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY, source) => {
				if (source) {
					for (const media of source) {
						media.$propertyView.classList.remove('moving');
						media.$timelineView.classList.remove('moving');
						media.$zoneView.classList.remove('moving');
					}
				}
			}
		} },
			this.$zoneViewName = $.span({ class: 'name' }),
			this.$zoneViewRunLength = $.span({ text: this.runlength, class: 'runlength' }),
			$.div({ class: 'dropbefore', droppable:
			{
				accept: ['content-group', 'content-leaf', 'widget', 'widgetbutton', 'moved-media'],
				drop: ContentDrop,
				region: 'content'
			} }),
			$.div({ class: 'dropafter', droppable:
			{
				accept: ['content-group', 'content-leaf', 'widget', 'widgetbutton', 'moved-media'],
				drop: ContentDrop,
				region: 'content'
			} })
		);
	}
	async youtubeURLToTitle() {
		// Cache the video or playlist title. This is asynchronous and might take a second or two to update
		// https://youtu.be/dQw4w9WgXcQ
		// https://www.youtube.com/watch?v=dQw4w9WgXcQ
		let videoIdMatches = this.attributes.url.match(/^.*?(?:v=|youtu\.be\/)([A-Za-z0-9-_]*)$/);
		// https://www.youtube.com/watch?v=2nXYbGmF3_Q&list=PLyqf6gJt7KuHBmeVzZteZUlNUQAVLwrZS
		let playlistIdMatches = this.attributes.url.match(/^.*?(?:v=)?[A-Za-z0-9-_]*?&?list=([A-Za-z0-9-_]*)$/);
		const key = 'AIzaSyDxh_IZrKbvOKlGJF75fVOQfg0Tnna9-rw';
		if (!this.hasOwnProperty('lastVideoId')) {
			this.lastVideoId = '';
		}
		if (!this.hasOwnProperty('lastPlaylistId')) {
			this.lastPlaylistId = '';
		}
		if (!this.hasOwnProperty('youtubeTitle')) {
			this.youtubeTitle = '';
		}
		if (this.attributes.url == '') {
			return;
		}
		if (videoIdMatches && videoIdMatches.length == 2) {
			const videoId = videoIdMatches[1];
			try {
				if (videoId != '' && this.lastVideoId != videoId) {
					this.lastVideoId = videoId;
					const info = await (await fetch(`https://www.googleapis.com/youtube/v3/videos?part=snippet&fields=items(localizations%2Csnippet%2Ftitle)&id=${videoId}&key=${key}`)).json();
					this.youtubeTitle = info.hasOwnProperty('items') && info.items.length > 0 && info.items[0].hasOwnProperty('snippet') && info.items[0].snippet.hasOwnProperty('title') ? info.items[0].snippet.title : '';
					this.UpdateNameStartTimeAndRunLength();
				}
			} catch (exception) {
				this.youtubeTitle = '';
			}
		}
		if (playlistIdMatches && playlistIdMatches.length == 2) {
			const playlistId = playlistIdMatches[1];
			try {
				if (playlistId != '' && this.lastPlaylistId != playlistId) {
					this.lastPlaylistId = playlistId;
					const info = await (await fetch(`https://www.googleapis.com/youtube/v3/playlists?part=snippet&fields=items(localizations%2Csnippet%2Ftitle)&id=${playlistId}&key=${key}`)).json();
					this.youtubeTitle = info.hasOwnProperty('items') && info.items.length > 0 && info.items[0].hasOwnProperty('snippet') && info.items[0].snippet.hasOwnProperty('title') ? info.items[0].snippet.title : '';
					this.UpdateNameStartTimeAndRunLength();
				}
			} catch (exception) {
				this.youtubeTitle = '';
			}
		}
	}
	get display() {
		var displayName = '';
		if (this.type == MEDIATYPE_WIDGET) {
			switch (this.widgettype) {
				case 'weather':
					displayName = 'Weather (' + this.attributes.location + ')';
					break;
				case 'weathercrawl':
					displayName = 'Weather Crawl (' + this.attributes.location + ')';
					break;
				case 'webpage':
					if (this.attributes.url == '') {
						displayName = this.name;
					} else {
						var truncatedText = this.attributes.url.substring(0, 128);
						displayName = `${truncatedText + (this.attributes.url.length > 128 ? '...' : '')} (Webpage)`;
					}
					break;
				case 'text':
					if (this.attributes.text == '') {
						displayName = this.name;
					} else {
						var truncatedText = this.attributes.text.substring(0, 128);
						displayName = `${truncatedText + (this.attributes.text.length > 128 ? '...' : '')} (Text)`;
					}
					break;
				case 'text2':
					const characterStyles = this.text.JSONToCharacterStyles(this.attributes.text);
					if (characterStyles.length == 0) {
						displayName = this.name;
					} else {
						var text = characterStyles.filter(c => c.character != String.fromCharCode(8203)).map(c => c.character).join('');
						var truncatedText = text.substring(0, 128);
						displayName = `${truncatedText + (text.length > 128 ? '...' : '')} (Advanced Text)`;
					}
					break;
				case 'buttongroup':
					displayName = this.name;
					break;
				case 'pdf':
					if (this.attributes.pdf == null) {
						displayName = this.name;
					} else {
						displayName = this.attributes.pdf.name;
					}
					break;
				case 'rss':
					if (this.attributes.rssfeed == 'useurl' && this.attributes.url == '') {
						displayName = this.name;
					} else if (this.attributes.rssfeed == 'useurl') {
						displayName = 'RSS (' + this.attributes.url + ')';
					} else {
						displayName = 'RSS (' + this.attributes.rssfeed + ')';
					}
					break;
				case 'rsscrawl':
					if (this.attributes.rssfeed == 'useurl' && this.attributes.url == '') {
						displayName = this.name;
					} else if (this.attributes.rssfeed == 'useurl') {
						displayName = 'RSS Crawl (' + this.attributes.url + ')';
					} else {
						displayName = 'RSS Crawl (' + this.attributes.rssfeed + ')';
					}
					break;
				case 'csvroll':
					if (this.attributes.csvfile == 'useurl' && this.attributes.url == '') {
						displayName = this.name;
					} else if (this.attributes.csvfile == 'useurl') {
						displayName = 'CSV Roll (' + this.attributes.url + ')';
					} else {
						displayName = 'CSV Roll (' + this.attributes.csvfile + ')';
					}
					break;
				case 'events':
					if (this.attributes.eventlist == '') {
						displayName = this.name;
					} else {
						displayName = 'Events (' + this.attributes.eventlist + ')';
					}
					break;
				case 'twitter':
					displayName = 'Twitter (' + this.attributes.follow + ')';
					break;
				case 'youtube':
					if (this.attributes.url == '') {
						displayName = this.name;
					} else {
						this.youtubeURLToTitle(); // async
						if (this.hasOwnProperty('youtubeTitle') && this.youtubeTitle != '') {
							var truncatedText = this.youtubeTitle.substring(0, 128);
							displayName = `${truncatedText + (this.youtubeTitle.length > 128 ? '...' : '')} (Youtube)`;
						} else {
							displayName = this.name;
						}
					}
					break;
				case 'galleryfrommetadata':
					displayName = `${FilterToName(this.attributes.filters)} (Gallery From Metadata)`;
					break;
				default:
					displayName = this.name;
			}
		} else {
			displayName = this.name;
		}
		return displayName;
	}
	Remove() {
		this.$propertyView.remove();
		this.$timelineView.remove();
		this.$zoneView.remove();
	}
	UpdateNameStartTimeAndRunLength() {
		const displayName = this.display;
		this.$timelineView.style.width = `${this.runlength * gTimelineScale}px`;
		this.$timelineView.hidden = !(!gShrink || this.runlength != 1);
		const cleanedDisplayName = displayName.replace(/\r?\n|\r/g, ' ');
		const newTitle = `Name: ${cleanedDisplayName} - Start Time: ${this.starttime} - Run Length: ${this.runlength}`;
		if (this.$timelineView.title != newTitle) this.$timelineView.title = newTitle;
		if (this.$timelineViewName.innerText != cleanedDisplayName) this.$timelineViewName.innerText = cleanedDisplayName;
		if (this.$zoneViewName.innerText != cleanedDisplayName) this.$zoneViewName.innerText = cleanedDisplayName;
		if (this.$zoneViewRunLength.innerText != this.runlength) this.$zoneViewRunLength.innerText = this.runlength;
		if (this.$propertyView.innerText != cleanedDisplayName) this.$propertyView.innerText = cleanedDisplayName;
	}
	Expand() {
		this.x = 0;
		this.y = 0;
		this.width = this.zone.width;
		this.height = this.zone.height;
		if (this.type == MEDIATYPE_WIDGET && this.HasAttribute('fillstyle')) {
			this.attributes.fillstyle = FILLSTYLE_NONE;
		} else if (this.type == MEDIATYPE_IMAGE || this.type == MEDIATYPE_VIDEO) {
			this.fillstyleid = FILLSTYLE_NONE;
		}
		UpdateProperties();
		this.zone.UpdateView();
	}
	// Determines if this media will be played in the future
	GetScheduledState() {
		var expired = false; // Expired
		var scheduled = false; // Schedule in the future
		if (this.scheduled) {
			const currentTime = moment.tz(configuration.timezone);
			const daypartingEnabled = this.sunday || this.monday || this.tuesday || this.wednesday || this.thursday || this.friday || this.saturday;
			const dayOfTheWeek = [this.sunday, this.monday, this.tuesday, this.wednesday, this.thursday, this.friday, this.saturday];
			const daypartStartAndEnd = (startOfWeek, dayItr) => {
				const daypartstarttime = this.utc ? moment(this.daypartstarttime, 'HH:mm:ss').add(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss') : this.daypartstarttime;
				const daypartendtime = this.utc ? moment(this.daypartendtime, 'HH:mm:ss').add(moment.tz(configuration.timezone).utcOffset(), 'minutes').format('HH:mm:ss') : this.daypartendtime
				const daypartStartTime = moment(startOfWeek.format('YYYY-MM-DD') + ' ' + daypartstarttime, 'YYYY-MM-DD HH:mm:ss').add(dayItr, 'days');
				const daypartEndTime =  moment(startOfWeek.format('YYYY-MM-DD') + ' ' + daypartendtime, 'YYYY-MM-DD HH:mm:ss').add(dayItr, 'days');
				// Handle the case where dayparting spans between days
				if (daypartEndTime.isBefore(daypartStartTime)) {
					daypartEndTime.add(1, 'days');
				}
				return { start: daypartStartTime, end: daypartEndTime };
			};
		
			if (this.startdate == null && this.enddate == null) {
				scheduled = true;
			} else {
				const startDate = this.utc ? moment.utc(this.startdate).tz(configuration.timezone) : moment.tz(this.startdate, configuration.timezone);
				const endDate = this.utc ? moment.utc(this.enddate).tz(configuration.timezone) : moment.tz(this.enddate, configuration.timezone);
				if (currentTime.isBefore(endDate)) {
					if (daypartingEnabled) {
						let dayPartExistsBetweenNowAndEnd = false;
						const timeItr = endDate.clone().startOf('week');
						while (!dayPartExistsBetweenNowAndEnd && timeItr.isSameOrAfter(startDate.clone().startOf('week')) && timeItr.isSameOrAfter(currentTime.clone().startOf('week'))) {
							for (const [dayItr, dayEnabled] of dayOfTheWeek.entries()) {
								if (!dayPartExistsBetweenNowAndEnd && dayEnabled) {
									const daypartTime = daypartStartAndEnd(timeItr, dayItr);
									// If UTC is enabled then we can also ignore daypart ranges that are expired which is what the last condition does
									if (daypartTime.start.isBefore(endDate) && daypartTime.end.isAfter(startDate) && (!this.utc || daypartTime.end.isSameOrAfter(currentTime))) {
										dayPartExistsBetweenNowAndEnd = true;
									}
								}
							}
							timeItr.subtract(1, 'week');
						}
						if (dayPartExistsBetweenNowAndEnd) {
							scheduled = true;
						} else {
							expired = true;
						}
					} else {
						scheduled = true;
					}
				} else {
					expired = true;
				}
			}
		}
		return {
			expired: expired,
			scheduled: scheduled
		};
	}
	ResizeMedia(oldWidth = this.zone.width, oldHeight = this.zone.height) {
		let actualWidth = this.actualwidth;
		let actualHeight = this.actualheight;
		if (this.type == MEDIATYPE_WIDGET && this.HasAttribute('fillstyle')) {
			actualWidth = this.attributes.cropwidth;
			actualHeight = this.attributes.cropheight;
		}
		if ((this.type == MEDIATYPE_IMAGE || this.type == MEDIATYPE_VIDEO) && this.fillstyleid == FILLSTYLE_NONE || this.type == MEDIATYPE_WIDGET && (!this.HasAttribute('fillstyle') || this.attributes.fillstyle == FILLSTYLE_NONE)) {
			if (this.x == 0 && this.width == oldWidth) {
				this.width = this.zone.width;
			}
			if (this.y == 0 && this.height == oldHeight) {
				this.height = this.zone.height;
			}
		} else if ((this.type == MEDIATYPE_IMAGE || this.type == MEDIATYPE_VIDEO) && this.fillstyleid == FILLSTYLE_COVER || this.type == MEDIATYPE_WIDGET && this.HasAttribute('fillstyle') && this.attributes.fillstyle == FILLSTYLE_COVER) {
			var mediaRatio = actualWidth / actualHeight;
			var zoneRatio = this.zone.width / this.zone.height;
			if (mediaRatio > zoneRatio) {
				this.width = Math.round(this.zone.height * (actualWidth / actualHeight));
				this.height = this.zone.height;
				this.x = Math.round((this.zone.width - this.width) / 2);
				this.y = 0;
			} else if (mediaRatio < zoneRatio) {
				this.width = this.zone.width;
				this.height = Math.round(this.zone.width * (actualHeight / actualWidth));
				this.x = 0;
				this.y = Math.round((this.zone.height - this.height) / 2);
			} else {
				this.width = this.zone.width;
				this.height = this.zone.height;
				this.x = 0;
				this.y = 0;
			}
		} else if ((this.type == MEDIATYPE_IMAGE || this.type == MEDIATYPE_VIDEO) && this.fillstyleid == FILLSTYLE_ASPECT_RATIO || this.type == MEDIATYPE_WIDGET && this.HasAttribute('fillstyle') && this.attributes.fillstyle == FILLSTYLE_ASPECT_RATIO) {
			var mediaRatio = actualWidth / actualHeight;
			var zoneRatio = this.zone.width / this.zone.height;
			if (mediaRatio > zoneRatio) {
				this.width = this.zone.width;
				this.height = Math.round(this.zone.width * (actualHeight / actualWidth));
				this.x = 0;
				this.y = Math.round((this.zone.height - this.height) / 2);
			} else if (mediaRatio < zoneRatio) {
				this.width = Math.round(this.zone.height * (actualWidth / actualHeight));
				this.height = this.zone.height;
				this.x = Math.round((this.zone.width - this.width) / 2);
				this.y = 0;
			} else {
				this.width = this.zone.width;
				this.height = this.zone.height;
				this.x = 0;
				this.y = 0;
			}
		}
	}
	ReloadTimelineImage() {
		this.$timelineView.querySelector('.image').style.backgroundImage = `url('${GetCachedMediaURL(account, this.icon, this.uploadedtimestamp, this.filesize)}')`;
	}
}

class MediaImage extends Media {
	constructor(zone, options) {
		super(zone, options);
		this.type = MEDIATYPE_IMAGE;
		
		this.cdnurl = options.cdnurl;
		this.actualwidth = options.actualwidth;
		this.actualheight = options.actualheight;
		this.contentid = options.contentid;
		this.fillstyleid = options.fillstyleid == undefined ? FILLSTYLE_NONE : options.fillstyleid;
		this.entrytransitiontype = options.entrytransitiontype == undefined ? 'None' : options.entrytransitiontype;
		this.entrytransitionduration = options.entrytransitionduration == undefined ? 2000 : options.entrytransitionduration;
		this.exittransitiontype = options.exittransitiontype == undefined ? 'None' : options.exittransitiontype;
		this.exittransitionduration = options.exittransitionduration == undefined ? 2000 : options.exittransitionduration;
		this.zoompan = options.zoompan == undefined ? 0 : options.zoompan;
		this.uploadedtimestamp = options.uploadedtimestamp;
		this.filesize = options.filesize;
		
		this.ReloadTimelineImage();
	}
	ReloadTimelineImage() {
		this.icon = this.cdnurl.split('.');
		this.icon.splice(-2, 0, 'thumbnail');
		this.icon = this.icon.join('.');
		super.ReloadTimelineImage();
	}
}

class MediaVideo extends Media {
	constructor(zone, options) {
		super(zone, options);
		this.type = MEDIATYPE_VIDEO;
		
		this.cdnurl = options.cdnurl;
		this.actualwidth = options.actualwidth;
		this.actualheight = options.actualheight;
		this.autoplay = options.autoplay == undefined ? true : options.autoplay;
		this.loop = options.loop == undefined ? true : options.loop;
		this.controls = options.controls == undefined ? false : options.controls;
		this.volume = options.hasaudio == undefined || options.hasaudio ? (options.volume == undefined ? 100 : options.volume) : 0;
		this.ducking = options.ducking == undefined ? 100 : options.ducking;
		this.contentid = options.contentid;
		this.fillstyleid = options.fillstyleid == undefined ? FILLSTYLE_NONE : options.fillstyleid;
		this.entrytransitiontype = options.entrytransitiontype == undefined ? 'None' : options.entrytransitiontype;
		this.entrytransitionduration = options.entrytransitionduration == undefined ? 2000 : options.entrytransitionduration;
		this.exittransitiontype = options.exittransitiontype == undefined ? 'None' : options.exittransitiontype;
		this.exittransitionduration = options.exittransitionduration == undefined ? 2000 : options.exittransitionduration;
		this.uploadedtimestamp = options.uploadedtimestamp;
		this.filesize = options.filesize;
		
		this.ReloadTimelineImage();
	}
	ReloadTimelineImage() {
		this.posterURL =  this.cdnurl.split('.');
		this.posterURL.splice(-2, 0, 'poster');
		this.posterURL.splice(-1, 1, 'jpg');
		this.posterURL = this.posterURL.join('.');
		this.icon = this.cdnurl.split('.');
		this.icon.splice(-2, 0, 'thumbnail');
		this.icon.splice(-1, 1, 'jpg');
		this.icon = this.icon.join('.');
		super.ReloadTimelineImage();
	}
}

class MediaAudio extends Media {
	constructor(zone, options) {
		super(zone, options);
		this.type = MEDIATYPE_AUDIO;
		
		this.cdnurl = options.cdnurl;
		this.autoplay = options.autoplay == undefined ? true : options.autoplay;
		this.loop = options.loop == undefined ? true : options.loop;
		this.controls = options.controls == undefined ? false : options.controls;
		this.contentid = options.contentid;
		this.volume = options.volume == undefined ? 100 : options.volume;
		this.ducking = options.ducking == undefined ? 100 : options.ducking;
		this.uploadedtimestamp = options.uploadedtimestamp;
		this.filesize = options.filesize;
		this.entrytransitiontype = 'None';
		this.entrytransitionduration = 0;
		this.exittransitiontype = 'None';
		this.exittransitionduration = 0;
		this.icon = 'mdi-music-note';
		
		this.$timelineView.querySelector('.image').classList.add('mdi', ...this.icon.split(' '));
		this.$zoneView.prepend($.i({ class: `mdi ${this.icon}` }));
	}
}

class MediaMisc extends Media {
	constructor(zone, options) {
		super(zone, options);
		this.type = MEDIATYPE_MISC;
		
		this.cdnurl = options.cdnurl;
		this.contentid = options.contentid;
		this.uploadedtimestamp = options.uploadedtimestamp;
		this.filesize = options.filesize;
		this.entrytransitiontype = options.entrytransitiontype == undefined ? 'None' : options.entrytransitiontype;
		this.entrytransitionduration = options.entrytransitionduration == undefined ? 2000 : options.entrytransitionduration;
		this.exittransitiontype = options.exittransitiontype == undefined ? 'None' : options.exittransitiontype;
		this.exittransitionduration = options.exittransitionduration == undefined ? 2000 : options.exittransitionduration;
		this.icon = 'mdi-file-outline';
		
		this.$timelineView.querySelector('.image').classList.add('mdi', ...this.icon.split(' '));
		this.$zoneView.prepend($.i({ class: `mdi ${this.icon}` }));
	}
}

class MediaWidget extends Media {
	constructor(zone, options) {
		super(zone, options);
		
		this.type = MEDIATYPE_WIDGET;
		this.widgettype = options.widgettype;
		if (this.widgettype == 'buttongroup') {
			this.name = options.attributes.name;
		}
		this.contentid = null;
		this.entrytransitiontype = options.entrytransitiontype == undefined ? 'None' : options.entrytransitiontype;
		this.entrytransitionduration = options.entrytransitionduration == undefined ? 2000 : options.entrytransitionduration;
		this.exittransitiontype = options.exittransitiontype == undefined ? 'None' : options.exittransitiontype;
		this.exittransitionduration = options.exittransitionduration == undefined ? 2000 : options.exittransitionduration;
		this.attributes = options.attributes;
		this.icon = options.icon;
		
		this.$timelineView.addEventListener('dblclick', e => {
			e.stopPropagation();
			EditWidget(zone, this);
		});
		this.$timelineView.querySelector('.image').classList.add('mdi', ...this.icon.split(' '));
		
		this.$zoneView.addEventListener('dblclick', () => {
			if (this.type == MEDIATYPE_WIDGET) {
				EditWidget(zone, this);
			}
		});
		this.$zoneView.prepend($.i({ class: `mdi ${this.icon}` }));
		
		if (this.widgettype == 'text2') {
			this.text = new Text(this);
		}
	}
	HasAttribute(property) {
		return WidgetInformation.some(widget => widget.name == this.widgettype && widget.attributes.some(attribute => attribute.name == property));
	}
}

function ClickButton(button) {
	if (DialogWindowOpen() || InPreviewMode()) {
		BlinkDialog();
		return;
	}
	gCurrentWidgetComponent = button;
	button.zones.forEach((zone, zoneItr) => {
		if (zone.statechange) {
			var playlistZone = gCurrentPlaylist.zones[zoneItr];
			playlistZone.visible = zone.show;
			SelectMedia(playlistZone, playlistZone.media.length != 0 ? [zone.trimleft == -1 ? 0 : zone.trimleft] : [], false, false, false, 'canvas', false, false);
		}
	});
	button.zones.forEach((zone, zoneItr) => {
		var playlistZone = gCurrentPlaylist.zones[zoneItr];
		playlistZone.media.forEach((media, mediaItr) => {
			if (zone.statechange) {
				media.$timelineView.style.opacity = zone.show && mediaItr >= (zone.trimleft == -1 ? 0 : zone.trimleft) && mediaItr <= (zone.trimright == -1 ? playlistZone.media.length - 1 : zone.trimright) ? 1 : 0.2;
			}
		});
	});
	UpdateZoneVisibility();
}

async function SelectZone(zone, focus, calledFromSelectMedia) {
	if (!calledFromSelectMedia) {
		document.querySelector('input:focus')?.blur();
	}
	for (const $element of document.getElementsByClassName('timelineItem')) {
		$element.style.opacity = 1;
	}
	
	if (zone && focus) {
		if (zone.mediaSelected.length != 0) {
			setTimeout(() => {
				if (zone.GetSelectedMedia() && gCurrentWidget == null) {
					if (document.getElementById('editor-view-tabs-timeline').classList.contains('selected')) {
						zone.GetSelectedMedia().$timelineView.focus();
					} else {
						zone.GetSelectedMedia().$zoneView.focus();
					}
				}
			}, 0);
		} else {
			setTimeout(() => {
				zone.$timelineName.focus();
			}, 0);
		}
	}
	if (gSelectedZone !== zone) {
		gSelectedZone = zone;
		
		if (gSelectedZone == null || gSelectedZone && gCurrentWidget !== gSelectedZone.GetSelectedMedia()) {
			ExitWidgetMode();
		}
		UpdateZoneAndMediaDOM();
		if (!calledFromSelectMedia) {
			UpdateSelection();
		}
		
		LineLayerClear();
		if (gSelectedZone) {
			LineLayerHighlightZone(gSelectedZone, true);
			if (focus) {
				if (gSelectedZone.mediaSelected.length != 0) {
					setTimeout(() => {
						if (gSelectedZone && gSelectedZone.GetSelectedMedia()) {
							if (document.getElementById('editor-view-tabs-timeline').classList.contains('selected')) {
								gSelectedZone.GetSelectedMedia().$timelineView.focus();
							} else {
								gSelectedZone.GetSelectedMedia().$zoneView.focus();
							}
						}
					}, 0);
				} else {
					setTimeout(() => {
						if (gSelectedZone) {
							gSelectedZone.$timelineName.focus();
						}
					}, 0);
				}
			}
		}
	}
	
	if (!calledFromSelectMedia) {
		if (focus) {
			UpdateProperties('zone');
		}
	}
}

async function SelectMedia(zone, indexes, preserve, preserveShift, moveToFront, selectMediaLocation, focus, selectZone) {
	document.querySelector('input:focus')?.blur();
	if (selectMediaLocation != null) {
		gSelectMediaLocation = selectMediaLocation;
	}
	const zoneSelectionChanged = gSelectedZone !== zone;
	if (selectZone) {
		SelectZone(zone, focus, true);
	} else {
		for (const $element of document.getElementsByClassName('timelineItem')) {
			$element.style.opacity = 1;
		}
	}
	// Clone the array and make changes to it instead so we can track if it's changed
	let mediaSelected = zone.mediaSelected.map(media => media);
	
	if (moveToFront) {
		mediaSelected.splice(mediaSelected.indexOf(zone.media[indexes[0]]), 1);
		mediaSelected.push(zone.media[indexes[0]]);
	} else if (mediaSelected.length != 0 && preserveShift) {
		mediaSelected = [mediaSelected[mediaSelected.length - 1]];
		var index = zone.media.indexOf(mediaSelected[0]);
		var start = Math.min(index, indexes[0]);
		var end = Math.max(index, indexes[0]);
		for (var itr = start; itr <= end; ++itr) {
			if (itr != index) {
				mediaSelected.unshift(zone.media[itr]);
			}
		}
	} else if (preserve) {
		var exists = false;
		if (mediaSelected.indexOf(zone.media[indexes[0]]) != -1) {
			mediaSelected.splice(mediaSelected.indexOf(zone.media[indexes[0]]), 1);
			exists = true;
		}
		if (mediaSelected.length == 0 || !exists) {
			mediaSelected.push(zone.media[indexes[0]]);
		}
	} else {
		mediaSelected = indexes.map(index => zone.media[index]);
	}
	
	var mediaSelectionChanged = mediaSelected.length != zone.mediaSelected.length || mediaSelected.some((media, mediaSelectedItr) => media !== zone.mediaSelected[mediaSelectedItr]);
	
	if (mediaSelectionChanged) {
		zone.mediaSelected = mediaSelected;
		for (const media of zone.media) {
			media.selected = false;
		}
		for (const media of zone.mediaSelected) {
			media.selected = true;
		}
		LineLayerClear();
		if (gSelectedZone) {
			LineLayerHighlightZone(gSelectedZone, true);
			if (focus) {
				if (gSelectedZone.mediaSelected.length != 0) {
					setTimeout(() => {
						if (gSelectedZone && gSelectedZone.GetSelectedMedia()) {
							if (document.getElementById('editor-view-tabs-timeline').classList.contains('selected')) {
								gSelectedZone.GetSelectedMedia().$timelineView.focus();
							} else {
								gSelectedZone.GetSelectedMedia().$zoneView.focus();
							}
						}
					}, 0);
				} else {
					setTimeout(() => {
						if (gSelectedZone) {
							gSelectedZone.$timelineName.focus();
						}
					}, 0);
				}
			}
		}
		
		if (gSelectedZone && gCurrentWidget != gSelectedZone.GetSelectedMedia()) {
			ExitWidgetMode();
		}
	}
	
	if (mediaSelectionChanged || zoneSelectionChanged || gPropertiesView == 'zone') {
		UpdateProperties('media');
	}
	if (mediaSelectionChanged || zoneSelectionChanged) {
		UpdateSelection();
	}
	
	if (mediaSelectionChanged) {
		zone.ReloadView();
	}
}

function toCanvasScale(value) {
	const canvasDimensions = CalculateCanvasDimensions();
	return value * gCurrentPlaylist.pagewidth / canvasDimensions.width;
}

function toCanvasWidth(value) {
	const canvasDimensions = CalculateCanvasDimensions();
	return (value * canvasDimensions.width + 0.5) / gCurrentPlaylist.pagewidth;
}

function toCanvasHeight(value) {
	const canvasDimensions = CalculateCanvasDimensions();
	return (value * canvasDimensions.height + 0.5) / gCurrentPlaylist.pageheight;
}

function DrawEmptyZone(context, x, y, width, height) {
	context.save();
	var gradient = context.createLinearGradient(x, y, width, height);
	gradient.addColorStop(0, '#1e94cc');
	gradient.addColorStop(0.4, '#fbfbef');
	gradient.addColorStop(0.6, '#fbfbef');
	gradient.addColorStop(1, '#1e94cc');
	context.fillStyle = gradient;
	context.fillRect(x, y, width, height);
	context.restore();
}

async function ContentDrop(e, source) {
	if (DialogWindowOpen() || gCurrentPlaylist == null) {
		BlinkDialog();
		return;
	}
	const isMedia = Array.isArray(source) && !(source[0] instanceof Media);
	const isMovedMedia = Array.isArray(source) && source[0] instanceof Media;
	const isWidget = source?.classList?.contains('widget') ?? false;
	const isWidgetButton = source?.classList?.contains('widgetbutton') ?? false;
	
	var pointerInCanvas = e.target.closest('#editor-canvas') != null;
	var pointerInTimelineView = e.target.closest('#timeline-view') != null;
	var pointerInZoneView = e.target.closest('#zone-view') != null;
	let zone = getZoneAt(e);
	
	if (isWidgetButton) {
		if (pointerInCanvas && gCurrentWidget) {
			if (zone != null) {
				if (zone.IsSignwaveZone()) {
					return;
				}
				const media = zone.GetSelectedMedia();
				if (media != null && media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
					const canvasViewOffset = $.offset(zone.$canvasView);
					const button = await CreateButton(toCanvasScale(e.pageX - canvasViewOffset.left), toCanvasScale(e.pageY - canvasViewOffset.top));
					const buttonWidgetAttributes = JSON.parse(JSON.stringify(await GetWidgetAttributes(WidgetInformation.find(widget => widget.name == 'button'))));
					delete buttonWidgetAttributes.name;
					Object.assign(button, buttonWidgetAttributes);
					button.default = false;
					button.globalidle = false;
					gSelectedZone.GetSelectedMedia().attributes.buttons.push(button);
					gSelectedZone.UpdateView();
					if (await EditWidget(gSelectedZone, {
						button,
						widgettype: 'button',
						get attributes() { return button; },
						set attributes(attributes) { Object.assign(button, attributes); },
						ResizeMedia: () => {}
					}, gSelectedZone.x + gCurrentWidget.x + button.x, gSelectedZone.y + gCurrentWidget.y + button.y, button.width, button.height)) {
						// User pressed OK
					} else {
						// User cancelled, so delete the button
						gSelectedZone.GetSelectedMedia().attributes.buttons.pop();
						gSelectedZone.UpdateView();
					}
				}
			}
		}
		return;
	}
	
	let insertIndex = null;
	if (e.target.classList.contains('dropbefore') || e.target.classList.contains('dropafter')) {
		const $timelineItemOrZoneItem = e.target.parentNode;
		insertIndex = [...$timelineItemOrZoneItem.parentNode.children].indexOf($timelineItemOrZoneItem); // TODO: lol, this is needlessly expensive
		if (e.target.classList.contains('dropafter')) {
			insertIndex++;
		}
	}
	
	if (zone && zone.IsSignwaveZone()) {
		return;
	}
	
	const interleave = document.getElementById('interleave').classList.contains('selected');
	const selectedNonPendingNodes = isMedia ? $contentTree.getAllChildrenFromNodes(source).filter(node => node.type != 1).filter($contentTree.filterDraggable) : [];
	
	if (isMedia && selectedNonPendingNodes.length == 0) {
		return;
	}
	
	// Turn a dropped PDF into a widget
	const pdf = selectedNonPendingNodes.find(node => node.contenttreetype == 9);
	if (isMedia && pdf) {
		isMedia = false;
		isWidget = true;
		source.widget = WidgetInformation.find(widget => widget.name == 'pdf');
		await document.getElementById('pdfpdf').setSelected(node => node.contenttreetype == 9 && node.contentid == pdf.contentid);
	}
	
	if (isWidget) {
		const widget = source.widget;
		if (zone != null && zone.media.length != 0 && widget.name == 'signwave') {
			toastError('The Signwave widget must be placed in zone by itself. Either empty the zone or create a new zone for the Signwave widget.');
			return;
		}
	}
	
	if (interleave) {
		if (zone == null || zone.mediaSelected.length < 2) {
			toastError('You have enabled interleave mode, but you have not selected two pieces of sequential media to insert between. You can hold shift to select a range of media in a zone in the timeline. If you enabled interlave mode by accident then click to turn it off. (It is next to the search media button).');
			return;
		} else if (!zone.IsMediaSelectedSequential()) {
			toastError('You have enabled interleave mode, but you have not selected sequential media to insert between. You can hold shift to select a range of media in a zone in the timeline. If you enabled interlave mode by accident then click to turn it off. (It is next to the search media button).');
			return;
		}
	}
	
	// Audio
	const mediaAudio = selectedNonPendingNodes.filter(media => media.type == MEDIATYPE_AUDIO);
	if (isMedia && mediaAudio.length != 0) {
		var audioZone = gCurrentPlaylist.zones.find(zone => zone.name == 'Audio');
		if (audioZone == undefined) {
			audioZone = new Zone(-1, 'Audio', 0, 0, 0, 0, 0, 100, false);
			AddZone(audioZone);
			MoveZoneLayer(audioZone, -(gCurrentPlaylist.zones.length - 1));
		}
		
		var insertedAudio = [];
		for (const media of mediaAudio) {
			const mediaAudio = new MediaAudio(audioZone, {
				name: media.name,
				cdnurl: media.cdnurl,
				runlength: media.runlength,
				contentid: media.contentid,
				uploadedtimestamp: media.uploadedtimestamp,
				filesize: media.filesize
			});
			// Add the media where the user dropped either in the timeline or zone view
			insertedAudio.push(mediaAudio);
			audioZone.ConvertSelectedButtonZoneTrimToReferences();
			if (insertIndex == null) {
				audioZone.media.push(mediaAudio);
			} else {
				audioZone.media.splice(insertIndex, 0, mediaAudio);
				insertIndex++;
			}
			audioZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
		}
		
		var indexes = insertedAudio.map(media => audioZone.media.indexOf(media));
		UpdateZoneAndMediaDOM();
		SelectMedia(audioZone, indexes, false, false, false, pointerInCanvas ? 'canvas' : (pointerInTimelineView ? 'timeline' : 'zone'), true, true);
		SetSaveFlag(true);
	}
	
	// Image, Video, or Misc
	const imageVideoMisc = selectedNonPendingNodes.filter(node => !node.group && (node.contenttreetype == 3 || node.contenttreetype == 5 || node.contenttreetype == 8));
	if (isMedia && imageVideoMisc.length != 0 || isMovedMedia || isWidget) {
		if (isMovedMedia) {
			for (const findZone of gCurrentPlaylist.zones) {
				let foundMedia = false;
				for (let i = 0; i < findZone.media.length; ++i) {
					if (source.includes(findZone.media[i])) {
						findZone.media[i].Remove();
						findZone.media.splice(i--, 1);
						foundMedia = true;
					}
				}
				if (foundMedia) {
					findZone.mediaSelected = [];
				}
			}
		}
		if (zone == null) {
			zone = new Zone(-1, 'Zone0', 0, 0, 0, gCurrentPlaylist.pagewidth, gCurrentPlaylist.pageheight, 100, false);
			AddZone(zone);
		}
		
		const AddMediaToZone = async (interleaveInterval, interleaveRunlength) => {
			const interleaveCreators = [];
			const interleaveMedia = [];
			
			const insertedMedia = [];
			
			const InsertContent = async mediaCreator => {
				if (interleave) {
					interleaveCreators.push(mediaCreator);
				} else {
					// Add the media where the user dropped either in the timeline or zone view
					const media = await mediaCreator();
					insertedMedia.push(media);
					zone.ConvertSelectedButtonZoneTrimToReferences();
					if (insertIndex == null) {
						zone.media.push(media);
					} else {
						zone.media.splice(insertIndex, 0, media);
						insertIndex++;
					}
					zone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
				}
			};
			
			if (isMedia || isMovedMedia) {
				for (const media of isMedia ? imageVideoMisc : source) {
					switch (media.type) {
						case MEDIATYPE_IMAGE:
							await InsertContent(async () => {
								return new MediaImage(zone, {
									name: media.name,
									runlength: isMovedMedia ? media.runlength : configuration.defaultrunlength,
									cdnurl: media.cdnurl,
									actualwidth: isMovedMedia ? media.actualwidth : media.width,
									actualheight: isMovedMedia ? media.actualheight : media.height,
									contentid: media.contentid,
									uploadedtimestamp: media.uploadedtimestamp,
									filesize: media.filesize
								});
							});
							break;
						case MEDIATYPE_VIDEO:
							await InsertContent(async () => {
								return new MediaVideo(zone, {
									name: media.name,
									runlength: isMovedMedia ? media.runlength : media.runlength,
									cdnurl: media.cdnurl,
									actualwidth: isMovedMedia ? media.actualwidth : media.width,
									actualheight: isMovedMedia ? media.actualheight : media.height,
									contentid: media.contentid,
									uploadedtimestamp: media.uploadedtimestamp,
									filesize: media.filesize,
									hasaudio: media.hasaudio
								});
							});
							break;
						case MEDIATYPE_MISC:
							await InsertContent(async () => {
								return new MediaMisc(zone, {
									name: media.name,
									runlength: isMovedMedia ? media.runlength : configuration.defaultrunlength,
									cdnurl: media.cdnurl,
									contentid: media.contentid,
									uploadedtimestamp: media.uploadedtimestamp,
									filesize: media.filesize
								});
							});
							break;
						case MEDIATYPE_WIDGET:
							await InsertContent(async () => {
								return new MediaWidget(zone, {
									name: media.name,
									widgettype: media.widgettype,
									icon: media.icon,
									runlength: media.runlength,
									attributes: JSON.parse(JSON.stringify(media.attributes))
								});
							});
							break;
					}
				}
			} else if (isWidget) {
				await InsertContent(async () => {
					const widget = source.widget;
					const media = new MediaWidget(zone, {
						name: widget.display,
						widgettype: widget.name,
						icon: widget.icon,
						runlength: configuration.defaultrunlength,
						attributes: JSON.parse(JSON.stringify(await GetWidgetAttributes(widget)))
					});
					// Button Groups have a lot of zone settings that need to be done when a button group is created
					if (media.widgettype == 'buttongroup') {
						// Button groups have buttons that refer to zones which refer to media. The button group widget remembers the state of the last edited button group. In certain cases the zone numbers and media has changed. This code fixes those errors.
						var buttonGroupNames = {};
						for (const [buttonGroupMedia] of GetAllButtonGroups()) {
							buttonGroupNames[buttonGroupMedia.attributes.name] = true;
						}
						let uniqueButtonGroupName = 'Button Group';
						let uniqueButtonGroupId = 2;
						while (buttonGroupNames.hasOwnProperty(uniqueButtonGroupName)) {
							uniqueButtonGroupName = `Button Group ${uniqueButtonGroupId++}`;
						}
						media.attributes.name = uniqueButtonGroupName;
						/*
						let defaultExists = false;
						for (const [button, buttonMedia, buttonZone] of GetAllButtons()) {
							if (button.default) {
								defaultExists = true;
							}
						}
						*/
						const buttonNameExists = name => {
							for (const [button, buttonMedia, buttonZone] of GetAllButtons()) {
								if (button.name == name) {
									return true;
								}
							}
							return media.attributes.buttons.some(button => button.name == name);
						};
						
						for (const button of media.attributes.buttons) {
							button.default = false;
							let uniqueButtonName = button.name;
							if (/^Button [1-9][0-9]*$/.test(button.name)) {
								let uniqueButtonId = 1;
								// If the name is of the form "Button n" then find the next available "Button n" available.
								while (buttonNameExists(uniqueButtonName)) {
									uniqueButtonName = 'Button ' + uniqueButtonId++;
								}
							} else {
								// If the button has a user-defined name then add a number after it or find an unused number
								let uniqueButtonId = 2;
								while (buttonNameExists(uniqueButtonName)) {
									uniqueButtonName = button.name.match(/^(.*?)(?: [1-9][0-9]*)?$/)[1] + ' ' + uniqueButtonId++;
								}
							}
							button.name = uniqueButtonName;
							// Remove button zones if the dragged in buttons have more zones than what exist in the playlist
							while (button.zones.length > gCurrentPlaylist.zones.length) {
								button.zones.pop();
							}
							// Add button zones if the dragged in buttons have less zones than what exists in the playlist
							while (button.zones.length < gCurrentPlaylist.zones.length) {
								button.zones.push(
								{
									statechange: false,
									show: false,
									stop: false,
									restart: true,
									trimleft: -1,
									trimright: -1
								});
							}
							
							button.zones.forEach((button, zoneItr) => {
								var playlistZone = gCurrentPlaylist.zones[zoneItr];
								button.trimleft = playlistZone.media.length == 0 ? -1 : Math.min(button.trimleft, playlistZone.media.length - 1);
								button.trimright = playlistZone.media.length == 0 ? -1 : Math.min(button.trimright, playlistZone.media.length - 1);
							});
						}
						for (const button of media.attributes.buttons) {
							// Make sure the button name exists
							if (!buttonNameExists(button.idleactionbutton)) {
								button.idleactionbutton = 'none';
							}
						}
					}
					return media;
				});
			}
			
			if (interleave) {
				var sequentialIndexes = gSelectedZone.GetSortedMediaSelectedIndexes();
				var index = sequentialIndexes[0];
				var length = sequentialIndexes.length;
				var interval = interleaveInterval;
				var mediaToSelect = [];
				var offsetFromIndex = index;
				while (offsetFromIndex + interval < index + length) {
					offsetFromIndex += interval;
					for (const mediaCreator of interleaveCreators) {
						var media = await mediaCreator();
						interleaveMedia.push(media);
						if (media.type == MEDIATYPE_IMAGE || media.type == MEDIATYPE_MISC) {
							media.runlength = interleaveRunlength;
						}
						zone.ConvertSelectedButtonZoneTrimToReferences();
						zone.media.splice(offsetFromIndex, 0, media);
						zone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
						mediaToSelect.push(offsetFromIndex);
						offsetFromIndex++;
						length++;
					}
				}
				UpdateWidgetNames();
				UpdateZoneAndMediaDOM();
				SelectMedia(zone, mediaToSelect, false, false, false, pointerInCanvas ? 'canvas' : (pointerInTimelineView ? 'timeline' : 'zone'), true, true);
			} else {
				UpdateWidgetNames();
				UpdateZoneAndMediaDOM();
				var indexes = insertedMedia.map(media => zone.media.indexOf(media));
				SelectMedia(zone, indexes, false, false, false, pointerInCanvas ? 'canvas' : (pointerInTimelineView ? 'timeline' : 'zone'), true, true);
			}
			
			if (isWidget) {
				UpdateWidgetNames();
				
				$buttonTab.hidden = !gCurrentPlaylist.IsInteractive();
				if (!gCurrentPlaylist.IsInteractive() && $buttonTab.classList.contains('selected')) {
					document.getElementById('mediatab').click();
				}
				try {
					if (await EditWidget(zone, interleave ? interleaveMedia[0] : zone.mediaSelected[0])) {
						for (const media of interleaveMedia) {
							media.attributes = JSON.parse(JSON.stringify(interleaveMedia[0].attributes));
						}
					} else {
						deleteCurrentMedia(false);
					}
				} catch (exception) {
					// User cancelled edit
				}
			}
			SetSaveFlag(true);
		};
		
		if (interleave) {
			const $dialog = document.body.appendChild($.uidialog({ movable: { handle: ':scope > header' }, css: { color: 'white' } },
				$.header({ text: 'Interleave Media' }),
				$.div({ class: 'formgrid' },
					$.label({ for: 'interleaveInterval', text: 'Interval' }),
					$.input({ type: 'text', id: 'interleaveInterval', size: 4, maxlength: 4, title: 'The amoung of media items to skip in the selection before inserting.', class: 'tooltip', value: 1 }),
					$.label({ for: 'interleaveRunLength', text: 'Run-length' }),
					$.input({ type: 'text', id: 'interleaveRunLength', size: 4, maxlength: 4, title: 'The run-length of all inserted items except for video and audio.', class: 'tooltip', value: 10 })
				),
				$.div({ class: 'panel-buttons' },
					$.button({ text: 'Cancel', click: function() {
						$dialog.remove();
					} }),
					$.button({ text: 'OK', click: async function() {
						await AddMediaToZone(Math.max(1, parseInt(document.getElementById('interleaveInterval').value, 10)), Math.max(1, parseInt(document.getElementById('interleaveRunLength').value, 10)));
						$dialog.remove();
					} })
				)
			));
			centerOnScreen($dialog);
		} else {
			await AddMediaToZone();
		}
	}
}

function renderCanvasImage(context, image, x, y, width, height) {
	try {
		context.drawImage(image, x, y, width, height);
	} catch(e) {
	}
}

function hideAccordion() {
	document.getElementById('playlistPropertiesHeader').classList.add('collapsed');
	document.getElementById('propertiesHeader').classList.add('collapsed');
}

function getZoneAt(e) {
	if (gCurrentPlaylist == null) {
		return null;
	}
	// Check if the mouse is over a canvas view zone
	var zone = gCurrentPlaylist.zones.find(zone => zone.$canvasView.contains(e.target));
	if (zone != undefined) {
		return zone;
	}
	// Check if the mouse is over a timeline view zone
	zone = gCurrentPlaylist.zones.find(zone => zone.$timelineBar.contains(e.target));
	if (zone != undefined) {
		return zone;
	}
	// Check if the mouse is over a zone view zone
	zone = gCurrentPlaylist.zones.find(zone => zone.$zoneView.contains(e.target));
	if (zone != undefined) {
		return zone;
	}
	return null;
}

function TestFeatureResizeBox(move, x, y, mediaSelection, featureZone, featureMedia = null, featureComponent = null) {
	const maximumSnapDistance = 12;
	const insideZone = isPointInRect(x, y, featureZone.x - maximumSnapDistance, featureZone.y - maximumSnapDistance, featureZone.width + maximumSnapDistance * 2, featureZone.height + maximumSnapDistance * 2);
	
	const feature = featureComponent ? featureComponent : (mediaSelection ? featureMedia : featureZone);
	const featureX = featureComponent ? featureZone.x + featureMedia.x + featureComponent.x : (mediaSelection ? featureZone.x + featureMedia.x : featureZone.x);
	const featureY = featureComponent ? featureZone.y + featureMedia.y + featureComponent.y : (mediaSelection ? featureZone.y + featureMedia.y : featureZone.y);
	
	var moveTypeX = 0;
	var moveTypeY = 0;
	if (featureMedia == null || !featureZone.IsMediaLayoutWidthAndHeightReadonly()) {
		if (y >= featureY - maximumSnapDistance && y < featureY + feature.height + maximumSnapDistance) {
			if (Math.abs(featureX - x) < maximumSnapDistance) {
				moveTypeX = MOVETYPE_RESIZE_LEFT;
			}
			if (Math.abs(featureX + feature.width - x) < maximumSnapDistance) {
				moveTypeX = MOVETYPE_RESIZE_RIGHT;
			}
		}
		if (x >= featureX - maximumSnapDistance && x < featureX + feature.width + maximumSnapDistance) {
			if (Math.abs(featureY - y) < maximumSnapDistance) {
				moveTypeY = MOVETYPE_RESIZE_TOP;
			}
			if (Math.abs(featureY + feature.height - y) < maximumSnapDistance) {
				moveTypeY = MOVETYPE_RESIZE_BOTTOM;
			}
		}
	}
	const insideFeature = isPointInRect(x, y, featureX, featureY, feature.width, feature.height);
	if (moveTypeX == 0 && moveTypeY == 0 && insideFeature) {
		if (featureMedia) {
			if (!featureZone.IsMediaLayoutXReadonly()) {
				moveTypeX = MOVETYPE_MOVE_X;
			}
			if (!featureZone.IsMediaLayoutYReadonly()) {
				moveTypeY = MOVETYPE_MOVE_Y;
			}
		} else {
			moveTypeX = MOVETYPE_MOVE_X | MOVETYPE_MOVE_Y;
		}
	}
	let distance = Math.min
	(
		Math.abs(featureX - x),
		Math.abs(featureY - y),
		Math.abs(featureX + feature.width - x),
		Math.abs(featureY + feature.height - y)
	);
	// Give priority to the zone or media if the mouse is inside of it. When zones are snapped to an edge the mouse is equal distance to both edges so this breaks the tie
	if ((moveTypeX || moveTypeY) && distance < maximumSnapDistance && (distance < move.closestDistance || distance == move.closestDistance && insideFeature)) {
		move.closestDistance = distance;
		move.closestDistanceIsInsideFeature = insideFeature;
		move.zone = featureZone;
		move.media = featureMedia;
		move.component = featureComponent;
		move.type = moveTypeX | moveTypeY;
	}
	if (move.zone == null && move.media == null && insideFeature) {
		move.closestDistance = 0;
		move.zone = featureZone;
		move.media = featureMedia;
		move.component = featureComponent;
		move.type = moveTypeX | moveTypeY;
	}
}

function getZoneMediaMoveAt(move, x, y, mediaSelection) {
	for (const zone of gCurrentPlaylist.zones.filter(zone => zone.visible).map(zone => zone).reverse()) {
		if (zone.locked && isPointInRect(x, y, zone.x, zone.y, zone.width, zone.height) && !mediaSelection) {
			Object.assign(move, {
				zone: zone,
				media: null,
				component: null,
				type: MOVETYPE_NONE
			});
			return move;
		}
		
		if (mediaSelection) {
			var media = zone.GetSelectedMedia();
			if (media) {
				TestFeatureResizeBox(move, x, y, mediaSelection, zone, media);
			}
		} else if (!zone.locked) {
			TestFeatureResizeBox(move, x, y, mediaSelection, zone);
		}
		if (move.type != MOVETYPE_NONE && isPointInRect(x, y, zone.x, zone.y, zone.width, zone.height)) {
			break;
		}
	}
}

// Get the zone or media and move type based on where the mouse is over the playlist. Media is only returned if mediaSelection is true
function getMoveAt(x, y, mediaSelection) {
	let move =
	{
		zone: null,
		media: null,
		component: null,
		type: MOVETYPE_NONE,
		closestDistance: Number.POSITIVE_INFINITY,
		closestDistanceIsInsideFeature: false
	};
	if (gCurrentPlaylist) {
		getZoneMediaMoveAt(move, x, y, mediaSelection);
		// Only look at the current widget if the user is clicking inside, but outside of the outer resize zone regions
		if (gCurrentWidget && (move.type & MOVETYPE_MOVE_X || move.type & MOVETYPE_MOVE_Y)) {
			if (gCurrentWidget.widgettype == 'buttongroup') {
				// Ignore zone and media and check the widget components
				let componentMove =
				{
					zone: null,
					media: null,
					component: null,
					type: MOVETYPE_NONE,
					closestDistance: Number.POSITIVE_INFINITY,
					closestDistanceIsInsideFeature: false
				};
				for (const button of gCurrentWidget.attributes.buttons.map(button => button).reverse()) {
					TestFeatureResizeBox(componentMove, x, y, mediaSelection, gSelectedZone, gCurrentWidget, button);
					if (componentMove.type & MOVETYPE_MOVE_X || componentMove.type & MOVETYPE_MOVE_Y) {
						break;
					}
				}
				if (componentMove.component != null) {
					move = componentMove;
				}
			}
		}
	}
	return move;
}

function updateZoneLayers() {
	var visualZoneOrdering = GetVisualZoneOrder();
	var layer = 1;
	for (const zone of gCurrentPlaylist.zones) {
		if (zone.name == 'Audio') {
			zone.z = 0;
		} else {
			zone.z = layer;
			// Only update zones that follow the Zone123 format. Custom names are left custom.
			if (/^Zone[0-9]+$/.test(zone.name)) {
				zone.name = 'Zone' + (visualZoneOrdering.indexOf(zone) + 1 - (gCurrentPlaylist.zones[0].name == 'Audio' ? 1 : 0));
			}
			zone.$propertyView.value = zone.name;
			zone.$propertyView.innerText = zone.name;
			zone.$timelineName.querySelector('.name').value = zone.name;
			zone.$timelineName.querySelector('.order').innerText = visualZoneOrdering.indexOf(zone) + 1 - (gCurrentPlaylist.zones[0].name == 'Audio' ? 1 : 0);
			zone.$zoneViewName.querySelector(':scope > .name').value = zone.name;
			layer++;
		}
	}
}

function progressiveLoadImage(canvas, zone, media) {
	var context = canvas.getContext('2d');
	
	if ('imageLoaded' in media) {
		context.clearRect(0, 0, canvas.width, canvas.height);
		context.drawImage(media.image, 0, 0, media.width, media.height);
	} else if ('thumbnailImageLoaded' in media) {
		context.clearRect(0, 0, canvas.width, canvas.height);
		context.drawImage(media.thumbnailImage, 0, 0, media.width, media.height);
		media.image = new Image();
		media.image.crossOrigin = 'Anonymous';
		media.image.onload = () => {
			media.imageLoaded = true;
			context.clearRect(0, 0, canvas.width, canvas.height);
			context.drawImage(media.image, 0, 0, media.width, media.height);
		};
		media.image.src = GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize);
	} else {
		media.thumbnailImage = new Image();
		media.thumbnailImage.crossOrigin = 'Anonymous';
		media.thumbnailImage.onload = () => {
			media.thumbnailImageLoaded = true;
			context.clearRect(0, 0, canvas.width, canvas.height);
			context.drawImage(media.thumbnailImage, 0, 0, media.width, media.height);
			media.image = new Image();
			media.image.crossOrigin = 'Anonymous';
			media.image.onload = () => {
				media.imageLoaded = true;
				context.clearRect(0, 0, canvas.width, canvas.height);
				context.drawImage(media.image, 0, 0, media.width, media.height);
			};
			if (media.cdnurl != null && media.cdnurl != '') {
				media.image.src = GetCachedMediaURL(account, media.cdnurl, media.uploadedtimestamp, media.filesize);
			}
		};
		media.thumbnailImage.src = GetCachedMediaURL(account, media.icon, media.uploadedtimestamp, media.filesize);
	}
}

function progressiveLoadVideoPoster(canvas, zone, media) {
	var context = canvas.getContext('2d');
	
	if ('imageLoaded' in media) {
		context.clearRect(0, 0, canvas.width, canvas.height);
		context.drawImage(media.image, 0, 0, media.width, media.height);
	} else if ('thumbnailImageLoaded' in media) {
		context.clearRect(0, 0, canvas.width, canvas.height);
		context.drawImage(media.thumbnailImage, 0, 0, media.width, media.height);
		media.image = new Image();
		media.image.crossOrigin = 'Anonymous';
		media.image.onload = () => {
			media.imageLoaded = true;
			context.clearRect(0, 0, canvas.width, canvas.height);
			context.drawImage(media.image, 0, 0, media.width, media.height);
		};
		media.image.src = GetCachedMediaURL(account, media.posterURL, media.uploadedtimestamp, media.filesize);
	} else {
		media.thumbnailImage = new Image();
		media.thumbnailImage.crossOrigin = 'Anonymous';
		media.thumbnailImage.onload = () => {
			media.thumbnailImageLoaded = true;
			context.clearRect(0, 0, canvas.width, canvas.height);
			context.drawImage(media.thumbnailImage, 0, 0, media.width, media.height);
			media.image = new Image();
			media.image.crossOrigin = 'Anonymous';
			media.image.onload = () => {
				media.imageLoaded = true;
				context.clearRect(0, 0, canvas.width, canvas.height);
				context.drawImage(media.image, 0, 0, media.width, media.height);
			};
			media.image.src = GetCachedMediaURL(account, media.posterURL, media.uploadedtimestamp, media.filesize);
		};
		media.thumbnailImage.src = GetCachedMediaURL(account, media.icon, media.uploadedtimestamp, media.filesize);
	}
}

function reloadAllZoneViews() {
	if (gCurrentPlaylist == null) {
		return;
	}
	for (let i = 0; i < gCurrentPlaylist.zones.length; ++i) {
		gCurrentPlaylist.zones[i].ReloadView();
	}
}

function ResizePlaylist(playlistWidth, playlistHeight) {
	for (const zone of gCurrentPlaylist.zones) {
		const zoneWidth = zone.width;
		const zoneHeight = zone.height;
		zone.x = Math.round(zone.x / playlistWidth * gCurrentPlaylist.pagewidth);
		zone.y = Math.round(zone.y / playlistHeight * gCurrentPlaylist.pageheight);
		zone.width = Math.round(zone.width / playlistWidth * gCurrentPlaylist.pagewidth);
		zone.height = Math.round(zone.height / playlistHeight * gCurrentPlaylist.pageheight);
		for (const media of zone.media) {
			media.ResizeMedia(zoneWidth, zoneHeight);
			if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
				for (const button of media.attributes.buttons) {
					button.x = Math.round(button.x / playlistWidth * gCurrentPlaylist.pagewidth);
					button.y = Math.round(button.y / playlistHeight * gCurrentPlaylist.pageheight);
					button.width = Math.round(button.width / playlistWidth * gCurrentPlaylist.pagewidth);
					button.height = Math.round(button.height / playlistHeight * gCurrentPlaylist.pageheight);
				}
			}
		}
	}
}

const playlistPropertiesChange = ['pagewidth', 'pageheight'];
function PopulatePlaylistProperty(property, defaultValue) {
	const $tag = document.getElementById(`playlist-${property}`);
	const events = playlistPropertiesChange.indexOf(property) != -1 ? ['change', 'keydown'] : '';
	if ($tag.lastAlterProperty) {
		for (const event of events) {
			$tag.removeEventListener(event, $tag.lastAlterProperty);
		}
		delete $tag.lastAlterProperty;
	}
	
	let value = gCurrentPlaylist[property];
	if (typeof defaultValue == 'boolean') {
		$tag.checked = value == null ? false : value;
	} else {
		$tag.value = value == null ? defaultValue : value;
	}
	
	$tag.lastAlterProperty = async e => {
		if (e.type == 'keydown') {
			if (e.key == 'Enter') {
				$tag.blur();
			}
			return;
		}
		let value = (() => {
			switch (typeof defaultValue) {
				case 'string':
					return $tag.value;
				case 'number':
					return parseInt($tag.value, 10);
				case 'boolean':
					return $tag.checked;
			}
		})();
		
		if (['pagewidth', 'pageheight'].indexOf(property) != -1) {
			if (!await QuestionBox('Modify playlist dimensions?', `Are you sure you want to modify the playlist dimensions to ${property == 'pagewidth' ? value : gCurrentPlaylist.pagewidth}x${property == 'pageheight' ? value : gCurrentPlaylist.pageheight}?`)) {
				return;
			}
		}
		
		if (typeof defaultValue == 'number' && isNaN(value) || value === '' || value == gCurrentPlaylist[property]) { // Need to use value === since we care about 0
			if (e.type == 'blur') {
				PopulatePlaylistProperty(property, defaultValue);
			}
			return;
		}
		
		switch (property) {
			case 'name':
				if (value == 'Audio') {
					return;
				}
				break;
		}
		
		for (const event of events) {
			$tag.removeEventListener(event, $tag.lastAlterProperty);
		}
		
		const playlistWidth = gCurrentPlaylist.pagewidth;
		const playlistHeight = gCurrentPlaylist.pageheight;
		
		gCurrentPlaylist[property] = value;
		
		switch (property) {
			case 'pagewidth':
			case 'pageheight':
				ResizePlaylist(playlistWidth, playlistHeight);
				
				//PopulateZoneLayoutProperties(zone);
				//PopulateMediaLayoutProperties(zone);
				//zone.UpdateView();
				LineLayerClear();
				SetSaveFlag(true);
				return;
		}
		for (const event of events) {
			$tag.addEventListener(event, $tag.lastAlterProperty);
		}
		SetSaveFlag(true);
	};
	
	for (const event of events) {
		$tag.addEventListener(event, $tag.lastAlterProperty);
	}
}

const zonePropertiesInputBlur = ['x', 'y', 'width', 'height', 'name'];
const zonePropertiesChange = ['locked', 'volume'];
function PopulateZoneProperty(zone, property, defaultValue) {
	const $tag = document.querySelector(`#zone-${property}`);
	const events = zonePropertiesInputBlur.indexOf(property) != -1 ? ['input', 'blur'] : ['change', 'keydown'];
	if ($tag.lastAlterProperty) {
		for (const event of events) {
			$tag.removeEventListener(event, $tag.lastAlterProperty);
		}
		delete $tag.lastAlterProperty;
	}
	
	let value;
	if (property == 'duration') {
		value = zone.media.reduce((duration, media) => duration + media.runlength, 0);
	} else {
		value = zone[property];
	}
	if (typeof defaultValue == 'boolean') {
		$tag.checked = value == null ? false : value;
	} else {
		$tag.value = value == null ? (zone.mediaSelected.length == 1 ? defaultValue : '') : value;
	}
	
	switch (property) {
		case 'x':
		case 'y':
		case 'width':
		case 'height':
			$tag.readonly = zone.locked;
			$tag.title = zone.locked ? 'The layout cannot be changed when the zone layout is locked.' : '';
			break;
	}
	
	if (['x', 'y', 'width', 'height'].indexOf(property) != -1) {
		const $layout = document.querySelector('#zone-header-layout > span:not(.mdi)');
		$layout.querySelector(':scope > label[for="zone-x"]').innerText = zone.x;
		$layout.querySelector(':scope > label[for="zone-y"]').innerText = zone.y;
		$layout.querySelector(':scope > label[for="zone-width"]').innerText = zone.width;
		$layout.querySelector(':scope > label[for="zone-height"]').innerText = zone.height;
	}
	
	$tag.lastAlterProperty = e => {
		if (e.type == 'keydown') {
			if (e.key == 'Enter') {
				$tag.blur();
			}
			return;
		}
		let value = (() => {
			switch (typeof defaultValue) {
				case 'string':
					return $tag.value;
				case 'number':
					return parseInt($tag.value, 10);
				case 'boolean':
					return $tag.checked;
			}
		})();
		
		if (typeof defaultValue == 'number' && isNaN(value) || value === '' || value == zone[property]) { // Need to use value === since we care about 0
			if (e.type == 'blur') {
				PopulateZoneProperty(zone, property, defaultValue);
			}
			return;
		}
		
		switch (property) {
			case 'name':
				if (value == 'Audio') {
					return;
				}
				break;
		}
		
		for (const event of events) {
			$tag.removeEventListener(event, $tag.lastAlterProperty);
		}
		
		const zoneWidth = zone.width;
		const zoneHeight = zone.height;
		
		zone[property] = value;
		
		switch (property) {
			case 'x':
			case 'y':
			case 'width':
			case 'height':
				zone.width = zone.width > gCurrentPlaylist.pagewidth ? gCurrentPlaylist.pagewidth : zone.width;
				zone.height = zone.height > gCurrentPlaylist.pageheight ? gCurrentPlaylist.pageheight : zone.height;
				zone.x = zone.x + zone.width > gCurrentPlaylist.pagewidth ? gCurrentPlaylist.pagewidth - zone.width : zone.x;
				zone.y = zone.y + zone.height > gCurrentPlaylist.pageheight ? gCurrentPlaylist.pageheight - zone.height : zone.y;
				zone.x = zone.x < 0 ? 0 : zone.x;
				zone.y = zone.y < 0 ? 0 : zone.y;
				
				for (const media of zone.media) {
					media.ResizeMedia(zoneWidth, zoneHeight);
				}
				
				PopulateZoneLayoutProperties(zone);
				PopulateMediaLayoutProperties(zone);
				zone.UpdateView();
				LineLayerClear();
				LineLayerHighlightZone(zone, true);
				SetSaveFlag(true);
				return;
			case 'name':
				updateZoneLayers();
				UpdateZoneAndMediaDOM();
				UpdateButtonList();
				if (e.type == 'blur') {
					PopulateZoneProperty(zone, 'name', '');
				} else {
					for (const event of events) {
						$tag.addEventListener(event, $tag.lastAlterProperty);
					}
				}
				SetSaveFlag(true);
				return;
			case 'locked':
				PopulateZoneLayoutProperties(zone);
				break;
		}
		for (const event of events) {
			$tag.addEventListener(event, $tag.lastAlterProperty);
		}
		SetSaveFlag(true);
	};
	for (const event of events) {
		$tag.addEventListener(event, $tag.lastAlterProperty);
	}
}

function PopulatePlaylistLayoutProperties() {
	PopulatePlaylistProperty('pagewidth', 0);
	PopulatePlaylistProperty('pageheight', 0);
}

function PopulateZoneLayoutProperties(zone) {
	PopulateZoneProperty(zone, 'x', 0);
	PopulateZoneProperty(zone, 'y', 0);
	PopulateZoneProperty(zone, 'width', 0);
	PopulateZoneProperty(zone, 'height', 0);
}

function applyStartEndDateClamping(property, value) {
	const $element = document.getElementById(`media-${property}`);
	const parsedDate = moment(value, 'YYYY-MM-DD HH:mm:ss');
	if (parsedDate.isBefore(moment($element.getAttribute('min'), 'YYYY-MM-DD'))) {
		return moment($element.getAttribute('min') + ' ' + moment.utc(value).local().format('HH:mm:ss'), 'YYYY-MM-DD HH:mm:ss').utc().format('YYYY-MM-DD HH:mm:ss');
	}
	if (parsedDate.isAfter(moment($element.getAttribute('max'), 'YYYY-MM-DD'))) {
		return moment($element.getAttribute('max') + ' ' + moment.utc(value).local().format('HH:mm:ss'), 'YYYY-MM-DD HH:mm:ss').utc().format('YYYY-MM-DD HH:mm:ss');
	}
	return value;
}

// TODO: Modify these to use the widget type format
const mediaPropertiesInputBlur = ['x', 'y', 'width', 'height', 'button-x', 'button-y', 'button-width', 'button-height', 'starttime', 'runlength'];
const mediaPropertiesChange = ['entrytransitiontype', 'entrytransitionduration', 'exittransitiontype', 'exittransitionduration', 'fillstyleid', 'zoompan', 'volume', 'ducking', 'scheduled', 'utc', 'scheduledaterange', 'scheduledayparting', 'sunday', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'button-metadata'];
const mediaPropertiesBlur = ['schedulestartdate', 'schedulestarttime', 'scheduleenddate', 'scheduleendtime', 'daypartstarttime', 'daypartendtime'];

function PopulateMediaProperty(zone, property, defaultValue) {
	const $tag = document.getElementById(`media-${property}`);
	
	const selectedMedia = zone.GetAllSelectedMedia();
	
	const value = zone.GetMediaValues(selectedMedia, property);
	// If multiselect is used and the property values aren't the same (value is null) then use keydown instead as we only update on enter and blur rather than any key change
	let events = [];
	if (mediaPropertiesInputBlur.includes(property)) {
		events = ['keydown','input', 'blur'];
	} else if (mediaPropertiesChange.includes(property)) {
		events = ['change'];
	} else if (mediaPropertiesBlur.includes(property)) {
		events = ['blur'];
	} else {
		throw `Property ${property} not associated with event.`;
	}
	if ($tag.lastAlterProperty) {
		for (const event of events) {
			$tag.removeEventListener(event, $tag.lastAlterProperty);
		}
		delete $tag.lastAlterProperty;
	}
	
	if (typeof defaultValue == 'object') {
		if (permissions['Metadata']) {
			if (JSON.stringify(value) != JSON.stringify($tag.data)) {
				Metadata($tag, value, []);
				$tag.data = value;
			}
		}
	} else if (typeof defaultValue == 'boolean') {
		$tag.checked = value == null ? false : value;
		$tag.indeterminate = false;
		if (value == null && zone.mediaSelected.length != 1) {
			$tag.checked = true;
			$tag.indeterminate = true;
		}
	} else {
		$tag.value = value == null ? (zone.mediaSelected.length == 1 ? defaultValue : '') : value;
	}
	
	switch (property) {
		case 'x':
			$tag.readOnly = zone.GetSelectedMediaXReadOnly();
			break;
		case 'y':
			$tag.readOnly = zone.GetSelectedMediaYReadOnly();
			break;
		case 'width':
			$tag.readOnly = zone.GetSelectedMediaWidthOrHeightReadOnly();
			break;
		case 'height':
			$tag.readOnly = zone.GetSelectedMediaWidthOrHeightReadOnly();
			break;
		case 'sunday':
		case 'monday':
		case 'tuesday':
		case 'wednesday':
		case 'thursday':
		case 'friday':
		case 'saturday':
			$tag.title = `${value == null || value == false ? 'Enable' : 'Disable'} dayparting for ${property}`;
			break;
	}
	
	const displayMultiselectHeaderProperty = (property, modifier = value => value) => {
		const value = zone.GetMediaValues(selectedMedia, property);
		return value == null ? '?' : modifier(value);
	};
	
	if (['x', 'y', 'width', 'height'].indexOf(property) != -1) {
		const $layout = document.querySelector('#media-header-layout > span:not(.mdi)');
		$layout.querySelector(':scope > label[for="media-x"]').innerText = displayMultiselectHeaderProperty('x');
		$layout.querySelector(':scope > label[for="media-y"]').innerText = displayMultiselectHeaderProperty('y');
		$layout.querySelector(':scope > label[for="media-width"]').innerText = displayMultiselectHeaderProperty('width');
		$layout.querySelector(':scope > label[for="media-height"]').innerText = displayMultiselectHeaderProperty('height');
	}
	
	if (['starttime', 'length', 'width', 'height'].indexOf(property) != -1) {
		const layoutTag = document.querySelector('#media-header-time > span:not(.mdi)');
		layoutTag.querySelector(':scope > label[for="media-starttime"]').innerText = `Start: ${displayMultiselectHeaderProperty('starttime')}s`;
		layoutTag.querySelector(':scope > label[for="media-runlength"]').innerText = `Run-length: ${displayMultiselectHeaderProperty('runlength')}s`;
		const $mediaRunlength = document.getElementById('media-runlength');
		$mediaRunlength.title = displayMultiselectHeaderProperty('runlength', value => humanizeDuration(value * 1000));
	}
	
	if (['entrytransitiontype', 'entrytransitionduration', 'exittransitiontype', 'exittransitionduration'].indexOf(property) != -1) {
		var layoutTag = document.querySelector('#media-header-transition > span:not(.mdi)');
		layoutTag.querySelector(':scope > label[for="media-entrytransitiontype"]').innerText = displayMultiselectHeaderProperty('entrytransitiontype');
		layoutTag.querySelector(':scope > label[for="media-entrytransitionduration"]').innerText = `${displayMultiselectHeaderProperty('entrytransitionduration', value => parseInt(value, 10) / 1000)}s`;
		layoutTag.querySelector(':scope > label[for="media-exittransitiontype"]').innerText = displayMultiselectHeaderProperty('exittransitiontype');
		layoutTag.querySelector(':scope > label[for="media-exittransitionduration"]').innerText = `${displayMultiselectHeaderProperty('exittransitionduration', value => parseInt(value, 10) / 1000)}s`;
	}
	
	if (['audio', 'ducking'].indexOf(property) != -1) {
		var layoutTag = document.querySelector('#media-header-audio > span:not(.mdi)');
		layoutTag.querySelector(':scope > label[for="media-volume"]').innerText = `Volume: ${displayMultiselectHeaderProperty('volume', value => value == 0 ? 'Mute' : `${value}%`)}`;
		layoutTag.querySelector(':scope > label[for="media-ducking"]').innerText = `Ducking: ${displayMultiselectHeaderProperty('ducking', value => value == 0 ? 'Mute' : `${value}%`)}`;
	}
	
	if (['scheduled'].indexOf(property) != -1) {
		var layoutTag = document.querySelector('#media-header-scheduling > span:not(.mdi)');
		layoutTag.innerText = displayMultiselectHeaderProperty('scheduled', value => value ? 'Enabled' : 'Disabled');
	}
	
	if (['button-x', 'button-y', 'button-width', 'button-height'].indexOf(property) != -1) {
		var layoutTag = document.querySelector('#media-header-buttonlayout > span:not(.mdi)');
		layoutTag.querySelector(':scope > label[for="media-button-x"]').innerText = displayMultiselectHeaderProperty('button-x');
		layoutTag.querySelector(':scope > label[for="media-button-y"]').innerText = displayMultiselectHeaderProperty('button-y');
		layoutTag.querySelector(':scope > label[for="media-button-width"]').innerText = displayMultiselectHeaderProperty('button-width');
		layoutTag.querySelector(':scope > label[for="media-button-height"]').innerText = displayMultiselectHeaderProperty('button-height');
	}
	
	$tag.lastAlterProperty = async e => {
		if (e.type == 'focus') {
			// Ensures that blur events are only followed by associated focus events. Any call to UpdateProperties while the user is focused on an input would fire a blur event if this isn't done.
			// More specifically imagine the user selects media in one zone (bind focus on the input) clicks on the input box (focus fired and bind blur) and then clicks on another media item (SelectMedia fires the last associated blur and calls UpdateProperties which unbinds blur and focus and registers a focus associated with the new media item).
			// A subtle problem is that pointerdown when selecting the second media item also queues a blur event. Without this order of event binding the second media item would capture this blur calling alterProperty needlessly.
			$tag.removeEventListener('focus', $tag.lastAlterProperty);
			$tag.addEventListener('blur', $tag.lastAlterProperty);
			return;
		} else if (e.type == 'blur') {
			$tag.removeEventListener('blur', $tag.lastAlterProperty);
			$tag.addEventListener('focus', $tag.lastAlterProperty);
		} else if (e.type == 'keydown') {
			if (e.key == 'Enter') {
				$tag.blur();
				e.stopImmediatePropagation();
				// When multiple media are selected and the question box is shown the yes button is focused. Without prevent default the enter action dispatches a click event. Stop propagation doesn't stop this.
				e.preventDefault();
			}
			return;
		}
		
		let value = (() => {
			switch (typeof defaultValue) {
				case 'string':
					return $tag.value;
				case 'number':
					return parseInt($tag.value, 10);
				case 'boolean':
					return $tag.checked;
				case 'object':
					return $tag.data;
			}
		})();
		
		const currentValue = zone.GetMediaValues(selectedMedia, property);
		if (typeof defaultValue == 'number' && isNaN(value) || value === '' || value == currentValue) { // Need to use value === since we care about 0
			if (e.type == 'blur') {
				PopulateMediaProperty(zone, property, defaultValue);
			}
			return;
		}
		
		try {
			if (currentValue == null && e.type == 'blur') {
				await new Promise(async (resolve, reject) => {
					var question = (() => {
						var optionValueToName = (id, value) => {
							return document.querySelector(`#${id} option[value="${value}"]`).innerText;
						};
						switch (property) {
							case 'x': return `Are you sure you want to set the left offset to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'y': return `Are you sure you want to set the top offset to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'width': return `Are you sure you want to set the width to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'height': return `Are you sure you want to set the height to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'runlength': return `Are you sure you want to set the run-length to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'entrytransitiontype': return `Are you sure you want to set the entry transition to '${optionValueToName('media-entrytransitiontype', value)}' for all the selected media in the zone '${zone.name}'?`;
							case 'entrytransitionduration': return `Are you sure you want to set the entry transition duration to ${optionValueToName('media-entrytransitionduration', value)} for all the selected media in the zone '${zone.name}'?`;
							case 'exittransitiontype': return `Are you sure you want to set the exit transition to '${optionValueToName('media-exittransitiontype', value)}' for all the selected media in the zone?`;
							case 'exittransitionduration': return `Are you sure you want to set the exit transition duration to ${optionValueToName('media-exittransitionduration', value)} for all the selected media in the zone '${zone.name}'?`;
							case 'schedulestartdate': return `Are you sure you want to set the schedule start date for all the selected media in the zone '${zone.name}'?`;
							case 'schedulestarttime': return `Are you sure you want to set the schedule start time for all the selected media in the zone '${zone.name}'?`;
							case 'scheduleenddate': return `Are you sure you want to set the schedule end date for all the selected media in the zone '${zone.name}'?`;
							case 'scheduleendtime': return `Are you sure you want to set the schedule end time for all the selected media in the zone '${zone.name}'?`;
							case 'daypartstarttime': return `Are you sure you want to set the schedule dayparting start time for all the selected media in the zone '${zone.name}'?`;
							case 'daypartendtime': return `Are you sure you want to set the schedule dayparting end time for all the selected media in the zone '${zone.name}'?`;
							case 'sunday':
							case 'monday':
							case 'tuesday':
							case 'wednesday':
							case 'thursday':
							case 'friday':
							case 'saturday':
								return `Are you sure you want to ${value ? 'enable' : 'disable'} dayparting for ${property} for all selected media in the zone '${zone.name}'?`;
							case 'scheduledayparting': return `Are you sure you want to ${value ? 'enable' : 'disable'} schedule dayparting for all selected media in the zone '${zone.name}'?`;
							case 'utc': return `Are you sure you want to ${value ? 'enable' : 'disable'} UTC for all the selected media in the zone '${zone.name}'?`;
							case 'scheduledaterange': return `Are you sure you want to ${value ? 'enable' : 'disable'} schedule date ranges for all the selected media in the zone '${zone.name}'?`;
							case 'scheduled': return `Are you sure you want to ${value ? 'enable' : 'disable'} the scheduling for all selected media in the zone '${zone.name}'? Disabling resets the scheduled ranges and dayparting settings to their defaults.`;
							case 'volume': return `Are you sure you want to set the volume to ${value == 0 ? 'mute' : `${value}%`} for all the selected media in the zone '${zone.name}'?`;
							case 'ducking': return `Are you sure you want to set the ducking to ${value == 0 ? 'mute' : `${value}%`} for all the selected media in the zone '${zone.name}'?`;
							case 'zoompan': return `Are you sure you want to ${value ? 'enable' : 'disable'} zoom pan for all the selected media in the zone '${zone.name}'?`;
							case 'fillstyleid': return `Are you sure you want to set the fill style to ${optionValueToName('media-fillstyleid', value)} for all the selected media in the zone '${zone.name}'?`;
							default: throw true;
						}
					})();
					if (await QuestionBox('Are you sure?', question)) {
						resolve();
					} else {
						reject();
					}
				});
			}
			if (currentValue != null || currentValue == null && (e.type == 'blur' || e.type == 'change')) {
				zone.SetMediaValues(selectedMedia, property, value);
				zone.UpdateView();
				UpdateZoneAndMediaDOM();
				UpdateProperties();
				LineLayerClear();
				SetSaveFlag(true);
			}
			return false;
		} catch (exception) {
			// User cancelled the change
			UpdateProperties();
			return true;
		}
	};
	for (const event of events) {
		$tag.addEventListener(event, $tag.lastAlterProperty);
	}
}

function PopulateMediaLayoutProperties(zone) {
	PopulateMediaProperty(zone, 'x', 0);
	PopulateMediaProperty(zone, 'y', 0);
	PopulateMediaProperty(zone, 'width', 100);
	PopulateMediaProperty(zone, 'height', 100);
}

function PopulateMediaStartRunlengthProperties(zone) {
	PopulateMediaProperty(zone, 'starttime', 0);
	PopulateMediaProperty(zone, 'runlength', 10);
	document.getElementById('media-runlength').readOnly = zone.GetSelectedMediaRunlengthReadOnly();
}

function PopulateButtonLayoutProperties(zone) {
	PopulateMediaProperty(zone, 'button-x', 0);
	PopulateMediaProperty(zone, 'button-y', 0);
	PopulateMediaProperty(zone, 'button-width', 0);
	PopulateMediaProperty(zone, 'button-height', 0);
}

function PopulateButtonMetadataProperties(zone) {
	PopulateMediaProperty(zone, 'button-metadata', []);
}

function PopulateWidgetProperty(zone, attribute) {
	const selectedMedia = zone.GetAllSelectedMedia();
	
	const value = zone.GetWidgetValues(selectedMedia, attribute);
	// If multiselect is used and the property values aren't the same (value is null) then use keydown instead as we only update on enter and blur rather than any key change
	const events = (() => {
		switch (attribute.type) {
			case 'bool':
				return ['change'];
			case 'string':
				if (attribute.hasOwnProperty('design') && (attribute.design == 'fontfamily' || attribute.design == 'rssfeed' || attribute.design == 'csvfile' || attribute.design == 'eventlist')) {
					return ['change'];
				} else if (attribute.hasOwnProperty('options')) {
					return ['change'];
				}
				return value == null ? ['keydown', 'focus'] : ['input', 'focus'];
			case 'int':
				if (attribute.hasOwnProperty('options')) {
					return ['change'];
				}
				return value == null ? ['keydown', 'focus'] : ['input', 'focus'];
			case 'float':
				if (attribute.hasOwnProperty('options')) {
					return ['change'];
				}
				return value == null ? ['keydown', 'focus'] : ['input', 'focus'];
			case 'color':
				return ['change'];
			case 'date':
				return ['change'];
			case 'time':
				return ['change'];
			case 'json':
				return ['change'];
			default:
				throw 'attribute type needs to be handled';
		}
	})();
	if (attribute.$input.lastAlterProperty) {
		for (const event of ['keydown', 'input', 'focus', 'blur', 'change']) {
			attribute.$input.removeEventListener(event, attribute.$input.lastAlterProperty);
		}
		delete attribute.$input.lastAlterProperty;
	}
	
	attribute.value = value;
	
	const displayMultiselectHeaderProperty = (attribute, modifier = value => value) => {
		const value = zone.GetWidgetValues(selectedMedia, attribute);
		return value == null ? '?' : modifier(value);
	};
	
	if (attribute.hasOwnProperty('$headerLabel')) {
		attribute.$headerLabel.innerText = displayMultiselectHeaderProperty(attribute);
	}
	
	attribute.$input.lastAlterProperty = async e => {
		if (e.type == 'focus') {
			attribute.$input.removeEventListener('focus', attribute.$input.lastAlterProperty);
			attribute.$input.addEventListener('blur', attribute.$input.lastAlterProperty);
			return;
		} else if (e.type == 'blur') {
			attribute.$input.removeEventListener('blur', attribute.$input.lastAlterProperty);
			attribute.$input.addEventListener('focus', attribute.$input.lastAlterProperty);
		} else if (e.type == 'keydown') {
			if (e.key == 'Enter') {
				attribute.$input.blur();
			}
			return;
		}
		
		let value = await attribute.value;
		
		const currentValue = zone.GetWidgetValues(selectedMedia, attribute);
		if (typeof attribute.default == 'number' && isNaN(value) || value === '' || value == currentValue) { // Need to use value === since we care about 0
			if (e.type == 'blur') {
				PopulateWidgetProperty(zone, attribute);
			}
			return;
		}
		
		try {
			if (currentValue == null) {
				await new Promise(async (resolve, reject) => {
					var question = (() => {
						var optionValueToName = () => {
							const name = attribute.$input.querySelector(`option[value="${value}"]`).innerText;
							return `Are you sure you want to set the ${attribute.display} to '${name}' for all the selected media in the zone '${zone.name}'?`;
						};
						switch (attribute.type) {
							case 'bool':
								return `Are you sure you want to ${value ? 'enable' : 'disable'} the ${attribute.display} for all the selected media in the zone '${zone.name}'?`;
								break;
							case 'string':
								if (attribute.hasOwnProperty('options')) {
									return handleOption();
								} else if (attribute.hasOwnProperty('design') && attribute.design == 'fontfamily') {
									throw 'Not implemented yet...';
								}
								return `Are you sure you want to set the ${attribute.display} to '${value}' for all the selected media in the zone '${zone.name}'?`;
							case 'int':
								if (attribute.hasOwnProperty('options')) {
									return handleOption();
								}
								return `Are you sure you want to set the ${attribute.display} to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'float':
								if (attribute.hasOwnProperty('options')) {
									return handleOption();
								}
								return `Are you sure you want to set the ${attribute.display} to ${value} for all the selected media in the zone '${zone.name}'?`;
							case 'color':
								return `Are you sure you want to set the ${attribute.display} to ${value} for all the selected media in the zone '${zone.name}'?`;
								break;
							case 'date':
								throw 'Not implemented yet...';
								break;
							case 'time':
								throw 'Not implemented yet...';
								break;
							case 'json':
								throw 'Not implemented yet...';
								break;
							default:
								throw 'attribute type needs to be handled';
						}
					})();
					if (await QuestionBox('Are you sure?', question)) {
						resolve();
					} else {
						reject();
					}
				});
			}
			zone.SetWidgetValues(selectedMedia, attribute, value);
			
			zone.UpdateView();
			UpdateZoneAndMediaDOM();
			UpdateProperties();
			LineLayerClear();
			SetSaveFlag(true);
			return false;
		} catch (exception) {
			// User cancelled the change
			UpdateProperties();
			return true;
		}
	};
	for (const event of events) {
		attribute.$input.addEventListener(event, attribute.$input.lastAlterProperty);
	}
}

function ZoneSelectChanged() {
	SelectZone(gCurrentPlaylist.zones.find(zone => zone.name == this.value), true, false);
}

async function UpdateProperties(view) {
	if (view != undefined) {
		gPropertiesView = view;
	}
	
	// Record changes in hidden state
	const hiddenChanges = new Set();
	function setHidden($element, hidden) {
		if (hiddenChanges.has($element) && $element.hidden == hidden) {
			hiddenChanges.delete($element);
		} else if (!hiddenChanges.has($element) && $element.hidden != hidden) {
			hiddenChanges.add($element);
		}
	}
	
	const zoneView = gPropertiesView == 'zone';
	const mediaView = gPropertiesView == 'media';
	const metadataView = gPropertiesView == 'metadata';
	
	setHidden(document.getElementById('propertiesHeader'), true);
	document.getElementById('propertiesHeader').classList.add('collapsed');
	setHidden(document.getElementById('metadataHeader'), true);
	document.getElementById('metadataHeader').classList.add('collapsed');
	
	setHidden(document.getElementById('playlistPropertiesHeader'), gCurrentPlaylist == null);
	setHidden(document.getElementById('playlistProperties'), gCurrentPlaylist == null);
	if (gCurrentPlaylist) {
		PopulatePlaylistLayoutProperties();
	}
	
	if (zoneView || mediaView) {
		const zone = gSelectedZone;
		const $propertiesHeader = document.getElementById('propertiesHeader');
		setHidden($propertiesHeader, false);
		$propertiesHeader.classList.remove('collapsed');
		$propertiesHeader.title = `Expand this to view and change the ${zoneView ? 'zone' : 'media'} properties`;
		$propertiesHeader.querySelector(':scope > span').innerText = zone ? (zoneView ? 'Zone Properties' : 'Media Properties') : 'No Selection';
		
		const $collapsedHeaders = document.querySelectorAll('#properties .header.collapsed');
		for (const $collapsedHeader of $collapsedHeaders) { $collapsedHeader.classList.remove('collapsed'); }
		
		for (const $element of document.querySelectorAll('#properties > header')) {
			setHidden($element, true);
		}
		setHidden(document.querySelector('#properties > .iconlist'), true);
		setHidden($zoneSelect, true);
		setHidden($mediaSelect, true);
		setHidden(document.getElementById('media-header-buttonmetadata'), true);
		setHidden(document.getElementById('media-header-buttonoptions'), true);
		setHidden(document.getElementById('media-header-widgetnooptions'), true);
		
		if (zone) {
			setHidden($zoneSelect, false);
			setHidden($mediaSelect, false);
			
			setHidden(document.querySelector('#properties > .iconlist'), !mediaView);
			setHidden(document.getElementById('zone-header-layout'), !zoneView);
			setHidden(document.getElementById('zone-header-details'), !zoneView);
			
			if (zoneView) {
				PopulateZoneProperty(zone, 'id', 0);
				PopulateZoneProperty(zone, 'name', '');
				PopulateZoneProperty(zone, 'duration', 0);
				PopulateZoneLayoutProperties(zone);
				PopulateZoneProperty(zone, 'locked', false);
				PopulateZoneProperty(zone, 'volume', 100);
			}
			if (zone.mediaSelected.length > 0 && mediaView) {
				const populateMediaLayout = () => {
					setHidden(document.getElementById('media-header-layout'), false);
					PopulateMediaLayoutProperties(zone);
				};
				const populateMediaStartRunlength = () => {
					setHidden(document.getElementById('media-header-time'), false);
					PopulateMediaStartRunlengthProperties(zone);
				};
				const populateMediaVolume = () => {
					setHidden(document.getElementById('media-header-audio'), false);
					PopulateMediaProperty(zone, 'volume', 0);
				};
				const populateMediaDucking = () => {
					setHidden(document.getElementById('media-header-audio'), false);
					PopulateMediaProperty(zone, 'ducking', 0);
				};
				const populateMediaTransitions = () => {
					setHidden(document.getElementById('media-header-transition'), false);
					PopulateMediaProperty(zone, 'entrytransitiontype', 'none');
					PopulateMediaProperty(zone, 'entrytransitionduration', 0);
					PopulateMediaProperty(zone, 'exittransitiontype', 'none');
					PopulateMediaProperty(zone, 'exittransitionduration', 0);
				};
				const populateMediaSchedule = () => {
					setHidden(document.getElementById('media-header-scheduling'), false);
					//const media = zone.GetSelectedMedia();
					const $mediaScheduled = document.getElementById('media-scheduled');
					PopulateMediaProperty(zone, 'scheduled', false);
					PopulateMediaProperty(zone, 'utc', false);
					PopulateMediaProperty(zone, 'scheduledaterange', false);
					PopulateMediaProperty(zone, 'schedulestartdate', '');
					PopulateMediaProperty(zone, 'schedulestarttime', '');
					PopulateMediaProperty(zone, 'scheduleenddate', '');
					PopulateMediaProperty(zone, 'scheduleendtime', '');
					for (const element of ['media-utc', 'media-scheduledaterange', 'media-scheduledayparting']) {
						toggleRow(document.getElementById(element), $mediaScheduled.checked && !$mediaScheduled.indeterminate, setHidden);
					}
					for (const element of ['media-schedulestartdate', 'media-schedulestarttime', 'media-scheduleenddate', 'media-scheduleendtime']) {
						toggleRow(document.getElementById(element), $mediaScheduled.checked && !$mediaScheduled.indeterminate && document.getElementById('media-scheduledaterange').checked, setHidden);
					}
					PopulateMediaProperty(zone, 'scheduledayparting', false);
					for (const element of ['media-scheduledow', 'media-daypartstarttime', 'media-daypartendtime']) {
						toggleRow(document.getElementById(element), $mediaScheduled.checked && !$mediaScheduled.indeterminate && document.getElementById('media-scheduledayparting').checked && !document.getElementById('media-scheduledayparting').indeterminate, setHidden);
					}
					PopulateMediaProperty(zone, 'daypartstarttime', '');
					PopulateMediaProperty(zone, 'daypartendtime', '');
					PopulateMediaProperty(zone, 'sunday', false);
					PopulateMediaProperty(zone, 'monday', false);
					PopulateMediaProperty(zone, 'tuesday', false);
					PopulateMediaProperty(zone, 'wednesday', false);
					PopulateMediaProperty(zone, 'thursday', false);
					PopulateMediaProperty(zone, 'friday', false);
					PopulateMediaProperty(zone, 'saturday', false);
				};
				
				switch (zone.GetSelectedMediaType()) {
					case MEDIATYPE_IMAGE:
						setHidden(document.getElementById('media-header-mediaoptions'), false);
						PopulateMediaProperty(zone, 'fillstyleid', FILLSTYLE_NONE);
						populateMediaLayout();
						populateMediaStartRunlength();
						populateMediaTransitions();
						populateMediaSchedule();
						PopulateMediaProperty(zone, 'zoompan', 0);
						break;
					case MEDIATYPE_VIDEO:
						setHidden(document.getElementById('media-header-mediaoptions'), false);
						PopulateMediaProperty(zone, 'fillstyleid', FILLSTYLE_NONE);
						populateMediaLayout();
						populateMediaStartRunlength();
						populateMediaTransitions();
						populateMediaSchedule();
						populateMediaVolume();
						populateMediaDucking();
						break;
					case MEDIATYPE_AUDIO:
						populateMediaStartRunlength();
						populateMediaVolume();
						populateMediaDucking();
						populateMediaSchedule();
						break;
					case MEDIATYPE_WIDGET:
						var selectedWidgetType = zone.GetSelectedWidgetType();
						if (gCurrentWidgetComponent == null) {
							populateMediaStartRunlength();
							populateMediaSchedule();
							if (selectedWidgetType != 'blank') {
								populateMediaLayout();
								populateMediaTransitions();
							}
						}
						
						if (gCurrentWidget && gCurrentWidget.widgettype == 'buttongroup' && gCurrentWidgetComponent) {
							setHidden(document.getElementById('media-header-buttonlayout'), false);
							setHidden(document.getElementById('media-header-buttonoptions'), false);
							if (permissions['Metadata']) {
								setHidden(document.getElementById('media-header-buttonmetadata'), false);
							}
							PopulateButtonLayoutProperties(zone);
							PopulateButtonMetadataProperties(zone);
						} else if (gCurrentWidget && gCurrentWidget.widgettype == 'text2') {
							// No media properties for text2
						} else if (selectedWidgetType != 'blank') {
							setHidden(document.getElementById('media-header-widgetoptions'), false);
							setHidden(document.getElementById('media-header-widgetnooptions'), false);
						}
						if (selectedWidgetType) {
							var widget = WidgetInformation.find(widget => widget.name == selectedWidgetType);
							for (const attribute of widget.attributes) {
								if (attribute.hasOwnProperty('$header')) {
									setHidden(attribute.$header, false);
								}
								if (attribute.hasOwnProperty('view') && attribute.view == 'property') {
									PopulateWidgetProperty(zone, attribute);
								}
							}
						}
						break;
					case MEDIATYPE_MISC:
						populateMediaLayout();
						populateMediaStartRunlength();
						populateMediaTransitions();
						populateMediaSchedule();
						break;
					default:
						if (zone.HasLayout()) {
							populateMediaLayout();
						}
						if (zone.HasRuntime()) {
							populateMediaStartRunlength();
						}
						if (zone.HasTransitions()) {
							populateMediaTransitions();
						}
						if (zone.HasVolume()) {
							populateMediaVolume();
						}
						if (zone.HasDucking()) {
							populateMediaDucking();
						}
						populateMediaSchedule();
						break;
				}
			}
		}
		
		for (const $collapsedHeader of $collapsedHeaders) { $collapsedHeader.classList.add('collapsed'); }
	} else if (metadataView) {
		setHidden(document.getElementById('metadataHeader'), false);
		document.getElementById('metadataHeader').classList.remove('collapsed');
		const selectedNodes = await $contentTree.getSelected();
		// Handle multiselect as well as single select
		if (selectedNodes.length > 1) {
			
		} else if (selectedNodes.length == 1) {
			const selectedNode = selectedNodes[0];
			document.getElementById('contentTree-metadata').classList.add('busy');
			const data = await get('/get/treemetadata.php',
			{
				tree: 'content',
				id: selectedNode.id
			});
			document.getElementById('contentTree-metadata').classList.remove('busy');
			if (data.success) {
				Metadata(document.getElementById('contentTree-metadata'), data.metadata, data.parentmetadata, async metadataToSave => {
					try {
						const postData = await post('/post/treemetadata.php',
						{
							tree: 'content',
							id: selectedNode.id,
							metadata: JSON.stringify(metadataToSave)
						});
						if (!postData.success) {
							toastError(postData.error);
						}
					} catch (exception) {
						NetworkingError();
					}
				}, selectedNode);
			}
		}
	}
	// Toggle all the recorded hidden changes
	for (const $e of hiddenChanges) $e.hidden = !$e.hidden;
}

async function RequiresSaveCheck() {
	if (gSaveFlag) {
		switch (await ConfirmationBox('Save playlist?', 'Do you wish to save your current playlist before you open a new playlist?', 'Yes', 'No', 'Cancel')) {
			case 0:
				await SavePlaylist();
				clearPlaylist();
				break;
			case 1:
				clearPlaylist();
				$buttonTab.hidden = true;
				document.getElementById('mediatab').click();
				break;
		}
	}
}

async function newPlaylistPressed() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	stopPreviewPressed();
	createThumbnail();
	if (!gSaveFlag) {
		var templates = [];
		const fixScreenOptions = () => {
			let width = parseInt($width.value, 10);
			let height = parseInt($height.value, 10);
			
			if ($landscape.checked && width < height) {
				$portrait.checked = true;
			} else if ($portrait.checked && width > height) {
				$landscape.checked = true;
			}
			
			if ($landscape.checked && Math.abs(width / height - 16 / 9) < 0.1 || $portrait.checked && $16x9.checked && Math.abs(width / height - 9 / 16) < 0.1) {
				$16x9.checked = true;
			} else if ($landscape.checked && Math.abs(width / height - 4 / 3) < 0.1 || $portrait.checked && $4x3.checked && Math.abs(width / height - 3 / 4) < 0.1) {
				$4x3.checked = true;
			} else {
				$customRatio.checked = true;
			}
		};
		
		const resizeAllTemplates = () => {
			fixScreenOptions();
			for (const $template of document.getElementsByClassName('template')) {
				const template = $template.data;
				var canvasWidth = Math.max(32, parseInt($width.value, 10));
				var canvasHeight = Math.max(32, parseInt($height.value, 10));
				var scale = 1;
				if (Math.max(canvasWidth, canvasHeight) > 170) {
					scale = 170 / Math.max(canvasWidth, canvasHeight);
				}
				var isRealTemplate = /^Template (1?[1-9]|10|20|21)$/.test(template.name);
				if (isRealTemplate) {
					const canvas = $template.querySelector(':scope > canvas');
					canvas.width = canvasWidth * scale;
					canvas.height = canvasHeight * scale;
					Object.assign(canvas.style, {
						width: `${canvasWidth * scale}px`,
						height: `${canvasHeight * scale}px`
					});
					var context = canvas.getContext('2d');
					context.save();
					const colors = [
						getComputedStyle(document.documentElement).getPropertyValue('--new-playlist-color1'),
						getComputedStyle(document.documentElement).getPropertyValue('--new-playlist-color2'),
						getComputedStyle(document.documentElement).getPropertyValue('--new-playlist-color3')
					];
					template.zones.forEach((zone, zoneItr) => {
						context.strokeStyle = 'black';
						context.fillStyle = colors[zoneItr % 3];
						const x = Math.round(parseInt(zone.x, 10) / parseInt(template.pagewidth, 10) * canvas.width);
						const y = Math.round(parseInt(zone.y, 10) / parseInt(template.pageheight, 10) * canvas.height);
						const width = zone.width / parseInt(template.pagewidth, 10) * canvas.width;
						const height = zone.height / parseInt(template.pageheight, 10) * canvas.height;
						context.fillRect(x, y, width, height);
						context.fillStyle = 'black';
						context.textBaseline = 'middle';
						context.textAlign = 'center';
						context.font = `12px ${getComputedStyle(document.documentElement).getPropertyValue('--font')}`;
						context.fillText(String.fromCharCode(65 + zoneItr), x + width / 2, y + height / 2);
					});
					context.strokeStyle = 'black';
					context.translate(0.5, 0.5);
					context.strokeRect(0, 0, canvas.width - 1, canvas.height - 1);
					context.restore();
				} else {
					Object.assign($template.querySelector(':scope > img').style, {
						width: `${canvasWidth * scale}px`,
						height: `${canvasHeight * scale}px`
					});
				}
			}
		};
		
		const onScreenOptionChange = () => {
			let width = parseInt($width.value, 10);
			let height = parseInt($height.value, 10);
			
			// Swap the width and height if they don't match landscape or portrait expectations
			if ($landscape.checked && width < height || $portrait.checked && width > height) {
				[width, height, $width.value, $height.value] = [height, width, $height.value, $width.value];
			}
			
			if ($16x9.checked) {
				$height.value = $landscape.checked ? Math.round(width / (16 / 9)) : Math.round(width * (16 / 9));
			} else if ($4x3.checked) {
				$height.value = $landscape.checked ? Math.round(width / (4 / 3)) : Math.round(width * (4 / 3));
			} else {
				// Custom ratio
				// Do nothing
			}
			resizeAllTemplates();
		}
		
		const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: {
			width: '815px',
			height: '80%'
		} },
			$.header({ text: 'New Playlist' }),
			$.div({ class: 'formgrid' },
				$.label({ for: 'aspectRatio', text: 'Aspect Ratio' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'radio', id: 'newplaylist-16x9', name: 'aspectRatio', value: '16x9', checked: true, change: onScreenOptionChange }),
					$.label({ for: 'newplaylist-16x9', text: '16x9', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'newplaylist-4x3', name: 'aspectRatio', value: '4x3', change: onScreenOptionChange }),
					$.label({ for: 'newplaylist-4x3', text: '4x3', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'newplaylist-customratio', name: 'aspectRatio', value: 'custom', change: onScreenOptionChange }),
					$.label({ for: 'newplaylist-customratio', text: 'Custom' })
				),
				$.label({ for: 'orientation', text: 'Orientation' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'radio', id: 'newplaylist-landscape', name: 'orientation', value: 'landscape', checked: true, change: onScreenOptionChange }),
					$.label({ for: 'newplaylist-landscape', text: 'Landscape', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'newplaylist-portrait', name: 'orientation', value: 'portrait', change: onScreenOptionChange }),
					$.label({ for: 'newplaylist-portrait', text: 'Portrait' }),
				),
				$.label({ text: 'Screen Units' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'text', id: 'newplaylist-width', size: 4, maxlength: 6, class: 'sampleInput', value: configuration.defaultplaylistwidth, input: resizeAllTemplates, change: resizeAllTemplates }),
					$.span({ text: 'x', css: { margin: '0 5px' } }),
					$.input({ type: 'text', id: 'newplaylist-height', size: 4, maxlength: 6, class: 'sampleInput', value: configuration.defaultplaylistheight, input: resizeAllTemplates, change: resizeAllTemplates })
				),
				$.label({ for: 'templateSelect', text: 'Select a Layout or Custom Template (Optional):', css: { gridColumn: '1 / 3' } })
			),
			$.div({ id: 'template-tabs', tabs: {}, css: {
				flex: '1 0 0px',
				minHeight: '0px'
			} },
				$.ul({},
					$.li({ text: 'Layouts' }),
					$.li({ id: 'template-tabs-customtemplates', text: 'Custom Templates' })
				),
				$.div({ id: 'templates', class: 'templates busy' }),
				$.div({ id: 'custom-playlists' },
					$.ul({ class: 'iconlist' },
						$.li({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: async () => {
							const $template = document.querySelector('.template.selected');
							if ($template == null) {
								toastError('Select a template first.');
							} else if (parseInt($template.data.id, 10) <= 21) {
								toastError('Select a user created template.');
							} else {
								const selectedTemplate = $template.data;
								if (await QuestionBox('Delete template?', 'Are you sure you want to delete the selected template?')) {
									const $templates = document.getElementById('customtemplates');
									$templates.classList.add('busy');
									try {
										const data = await post('/post/deleteplaylist.php', {
											playlistid: selectedTemplate.id
										});
										CheckSessionExpired(data);
										if (data.success) {
											templates = templates.filter(template => template.id != selectedTemplate.id);
											$template.remove();
										} else {
											toastError(`There was a problem deleting the template. ${data.error}`);
										}
									} catch (exception) {
										NetworkingError();
									} finally {
										$templates.classList.remove('busy');
									}
								}
							}
						} })
					),
					$.div({ id: 'customtemplates', class: 'templates busy' })
				)
			),
			$.div({ class: 'panel-buttons' },
				$.button({ text: 'Cancel', click: function() {
					$dialog.remove();
				} }),
				$.button({ text: 'OK', click: function() {
					const $template = document.querySelector('.template.selected');
					createPlaylist($template?.data);
					$dialog.remove();
				} })
			)
		));
		centerOnScreen($dialog);
		
		const $width = document.getElementById('newplaylist-width');
		const $height = document.getElementById('newplaylist-height');
		const $16x9 = document.getElementById('newplaylist-16x9');
		const $4x3 = document.getElementById('newplaylist-4x3');
		const $customRatio = document.getElementById('newplaylist-customratio');
		const $landscape = document.getElementById('newplaylist-landscape');
		const $portrait = document.getElementById('newplaylist-portrait');
		
		try {
			const data = await get('/get/templatedialogdata.php');
			document.getElementById('templates').classList.remove('busy');
			document.getElementById('customtemplates').classList.remove('busy');
			CheckSessionExpired(data);
			if (data.success) {
				templates = data.data;
				templates.sort((a, b) => {
					var aTemplate = /^Template (1?[1-9]|10|20|21)$/.test(a.name);
					var bTemplate = /^Template (1?[1-9]|10|20|21)$/.test(b.name);
					if (aTemplate && !bTemplate) {
						return -1;
					} else if (!aTemplate && bTemplate) {
						return 1;
					} else if (aTemplate && bTemplate) {
						var aValue = parseInt(/([0-9]+)/.exec(a.name)[1], 10);
						var bValue = parseInt(/([0-9]+)/.exec(b.name)[1], 10);
						return aValue - bValue;
					} else {
						return a.name.localeCompare(b.name)
					}
				});
				templates.forEach((template, templateItr) => {
					var isRealTemplate = /^Template (1?[1-9]|10|20|21)$/.test(template.name);
					var loadImageAttempt = 0;
					const $template = document.getElementById(isRealTemplate ? 'templates' : 'customtemplates').appendChild($.div({ id: `template-${template.id}`, class: 'template', click: async function() {
						if (this.classList.contains('selected')) {
							this.classList.remove('selected');
						} else {
							if (!isRealTemplate && ($width.value != template.pagewidth || $height.value != template.pageheight)) {
								if (await QuestionBox('Template width and height defined', `This template has a defined width and height. Do you want to use the template's width and height?`)) {
									$width.value = template.pagewidth;
									$height.value = template.pageheight;
									resizeAllTemplates();
									document.getElementById(`template-${template.id}`).classList.add('selected');
								}
							}
							document.querySelector('.template.selected')?.classList?.remove('selected');
							this.classList.add('selected');
						}
					}, dblclick: function() {
						for (const $e of document.getElementsByClassName('template')) $e.classList.remove('selected');
						this.classList.add('selected');
						createPlaylist(this.data);
						$dialog.remove();
					} },
						isRealTemplate ? $.canvas({ class: 'openTemplateImage' }) : $.img({ src: `thumbnails/${account}/${template.id}.jpg?random=${Math.floor(Math.random() * 100000)}`, class: 'openTemplateImage', load: function() {
							this.parentNode.classList.remove('busy');
						}, error: function() {
							loadImageAttempt++;
							this.parentNode.classList.add('busy');
							if (loadImageAttempt < 3) {
								setTimeout(() => {
									this.setAttribute('src', `thumbnails/${account}/${template.id}.jpg?random=${Math.floor(Math.random() * 100000)}`);
								}, 30 * 1000);
							}
						} }),
						$.div({ text: isRealTemplate ? 'Layout ' + (templateItr + 1) : template.name })
					));
					$template.data = template;
				});
				resizeAllTemplates();
			}
		} catch (exception) {
			NetworkingError();
		}
	} else {
		try {
			await RequiresSaveCheck();
			newPlaylistPressed();
		} catch (exception) {
			// User cancelled save
		}
	}
}

async function openPlaylistPressed() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	stopPreviewPressed();
	if (!gSaveFlag) {
		createThumbnail();
		const searchInputChange = function() {
			this.previousElementSibling.hidden = this.value != '';
			this.previousElementSibling.previousElementSibling.hidden = this.value == '';
			if (this.value == '') {
				document.getElementById('playlists').removeFilterView();
			} else {
				document.getElementById('playlists').filterView(node => {
					if (node.group) {
						return false;
					}
					return node.name.toLowerCase().indexOf(this.value.toLowerCase()) != -1;
				});
			}
		};
		const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: {
			width: '815px',
			height: '80%'
		} },
			$.header({ text: 'Open Playlist' }),
			$.div({ class: 'iconlist' },
				$.i({ title: 'Remove filter', class: 'tooltip-button mdi mdi-close-circle-outline removefilter', hidden: true, click: function() {
					this.hidden = true;
					this.nextElementSibling.hidden = false;
					this.nextElementSibling.nextElementSibling.value = '';
					this.nextElementSibling.nextElementSibling.dispatchEvent(new Event('change'));
					document.getElementById('playlists').removeFilterView();
				} }),
				$.label({ for: 'searchPlaylistFilter', title: 'Search', class: 'tooltip-button mdi mdi-magnify addfilter' }),
				$.input({ type: 'text', id: 'searchPlaylistFilter', placeholder: 'Search', class: 'sampleInput', input: searchInputChange, change: searchInputChange }),
				$.i({ title: 'Add group', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => {
					document.getElementById('playlists').addGroup();
				} }),
				$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => {
					document.getElementById('playlists').deleteSelected();
				} })
			),
			$.uitree({ id: 'playlists', config: {
				tree: 'playlist',
				thumbnail: PlaylistTreeThumbnail,
				overrideThumbnailDimensions: PlaylistTreeThumbnailThumbnailDimensions,
				editable: node => node.group && node.parentNode != null,
				showSmallLargeThumbnail: true,
				deleteConfirmation: async function(selectedNodes, selectedAllChildren, deleteNodes) {
					if (selectedNodes.length == 0) {
						toastError('Select a group or playlist to delete.');
						return;
					}
					var selectedNode = selectedNodes[0];
					if (selectedNode.group) {
						var playlistCount = selectedAllChildren.filter(node => !node.group).length;
						if (playlistCount != 0) {
							await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${playlistCount} playlist${playlistCount == 1 ? '' : 's'}?`, 'Yes', 'No');
						}
					} else {
						await ConfirmationBox("Remove playlist?", `Are you sure you want to remove the playlist '${selectedNode.name}'?`, 'Yes', 'No');
					}
						
					await deleteNodes(selectedAllChildren);
						
					if (gCurrentPlaylist && selectedAllChildren.some(playlist => gCurrentPlaylist.id == playlist.playlistid)) {
						gCurrentPlaylist.id = -1;
						SetSaveFlag(true);
					}
				},
				doubleClick: async function(e, node) {
					if (!node.group) {
						await document.getElementById('playlists').setSelected(playlist => playlist.playlistid == node.playlistid);
						document.getElementById('openOk').click();
						return true;
					}
					return false;
				}
			} }),
			$.div({ class: 'panel-buttons' },
				$.button({ text: 'Cancel', click: function() {
					$dialog.remove();
				} }),
				$.button({ id: 'openOk', text: 'Open', click: async function() {
					if (gSaveFlag) {
						try {
							await RequiresSaveCheck();
						} catch (exception) {
							return;
						}
					}
					var selected = (await document.getElementById('playlists').getSelected()).filter(node => !node.group);
					if (selected.length == 1) {
						LoadPlaylistById(selected[0].playlistid);
						$dialog.remove();
					}
				} })
			)
		));
		centerOnScreen($dialog);
		
		stopPreviewPressed();
	} else {
		try {
			await RequiresSaveCheck();
			openPlaylistPressed();
		} catch (exception) {
			// User cancelled the save
		}
	}
}

function getPlaylistJSON() {
	if (gCurrentPlaylist == null) {
		return;
	}
	gCurrentPlaylist.runlength = gCurrentPlaylist.zones.reduce((maxRunLength, zone) => Math.max(maxRunLength, zone.media.reduce((runlength, media) => runlength + media.runlength, 0)), 0);
	var copyProperties = (from, to) => {
		for (var property in from) {
			if (typeof from[property] == 'boolean' || typeof from[property] == 'string' || typeof from[property] == 'number' || from[property] == null) {
				to[property] = from[property];
			}
		}
	};
	var playlistJSON = {};
	copyProperties(gCurrentPlaylist, playlistJSON);
	playlistJSON.zones = [];
	for (const zone of gCurrentPlaylist.zones) {
		var zoneJSON = {};
		zoneJSON.media = [];
		copyProperties(zone, zoneJSON);
		for (const media of zone.media) {
			var mediaJSON = {};
			copyProperties(media, mediaJSON);
			if (media.type == MEDIATYPE_WIDGET) {
				mediaJSON.attributes = JSON.parse(JSON.stringify(media.attributes));
			}
			zoneJSON.media.push(mediaJSON);
		}
		playlistJSON.zones.push(zoneJSON);
	}
	return JSON.stringify(playlistJSON);
}

async function ShowSaveDialog(header, saveType) {
	return new Promise(async (resolve, reject) => {
		if (gCurrentPlaylist == null) {
			reject();
			return;
		}
		stopPreviewPressed();
		
		if (saveType == 'save' && gCurrentPlaylist.name != 'Untitled' && gCurrentPlaylist.id != -1) {
			// Already has a name
			await SavePlaylistOrTemplate(null, gCurrentPlaylist.name, saveType);
			resolve();
		} else {
			if (saveType == 'save') {
				saveType = 'saveas';
			}
			let $name;
			let $playlists;
			let $ok;
			const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: saveType != 'saveastemplate' ? {
				width: '815px',
				height: '80%'
			} : {
				width: '350px'
			} },
				$.header({ text: header }),
				$.div({ class: 'formgrid' },
					$.label({ text: 'Name' }),
					$name = $.input({ type: 'text', size: 30, maxlength: 80, class: 'sampleInput name', value: gCurrentPlaylist.name == 'Untitled' ? '' : gCurrentPlaylist.name, keydown: e => {
						if (e.key == 'Enter') {
							$ok.click();
						}
					} })
				),
				$.ul({ class: 'iconlist' },
					$.li({ title: 'Add group', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => document.getElementById('playlists').addGroup() }),
					$.li({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => document.getElementById('playlists').deleteSelected() })
				),
				$playlists = $.uitree({ id: 'playlists', config: {
					tree: 'playlist',
					thumbnail: PlaylistTreeThumbnail,
					overrideThumbnailDimensions: PlaylistTreeThumbnailThumbnailDimensions,
					editable: node => node.group && node.parentNode != null,
					showSmallLargeThumbnail: true,
					deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
						if (selectedNodes.length == 0) {
							toastError('Select a group or playlist to delete.');
							return;
						}
						var selectedNode = selectedNodes[0];
						if (selectedNode.group) {
							var playlistCount = selectedAllChildren.filter(node => !node.group).length;
							if (playlistCount != 0) {
								await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${playlistCount} playlist${playlistCount == 1 ? '' : 's'}?`, 'Yes', 'No');
							}
						} else {
							await ConfirmationBox("Remove playlist?", `Are you sure you want to remove the playlist '${selectedNode.name}'?`, 'Yes', 'No');
						}
						await deleteNodes(selectedAllChildren);
					}
				} }),
				$.div({ class: 'panel-buttons' },
					$.button({ text: 'Cancel', click: function() {
						$dialog.remove();
						reject();
					} }),
					$ok = $.button({ text: 'OK', click: async function() {
						if (await SavePlaylistOrTemplate(saveType == 'saveastemplate' || (await $playlists.getSelected())[0], $name.value, saveType)) {
							$dialog.remove();
						}
						resolve();
					} })
				)
			));
			centerOnScreen($dialog);
			for (const $e of $dialog.querySelectorAll('.iconlist, #playlists')) $e.hidden = saveType == 'saveastemplate';
			$name.focus();
		}
	});
}

async function SavePlaylistOrTemplate(parent, name, saveType) {
	var save = async (type, template, playlistJSON) => {
		LockBody();
		try {
			var data = await post('/post/saveplaylist.php',
			{
				savetype: type,
				parent: parent == null ? null : parent.id,
				template: template,
				playlist: playlistJSON,
				preview: false
			});
			UnlockBody();
			CheckSessionExpired(data);
			if (data.success) {
				return data;
			} else {
				if (data.hasOwnProperty('nameexists')) {
					// Rare edge case where the user deletes a playlist with it still open in the editor and someone else saves a new playlist with the same name
					gCurrentPlaylist.name = 'Untitled';
					setWindowHeader();
				}
				toastError(`The playlist or template could not be saved. ${data.error}`);
			}
		} catch (exception) {
			UnlockBody();
			NetworkingError();
			throw exception;
		}
	};
	
	try {
		if (saveType == 'saveas' || saveType == 'saveastemplate') {
			if (name.length == 0) {
				toastError('Please enter a name.');
				return false;
			} else if (name == 'Untitled' || !IsAlphaNumericSpacesHyphensUnderscoresApostrophesNotEmptyLessThan60Characters(name)) {
				toastError('The name can only contain letters, numbers, spaces, hyphens, underscores, or apostrophes and must be between 1 and 60 characters.');
				return false;
			} else if (saveType == 'saveas' && (parent == null || !parent.group)) {
				toastError('Please select a folder.');
				return false;
			} else {
				const exists = await checkPlaylistNameExists(name);
				if (exists) {
					toastError('A playlist or template with that name already exists.');
					return false;
				}
				if (saveType == 'saveas') {
					var oldName = gCurrentPlaylist.name;
					gCurrentPlaylist.name = name;
					var playlistJSON = getPlaylistJSON();
					gCurrentPlaylist.name = oldName;
					var data = await save(1, false, playlistJSON);
					data = await LoadPlaylistById(data.playlistid);
					if (data.success) {
						SetSaveFlag(false);
					}
				} else if (saveType == 'saveastemplate') {
					var playlistJSON = JSON.parse(getPlaylistJSON());
					playlistJSON.name = name;
					playlistJSON = JSON.stringify(playlistJSON);
					var data = await save(1, true, playlistJSON);
					toastSuccess('The template has been saved successfully.');
				}
				createThumbnail(); // Do not wait for this to finish
			}
		} else if (saveType == 'save') {
			var playlistJSON = getPlaylistJSON();
			var data = await save(0, false, playlistJSON);
			data = await LoadPlaylistById(data.playlistid);
			if (data.success) {
				await new Promise(resolve => setTimeout(resolve, 1000));
				createThumbnail(); // Do not wait for this to finish
				SetSaveFlag(false);
			}
		}
	} catch (exception) {
		return false;
	}
	return true;
}

async function SavePlaylist(fromSavePlaylist = false) {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	if (!fromSavePlaylist && document.getElementById('editor-saveonpublish').checked) {
		await UpdateScheduledPlayers(true);
	} else {
		await ShowSaveDialog('Save', 'save');
	}
}

async function SaveAsPlaylist() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	await ShowSaveDialog('Save As', 'saveas');
}

async function SaveAsTemplate() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	await ShowSaveDialog('Save As Template', 'saveastemplate');
}

function exportPlaylistPressed() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	
	var updatePercentageInterval = null;
	
	var updatePercentage = () => {
		updatePercentageInterval = setInterval(async () => {
			try {
				var data = await get('/get/exportedplaylistspercentage.php');
				if (data.success) {
					document.getElementById('export-percentage').innerText = `Exporting... ${data.percentage} Done`;
					if (data.percentage == '100%') {
						clearInterval(updatePercentageInterval);
					}
				} else {
					clearInterval(updatePercentageInterval);
				}
			} catch (exception) {
				clearInterval(updatePercentageInterval);
			}
		}, 1000);
	};
	
	const inputChange = function() {
		const $exportPlaylists = document.getElementById('export-playlists');
		this.previousElementSibling.hidden = this.value != '';
		this.previousElementSibling.previousElementSibling.hidden = this.value == '';
		if (this.value == '') {
			$exportPlaylists.removeFilterView();
		} else {
			$exportPlaylists.filterView(node => {
				if (node.group) {
					return false;
				}
				return node.name.toLowerCase().indexOf(this.value.toLowerCase()) != -1;
			});
		}
	};
	
	const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: {
		width: '815px',
		height: '80%'
	} },
		$.header({ text: 'Export Playlist(s)' }),
		$.div({ class: 'iconlist' },
			$.i({ title: 'Remove filter', class: 'tooltip-button mdi mdi-close-circle-outline removefilter', hidden: true, click: function() {
				this.hidden = true;
				this.nextElementSibling.hidden = false;
				this.nextElementSibling.nextElementSibling.querySelector('input').value = '';
				this.nextElementSibling.nextElementSibling.dispatchEvent(new Event('change'));
				document.getElementById('export-playlists').removeFilterView();
			} }),
			$.label({ for: 'searchExportPlaylistFilter', title: 'Search', class: 'tooltip-button mdi mdi-magnify addfilter' }),
			$.input({ type: 'text', id: 'searchExportPlaylistFilter', placeholder: 'Search', class: 'sampleInput', input: inputChange, change: inputChange }),
			$.i({ title: 'Add group', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => document.getElementById('export-playlists').addGroup() }),
			$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => document.getElementById('export-playlists').deleteSelected() })
		),
		$.uitree({ id: 'export-playlists', config: {
			tree: 'playlist',
			thumbnail: PlaylistTreeThumbnail,
			overrideThumbnailDimensions: PlaylistTreeThumbnailThumbnailDimensions,
			checkbox: true,
			editable: node => node.group && node.parentNode != null,
			showSmallLargeThumbnail: true,
			deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
				if (selectedNodes.length == 0) {
					toastError('Select a group or playlist to delete.');
					return;
				}
				const selectedNode = selectedNodes[0];
				if (selectedNode.group) {
					var playlistCount = selectedAllChildren.filter(node => !node.group).length;
					if (playlistCount != 0) {
						await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${playlistCount} playlist${playlistCount == 1 ? '' : 's'}?`, 'Yes', 'No');
					}
				} else {
					await ConfirmationBox("Remove playlist?", `Are you sure you want to remove the playlist '${selectedNode.name}'?`, 'Yes', 'No');
				}
				await deleteNodes(selectedAllChildren);
			}
		} }),
		$.div({ class: 'panel-buttons' },
			$.span({ id: 'export-percentage' }),
			$.button({ text: 'Close', click: function() {
				clearInterval(updatePercentageInterval);
				$dialog.remove();
			} }),
			$.button({ text: 'Clean Up', title: 'Clean up any exported playlist archive on the server. The file is not automatically removed after downloading. Only one archive file can exist at a time on the server, but these can be large so cleaning them up after downloading is advised.', click: async () => {
				LockBody();
				try {
					const data = await post('/post/deleteexport.php');
					UnlockBody();
					CheckSessionExpired(data);
					if (data.success) {
						document.getElementById('export-percentage').innerText = '';
						toastSuccess('Export file removed.');
					} else {
						toastError(`Clean Export File: ${data.error}`);
					}
				} catch (exception) {
					UnlockBody();
					NetworkingError();
				}
			} }),
			$.button({ text: 'Download', class: 'primary', click: async () => {
				LockBody();
				try {
					var data = await get('/get/exportedplaylists.php');
					UnlockBody();
					CheckSessionExpired(data);
					if (data.success) {
						window.location.assign(window.location.protocol + '//' + window.location.host + '/exportedplaylists/' + account + '/' + data.filename);
					} else {
						toastError(`Exported Playlist: ${data.error}`);
					}
				} catch (exception) {
					UnlockBody();
					NetworkingError();
				}
			} }),
			$.button({ text: 'Export', click: async () => {
				const checked = (await document.getElementById('export-playlists').getCheckedAllChildren()).filter(node => !node.group);
				if (checked.length == 0) {
					toastError('Check at least one playlist to export.');
				} else {
					await InputBox('ZIP name', 'Enter a name for the ZIP archive that holds the playlist(s).', 'Ok', 'Cancel', '', async name => {
						try {
							var data = await post('/post/exportplaylists.php',
							{
								name: name,
								ids: JSON.stringify(checked.map(playlist => playlist.playlistid))
							});
							CheckSessionExpired(data);
							if (data.success) {
								toastSuccess('Export has begun. Clicking the download exported playlist button will tell you if it is ready.');
								document.getElementById('export-percentage').innerText = 'Exporting... 0% Done';
								updatePercentage();
								return true;
							} else {
								toastError(`Exporting Busy: ${data.error}`);
							}
						} catch (exception) {
							NetworkingError();
						}
					}, () => true, null);
				}
			} })
		)
	));
	centerOnScreen($dialog);
	
	// Check on previously created exports
	updatePercentage();
}

function importPlaylistPressed() {
	ShowFileDialog();
}

async function playPreviewPressed() {
	if (gCurrentPlaylist == null) {
		return;
	}
	LockBody();
	ExitWidgetMode();
	var playlistJSON = getPlaylistJSON();
	try {
		var data = await post('/post/saveplaylist.php',
		{
			savetype: 1,
			parent: null,
			template: false,
			playlist: playlistJSON,
			preview: true
		});
		CheckSessionExpired(data);
		if (data.success) {
			let url = 'player.php?preview=true&account=' + account + '&playlistid=' + data.playlistid + '&starttime=' + gTimeBarSelection;
			if (document.getElementById('changeCursor').classList.contains('mdi-cursor-default')) {
				url = url + '&cursor=default';
			}
			for (const $e of document.getElementsByClassName('timelineItem')) { $e.style.opacity = 1; }
			for (const $e of document.querySelectorAll('.timelineName > .hide-button')) { $e.hidden = true; }
			$playStop.removeEventListener('click', playPreviewPressed);
			$playStop.addEventListener('click', stopPreviewPressed);
			$playStop.classList.remove('mdi-play');
			$playStop.classList.add('mdi-stop');
			$playStop.title = 'Stop preview';
			const $previewIframe = document.getElementById('previewIframe');
			if ($previewIframe.hidden) {
				$editorCanvasShadow.hidden = true;
				document.getElementById('changeCursor').hidden = false;
				$previewIframe.hidden = false;
				$previewIframe.setAttribute('src', url);
				$previewIframe.style.width = `${gCurrentPlaylist.pagewidth}px`;
				$previewIframe.style.height = `${gCurrentPlaylist.pageheight}px`;
			}
			UnlockBody();
		} else {
			UnlockBody();
			toastError(`A preview could not be generated for the playlist. ${data.error}`);
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
	}
}

function stopPreviewPressed() {
	const $previewIframe = document.getElementById('previewIframe');
	if (!$previewIframe.hidden) {
		$playStop.removeEventListener('click', stopPreviewPressed);
		$playStop.addEventListener('click', playPreviewPressed);
		$playStop.classList.remove('mdi-stop');
		$playStop.classList.add('mdi-play');
		$playStop.title = 'Play preview';
		for (const $timelineName of document.getElementsByClassName('timelineName')) {
			$timelineName.hidden = false;
			$timelineName.querySelector('.hide-button').hidden = false;
		}
		for (const $timelineBar of document.getElementsByClassName('timelineBar')) {
			$timelineBar.hidden = false;
			$timelineBar.querySelector('.marker')?.remove();
		}
		for (const $timelineItem of document.getElementsByClassName('timelineItem')) $timelineItem.hidden = false;
		document.getElementById('changeCursor').hidden = true;
		$editorCanvasShadow.hidden = false;
		$previewIframe.hidden = true;
		$previewIframe.setAttribute('src', '');
	}
}

function clearPlaylist() {
	if (gCurrentPlaylist) {
		for (const zone of gCurrentPlaylist.zones) {
			for (const media of zone.media) {
				media.Remove();
			}
			zone.Remove();
		}
	}
	gCurrentPlaylist = null;
	gSelectedZone = null;
	setWindowHeader();
	clearEditorCanvas();
	LineLayerClear();
	$editorCanvasShadow.hidden = true;
	document.getElementById('changeCursor').hidden = true;
	document.getElementById('editor-canvastip').hidden = false;
	UpdateProperties('zone');
	gSelectedZone = null;
	gExpand = false;
	ExitWidgetMode();
	document.getElementById('interactivebuttons').innerHTML = '';
	SetSaveFlag(false);
	resizeCanvas();
}

async function checkPlaylistNameExists(playlistName) {
	LockBody();
	try {
		var data = await get('/get/checkplaylistnameexists.php',
		{
			name: playlistName
		});
		UnlockBody();
		CheckSessionExpired(data);
		if (data.success) {
			return data.exists;
		}
	} catch (exception) {
		UnlockBody();
		NetworkingError();
		throw exception;
	}
}

function createPlaylist(template = null) {
	delete localStorage.lastplaylistid;
	delete localStorage.zonestate;
	delete localStorage.zoneresize;
	const $width = document.getElementById('newplaylist-width');
	const $height = document.getElementById('newplaylist-height');
	var playlistWidth = parseInt($width.value, 10);
	var playlistHeight = parseInt($height.value, 10);
	if (isNaN(playlistWidth) || isNaN(playlistHeight)) {
		toastError('Invalid width or height entered.');
		$width.value = configuration.defaultplaylistwidth;
		$height.value = configuration.defaultplaylistheight;
		return;
	}
	
	gTimeBarSelection = 0;
	UpdateTimeBar();
	clearPlaylist();
	if (template != null) {
		populatePlaylist(template, true, playlistWidth, playlistHeight);
		gTimeBarSelection = 0;
		UpdateTimeBar();
		SetSaveFlag(true);
	} else {
		gCurrentPlaylist = createPlaylistObject(-1, 'Untitled', 0, playlistWidth, playlistHeight);
		populatePlaylist(gCurrentPlaylist, false, playlistWidth, playlistHeight);
		resizeCanvas();
		SetSaveFlag(true);
	}
}

async function LoadPlaylistById(playlistId) {
	localStorage.lastplaylistid = playlistId;
	clearPlaylist();
	if (playlistId == null) {
		return;
	}
	LockBody();
	try {
		var data = await get('/get/playlist.php', {
			id: playlistId
		});
		UnlockBody();
		if (data.success) {
			populatePlaylist(data.playlist, false, data.playlist.pagewidth, data.playlist.pageheight);
			gTimeBarSelection = 0;
			UpdateTimeBar();
		}
		return data;
	} catch (exception) {
		UnlockBody();
		NetworkingError();
		throw exception;
	}
}

function populatePlaylist(playlist, isTemplate, playlistWidth, playlistHeight) {
	if (!isTemplate) {
		playlistWidth = playlist.pagewidth;
		playlistHeight = playlist.pageheight;
	}
	
	var scaleX = playlistWidth / playlist.pagewidth;
	var scaleY = playlistHeight / playlist.pageheight;
	
	// DO NOT CLAMP. THESE ARE USED FOR BOTH ZONE AND MEDIA.
	var ScaleX = value => {
		return isTemplate ? Math.round(value * scaleX) : value;
	};
	var ScaleY = value => {
		return isTemplate ? Math.round(value * scaleY) : value;
	};
	
	document.getElementById('editor-canvastip').hidden = true;
	$editorCanvasShadow.hidden = false;
	
	gCurrentPlaylist = createPlaylistObject(isTemplate ? -1 : playlist.id, isTemplate ? 'Untitled' : playlist.name, 0, playlistWidth, playlistHeight);
	document.getElementById('mediatab').click();
	
	for (const playlistZone of playlist.zones) {
		var zone = new Zone(playlistZone.id, playlistZone.name, ScaleX(playlistZone.x), ScaleY(playlistZone.y), playlistZone.z, ScaleX(playlistZone.width), ScaleY(playlistZone.height), playlistZone.volume, playlistZone.locked);
		for (const media of playlistZone.media) {
			// Note the use of (({a}) => ({a}))(o) is just picking syntax to pull out a subset of properties from an object
			const commonProperties = Object.assign((({ id, name, runlength, scheduled, utc, startdate, enddate, daypartstarttime, daypartendtime, sunday, monday, tuesday, wednesday, thursday, friday, saturday }) => ({ id, name, runlength, scheduled, utc, startdate, enddate, daypartstarttime, daypartendtime, sunday, monday, tuesday, wednesday, thursday, friday, saturday }))(media),
			{
				x: ScaleX(media.x),
				y: ScaleY(media.y),
				width: ScaleX(media.width),
				height: ScaleY(media.height)
			});
			if (media.type != MEDIATYPE_AUDIO) {
				// Copy transition properties
				Object.assign(commonProperties, (({ entrytransitiontype, entrytransitionduration, exittransitiontype, exittransitionduration }) => ({ entrytransitiontype, entrytransitionduration, exittransitiontype, exittransitionduration }))(media));
			}
			switch (media.type) {
				case MEDIATYPE_IMAGE:
					zone.media.push(new MediaImage(zone, Object.assign(commonProperties, (({ cdnurl, actualwidth, actualheight, contentid, fillstyleid , zoompan, uploadedtimestamp, filesize }) => ({ cdnurl, actualwidth, actualheight, contentid, fillstyleid, zoompan, uploadedtimestamp, filesize }))(media))));
					break;
				case MEDIATYPE_VIDEO:
					zone.media.push(new MediaVideo(zone, Object.assign(commonProperties, (({ cdnurl, actualwidth, actualheight, autoplay, loop, controls, volume, ducking, contentid, fillstyleid, uploadedtimestamp, filesize }) => ({ cdnurl, actualwidth, actualheight, autoplay, loop, controls, volume, ducking, contentid, fillstyleid, uploadedtimestamp, filesize }))(media))));
					break;
				case MEDIATYPE_AUDIO:
					zone.media.push(new MediaAudio(zone, Object.assign(commonProperties, (({ cdnurl, autoplay, loop, controls, contentid, volume, ducking, uploadedtimestamp, filesize }) => ({ cdnurl, autoplay, loop, controls, contentid, volume, ducking, uploadedtimestamp, filesize }))(media))));
					break;
				case MEDIATYPE_MISC:
					zone.media.push(new MediaMisc(zone, Object.assign(commonProperties, (({ cdnurl, contentid, uploadedtimestamp, filesize }) => ({ cdnurl, contentid, uploadedtimestamp, filesize }))(media))));
					break;
				case MEDIATYPE_WIDGET:
					RepairWidget(media);
					zone.media.push(new MediaWidget(zone, Object.assign(commonProperties, (({ widgettype, icon, attributes }) => ({ widgettype, icon, attributes }))(media))));
					break;
			}
		}
		AddZone(zone, true);
	}
	setWindowHeader();
	UpdateWidgetNames();
	if (!localStorage.hasOwnProperty('swapzoneorder')) {
		localStorage.swapzoneorder = false;
	}
	if (!localStorage.hasOwnProperty('zoneresize')) {
		localStorage.zoneresize = JSON.stringify([]);
	}
	UpdateZoneAndMediaDOM();
	for (const zone of gCurrentPlaylist.zones) {
		SelectMedia(zone, zone.media.length != 0 ? [0] : [], false, false, false, '', false, false);
	}
	SelectZone(null, true, false);
	
	reloadAllZoneViews();
	$buttonTab.hidden = !gCurrentPlaylist.IsInteractive();
	if (!gCurrentPlaylist.IsInteractive() && $buttonTab.classList.contains('selected')) {
		document.getElementById('mediatab').click();
	}
	UpdateButtonList();
	
	if (!localStorage.hasOwnProperty('zonestate')) {
		localStorage.zonestate = JSON.stringify([]);
	}
	var zoneState = JSON.parse(localStorage.zonestate);
	
	gCurrentPlaylist.zones.forEach((zone, zoneItr) => {
		zone.visible = zoneItr < zoneState.length ? zoneState[zoneItr] : true;
	});
	
	UpdateZoneVisibility();
	resizeCanvas();
}

async function RequestWebContent(selectedWebpage) {
	const $webpages = document.getElementById('webpages');
	$webpages.classList.add('busy');
	try {
		var data = await get('/get/contentweb.php');
		$webpages.innerHTML = '';
		CheckSessionExpired(data);
		if (data.success) {
			for (const webpage of data.webpages) {
				const $webpage = $webpages.appendChild($.li({ class: [selectedWebpage == webpage['url'], 'selected'], text: webpage['url'], title: webpage['url'], click: function() {
					for (const $e of document.querySelectorAll('#webpages > li')) { $e.classList.remove('selected'); }
					this.classList.add('selected');
					document.getElementById('webpageurl').value = this.innerText;
					document.getElementById('webpageurl').dispatchEvent(new Event('change'));
				} }));
				$webpage.webpageid = webpage.id;
			}
		}
	} catch (exception) {
		NetworkingError();
	} finally {
		$webpages.classList.remove('busy');
	}
}

async function GetWidgetAttributes(widget) {
	const attributes = {};
	for (const attribute of widget.attributes) {
		attributes[attribute.name] = await attribute.value;
	}
	// Hacky addition for inlineimages that stores images used in the text
	for (const attributeName in attributes) {
		if (widget.name == 'text' && attributeName == 'inlineimages') {
			// This is convoluted because we're doing it here instead of server-side. Because someone might have a restricted content tree view we don't have all the images to update in some cases so we update only the ones we can
			const imageNames = [];
			const imagePromises = [];
			attributes.text.replace(/\[image="[^"]*?"[^\]]*?\]/g, imageCode => {
				imagePromises.push(new Promise(async resolve => {
					var imageName = imageCode.match(/image="([^"]*)"/);
					imageName = imageName.length == 2 ? imageName[1] : '';
					imageNames.push(imageName);
					for (const node of await $contentTree.getAllNodes()) {
						if (node.name == imageName) {
							resolve(ContentNodeToMinimalImageContent(node));
						}
					}
					resolve(null);
				}));
				return imageCode;
			});
			const images = (await Promise.all(imagePromises)).filter(image => image);
			const oldImages = document.getElementById('textinlineimages').data;
			// Replace all the oldImage references with newer copies
			for (const image of images) {
				var index = oldImages.findIndex(oldImage => oldImage.name == image.name);
				if (index != -1) {
					oldImages.splice(index, 1);
				}
			}
			var newImages = oldImages.concat(images);
			// Remove any image references that are no longer used in the text
			for (var itr = 0; itr < newImages.length; ++itr) {
				if (!imageNames.some(imageName => newImages[itr].name == imageName)) {
					newImages.splice(itr--, 1);
				}
			}
			attributes.inlineimages = newImages;
		}
	}
	for (const attribute of widget.attributes) {
		if (attribute.hasOwnProperty('$metadataInput')) {
			attributes[`${attribute.name}metadata`] = attribute.$metadataInput.value;
		}
		if (attribute.hasOwnProperty('$generatedMetadataInput')) {
			attributes[`${attribute.name}generatedmetadata`] = attribute.$generatedMetadataInput.value;
		}
	}
	return attributes;
}

function RemoveValidationError() {
	this.classList.remove('validationError');
	this.removeAttribute('title');
}

function SetWidgetDefaults(widget) {
	for (const attribute of widget.attributes) {
		attribute.value = JSON.parse(JSON.stringify(attribute.default));
		if (attribute.hasOwnProperty('$metadataInput')) {
			attribute.$metadataInput.value = '';
		}
		if (attribute.hasOwnProperty('$generatedMetadataInput')) {
			attribute.$generatedMetadataInput.value = '';
		}
	}
}

async function CallWidgetChange(widget) {
	var attributeValues = await GetWidgetAttributes(widget);
	for (const attribute of widget.attributes) {
		attribute.$input.oldvalue = attributeValues[attribute.name];
		switch (attribute.type) {
			case 'string':
			case 'bool':
			case 'int':
			case 'float':
			case 'color':
			case 'date':
			case 'time':
			case 'json':
				attribute.$input.dispatchEvent(new Event('change'));
				break;
			default:
				throw 'attribute type needs to be handled';
		}
	}
}

function powerBIFilterName(filter, filterIndex) {
	/*
	https://github.com/Microsoft/PowerBI-JavaScript/wiki/Filters#filter-types
	Advanced = 0,
	Basic = 1,
	Unknown = 2,
	IncludeExclude = 3,
	RelativeDate = 4,
	TopN = 5,
	Tuple = 6,
	RelativeTime = 7
	*/
	// Has a custom name
	if (filter.hasOwnProperty('displaySettings') && filter.displaySettings.hasOwnProperty('displayName')) {
		return filter.displaySettings.displayName;
	} else if (filter.filterType == 0) {
		return `${filter.target.table} (${filter.target.column}) Advanced (Requires a custom name to use)`;
	} else if (filter.filterType == 1) {
		return `${filter.target.table} (${filter.target.column}) (${filter.operator}) Basic (Requires a custom name to use)`;
	} else if (filter.filterType == 2) {
		return `${filter.target.table} (${filter.target.column}) Unknown (Requires a custom name to use)`;
	} else if (filter.filterType == 3) {
		return `${filter.target.table} (${filter.target.column}) IncludeExclude (Requires a custom name to use)`;
	} else if (filter.filterType == 4) {
		return `${filter.target.table} (${filter.target.column}) RelativeDate (Requires a custom name to use)`;
	} else if (filter.filterType == 5) {
		return `${filter.target.table} (${filter.target.column}) TopN (Requires a custom name to use)`;
	} else if (filter.filterType == 6) {
		return `${filter.target.table} (${filter.target.column}) Tuple (Requires a custom name to use)`;
	} else if (filter.filterType == 7) {
		return `${filter.target.table} (${filter.target.column}) RelativeTime (Requires a custom name to use)`;
	}
	return `Filter ${filterIndex}`;
}

function InitializeWidgets() {
	for (const widget of WidgetInformation) {
		widget.attr = {};
		// Format validation
		for (const attribute of widget.attributes) {
			if (!attribute.hasOwnProperty('default')) {
				throw 'Attribute must have default: ' + attribute.name;
			}
			// TODO type and property requirement validation
			widget.attr[attribute.name] = attribute;
		}
		const attr = widget.attr;
		// Create the widget panel item
		const widgetDragTag = document.getElementById('widgets').appendChild($.div({ id: `div${widget.name}`, title: widget.description, class: 'widget dragWidget button-choices', draggable: {
			thresholdX: 3,
			thresholdY: 3,
			start: e => {
				if (DialogWindowOpen() || InPreviewMode() || gCurrentPlaylist == null) {
					BlinkDialog();
					return false;
				}
				return widgetDragTag;
			},
			drag: e => {
				var zone = getZoneAt(e);
				LineLayerClear();
				LineLayerHighlightZone(zone, true);
			}
		} },
			$.i({ class: `mdi ${widget.icon}` }),
			$.span({ text: widget.display })
		));
		widgetDragTag.widget = widget;
		
		if (widget.hasOwnProperty('parent') || widget.name == 'exchange' || widget.name == 'stockcrawl') {
			widgetDragTag.remove();
		}
		// Some widgets are flagged for advanced users and require certain user permissions to be set. Hide widgets from non-advanced users.
		if (widget.hasOwnProperty('permissions') && widget.permissions.some(permission => !permissions[permission])) {
			widgetDragTag.remove();
		}
		
		widget.$dialog = document.body.appendChild($.uidialog({ id: widget.name, class: 'widget-dialog', movable: { handle: ':scope > header' } },
			$.header({ text: widget.display }),
			$.ul({ class: 'iconlist' },
				$.li({ title: 'Reset Widget to Defaults', class: 'tooltip-button mdi mdi-eraser', click: async () => {
					if (await QuestionBox('Reset Widgets to Defaults?', 'Are you sure you want to reset the widget values in the form to their defaults?')) {
						SetWidgetDefaults(widget);
						CallWidgetChange(widget);
					}
				} })
			)
		));
		widget.$dialog.hidden = true;
		
		Object.assign(widget.$dialog.style, {
			minWidth: '300px'
		});
		if (widget.hasOwnProperty('width')) {
			widget.$dialog.style.width = `${widget.width}px`;
		}
		if (widget.hasOwnProperty('height')) {
			widget.$dialog.style.height = `${widget.height}px`;
		}
		
		const CreateAttributeInput = (attribute, id) => {
			const buildOptions = () => {
				if (attribute.hasOwnProperty('groupedbutton')) {
					return $.fieldset({ id });
				} else {
					return $.select({ id });
				}
			};
			const placeholder = attribute.hasOwnProperty('placeholder') ? `${attribute.placeholder}${attribute.hasOwnProperty('required') ? ' (Required)' : '' }`: (attribute.hasOwnProperty('required') ? 'Required' : '');
			switch (attribute.type) {
				case 'bool':
					if (attribute.hasOwnProperty('icon')) {
						return $.uicheckbox({ id, config: { icon: attribute.icon } });
					}
					return $.uicheckbox({ id });
				case 'string':
					if (attribute.hasOwnProperty('options')) {
						return buildOptions();
					} else if (attribute.hasOwnProperty('password')) {
						return $.input({ type: 'password', id });
					}
					return $.input({ type: 'text', id, placeholder });
				case 'int':
					if (attribute.hasOwnProperty('options')) {
						return buildOptions();
					}
					return $.input({ type: 'text', id, placeholder });
				case 'float':
					if (attribute.hasOwnProperty('options')) {
						return buildOptions();
					}
					return $.input({ type: 'text', id, placeholder });
				case 'color':
					var configuration = { ...(attribute.hasOwnProperty('options') ? attribute.options : []) };
					configuration.previewChange = function() { this.dispatchEvent(new Event('change')); };
					if (!configuration.hasOwnProperty('palette')) {
						configuration.showPalette = true;
						if (!configuration.hasOwnProperty('localStorageKey')) {
							configuration.localStorageKey = 'globalcolor';
						}
					}
					return $.uicolor({ id, config: configuration });
				case 'date':
					return $.input({ id });
				case 'time':
					return $.input({ id });
				case 'json':
					return $.input({ id });
				default:
					throw 'attribute type needs to be handled';
			}
		};
		const $mediaHeaderWidgetOptions = document.getElementById('media-header-widgetoptions');
		let $lastGroup;
		// Find attributes that are placed in the left property panel
		for (const attribute of widget.attributes) {
			const id = widget.name + attribute.name;
			const headerId = `media-header-${widget.name}-${attribute.headergroup}`;
			if (attribute.hasOwnProperty('viewheader')) {
				if (typeof attribute.viewheader == 'object') {
					attribute.$header = $mediaHeaderWidgetOptions.insertAdjacentElement('beforebegin', $.header({ id: headerId, text: attribute.viewheader.name },
						$.span({}, ...attribute.viewheader.labels.map(label => {
							if (attr.hasOwnProperty(label)) {
								const labelAttribute = widget.attr[label];
								labelAttribute.$headerLabel = $.label({ for: widget.name + labelAttribute.name, title: labelAttribute.display, class: 'tooltip' });
								return labelAttribute.$headerLabel;
							} else {
								return $.span({ text: label });
							}
						})),
						$.span({ title: `Expand/Collapse ${attribute.viewheader.name}`, class: 'mdi mdi-play mdi-rotate-90 tooltip' })
					));
				} else {
					attribute.$header = $mediaHeaderWidgetOptions.insertAdjacentElement('beforebegin', $.header({ id: headerId, text: attribute.viewheader },
						$.span({}),
						$.span({ title: `Expand/Collapse ${attribute.viewheader}`, class: 'mdi mdi-play mdi-rotate-90 tooltip' })
					));
				}
				if (!localStorage.hasOwnProperty(headerId)) {
					localStorage[headerId] = true;
				}
				attribute.$header.classList.toggle('collapsed', localStorage[headerId] == 'true');
				$lastGroup = $mediaHeaderWidgetOptions.insertAdjacentElement('beforebegin', $.div({ class: 'group' }));
			}
			if (attribute.hasOwnProperty('view') && attribute.view == 'property') {
				attribute.$label = $lastGroup.appendChild($.label({ for: id, text: attribute.display, class: `tooltip-help ${widget.name}-${attribute.headergroup}` }));
				const $attributeInput = CreateAttributeInput(attribute, id);
				$attributeInput.classList.add('property-widget-view-option', `${widget.name}-${attribute.headergroup}`);
				attribute.$inputContainer = $lastGroup.appendChild($attributeInput);
				attribute.$input = attribute.$inputContainer.id == id ? attribute.$inputContainer : document.getElementById(id);
			}
		}
		
		// Properties placed in the pop up dialog
		if (!widget.hasOwnProperty('design') || widget.design != 'custom') {
			const $widgetDialogTabs = $.div();
			widget.$dialog.appendChild($widgetDialogTabs);
			const $tabs = $.ul();
			$widgetDialogTabs.appendChild($tabs);
			
			let $currentFormGrid = null;
			for (const attribute of widget.attributes) {
				if (attribute.hasOwnProperty('view') && attribute.view == 'property') {
					continue;
				}
				var id = widget.name + attribute.name;
				if (attribute.hasOwnProperty('group')) {
					$currentFormGrid = $.div({ class: 'formgrid' });
					$widgetDialogTabs.appendChild($currentFormGrid);
					$tabs.appendChild($.li({ text: attribute.group }));
				}
				// If no groups are used then a default form grid is made with no associated tab
				if ($currentFormGrid == null) {
					$currentFormGrid = $.div({ class: 'formgrid' });
					$widgetDialogTabs.appendChild($currentFormGrid);
				}
				
				attribute.$label = $currentFormGrid.appendChild($.label({ for: id, title: attribute.description, text: attribute.display, class: 'tooltip-help', css: {
					paddingLeft: attribute.hasOwnProperty('indent') ? `${attribute.indent * 10}px` : 0
				} }));
				const createBasicInput = () => {
					// The actual setup code runs later since it sets up both automatically generated forms and custom forms
					attribute.$inputContainer = CreateAttributeInput(attribute, id);
					if (attribute.$inputContainer) {
						$currentFormGrid.appendChild(attribute.$inputContainer);
						attribute.$input = attribute.$inputContainer.id == id ? attribute.$inputContainer : document.getElementById(id);
					}
				};
				if (attribute.hasOwnProperty('design')) {
					switch (attribute.design) {
						case 'fontfamily':
							attribute.$input = $currentFormGrid.appendChild($.uiinputselect({ id, config: {
								options: fonts,
								history: 10,
								processOption: function() {
									this.style.fontFamily = `'${this.value}'`;
								}
							}, textchanged: function() {
								this.$input.style.fontFamily = `'${this.value}'`;
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'rssfeed':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'rss',
								defaultOptions: [{ id: 0, name: 'Use URL', value: 'useurl' }],
								leafIcon: node => 'schedule',
								filter: node => !node.group
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'csvfile':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'csv',
								defaultOptions: [{ id: 0, name: 'Use URL', value: 'useurl' }],
								leafIcon: node => 'schedule',
								filter: node => !node.group
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'eventlist':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'event',
								leafIcon: node => 'schedule',
								filter: node => !node.group
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'location':
							{
							createBasicInput();
							let $spinner = null;
							let $greenOk = null;
							let $redError = null;
							let controller = null;
							var locationBouncerTimeout = null;
							attribute.$input.classList.add('sampleInput');
							Object.assign(attribute.$input.style, {
								paddingRight: '25px',
								width: '100%'
							});
							
							for (const event of ['change', 'input']) {
								attribute.$input.addEventListener(event, function() {
									var location = this.value;
									if (location == '') {
										for (const $e of [$spinner, $greenOk]) $e.hidden = true;
										$redError.hidden = false;
										return;
									}
									clearTimeout(locationBouncerTimeout);
									locationBouncerTimeout = setTimeout(async () => {
										for (const $e of [$greenOk, $redError]) $e.hidden = true;
										$spinner.hidden = false;
										if (controller) {
											controller.abort();
										}
										controller = new AbortController();
										try {
											var data = await post('/post/checkvalidweather.php', {
												location: location
											}, controller);
											if (data.success) {
												for (const $e of [$spinner, $redError]) $e.hidden = true;
												$greenOk.hidden = false;
											} else {
												for (const $e of [$spinner, $greenOk]) $e.hidden = true;
												$redError.hidden = false;
											}
										} catch (exception) {
										}
									}, 500);
								});
							}
							$currentFormGrid.appendChild($.span({ css: {
								display: 'flex',
								flexFlow: 'row nowrap',
								alignItems: 'center',
								position: 'relative'
							} },
								attribute.$input,
								$spinner = $.i({ class: 'spinner mdi mdi-loading mdi-spin', hidden: true }),
								$greenOk = $.i({ class: 'greenOk mdi mdi-check-circle-outline', hidden: true }),
								$redError = $.i({ class: 'redError mdi mdi-alert-circle-outline', hidden: true })
							));
							attribute.$inputContainer = attribute.$input.parentNode;
							}
							break;
						case 'rssfeedurl':
							{
							createBasicInput();
							let $spinner = null;
							let $greenOk = null;
							let $redError = null;
							let controller = null;
							attribute.$input.classList.add('sampleInput');
							Object.assign(attribute.$input.style, {
								paddingRight: '25px',
								width: '100%'
							});
							for (const event of ['change', 'input']) {
								attribute.$input.addEventListener(event, async function() {
									var rssUrl = this.value;
									if (rssUrl == '') {
										for (const $e of [$spinner, $greenOk]) $e.hidden = true;
										$redError.hidden = false;
										return;
									}
									for (const $e of [$greenOk, $redError]) $e.hidden = true;
									$spinner.hidden = false;
									if (controller) {
										controller.abort();
									}
									controller = new AbortController();
									try {
										var data = await post('/post/checkvalidrssfeed.php', {
											url: rssUrl
										}, controller);
										if (data.success) {
											for (const $e of [$spinner, $redError]) $e.hidden = true;
											$greenOk.hidden = false;
										} else {
											for (const $e of [$spinner, $greenOk]) $e.hidden = true;
											$redError.hidden = false;
										}
									} catch (exception) {
									}
								});
							}
							$currentFormGrid.appendChild($.span({ css: {
								display: 'flex',
								flexFlow: 'row nowrap',
								alignItems: 'center',
								position: 'relative'
							} },
								attribute.$input,
								$spinner = $.i({ class: 'spinner mdi mdi-loading mdi-spin', hidden: true }),
								$greenOk = $.i({ class: 'greenOk mdi mdi-check-circle-outline', hidden: true }),
								$redError = $.i({ class: 'redError mdi mdi-alert-circle-outline', hidden: true })
							));
							attribute.$inputContainer = attribute.$input.parentNode;
							}
							break;
						case 'media':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree:'content',
								leafIcon: node => GetContentIcon(node),
								filter: node => attribute.hasOwnProperty('designfilter') ? node.contenttreetype == attribute.designfilter : true
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'mediaarray':
							attribute.$input = $currentFormGrid.appendChild($.div({ id, class: 'formgrid compact' }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'masterpins':
							attribute.$input = $currentFormGrid.appendChild($.div({ id, class: 'formgrid compact' }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'predefinedtitles':
							attribute.$input = $currentFormGrid.appendChild($.div({ id, class: 'formgrid compact' }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'metadatafilters':
							attribute.$input = $currentFormGrid.appendChild($.ul({ id, css: { width: '100%', height: '250px' }, change: function() {
								MetadataFilters(this, this.data);
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'powerbireportfilters':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'powerbireportfilters',
								showRoot: true,
								src: async () => {
									for (let i = 0; i < 3; ++i) {
										if (gCurrentWidget == null) return [];
										try {
											const filterAndPages = await gCurrentWidget.widget.GetFilterTree();
											return filterAndPages.filters.map((filter, filterIndex) => ({
												id: filterIndex,
												parent: null,
												type: 1,
												name: powerBIFilterName(filter, filterIndex),
												filter
											}));
										} catch (exception) {
											await new Promise(resolve => setTimeout(resolve, 1000));
										}
									}
									return [];
								},
								disable: node => !node.filter.hasOwnProperty('displaySettings') || !node.filter.displaySettings.hasOwnProperty('displayName'),
								checkbox: true,
								groupTypes: [],
								leafIcon: node => 'filter'
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'powerbipagefilters':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'powerbipagefilters',
								showRoot: true,
								src: async () => {
									for (let i = 0; i < 3; ++i) {
										try {
											if (gCurrentWidget == null) return [];
											const tree = [];
											let sequence = 0;
											const filterAndPages = await gCurrentWidget.widget.GetFilterTree();
											for (const [pageIndex, { page, filters }] of filterAndPages.pages.entries()) {
												const pageId = sequence++;
												tree.push({ id: pageId, parent: null, type: 1, name: `${page.displayName} (${page.name})` });
												for (const [filterIndex, filter] of filters.entries()) {
													tree.push({ id: sequence++, parent: pageId, type: 2, name: powerBIFilterName(filter, filterIndex), filter });
												}
											}
											return tree;
										} catch (exception) {
											await new Promise(resolve => setTimeout(resolve, 1000));
										}
									}
									return [];
								},
								disable: node => node.type != 2 || !node.filter.hasOwnProperty('displaySettings') || !node.filter.displaySettings.hasOwnProperty('displayName'),
								checkbox: true,
								groupTypes: [1],
								groupIcon: node => node.type == 1 ? 'page' : '',
								leafIcon: node => node.type == 2 ? 'filter' : ''
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'powerbivisualfilters':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'powerbivisualfilters',
								showRoot: true,
								src: async () => {
									for (let i = 0; i < 3; ++i) {
										try {
											if (gCurrentWidget == null) return [];
											const tree = [];
											let sequence = 0;
											const filterAndPages = await gCurrentWidget.widget.GetFilterTree();
											for (const [pageIndex, { page, visuals }] of filterAndPages.pages.entries()) {
												const pageId = sequence++;
												tree.push({ id: pageId, parent: null, type: 1, name: page.displayName });
												for (const [visualIndex, { visual, filters }] of visuals.entries()) {
													if (filters.length != 0) {
														const visualId = sequence++;
														tree.push({ id: visualId, parent: pageId, type: 2, name: `${visual.title == undefined || visual.title == '' ? visual.name : visual.title} (${visual.type})` });
														for (const [filterIndex, filter] of filters.entries()) {
															tree.push({ id: sequence++, parent: visualId, type: 3, name: powerBIFilterName(filter, filterIndex), filter });
														}
													}
												}
											}
											return tree;
										} catch (exception) {
											await new Promise(resolve => setTimeout(resolve, 1000));
										}
									}
									return [];
								},
								disable: node => node.type != 3 || !node.filter.hasOwnProperty('displaySettings') || !node.filter.displaySettings.hasOwnProperty('displayName'),
								checkbox: true,
								groupTypes: [1, 2],
								groupIcon: node => {
									switch (node.type) {
										case 1:
											return 'page';
										case 2:
											return 'visual';
									}
								},
								leafIcon: node => node.type == 3 ? 'filter' : ''
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'powerbireportfiltervalues':
							attribute.$input = $currentFormGrid.appendChild($.div({ class: 'formgrid compact', css: { gridTemplateColumns: '1fr max-content 1fr', maxWidth: '300px' } }));
							attribute.$input.renderFilters = async () => {
								// GetFilterTree requires the widget to be loaded and can throw exceptions
								for (let i = 0; i < 3; ++i) {
									try {
										if (gCurrentWidget == null) return;
										const filterNameToValues = attribute.$input.data;
										attribute.$input.innerHTML = '';
										attribute.$input.append($.header({},
											$.label({ text: 'Default', title: 'The default value set for the filter in the report' }),
											$.label({ text: 'Override', title: 'Change the filter value used. Ensure that the filter is selected in the above drop downs' }),
											$.label({ text: 'Value(s)', title: 'The value to substitute into the filter when it is applied to the report' })
										));
										// Render all the pages, visuals, and filters for filters with names. This is formatted in an indented tree
										const filterTree = await gCurrentWidget.widget.GetFilterTree();
										// Clean up deleted filters that we might be storing values for
										const filterNameToDefaultValues = filterTree.filterNameToDefaultValues;
										for (const key in filterNameToValues) {
											if (!filterNameToDefaultValues.hasOwnProperty(key)) {
												delete filterNameToValues[key];
											}
										}
										const addFilterRows = (filter, defaultFilterValues, indent) => {
											filterNameToValues[filter.displaySettings.displayName] ??= [];
											const filterValues = filterNameToValues[filter.displaySettings.displayName];
											while (filterValues.length < defaultFilterValues.length) {
												filterValues.push({ override: false, value: '' });
											}
											attribute.$input.append(
												$.label({ class: 'powerbilabel', text: filter.displaySettings.displayName, css: { marginLeft: `${indent * 5}px` } }),
												...defaultFilterValues.map((defaultFilterValue, index) => {
													const filterValue = filterValues[index];
													const defaultValue = Array.isArray(defaultFilterValue) ? defaultFilterValue.join(',') : defaultFilterValue;
													return [
														$.input({ type: 'text', value: defaultValue, title: defaultValue, readonly: true, css: { height: '22px', marginLeft: `${(indent + 1) * 5}px` } }),
														$.uicheckbox({ type: 'text', checked: filterValue.override, css: { height: '22px' }, change: function(e) {
															e.stopPropagation();
															filterValue.override = this.checked;
															attribute.$input.renderFilters();
														} }),
														$.input({ type: 'text', value: filterValue.value, readonly: !filterValue.override, css: { height: '22px' }, change: function() {
															filterValue.value = this.value;
														} })
													];
												}).flat()
											);
										}
										for (const [index, filter] of filterTree.filters.entries()) {
											if (filterTree.filterValues[index].length != 0) {
												addFilterRows(filter, filterTree.filterValues[index], 0);
											}
										}
										for (const page of filterTree.pages) {
											if (page.hasCustomFilterWithValues) {
												attribute.$input.appendChild(
													$.label({ class: 'powerbilabel', text: `${page.page.displayName} (${page.page.name})`, title: 'Page' })
												);
												for (const [index, filter] of page.filters.entries()) {
													addFilterRows(filter, page.filterValues[index], 1);
												}
												for (const visual of page.visuals) {
													if (visual.hasCustomFilterWithValues) {
														attribute.$input.appendChild(
															$.label({ class: 'powerbilabel', text: `${visual.visual.title == undefined || visual.visual.title == '' ? visual.visual.name : visual.visual.title} (${visual.visual.type})`, title: 'Visual', css: { marginLeft: '5px' } })
														);
														for (const [index, filter] of visual.filters.entries()) {
															if (visual.filterValues[index].length != 0) {
																addFilterRows(filter, visual.filterValues[index], 2);
															}
														}
													}
												}
											}
										}
										return;
									} catch (exception) {
										await new Promise(resolve => setTimeout(resolve, 1000));
									}
								}
							};
							attribute.$inputContainer = attribute.$input;
							break;
						case 'networkcommands':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'networkcommand',
								fullSelect: true,
								checkbox: true,
								groupTypes: [1, 2, 3],
								groupIcon: node => node.type == 2 ? 'networkcommand' : '',
								disable: node => node.type == 1,
								filter: node => node.type == 1 || node.type == 2
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'apicommands':
							attribute.$input = $currentFormGrid.appendChild($.uitreeselect({ id, config: {
								tree: 'apitokens',
								showRoot: true,
								checkbox: true,
								src: async () => {
									const response = await get('/get/apitoken.php');
									if (!response.success) {
										toastError('Error getting API tokens.');
										return [];
									}
									return response.apitokens.map(token => ({ id: token.id, parent: null, type: 1, name: token.display, token }));
								}
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
						case 'idledelay':
							attribute.$input = $currentFormGrid.appendChild($.uiinputselect({ id, config: {
								options: (() => {
									const delays = [];
									for (let value = 0; value <= 150; ++value) {
										delays.push({ name: value, value });
									}
									return delays;
								})(),
								allowInput: true
							} }));
							attribute.$inputContainer = attribute.$input;
							break;
					}
				} else {
					createBasicInput();
				}
				
				const validMetadataType = metadataDataTypes.find(type => type.value == attribute.type);
				if (validMetadataType && (!attribute.hasOwnProperty('metadata') || attribute.metadata)) {
					const metadataId = `${id}metadata`;
					const generatedMetadataId = `${id}generatedmetadata`;
					attribute.$metadataToggle = $currentFormGrid.appendChild($.i({ title: 'Use metadata value. Enabling this turns the value to the left into the default value. If no metadata is found on this widget matching the name then the default value is used.', class: 'mdi mdi-code-brackets tooltip-button metadatatoggle', click: () => {
						const $metadata = document.getElementById(metadataId);
						if (!$metadata.hidden) {
							$metadata.value = '';
							attribute.$metadataLabel.hidden = true;
							attribute.$metadataInput.hidden = true;
						} else {
							attribute.$metadataLabel.hidden = false;
							attribute.$metadataInput.hidden = false;
						}
					} }));
					if (attribute.hasOwnProperty('generatedmetadata') && attribute.generatedmetadata)
					{
						attribute.$generatedMetadataToggle = $currentFormGrid.appendChild($.i({ title: 'Use generated item metadata value. Enabling this turns the value to the left into the default value. If no metadata property that matches the name is found on the generated item the default value is used. (Assuming the user has not enabled a metadata toggle', class: 'mdi mdi-code-array tooltip-button metadatatoggle', click: () => {
						const $generatedMetadata = document.getElementById(generatedMetadataId);
						if (!$generatedMetadata.hidden) {
								$generatedMetadata.value = '';
								attribute.$generatedMetadataLabel.hidden = true;
								attribute.$generatedMetadataInput.hidden = true;
							} else {
								attribute.$generatedMetadataLabel.hidden = false;
								attribute.$generatedMetadataInput.hidden = false;
							}
						} }));
					}
					attribute.$metadataInput = $currentFormGrid.appendChild($.input({ type: 'text', id: metadataId, placeholder: 'Metadata Name', css: { gridColumn: 2 } }));
					attribute.$metadataLabel = $currentFormGrid.appendChild($.label({ for: metadataId, title: `Metadata reference name. Note: The type must match for the metadata. For example, if the widget property expects a color you cannot assign a text value.`, class: 'mdi mdi-subdirectory-arrow-left tooltip-help metadataarrow' }));
					if (attribute.hasOwnProperty('generatedmetadata') && attribute.generatedmetadata) {
						attribute.$generatedMetadataInput = $currentFormGrid.appendChild($.input({ type: 'text', id: generatedMetadataId, placeholder: 'Generated Item Metadata Name', css: { gridColumn: '2' } }));
						attribute.$generatedMetadataLabel = $currentFormGrid.appendChild($.label({ for: generatedMetadataId, title: `Metadata reference name. Note: The type must match for the metadata. For example, if the widget property expects a color you cannot assign a text value.`, class: 'mdi mdi-subdirectory-arrow-left tooltip-help generatedmetadataarrow' }));
					} else {
						Object.assign(attribute.$metadataToggle.style, {
							gridColumnEnd: 'span 2'
						});
					}
				}
			}
			$widgetDialogTabs.classList.add('horizontal');
			let beforeOffset;
			tabs($widgetDialogTabs, {
				before: () => {
					beforeOffset = $.offset(widget.$dialog)
					widget.$dialog.style.left = 0;
					widget.$dialog.style.top = 0;
				},
				after: () => {
					movableConstrainInWindow(widget.$dialog, beforeOffset);
				}
			});
		}
		switch (widget.name) {
			case 'text':
				widget.$dialog.style.width = '990px';
				widget.$dialog.append(
					$.div({ css: {
						display: 'flex',
						flexFlow: 'row nowrap',
						alignItems: 'center',
						padding: '10px'
					} },
						$.uiinputselect({ id: 'textfontfamily', title: 'Font', css: { flex: '1 0 auto', marginRight: '5px' }, config: { options: fonts } }),
						$.uiinputselect({ id: 'textfontsize', title: 'Font Size', css: { marginRight: '5px', minWidth: '40px' }, config: {
							options: (() => {
								const fontSizes = [];
								for (let value = 8; value <= 400; ++value) {
									fontSizes.push({ name: value, value });
								}
								return fontSizes;
							})()
						} }),
						$.select({ id: 'textlineheight', title: 'Line Height', css: { marginRight: '5px' } }),
						$.uicolor({ id: 'textforeground', title: 'Font Color', css: { marginRight: '5px' }, config: {
							compact: true,
							showPalette: true,
							localStorageKey: 'globalcolor',
							previewChange: function() { this.dispatchEvent(new Event('change')); }
						} }),
						$.uicolor({ id: 'textbackground', title: 'Background Color', css: { marginRight: '5px' }, config: {
							compact: true,
							showPalette: true,
							localStorageKey: 'globalcolor',
							previewChange: function() { this.dispatchEvent(new Event('change')); }
						} }),
						$.uicheckbox({ id: 'textbold', class: 'toolbar-checkbox', title: 'Bold', config: { icon: 'format-bold' }, css: {
							marginRight: '5px'
						} }),
						$.uicheckbox({ id: 'textitalic', class: 'toolbar-checkbox', title: 'Italic', config: { icon: 'format-italic' }, css: {
							marginRight: '5px'
						} }),
						$.fieldset({ id: 'texthorizontal', css: { marginRight: '6px' } }),
						$.fieldset({ id: 'textvertical' }),
						$.label({ title: 'Edit text and inline image filters', class: 'text-opentextfilters mdi mdi-brush', css: {
							fontSize: '22px',
							cursor: 'pointer',
							margin: '0 5px'
						}, click: () => {
							var saveFilters = () => {
								var filters = document.getElementById('textfilters').data;
								filters.text = [];
								filters.inlineimages = [];
								for (const $filter of document.querySelectorAll('#textEditTextFiltersDialog-textlist > div')) {
									switch ($filter.type) {
										case 'dropshadow':
											filters.text.push({
												type: 'dropshadow',
												color: $filter.querySelector('.dropshadowcolor').value,
												offset:  $filter.querySelector('.dropshadowoffset').value,
												blurradius: $filter.querySelector('.dropshadowblurradius').value
											});
											break;
										case 'outline':
											filters.text.push({
												type: 'outline',
												color: $filter.querySelector('.outlinecolor').value,
												width: $filter.querySelector('.outlinewidth').value
											});
											break;
									}
								}
								for (const $filter of document.querySelectorAll('#textEditTextFiltersDialog-inlineimageslist > div')) {
									switch ($filter.type) {
										case 'dropshadow':
											filters.inlineimages.push({
												type: 'dropshadow',
												color: $filter.querySelector('.dropshadowcolor').value,
												offset:  $filter.querySelector('.dropshadowoffset').value,
												blurradius: $filter.querySelector('.dropshadowblurradius').value
											});
											break;
									}
								}
								document.getElementById('textfilters').dispatchEvent(new Event('change'));
							};
							const createDropShadowItem = (dropShadowColor, dropShadowOffset, dropShadowBlurRadius) => {
								const $dropShadowItem = $.div({ class: 'dropshadowitem' },
									$.div({},
										$.span({ text: 'Drop Shadow' }),
										$.i({ title: 'Delete', class: 'mdi mdi-trash-can-outline tooltip-button', click: function() {
											this.closest('.dropshadowitem').remove();
											saveFilters();
										} })
									),
									$.div({ class: 'propertygrid' },
										$.label({ text: 'Color' }),
										$.uicolor({ class: 'dropshadowcolor', value: dropShadowColor, config: {
											showPalette: true,
											localStorageKey: 'globalcolor',
											previewChange: () => saveFilters(),
											opacity: true
										}, change: saveFilters, input: saveFilters }),
											
										$.label({ text: 'Offset' }),
										$.select({ class: 'dropshadowoffset', change: saveFilters, input: saveFilters, value: dropShadowOffset },
											...range(100).map(i => $.option({ text: i, value: i }))
										),
											
										$.label({ text: 'Blur Radius' }),
										$.select({ class: 'dropshadowblurradius', value: dropShadowBlurRadius, change: saveFilters, input: saveFilters },
											...range(100).map(i => $.option({ text: i, value: i }))
										)
									)
								);
								$dropShadowItem.type = 'dropshadow';
								return $dropShadowItem;
							};
							const createOutlineItem = (outlineColor, outlineWidth) => {
								const $outlineItem = $.div({ class: 'outlineitem' },
									$.div({},
										$.span({ text: 'Outline' }),
										$.i({ title: 'Delete', class: 'mdi mdi-trash-can-outline tooltip-button', click: function() {
											this.closest('.outlineitem').remove();
											saveFilters();
										} })
									),
									$.div({ class: 'propertygrid' },
										$.label({ text: 'Color' }),
										$outlineColor = $.uicolor({ class: 'outlinecolor', value: outlineColor, config: {
											showPalette: true,
											localStorageKey: 'globalcolor',
											previewChange: () => saveFilters(),
											opacity: true
										}, change: saveFilters }),
										$.label({ text: 'Width' }),
										$.select({ title: 'Outline Width', class: 'outlinewidth', value: outlineWidth, change: saveFilters, input: saveFilters },
											...range(99, 1).map(i => $.option({ text: i, value: i }))
										)
									)
								);
								$outlineItem.type = 'outline';
								return $outlineItem;
							};
							for (const $e of document.getElementById('text').querySelectorAll('.text-opentextfilters, .widget-dialog > div:last-child')) $e.style.pointerEvents = 'none';
							const removeInputBlocksOnTextDialog = () => { for (const $e of document.getElementById('text').querySelectorAll('.text-opentextfilters, .widget-dialog > div:last-child')) $e.style.pointerEvents = 'all'; }
							const copyOfFilters = JSON.parse(JSON.stringify(document.getElementById('textfilters').data));
							const $dialog = document.body.appendChild($.uidialog({ id: 'textEditTextFiltersDialog', css: { color: 'white', width: '750px' }, movable: { handle: ':scope > header' } },
								$.header({ text: 'Text and Inline Image Filters', css: { cursor: 'move' } }),
								$.div({ text: 'Text Filters (Note: Only one outline is supported and drop shadows are stacked top to bottom in the list).', css: { margin: '10px' } }),
								$.div({ css: { margin: '0 10px' } },
									$.button({ id: 'textEditTextFiltersDialog-addtextdropshadow', text: 'Add Drop Shadow', title: 'Add Drop Shadow', click: () => {
										document.getElementById('textEditTextFiltersDialog-textlist').appendChild(createDropShadowItem('rgba(0, 0, 0, 1)', 5, 5));
										saveFilters();
									} }),
									$.button({ id: 'textEditTextFiltersDialog-addtextoutline', text: 'Add Outline', title: 'Add Outline', css: { marginLeft: '5px' }, click: () => {
										document.getElementById('textEditTextFiltersDialog-textlist').appendChild(createOutlineItem('rgba(0, 0, 0, 1)', 5));
										saveFilters();
									} })
								),
								$.div({ id: 'textEditTextFiltersDialog-textlist', class: 'custom-propertylist', css: { margin: '10px' } },
									...(copyOfFilters.hasOwnProperty('text') ? copyOfFilters.text : []).map(filter => {
										switch (filter.type) {
											case 'dropshadow':
												return createDropShadowItem(filter.color, filter.offset, filter.blurradius);
												break;
											case 'outline':
												return createOutlineItem(filter.color, filter.width);
												break;
										}
									})
								),
								$.div({ text: 'Inline Image Filters', css: { margin: '10px' } }),
								$.div({ css: { margin: '0 10px' } },
									$.button({ id: 'textEditTextFiltersDialog-addinlineimagesdropshadow', text: 'Add Drop Shadow', title: 'Add Drop Shadow', click: function() {
										document.getElementById('textEditTextFiltersDialog-inlineimageslist').appendChild(createDropShadowItem('rgba(0, 0, 0, 1)', 5, 5));
									} })
								),
								$.div({ id: 'textEditTextFiltersDialog-inlineimageslist', class: 'custom-propertylist', css: { margin: '10px' } },
									...(copyOfFilters.hasOwnProperty('inlineimages') ? copyOfFilters.inlineimages : []).map(filter => {
										switch (filter.type) {
											case 'dropshadow':
												return createDropShadowItem(filter.color, filter.offset, filter.blurradius);
												break;
										}
									})
								),
								$.div({ class: 'panel-buttons' },
									$.button({ text: 'Cancel', click: function() {
										document.getElementById('textfilters').data = copyOfFilters;
										document.getElementById('textfilters').dispatchEvent(new Event('change'));
										$dialog.remove();
										removeInputBlocksOnTextDialog();
									} }),
									$.button({ text: 'OK', click: function() {
										saveFilters();
										$dialog.remove();
										removeInputBlocksOnTextDialog();
									} })
								)
							));
							centerOnScreen($dialog);
						} }),
						$.label({ title: 'Insert inline image at cursor', class: 'mdi mdi-image-plus', css: {
							fontSize: '22px',
							cursor: 'pointer',
							marginRight: '5px'
						}, click: () => {
							const $textText = document.getElementById('texttext');
							const caretPosition = $textText.selectionStart;
							let $textInsertInlineImageDialogImage;
							const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '750px' } },
								$.header({ text: 'Insert Inline Image' }),
								$.div({ class: 'formgrid' },
									$.label({ for: 'textInsertInlineImageDialogImage', text: 'Image' }),
									$textInsertInlineImageDialogImage = $.uitreeselect({
										id: 'textInsertInlineImageDialogImage',
										config: {
											tree: 'content',
											leafIcon: node => GetContentIcon(node),
											filter: node => node.contenttreetype == 3
										}
									}),
									$.label({ for: 'textInsertInlineImageDialogHeight', text: 'Height' }),
									$.input({ type: 'text', id: 'textInsertInlineImageDialogHeight', class: 'sampleInput', value: 100 }),
									$.label({ for: 'textInsertInlineImageDialogAlignment', text: 'Vertical Alignment' }),
									$.select({ id: 'textInsertInlineImageDialogAlignment', class: 'sampleSelect', value: 'middle' },
										$.option({ text: 'Top', value: 'top' }),
										$.option({ text: 'Middle', value: 'middle' }),
										$.option({ text: 'Bottom', value: 'bottom' })
									)
								),
								$.div({ class: 'panel-buttons' },
									$.button({ text: 'Cancel', click: function() {
										$dialog.remove();
									} }),
									$.button({ text: 'OK', click: async function() {
										const selectedNode = (await $textInsertInlineImageDialogImage.getSelected())[0];
										const image = selectedNode ? ContentNodeToMinimalImageContent(selectedNode) : null;
										if (image == null) {
											toastError('Select an image.');
										} else {
											let height = parseInt(document.getElementById('textInsertInlineImageDialogHeight').value, 10);
											height = isNaN(height) ? 100 : height;
											const alignment = document.getElementById('textInsertInlineImageDialogAlignment').value;
											const text = $textText.value;
											const textToAdd = `[image="${image.name}" height="${height}" align="${alignment}"]`;
											$textText.value = text.substring(0, caretPosition) + textToAdd + text.substring(caretPosition);
											$textText.dispatchEvent(new Event('change'));
											setTimeout(() => $textText.selectionStart = caretPosition + textToAdd.length, 0);
											$dialog.remove();
										}
									} })
								)
							));
							centerOnScreen($dialog);
						} }),
						$.input({ type: 'text', id: 'textfilters', hidden: true }),
						$.input({ type: 'text', id: 'textinlineimages', hidden: true }),
						$.label({ title: 'Insert metadata text reference at cursor', class: 'mdi mdi-code-brackets', hidden: !permissions['Metadata'], css: {
							fontSize: '22px',
							cursor: 'pointer'
						}, click: () => {
							const caretPosition = document.getElementById('texttext').selectionStart;
							const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '750px' } },
								$.header({ text: 'Insert Inline Text' }),
								$.div({ class: 'formgrid' },
									$.label({ for: 'textInsertMetadata-text', text: 'Metadata Name' }),
									$.input({ type: 'text', id: 'textInsertMetadata-text' }),
									$.label({ for: 'textInsertMetadata-default', text: 'Default Text (Optional)' }),
									$.input({ type: 'text', id: 'textInsertMetadata-default' })
								),
								$.div({ class: 'panel-buttons' },
									$.button({ text: 'Cancel', click: function() {
										$dialog.remove();
									} }),
									$.button({ text: 'OK', click: function() {
										const metadataName = document.getElementById('textInsertMetadata-text').value;
										const defaultValue = document.getElementById('textInsertMetadata-default').value;
										if (metadataName == '') {
											toastError('Enter a metadata property name.');
										} else {
											const text = document.getElementById('texttext').value;
											const textToAdd = `[metadata="${metadataName}"${defaultValue == '' ? '' : ` default="${defaultValue}"`}]`;
											document.getElementById('texttext').value = text.substring(0, caretPosition) + textToAdd + text.substring(caretPosition);
											document.getElementById('texttext').dispatchEvent(new Event('change'));
											setTimeout(() => document.getElementById('texttext').selectionStart = caretPosition + textToAdd.length, 0);
											$dialog.remove();
										}
									} })
								)
							));
							centerOnScreen($dialog);
						} })
					),
					$.textarea({ id: 'texttext', rows: 10 })
				);
				break;
			case 'webpage':
				Object.assign(widget.$dialog.style, {
					width: '60%',
					height: '80%'
				});
				widget.$dialog.append(
					$.ul({ class: 'widget-accordion', css: {
						flex: '1 0 0px',
						minHeight: '0px'
					} },
						$.li({ text: 'URL', class: 'widget-accordion-header' }),
						$.li({ css: {
							flex: '1 0 0px',
							minHeight: '0px',
							display: 'flex',
							flexFlow: 'column nowrap',
							padding: 0
						} },
							$.div({ css: {
								flex: '0 0 auto',
								display: 'flex',
								flexFlow: 'column nowrap'
							} },
								$.div({ class: 'formgrid' },
									$.label({ for: 'webpageurl', text: 'URL', css: { marginRight: '5px' } }),
									$.input({ type: 'text', id: 'webpageurl', maxlength: 512, css: { flex: '1 0 0px', minHeight: '0px', marginRight: '5px' } }),
									$.button({ text: 'Preview...', css: { marginRight: '5px', gridColumn: '3' }, click: () => {
										var url = document.getElementById('webpageurl').value;
										if (!/^http.*$/.test(url)) {
											toastError('Webpages must define their protocol. Enter either https:// or http:// before the URL. (Ideally use https:// if the website supports it).');
										} else {
											window.open(url, '', 'scrollbars=yes,menubar=no,height=768,width=1024,resizable=yes,toolbar=no,location=no,status=no');
										}
									} }),
									$.button({ id: 'webpage-add', text: 'Add Web Page', css: { gridColumn: '4' }, click: async () => {
										var webpage = document.getElementById('webpageurl').value;
										if (webpage == '') {
											toastError('Enter a webpage URL.');
										} else if (!/^http.*$/.test(webpage)) {
											toastError('Webpages must define their protocol. Enter either https:// or http:// before the URL. (Ideally use https:// if the website supports it).');
										} else {
											var videoId = checkYoutubeURL(webpage);
											if (videoId) {
												webpage = formatYoutubeURL(webpage, videoId);
											}
											document.getElementById('webpages').classList.add('busy');
											try {
												var data = await post('/post/addwebpage.php', {
													webpage: webpage
												});
												document.getElementById('webpages').classList.remove('busy');
												CheckSessionExpired(data);
												if (data.success) {
													RequestWebContent(webpage);
												}
											} catch (exception) {
												document.getElementById('webpages').classList.remove('busy');
												NetworkingError();
											}
										}
									} }),
									
									$.div({ class: 'formgrid-label', css: { display: 'flex', flexFlow: 'row nowrap', alignItems: 'center' } },
										$.label({ text: 'Live Webpage', title: 'Enabling this causes the webpage to be pulled directly from the site rather than generating a static screenshot. Note that live webpages must use https (secure) URLs.' }),
										$.i({ title: 'Click for live webpage information', class: 'tooltip-button mdi mdi-information-outline', css: { fontSize: '22px', margin: '0 5px' }, click: function() {
											const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '800px' } },
												$.header({ text: 'Live Webpage Information' }),
												$.ul({ id: 'customwebpages-directions' },
													$.li({ text: 'Live webpage information:' }),
													$.li({},
														$.ul({ css: { marginLeft: '25px', listStyle: 'disc' } },
															$.li({ text: 'Live webpages must use https (secure) URLs. That is they must be secure in order to run on the players. If you want to display a site using http and are fine with just a screenshot of the page leave this unchecked.' }),
															$.li({ text: 'Live webpages can render slightly differently on Android, Linux, and Windows players. What this means is the crop settings might need to be changed slightly after viewing the player to correct visual errors. (You might notice small changes in how text is rendered or wrapped on a page such that the height needs to be slightly smaller or larger to look right on the player).' })
														)
													)
												),
												$.div({ class: 'panel-buttons' },
													$.button({ text: 'OK', click: function() {
														$dialog.remove();
													} })
												)
											));
											centerOnScreen($dialog);
										} })
									),
									$.uicheckbox({ id: 'webpagelive' }),
									
									$.label({ for: 'webpagerefresh', text: 'Refresh Interval' }),
									$.select({ id: 'webpagerefresh', css: { margin: '1px 5px 5px 5px' } }),
									
									$.label({ for: 'webpagemuteonhideandreloadonshow', text: 'Mute on Hide and Reload on Show (Also Android Scroll Fix)', title: `If the URL is playing audio this option will mute it by unloading the page when the widget is hidden and then reload the page when it's shown. Note that depending on how long the zone duration is this could cause the page to reload a lot during the day. Be mindful of this. This also fixes a problem with interactive scroll on for Android device webpages.` }),
									$.uicheckbox({ id: 'webpagemuteonhideandreloadonshow' })
								)
							),
							$.div({ css: { flex: '1 0 0px', minHeight: '0px' }, tabs: {} },
								$.ul({},
									$.li({ text: 'Saved Webpages' }),
									$.li({ text: 'Custom Webpages' })
								),
								$.div({ class: 'customwebpages-tab-body' },
									$.ul({ class: 'iconlist' },
										$.li({ title: 'Delete Selected Web Page', class: 'tooltip-button mdi mdi-trash-can-outline', click: async () => {
											const $selectedWebpage = document.querySelector('#webpages > li.selected');
											if ($selectedWebpage != null) {
												if (await QuestionBox('Delete webpage?', 'Are you sure you want to delete the selected web page?')) {
													const id = $selectedWebpage.webpageid;
													document.getElementById('webpages').classList.add('busy');
													try {
														var data = await post('/post/deleteWebpage.php', {
															id: id
														});
														document.getElementById('webpages').classList.remove('busy');
														CheckSessionExpired(data);
														if (data.success) {
															RequestWebContent();
														}
													} catch (exception) {
														document.getElementById('webpages').classList.remove('busy');
														NetworkingError();
													}
												}
											}
										} }),
										$.li({ class: 'spacer' })
									),
									$.ul({ id: 'webpages', css: { flex: '1 0 0px', minHeight: '0px' } })
								),
								$.div({ class: 'customwebpages-tab-body' },
									$.input({ type: 'file', id: 'customwebpage-file', hidden: true }),
									$.ul({ id: 'customwebpages-options', class: 'iconlist' },
										$.li({ title: 'Delete selected custom webpage', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => document.getElementById('customwebpages').deleteSelected() }),
										$.li({ class: 'spacer' }),
										$.li({ title: 'Add group', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => document.getElementById('customwebpages').addGroup() }),
										$.li({ title: 'New custom webpage', class: 'tooltip-button mdi mdi-playlist-plus', click: async function() {
											const $customWebpages = document.getElementById('customwebpages');
											let selectedNode = (await $customWebpages.getSelected())[0];
											if (selectedNode == undefined) {
												toastError('Select a group to add the new webpage to.');
												return;
											}
											if (!selectedNode.group && selectedNode.parentNode) {
												selectedNode = selectedNode.parentNode;
												await $customWebpages.setSelected([selectedNode.id]);
											}
											if (!selectedNode.group) {
												toastError('Invalid add permissions for parent group.');
												return;
											}
											document.getElementById('customwebpage-file').replaceWith(
												$.input({ type: 'file', id: 'customwebpage-file', hidden: true, change: async function() {
													if (this.files.length == 0) {
														toastError('Select a zip file to upload.');
													} else {
														await InputBox('Webpage name', 'Enter a name for the webpage.', 'Ok', 'Cancel', '', async name => {
															if (!IsAlphaNumericSpacesHyphensUnderscoresApostrophesNotEmptyLessThan60Characters(name)) {
																toastError('The webpage name can only container letters, numbers, spaces, hyphens, underscores, or apostrophes and must be between 1 and 40 characters.');
																return false;
															}
															
															const upload = async (overwrite, id = -1) => {
																const data = await $customWebpages.addItems([{
																	type: 2,
																	id,
																	name: name,
																	parent: selectedNode.id,
																	file: this.files[0],
																	overwrite: overwrite,
																	progress: (node, loaded, total) => {
																		document.getElementById('status').innerText = `Uploading webpage: ${(loaded / total * 100).toFixed(2)}%.`;
																		StartClearStatusTimer();
																	}
																}]);
																if (data.success) {
																	toastSuccess('Custom webpage uploading.');
																	return true;
																} else if (data.errors.length != 0 && data.errors[0].hasOwnProperty('overwrite')) {
																	if (await QuestionBox('Overwrite?', 'A custom webpage with that name already exist. Do you want to overwrite?')) {
																		return await upload(true, data.errors[0].id);
																	} else {
																		return false;
																	}
																} else {
																	return false;
																}
															};
															return await upload(false);
														}, () => {
															this.value = '';
															return true;
														}, null);
													}
												} })
											);
											document.getElementById('customwebpage-file').click();
										} }),
										$.li({ title: 'Directions', class: 'tooltip-button mdi mdi-information-outline', click: () => {
											const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '800px' } },
												$.header({ text: 'Custom Webpage Directions' }),
												$.ul({ id: 'customwebpages-directions' },
													$.li({ text: 'This custom webpage tool allows one to upload and host a webpage on this server.' }),
													$.li({},
														$.ul({ css: { marginLeft: '25px', listStyle: 'disc' } },
															$.li({ text: 'The upload format is a zip archive with an index.html file at the top level of the archive along with the content. (Be careful not to have the index.html in a folder. Create the archive by selecting the index.html and content and then creating the zip rather than zipping the folder they exist in).' }),
															$.li({ text: 'All references to content must be relative to the index.html file. So for instance if you reference images/example.jpg then in your zip archive there should be a folder called images with the example.jpg file in it.' }),
															$.li({ text: 'Depending on the size of the uploaded files this may take a few minutes. When the upload is complete it will say the upload succeeded or tell you about any errors encountered.' }),
															$.li({ text: 'After you have uploaded your files you must create a playlist and schedule with that playlist before you may send it to the player.' })
														)
													)
												),
												$.div({ class: 'panel-buttons' },
													$.button({ text: 'OK', click: function() {
														$dialog.remove();
													} })
												)
											));
											centerOnScreen($dialog);
										} })
									),
									$.uitree({ id: 'customwebpages', css: { flex: '1 0 0px', minHeight: '0px' }, config: {
										tree: 'customwebpage',
										editable: node => node.group && node.parentNode != null,
										deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
											if (selectedNodes.length == 0) {
												toastError('Select a custom webpage to delete.');
												return;
											}
											var selectedNode = selectedNodes[0];
											if (selectedNode.group) {
												var customWebpageCount = selectedAllChildren.filter(node => !node.group).length;
												if (customWebpageCount != 0) {
													await ConfirmationBox("Remove group?", `Are you sure you want to remove the group '${selectedNode.name}' which has ${customWebpageCount} custom webpage${customWebpageCount == 1 ? '' : 's'}?`, 'Yes', 'No');
												}
											} else {
												await ConfirmationBox("Remove custom webpage?", `Are you sure you want to remove the custom webpage '${selectedNode.name}'?`, 'Yes', 'No');
											}
											await deleteNodes(selectedAllChildren);
										}
									}, change: async function(e) {
										const selectedNode = (await this.getSelected())[0];
										if (!selectedNode?.group) {
											document.getElementById('webpageurl').value = selectedNode.url;
											document.getElementById('webpageurl').dispatchEvent(new Event('change'));
											if (!document.getElementById('webpagelive').checked) {
												if (await QuestionBox('Enable live webpage?', 'You have selected a custom webpage which can and should use the live webpage mode. Do you want to enable live webpage?')) {
													document.getElementById('webpagelive').checked = true;
													document.getElementById('webpagelive').dispatchEvent(new Event('change'));
												}
											}
										}
									} })
								)
							)
						),
						$.li({ id: 'webpage-crop', text: 'Crop', class: 'widget-accordion-header' }),
						$.li({ css: {
							display: 'flex',
							flexFlow: 'row nowrap',
							padding: 0
						} },
							$.div({ css: {
								flex: '0 0 auto',
								display: 'flex',
								flexFlow: 'column nowrap',
								overflowY: 'auto',
								backgroundColor: 'var(--formgrid-background-color)',
								color: 'var(--formgrid-font-color)',
								padding: '10px'
							} },
								$.div({ text: 'Viewport', title: 'The render width and height are how large the server should render the screenshot of the webpage. Set this before setting the crop rectangle. The crop rectangle must be inside of the render width and height. The webpage on the right might take a few seconds to load or resize.', class: 'tooltip-help' },
									$.i({ class: 'mdi mdi-information-outline', css: {
										margin: '0 10px',
										verticalAlign: 'middle'
									} })
								),
								$.div({ css: {
									flex: '0 0 auto',
									display: 'flex',
									flexFlow: 'column nowrap',
									padding: '5px 5px 10px 10px'
								} },
									$.label({ for: 'webpagerenderwidth', text: 'Render Width:' }),
									$.input({ type: 'text', id: 'webpagerenderwidth', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagerenderheight', text: 'Render Height:' }),
									$.input({ type: 'text', id: 'webpagerenderheight', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagescrollleft', text: 'Scroll Left:' }),
									$.input({ type: 'text', id: 'webpagescrollleft', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagescrolltop', text: 'Scroll Top:' }),
									$.input({ type: 'text', id: 'webpagescrolltop', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } })
								),
								$.div({ text: 'Crop', title: 'You can change the crop rectangle on the right by dragging a rectangle. Note that if crop width and crop height do not match the media width and height then the area will be stretched to fit the media width and height. (You can use the fill style property in the media properties to maintain aspect ratio).', class: 'tooltip-help' },
									$.i({ class: 'mdi mdi-information-outline', css: {
										margin: '0 10px',
										verticalAlign: 'middle'
									} })
								),
								$.div({ css: {
									display: 'flex',
									flexFlow: 'column nowrap',
									padding: '5px 5px 10px 10px'
								} },
									$.label({ for: 'webpagecropx', text: 'Left:' }),
									$.input({ type: 'text', id: 'webpagecropx', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagecropy', text: 'Top:' }),
									$.input({ type: 'text', id: 'webpagecropy', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagecropwidth', text: 'Width:' }),
									$.input({ type: 'text', id: 'webpagecropwidth', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
									$.label({ for: 'webpagecropheight', text: 'Height:' }),
									$.input({ type: 'text', id: 'webpagecropheight', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } })
								)
							),
							widget.$webpageCrop = $.div({ css: {
								flex: '1 0 0px',
								minHeight: '0px',
								overflow: 'auto',
								position: 'relative',
								border: '1px solid white',
								marginBottom: 0
							} },
								widget.$webpageCropPreview = $.iframe({ sandbox: 'allow-scripts allow-same-origin', css: { border: 'none', position: 'absolute', left: 0, top: 0, pointerEvents: 'none' } }),
								widget.$webpageCropBox = $.canvas({ id: 'webpageCropBox', css: {
									position: 'absolute',
									left: 0,
									right: 0
								} })
							)
						)
					)
				);
				RequestWebContent();
				for (const $e of document.querySelectorAll('#customwebpages-directions, #customwebpages-options')) $e.hidden = !permissions['Custom Webpage'];
				break;
			case 'pdf':
				Object.assign(widget.$dialog.style, {
					width: '60%',
					height: '80%'
				});
				widget.$dialog.append(
					$.li({ css: {
						flex: '1 0 0px',
						display: 'flex',
						flexFlow: 'row nowrap',
						minHeight: '300px'
					} },
						$.div({ css: {
							flex: '0 0 auto',
							display: 'flex',
							flexFlow: 'column nowrap',
							padding: '10px',
							overflowY: 'auto',
							backgroundColor: 'var(--formgrid-background-color)',
							color: 'var(--formgrid-font-color)'
						} },
							$.div({ css: {
								flex: '0 0 auto',
								display: 'flex',
								flexFlow: 'row nowrap',
								alignItems: 'center',
								marginBottom: '10px'
							} },
								$.label({ for: 'pdfpdf', text: 'PDF', title: 'PDF file to use', class: 'tooltip', css: { marginRight: '5px' } }),
								$.uitreeselect({ id: 'pdfpdf', css: { width: '150px' }, config: {
									tree: 'content',
									leafIcon: node => GetContentIcon(node),
									filter: node => node.contenttreetype == 9
								} })
							),
							$.div({ text: 'Viewport', title: 'The render width and height are how large the pdf viewport should be. Set this before setting the crop rectangle. The crop rectangle must be inside of the render width and height. The PDF viewer on the right might take a few seconds to load or resize. The scale feature has a hard limit to how large the pages can be scaled (1920px).', class: 'tooltip-help' },
								$.i({ class: 'mdi mdi-information-outline', css: {
									margin: '0 10px',
									verticalAlign: 'middle'
								} })
							),
							$.div({ css: {
								flex: '0 0 auto',
								display: 'flex',
								flexFlow: 'column nowrap',
								padding: '5px 5px 10px 10px'
							} },
								$.label({ for: 'pdfviewportwidth', text: 'Viewport Width:' }),
								$.input({ type: 'text', id: 'pdfviewportwidth', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfviewportheight', text: 'Viewport Height:' }),
								$.input({ type: 'text', id: 'pdfviewportheight', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfscrollleft', text: 'Scroll Left:' }),
								$.input({ type: 'text', id: 'pdfscrollleft', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfscrolltop', text: 'Scroll Top:' }),
								$.input({ type: 'text', id: 'pdfscrolltop', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfpagenumber', text: 'Scroll to Page Number:' }),
								$.select({ id: 'pdfpagenumber', class: 'sampleSelect', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfscale', text: 'Scale:' }),
								$.select({ id: 'pdfscale', class: 'sampleSelect', css: { margin: '1px 5px 5px 5px' } })
							),
							$.div({ text: 'Crop', title: 'You can change the crop rectangle on the right by dragging a rectangle. Note that if crop width and crop height do not match the media width and height then the area will be stretched to fit the media width and height. (You can use the fill style property in the media properties to maintain aspect ratio).', class: 'tooltip-help' },
								$.i({ class: 'mdi mdi-information-outline', css: {
									margin: '0 10px',
									verticalAlign: 'middle'
								} })
							),
							$.div({ css: {
								flex: '0 0 auto',
								display: 'flex',
								flexFlow: 'column nowrap',
								padding: '5px 5px 10px 10px'
							} },
								$.label({ for: 'pdfcropx', text: 'Left:' }),
								$.input({ type: 'text', id: 'pdfcropx', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfcropy', text: 'Top:' }),
								$.input({ type: 'text', id: 'pdfcropy', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfcropwidth', text: 'Width:' }),
								$.input({ type: 'text', id: 'pdfcropwidth', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } }),
								$.label({ for: 'pdfcropheight', text: 'Height:' }),
								$.input({ type: 'text', id: 'pdfcropheight', size: 4, maxlength: 4, class: 'sampleInput', css: { margin: '1px 5px 5px 5px' } })
							),
							$.div({ class: 'formgrid', css: { marginTop: 0 } },
								$.label({ for: 'pdfallowscroll', text: 'Allow Interactive Scroll', title: 'On touch displays whether this PDF should be allowed to scroll' }),
								$.uicheckbox({ id: 'pdfallowscroll' })
							)
						),
						widget.$pdfCrop = $.div({ css: {
							flex: '1 0 0px',
							minHeight: '0px',
							overflow: 'auto',
							position: 'relative',
							border: '1px solid white',
							marginBottom: 0
						} },
							widget.$pdfCropPreview = $.div({ css: { position: 'absolute', left: 0, top: 0, pointerEvents: 'none' } }),
							widget.$pdfCropBox = $.canvas({ id: 'pdfCropBox', css: {
								position: 'absolute',
								left: 0,
								right: 0
							} })
						)
					)
				);
				break;
		}
		
		for (const attribute of widget.attributes) {
			if (!attribute.hasOwnProperty('$input')) {
				attribute.$input = document.getElementById(`${widget.name}${attribute.name}`);
			}
			Object.defineProperty(attribute, 'value',
			{
				get: function() {
					return new Promise(async resolve => {
						switch (this.type) {
							case 'bool':
								resolve(this.$input.checked);
								break;
							case 'int':
								if (this.hasOwnProperty('groupedbutton')) {
									resolve(parseInt(this.$input.querySelector(':radio:checked').value, 10));
								} else if (this.hasOwnProperty('design') && this.design == 'idledelay') {
									resolve(this.$input.value);
								} else {
									var value = this.$input.value == undefined ? this.default : parseInt(this.$input.value, 10);
									if (isNaN(value)) {
										value = this.default;
									}
									if (this.hasOwnProperty('minimum')) {
										value = Math.max(value, this.minimum);
									}
									if (this.hasOwnProperty('maximum')) {
										value = Math.min(value, this.maximum);
									}
									resolve(value);
								}
								break;
							case 'float':
								if (this.hasOwnProperty('groupedbutton')) {
									resolve(parseFloat(this.$input.querySelector(':scope > :checked').value));
								} else {
									var value = this.$input.value == undefined ? this.default : parseFloat(this.$input.value);
									if (isNaN(value)) {
										value = this.default;
									}
									if (this.hasOwnProperty('minimum')) {
										value = Math.max(value, this.minimum);
									}
									if (this.hasOwnProperty('maximum')) {
										value = Math.min(value, this.maximum);
									}
									resolve(value);
								}
								break;
							case 'string':
								if (this.hasOwnProperty('design') && this.design == 'rssfeed') {
									const selectedNode = (await this.$input.getSelected())[0];
									resolve(selectedNode == undefined || selectedNode.id == 0 ? 'useurl' : selectedNode.name);
								} else if (this.hasOwnProperty('design') && this.design == 'csvfile') {
									const selectedNode = (await this.$input.getSelected())[0];
									resolve(selectedNode == undefined || selectedNode.id == 0 ? 'useurl' : selectedNode.name);
								} else if (this.hasOwnProperty('design') && this.design == 'eventlist') {
									resolve((await this.$input.getSelected())[0]?.name ?? '');
								} else if (this.hasOwnProperty('groupedbutton')) {
									resolve(this.$input.querySelector(':scope > :checked').value);
								} else if (this.hasOwnProperty('design') && this.design == 'fontfamily') {
									resolve(this.$input.value == '' ? this.default : this.$input.value);
								} else if (this.hasOwnProperty('design') && (this.design == 'powerbireportfilters' || this.design == 'powerbipagefilters' || this.design == 'powerbivisualfilters')) {
									const checked = await this.$input.getCheckedAllChildren();
									resolve(checked.map(node => node.name).join(','));
								} else {
									resolve(this.$input.value == undefined || this.$input.value == null ? this.default : this.$input.value);
								}
								break;
							case 'color':
								if (this.$input.value == 'rgba(255, 255, 255, 0)') {
									resolve('');
								} else {
									resolve(this.$input.value);
								}
								break;
							case 'date':
								resolve(moment(this.$input.value, 'YYYY-MM-DD').format('MM/DD/YYYY'));
								break;
							case 'time':
								resolve(moment(this.$input.value, 'HH:mm').format('hh:mma'));
								break;
							case 'json':
								if (this.hasOwnProperty('design') && this.design == 'media') {
									const selectedNode = (await this.$input.getSelected())[0];
									resolve(selectedNode ? ContentNodeToMinimalImageContent(selectedNode) : null);
								} else if (attribute.hasOwnProperty('design') && attribute.design == 'networkcommands') {
									resolve((await this.$input.getCheckedAllChildren()).filter(node => node.type == 2).map(node => node.networkcommandid));
								} else if (attribute.hasOwnProperty('design') && attribute.design == 'apicommands') {
									resolve((await this.$input.getCheckedAllChildren()).map(node => node.id));
								} else {
									resolve(JSON.parse(JSON.stringify(await this.$input.data)));
								}
								break;
							default:
								reject(new Error('attribute type needs to be handled'));
								break;
						}
					});
				},
				set: async function(value) {
					switch (this.type) {
						case 'bool':
							this.$input.checked = value == null ? true : value;
							this.$input.indeterminate = value == null;
							break;
						case 'string':
							if (this.hasOwnProperty('design') && this.design == 'rssfeed') {
								await this.$input.setSelected(node => node.hasOwnProperty('value') && value.replace('.rss', '') == node.value || node.name == value.replace('.rss', ''));
							} else if (this.hasOwnProperty('design') && this.design == 'csvfile') {
								await this.$input.setSelected(node => node.hasOwnProperty('value') && value.replace('.csv', '') == node.value || node.name == value.replace('.csv', ''));
							} else if (this.hasOwnProperty('design') && this.design == 'eventlist') {
								await this.$input.setSelected(node => node.name == value);
							} else if (this.hasOwnProperty('groupedbutton')) {
								this.$input.querySelector(`input[value="${value}"]`).checked = true;
							} else if (this.hasOwnProperty('design') && this.design == 'fontfamily') {
								this.$input.value = value == 'default' ? defaultFont : (value == null ? '' : value);
							} else if (this.design == 'powerbireportfilters' || this.design == 'powerbipagefilters' || this.design == 'powerbivisualfilters') {
								if (value == '') {
									this.$input.setChecked(() => false);
								} else {
									for (let i = 0; i < 3; ++i) {
										try {
											if (gCurrentWidget == null) break;
											await gCurrentWidget.widget.loaded;
											await this.$input.reload();
											await this.$input.setChecked(node => value.split(',').map(name => name.trim()).includes(node.name));
											break;
										} catch (exception) {
											console.log(exception);
											await new Promise(resolve => setTimeout(resolve, 2000));
										}
									}
								}
							} else if (this.hasOwnProperty('options') && this.options == 'button') {
								this.$input.innerHTML = '';
								this.$input.appendChild($.option({ text: 'None', value: 'none' }));
								let firstButtonGroup = true;
								for (const [media, zone, buttons] of GetAllButtonsGrouped()) {
									if (!firstButtonGroup) {
										this.$input.appendChild($.option({ text: '-------------', disabled: true }));
									}
									firstButtonGroup = false;
									for (const button of buttons) {
										this.$input.appendChild($.option({ text: `${button.name}${button.default ? ' (Default)' : ''}`, value: button.name }));
									}
								}
								this.$input.value = value == null ? '' : value;
							} else {
								this.$input.value = value == null ? '' : value;
							}
							break;
						case 'int':
							if (this.hasOwnProperty('design') && this.design == 'idledelay') {
								this.$input.value = value;
								break;
							}
						case 'float':
							if (this.hasOwnProperty('groupedbutton')) {
								this.$input.querySelector(`input[value="${value}"]`).checked = true;
							} else {
								this.$input.value = value == null ? '' : value;
							}
							break;
						case 'color':
							this.$input.value = value == null ? '' : (value == '' ? 'rgba(255, 255, 255, 0)' : value);
							break;
						case 'date':
							if (value == '01/01/2017') { // This is a magic number in widgets.json that indicates to use the current date
								value = moment().format('YYYY-MM-DD');
							}
							if (/^[0-9]{1,2}\/[0-9]{1,2}\/[0-9]{4}$/.test(value)) {
								value = moment(value, 'MM/DD/YYYY').format('YYYY-MM-DD')
							}
							this.$input.value = value == null ? '' : value;
							break;
						case 'time':
							if (/^[0-9]{1,2}:[0-9]{1,2}(am|pm)$/.test(value)) {
								value = moment(value, 'hh:mma').format('HH:mm');
							}
							this.$input.value = value == null ? '' : value;
							break;
						case 'json':
							if (this.hasOwnProperty('design') && this.design == 'media') {
								await this.$input.setSelected(node => value != null && node.type == value.type && node.contentid == value.contentid);
							} else if (this.hasOwnProperty('design') && this.design == 'networkcommands') {
								await this.$input.setChecked(node => node.type == 2 && value.includes(node.networkcommandid));
							} else if (this.hasOwnProperty('design') && this.design == 'apicommands') {
								await this.$input.setChecked(node => value.includes(node.id));
							} else if (this.hasOwnProperty('design') && this.design == 'powerbireportfiltervalues') {
								this.$input.data = value;
								this.$input.renderFilters();
							} else {
								this.$input.data = value;
							}
							break;
						default:
							throw 'attribute type needs to be handled';
					}
				}
			});
			attribute.toggle = state => {
				if (attribute.hasOwnProperty('$label')) {
					attribute.$label.hidden = !state;
				}
				if (attribute.hasOwnProperty('$inputContainer')) {
					attribute.$inputContainer.hidden = !state;
				}
				if (attribute.hasOwnProperty('$metadataToggle')) {
					attribute.$metadataToggle.hidden = !(state && permissions['Metadata']);
				}
				if (attribute.hasOwnProperty('$metadataLabel') && attribute.hasOwnProperty('$metadataInput')) {
					attribute.$metadataLabel.hidden = !(state && attribute.$metadataInput.value != '');
				}
				if (attribute.hasOwnProperty('$metadataInput')) {
					attribute.$metadataInput.hidden = !(state && attribute.$metadataInput.value != '');
				}
				if (attribute.hasOwnProperty('$generatedMetadataToggle')) {
					attribute.$generatedMetadataToggle.hidden = !(state && permissions['Metadata']);
				}
				if (attribute.hasOwnProperty('$generatedMetadataLabel') && attribute.hasOwnProperty('$generatedMetadataInput')) {
					attribute.$generatedMetadataLabel.hidden = !(state && attribute.$generatedMetadataInput.value != '');
				}
				if (attribute.hasOwnProperty('$generatedMetadataInput')) {
					attribute.$generatedMetadataInput.hidden = !(state && attribute.$generatedMetadataInput.value != '');
				}
			};
		}
		
		for (const attribute of widget.attributes) {
			var buildOptions = () => {
				var options = attribute.options;
				if (!Array.isArray(options)) {
					switch (options) {
						case 'fontsize':
							options = [];
							for (var i = 6; i < 401; ++i) {
								options.push({ name: i, value: i });
							}
							break;
						case 'lineheight':
							options = [{ name: 'Default', value: -1 }];
							for (var i = 6; i < 401; ++i) {
								options.push({ name: i, value: i });
							}
							break;
						case 'providers':
							options = [
								{ name: 'Microsoft Office 365 Exchange', value: 'microsoft office 365 exchange' },
								{ name: 'Microsoft Server Exchange', value: 'microsoft exchange' },
								{ name: 'Google Calendar', value: 'google calendar' },
								{ name: 'Delphi EMS', value: 'delphi ems' },
								{ name: 'MICROS Fidelio Opera PMS', value: 'micros fidelio opera pms' },
								{ name: 'MICROS Opera Hotel Edition', value: 'micros opera hotel edition' },
								{ name: 'Dean Evans EMS', value: 'dean evans ems' },
								{ name: 'EMS Platform Services', value: 'ems platform services' },
								{ name: 'Ad Astra', value: 'ad astra' },
								{ name: 'Hyatt Readerboard', value: 'hyatt' },
								{ name: '25 Live', value: '25live' },
								{ name: 'Civic Engage 6', value: 'civicengage 6' },
								{ name: 'Nexudus', value: 'nexudus' },
								{ name: 'Events', value: 'events' },
								{ name: 'LibraryMarket', value: 'librarymarket' },
								{ name: 'iCalendar', value: 'icalendar' }
							];
							break;
						case 'alignment':
							options = [
								{ name: 'Left', value: 'left' },
								{ name: 'Center', value: 'center' },
								{ name: 'Right', value: 'right' }
							];
							break;
						case 'fillstyle':
							options = [
								{ name: 'None', value: 1 },
								{ name: 'Cover', value: 2 },
								{ name: 'Aspect Ratio', value: 3 }
							];
							break;
						case 'eventlistcolumn':
							options = [
								{ name: 'None', value: 'none' },
								{ name: 'Title', value: 'title' },
								{ name: 'Owner/Organizer/Booker', value: 'name' },
								{ name: 'Room', value: 'room' },
								{ name: 'Start Date Time - End Date Time', value: 'time' },
								{ name: 'Start Date', value: 'startdate' },
								{ name: 'Start Time', value: 'starttime' },
								{ name: 'Hyphen', value: 'hyphen' },
								{ name: 'End Date', value: 'enddate' },
								{ name: 'End Time', value: 'endtime' }
							];
							break;
						case 'borderthickness':
							options = [];
							for (var i = 0; i <= 10; ++i) {
								options.push({ name: i, value: i });
							}
							break;
						case 'borderradius':
							options = [];
							for (var i = 0; i <= 50; ++i) {
								options.push({ name: i, value: i });
							}
							break;
						case 'button':
							// Handled in the set function to rebuild every time
							options = [];
							break;
					}
				}
				if (Array.isArray(options)) {
					if (attribute.hasOwnProperty('groupedbutton')) {
						for (const option of options) {
							const optionId = `${widget.name}${attribute.name}${option.value}`;
							attribute.$input.append(
								$.input({ type: 'radio', id: optionId, value: option.value, name: `${widget.name}${attribute.name}` }),
								$.label({ for: optionId, class: [option.hasOwnProperty('icon'), `mdi mdi-${option.icon}`], text: option.hasOwnProperty('icon') ? '' : option.name })
							);
						}
						attribute.$input.classList.add('sampleGroupedButtons');
					} else {
						for (const option of options) {
							attribute.$input.appendChild($.option({ text: option.name, value: option.value }));
						}
						attribute.$input.classList.toggle('sampleSelect', !attribute.hasOwnProperty('view') || attribute.view != 'property');
					}
				} else {
					throw 'Options list not defined';
				}
			};
			switch (attribute.type) {
				case 'bool':
					attribute.$input.checked = attribute.default;
					break;
				case 'string':
					if (attribute.hasOwnProperty('design') && (attribute.design == 'rssfeed' || attribute.design == 'csvfile' || attribute.design == 'eventlist')) {
						// Do Nothing
					} else if (attribute.hasOwnProperty('options')) {
						buildOptions();
					} else if (attribute.hasOwnProperty('design') && attribute.design == 'fontfamily') {
						attribute.$inputContainer = attribute.$input;
					} else if (attribute.hasOwnProperty('password')) {
						attribute.$input.type = 'password';
						attribute.$input.placeholder = attribute.hasOwnProperty('required') ? 'Required field' : '';
					} else if (attribute.hasOwnProperty('design') && (attribute.design == 'powerbireportfilters' || attribute.design == 'powerbipagefilters' || attribute.design == 'powerbivisualfilters')) {
						attribute.$input.setChecked([]);
					}
					break;
				case 'int':
					if (attribute.hasOwnProperty('options')) {
						buildOptions();
					}
					break;
				case 'float':
					if (attribute.hasOwnProperty('options')) {
						buildOptions();
					}
					break;
				case 'color':
					attribute.$input.value = attribute.default == '' ? 'rgba(255, 255, 255, 0)' : attribute.default;
					break;
				case 'date':
					attribute.$input.type = 'date';
					break;
				case 'time':
					attribute.$input.type = 'time';
					break;
				case 'json':
					if (attribute.hasOwnProperty('design') && attribute.design == 'media') {
						attribute.$input.setSelected([]);
					} else {
						attribute.$input.type = 'text';
						attribute.$input.data = JSON.parse(JSON.stringify(attribute.default));
					}
					break;
				default:
					throw 'attribute type needs to be handled';
			}
		}
		
		SetWidgetDefaults(widget);
		
		widget.$dialog.appendChild(
			$.div({ class: 'panel-buttons' },
				widget.$cancel = $.button({ text: 'Cancel' }),
				widget.$ok = $.button({ text: 'OK' })
			)
		);
		
		const toggle = (attributes, state) => {
			for (const attribute of attributes) {
				attribute.toggle(state);
			}
			movableConstrainInWindow(widget.$dialog);
		};
		
		switch (widget.name) {
			case 'rsscrawl':
				for (const attribute of [attr.filter, attr.rssfeed]) {
					attribute.$input.addEventListener('change', async () => {
						const selectedNode = (await attr.rssfeed.$input.getSelected())[0];
						const useURL = selectedNode != undefined && selectedNode.hasOwnProperty('value') && selectedNode.id == 0;
						if (useURL || !attr.filter.$input.checked) {
							attr.start.$input.value = moment().format('YYYY-MM-DD');
							attr.end.$input.value = moment().format('YYYY-MM-DD');
						}
						
						attr.url.toggle(useURL);
						attr.filter.toggle(!useURL);
						attr.start.toggle(!useURL && attr.filter.$input.checked);
						attr.end.toggle(!useURL && attr.filter.$input.checked);
					});
				}
				break;
			case 'webpage':
				var webpageCropTimeout = null;
				
				var webpageRenderCrop = (cropX, cropY, cropWidth, cropHeight) => {
					var clientWidth = widget.$webpageCrop.clientWidth;
					var clientHeight = widget.$webpageCrop.clientHeight;
					if (clientWidth != 0 && clientHeight != 0 &&
					    (widget.$webpageCropBox.width != clientWidth || widget.$webpageCropBox.height != clientHeight)) {
						widget.$webpageCropBox.width = clientWidth;
						widget.$webpageCropBox.height = clientHeight;
					}
					
					var context = widget.$webpageCropBox.getContext('2d');
					context.clearRect(0, 0, widget.$webpageCropBox.width, widget.$webpageCropBox.height);
					context.save();
					
					context.translate(-widget.$webpageCrop.scrollLeft, -widget.$webpageCrop.scrollTop);
					context.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-fill-color');
					context.fillRect(cropX, cropY, cropWidth, cropHeight);
					context.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-stroke-color');
					context.translate(0.5, 0.5);
					context.strokeRect(cropX, cropY, cropWidth, cropHeight);
					context.restore();
				};
				
				for (const attribute of [attr.url, attr.renderwidth, attr.renderheight, attr.scrollleft, attr.scrolltop, attr.live]) {
					const inputChange = async () => {
						var attributes = await GetWidgetAttributes(widget);
						widget.$webpageCropPreview.width = attributes.renderwidth;
						widget.$webpageCropPreview.height = attributes.renderheight;
						widget.$webpageCropPreview.src = 'about:blank';
						
						clearTimeout(webpageCropTimeout);
						webpageCropTimeout = setTimeout(async () => {
							var attributes = await GetWidgetAttributes(widget);
							if (attributes.url) {
								widget.$webpageCropPreview.style.border = 'none';
								
								if (attributes.live) {
									widget.$webpageCropPreview.src = attributes.url;
								} else {
									widget.$webpageCropPreview.src = `/screenshots?url=${encodeURIComponent(attributes.url)}&width=${attributes.renderwidth}&height=${attributes.renderheight}&scrolltop=${attributes.scrolltop}`;
								}
								
								webpageRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
							}
						}, 2500);
					};
					attribute.$input.addEventListener('input', inputChange);
					attribute.$input.addEventListener('change', inputChange);
				}
				for (const attribute of [attr.cropx, attr.cropy, attr.cropwidth, attr.cropheight]) {
					const inputChange = async () => {
						const attributes = await GetWidgetAttributes(widget);
						webpageRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
					};
					attribute.$input.addEventListener('input', inputChange);
					attribute.$input.addEventListener('change', inputChange);
				}
				
				document.getElementById('webpage-crop').addEventListener('click', () => {
					widget.$webpageCrop.scroll();
				});
				var startCropX;
				var startCropY;
				draggable(widget.$webpageCrop, {
					handle: '#webpageCropBox',
					scrollParent: widget.$webpageCrop,
					start: (e, startX, startY, startOffsetX, startOffsetY) => {
						startCropX = Math.floor(startOffsetX + widget.$webpageCrop.scrollLeft);
						startCropY = Math.floor(startOffsetY + widget.$webpageCrop.scrollTop);
						attr.cropx.$input.value = startCropX;
						attr.cropy.$input.value = startCropY;
						attr.cropwidth.$input.value = 1;
						attr.cropheight.$input.value = 1;
						attr.cropx.$input.dispatchEvent(new Event('change'));
					},
					drag: (e, startX, startY, startOffsetX, startOffsetY) => {
						const webpageCropBoxOffset = $.offset(widget.$webpageCropBox);
						var relativeX = Math.max(0, Math.floor(e.pageX - webpageCropBoxOffset.left + widget.$webpageCrop.scrollLeft));
						var relativeY = Math.max(0, Math.floor(e.pageY - webpageCropBoxOffset.top + widget.$webpageCrop.scrollTop));
						var cropX = Math.min(startCropX, relativeX);
						var cropY = Math.min(startCropY, relativeY);
						var cropWidth = Math.max(1, Math.abs(relativeX - startCropX));
						var cropHeight = Math.max(1, Math.abs(relativeY - startCropY));
						if (cropX + cropWidth > parseInt(attr.renderwidth.$input.value, 10)) {
							cropWidth = parseInt(attr.renderwidth.$input.value, 10) - cropX;
						}
						if (cropY + cropHeight > parseInt(attr.renderheight.$input.value, 10)) {
							cropHeight = parseInt(attr.renderheight.$input.value, 10) - cropY;
						}
						attr.cropx.$input.value = cropX;
						attr.cropy.$input.value = cropY;
						attr.cropwidth.$input.value = cropWidth;
						attr.cropheight.$input.value = cropHeight;
						attr.cropx.$input.dispatchEvent(new Event('change'));
					}
				});
				
				widget.$webpageCrop.addEventListener('scroll', async function() {
					Object.assign(widget.$webpageCropBox.style, {
						left: `${this.scrollLeft}px`,
						top: `${this.scrollTop}px`
					});
					const attributes = await GetWidgetAttributes(widget);
					webpageRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
				});
				
				attr.url.$input.addEventListener('change', function() {
					document.getElementById('webpage-add').hidden = this.value.indexOf('customwebpages') != -1;
				});
				break;
			case 'youtube':
				attr.livestream.$input.addEventListener('change', function() {
					toggle([attr.starttime, attr.loop, attr.seektobeginning], !attr.livestream.$input.checked);
				});
				break;
			case 'pdf':
				var pdfCropTimeout = null;
				var pdfRenderCrop = (cropX, cropY, cropWidth, cropHeight) => {
					var clientWidth = widget.$pdfCrop.clientWidth;
					var clientHeight = widget.$pdfCrop.clientHeight;
					if (clientWidth != 0 && clientHeight != 0 &&
					    (widget.$pdfCropBox.width != clientWidth ||
					     widget.$pdfCropBox.height != clientHeight)) {
						widget.$pdfCropBox.width = clientWidth;
						widget.$pdfCropBox.height = clientHeight;
					}
					
					const context = widget.$pdfCropBox.getContext('2d');
					context.clearRect(0, 0, widget.$pdfCropBox.width, widget.$pdfCropBox.height);
					context.save();
					context.translate(-widget.$pdfCrop.scrollLeft, -widget.$pdfCrop.scrollTop);
					context.fillStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-fill-color');
					context.fillRect(cropX, cropY, cropWidth, cropHeight);
					context.strokeStyle = getComputedStyle(document.documentElement).getPropertyValue('--editor-crop-stroke-color');
					context.translate(0.5, 0.5);
					context.strokeRect(cropX, cropY, cropWidth, cropHeight);
					context.restore();
				};
				
				for (const attribute of [attr.pdf, attr.viewportwidth, attr.viewportheight, attr.scrollleft, attr.scrolltop, attr.scale]) {
					const inputChange = () => {
						// Clear this instantly so it looks visually immediate
						widget.$pdfCropPreview.innerHTML = '';
					
						clearTimeout(pdfCropTimeout);
						pdfCropTimeout = setTimeout(async () => {
							const attributes = await GetWidgetAttributes(widget);
							if (attributes.pdf) {
								const [pages, currentPageNumber] = await RenderPDF(widget.$pdfCropPreview, GetCachedMediaURL(account, attributes.pdf.cdnurl, attributes.pdf.uploadedtimestamp, attributes.pdf.filesize), attributes.scrollleft, attributes.scrolltop, false, false, attributes.viewportwidth, attributes.viewportheight, attributes.scale);
								const $pdfPageNumber = document.getElementById('pdfpagenumber');
								$pdfPageNumber.removeEventListener('change', $pdfPageNumber.lastChange);
								$pdfPageNumber.innerHTML = '';
								pages.forEach((page, pageIndex) => {
									$pdfPageNumber.appendChild($.option({ text: pageIndex + 1, value: pageIndex }));
								});
								$pdfPageNumber.value = pages.length != 0 && attributes.scrolltop >= pages[pages.length - 1].top ? pages.length - 1 : currentPageNumber;
								$pdfPageNumber.addEventListener('change', $pdfPageNumber.lastChange = function() {
									const pageNumber = this.value;
									document.getElementById('pdfscrolltop').value = pages[pageNumber].top;
									document.getElementById('pdfscrolltop').dispatchEvent(new Event('change'));
								});
								pdfRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
							}
						}, 2500);
					};
					attribute.$input.addEventListener('input', inputChange);
					attribute.$input.addEventListener('change', inputChange);
				}
				for (const attribute of [attr.cropx, attr.cropy, attr.cropwidth, attr.cropheight]) {
					const inputChange = async () => {
						const attributes = await GetWidgetAttributes(widget);
						pdfRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
					};
					attribute.$input.addEventListener('input', inputChange);
					attribute.$input.addEventListener('change', inputChange);
				}
				
				widget.$pdfCrop.addEventListener('click', function() {
					widget.$pdfCrop.scroll();
				});
				var startCropX;
				var startCropY;
				draggable(widget.$pdfCrop, {
					handle: '#pdfCropBox',
					scrollParent: widget.$pdfCrop,
					start: (e, startX, startY, startOffsetX, startOffsetY) => {
						startCropX = Math.floor(startOffsetX + widget.$pdfCrop.scrollLeft);
						startCropY = Math.floor(startOffsetY + widget.$pdfCrop.scrollTop);
						attr.cropx.$input.value = startCropX;
						attr.cropy.$input.value = startCropY;
						attr.cropwidth.$input.value = 1;
						attr.cropheight.$input.value = 1;
						attr.cropx.$input.dispatchEvent(new Event('change'));
					},
					drag: (e, startX, startY, startOffsetX, startOffsetY) => {
						const pdfCropBoxOffset = $.offset(widget.$pdfCropBox);
						var relativeX = Math.max(0, Math.floor(e.pageX - pdfCropBoxOffset.left + widget.$pdfCrop.scrollLeft));
						var relativeY = Math.max(0, Math.floor(e.pageY - pdfCropBoxOffset.top + widget.$pdfCrop.scrollTop));
						var cropX = Math.min(startCropX, relativeX);
						var cropY = Math.min(startCropY, relativeY);
						var cropWidth = Math.max(1, Math.abs(relativeX - startCropX));
						var cropHeight = Math.max(1, Math.abs(relativeY - startCropY));
						if (cropX + cropWidth > parseInt(attr.viewportwidth.$input.value, 10)) {
							cropWidth = parseInt(attr.viewportwidth.$input.value, 10) - cropX;
						}
						if (cropY + cropHeight > parseInt(attr.viewportheight.$input.value, 10)) {
							cropHeight = parseInt(attr.viewportheight.$input.value, 10) - cropY;
						}
						attr.cropx.$input.value = cropX;
						attr.cropy.$input.value = cropY;
						attr.cropwidth.$input.value = cropWidth;
						attr.cropheight.$input.value = cropHeight;
						attr.cropx.$input.dispatchEvent(new Event('change'));
					}
				});
				
				widget.$pdfCrop.addEventListener('scroll', async function() {
					Object.assign(widget.$pdfCropBox.style, {
						left: `${this.scrollLeft}px`,
						top: `${this.scrollTop}px`
					});
					const attributes = await GetWidgetAttributes(widget);
					pdfRenderCrop(attributes.cropx, attributes.cropy, attributes.cropwidth, attributes.cropheight);
				});
				break;
			case 'rss':
				for (const attribute of [attr.filter, attr.rssfeed]) {
					attribute.$input.addEventListener('change', async () => {
						const selectedNode = (await attr.rssfeed.$input.getSelected())[0];
						const useURL = selectedNode != undefined && selectedNode.hasOwnProperty('value') && selectedNode.id == 0;
						if (useURL) {
							attr.filter.$input.checked = false;
						}
						if (useURL || !attr.filter.$input.checked) {
							attr.start.$input.value = moment().format('YYYY-MM-DD');
							attr.end.$input.value = moment().format('YYYY-MM-DD');
						}
						attr.url.toggle(useURL);
						attr.filter.toggle(!useURL);
						attr.start.toggle(!useURL && attr.filter.$input.checked);
						attr.end.toggle(!useURL && attr.filter.$input.checked);
					});
				}
				break;
			case 'stockcrawl':
				for (const stock of [
					{ name: 'intc',  ticker: 'intc' },
					{ name: 'goog',  ticker: 'goog' },
					{ name: 't',  ticker: 't' },
					{ name: 'aapl',  ticker: 'aapl' },
					{ name: 'orcl',  ticker: 'orcl' },
					{ name: 'amd',  ticker: 'amd' },
					{ name: 'ge',  ticker: 'ge' }
				]) {
					document.getElementById(`stockcrawl${stock.name}`).addEventListener('change', function() {
						attr.text.$input.value = attr.text.$input.value.replace(new RegExp(stock.ticker + ',', 'g'), '');
						if (this.checked) {
							attr.text.$input.value = attr.text.$input.value + stock.ticker + ',';
						}
						attr.text.$input.dispatchEvent(new Event('change'));
					});
				}
				break;
			case 'csvroll':
				attr.csvfile.$input.addEventListener('change', async function() {
					const selectedNode = (await this.getSelected())[0];
					attr.url.toggle(selectedNode == undefined || selectedNode.id == 0);
				});
				break;
			case 'videoinput':
				attr.inputtype.$input.addEventListener('change', function() {
					attr.framerate.toggle(this.value == 'blackmagicinput');
					attr.inputframerate.toggle(this.value == 'blackmagicinput');
					attr.pixelformat.toggle(this.value == 'blackmagicinput');
					attr.inputpixelformat.toggle(this.value == 'blackmagicinput');
					attr.fillstyle.toggle(this.value == 'blackmagicinput');
					if (this.value == 'blackmagicinput') {
						attr.framerate.$input.dispatchEvent(new Event('change'));
						attr.pixelformat.$input.dispatchEvent(new Event('change'));
					}
				});
				attr.resolution.$input.addEventListener('change', function() {
					if (this.value != 'custom') {
						var resolution = this.value.split('x');
						attr.inputwidth.$input.value = resolution[0];
						attr.inputheight.$input.value = resolution[1];
					}
					attr.inputwidth.toggle(this.value == 'custom');
					attr.inputheight.toggle(this.value == 'custom');
				});
				attr.framerate.$input.addEventListener('change', function() {
					if (this.value != 'custom') {
						attr.inputframerate.$input.value = this.value;
					}
					attr.inputframerate.toggle(this.value == 'custom');
				});
				attr.pixelformat.$input.addEventListener('change', function() {
					if (this.value != 'custom') {
						attr.inputpixelformat.$input.value = this.value;
					}
					attr.inputpixelformat.toggle(this.value == 'custom');
				});
				break;
			case 'signwave':
				attr.usedefaultbackgroundimage.$input.addEventListener('change', () => {
					toggle([attr.defaultbackgroundimage, attr.defaultbackgroundimagecoverfill], attr.usedefaultbackgroundimage.$input.checked);
				});
				for (const attribute of [attr.showcurrentevent, attr.showimagefornocurrentevent, attr.enableswipepreviousandnext]) {
					attribute.$input.addEventListener('change', () => {
						const viewByHour = attr.showcurrentevent.$input.checked;
						toggle(
						[
							attr.currenteventalignment,
							attr.currenteventfirstline,
							attr.currenteventfirstlinefont,
							attr.currenteventfirstlinefontsize,
							attr.currenteventfirstlinecolor,
							attr.currenteventsecondline,
							attr.currenteventsecondlinefont,
							attr.currenteventsecondlinefontsize,
							attr.currenteventsecondlinecolor,
							attr.currenteventthirdline,
							attr.currenteventthirdlinefont,
							attr.currenteventthirdlinefontsize,
							attr.currenteventthirdlinecolor,
							attr.showimagefornocurrentevent,
							attr.upcomingeventthreshold,
							attr.enableswipepreviousandnext
						], viewByHour);
						toggle([attr.nocurrenteventimage, attr.nocurrenteventimagefillstyle], viewByHour && attr.showimagefornocurrentevent.$input.checked);
						toggle([attr.swipepreviousandnextarrowcolor], viewByHour && attr.enableswipepreviousandnext.$input.checked);
					});
				}
				for (const attribute of [attr.enableviewbydaybutton, attr.enableviewbyweekbutton, attr.eventlistviewenablescrollarrows, attr.eventlistshowimagesfornoevents]) {
					attribute.$input.addEventListener('change', () => {
						var viewByDay = attr.enableviewbydaybutton.$input.checked;
						var viewByWeek = attr.enableviewbyweekbutton.$input.checked;
						toggle(
						[
							attr.eventlistfont,
							attr.eventlistfontsize,
							attr.eventlistcolor,
							attr.eventlistcurrenteventarrowcolor,
							attr.eventlistcolumnfirst,
							attr.eventlistcolumnsecondary,
							attr.eventlistcolumntertiary
						], viewByDay || viewByWeek);
						toggle([attr.eventlistscrollarrowcolor], (viewByDay || viewByWeek) && attr.eventlistviewenablescrollarrows.$input.checked);
						toggle([attr.eventlistnoeventsimage, attr.eventlistnoeventsimagefill], (viewByDay || viewByWeek) && attr.eventlistshowimagesfornoevents.$input.checked);
					});
				}
				attr.showroom.$input.addEventListener('change', function() {
					toggle([attr.roomfont, attr.roomfontsize, attr.roomcolor, attr.roomhorizontalalignment], this.checked);
				});
				attr.showtime.$input.addEventListener('change', function() {
					toggle([attr.timefont, attr.timefontsize, attr.timecolor, attr.timehorizontalalignment], this.checked);
				});
				attr.showlogo.$input.addEventListener('change', function() {
					toggle(
					[
						attr.logoimage,
						attr.logoheight,
						attr.logohorizontalalignment
					], this.checked);
				});
				attr.enableinstanteventbutton.$input.addEventListener('change', function() {
					toggle(
					[
						attr.instanteventrequirepin,
						attr.instanteventcustomstartdate,
						attr.instanteventcustomenddate,
						attr.intanteventrequirename,
						attr.instanteventcustomtitle,
						attr.instanteventpredefinedtitles,
						attr.instanteventfont,
						attr.instanteventarrowcolor
					], this.checked);
				});
				attr.enablereportanissuebutton.$input.addEventListener('change', function() {
					toggle(
					[
						attr.reportanissuerequirepin,
						attr.reportanissuebuttontext,
						attr.reportanissueconfirmationtext,
						attr.reportanissueconfirmationfont,
						attr.reportanissueconfirmationfontsize,
						attr.reportanissueconfirmationcolor,
						attr.reportanissuetoemail
					], this.checked);
				});
				const renderMasterPINs = () => {
					attr.masterpins.$input.innerHTML = '';
					attr.masterpins.$input.append(
						$.header({},
							$.label({ text: 'PIN' }),
							$.i({ class: 'tooltip-button mdi mdi-plus-circle-outline', title: 'Add PIN', click: () => {
								attr.masterpins.$input.data.push('');
								renderMasterPINs();
							} })
						),
						...attr.masterpins.$input.data.map((pin, index) => {
							const pinInputChange = function() {
								attr.masterpins.$input.data[index] = this.value;
							};
							return [
								$.input({ type: 'text', class: 'sampleInput', placeholder: 'pin', css: { gridColumn: 1 }, value: pin, input: pinInputChange, change: pinInputChange }),
								$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: function() {
									attr.masterpins.$input.data.splice(index, 1);
									renderMasterPINs();
									attr.masterpins.$input.dispatchEvent(new Event('change'));
								} })
							];
						}).flat()
					);
				};
				attr.masterpins.$input.addEventListener('change', renderMasterPINs);
				
				const renderPredefinedTitles = () => {
					attr.instanteventpredefinedtitles.$input.innerHTML = '';
					attr.instanteventpredefinedtitles.$input.append(
						$.header({},
							$.label({ text: 'Title' }),
							$.i({ class: "tooltip-button mdi mdi-plus-circle-outline", title: "Add Title", click: () => {
								attr.instanteventpredefinedtitles.$input.data.push('');
								renderPredefinedTitles();
							} })
						),
						...attr.instanteventpredefinedtitles.$input.data.map((pin, index) => {
							const pinInputChange = function() {
								attr.instanteventpredefinedtitles.$input.data[index] = this.value;
							};
							return [
								$.input({ type: 'text', class: 'sampleInput', css: { gridColumn: 1 }, value: pin, input: pinInputChange, change: pinInputChange }),
								$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: function() {
									attr.instanteventpredefinedtitles.$input.data.splice(index, 1);
									renderPredefinedTitles();
									attr.instanteventpredefinedtitles.$input.dispatchEvent(new Event('change'));
								} })
							];
						}).flat()
					);
				};
				attr.instanteventpredefinedtitles.$input.addEventListener('change', renderPredefinedTitles);
				break;
			case 'readerboard':
				attr.usedefaultbackgroundimage.$input.addEventListener('change', function() {
					toggle([attr.defaultbackgroundimage, attr.defaultbackgroundimagecoverfill], this.checked);
				});
				attr.showname.$input.addEventListener('change', function() {
					toggle([attr.namealignment, attr.namefont, attr.namefontsize, attr.namecolor], this.checked);
				});
				attr.groupby.$input.addEventListener('change', function() {
					toggle(
					[
						attr.upcomingeventcountpergroup,
						attr.groupfont,
						attr.groupfontsize,
						attr.groupcolor,
						attr.groupbackgroundcolor,
						attr.groupnoeventstext
					], attr.groupby.$input.value != 'none');
					toggle([attr.upcomingeventcount], this.value == 'none');
				});
				attr.groupshownoevents.$input.addEventListener('change', function() {
					toggle([attr.groupnoeventstext], this.checked);
				});
				attr.eventlistusenoeventsimages.$input.addEventListener('change', function() {
					toggle([attr.eventlistnoeventimages, attr.eventlistnoeventimagescoverfill], this.checked);
				});
				
				const renderNoEventsImages = () => {
					attr.eventlistnoeventimages.$input.innerHTML = '';
					attr.eventlistnoeventimages.$input.append(
						$.header({},
							$.label({ text: 'Image' }),
							$.i({ class: "tooltip-button mdi mdi-plus-circle-outline", title: "Add another image", click: () => {
								attr.eventlistnoeventimages.$input.data.push(null);
								renderNoEventsImages();
							} })
						)
					);
					attr.eventlistnoeventimages.$input.append(...attr.eventlistnoeventimages.$input.data.map((media, index) => {
						return [
							$.uitreeselect({ css: { gridColumn: 1 }, config: {
								tree: 'content',
								leafIcon: node => GetContentIcon(node),
								filter: node => node.contenttreetype == 3
							}, change: async function(e) {
								e.stopPropagation();
								const selectedNode = (await this.getSelected())[0];
								const image = selectedNode ? ContentNodeToMinimalImageContent(selectedNode) : null;
								attr.eventlistnoeventimages.$input.data[index] = image;
								attr.eventlistnoeventimages.$input.dispatchEvent(new CustomEvent('change'));
							}, process: async function() {
								await this.setSelected(node => media != null && node.type == media.type && node.contentid == media.contentid);
							} }),
							$.i({ title: 'Delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: function() {
								attr.eventlistnoeventimages.$input.data.splice(index, 1);
								renderNoEventsImages();
								attr.eventlistnoeventimages.$input.dispatchEvent(new CustomEvent('change'));
							} })
						];
					}).flat());
				};
				attr.eventlistnoeventimages.$input.addEventListener('change', renderNoEventsImages);
				break;
			case 'powerbireport':
				attr.pagechangedelay.$input.addEventListener('change', function() {
					toggle([attr.pagenumber], attr.pagechangedelay.$input.value == '-1');
				});
				attr.togglereportfilters.$input.addEventListener('change', function() {
					toggle([attr.reportfilters], this.checked);
				});
				attr.togglepagefilters.$input.addEventListener('change', function() {
					toggle([attr.pagefilters], this.checked);
				});
				attr.togglevisualfilters.$input.addEventListener('change', function() {
					toggle([attr.visualfilters], this.checked);
				});
				break;
			case 'screenfeedsocialapp':
				attr.layout.$input.addEventListener('change', async function() {
					attr.columns.toggle(this.value == 'stream');
					attr.rows.toggle(this.value == 'photogrid');
					attr.margins.toggle(this.value == 'stream' || this.value == 'photogrid');
					attr.showimage.toggle(this.value == 'post' || this.value == 'stream');
					attr.imagelocation.toggle((this.value == 'post' || this.value == 'stream') && attr.showimage.$input.checked);
					if (attr.theme.$input.value != 'custom') {
						if (await QuestionBox('Reapply theme?', 'A theme is selected. The theme settings need to be reapplied for different layouts. Reapply the theme?')) {
							attr.theme.$input.dispatchEvent(new Event('change'));
						} else {
							attr.theme.$input.value = 'custom';
						}
					}
				});
				attr.theme.$input.addEventListener('change', function() {
					var theme = null;
					switch (this.value) {
						case 'Bluebird':
							theme = {
								base: '#0099e5',
								accent: '#ffffff',
								containers: '#ffffff',
								primaryText: '#ffffff',
								secondaryText: '#9dd9f7'
							};
							break;
						case 'Simple Gray':
							theme = {
								base: '#292929',
								accent: '#b8b8b8',
								containers: '#666666',
								primaryText: '#ffffff',
								secondaryText: '#b8b8b8'
							};
							break;
						case 'Clean Accent':
							theme = {
								base: '#ffffff',
								accent: '#39c088',
								containers: '#404040',
								primaryText: '#262626',
								secondaryText: '#808080'
							};
							break;
						case 'Red Mono':
							theme = {
								base: '#a90e0e',
								accent: '#ffffff',
								containers: '#e8a8a8',
								primaryText: '#ffffff',
								secondaryText: '#e8a8a8'
							};
							break;
						case 'Clean Combos':
							theme = {
								base: '#fbfaee',
								accent: '#ca2400',
								containers: '#868582',
								primaryText: '#0e3a5c',
								secondaryText: '#5d95d4'
							};
							break;
						case 'Dark Accent':
							theme = {
								base: '#000000',
								accent: '#fd7c00',
								containers: '#737373',
								primaryText: '#ffffff',
								secondaryText: '#a6a6a6'
							};
							break;
					}
					if (theme) {
						attr.background.$input.value = theme.base;
						for (const attribute of [attr.nameforeground, attr.posttextforeground]) {
							attribute.$input.value = theme.primaryText;
						}
						for (const attribute of [attr.usernameforeground, attr.detailsforeground]) {
							attribute.$input.value = theme.secondaryText;
						}
						switch (attr.layout.$input.value) {
							case 'post':
								attr.posttextfontsize.$input.value = 40;
								attr.containercolor.$input.value = hexToRGBA(theme.accent, 1);
								break;
							case 'stream':
								attr.posttextfontsize.$input.value = 30;
								attr.detailsforeground.$input.value = theme.accent;
								attr.containercolor.$input.value = hexToRGBA(theme.containers, 0.4);
								break;
							case 'photowall':
								attr.posttextfontsize.$input.value = 30;
								attr.containercolor.$input.value = 'rgba(0,0,0,0.4)';
								break;
							case 'photogrid':
								attr.posttextfontsize.$input.value = 30;
								attr.containercolor.$input.value = 'rgba(0,0,0,0.4)';
								break;
						}
					}
				});
				attr.showname.$input.addEventListener('change', function() {
					toggle([attr.namefont, attr.namefontsize, attr.nameforeground], this.checked);
				});
				attr.showusername.$input.addEventListener('change', function() {
					toggle([attr.usernamefont, attr.usernamefontsize, attr.usernameforeground], this.checked);
				});
				attr.showposttext.$input.addEventListener('change', function() {
					toggle([attr.posttextfont, attr.posttextfontsize, attr.posttextforeground], this.checked);
				});
				attr.backgroundtype.$input.addEventListener('change', function() {
					toggle([attr.backgroundimage, attr.backgroundimagecoverfill], this.value == 'image');
				});
				attr.showimage.$input.addEventListener('change', function() {
					attr.imagelocation.toggle((attr.layout.$input.value == 'post' || attr.layout.$input.value == 'stream') && this.checked);
				});
				break;
			case 'googlemaps':
				attr.provider.$input.addEventListener('change', function() {
					toggle([attr.zoom, attr.apikey, attr.maptype, attr.traffic, attr.nearbysearch, attr.marker], this.value == 'gm');
					attr.osmzoom.toggle(this.value == 'osm');
					attr.markertext.toggle(this.value == 'gm' && attr.marker.$input.checked);
				});
				attr.marker.$input.addEventListener('change', function() {
					attr.markertext.toggle(attr.provider.$input.value == 'gm' && attr.marker.$input.checked);
				});
				break;
			case 'buttonsfromcontenttreemetadata':
				attr.theme.$input.addEventListener('change', function() {
					toggle([attr.buttontextcolor, attr.buttonbackgroundcolor, attr.useimage, attr.buttonborderthickness, attr.buttondropshadow], attr.theme.$input.value == 'custom');
					toggle([attr.image, attr.fillstyle, attr.crop], attr.theme.$input.value == 'custom' && attr.useimage.$input.checked);
					toggle([attr.buttonbordercolor, attr.buttonborderradius], attr.theme.$input.value == 'custom' && attr.buttonborderthickness.$input.value != "0");
					toggle([attr.buttondropshadowleft, attr.buttondropshadowtop, attr.buttondropshadowblurradius, attr.buttondropshadowcolor], attr.theme.$input.value == 'custom' && attr.buttondropshadow.$input.checked);
				});
				attr.useimage.$input.addEventListener('change', function() {
					toggle([attr.image, attr.fillstyle, attr.crop], attr.theme.$input.value == 'custom' && attr.useimage.$input.checked);
				});
				attr.buttonborderthickness.$input.addEventListener('change', function() {
					toggle([attr.buttonbordercolor, attr.buttonborderradius], attr.theme.$input.value == 'custom' && attr.buttonborderthickness.$input.value != "0");
				});
				attr.buttondropshadow.$input.addEventListener('change', function() {
					toggle([attr.buttondropshadowleft, attr.buttondropshadowtop, attr.buttondropshadowblurradius, attr.buttondropshadowcolor], attr.theme.$input.value == 'custom' && attr.buttondropshadow.$input.checked);
				});
				break;
			case 'mediafrommetadata':
				attr.usedefaultimage.$input.addEventListener('change', function() {
					toggle([attr.defaultimage], attr.usedefaultimage.$input.checked);
				});
				break;
			case 'galleryfrommetadata':
				attr.usedefaultimage.$input.addEventListener('change', function() {
					toggle([attr.defaultimage], attr.usedefaultimage.$input.checked);
				});
				attr.mediaborderthickness.$input.addEventListener('change', function() {
					toggle([attr.mediabordercolor, attr.mediaborderradius], attr.mediaborderthickness.$input.value != "0");
				});
				attr.showarrows.$input.addEventListener('change', function() {
					toggle([attr.popuparrowclosecolor], attr.showarrows.$input.checked || attr.showclose.$input.checked);
				});
				attr.showclose.$input.addEventListener('change', function() {
					toggle([attr.popuparrowclosecolor], attr.showarrows.$input.checked || attr.showclose.$input.checked);
				});
				attr.mediadropshadow.$input.addEventListener('change', function() {
					toggle([attr.mediadropshadowleft, attr.mediadropshadowtop, attr.mediadropshadowblurradius, attr.mediadropshadowcolor], attr.mediadropshadow.$input.checked);
				});
				break;
			case 'ce6leadership':
				attr.usebackgroundimage.$input.addEventListener('change', function() {
					toggle([attr.backgroundimage, attr.backgroundimagecoverfill], attr.usebackgroundimage.$input.checked);
				});
				attr.showtitle.$input.addEventListener('change', function() {
					toggle([attr.titlefont, attr.titlefontsize, attr.titlecolor], attr.showtitle.$input.checked);
				});
				attr.showphonenumber.$input.addEventListener('change', function() {
					toggle([attr.phonenumberfont, attr.phonenumberfontsize, attr.phonenumbercolor], attr.showphonenumber.$input.checked);
				});
				attr.showemail.$input.addEventListener('change', function() {
					toggle([attr.emailfont, attr.emailfontsize, attr.emailcolor], attr.showemail.$input.checked);
				});
				break;
			case 'ce6events':
				attr.usebackgroundimage.$input.addEventListener('change', function() {
					toggle([attr.backgroundimage, attr.backgroundimagecoverfill], attr.usebackgroundimage.$input.checked);
				});
				attr.showdate.$input.addEventListener('change', function() {
					toggle([attr.groupbydayfont, attr.groupbydayfontsize, attr.groupbydaycolor, attr.datefont, attr.datefontsize, attr.datecolor], attr.showdate.$input.checked);
				});
				attr.showlocation.$input.addEventListener('change', function() {
					toggle([attr.locationfont, attr.locationfontsize, attr.locationcolor], attr.showlocation.$input.checked);
				});
				attr.showdescription.$input.addEventListener('change', function() {
					toggle([attr.descriptionfont, attr.descriptionfontsize, attr.descriptioncolor], attr.showdescription.$input.checked);
				});
				break;
			case 'ce6news':
				attr.usebackgroundimage.$input.addEventListener('change', function() {
					toggle([attr.backgroundimage, attr.backgroundimagecoverfill], attr.usebackgroundimage.$input.checked);
				});
				attr.showdate.$input.addEventListener('change', function() {
					toggle([attr.datefont, attr.datefontsize, attr.datecolor], attr.showdate.$input.checked);
				});
				attr.showauthor.$input.addEventListener('change', function() {
					toggle([attr.authorfont, attr.authorfontsize, attr.authorcolor], attr.showauthor.$input.checked);
				});
				attr.showdescription.$input.addEventListener('change', function() {
					toggle([attr.descriptionfont, attr.descriptionfontsize, attr.descriptioncolor, attr.resizefontsize], attr.showdescription.$input.checked);
				});
				attr.imagelayout.$input.addEventListener('change', function() {
					attr.imageminimumheight.toggle(this.value == 'bottom');
				});
				break;
			case 'button':
				attr.globalidle.$input.addEventListener('change', function() {
					toggle([attr.globalidledelay], attr.globalidle.$input.checked);
				});
				attr.idleactionbutton.$input.addEventListener('change', function() {
					toggle([attr.idledelay], attr.idleactionbutton.$input.value != 'none');
				});
				for (const attribute of [attr.clickable, attr.theme, attr.usefocustheme, attr.focustheme, attr.usebackgroundimage, attr.focususebackgroundimage]) {
					attribute.$input.addEventListener('change', function() {
						toggle([attr.text, attr.font, attr.fontsize, attr.theme, attr.focustheme], attr.clickable.$input.checked);
						toggle([attr.fontcolor, attr.backgroundcolor, attr.bordercolor, attr.borderthickness, attr.borderradius, attr.usebackgroundimage], attr.clickable.$input.checked && attr.theme.$input.value == 'custom');
						toggle([attr.backgroundimage, attr.backgroundimagecoverfill], attr.clickable.$input.checked && attr.theme.$input.value == 'custom' && attr.usebackgroundimage.$input.checked);
						toggle([attr.focustheme], attr.clickable.$input.checked && attr.usefocustheme.$input.checked);
						toggle([attr.focusfontcolor, attr.focusbackgroundcolor, attr.focusbordercolor, attr.focusborderthickness, attr.focusborderradius, attr.focususebackgroundimage], attr.clickable.$input.checked && attr.usefocustheme.$input.checked && attr.focustheme.$input.value == 'custom');
						toggle([attr.focusbackgroundimage, attr.focusbackgroundimagecoverfill], attr.clickable.$input.checked && attr.usefocustheme.$input.checked && attr.focustheme.$input.value == 'custom' && attr.focususebackgroundimage.$input.checked);
					});
				}
				break;
		}
		// The change() event is called for all widget controls so everything is in the correct state
		CallWidgetChange(widget);
	}
}

function isYoutube(element, index, array) {
	return (element.toLowerCase() != 'youtube');
}

function checkYoutubeURL(userUrl) {
	var link = $.a({ href: userUrl });
	var foundYoutube = false;
	var foundYoutu = false;
	var foundBe = false;
	var tokens = link.hostname.split('.');
	for (var i = 0; i < tokens.length; ++i) {
		var lowerCase = tokens[i].toLowerCase();
		if (lowerCase == 'youtube') {
			foundYoutube = true;
		} else if (lowerCase == 'youtu') {
			foundYoutu = true;
		} else if (lowerCase == 'be') {
			foundBe = true;
		}
	}
	if (foundYoutu && foundBe) {
		foundYoutube = true;
	}
	
	if (foundYoutube) {
		var regExp = /^.*(youtu.be\/|v\/|u\/\w\/|embed\/|watch\?v=|\&v=)([^#\&\?]*).*/;
		var match = userUrl.match(regExp);
		if (match && match[2].length == 11) {
			return match[2];
		} else {
			return null;
		}
	}
	return null;
}

function formatYoutubeURL(userUrl, videoId) {
	// https://www.youtube.com/embed/VIDEO_ID?playlist=VIDEO_ID&autoplay=1&rel=0&loop=1&showinfo=0&controls=0
	var construct = 'https://www.youtube.com/embed/' + videoId + '?playlist=' + videoId + '&autoplay=1&rel=0&loop=1&showinfo=0&controls=0';
	return construct;
}

function RepairWidget(media) {
	const widget = WidgetInformation.find(widget => widget.name == media.widgettype);
	for (const attribute of widget.attributes) {
		if (!media.attributes.hasOwnProperty(attribute.name)) {
			media.attributes[attribute.name] = attribute.default;
		}
	}
}

async function EditWidget(zone, media, x, y, width, height) {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return true;
	}
	stopPreviewPressed();
	
	const widget = WidgetInformation.find(widget => widget.name == media.widgettype);
	
	if (widget.attributes.length == 0) {
		return true;
	}
	
	RepairWidget(media);
	
	if (widget.name == 'pdf') {
		if ((await $contentTree.getAllNodes()).filter(node => node.type == MEDIATYPE_PDF).length == 0) {
			toastWarning('You must upload PDFs to the content tree before using this widget.');
		}
	}
	
	const lockWidgetDialog = () => {
		widget.$dialog.classList.add('busy');
	};
	
	const unlockWidgetDialog = () => {
		widget.$dialog.classList.remove('busy');
	};
	
	const redrawWidget = async (e = null) => {
		if (e) {
			e.stopPropagation();
		}
		// Temporarily use the current dialog values to render the widget
		var oldAttributes = JSON.parse(JSON.stringify(media.attributes));
		media.attributes = await GetWidgetAttributes(widget);
		media.ResizeMedia();
		zone.UpdateView();
		media.attributes = oldAttributes;
		media.ResizeMedia();
	};
	
	// This registers or unregisters the change callbacks on the widget dialogs to redraw the widget
	const toggleRedrawWidget = on => {
		for (const attribute of widget.attributes) {
			// Attributes moved to the left property panel use the PopulateMediaProperty and alterProperty code
			if (attribute.hasOwnProperty('view') && attribute.view == 'property') {
				continue;
			}
			var toggleEvents = (...events) => {
				for (const event of events) {
					if (on) {
						attribute.$input.addEventListener(event, redrawWidget);
					} else {
						attribute.$input.removeEventListener(event, redrawWidget);
					}
				}
			};
			switch (attribute.type) {
				case 'bool':
					toggleEvents('change');
					break;
				case 'string':
					if (attribute.hasOwnProperty('options')) {
						toggleEvents('change');
					} else if (attribute.hasOwnProperty('password')) {
						toggleEvents('change', 'input');
					} else {
						toggleEvents('change', 'input');
					}
					break;
				case 'int':
					if (attribute.hasOwnProperty('options')) {
						toggleEvents('change');
					} else {
						toggleEvents('change', 'input');
					}
					break;
				case 'float':
					if (attribute.hasOwnProperty('options')) {
						toggleEvents('change');
					} else {
						toggleEvents('change', 'input');
					}
					break;
				case 'color':
					toggleEvents('change');
					break;
				case 'date':
					toggleEvents('change', 'input');
					break;
				case 'time':
					toggleEvents('change', 'input');
					break;
				case 'json':
					toggleEvents('change');
					break;
				default:
					throw 'attribute type needs to be handled';
			}
		}
	};
	function ShowToolbar(id) {
		document.getElementById('toolbars').hidden = false;
		for (const $toolbar of document.querySelectorAll('#toolbars > div')) $toolbar.hidden = true;
		document.getElementById(id).hidden = false;
		UpdateLineLayer();
	}
	// Show the widget dialogs. Some of them require that they be visible to get dimension information like the webpage and pdf dialogs
	if (widget.name == 'buttongroup') {
		if (gCurrentWidget !== media) {
			gCurrentWidget = media;
			UpdateProperties('media');
			ShowToolbar('buttongrouptoolbar');
			resizeCanvas();
			if (!zone.visible) {
				ToggleZoneVisibility(zone);
			}
			zone.ReloadView();
		}
	} else if (widget.name == 'text2') {
		gCurrentWidget = media;
		UpdateProperties('media');
		ShowToolbar('text2toolbar');
		resizeCanvas();
		zone.ReloadView();
	} else {
		if (widget.hasOwnProperty('parent')) {
			gCurrentWidgetComponent = media;
		} else {
			gCurrentWidget = media;
		}
		widget.$dialog.hidden = false;
		// Buttons position the dialog relative to the button so the dialog is not over the button
		if (x != undefined && y != undefined && width != undefined && height != undefined) {
			const editorCanvasOffset = $.offset($editorCanvas);
			let dialogX = editorCanvasOffset.left + Math.round(toCanvasWidth(x + width)) + 5;
			let dialogY = editorCanvasOffset.top + Math.round(toCanvasHeight(y));
			if (dialogX + widget.$dialog.offsetWidth > document.documentElement.clientWidth) {
				dialogX = editorCanvasOffset.left + toCanvasWidth(x) - widget.$dialog.offsetWidth - 5;
			}
			if (dialogX < 0) {
				dialogX = 0;
			}
			if (dialogY + widget.$dialog.offsetHeight > document.documentElement.clientHeight) {
				dialogY = document.documentElement.clientHeight - widget.$dialog.offsetHeight;
			}
			if (dialogY < 0) {
				dialogY = 0;
			}
			Object.assign(widget.$dialog.style, {
				left: `${dialogX}px`,
				top: `${dialogY}px`
			});
		} else {
			centerOnScreen(widget.$dialog);
		}
	}
	
	// Load the widget values
	for (const attribute of widget.attributes) {
		attribute.$input.oldvalue = media.attributes[attribute.name];
		
		attribute.value = media.attributes[attribute.name];
		if (attribute.hasOwnProperty('$metadataInput')) {
			attribute.$metadataInput.value = media.attributes.hasOwnProperty(`${attribute.name}metadata`) ? media.attributes[`${attribute.name}metadata`] : '';
		}
		if (attribute.hasOwnProperty('$generatedMetadataInput')) {
			attribute.$generatedMetadataInput.value = media.attributes.hasOwnProperty(`${attribute.name}generatedmetadata`) ? media.attributes[`${attribute.name}generatedmetadata`] : '';
		}
		attribute.toggle(true); // This shows/hides metadata fields
	}
	
	for (const attribute of widget.attributes) {
		attribute.$input.dispatchEvent(new Event('change'));
	}
	
	toggleRedrawWidget(true);
	let cancelClickHandler;
	let okClickHandler;
	const cancelled = await new Promise(resolve => {
		widget.$cancel.addEventListener('click', cancelClickHandler = async () => {
			// Only update if something changed
			if (JSON.stringify(media.attributes) != JSON.stringify(await GetWidgetAttributes(widget))) {
				zone.UpdateView();
			}
			if (widget.name == 'webpage') {
				widget.$webpageCropPreview.src = 'about:blank';
			}
			resolve(true);
		});
		widget.$ok.addEventListener('click', okClickHandler = async () => {
			widget.$ok.style.pointerEvents = 'none';
			widget.$cancel.style.pointerEvents = 'none';
			const resetFieldValidation = () => {
				for (const attribute of widget.attributes) {
					if (attribute.hasOwnProperty('required')) {
						switch (attribute.type) {
							case 'string':
							case 'int':
							case 'float':
							case 'date':
							case 'time':
								RemoveValidationError.call(attribute.$input);
								break;
							case 'json':
								break;
							case 'bool':
								break;
							default:
								throw 'attribute type needs to be handled';
						}
					}
				}
			};
			
			const verifyWidgetFields = async () => {
				resetFieldValidation();
				var pass = [];
				for (const attribute of widget.attributes) {
					const error = message => {
						attribute.$input.classList.add('validationError');
						attribute.$input.title = message;
						attribute.$input.removeEventListener('focus', RemoveValidationError, { once: true });
						attribute.$input.addEventListener('focus', RemoveValidationError, { once: true });
					};
					switch (attribute.type) {
						case 'bool':
							break;
						case 'int':
							var parsedInt = parseInt(attribute.$input.value, 10);
							if (isNaN(parsedInt) || parsedInt != parseFloat(attribute.$input.value)) {
								error('Integer required, e.g. 42');
								pass.push(attribute);
							} else if (attribute.hasOwnProperty('minimum') && parsedInt < attribute.minimum) {
								error(`Input must be greater than or equal to ${attribute.minimum}`);
								pass.push(attribute);
							} else if (attribute.hasOwnProperty('maximum') && parsedInt > attribute.maximum) {
								error(`Input must be less than or equal to ${attribute.maximum}`);
								pass.push(attribute);
							}
							break;
						case 'float':
							if (isNaN(parseFloat(attribute.$input.value))) {
								error('Number required. e.g. 1.23');
								pass.push(attribute);
							} else if (attribute.hasOwnProperty('minimum') && parsedInt < attribute.minimum) {
								error(`Input must be greater than ${attribute.minimum}`);
								pass.push(attribute);
							} else if (attribute.hasOwnProperty('maximum') && parsedInt > attribute.maximum) {
								error(`Input must be less than ${attribute.maximum}`);
								pass.push(attribute);
							}
							break;
						case 'string':
							if (attribute.hasOwnProperty('required') && attribute.hasOwnProperty('design') && (attribute.design == 'fontfamily' || attribute.design == 'rssfeed' || attribute.design == 'csvfile' || attribute.design == 'eventlist') && attribute.$input.data == null) {
								error('Input required');
								pass.push(attribute);
							}
							if (attribute.hasOwnProperty('required') && attribute.$input.value == '') {
								error('Input required');
								pass.push(attribute);
							} else if (attribute.hasOwnProperty('match')) {
								throw 'Not implemented yet.';
							}
							break;
						case 'date':
							if (attribute.$input.value == '') {
								error('Input date is required');
								pass.push(attribute);
							}
							break;
						case 'time':
							if (attribute.$input.value == '') {
								error('Input time is required');
								pass.push(attribute);
							}
							break;
						case 'color':
							break;
						case 'json':
							if (attribute.hasOwnProperty('required') && attribute.hasOwnProperty('design') && attribute.design == 'media') {
								if ((await attribute.$input.getSelected()).length == 0) {
									error('Select required.');
									pass.push(attribute);
								}
							} else if (attribute.hasOwnProperty('required') && attribute.$input.data == null) {
								error('Select required.');
								pass.push(attribute);
							}
							break;
						default:
							throw 'attribute type needs to be handled';
					}
				}
				return pass;
			};
			
			try {
				lockWidgetDialog();
				const failedFields = await verifyWidgetFields();
				if (failedFields.length != 0) {
					toastError(`Invalid entries in the following fields: "${failedFields.map(attribute => attribute.display).join("', '")}". Hover over the corresponding field for more information.`);
					throw new Error();
				}
				let selectedNode;
				switch (widget.name) {
					case 'webpage':
						if (/^http:\/\/.*$/.test(document.getElementById('webpageurl').value) && document.getElementById('webpagelive').checked) {
							toastError('Webpages displayed live (in an iframe) on players must use https. The non-live option can render a non-secure http site and display it on players.');
							throw new Error();
						} else if (!/^http.*$/.test(document.getElementById('webpageurl').value)) {
							toastError('Webpages must define their protocol. Enter either https:// or http:// before the URL. (Ideally use https:// if the website supports it).');
							throw new Error();
						}
						break;
					case 'csvroll':
						selectedNode = (await document.getElementById('csvrollcsvfile').getSelected())[0];
						if (selectedNode != undefined && selectedNode.id == 0 && document.getElementById('csvrollurl').value == '') {
							toastError('You have selected to use a URL, but have not provided a URL.');
							throw new Error();
						}
						break;
					case 'rss':
						selectedNode = (await document.getElementById('rssrssfeed').getSelected())[0];
						if (selectedNode != undefined && selectedNode.id == 0 && document.getElementById('rssurl').value == '') {
							toastError('You have selected to use a URL but have not provided a URL.');
							throw new Error();
						}
						if (document.getElementById('rssfilter').checked && document.getElementById('rssstart').value == '') {
							toastError('You have selected to use a filter but have not provided a start.');
							throw new Error();
						}
						if (document.getElementById('rssfilter').checked && document.getElementById('rssstart').value == '') {
							toastError('You have selected to use a filter but have not provided an end.');
							throw new Error();
						}
						break;
					case 'rsscrawl':
						selectedNode = (await document.getElementById('rsscrawlrssfeed').getSelected())[0];
						if (selectedNode != undefined && selectedNode.id == 0 && document.getElementById('rsscrawlurl').value == '') {
							toastError('You have selected to use a URL but have not provided a URL.');
							throw new Error();
						}
						if (document.getElementById('rsscrawlfilter').checked && document.getElementById('rsscrawlstart').value == '') {
							toastError('You have selected to use a filter but have not provided a start.');
							throw new Error();
						}
						if (document.getElementById('rsscrawlfilter').checked && document.getElementById('rsscrawlstart').value == '') {
							toastError('You have selected to use a filter but have not provided an end.');
							throw new Error();
						}
						break;
					case 'events':
						if (document.getElementById('eventseventlist').data == '') {
							toastError('Select an event list in the drop down that you created in the Events editor. If none is listed you must go to the event editor in the left navigation area to create one first.');
							throw new Error();
						}
						break;
					case 'signwave':
						if (document.getElementById('signwaveusedefaultbackgroundimage').checked && (await document.getElementById('signwavedefaultbackgroundimage').getSelected()).length == 0) {
							toastError('Basic Details > Default Background Image: You have selected to use a default background image but have no provided an image.');
							throw new Error();
						}
						if (document.getElementById('signwaveshowlogo').checked && (await document.getElementById('signwavelogoimage').getSelected()).length == 0) {
							toastError('Logo > Image: You have selected to use a logo but have not provided an image.');
							throw new Error();
						}
						if (document.getElementById('signwaveshowlogo').checked && document.getElementById('signwavelogoheight').value == '') {
							toastError('Logo > Height: You have selected to use a logo but have not provided a logo image height.');
							throw new Error();
						}
						if (document.getElementById('signwaveshowcurrentevent').checked && document.getElementById('signwaveshowimagefornocurrentevent').checked && (await document.getElementById('signwavenocurrenteventimage').getSelected()).length == 0) {
							toastError('Current Event > Image: You have selected to use a no current event image but have not provided an image.');
							throw new Error();
						}
						if (document.getElementById('signwaveenablereportanissuebutton').checked && document.getElementById('signwavereportanissuebuttontext').value == '') {
							toastError('Report Issue > Button Text: You have enabled reporting an issue but have not provided text for the button.');
							throw new Error();
						}
						var viewByHour = document.getElementById('signwaveshowcurrentevent').checked;
						var viewByDay = document.getElementById('signwaveenableviewbydaybutton').checked;
						var viewByWeek = document.getElementById('signwaveenableviewbyweekbutton').checked;
						if (!viewByHour && !viewByDay && !viewByWeek) {
							toastError('Either current event, view by day, view by week, or any combination of the three must be enabled. Currently you have none of these options checked.');
							throw new Error();
						}
						if (document.getElementById('signwavereportanissuerequirepin').checked) {
							if (document.getElementById('signwavemasterpins').data.length == 0) {
								toastError('Master PINs > Master PINs: You have enabled reporting an issue to require a pin but have not entered one.');
								throw new Error();
							} else if (document.getElementById('signwavemasterpins').data.some(pin => pin == '')) {
								toastError('Master Pins > Master PINs: You have enabled reporting an issue to require a pin but have left a pin empty.');
								throw new Error();
							}
						}
						if (document.getElementById('signwaveenablereportanissuebutton').checked && document.getElementById('signwavereportanissueconfirmationtext').value == '') {
							toastError('Report Issue > Confirmation Text: You have enabled reporting an issue but have not provided confirmation text before a report is sent.');
							throw new Error();
						}
						if (document.getElementById('signwaveenablereportanissuebutton').checked && (document.getElementById('signwavereportanissuetoemail').value == '' || document.getElementById('signwavereportanissuetoemail').value.indexOf('@') == -1)) {
							toastError('Report Issue > Email: You have enabled reporting an issue but have not provided a valid email to report to.');
							throw new Error();
						}
						break;
					case 'readerboard':
						if (document.getElementById('readerboardusedefaultbackgroundimage').checked && (await document.getElementById('readerboarddefaultbackgroundimage').getSelected()).length == 0) {
							toastError('Basic Details > Default Background Image: You have selected to use a default background image but have no provided an image.');
							throw new Error();
						}
						break;
					case 'googlemaps':
						if (document.getElementById('googlemapsprovider').value == 'gm' && document.getElementById('googlemapsapikey').value == '') {
							toastError('Google Maps API Key: Google Maps requires an API Key tied to a billing account to work.');
							throw new Error();
						}
						break;
					case 'button':
						if (!/^[A-Za-z0-9_\-',. ]{1,30}$/.test(await widget.attr.name.value)) {
							toastError('Button names can only contain alphanumeric characters, underscores, commas, and periods and must be between 1 and 30 characters.');
							widget.attr.name.value = media.attributes.name;
							throw new Error();
						}
						await new Promise(async (resolve, reject) => {
							if (await widget.attr.default.value && !media.attributes.default) {
								let defaultButtonExists = false;
								for (const [otherButton] of GetAllButtons()) {
									if (otherButton.default) {
										defaultButtonExists = true;
									}
								}
								if (defaultButtonExists) {
									try {
										switch (await ConfirmationBox('Make this the default button?', 'Are you sure you want to make this button default?', 'Yes', 'No', 'Cancel')) {
											case 0:
												resolve();
												break;
											case 1:
												widget.attr.default.value = false;
												resolve();
												break;
										}
									} catch (exception) {
										reject();
									}
								} else {
									resolve();
								}
							} else {
								resolve();
							}
						});
						await new Promise(async (resolve, reject) => {
							if (await widget.attr.globalidle.value && !media.attributes.globalidle) {
								let globalIdleButtonExists = false;
								for (const [otherButton] of GetAllButtons()) {
									if (otherButton.globalidle) {
										globalIdleButtonExists = true;
									}
								}
								if (globalIdleButtonExists) {
									try {
										switch (await ConfirmationBox('Make this the global idle button?', 'Are you sure you want to make this button the global idle button?', 'Yes', 'No', 'Cancel')) {
											case 0:
												resolve();
												break;
											case 1:
												widget.attr.globalidle.value = false;
												resolve();
												break;
										}
									} catch (exception) {
										reject();
									}
								} else {
									resolve();
								}
							} else {
								resolve();
							}
						});
						await new Promise(async (resolve, reject) => {
							var conflict = false;
							for (const [otherButton] of GetAllButtons()) {
								if (media.button !== otherButton && otherButton.name == await widget.attr.name.value) {
									conflict = true;
									if (await QuestionBox('Button name in use', 'A button with that name already exists. Do you want to take the name and rename the other button?')) {
										ChangeButtonName(otherButton, GetUniqueButtonName());
										resolve();
									} else {
										widget.attr.name.value = media.attributes.name;
										reject();
									}
								}
							}
							if (!conflict) {
								resolve();
							}
						});
						// Save changes
						if (await widget.attr.default.value) {
							for (const [otherButton] of GetAllButtons()) {
								otherButton.default = false;
							}
						}
						if (await widget.attr.globalidle.value) {
							for (const [otherButton] of GetAllButtons()) {
								otherButton.globalidle = false;
							}
						}
						ChangeButtonName(media.button, await widget.attr.name.value);
						break;
				}
				media.attributes = await GetWidgetAttributes(widget);
				if (widget.name == 'button') {
					UpdateButtonList();
				}
				UpdateZoneAndMediaDOM();
				media.ResizeMedia();
				zone.ReloadView();
				SetSaveFlag(true);
				resetFieldValidation();
				if (widget.name == 'webpage') {
					widget.$webpageCropPreview.src = 'about:blank';
				}
				resolve(false);
			} catch (exception) {
				// There was an error and the user must correct it to continue
			} finally {
				unlockWidgetDialog();
				widget.$ok.style.pointerEvents = 'all';
				widget.$cancel.style.pointerEvents = 'all';
			}
		});
		// These widgets have no attributes so the dialog is skipped by clicking ok. This option can be moved as a setting in the widgets.json
		if (widget.name == 'blank' || widget.name == 'buttongroup' || widget.name == 'text2') {
			widget.$ok.click();
			if (widget.name == 'buttongroup') {
				UpdateButtonList();
			}
		}
	});
	widget.$cancel.removeEventListener('click', cancelClickHandler);
	widget.$ok.removeEventListener('click', okClickHandler);
	toggleRedrawWidget(false);
	widget.$dialog.hidden = true;
	return !cancelled;
}

function* GetAllButtonGroups() {
	if (gCurrentPlaylist == null) {
		return;
	}
	for (const zone of gCurrentPlaylist.zones) {
		if (zone.name != 'Audio') {
			for (const media of zone.media) {
				if (media.type == MEDIATYPE_WIDGET && media.widgettype == 'buttongroup') {
					yield [media, zone];
				}
			}
		}
	}
}

function* GetAllButtonsGrouped() {
	for (const [media, zone] of GetAllButtonGroups()) {
		yield [
			media,
			zone,
			media.attributes.buttons.map(button => button).sort((a, b) => {
				// Default button is always on the top
				if (a.default) {
					return -1;
				} else if (b.default) {
					return 1;
				} else {
					return a.name.localeCompare(b.name);
				}
			})
		];
	}
}

// This could be expensive and show up in the profiler
function* GetAllButtons() {
	for (const [media, zone] of GetAllButtonGroups()) {
		for (const button of media.attributes.buttons) {
			yield [button, media, zone];
		}
	}
}

function GetUniqueButtonName() {
	var pattern = /^Button ([1-9][0-9]*)$/;
	var nextValue = 1;
	for (const [button] of GetAllButtons()) {
		if (pattern.test(button.name)) {
			nextValue = Math.max(nextValue, parseInt(button.name.match(pattern)[1], 10) + 1);
		}
	}
	return 'Button ' + nextValue;
}

function ChangeButtonName(currentButton, newName) {
	for (const [button] of GetAllButtons()) {
		if (button.idleactionbutton == currentButton.name) {
			button.idleactionbutton = newName;
		}
	}
	currentButton.name = newName;
}

async function CreateButton(x, y) {
	var button = {
		name: GetUniqueButtonName(),
		x: Math.floor(x),
		y: Math.floor(y),
		width: 200,
		height: 100,
		metadata: [],
		zones: []
	};
	for (const zone of gCurrentPlaylist.zones) {
		button.zones.push({
			statechange: false,
			show: false,
			stop: false,
			restart: true,
			trimleft: -1,
			trimright: -1
		});
	}
	return button;
}

function SetDefaultButtonIfNoneExists()
{
	var defaultExists = false;
	for (const [button] of GetAllButtons()) {
		if (button.default) {
			defaultExists = true;
		}
	}
	if (!defaultExists) {
		var setDefault = false;
		for (const [button] of GetAllButtons()) {
			if (!setDefault) {
				button.default = true;
				setDefault = true;
			}
		}
	}
}

function DeleteButton(deleteButton) {
	if (deleteButton == null) {
		return;
	}
	for (const [media] of GetAllButtonGroups()) {
		for (var buttonItr = 0; buttonItr < media.attributes.buttons.length; ++buttonItr) {
			if (media.attributes.buttons[buttonItr] === deleteButton) {
				media.attributes.buttons.splice(buttonItr--, 1);
			}
		}
		for (const button of media.attributes.buttons) {
			if (button.idleactionbutton == deleteButton.name) {
				button.idleactionbutton = 'none';
			}
		}
	}
	SetDefaultButtonIfNoneExists();
	gCurrentWidgetComponent = null;
	reloadAllZoneViews();
	UpdateButtonList();
}

function BlinkDialog() {
	const $dialog = document.querySelector('ui-dialog:not([hidden])');
	if ($dialog && $dialog.classList.contains('blink')) {
		$dialog.addEventListener('animationend', function() {
			this.classList.remove('blink');
		}, { once: true });
		$dialog.classList.add('blink');
	}
}

function ExitWidgetMode(force = false) {
	if (gCurrentWidget) {
		gCurrentWidget.UpdateNameStartTimeAndRunLength();
	}
	if (force || gCurrentWidget || gCurrentWidgetComponent) {
		gCurrentWidget = null;
		gCurrentWidgetComponent = null;
		document.getElementById('toolbars').hidden = true;
		resizeCanvas();
		reloadAllZoneViews();
		UpdateProperties();
	}
}

function UpdateButtonList() {
	const oldScrollTop = document.querySelector('#interactivebuttons > .button-accordion') ? document.querySelector('#interactivebuttons > .button-accordion').scrollTop : 0;
	for (const [media, zone, buttons] of GetAllButtonsGrouped()) {
		if (!media.hasOwnProperty('collapsed')) {
			media.collapsed = false;
		}
	}
	document.getElementById('interactivebuttons').innerHTML = '';
	const accordion = document.getElementById('interactivebuttons').appendChild(
		$.ul({ class: 'button-accordion', focus: hideAccordion, scroll: hideAccordion })
	);
	let buttonItr = 0;
	for (const [media, zone, buttons] of GetAllButtonsGrouped()) {
		const buttonGroupHeaderToggle = function() {
			this.classList.toggle('collapsed');
			media.collapsed = this.classList.contains('collapsed');
			UpdateButtonList();
		};
		
		const buttonGroupHeader = accordion.appendChild($.li({ class: ['button-accordion-group-header', media.collapsed, 'collapsed'], multiclick: function(e, clicks) {
			if (DialogWindowOpen()) {
				BlinkDialog();
				return;
			}
			var inputTag = this.querySelector('input');
			if (!inputTag.classList.contains('edit')) {
				if (clicks == 1) {
					SelectMedia(zone, [zone.media.indexOf(media)], false, false, false, 'canvas', true, true);
					buttonGroupHeaderToggle.call(this);
				} else if (clicks == 2) {
					inputTag.classList.add('edit');
					const saveButtonGroupName = e => {
						if (e.type == 'blur' || e.type == 'keydown' && e.key == 'Enter') {
							inputTag.removeEventListener('blur', saveButtonGroupName);
							inputTag.removeEventListener('keydown', saveButtonGroupName);
							inputTag.classList.remove('edit');
							if (inputTag.value != '') {
								media.attributes.name = inputTag.value;
								media.name = inputTag.value;
							} else {
								media.attributes.name = 'Button Group';
								media.name = 'Button Group';
							}
							inputTag.blur();
							e.preventDefault();
							UpdateProperties();
							UpdateZoneAndMediaDOM();
							UpdateButtonList();
						}
					};
					inputTag.addEventListener('blur', saveButtonGroupName);
					inputTag.addEventListener('keydown', saveButtonGroupName);
					inputTag.focus();
				}
			}
		} },
			$.input({ value: media.attributes.name, title: `Double click to edit name: ${media.attributes.name}` }),
			$.i({ class: 'expander mdi mdi-format-vertical-align-center', title: 'Collapse all other button groups', css: { fontSize: '16px' }, click: e => {
				e.stopPropagation();
				for (const [otherMedia] of GetAllButtonsGrouped()) {
					if (otherMedia != media) {
						otherMedia.collapsed = true;
					}
				}
				media.collapsed = false;
				UpdateButtonList();
			} }),
			$.i({ class: 'expander mdi mdi-play mdi-rotate-90', click: function(e) {
				e.stopPropagation();
				buttonGroupHeaderToggle.call(this.parentNode);
			} })
		));
		if (!media.collapsed) {
			for (const button of buttons) {
				// Build accordion view for each button
				var buttonHeader = accordion.appendChild($.li({ id: `button-${buttonItr}`, text: `${button.name}${button.default ? ' (Default)' : ''}${button.globalidle ? ' (Global Idle)' : ''}`, class: ['button-accordion-header', gCurrentWidgetComponent !== button, 'collapsed'] }));
				if (gCurrentWidgetComponent === button) {
					const buttonZones = $.li({ class: 'button-accordion-body' });
					buttonZones.appendChild(
						$.div({ class: 'button-zone-header' },
							$.span({ text: 'Zone' }),
							$.i({ title: 'Enable zone state modifications. Enabling this allows a button click to change the visibility of this zone, the media range, whether to restart the zone, or play once depending on the selected options. Not enabling this does nothing to the zone when the button is clicked, keeping its current state. If enabled, any metadata associated with the button is copied to each of trimmed media. This allows buttons to change the state of widgets that reference metadata.', class: 'mdi mdi-check' }),
							$.i({ title: 'Whether to show the zone when the button is pressed', class: 'mdi mdi-eye-outline' }),
							$.i({ title: 'Trim left. Shows the zone as if there were no media before this media index starting at 1.', class: 'mdi mdi-arrow-collapse-right' }),
							$.i({ title: 'Trim right. Shows the zone as if there were no media after this media index starting at 1.', class: 'mdi mdi-arrow-collapse-left' }),
							$.i({ title: 'Start the zone from the beginning when the button is pressed', class: 'mdi mdi-skip-backward' }),
							$.i({ title: 'Play once. Stop at the end of the zone. If this is not enabled then the zone will loop', class: 'mdi mdi-stop' })
						)
					);
					button.zones.forEach((zone, zoneItr) => {
						var zoneName = gCurrentPlaylist.zones[zoneItr].name;
						var buttonZoneId = buttonItr + '-' + zoneItr;
						var editTrim = () => {
							const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '300px' } },
								$.header({ text: 'Trim Zone' }),
								$.div({ class: 'formgrid' },
									$.label({ for: 'buttonTrimZoneLeft', text: 'Trim Left Media' }),
									$.select({ id: 'buttonTrimZoneLeft', title: 'The first media in the zone to show. All media before this are not displayed.', value: zone.trimleft },
										$.option({ text: 'No Trim', value: -1 }),
										...gCurrentPlaylist.zones[zoneItr].media.map((media, mediaItr) => $.option({ text: media.display, value: mediaItr }))
									),
										
									$.label({ for: 'buttonTrimZoneRight', text: 'Trim Right Media' }),
									$.select({ id: 'buttonTrimZoneRight', title: 'The last media in the zone to show. All media after this are not displayed. This zone will loop back to the trim left media item defined above.', value: zone.trimright },
										$.option({ text: 'No Trim', value: -1 }),
										...gCurrentPlaylist.zones[zoneItr].media.map((media, mediaItr) => $.option({ text: media.display, value: mediaItr }))
									)
								),
								$.div({ class: 'panel-buttons' },
									$.button({ text: 'Cancel', click: function() {
										$dialog.remove();
									} }),
									$.button({ text: 'OK', click: function() {
										const newTrimLeft = parseInt(document.getElementById('buttonTrimZoneLeft').value, 10);
										const newTrimRight = parseInt(document.getElementById('buttonTrimZoneRight').value, 10);
										if (newTrimLeft == -1 || newTrimRight == -1 || newTrimLeft <= newTrimRight) {
											zone.trimleft = newTrimLeft;
											zone.trimright = newTrimRight;
										} else {
											zone.trimleft = newTrimRight;
											zone.trimright = newTrimLeft;
										}
										UpdateButtonList();
										$dialog.remove();
									} })
								)
							));
							centerOnScreen($dialog);
						};
						const trimLeftPointerEnter = function() {
							this.removeEventListener('pointerenter', trimLeftPointerEnter);
							this.title = zone.trimleft == -1 ? 'No Trim' : gCurrentPlaylist.zones[zoneItr].media[zone.trimleft].name;
							this.dispatchEvent(new Event('pointerenter'));
							this.addEventListener('pointerenter', trimLeftPointerEnter);
						};
						const trimRightPointerEnter = function() {
							this.removeEventListener('pointerenter', trimRightPointerEnter);
							this.title = zone.trimright == -1 ? 'No Trim' : gCurrentPlaylist.zones[zoneItr].media[zone.trimright].name;
							this.dispatchEvent(new Event('pointerenter'));
							this.addEventListener('pointerenter', trimRightPointerEnter);
						};
						buttonZones.appendChild(
							$.div({ class: [ 'button-zone', !zone.statechange, 'disabled', !zone.statechange || !zone.show, 'show-disabled' ], pointerenter: () => {
								const zoneMedia = gCurrentPlaylist.zones[zoneItr].media;
								const trimLeft = zone.trimleft == -1 ? 0 : zone.trimleft;
								const trimRight = zone.trimright == -1 ? zoneMedia.length - 1 : zone.trimright;
								if (trimLeft >= 0 && trimLeft < zoneMedia.length &&
								    trimRight >= 0 && trimRight < zoneMedia.length) {
									const trimLeftTag = zoneMedia[trimLeft].$timelineView;
									const trimRightTag = zoneMedia[trimRight].$timelineView;
									Object.assign($timeBarButtonZoneHighlight.style, {
										left: `${$timelineOptions.offsetWidth + 10 + trimLeftTag.offsetLeft}px`,
										top: `${trimLeftTag.closest('.timelineBar').offsetTop}px`,
										width: `${trimRightTag.offsetLeft + trimRightTag.offsetWidth - trimLeftTag.offsetLeft}px`,
										height: `${trimLeftTag.offsetHeight}px`
									});
									$.fadeIn($timeBarButtonZoneHighlight).catch(() => {});
								}
							}, pointerleave: () => $.fadeOut($timeBarButtonZoneHighlight).catch(() => {}) },
								$.span({ text: zoneName, title: zoneName }),
								$.uicheckbox({ id: 'button-statechange-' + buttonZoneId, title: !zone.statechange ? 'Enable the zone state change checkbox to enable zone visibility, trim, or play settings.' : '', checked: zone.statechange, change: function() {
									zone.statechange = this.checked;
									if (!zone.statechange) {
										zone.show = false;
										zone.stop = false;
										zone.restart = true;
										zone.trimleft = -1;
										zone.trimright = -1;
									}
									UpdateButtonList();
								} }),
								$.uicheckbox({ id: `button-show-${buttonZoneId}`, title: zone.show ? 'Enable the zone show checkbox to enable zone trim or play settings.' : '', checked: zone.show, change: function() {
									zone.show = this.checked;
									if (!zone.show) {
										zone.stop = false;
										zone.restart = true;
										zone.trimleft = -1;
										zone.trimright = -1;
									}
									UpdateButtonList();
								} }),
								$.span({ text: zone.trimleft == -1 ? '-' : zone.trimleft + 1, css: { cursor: 'pointer' }, click: editTrim, pointerenter: trimLeftPointerEnter }),
								$.span({ text: zone.trimright == -1 ? '-' : zone.trimright + 1, css: { cursor: 'pointer' }, click: editTrim, pointerenter: trimRightPointerEnter, css: {
									'cursor': 'pointer'
								} }),
								$.uicheckbox({ id: `button-restart-${buttonZoneId}`, checked: zone.restart, change: function() {
									zone.restart = this.checked;
								} }),
								$.uicheckbox({ id: `button-stop-${buttonZoneId}`, checked: zone.stop, change: function() {
									zone.stop = this.checked;
								} })
							)
						);
					});
					accordion.append(buttonZones);
				}
				
				const headerText = buttonHeader.innerText;
				buttonHeader.innerText = '';
				buttonHeader.append(
					$.span({ title: 'Click the header to preview a button click', text: headerText }),
					$.i({ title: 'Edit button attributes', class: 'mdi mdi-lead-pencil', css: { marginRight: '10px' }, click: e => {
						e.stopPropagation();
						EditWidget(zone, {
							button,
							widgettype: 'button',
							get attributes() { return button; },
							set attributes(attributes) { Object.assign(button, attributes); },
							ResizeMedia: () => {}
						});
					} }),
					$.i({ title: 'Delete button', class: 'mdi mdi-trash-can-outline', css: { marginRight: '5px' }, click: async (e) => {
						e.stopPropagation();
						if (DialogWindowOpen()) {
							BlinkDialog();
							return;
						}
						if (await QuestionBox('Delete button', `Are you sure you want to delete the button '${button.name}'.`)) {
							DeleteButton(button);
						}
					} }),
					$.i({ class: 'expander mdi mdi-play mdi-rotate-90', click: e => {
						e.stopPropagation();
						gCurrentWidgetComponent = gCurrentWidgetComponent !== button ? button : null;
						UpdateButtonList();
					} })
				);
				buttonHeader.addEventListener('click', () => {
					ClickButton(button);
					UpdateButtonList();
				});
				buttonItr++;
			}
		}
	}
	document.querySelector('#interactivebuttons > .button-accordion').scrollTop = oldScrollTop;
}

function UpdateSelection() {
	if (gSelectedZone != null) {
		$zoneSelect.removeEventListener('change', ZoneSelectChanged);
		$zoneSelect.value = gSelectedZone.name;
		$zoneSelect.addEventListener('change', ZoneSelectChanged);
	}
	for (const zone of gCurrentPlaylist.zones) {
		zone.$timelineName.classList.toggle('selected', zone === gSelectedZone);
		zone.$timelineBar.classList.toggle('selected', zone === gSelectedZone);
		for (const media of zone.media) {
			const scheduledState = media.GetScheduledState();
			media.$propertyView.classList.toggle('selected', media.selected)
			media.$propertyView.classList.toggle('scheduled', scheduledState.scheduled)
			media.$propertyView.classList.toggle('expired', scheduledState.expired);
			
			media.$timelineView.classList.toggle('selected', media.selected)
			media.$timelineView.classList.toggle('scheduled', scheduledState.scheduled)
			media.$timelineView.classList.toggle('expired', scheduledState.expired);
			
			media.$zoneView.classList.toggle('selected', media.selected)
			media.$zoneView.classList.toggle('scheduled', scheduledState.scheduled)
			media.$zoneView.classList.toggle('expired', scheduledState.expired);
		}
	}
}

function SaveZoneResize() {
	localStorage.zoneresize = JSON.stringify(gCurrentPlaylist.zones.map(zone => zone.hasOwnProperty('zoneViewWidth') ? zone.zoneViewWidth : 0));
}

function ToggleZoneVisibility(zone) {
	zone.visible = !zone.visible;
	var zoneState = [];
	for (const zone of gCurrentPlaylist.zones) {
		zoneState.push(zone.visible);
	}
	localStorage.zonestate = JSON.stringify(zoneState);
	UpdateZoneVisibility();
}

/*
async function GetRSSFeeds() {
	return new Promise((resolve, reject) => {
		try {
			var data = await get('/get/rssfeed.php');
			CheckSessionExpired(data);
			if (data.success) {
				resolve(false, data.data);
			} else {
				reject();
			}
		} catch (exception) {
			NetworkingError();
			reject();
		}
	});
}

async function GetEventLists() {
	return new Promise((resolve, reject) => {
		try {
			var data = await get('/get/eventlist.php');
			CheckSessionExpired(data);
			if (data.success) {
				resolve(false, data.data);
			} else {
				reject();
			}
		} catch (exception) {
			NetworkingError();
			reject();
		}
	});
}

async function GetCSVFiles() {
	return new Promise((resolve, reject) => {
		try {
			var data = await get('/get/csvfile.php');
			CheckSessionExpired(data);
			if (data.success) {
				resolve(false, data.data);
			} else {
				reject();
			}
		} catch (exception) {
			NetworkingError();
			reject();
		}
	});
}
*/

function GetVisualZoneOrder() {
	var visualZoneOrder = gCurrentPlaylist.zones.map(zone => zone);
	if (localStorage.swapzoneorder == 'true') {
		visualZoneOrder.reverse();
		// Move the audio zone, if it exists, to the front
		if (gCurrentPlaylist.zones.length > 0 && gCurrentPlaylist.zones[0].name == 'Audio') {
			visualZoneOrder.unshift(visualZoneOrder.pop());
		}
	}
	return visualZoneOrder;
}

function UpdateZoneAndMediaDOM() {
	var previousScrollLeft = $timelineView.scrollLeft;
	
	var zoneResize = JSON.parse(localStorage.zoneresize);
	zoneResize.forEach((width, zoneItr) => {
		if (zoneItr < gCurrentPlaylist.zones.length) {
			gCurrentPlaylist.zones[zoneItr].zoneViewWidth = width;
		}
	});
	
	var expiredMediaExists = false;
	
	var visualZoneOrdering = GetVisualZoneOrder();
	// Preverse the scrollTop
	visualZoneOrdering.forEach(zone => zone.zoneViewMediaScrollTop = zone.$zoneViewMedia.scrollTop);
	document.querySelector('#zone-view > div').append(...visualZoneOrdering.map(zone => zone.$zoneView), document.getElementById('zone-view-addzone'));
	visualZoneOrdering.forEach(zone => zone.$zoneViewMedia.scrollTop = zone.zoneViewMediaScrollTop);
	$timelineView.append(...visualZoneOrdering.map(zone => [zone.$timelineName, zone.$timelineBar]).flat());
	$zoneSelect.append(...visualZoneOrdering.map(zone => zone.$propertyView));
	
	for (const zone of gCurrentPlaylist.zones) {
		zone.$propertyView.value = zone.name;
		zone.$propertyView.innerText = zone.name;
		zone.$timelineBar.append(...zone.media.map(media => media.$timelineView));
		zone.$zoneViewMedia.append(...zone.media.map(media => media.$zoneView));
		
		var time = 0;
		for (const media of zone.media) {
			media.starttime = time;
			time += media.runlength;
			
			const scheduledState = media.GetScheduledState();
			if (scheduledState.expired) {
				expiredMediaExists = true;
			}
			media.$timelineView.hidden = !(!gShrink || media.runlength != 1);
		}
		
		for (const media of zone.media) {
			media.UpdateNameStartTimeAndRunLength();
		}
	}
	
	if (gSelectedZone) {
		for (const media of gSelectedZone.media) {
			media.UpdateNameStartTimeAndRunLength();
		}
		while($mediaSelect.firstChild) {
			$mediaSelect.removeChild($mediaSelect.firstChild);
		}
		$mediaSelect.append(...gSelectedZone.media.map(media => media.$propertyView));
	}
	updateZoneLayers();
	UpdateSelection();
	ResizeTimelineBars();
	$timelineView.scrollLeft = previousScrollLeft;
	UpdateZoneVisibility();
	UpdateTimeBar();
	
	document.getElementById('editor-removeexpiredmedia').hidden = !expiredMediaExists;
}

function UpdateZoneVisibility() {
	const zonesVisible = gCurrentPlaylist.zones.every(zone => zone.visible);
	const $toggleAllZones = document.getElementById('toggleallzones');
	$toggleAllZones.classList.toggle('mdi-eye-outline', zonesVisible);
	$toggleAllZones.classList.toggle('mdi-eye-off-outline', !zonesVisible);
	$toggleAllZones.title = zonesVisible ? 'Hide all zones' : 'Show all zones';
	for (const zone of gCurrentPlaylist.zones) {
		const $timelineViewHideButton = zone.$timelineName.querySelector('.hide-button');
		$timelineViewHideButton.classList.toggle('mdi-eye-outline', zone.visible);
		$timelineViewHideButton.classList.toggle('mdi-eye-off-outline', !zone.visible);
		$timelineViewHideButton.title = zone.visible ? 'Hide zone' : 'Show zone';
		
		const $zoneViewHideButton = zone.$zoneView.querySelector('.hide-button');
		$zoneViewHideButton.classList.toggle('mdi-eye-outline', zone.visible);
		$zoneViewHideButton.classList.toggle('mdi-eye-off-outline', !zone.visible);
		$zoneViewHideButton.title = zone.visible ? 'Hide zone' : 'Show zone';
		
		zone.$canvasView.hidden = !zone.visible;
		if (!zone.visible && zone.GetSelectedMedia() == gCurrentWidget) {
			ExitWidgetMode();
		}
	}
}

function UpdateTimeBar() {
	const timelineNameResizeWidth = $timelineNameResize.offsetWidth;
	$timelineNameResize.hidden = true;
	$.fadeOut($timeBarButtonZoneHighlight).catch(() => {});
	const $timeBar = document.getElementById('timeBar');
	const $timeBarLine = document.getElementById('timeBarLine');
	
	const smallTicks = gShrink ? gTimelineScale * 2 : gTimelineScale;
	const largeTicks = gShrink ? 20 * gTimelineScale : 10 * gTimelineScale;
	const largerTicks = gShrink ? 40 * gTimelineScale : 20 * gTimelineScale;
	
	const scrollLeft = $timelineView.scrollLeft;
	const scrollTop = $timelineView.scrollTop;
	// Draw the ticks on the timebar
	$timeBar.width = $timelineView.clientWidth - $timelineOptions.offsetWidth - timelineNameResizeWidth + 10;
	Object.assign($timeBar.style, {
		left: `${$timelineOptions.offsetWidth}px`,
		width: `${$timeBar.width}px`
	});
	
	const timeBarContext = $timeBar.getContext('2d');
	timeBarContext.clearRect(0, 0, $timeBar.width, $timeBar.height);
	timeBarContext.save();
	timeBarContext.translate(10, 0);
	const color = getComputedStyle(document.documentElement).getPropertyValue('--timeline-font-color');
	timeBarContext.strokeStyle = color;
	timeBarContext.fillStyle = color;
	timeBarContext.beginPath();
	for (var x = scrollLeft - scrollLeft % smallTicks - (1 * largeTicks); x < scrollLeft + $timeBar.width; x += smallTicks) {
		timeBarContext.moveTo(x + 0.5 - scrollLeft, 32 - (x % largerTicks == 0 ? 12 : (x % largeTicks == 0 ? 9 : 6)) + 0.5);
		timeBarContext.lineTo(x + 0.5 - scrollLeft, 32 + 0.5);
	}
	timeBarContext.stroke();
	timeBarContext.font = `12px ${getComputedStyle(document.documentElement).getPropertyValue('--font')}`;
	timeBarContext.textBaseline = 'top';
	for (var x = scrollLeft - scrollLeft % smallTicks - (1 * largeTicks); x < scrollLeft + $timeBar.width; x += smallTicks) {
		if (x % largeTicks == 0) {
			var time = secondsToTime(x / gTimelineScale);
			var timeStr = '';
			if (time.h > 0) {
				timeStr = time.h + ':' + pad2(time.m) + ':' + pad2(time.s);
			} else if (time.m > 0) {
				timeStr = time.m + ':' + pad2(time.s);
			} else {
				timeStr = '0:' + pad2(time.s);
			}
			timeBarContext.fillText(timeStr, x - scrollLeft, 2);
		}
	}
	timeBarContext.clearRect(-10, 0, 10, $timeBar.height);
	// Draw the chevron at the current time
	const positionX = gTimeBarSelection * gTimelineScale - scrollLeft;
	if (positionX >= 0) {
		timeBarContext.lineWidth  = 2;
		timeBarContext.beginPath();
		timeBarContext.moveTo(positionX - 5.5, 18);
		timeBarContext.lineTo(positionX + 0.5, 24);
		timeBarContext.lineTo(positionX + 6.5, 18);
		timeBarContext.stroke();
	}
	timeBarContext.restore();
	
	Object.assign($timeBarLine.style, {
		left: `${$timelineOptions.offsetWidth + timelineNameResizeWidth + gTimeBarSelection * gTimelineScale}px`,
		top: `${scrollTop + $timeBar.height}px`,
		height: `${$timelineView.clientHeight - $timeBar.height}px`
	});
	$timeBarLine.hidden = positionX <= 0;
	
	Object.assign($timelineNameResize.style, {
		left: `${$timelineOptions.offsetWidth}px`,
		height: `${$timelineView.clientHeight - $timelineOptions.offsetHeight}px`
	});
	$timelineNameResize.hidden = false;
}

function AddZone(zone, populatingPlaylist = false) {
	ExitWidgetMode();
	gCurrentPlaylist.zones.push(zone);
	updateZoneLayers();
	if (!populatingPlaylist) {
		reloadAllZoneViews();
		SaveZoneResize();
		UpdateZoneAndMediaDOM();
		UpdateProperties();
		for (const [button] of GetAllButtons()) {
			button.zones.push(
			{
				statechange: false,
				show: false,
				stop: false,
				restart: true,
				trimleft: -1,
				trimright: -1
			});
		}
		UpdateButtonList();
	}
}

function RemoveZone(zone) {
	var index = gCurrentPlaylist.zones.indexOf(zone);
	gCurrentPlaylist.zones.splice(index, 1);
	zone.Remove();
	updateZoneLayers();
	reloadAllZoneViews();
	SaveZoneResize();
	UpdateZoneAndMediaDOM();
	
	UpdateProperties();
	for (const [button] of GetAllButtons()) {
		button.zones.splice(index, 1);
	}
	UpdateButtonList();
}

function MoveZoneLayer(zone, offset) {
	var index = gCurrentPlaylist.zones.indexOf(zone);
	gCurrentPlaylist.zones.splice(index + offset, 0, gCurrentPlaylist.zones.splice(index, 1)[0]);
	updateZoneLayers();
	reloadAllZoneViews();
	SaveZoneResize();
	UpdateProperties();
	UpdateZoneAndMediaDOM();
	for (const [button] of GetAllButtons()) {
		button.zones.splice(index + offset, 0, button.zones.splice(index, 1)[0]);
	}
	UpdateButtonList();
	SetSaveFlag(true);
}

function addNewZone() {
	if (gCurrentPlaylist == null) {
		toastError('Create or open a playlist to add a new zone.');
		return;
	}
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	stopPreviewPressed();
	var marginX = gCurrentPlaylist.pagewidth / 5;
	var marginY = gCurrentPlaylist.pageheight / 5;
	var zone = new Zone(-1, 'Zone0', marginX, marginY, 0, gCurrentPlaylist.pagewidth - (2 * marginX), gCurrentPlaylist.pageheight - (2 * marginY), 100, false);
	AddZone(zone);
	
	SelectZone(gCurrentPlaylist.zones[gCurrentPlaylist.zones.length - 1], true, false);
	
	SetSaveFlag(true);
}

function toggleAllZonesPressed() {
	const zonesVisible = gCurrentPlaylist.zones.every(zone => zone.visible);
	for (const zone of gCurrentPlaylist.zones) {
		zone.visible = !zonesVisible;
	}
	var zoneState = [];
	for (const zone of gCurrentPlaylist.zones) {
		zoneState.push(zone.visible);
	}
	localStorage.zonestate = JSON.stringify(zoneState);
	UpdateZoneVisibility();
}

async function deleteCurrentZonePressed() {
	if (gSelectedZone) {
		if (gSelectedZone.media.length == 0 || await QuestionBox('Delete zone?', 'Are you sure you want to delete the zone?')) {
			deleteCurrentZone();
		}
	}
}

function deleteCurrentZone() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	stopPreviewPressed();
	if (gSelectedZone) {
		var zoneIndex = gCurrentPlaylist.zones.indexOf(gSelectedZone);
		RemoveZone(gSelectedZone);
		// Select the previous zone or another zone or no zone depending on the zones available
		SelectZone(gCurrentPlaylist.zones.length == 0 ? null : (zoneIndex == 0 ? gCurrentPlaylist.zones[0] : gCurrentPlaylist.zones[zoneIndex - 1]), true, false);
		SetSaveFlag(true);
	}
}

async function deleteCurrentMedia(showConfirmation) {
	if (showConfirmation) {
		if (gSelectedZone.mediaSelected.length == 1) {
			var mediaSelected = gSelectedZone.mediaSelected[0];
			if (mediaSelected.type == MEDIATYPE_WIDGET && mediaSelected.name == 'signwave') {
				if (await QuestionBox('Delete Signwave widget?', 'Are you sure you want to delete the Signwave widget?')) {
					deleteCurrentMedia(false);
				}
			} else if (mediaSelected.type == MEDIATYPE_WIDGET && mediaSelected.name == 'readerboard') {
				if (await QuestionBox('Delete Readerboard widget?', 'Are you sure you want to delete the Readerboard widget?')) {
					deleteCurrentMedia(false);
				}
			} else if (mediaSelected.type == MEDIATYPE_WIDGET && mediaSelected.widgettype == 'buttongroup') {
				if (gCurrentWidgetComponent) {
					try {
						switch (await ConfirmationBox('Delete Button or Button Group widget?', 'Do you want to delete the selected Button or delete the Button Group widget?', 'Button', 'Button Group', 'Cancel')) {
							case 0:
								DeleteButton(gCurrentWidgetComponent);
								break;
							case 1:
								deleteCurrentMedia(false);
								break;
						}
					} catch (exception) {
						// Cancelled
					}
				} else {
					if (await QuestionBox('Delete Button Group widget?', 'Are you sure you want to delete the Button Group widget?')) {
						deleteCurrentMedia(false);
					}
				}
			} else {
				deleteCurrentMedia(false);
			}
		} else {
			var readerboardCount = 0;
			var buttonGroupCount = 0;
			for (const mediaSelected of gSelectedZone.mediaSelected) {
				if (mediaSelected.type == MEDIATYPE_WIDGET && mediaSelected.name == 'readerboard') {
					readerboardCount++;
				}
				if (mediaSelected.type == MEDIATYPE_WIDGET && mediaSelected.widgettype == 'buttongroup' && gCurrentWidget != null) {
					buttonGroupCount++;
				}
			}
			if (readerboardCount != 0 || buttonGroupCount != 0) {
				var confirmationMessage = '';
				// Readerboard
				if (readerboardCount == 1) {
					confirmationMessage += 'Readerboard';
				} else if (readerboardCount >= 2) {
					confirmationMessage += readerboardCount + ' Readerboard';
				}
				if (readerboardCount != 0 && buttonGroupCount != 0) {
					confirmationMessage += ' and ';
				}
				// Button Group
				if (buttonGroupCount == 1) {
					confirmationMessage += 'Button Group';
				} else if (buttonGroupCount >= 2) {
					confirmationMessage += buttonGroupCount + ' Button Group';
				}
				if (await QuestionBox(`Delete widget${readerboardCount + buttonGroupCount == 1 ? '' : 's'}?`, `Are you sure you want to delete the ${confirmationMessage} widget${readerboardCount + buttonGroupCount == 1 ? '' : 's'}?`)) {
					deleteCurrentMedia(false);
				}
			} else {
				// No Readerboard or Button Group widgets in selection
				deleteCurrentMedia(false);
			}
		}
	} else {
		// For all the button groups in the playlist reset any trim information and display a message indicating this change
		var requiresUpdateButtonList = false;
		for (const [button] of GetAllButtons()) {
			var buttonZone = button.zones[gCurrentPlaylist.zones.indexOf(gSelectedZone)];
			for (const mediaSelected of gSelectedZone.mediaSelected) {
				var mediaIndex = gSelectedZone.media.indexOf(mediaSelected);
				if (buttonZone.trimleft == mediaIndex) {
					buttonZone.trimleft = -1;
					requiresUpdateButtonList = true;
				}
				if (buttonZone.trimright == mediaIndex) {
					buttonZone.trimright = -1;
					requiresUpdateButtonList = true;
				}
			}
		}
		gSelectedZone.ConvertSelectedButtonZoneTrimToReferences();
		var minSelectedIndex = Number.POSITIVE_INFINITY;
		for (const mediaSelected of gSelectedZone.mediaSelected) {
			minSelectedIndex = Math.min(minSelectedIndex, gSelectedZone.media.indexOf(mediaSelected));
			gSelectedZone.media.splice(gSelectedZone.media.indexOf(mediaSelected), 1);
			mediaSelected.Remove();
		}
		gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
		SetSaveFlag(true);
		SelectMedia(gSelectedZone, gSelectedZone.media.length == 0 ? [] : [minSelectedIndex != 0 ? minSelectedIndex - 1 : minSelectedIndex], false, false, false, null, true, true);
		if (gSelectedZone.media.length == 0) {
			setTimeout(function() {
				if (gSelectedZone) {
					gSelectedZone.$timelineName.focus();
				}
			}, 0);
		}
		
		if (gSelectedZone.media.length == 0 && gSelectedZone.name == 'Audio') {
			deleteCurrentZone();
		} else {
			gSelectedZone.ReloadView();
		}
		
		gSelectedZone.UpdateMediaStartTimes();
		ExitWidgetMode();
		$buttonTab.hidden = !gCurrentPlaylist.IsInteractive();
		if (!gCurrentPlaylist.IsInteractive() && $buttonTab.classList.contains('selected')) {
			document.getElementById('mediatab').click();
		}
		UpdateWidgetNames();
		SetDefaultButtonIfNoneExists();
		UpdateButtonList();
		UpdateProperties(gSelectedZone == null || gSelectedZone.media.length == 0 ? 'zone' : 'media');
		UpdateZoneAndMediaDOM();
		
		if (requiresUpdateButtonList) {
			UpdateButtonList();
		}
	}
}

function UpdateWidgetNames() {
	if (gCurrentPlaylist == null) {
		return;
	}
	var typeToDisplay = {};
	for (const widget of WidgetInformation) {
		typeToDisplay[widget.name] = widget.display;
	}
	for (const zone of gCurrentPlaylist.zones) {
		var widgetSequenceNumbers = {};
		for (const media of zone.media) {
			if (media.type == MEDIATYPE_WIDGET && media.widgettype != 'buttongroup') {
				if (!widgetSequenceNumbers.hasOwnProperty(media.widgettype)) {
					widgetSequenceNumbers[media.widgettype] = 1;
				}
				media.name = typeToDisplay[media.widgettype] + ' ' + widgetSequenceNumbers[media.widgettype]++;
			}
		}
	}
}

function updateCurrentZoneMediaOrder() {
	if (gCurrentPlaylist && gSelectedZone) {
		gSelectedZone.$timelineBar.append(...gSelectedZone.media.map(media => media.$timelineView));
		gSelectedZone.$zoneViewMedia.append(...gSelectedZone.media.map(media => media.$zoneView));
		$mediaSelect.append(...gSelectedZone.media.map(media => media.$propertyView));
	}
}

function moveCurrentMediaBefore(e) {
	gSelectedZone.ConvertSelectedButtonZoneTrimToReferences();
	
	var mediaSelectedIndexes = gSelectedZone.GetSortedMediaSelectedIndexes();
	var selectedIndex = gSelectedZone.GetSelectedMediaIndex();
	
	var min = mediaSelectedIndexes[0];
	var max = mediaSelectedIndexes[mediaSelectedIndexes.length - 1];
	
	if (e.ctrlKey) {
		var selectedMedia = gSelectedZone.GetAllSelectedMedia(); // Holds a reference to the selected items. Useful since we'll be copying over their place in the media array.
		
		var selectedIndexesAfter = 0;
		for (const index of mediaSelectedIndexes) {
			if (index > selectedIndex) {
				selectedIndexesAfter++;
			}
		}
		
		// Shift the non-selected items right to make room for the selected items on the left
		for (var index = max; index > selectedIndex; --index) {
			// If the index is selected or null then copy over it with the previous item and set the previous item to null
			if (mediaSelectedIndexes.indexOf(index) != -1 || gSelectedZone.media[index] == null) {
				for (var index2 = index - 1; index2 > selectedIndex; --index2) {
					if (mediaSelectedIndexes.indexOf(index2) == -1 && gSelectedZone.media[index2] != null) {
						gSelectedZone.media[index] = gSelectedZone.media[index2];
						gSelectedZone.media[index2] = null;
						break;
					}
				}
			}
		}
		
		// Move the selected items after selectedIndex
		for (var index = selectedIndex + 1; index < selectedIndex + 1 + selectedIndexesAfter; ++index) {
			var selectedMediaOffset = selectedMedia.indexOf(gSelectedZone.media[selectedIndex]);
			gSelectedZone.media[index] = selectedMedia[selectedMediaOffset + (index - selectedIndex)];
		}
		
		mediaSelectedIndexes = gSelectedZone.GetSortedMediaSelectedIndexes();
		
		var canMoveLeft = false;
		for (var index = selectedIndex - 1; index >= 0; --index) {
			if (mediaSelectedIndexes.indexOf(index) == -1) {
				canMoveLeft = true;
				break;
			}
		}
		
		if (canMoveLeft) {
			max = mediaSelectedIndexes[mediaSelectedIndexes.length - 1];
			// Move the selectedIndex and all following selected items to the left and move the displaced item in the front to the right of the selected items.
			for (var index = selectedIndex - 1; index >= 0; --index) {
				if (mediaSelectedIndexes.indexOf(index) == -1) {
					var notSelected = gSelectedZone.media[index];
					for (; index < max; ++index) {
						gSelectedZone.media[index] = gSelectedZone.media[index + 1];
					}
					gSelectedZone.media[max] = notSelected;
					break;
				}
			}
			SetSaveFlag(true);
		} else {
			gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
			return false;
		}
	} else {
		if (min != 0) {
			// Regular shift left. Move every selected item once left. This works with both sequential and non-sequential items.
			var movingMedia = gSelectedZone.media[min - 1];
			for (var index = min; index <= max + 1; ++index) {
				if (mediaSelectedIndexes.indexOf(index) == -1) {
					// Not selected media item
					if (movingMedia) {
						gSelectedZone.media[index - 1] = movingMedia;
					}
					if (index < gSelectedZone.media.length) {
						movingMedia = gSelectedZone.media[index];
					}
				} else {
					gSelectedZone.media[index - 1] = gSelectedZone.media[index];
				}
			}
			SetSaveFlag(true);
		} else {
			gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
			return false;
		}
	}
	
	gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
	
	gSelectedZone.UpdateMediaStartTimes();
	return true;
}

function moveCurrentMediaAfter(e) {
	gSelectedZone.ConvertSelectedButtonZoneTrimToReferences();
	
	var mediaSelectedIndexes = gSelectedZone.GetSortedMediaSelectedIndexes();
	var selectedIndex = gSelectedZone.GetSelectedMediaIndex();
	
	var min = mediaSelectedIndexes[0];
	var max = mediaSelectedIndexes[mediaSelectedIndexes.length - 1];
	
	if (e.ctrlKey) {
		var selectedMedia = gSelectedZone.GetAllSelectedMedia(); // Holds a reference to the selected items. Useful since we'll be copying over their place in the media array.
		
		var selectedIndexesBefore = 0;
		for (const index of mediaSelectedIndexes) {
			if (index < selectedIndex) {
				selectedIndexesBefore++;
			}
		}
		
		// Shift the non-selected items left to make room for the selected items on the right
		for (var index = min; index < selectedIndex; ++index) {
			// If the index is selected or null then copy over it with the previous item and set the previous item to null
			if (mediaSelectedIndexes.indexOf(index) != -1 || gSelectedZone.media[index] == null) {
				for (var index2 = index + 1; index2 < selectedIndex; ++index2) {
					if (mediaSelectedIndexes.indexOf(index2) == -1 && gSelectedZone.media[index2] != null) {
						gSelectedZone.media[index] = gSelectedZone.media[index2];
						gSelectedZone.media[index2] = null;
						break;
					}
				}
			}
		}
		
		// Move the selected items after selectedIndex
		for (var index = selectedIndex - selectedIndexesBefore; index < selectedIndex; ++index) {
			gSelectedZone.media[index] = selectedMedia[index - (selectedIndex - selectedIndexesBefore)];
		}
		
		mediaSelectedIndexes = gSelectedZone.GetSortedMediaSelectedIndexes();
		
		var canMoveRight = false;
		for (var index = selectedIndex + 1; index < gSelectedZone.media.length; ++index) {
			if (mediaSelectedIndexes.indexOf(index) == -1) {
				canMoveRight = true;
				break;
			}
		}
		
		if (canMoveRight) {
			min = mediaSelectedIndexes[0];
			// Move the selectedIndex and all following selected items to the right and move the displaced item in the front to the left of the selected items.
			for (var index = selectedIndex + 1; index < gSelectedZone.media.length; ++index) {
				if (mediaSelectedIndexes.indexOf(index) == -1) {
					var notSelected = gSelectedZone.media[index];
					for (; index > min; --index) {
						gSelectedZone.media[index] = gSelectedZone.media[index - 1];
					}
					gSelectedZone.media[min] = notSelected;
					break;
				}
			}
			SetSaveFlag(true);
		} else {
			gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
			return false;
		}
	} else {
		if (max != gSelectedZone.media.length - 1) {
			// Regular shift right. Move every selected item once right. This works with both sequential and non-sequential items.
			if (max != gSelectedZone.media.length - 1) {
				var movingMedia = gSelectedZone.media[max + 1];
				for (var index = max; index >= min - 1; --index) {
					if (mediaSelectedIndexes.indexOf(index) == -1) {
						// Not selected media item
						if (movingMedia) {
							gSelectedZone.media[index + 1] = movingMedia;
						}
						if (index >= 0) {
							movingMedia = gSelectedZone.media[index];
						}
					} else {
						gSelectedZone.media[index + 1] = gSelectedZone.media[index];
					}
				}
			}
			SetSaveFlag(true);
		} else {
			gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
			return false;
		}
	}
	
	gSelectedZone.ConvertSelectedButtonZoneTrimReferencesToIndexes();
	
	gSelectedZone.UpdateMediaStartTimes();
	return true;
}

function selectMediaAtTime(seconds) {
	var lastZoneSelected = null;
	for (const zone of gCurrentPlaylist.zones) {
		var zoneDuration = zone.media.reduce((duration, media) => duration + media.runlength, 0);
		var indexes = [];
		zone.media.forEach(function(media, mediaItr) {
			if (seconds % zoneDuration >= media.starttime && seconds % zoneDuration < media.starttime + media.runlength) {
				lastZoneSelected = zone;
				indexes = [mediaItr];
			}
		});
		SelectMedia(zone, indexes, false, false, false, 'timeline', false, false);
	}
	if (lastZoneSelected) {
		SelectZone(lastZoneSelected, true, false);
	}
}

function selectTimeBarLocation(e) {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	if (gCurrentPlaylist != null) {
		const $timeBar = document.getElementById('timeBar');
		const timeBarOffset = $.offset($timeBar);
		const seconds = Math.max(0, Math.round(((e.pageX - timeBarOffset.left - 10) + $timelineView.scrollLeft) / gTimelineScale));
		gTimeBarSelection = seconds;
		UpdateTimeBar();
		selectMediaAtTime(seconds);
	}
}

function CalculateCanvasDimensions(scaleValue = $canvasSlider.value) {
	if (gExpand) {
		const scale = scaleValue / 100;
		return {
			width: gCurrentPlaylist.pagewidth * scale,
			height: gCurrentPlaylist.pageheight * scale
		};
	} else {
		const viewportWidth = $editorCanvasContainer.offsetWidth;
		const viewportHeight = $editorCanvasContainer.offsetHeight;
		const canvasWidth = Math.min(gCurrentPlaylist.pagewidth, viewportWidth);
		const canvasHeight = Math.min(gCurrentPlaylist.pageheight, viewportHeight);
		
		return canvasWidth / gCurrentPlaylist.pagewidth > canvasHeight / gCurrentPlaylist.pageheight ?
		{
			width: gCurrentPlaylist.pagewidth / gCurrentPlaylist.pageheight * canvasHeight,
			height: canvasHeight
		} :
		{
			width: canvasWidth,
			height: gCurrentPlaylist.pageheight / gCurrentPlaylist.pagewidth * canvasWidth
		};
	}
}

let bounceUpdateLineLayer = null;
function UpdateLineLayer() {
	$lineLayer.hidden = true;
	clearTimeout(bounceUpdateLineLayer);
	bounceUpdateLineLayer = setTimeout(() => {
		bounceUpdateLineLayer = null;
		$lineLayer.width = Math.floor($editorCanvasContainer.clientWidth) - 1;
		$lineLayer.height = Math.floor($editorCanvasContainer.clientHeight) - 1;
		$lineLayer.hidden = false;
		LineLayerClear();
		if (gSelectedZone) {
			LineLayerHighlightZone(gSelectedZone, true);
		}
	}, 100);
}

function resizeCanvas(scaleValue) {
	if (gCurrentPlaylist == null) {
		return;
	}
	
	// Chrome has some weird layout issues where it thinks scrollbars are visible when they aren't.
	$editorCanvasContainer.style.overflow = gExpand ? '' : 'hidden';
	
	// Hide linelayer so it doesn't interfere with the width/height scaling
	// The bounce handles the hiding for us
	UpdateLineLayer();
	
	if (!gExpand) {
		$editorCanvasContainer.scrollLeft = 0;
		$editorCanvasContainer.scrollTop = 0;
	}
	
	const canvasDimensions = CalculateCanvasDimensions(scaleValue);
	
	$editorCanvas.style.width = `${gCurrentPlaylist.pagewidth}px`;
	$editorCanvas.style.height = `${gCurrentPlaylist.pageheight}px`;
	
	Object.assign($editorCanvasShadow.style, {
		boxShadow: canvasDimensions.width < 3840 && canvasDimensions.height < 2160 ? '0 0 10px black' : 'none',
		width: `${canvasDimensions.width}px`,
		height: `${canvasDimensions.height}px`
	});
	
	$canvasSlider.value = Math.round((canvasDimensions.width / gCurrentPlaylist.pagewidth) * 100);
	for (const $e of document.querySelectorAll('#editor-canvas, #previewIframe')) {
		Object.assign($e.style, {
			transformOrigin: '0 0',
			transform: `scale(${canvasDimensions.width / gCurrentPlaylist.pagewidth})`
		});
	}
	
	UpdateTimeBar();
	
	$editorCanvasContainer.scrollLeft = Math.min(canvasDimensions.width - $editorCanvasContainer.clientWidth, $editorCanvasContainer.scrollLeft);
	$editorCanvasContainer.scrollTop = Math.min(canvasDimensions.height - $editorCanvasContainer.clientHeight, $editorCanvasContainer.scrollTop);
	
	UpdateLineLayer();
	
	ResizeTimelineBars();
	
	// TODO: Is this still necessary?
	document.getElementById('webpageCropBox').parentNode.scroll();
}

function GetCachedMediaURL(account, cdnURL, uploadedtimestamp, filesize) {
	return '/get/mediaurl.php?account=' + account + '&cdnurl=' + encodeURIComponent(cdnURL) + '&t=' + moment(uploadedtimestamp + ' +0000', 'YYYY-MM-DD HH:mm:ss.S Z').unix() + '&size=' + filesize;
}

async function createThumbnail() {
	try {
		// Calling this starts generating any missing thumbnails
		var data = await get('/playlistthumbnail');
		CheckSessionExpired(data);
		if (data.success) {
			return;
		} else {
			throw new Error('playlistthumbnail failed');
		}
	} catch (exception) {
		// Do nothing
	}
}

function setWindowHeader() {
	document.getElementById('editor-header').innerText = gCurrentPlaylist == null ? 'Editor' : `Editor - ${gCurrentPlaylist.name}${gSaveFlag ? '*' : ''}, ${gCurrentPlaylist.pagewidth}x${gCurrentPlaylist.pageheight}`;
	document.getElementById('editor-interactiveflag').innerText = gCurrentPlaylist != null && gCurrentPlaylist.IsInteractive() ? '(Interactive Playlist)' : '';
}

function beforeUnload(e) {
	if (gSaveFlag) {
		e.preventDefault();
		e.returnValue = '';
	}
}

function BindOnBeforeUnload() {
	window.addEventListener('beforeunload', beforeUnload);
}

function UnbindOnBeforeUnload() {
	window.removeEventListener('beforeunload', beforeUnload);
}

function shrinkEditorTimeline() {
	const $shrinkEditorTimeline = document.getElementById('shrinkEditorTimeline');
	gShrink = !gShrink;
	gTimelineScale = gShrink ? 2 : 5;
	$shrinkEditorTimeline.title = gShrink ? 'Expand the timeline media items to see less (Default View).' : 'Shrink the timeline media items to see more. This shrunken view does not allow resizing via dragging.';
	$shrinkEditorTimeline.classList.toggle('mdi-arrow-expand-horizontal', gShrink);
	$shrinkEditorTimeline.classList.toggle('mdi-arrow-collapse-horizontal', !gShrink);
	UpdateZoneAndMediaDOM();
}

function swapZoneOrder() {
	localStorage.swapzoneorder = localStorage.swapzoneorder == 'false';
	UpdateZoneAndMediaDOM();
}

function changeCursor() {
	const $changeCursor = document.getElementById('changeCursor');
	$changeCursor.classList.toggle('mdi-cursor-default-outline');
	$changeCursor.classList.toggle('mdi-cursor-default');
	stopPreviewPressed();
	playPreviewPressed();
}

function toggleExpand(scaleValue) {
	gExpand = !gExpand;
	const $expandEditorCanvas = document.getElementById('expandEditorCanvas');
	$expandEditorCanvas.title = gExpand ? 'Fit the playlist to the canvas area.' : 'Expand the canvas area to the size of the playlist.';
	$expandEditorCanvas.classList.toggle('mdi-arrow-collapse-all', gExpand);
	$expandEditorCanvas.classList.toggle('mdi-arrow-expand-all', !gExpand);
	resizeCanvas(scaleValue);
}

function InPreviewMode() {
	return $playStop.classList.contains('mdi-stop');
}

function InitializeFonts() {
	for (var fontItr = 0; fontItr < fonts.length; ++fontItr) {
		if (fonts[fontItr] == 'Vegur') {
			fonts.splice(fontItr--, 1);
		}
	}
	
	FontLoader.loadFonts(fonts, function() {
		reloadAllZoneViews();
	});
}

function TestFilter(node) {
	document.getElementById('removeFilter').hidden = false;
	document.getElementById('showNameFilter').hidden = true;
	document.getElementById('interleave').hidden = true;
	document.getElementById('contentTree-addgroup').hidden = true;
	document.getElementById('contentTree-upload').hidden = true;
	document.getElementById('contentTree-delete').hidden = true;
	$nameFilter.hidden = false;
	if (node.group) {
		return false;
	}
	var name = node.name;
	var extension = name.replace(/^.*\./, '');
	var nameWidthoutExtension = name.replace(/\.[^\/.]+$/, '');
	var passesImageFilter = !$imageFilter.classList.contains('selected') || node.contenttreetype == 3 || ['png', 'jpg', 'jpeg', 'bmp'].indexOf(extension) != -1;
	var passesVideoFilter = !$videoFilter.classList.contains('selected') || node.contenttreetype == 5 || ['mp4', 'ogv', 'mov', 'mpg', 'mpeg', 'wmv'].indexOf(extension) != -1;
	var passesAudiofilter = !$audioFilter.classList.contains('selected') || node.contenttreetype == 4 || ['mp3', 'wav', 'ogg'].indexOf(extension) != -1;
	var passesPDFfilter = !$pdfFilter.classList.contains('selected') || node.contenttreetype == 9 || ['pdf'].indexOf(extension) != -1;
	let passesNameFilter = true;
	let textFilter = $nameFilter.value.toLowerCase();
	if (textFilter.indexOf('*') != -1 || textFilter.indexOf('.') != -1) {
		if (textFilter.indexOf('*') == -1) {
			textFilter = `*${textFilter}*`;
		}
		const escapeCharacters = /[-\/\\^$*+?.()|[\]{}]/g;
		passesNameFilter = (new RegExp(`^${textFilter.split('*').map(s => s.replace(escapeCharacters, '\\$&')).join('.*?')}$`, 'i')).test(name);
	} else if (textFilter != '') {
		passesNameFilter = nameWidthoutExtension.toLowerCase().indexOf(textFilter) != -1;
	}
	return passesImageFilter && passesVideoFilter && passesAudiofilter && passesPDFfilter && passesNameFilter;
}

async function ShowFileDialog() {
	if (DialogWindowOpen()) {
		BlinkDialog();
		return;
	}
	if ($contentTree.isBusyForUploads()) {
		toastWarning('Uploading is disabled until pending operations complete.');
		return;
	}
	let selected = await $contentTree.getSelected();
	if (selected.length != 1) {
		toastError('Select a group to upload to.');
	} else {
		document.getElementById('files').click();
	}
}

function filterByType(type) {
	// TODO: Rewrite this, lol
	switch (type) {
		case 'image':
			$imageFilter.classList.toggle('selected');
			$videoFilter.classList.remove('selected');
			$audioFilter.classList.remove('selected');
			$pdfFilter.classList.remove('selected');
			break;
		case 'video':
			$imageFilter.classList.remove('selected');
			$videoFilter.classList.toggle('selected');
			$audioFilter.classList.remove('selected');
			$pdfFilter.classList.remove('selected');
			break;
		case 'audio':
			$imageFilter.classList.remove('selected');
			$videoFilter.classList.remove('selected');
			$audioFilter.classList.toggle('selected');
			$pdfFilter.classList.remove('selected');
			break;
		case 'pdf':
			$imageFilter.classList.remove('selected');
			$videoFilter.classList.remove('selected');
			$audioFilter.classList.remove('selected');
			$pdfFilter.classList.toggle('selected');
	}
	if ($nameFilter.value == '' && [$imageFilter, $videoFilter, $audioFilter, $pdfFilter].every($filter => !$filter.classList.contains('selected'))) {
		removeFilter();
	} else {
		$contentTree.filterView(TestFilter);
	}
}

function removeFilter() {
	$contentTree.removeFilterView();
	
	document.getElementById('removeFilter').hidden = true;
	document.getElementById('showNameFilter').hidden = false;
	document.getElementById('interleave').hidden = false;
	document.getElementById('contentTree-addgroup').hidden = false;
	document.getElementById('contentTree-upload').hidden = false;
	document.getElementById('contentTree-delete').hidden = false;
	$nameFilter.hidden = true;
	$nameFilter.value = '';
	$imageFilter.classList.remove('selected');
	$videoFilter.classList.remove('selected');
	$audioFilter.classList.remove('selected');
	$pdfFilter.classList.remove('selected');
}

function onFilesChanged() {
	removeFilter();
	processFileUploads([].slice.call(document.getElementById('files').files));
	clearUploadFiles();
}

function handleFileSelect(e) {
	removeFilter();
	e.stopPropagation();
	e.preventDefault();
	processFileUploads([].slice.call(e.dataTransfer.files));
	clearUploadFiles();
}

async function showZIPConflictDialog(playlists) {
	const updateImportPlaylists = playlists => {
		const $importPlaylists = document.getElementById('import-playlists');
		let lastInputSelectionStart = null;
		let lastInputIndex = -1;
		const $input = document.activeElement;
		if ($input && $input.tagName == 'INPUT' && $importPlaylists.contains($input)) {
			lastInputSelectionStart = $input.selectionStart;
			lastInputIndex = [...$importPlaylists.children].indexOf($input);
		}
		$importPlaylists.innerHTML = '';
		$importPlaylists.append(
			$.header({},
				$.span({ text: 'Import', title: 'Whether to import the selected playlist from the archive', class: 'import-header-import' }),
				$.span({ text: 'Playlist Name', title: 'The name of the playlist when it was exported', class: 'import-header-name' }),
				$.span({ text: 'Imported Name', title: 'The name you want the imported playlist to have', class: 'import-header-importedname' }),
				$.span({ text: 'Overwrite Existing', title: 'If a playlist already exists with the entered name, then checking this will overwrite the existing playlist', class: 'import-header-overwrite' }),
				$.span({ text: 'Warnings', title: 'Any conflicts that exists such as a name being in use or a permission error', class: 'import-header-warnings' })
			),
			...playlists.map(playlist => {
				const importedNameInputInputChange = async function() {
					if (lastConflictsAbortController) {
						lastConflictsAbortController.abort();
					}
					playlist.newName = this.value;
					try {
						await checkForImportPlaylistConflicts(playlists);
						updateImportPlaylists(playlists);
					} catch (exception) {
						toastError('Error getting import conflicts.');
					}
				};
				let warning = '';
				if (playlist.newNameExists) {
					if (playlist.newNamePermission) {
						warning = 'A playlist with the entered import name already exists. You can disable the import, overwrite the existing playlist, or choose another name.';
					} else {
						warning = 'A playlist with the entered import name already exists and you do not have the permission to overwrite.';
					}
				}
				return [
					$.uicheckbox({ checked: playlist.import, change: function() {
						playlist.import = this.checked;
						if (!playlist.import) {
							playlist.newName = playlist.oldName;
							playlist.overwrite = false;
						}
						updateImportPlaylists(playlists);
					} }),
					$.input({ type: 'text', value: playlist.oldName, disabled: !playlist.import, readonly: true }),
					$.input({ type: 'text', value: playlist.newName, disabled: !playlist.import, input: importedNameInputInputChange, change: importedNameInputInputChange }),
					$.uicheckbox({ checked: playlist.overwrite, disabled: !playlist.import, change: function() {
						playlist.overwrite = this.checked;
						updateImportPlaylists(playlists);
					} }),
					$.input({ type: 'text', value: warning, title: warning, disabled: !playlist.import, readonly: true, css: { color: !playlist.import || playlist.newNameExists && playlist.newNamePermission && playlist.overwrite ? 'green' : 'red' } })
				];
			}).flat()
		);
		if (lastInputIndex != -1 && lastInputIndex < $importPlaylists.children.length) {
			$importPlaylists.children[lastInputIndex].focus();
			$importPlaylists.children[lastInputIndex].setSelectionRange(lastInputSelectionStart, lastInputSelectionStart);
		}
	};
	return await new Promise((resolve, reject) => {
		const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: {
			width: '800px'
		} },
			$.header({ text: 'Import Playlist(s)' }),
			$.div({ id: 'import-playlists', class: 'formgrid compact' }),
			$.div({ class: 'panel-buttons' },
				$.button({ text: 'Cancel', click: function() {
					$dialog.remove();
					reject();
				} }),
				$.button({ text: 'Import', click: async () => {
					try {
						await checkForImportPlaylistConflicts(playlists);
						updateImportPlaylists(playlists);
						const errors = [];
						for (const playlist of playlists) {
							if (playlist.import && playlist.newNameExists) {
								if (!playlist.newNamePermission) {
									errors.push(`A playlist with the name "${playlist.newName}" already exists and you do not have permission to overwrite it. You can choose another name or choose not to import that specific playlist.`);
								} else if (!playlist.overwrite) {
									errors.push(`A playlist with the name "${playlist.newName}" already exists. You can choose not to import the playlist "${playlist.oldName}", choose another name, or overwrite the existing playlist.`);
								}
							}
						}
						if (errors.length == 0) {
							toastSuccess('Importing has begun. Once the archive is uploaded the import process will add the media and playlist(s). This may take a few minutes.');
							$dialog.remove();
							resolve(playlists);
						} else {
							toastError(errors[0]);
							reject();
						}
					} catch (exception) {
						toastError('Error getting import conflicts.');
						reject();
					}
				} })
			)
		));
		centerOnScreen($dialog);
		updateImportPlaylists(playlists);
	});
}

async function processZIPArchive(zipArchive) {
	let zip;
	try {
		zip = await JSZip.loadAsync(zipArchive);
	} catch (exception) {
		if (exception.name === 'NotReadableError') {
			toastError('Could not open playlist archive. Note: A 2GB limit exists for archive files.');
		} else {
			toastError('Could not open playlist archive.');
		}
	}
	const playlistZIPFiles = [];
	zip.forEach(relativePath => {
		if (/^playlist[0-9]*\.json$/.test(relativePath)) {
			playlistZIPFiles.push(relativePath);
		}
	});
	if (playlistZIPFiles.length == 0) {
		toastError('Could not process playlist archive. Invalid format.');
		throw new Error();
	}
	
	const playlistsJSON = await Promise.all(playlistZIPFiles.map(async relativePath => {
		try {
			const blob = await zip.files[relativePath].async('blob');
			return JSON.parse(await blob.text());
		} catch (exception) {
			return null;
		}
	}));
	if (playlistsJSON.some(playlist => playlist == null || !playlist.hasOwnProperty('name'))) {
		// Invalid playlist
		toastError('Could not process playlist archive. Invalid format for playlists.');
		throw new Error();
	}
	
	const playlistNames = playlistsJSON.map(playlist => playlist.name);
	playlistNames.sort((a, b) => a.localeCompare(b));
	
	const playlists = playlistNames.map(playlistName => {
		return {
			import: true,
			oldName: playlistName,
			newName: playlistName,
			overwrite: false
		};
	});
	try {
		await checkForImportPlaylistConflicts(playlists);
	} catch (exception) {
		toastError('Could not get playlist name conflicts.');
		throw new Error();
	}
	return await showZIPConflictDialog(playlists);
}

let lastConflictsAbortController = null;
async function checkForImportPlaylistConflicts(playlists) {
	try {
		lastConflictsAbortController = new AbortController();
		const data = await get('/get/importplaylistsexists.php', {
			playlists: JSON.stringify(playlists)
		}, lastConflictsAbortController);
		CheckSessionExpired(data);
		if (data.success) {
			data.playlists.forEach((playlist, playlistIndex) => {
				Object.assign(playlists[playlistIndex], playlist);
			});
		} else {
			throw new Error('Unexpected success is false');
		}
	} catch (exception) {
		if (exception.name != 'AbortError') {
			NetworkingError();
		}
	}
}

function correctFilename(name) {
	let filename = name.split('.');
	if (filename.length > 1) {
		filename[filename.length - 1] = filename[filename.length - 1].toLowerCase();
		return filename.join('.');
	}
	return name;
}

async function processFileUploads(files) {
	if (files.length == 0) {
		return;
	}
	$contentTree.lock();
	// Display an error if the user uploads a folder
	try {
		for (const file of files) {
			try {
				const buffer = await file.slice(0, 100).arrayBuffer();
				if (buffer.byteLength != 100) {
					throw new Error('Folder');
				}
			} catch (exception) {
				console.log(exception);
				throw exception;
			}
		}
	} catch (exception) {
		toastError('Folders cannot be uploaded');
		$contentTree.unlock();
		return;
	}
	
	// Find the selected folder to upload into
	let selectedNode = (await $contentTree.getSelected())[0];
	if (selectedNode == undefined) {
		toastError('Select a folder to upload to and try again.');
		$contentTree.unlock();
		return;
	}
	// If the user selected a single file in the tree then upload to the file's folder.
	if (!selectedNode.group && selectedNode.parentNode) {
		selectedNode = selectedNode.parentNode;
		await $contentTree.setSelected([selectedNode.id]);
	}
	if (!selectedNode.group) {
		toastError('Invalid add permissions for parent group.');
		$contentTree.unlock();
		return;
	}
	selectedNode.tag.classList.remove('collapsed');
	
	let importPlaylists = null;
	const zipArchive = files.find(file => /^.*\.zip$/.test(file.name));
	if (zipArchive) {
		files = [zipArchive];
		try {
			importPlaylists = await processZIPArchive(zipArchive);
		} catch {
			// An archive was supplied, but an error occured and was displayed to the user
			$contentTree.unlock();
			return;
		}
	}
	const items = [];
	for (const file of files) {
		const item = {
			type: 6,
			name: correctFilename(file.name),
			parent: selectedNode.id,
			file: file
		};
		if (importPlaylists) {
			item.uploaddata = importPlaylists;
		}
		items.push(item);
	}
	$contentTree.unlock();
	let data = await $contentTree.addItems(items);
	if (!data.success) {
		if (data.errors.every(error => error.hasOwnProperty('overwrite'))) {
			if (await QuestionBox('Overwrite files?', `The following filenames exist in this folder: '${data.errors.map(error => {
				const node = items.find(node => node.addid == error.addid);
				return `${node.name}`;
			}).join("', '")}'. Do you want to overwrite the current files in this folder with the new files that match?`)) {
				// Mark all the existing items as overwrite
				for (const error of data.errors.filter(error => error.hasOwnProperty('overwrite'))) {
					const item = items.find(item => item.addid == error.addid);
					item.id = error.id;
					item.overwrite = true;
				}
				data = await $contentTree.addItems(items);
				if (!data.success) {
					document.getElementById('status').innerText = `Upload Failed: "${data.errors.map(error => error.error).join('", "')}"`;
					StartClearStatusTimer();
				}
			}
		} else {
			document.getElementById('status').innerText = `Upload Failed: "${data.errors.map(error => error.error).join('", "')}"`;
			StartClearStatusTimer();
		}
	}
}

function clearUploadFiles() {
	document.getElementById('files').value = '';
}

async function fileCheckLoop() {
	const changingFiles = (await $contentTree.getAllNodes()).filter(node => {
		// Ignore folders, error nodes, or nodes that are done
		return !node.group && !node.error && !node.done;
	});
	if (changingFiles.length != 0) {
		try {
			var data = await get('/get/contentfileviewupdate.php',
			{
				syncid: $contentTree.syncId,
				ids: JSON.stringify(changingFiles.map(node => node.id))
			});
			CheckSessionExpired(data);
			if (data.success) {
				var requiresRedraw = false;
				var requireTreeRerender = false;
				for (const media of data.media) {
					//console.log(`${media.name} ${media.contenttreetype}`);
					const oldNode = (await $contentTree.getAllNodes()).find(node => node.id == media.id);
					// If the upload hasn't started for an overwrite then it'll hold the old status. Don't overwrite the client-side data with that as we're waiting to upload an overwrite.
					if (oldNode && (!oldNode.hasOwnProperty('overwrite') || !oldNode.overwrite || (media?.data?.status && !media?.data?.status.done))) {
						// This overwrite flag is no longer needed as the status matches the server
						delete oldNode.overwrite;
						if (media.type != 1 && media.contenttreetype != 2 && !moment(oldNode.uploadedtimestamp + ' +0000', 'YYYY-MM-DD HH:mm:ss.S Z').isSame(moment(media.uploadedtimestamp + ' +0000', 'YYYY-MM-DD HH:mm:ss.S Z')) && gCurrentPlaylist != null) {
							for (var zoneItr = 0; zoneItr < gCurrentPlaylist.zones.length; ++zoneItr) {
								var zone = gCurrentPlaylist.zones[zoneItr];
								for (const playlistMedia of zone.media) {
									if (playlistMedia.type == media.type && playlistMedia.contentid == media.contentid) {
										// This operation also removes the cached images in the timeline and progressive image data so it'll reload correctly
										switch (media.type) {
											case MEDIATYPE_IMAGE:
												Object.assign(playlistMedia, (({ name, cdnurl, uploadedtimestamp, filesize }) => ({ name, cdnurl, uploadedtimestamp, filesize }))(media), { actualwidth: media.width, actualheight: media.height });
												playlistMedia.ReloadTimelineImage();
												break;
											case MEDIATYPE_VIDEO:
												Object.assign(playlistMedia, (({ name, cdnurl, uploadedtimestamp, filesize, runlength }) => ({ name, cdnurl, uploadedtimestamp, filesize, runlength }))(media), { actualwidth: media.width, actualheight: media.height });
												playlistMedia.ReloadTimelineImage();
												break;
											case MEDIATYPE_AUDIO:
												Object.assign(playlistMedia, (({ name, cdnurl, uploadedtimestamp, filesize, runlength }) => ({ name, cdnurl, uploadedtimestamp, filesize, runlength }))(media));
												break;
											case MEDIATYPE_MISC:
												Object.assign(playlistMedia, (({ name, cdnurl, uploadedtimestamp, filesize }) => ({ name, cdnurl, uploadedtimestamp, filesize }))(media));
												break;
										}
										delete playlistMedia.imageLoaded;
										delete playlistMedia.thumbnailImageLoaded;
										requiresRedraw = true;
									}
								}
							}
						}
						if (oldNode.contenttreetype != media.contenttreetype || oldNode.contentid != media.contentid || JSON.stringify(oldNode.data.status) != JSON.stringify(media.data.status)) {
							requireTreeRerender = true;
						}
						for (var property in media) {
							oldNode[property] = media[property];
						}
					}
					
					if (media.hasOwnProperty('uploadfailedrevert')) {
						toastError(`Error uploading file: '${media.name}', ${media.errorMessage}. Reverted to previous upload.`);
					} else if (media.hasOwnProperty('uploadfailed')) {
						toastError(`Error uploading file: '${media.name}', ${media.errorMessage}`);
					}
				}
				if (requiresRedraw) {
					reloadAllZoneViews();
					UpdateZoneAndMediaDOM();
				}
				if (requireTreeRerender) {
					// Slight overkill since we're just updating the leaf icon class
					$contentTree.render();
				}
			} else {
				if ('treeoutdated' in data && !('import' in data)) {
					if (!DialogWindowOpen()) {
						TreeOutOfDate('content');
					}
					await $contentTree.reload();
				}
				if ('import' in data) {
					toastSuccess('An import operation has changed the tree. A new folder for an imported template was created. The zip file in the content tree is no longer needed and was removed.');
					await $contentTree.reload();
				}
			}
		} catch (exception) {
			NetworkingError();
		}
	}
}

function StartClearStatusTimer() {
	clearTimeout(convertingFilesTimer);
	convertingFilesTimer = setTimeout(async () => {
		// Ignore content nodes that are folders, in an error state, and ones that are in the "upload ready" or "publish complete" states.
		const spinningFiles = (await $contentTree.getAllNodes()).filter(node => node.busy);
		if (spinningFiles.length != 0) {
			document.getElementById('status').innerText = `The content tree is busy with ${spinningFiles.length} file${spinningFiles.length > 1 ? 's' : ''}. Hover over them to view their status.`;
			StartClearStatusTimer();
		} else {
			document.getElementById('status').innerText = '';
		}
	}, 5000);
}

function handleDragOver(e) {
	e.stopPropagation();
	e.preventDefault();
	e.dataTransfer.dropEffect = 'copy'; // Explicitly show this is a copy.
}

async function PublishPlaylist() {
	if (gCurrentPlaylist == null) {
		toastError('Open a playlist to publish.');
		return;
	}
	const $playerPublishTree = document.getElementById('playerPublishTree');
	const selectedNode = (await $playerPublishTree.getSelected())[0];
	const selectedPlayers = (await $playerPublishTree.getSelectedAllChildren()).filter(node => !node.group);
	if (selectedPlayers == undefined) {
		toastError('Select a player or group to publish the current playlist to.');
	} else {
		try {
			// If any player the user is publishing to has a non-Playlist Only schedule and isn't set to no schedule then prompt the user that they're going to replace the schedule(s)
			await new Promise(async (resolve, reject) => {
				if (selectedPlayers.some(node => node.currentschedulename != '' && !/^.*\(Playlist Only\)$/.test(node.currentschedulename))) {
					if (await QuestionBox('Assign new playlist only schedule', 'The publish feature sends ONLY the current playlist, in a new schedule, to the selected player(s). This new "Playlist Only" schedule replaces any previously set schedule for the player(s). Are you sure you want to do this?')) {
						resolve();
					} else {
						reject();
					}
				} else {
					resolve();
				}
			});
			// The user might need to save before the playlist can be published. If the playlist is already saved then saving is optional
			await new Promise(async (resolve, reject) => {
				if (gSaveFlag) {
					// If the playlist has never been saved before then it has to be saved first before it can be published
					if (gCurrentPlaylist.id == -1) {
						try {
							await SavePlaylist(true);
							resolve();
						} catch (exception) {
							// SavePlaylist was cancelled by the user
							reject();
						}
					} else {
						try {
							switch (await ConfirmationBox('Save playlist?', 'Do you wish to save your playlist before publishing?', 'Yes', 'No', 'Cancel')) {
								case 0:
									await SavePlaylist();
									resolve();
									break;
								case 1:
									resolve();
									break;
							}
						} catch (exception) {
							reject();
						}
					}
				} else {
					resolve();
				}
			});
			// Publish
			try {
				var data = await post('/post/publishplaylist.php', {
					playlistid: gCurrentPlaylist.id,
					id: selectedNode.id
				});
				CheckSessionExpired(data);
				if (data.success) {
					toastSuccess(`Playlist published to players: "${data.updated.join('", "')}"`);
				} else {
					toastError(`There was a problem publishing the playlist. ${data.hasOwnProperty('errors') ? `Could not publish to the following players: "${data.errors.join('", "')}"` : data.error}`);
				}
			} catch (exception) {
				NetworkingError();
			}
		} catch (exception) {
			// The user cancelled the publish by either closing out of a save dialog or declining to publish when prompted about the warning on overwriting the current schedule
		}
	}
}

async function UpdateScheduledPlayers(fromSavePlaylist = false) {
	if (gCurrentPlaylist == null) {
		toastError('A playlist is required to update scheduled players.');
		return;
	}
	try {
		if (gSaveFlag) {
			// If the playlist has never been saved before then it has to be saved first before it can be pushed to players
			if (fromSavePlaylist || gCurrentPlaylist.id == -1) {
				await SavePlaylist(fromSavePlaylist);
			} else {
				switch (await ConfirmationBox('Save playlist?', 'Do you wish to save your playlist before updating players?', 'Yes', 'No', 'Cancel')) {
					case 0:
						await SavePlaylist(fromSavePlaylist);
						break;
				}
			}
		}
		try {
			var data = await post('/post/updatescheduledplayers.php', {
				playlistid: gCurrentPlaylist.id
			});
			CheckSessionExpired(data);
			if (data.success) {
				toastSuccess('Updated all players that are using this playlist');
			} else {
				toastError(`There was a problem updating the players. ${data.error}`);
			}
		} catch (exception) {
			NetworkingError();
		}
	} catch (exception) {
		// SavePlaylist was cancelled by the user
	}
}

function TogglePlaylistProperties() {
	$contentTree.removeEventListener('focus', hideAccordion);
	$contentTree.removeEventListener('scroll', hideAccordion);
	document.getElementById('playlistPropertiesHeader').classList.toggle('collapsed');
	setTimeout(() => {
		$contentTree.removeEventListener('focus', hideAccordion);
		$contentTree.removeEventListener('scroll', hideAccordion);
		$contentTree.addEventListener('focus', hideAccordion);
		$contentTree.addEventListener('scroll', hideAccordion);
	}, 100);
}

function ToggleProperties() {
	$contentTree.removeEventListener('focus', hideAccordion);
	$contentTree.removeEventListener('scroll', hideAccordion);
	document.getElementById('propertiesHeader').classList.toggle('collapsed');
	setTimeout(() => {
		$contentTree.removeEventListener('focus', hideAccordion);
		$contentTree.removeEventListener('scroll', hideAccordion);
		$contentTree.addEventListener('focus', hideAccordion);
		$contentTree.addEventListener('scroll', hideAccordion);
	}, 100);
}

function ToggleMetadata() {
	$contentTree.removeEventListener('focus', hideAccordion);
	$contentTree.removeEventListener('scroll', hideAccordion);
	document.getElementById('metadataHeader').classList.toggle('collapsed');
	setTimeout(() => {
		$contentTree.removeEventListener('focus', hideAccordion);
		$contentTree.removeEventListener('scroll', hideAccordion);
		$contentTree.addEventListener('focus', hideAccordion);
		$contentTree.addEventListener('scroll', hideAccordion);
	}, 100);
}

let startMoveType = MOVETYPE_NONE;
let canvasLastPageX = 0;
let canvasLastPageY = 0;
function CanvasPointerMove({ pageX, pageY, ctrlKey }) {
	canvasLastPageX = pageX;
	canvasLastPageY = pageY;
	if (startMoveType != MOVETYPE_NONE || gCurrentPlaylist == null || DialogWindowOpen() || InPreviewMode()) {
		return;
	}
	const thisOffset = $.offset($editorCanvas);
	var canvasX = Math.round(toCanvasScale(pageX - thisOffset.left));
	var canvasY = Math.round(toCanvasScale(pageY - thisOffset.top));
		
	const move = getMoveAt(canvasX, canvasY, ctrlKey);
	LineLayerClear();
	if (gSelectedZone !== move.zone) {
		LineLayerHighlightZone(gSelectedZone, true);
	}
	$editorCanvas.classList.remove('resize');
	if (move.zone || move.media || move.component) {
		if (move.zone.locked || move.type == MOVETYPE_NONE) {
			// The feature is not moveable
			$editorCanvas.style.cursor = '';
			LineLayerHighlightZone(move.zone, true);
			return;
		} else if (move.type & MOVETYPE_MOVE_X || move.type & MOVETYPE_MOVE_Y) {
			$editorCanvas.style.cursor = 'move';
		} else {
			$editorCanvas.style.cursor = (move.type & MOVETYPE_RESIZE_TOP ? 'n' : '') +
				                (move.type & MOVETYPE_RESIZE_BOTTOM ? 's' : '') +
				                (move.type & MOVETYPE_RESIZE_LEFT ? 'w' : '') +
				                (move.type & MOVETYPE_RESIZE_RIGHT ? 'e' : '') +
				                '-resize';
			$editorCanvas.classList.add('resize');
		}
			
		if (move.component) {
			LineLayerResizeBoxMediaComponent(move.zone, move.media, move.component, move.type);
		} else if (move.media) {
			LineLayerResizeBoxMedia(move.zone, move.media, move.type);
		} else {
			LineLayerResizeBoxZone(move.zone, move.type);
		}
	} else {
		$editorCanvas.style.cursor = '';
	}
}

window.addEventListener('load', () => {
	const dateMax = moment();
	
	if (!localStorage.hasOwnProperty('editorlayout')) {
		localStorage.editorlayout = 'landscape';
	}
	document.body.classList.add(localStorage.editorlayout);
	
	document.body.append(
		CreateNavigationMenu(),
		$.div({ id: 'editor-left' },
			$.div({ class: 'panel', css: { flex: '1 1 auto', border: 'none' } },
				$.header({ text: 'Content' }),
				$.ul({ css: { display: 'flex', flexFlow: 'row nowrap' } },
					$.li({ id: 'mediatab', text: 'Files' }),
					$.li({ id: 'widgettab', text: 'Widgets' }),
					$buttonTab = $.li({ id: 'buttontab', hidden: true },
						$.span({ text: 'Buttons' }),
						$collapseButtonGroups = $.i({ class: 'mdi mdi-format-vertical-align-center', title: 'Collapse all', css: {
							fontSize: '16px',
							marginLeft: '5px'
						} })
					)
				),
				$.div({ id: 'medias' },
					$.div({ class: 'iconlist' }, ...[
						$.div({ id: 'removeFilter', class: 'tooltip-button mdi mdi-filter-remove-outline', click: removeFilter, title: 'Remove filter', hidden: true }),
						$.div({ id: 'showNameFilter', class: 'tooltip-button mdi mdi-magnify', title: 'Search', click: function() {
							$contentTree.filterView(TestFilter);
							$nameFilter.focus();
						} }),
						$nameFilter = $.input({ type: 'text', id: 'nameFilter', placeholder: 'Search', hidden: true, input: () => {
							$contentTree.filterView(TestFilter);
						}, keydown: function(e) {
							if (e.key == 'Backspace' && this.value == '') {
								removeFilter();
							}
						} }),
						$imageFilter = $.i({ id: 'imageFilter', class: 'tooltip-button mdi mdi-file-image-outline', click: () => filterByType('image'), title: 'Show only images' }),
						$videoFilter = $.i({ id: 'videoFilter', class: 'tooltip-button mdi mdi-file-video-outline', click: () => filterByType('video'), title: 'Show only videos' }),
						$audioFilter = $.i({ id: 'audioFilter', class: 'tooltip-button mdi mdi-file-music-outline', click: () => filterByType('audio'), title: 'Show only audio' }),
						$pdfFilter = $.i({ id: 'pdfFilter', class: 'tooltip-button mdi mdi-file-pdf-outline', click: () => filterByType('pdf'), title: 'Show only PDF' }),
						$.i({ id: 'interleave', class: 'tooltip-button mdi mdi-layers-outline', title: 'Toggle interleave mode. Select a range of media then drop media onto the zone to interleave media between the selection.', css: { margin: '0 5px' }, click: function() {
							this.classList.toggle('selected');
						} }),
						permissions['Editor Content Tree'] ? [
							$.i({ id: 'contentTree-addgroup', class: 'tooltip-button mdi mdi-folder-plus-outline', click: () => $contentTree.addGroup(), title: 'Add group' }),
							$.i({ id: 'contentTree-upload', class: 'tooltip-button mdi mdi-upload', click: ShowFileDialog, title: 'Upload' }),
							$.i({ id: 'contentTree-delete', class: 'tooltip-button mdi mdi-trash-can-outline', click: () => $contentTree.deleteSelected(), title: 'Delete' })
						] : []
					].flat()),
					$contentTree = $.uitree({ id: 'contentTree', class: 'treeContainer', config: {
						tree: 'content',
						multiselect: true,
						filterDraggable: function(node, index, nodes) {
							if (!node.group) {
								// If any of the outputs had a conversion or CDN issue then the node isn't busy, it's in an error state
								if (node.busy && !node.error) {
									return false;
								}
							}
							// If every media is an image, video, audio, or misc then allow them all
							if (nodes.every(node => node.contenttreetype == 1 || node.contenttreetype == 3 || node.contenttreetype == 4 || node.contenttreetype == 5 || node.contenttreetype == 8)) {
								return true;
							} else if (nodes.some(node => node.contenttreetype == 9) &&
										nodes.some(node => node.contenttreetype == 3 || node.contenttreetype == 4 || node.contenttreetype == 5 || node.contenttreetype == 8)) {
								// If a PDF is selected along with other media then unselect the PDF
								return node.contenttreetype != 9;
							} else {
								const pdfs = nodes.filter(node => node.contenttreetype == 9);
								return node === pdfs[0]; // Only select the first pdf. Can't drop more than one PDF at a time
							}
						},
						dragHelper: async function(e, region, thumbnail) {
							if (region == 'content') {
								const selectedNodes = (await this.getSelectedFilteredChildren()).filter(node => !node.group);
								if (selectedNodes.length == 0) {
									return false;
								}
								const $drop = $.ul({ class: ['tree-dragging', thumbnail, 'leafs'] },
									...selectedNodes.map(node => node.tag.querySelector(':scope > .node').cloneNode(true))
								);
								return $drop;
							}
						},
						leafIcon: node => GetContentIcon(node),
						editable: node => node.group && node.parentNode != null,
						tooltip: function(node) {
							if (!node.group) {
								this.addEventListener('pointerenter', function() {
									document.getElementById('contentTreeThumbnail')?.remove();
									const contentTreeOffset = $.offset($contentTree);
									const $thumbnail = document.body.appendChild($.div({ id: 'contentTreeThumbnail', css: {
										position: 'absolute',
										display: 'inline-block',
										pointerEvents: 'none',
										left: `${contentTreeOffset.left + $contentTree.offsetWidth + 5}px`,
										top: 0,
										backgroundColor: 'var(--header-color)',
										color: 'white',
										padding: '3px 6px',
										borderRadius: '2px',
										fontSize: '14px',
										boxShadow: '2px 2px 3px rgba(0,0,0,0.5)'
									} },
										$.div({ text: node.name, css: { fontSize: '12px' } })
									));
									
									const setTop = () => {
										const thisOffset = $.offset(this);
										$thumbnail.style.top = `${thisOffset.top + $thumbnail.offsetHeight <= document.documentElement.clientHeight ? thisOffset.top : document.documentElement.clientHeight - $thumbnail.offsetHeight - 15}px`;
									};
									
									if (node.done) {
										switch (node.contenttreetype) {
											case 3:
												var thumbnailURL = node.cdnurl.split('.');
												thumbnailURL.splice(-2, 0, 'thumbnail');
												$thumbnail.hidden = true;
												$thumbnail.appendChild($.img({ src: GetCachedMediaURL(account, thumbnailURL.join('.'), node.uploadedtimestamp, node.filesize), css: {
													display: 'block',
													maxHeight: 0,
												}, load: () => {
													$thumbnail.style.top = 0;
													$thumbnail.querySelector('img').style.maxHeight = '500px';
													$thumbnail.hidden = false;
													setTop();
												} }));
												break;
											case 4:
												setTop();
												break;
											case 5:
												var thumbnailURL = node.cdnurl.split('.');
												thumbnailURL.splice(-2, 0, 'thumbnail');
												thumbnailURL.splice(-1, 1, 'jpg');
												$thumbnail.hidden = true;
												$thumbnail.appendChild($.img({ src: GetCachedMediaURL(account, thumbnailURL.join('.'), node.uploadedtimestamp, node.filesize), css: {
													display: 'block',
													maxHeight: 0
												}, load: () => {
													$thumbnail.style.top = 0;
													$thumbnail.querySelector('img').style.maxHeight = '500px';
													$thumbnail.hidden = false;
													setTop();
												} }));
												break;
											case 8:
												setTop();
												break;
											case 9:
												var thumbnailURL = node.cdnurl.split('.');
												thumbnailURL.splice(-2, 0, 'thumbnail');
												thumbnailURL.splice(-1, 1, 'jpg');
												$thumbnail.hidden = true;
												$thumbnail.appendChild($.img({ src: GetCachedMediaURL(account, thumbnailURL.join('.'), node.uploadedtimestamp, node.filesize), css: {
													display: 'block',
													maxHeight: 0,
												}, load: () => {
													$thumbnail.style.top = 0;
													$thumbnail.querySelector('img').style.maxHeight = '500px';
													$thumbnail.hidden = false;
													setTop();
												} }));
												break;
										}
										
										if (node.contenttreetype == 3 || node.contenttreetype == 5) {
											$thumbnail.appendChild($.div({ text: `Width: ${node.width}`, css: { fontSize: '14px' } }));
											$thumbnail.appendChild($.div({ text: `Height: ${node.height}`, css: { fontSize: '14px' } }));
										}
										if (node.contenttreetype == 4 || node.contenttreetype == 5) {
											var runLength = parseInt(node.runlength, 10);
											var time = secondsToTime(runLength);
											var timeStr = '';
											if (time.h > 0) {
												timeStr = time.h + ':' + pad2(time.m) + ':' + pad2(time.s);
											} else if (time.m > 0) {
												timeStr = time.m + ':' + pad2(time.s);
											} else {
												timeStr = '0:' + pad2(time.s);
											}
											$thumbnail.appendChild($.div({ text: `Duration: ${timeStr}`, css: { fontSize: '14px' } }));
										}
										if (!node.group && node.contenttreetype != 2 && node.contenttreetype != 6 && node.contenttreetype != 7) {
											$thumbnail.appendChild($.div({ text: `File Size: ${humanFileSize(node.filesize, true)}`, css: { fontSize: '12px' } }));
											$thumbnail.appendChild($.div({ text: `Upload Date: ${moment(`${node.uploadedtimestamp} +0000`, 'YYYY-MM-DD HH:mm:ss.S Z').format('M/D/YY h:mm a')}`, css: { fontSize: '12px' } }));
										}
										if (node?.data?.status?.overwrite) {
											$thumbnail.appendChild($.div({ text: 'Overwrite Failed. See below.', css: { fontSize: '10px', fontWeight: 'bold' } }));
											for (const output of ['original', ...node.data.status.overwrite.output]) {
												const outputTask = node.data.status.overwrite[output];
												if (outputTask.error) {
													$thumbnail.appendChild($.div({ text: `${output} ${outputTask.error}`, css: { fontSize: '10px' } }));
												} else {
													$thumbnail.append(
														$.div({ text: `Overwrite: ${output}`, css: { fontSize: '9px', fontWeight: 'bold', textTransform: 'capitalize' } }),
														$.div({ text: `Conversion: ${outputTask.convert.error ? outputTask.convert.error : (outputTask.convert.progress ? `${outputTask.convert.progress}%` : 'Not Started')}`, css: { fontSize: '8px', textIndent: '3px', color: outputTask.convert.error ? 'var(--tree-error-color)' : '' } }),
														$.div({ text: `CDN Upload: ${outputTask.cdn.error ? outputTask.cdn.error : (outputTask.cdn.progress ? `${outputTask.cdn.progress}%` : 'Not Started')}`, css: { fontSize: '8px', textIndent: '3px', color: outputTask.cdn.error ? 'var(--tree-error-color)' : '' } })
													);
												}
											}
										}
									} else {
										if (node.data.status.upload.error) {
											$thumbnail.appendChild(
												$.div({},
													$.span({ text: `Upload`, css: { fontSize: '12px', fontWeight: 'bold' } }),
													$.span({ text: node.data.status.upload.error, css: { fontSize: '12px', marginLeft: '1ch', color: 'var(--tree-error-color)' } })
												)
											);
										} else if (node.data.status.upload.progress) {
											$thumbnail.appendChild(
												$.div({},
													$.span({ text: `Upload`, css: { fontSize: '12px', fontWeight: 'bold' } }),
													$.span({ text: `${node.data.status.upload.progress}%`, css: { fontSize: '12px', marginLeft: '1ch' } })
												)
											);
										} else {
											if (node.hasOwnProperty('file')) {
												$thumbnail.appendChild(
													$.div({ text: 'Upload pending for local file...', css: { fontSize: '12px', fontWeight: 'bold' } })
												);
											} else {
												$thumbnail.append(
													$.div({ text: 'No local file. Another user might have uploaded.', css: { fontSize: '12px', fontWeight: 'bold' } }),
													$.div({ text: 'If you refreshed the page you need to upload again to reassociate the local file to this content item.', css: { fontSize: '12px', fontWeight: 'bold' } })
												);
											}
										}
										for (const output of ['original', ...node.data.status.output]) {
											const outputTask = node.data.status[output];
											if (outputTask.error) {
												$thumbnail.appendChild(
													$.div({ text: `${output} ${outputTask.error}`, css: { fontSize: '12px' } })
												);
											} else {
												$thumbnail.append(
													$.div({ text: `${output}`, css: { fontSize: '12px', fontWeight: 'bold', textTransform: 'capitalize' } }),
													$.div({ text: `Conversion: ${outputTask.convert.error ? outputTask.convert.error : (outputTask.convert.progress ? `${outputTask.convert.progress}%` : 'Not Started')}`, css: { fontSize: '10px', textIndent: '3px', color: outputTask.convert.error ? 'var(--tree-error-color)' : '' } }),
													$.div({ text: `CDN Upload: ${outputTask.cdn.error ? outputTask.cdn.error : (outputTask.cdn.progress ? `${outputTask.cdn.progress}%` : 'Not Started')}`, css: { fontSize: '10px', textIndent: '3px', color: outputTask.cdn.error ? 'var(--tree-error-color)' : '' } })
												);
											}
										}
										setTop();
									}
								});
								this.addEventListener('pointerleave', () => {
									document.getElementById('contentTreeThumbnail')?.remove();
								});
							}
						},
						dragAppendTo: 'body',
						deleteConfirmation: async (selectedNodes, selectedAllChildren, deleteNodes) => {
							if (selectedNodes.length == 0) {
								toastError('Select a group or media to delete.');
								return;
							}
							const selectedContent = selectedAllChildren.filter(node => !node.group);
							if (gCurrentPlaylist != null) {
								for (const node of selectedContent) {
									if (node.type != 6) {
										for (const zone of gCurrentPlaylist.zones) {
											for (const media of zone.media) {
												if (media.type == node.type && media.contentid == node.contentid) {
													toastError(`The media '${media.name}' is being used in the current playlist and can not be deleted until removed.`);
													return;
												}
											}
										}
									}
								}
							}
							if (selectedAllChildren.length > 1) {
								const folders = selectedAllChildren.filter(node => node.group);
								const media = selectedAllChildren.filter(node => !node.group);
								if (media.length == 0 || folders.length == 0 && media.length == 1) {
									// Single item, no confirmation
								} else if (folders.length == 0) {
									await ConfirmationBox('Remove all?', `You selected ${media.length} media items. Are you sure you want to remove them all?`, 'Yes', 'No');
								} else {
									await ConfirmationBox('Remove all?', `You selected ${folders.length} folder${folders.length == 1 ? '' : 's'} and ${media.length} media item${media.length == 1 ? '' : 's'}. Are you sure you want to remove them all?`, 'Yes', 'No');
								}
							}
							await deleteNodes(selectedAllChildren);
						},
						doubleClick: (e, node) => {
							if (!node.group) {
								PreviewContent(account, node);
								return true;
							}
							return false;
						}
					}, load: async () => {
						if ((await $contentTree.getAllNodes()).some(node => node.group)) {
							for (const $e of document.querySelectorAll('#contentTree-addgroup, #contentTree-delete, #contentTree-upload')) $e.hidden = false;
						}
					} }),
					$.input({ type: 'file', id: 'files', name: 'files[]', css: { display: 'none' }, change: onFilesChanged, multiple: true })
				),
				$.div({ id: 'widgets', scroll: hideAccordion }),
				$.div({ id: 'interactivebuttons' })
			),
			$.div({ id: 'playlistPropertiesHeader', class: 'accordion-header tooltip-button', click: TogglePlaylistProperties },
				$.span({ text: 'Playlist Properties' }),
				$.span({ id: 'playlist-editlayout', class: 'mdi mdi-lead-pencil tooltip', title: 'Edit dimensions' }),
				$.i({ class: 'mdi mdi-play mdi-rotate-90' })
			),
			$.div({ id: 'playlistProperties', class: 'propertygrid' },
				$.label({ for: 'playlist-pagewidth', class: 'tooltip-help playlist-layout', title: 'The horizontal dimensions in pixels', text: 'Width' }),
				$.input({ type: 'text', id: 'playlist-pagewidth', class: 'tooltip-help playlist-layout', maxlength: '5', title: 'Press the edit dimensions icon above to edit this value',  readonly: true }),
				$.label({ for: 'playlist-pageheight', class: 'tooltip-help playlist-layout', title: 'The vertical dimensions in pixels', text: 'Height' }),
				$.input({ type: 'text', id: 'playlist-pageheight', class: 'tooltip-help playlist-layout', maxlength: '5', title: 'Press the edit dimensions icon above to edit this value', readonly: true }),
			),
			$.div({ id: 'propertiesHeader', class: 'accordion-header tooltip-button', click: ToggleProperties },
				$.span({ text: 'Zone Properties' }),
				$.i({ class: 'mdi mdi-play mdi-rotate-90' })
			),
			$.div({ id: 'properties', class: 'propertygrid' },
				$.ul({ class: 'iconlist label' },
					$.li({ id: 'properties-expandmedia', class: 'tooltip-button mdi mdi-arrow-expand-all', click: expandCurrentMedia, title: 'Expand to fit zone' })
				),
				$zoneSelect = $.select({ id: 'zoneSelect', class: 'label' }),
				$mediaSelect = $.ul({ id: 'mediaSelect', class: 'label', keyup: e => {
					if (e.keyCode == 46) {
						deleteCurrentMedia(true);
						e.preventDefault();
					}
				} }),
				$.header({ id: 'media-header-layout', text: 'Layout' },
					$.span({},
						$.label({ for: 'media-x', title: 'Offset Left', class: 'tooltip' }),
						$.span({ text: ', ' }),
						$.label({ for: 'media-y', title: 'Offset Top', class: 'tooltip' }),
						$.span({ text: '   ' }),
						$.label({ for: 'media-width', title: 'Width', class: 'tooltip' }),
						$.span({ text: 'x' }),
						$.label({ for: 'media-height', title: 'Height', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Layout' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-x', class: 'tooltip-help media-layout', title: 'The offset in pixels from the left of the zone. Holding ctrl in the canvas allows editing this by dragging', text: 'Offset Left' }),
					$.input({ type: 'text', id: 'media-x', class: 'media-layout', maxlength: '5' }),
					$.label({ for: 'media-y', class: 'tooltip-help media-layout', title: 'The offset in pixels from the top of the zone. Holding ctrl in the canvas allows editing this by dragging', text: 'Offset Top' }),
					$.input({ type: 'text', id: 'media-y', class: 'media-layout', maxlength: '5' }),
					
					$.label({ for: 'media-width', class: 'tooltip-help media-layout', title: 'The horizontal dimensions in pixels. Holding ctrl in the canvas allows editing this by dragging', text: 'Width' }),
					$.input({ type: 'text', id: 'media-width', class: 'media-layout', maxlength: '5' }),
					$.label({ for: 'media-height', class: 'tooltip-help media-layout', title: 'The vertical dimensions in pixels. Holding ctrl in the canvas allows editing this by dragging', text: 'Height' }),
					$.input({ type: 'text', id: 'media-height', class: 'media-layout', maxlength: '5' })
				),
				
				$.header({ id: 'zone-header-layout', text: 'Layout' },
					$.span({},
						$.label({ for: 'zone-x', title: 'Offset Left', class: 'tooltip' }),
						$.span({ text: ', ' }),
						$.label({ for: 'zone-y', title: 'Offset Top', class: 'tooltip' }),
						$.span({ text: '   ' }),
						$.label({ for: 'zone-width', title: 'Width', class: 'tooltip' }),
						$.span({ text: 'x' }),
						$.label({ for: 'zone-height', title: 'Height', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Layout' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'zone-x', class: 'tooltip-help zone-layout', title: 'The offset in pixels from the left of the screen', text: 'Offset Left' }),
					$.input({ type: 'text', id: 'zone-x', class: 'zone-layout', maxlength: '5' }),
					$.label({ for: 'zone-y', class: 'tooltip-help zone-layout', title: 'The offset in pixels from the top of the screen', text: 'Offset Top' }),
					$.input({ type: 'text', id: 'zone-y', class: 'zone-layout', maxlength: '5' }),
					
					$.label({ for: 'zone-width', class: 'tooltip-help zone-layout', title: 'The horizontal dimensions in pixels', text: 'Width' }),
					$.input({ type: 'text', id: 'zone-width', class: 'zone-layout', maxlength: '5' }),
					$.label({ for: 'zone-height', class: 'tooltip-help zone-layout', title: 'The vertical dimensions in pixels', text: 'Height' }),
					$.input({ type: 'text', id: 'zone-height', class: 'zone-layout', maxlength: '5' })
				),
				
				$.header({ id: 'zone-header-details', text: 'Zone Details' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Details' })
				),
				$.div({ class: 'group' },
					$.label({ css: { display: 'none' }, for: 'zone-id', text: 'ID:' }),
					$.input({ css: { display: 'none' }, type: 'text', id: 'zone-id', size: '6', readonly: true }),
					
					$.label({ for: 'zone-name', class: 'tooltip-help details', title: 'The name of the zone', text: 'Name' }),
					$.input({ type: 'text', id: 'zone-name', class: 'details' }),
					
					$.label({ for: 'zone-duration', class: 'tooltip-help details', title: 'The sum of the runlength for all the media in this zone', text: 'Duration' }),
					$.input({ type: 'text', id: 'zone-duration', class: 'details', size: '17', readonly: true }),
					
					$.label({ for: 'zone-locked', class: 'tooltip-help details', title: 'Whether to lock the layout for the zone so that changes are not made', text: 'Lock Layout' }),
					$.uicheckbox({ id: 'zone-locked', class: 'details' }),
					
					$.label({ for: 'zone-volume', class: 'tooltip-help details', title: 'Scales all the volume in the zone (where 100% is full volume)', text: 'Volume' }),
					$.select({ id: 'zone-volume', class: 'details' },
						$.option({ value: '100', text: '100%' }),
						$.option({ value: '75', text: '75%' }),
						$.option({ value: '50', text: '50%' }),
						$.option({ value: '25', text: '25%' }),
						$.option({ value: '0', text: 'Mute' })
					)
				),
				
				$.header({ id: 'media-header-time', text: 'Time' },
					$.span({},
						$.label({ for: 'media-starttime', title: 'Start', class: 'tooltip' }),
						$.span({ text: ', ' }),
						$.label({ for: 'media-runlength', title: 'Run-Length', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Time' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-starttime', class: 'tooltip-help time', title: 'The time in seconds (relative to the start of the zone) before the media will show', text: 'Start' }),
					$.input({ type: 'text', id: 'media-starttime', class: 'time', maxlength: '5', readonly: true }),
					
					$.label({ for: 'media-runlength', class: 'tooltip-help time', title: 'The duration in seconds the content will show', text: 'Run-Length' }),
					$.input({ type: 'text', id: 'media-runlength', class: 'time', maxlength: '5' })
				),
				
				$.header({ id: 'media-header-audio', text: 'Audio' },
					$.span({},
						$.label({ for: 'media-volume', title: 'Volume', class: 'tooltip' }),
						$.span({ text: ', ' }),
						$.label({ for: 'media-ducking', title: 'Ducking', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Audio' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-volume', class: 'tooltip-help audio', title: 'Sets the volume of this media (where 100% is full volume)', text: 'Volume' }),
					$.select({ id: 'media-volume', class: 'audio' },
						$.option({ value: '100', text: '100%' }),
						$.option({ value: '75', text: '75%' }),
						$.option({ value: '50', text: '50%' }),
						$.option({ value: '25', text: '25%' }),
						$.option({ value: '0', text: 'Mute' })
					),
					
					$.label({ for: 'media-ducking', class: 'tooltip-help audio', title: 'The volume used for this media when other media with audio are playing. (100% means full volume)', text: 'Ducking' }),
					$.select({ id: 'media-ducking', class: 'audio' },
						$.option({ value: '100', text: '100%' }),
						$.option({ value: '75', text: '75%' }),
						$.option({ value: '50', text: '50%' }),
						$.option({ value: '25', text: '25%' }),
						$.option({ value: '0', text: 'Mute' })
					)
				),
				
				$.header({ id: 'media-header-transition', text: 'Transitions' },
					$.span({},
						$.label({ for: 'media-entrytransitiontype', title: 'Entry Transition', class: 'tooltip' }),
						$.span({ text: ' ' }),
						$.label({ for: 'media-entrytransitionduration', title: 'Entry Duration', class: 'tooltip' }),
						$.span({ text: ' - ' }),
						$.label({ for: 'media-exittransitiontype', title: 'Exit Transition', class: 'tooltip' }),
						$.span({ text: ' ' }),
						$.label({ for: 'media-exittransitionduration', title: 'Exit Duration', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Transitions' })
				),
				$.div({ class: 'group' },
					$.label({ class: 'tooltip-help transition', title: 'The transition played between the last media item', text: 'Entry Transition' }),
					$.select({ id: 'media-entrytransitiontype', class: 'transition' }),
					$.label({ class: 'tooltip-help transition', css: { paddingLeft: '16px' }, title: 'The transition played between the last media item', text: 'Entry Duration' }),
					$.select({ id: 'media-entrytransitionduration', class: 'transition' },
						$.option({ value: '3000', text: 'Slow' }),
						$.option({ value: '2000', text: 'Normal' }),
						$.option({ value: '1000', text: 'Fast' })
					),
					
					$.label({ class: 'tooltip-help transition', title: 'The transition played between the next media item', text: 'Exit Transition' }),
					$.select({ id: 'media-exittransitiontype', class: 'transition' }),
					$.label({ class: 'tooltip-help transition', css: { paddingLeft: '16px' }, title: 'The transition played between the next media item', text: 'Exit Duration' }),
					$.select({ id: 'media-exittransitionduration', class: 'transition' },
						$.option({ value: '3000', text: 'Slow' }),
						$.option({ value: '2000', text: 'Normal' }),
						$.option({ value: '1000', text: 'Fast' })
					)
				),
				
				$.header({ id: 'media-header-mediaoptions', text: 'Media Options' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Media Options' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-fillstyleid', class: 'tooltip-help mediaoptions', title: 'None uses the layout fields to position and size the media. Cover fills the zone while preserving the aspect ratio. Aspect ratio extends the width or height to the zone boundaries while preserving the aspect ratio.', text: 'Fill Style' }),
					$.select({ id: 'media-fillstyleid', class: 'mediaoptions' },
						$.option({ value: '1', text: 'None' }),
						$.option({ value: '2', text: 'Cover' }),
						$.option({ value: '3', text: 'Aspect Ratio' })
					),
					
					$.label({ for: 'media-zoompan', class: 'tooltip-help mediaoptions', title: 'Creates a slow zoom and pan animation that is useful for backgrounds', text: 'Zoom Pan' }),
					$.select({ id: 'media-zoompan', class: 'mediaoptions' },
						$.option({ value: '0', text: 'None' }),
						$.option({ value: '1', text: 'Standard' })
					)
				),
				
				$.header({ id: 'media-header-weboptions', text: 'Web Options' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Web Options' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-fillstyle', class: 'tooltip-help weboptions', title: 'None uses the layout fields to position and size the media. Cover fills the zone while preserving the aspect ratio. Aspect ratio extends the width or height to the zone boundaries while preserving the aspect ratio.', text: 'Fill Style' }),
					$.select({ id: 'media-fillstyle', class: 'weboptions' },
						$.option({ value: '1', text: 'None' }),
						$.option({ value: '2', text: 'Cover' }),
						$.option({ value: '3', text: 'Aspect Ratio' })
					),
					$.div({ css: { padding: '5px', fontStyle: 'italic' }, class: 'noselect label weboptions', text: 'Double click the web in the timeline to view crop options.' })
				),
				
				$.header({ id: 'media-header-scheduling', class: 'tooltip-help', title: 'Scheduling media is a way for playlists to selectively show media between time ranges or with dayparting between two dates.', text: 'Scheduling' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Scheduling' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-scheduled', class: 'tooltip-help scheduling', title: 'Enabling this allows the media to be scheduled between specific dates and optionally allows dayparting options for days of the week.', text: 'Enable' }),
					$.uicheckbox({ id: 'media-scheduled', class: 'scheduling' }),
					
					$.label({ for: 'media-utc', class: 'tooltip-help scheduling', title: 'Whether this scheduling is relative to UTC time. (Note that the times, below, are displayed using the timezone in the Configuration page for convenience when this is enabled). If the goal is to have the media schedule happen in every timezone at the same time, then use UTC. If the goal is to have a media schedule happen in its own timezone, then uncheck this. For example, if you need to show media between 12AM-1PM in each timezone, then you would uncheck UTC.', text: 'UTC' }),
					$.uicheckbox({ id: 'media-utc', class: 'scheduling' }),
					
					$.label({ for: 'media-scheduledaterange', class: 'tooltip-help scheduling', css: { paddingLeft: '16px' }, title: 'Enabling this allows the media to be scheduled between specific dates.', text: 'Date Range' }),
					$.uicheckbox({ id: 'media-scheduledaterange', class: 'scheduling' }),
					
					$.label({ for: 'media-schedulestartdate', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'Start date and time. The media will only display after this time.', text: 'Start Date' }),
					$.input({ type: 'date', id: 'media-schedulestartdate', class: 'scheduling', min: moment().subtract(1, 'year').format('YYYY-MM-DD'), max: moment().add(10, 'year').format('YYYY-MM-DD') }),
					$.label({ for: 'media-schedulestarttime', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'Start date and time. The media will only display after this time.', text: 'Start Time' }),
					$.input({ type: 'time', id: 'media-schedulestarttime', class: 'scheduling', step: '1' }),
					
					$.label({ for: 'media-scheduleenddate', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'End date and time. The media will only display before this time.', text: 'End Date' }),
					$.input({ type: 'date', id: 'media-scheduleenddate', class: 'scheduling', min: moment().subtract(1, 'year').format('YYYY-MM-DD'), max: moment().add(10, 'year').format('YYYY-MM-DD') }),
					$.label({ for: 'media-scheduleendtime', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'End date and time. The media will only display before this time.', text: 'End Time' }),
					$.input({ type: 'time', id: 'media-scheduleendtime', class: 'scheduling', step: '1' }),
					
					$.label({ for: 'media-scheduledayparting', class: 'tooltip-help scheduling', css: { paddingLeft: '16px' }, title: 'Dayparting schedules media to only play on selected days of the week between specific times.', text: 'Enable Dayparting' }),
					$.uicheckbox({ id: 'media-scheduledayparting', class: 'scheduling' }),
					
					$.label({ class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'Days the media should display on', text: 'Days' }),
					$.span({ id: 'media-scheduledow', class: 'checkbox-group scheduling' },
						$.uicheckbox({ id: 'media-sunday', class: 'propertygrid-checkbox', text: 'S' }),
						$.uicheckbox({ id: 'media-monday', class: 'propertygrid-checkbox', text: 'M' }),
						$.uicheckbox({ id: 'media-tuesday', class: 'propertygrid-checkbox', text: 'T' }),
						$.uicheckbox({ id: 'media-wednesday', class: 'propertygrid-checkbox', text: 'W' }),
						$.uicheckbox({ id: 'media-thursday', class: 'propertygrid-checkbox', text: 'T' }),
						$.uicheckbox({ id: 'media-friday', class: 'propertygrid-checkbox', text: 'F' }),
						$.uicheckbox({ id: 'media-saturday', class: 'propertygrid-checkbox', text: 'S' }),
					),
					
					$.label({ for: 'media-daypartstarttime', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'Dayparting start time. Media will only play for the selected days between these times.', text: 'Start Time' }),
					$.input({ type: 'time', id: 'media-daypartstarttime', class: 'scheduling', step: '1' }),
					$.label({ for: 'media-daypartendtime', class: 'tooltip-help scheduling', css: { paddingLeft: '24px' }, title: 'Dayparting end time. Media will only play for the selected days between these times.', text: 'End Time' }),
					$.input({ type: 'time', id: 'media-daypartendtime', class: 'scheduling', step: '1' })
				),
				
				$.header({ id: 'media-header-buttonlayout', text: 'Button Layout' },
					$.span({},
						$.label({ for: 'media-button-x', title: 'Offset Left', class: 'tooltip' }),
						$.span({ text: ', ' }),
						$.label({ for: 'media-button-y', title: 'Offset Top', class: 'tooltip' }),
						$.span({ text: '   ' }),
						$.label({ for: 'media-button-width', title: 'Width', class: 'tooltip' }),
						$.span({ text: 'x' }),
						$.label({ for: 'media-button-height', title: 'Height', class: 'tooltip' })
					),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Button Layout' })
				),
				$.div({ class: 'group' },
					$.label({ for: 'media-button-x', class: 'tooltip-help buttonlayout', title: 'The offset in pixels from the left of the button group', text: 'Offset Left' }),
					$.input({ type: 'text', id: 'media-button-x', class: 'buttonlayout', maxlength: '5' }),
					$.label({ for: 'media-button-y', class: 'tooltip-help buttonlayout', title: 'The offset in pixels from the top of the button group', text: 'Offset Top' }),
					$.input({ type: 'text', id: 'media-button-y', class: 'buttonlayout', maxlength: '5' }),
					
					$.label({ for: 'media-button-width', class: 'tooltip-help buttonlayout', title: 'The horizontal dimensions in pixels', text: 'Width' }),
					$.input({ type: 'text', id: 'media-button-width', class: 'buttonlayout', maxlength: '5' }),
					$.label({ for: 'media-button-height', class: 'tooltip-help buttonlayout', title: 'The vertical dimensions in pixels', text: 'Height' }),
					$.input({ type: 'text', id: 'media-button-height', class: 'buttonlayout', maxlength: '5' })
				),
				
				$.header({ id: 'media-header-buttonmetadata', text: 'Button Metadata' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Button Metadata' })
				),
				$.div({ class: 'group' },
					$.div({ id: 'media-button-metadata', class: 'label metadatagrid', css: { gridColumn: '1 / -1' } })
				),
				
				$.div({ id: 'media-header-buttonoptions', css: { padding: '5px', fontStyle: 'italic' }, class: 'label buttonlayout', text: 'When in button edit mode double click a button to edit it. More options are in the buttons tab.' }),
				
				$.header({ id: 'media-header-widgetoptions', text: 'Widget Options' },
					$.span(),
					$.span({ class: 'mdi mdi-play mdi-rotate-90 tooltip', title: 'Expand/Collapse Button Layout' })
				),
				$.div({ class: 'group' },
					$.div({ id: 'media-header-widgetnooptions', css: { gridColumn: '1 / -1', padding: '5px', fontStyle: 'italic' }, class: 'noselect label widgetoptions', text: 'Double click the widget in the timeline to view options.' })
				)
			),
			$.div({ id: 'metadataHeader', class: 'accordion-header tooltip-button', click: ToggleMetadata },
				$.span({ text: 'Metadata' }),
				$.i({ class: 'mdi mdi-play mdi-rotate-90' })
			),
			$.div({ id: 'contentTree-metadata', class: 'metadatagrid' })
		),
		$.div({ id: 'editor-content-resize', class: 'horizontal-resize' }),
		$.header({ id: 'editor-main-header' },
			$editorLayout = $.span({ id: 'editor-layout', class: 'tooltip-button mdi' }),
			$.span({ id: 'editor-header', class: 'noselect' }),
			$.span({ id: 'editor-interactiveflag' }),
			$.span({ id: 'status' }),
			$.button({ text: 'Publish on Save', title: 'If enabled, then every time the playlist is saved it will be pushed to any player that has it scheduled.', css: {
				display: 'flex',
				padding: '1px 4px',
				margin: '3px 5px'
			}, click: () => {
				document.getElementById('editor-saveonpublish').click();
			} },
				$.uicheckbox({ id: 'editor-saveonpublish', checked: !localStorage.hasOwnProperty('saveonpublish') ? false : localStorage['saveonpublish'] == 'true', css: {
					marginLeft: '4px'
				}, change: function() {
					localStorage['saveonpublish'] = this.checked;
				} })
			),
			$.span({ id: 'editor-new', class: 'tooltip-button mdi mdi-file-outline editor-header-requires-playlist-permission', title: 'New' }),
			$.span({ id: 'editor-open', class: 'tooltip-button mdi mdi-folder-open-outline editor-header-requires-playlist-permission', title: 'Open (Ctrl+O)' }),
			$.span({ id: 'editor-save', class: 'tooltip-button mdi mdi-content-save-outline editor-header-requires-playlist-permission', title: 'Save (Ctrl+S)' }),
			$.span({ id: 'editor-menu', class: 'mdi mdi-menu' })
		),
		$.div({ id: 'toolbars' },
			$.div({ id: 'text2toolbar', hidden : true },
				$.div({ id: 'text2text', hidden: true }),
				$.div({ id: 'text2inlineimages', hidden: true }),
				$.select({ id: 'text2stretch', title: 'Stretch', class: 'sampleSelect', change: function() {
					gCurrentWidget.text.ApplyStyle('stretch', this.value);
				} },
					$.option({ text: 'None', value: 'normal' }),
					$.option({ text: 'Condensed', value: 'condensed' }),
					$.option({ text: 'Expanded', value: 'expanded' })
				),
				$.uiinputselect({ id: 'text2family', title: 'Font', config: {
					options: fonts,
					history: 10,
					processOption: function() {
						this.style.fontFamily = `'${this.value}'`;
					}
				}, change: function() {
					gCurrentWidget.text.ApplyStyle('family', this.value);
				}, textchanged: function() {
					this.$input.style.fontFamily = `'${this.value}'`;
				} }),
				$.uiinputselect({ id: 'text2size', title: 'Font Size', css: {
					minWidth: '40px'
				}, config: {
					options: (() => {
						const fontSizes = [];
						for (let value = 8; value <= 400; ++value) {
							fontSizes.push({ name: value, value });
						}
						return fontSizes;
					})()
				}, change: function() {
					gCurrentWidget.text.ApplyStyle('size', parseInt(this.value, 10));
				} }),
				$.uiinputselect({ id: 'text2lineheight', title: 'Line Height', config: {
					options: (() => {
						const lineHeights = [{ name: 'Default', value: -1 }];
						for (let value = 8; value <= 400; ++value) {
							lineHeights.push({ name: value, value });
						}
						return lineHeights;
					})()
				}, change: function() {
					gCurrentWidget.text.ApplyStyle('lineheight', parseInt(this.value, 10));
				} }),
				$.uicheckbox({ id: 'text2bold', class: 'toolbar-checkbox', title: 'Bold (Ctrl+B)', config: { icon: 'format-bold' }, change: function() {
					gCurrentWidget.text.ApplyStyle('bold', this.checked);
				} }),
				$.uicheckbox({ id: 'text2italic', class: 'toolbar-checkbox', title: 'Italic (Ctrl+I)', config: { icon: 'format-italic' }, change: function() {
					gCurrentWidget.text.ApplyStyle('italic', this.checked);
				} }),
				$.uicheckbox({ id: 'text2underline', class: 'toolbar-checkbox', title: 'Underline (Ctrl+U)', config: { icon: 'format-underline' }, change: function() {
					gCurrentWidget.text.ApplyStyle('underline', this.checked);
				} }),
				$.uicheckbox({ id: 'text2strikethrough', class: 'toolbar-checkbox', title: 'Strikethrough', config: { icon: 'format-strikethrough' }, change: function() {
					gCurrentWidget.text.ApplyStyle('strikethrough', this.checked);
				} }),
				$.uicheckbox({ id: 'text2subscript', class: 'toolbar-checkbox', title: 'Subscript (Ctrl+=)', config: { icon: 'format-subscript' }, change: function() {
					gCurrentWidget.text.ApplyStyle('subscript', this.checked);
				} }),
				$.uicheckbox({ id: 'text2superscript', class: 'toolbar-checkbox', title: 'Superscript (Ctrl+Shift+=)', config: { icon: 'format-superscript' }, change: function() {
					gCurrentWidget.text.ApplyStyle('superscript', this.checked);
				} }),
				$.fieldset({ id: 'text2horizontal', class: 'sampleGroupedButtons' },
					$.input({ type: 'radio', id: 'text2left', value: 'left', name: 'text2horizontal' }),
					$.label({ for: 'text2left', title: 'Left', class: 'mdi mdi-format-align-left' }),
					$.input({ type: 'radio', id: 'text2center', value: 'center', name: 'text2horizontal' }),
					$.label({ for: 'text2center', title: 'Center', class: 'mdi mdi-format-align-center' }),
					$.input({ type: 'radio', id: 'text2right', value: 'right', name: 'text2horizontal' }),
					$.label({ for: 'text2right', title: 'Right', class: 'mdi mdi-format-align-right' })
				),
				$.uicolor({ id: 'text2color', title: 'Font Color', config: {
					compact: true,
					showPalette: true,
					localStorageKey: 'globalcolor',
					previewChange: function() {
						gCurrentWidget.text.ApplyStyle('color', this.value);
					},
					opacity: true
				}, change: function() {
					if (gCurrentWidget) {
						gCurrentWidget.text.ApplyStyle('color', this.value);
					}
				} }),
				$.uicolor({ id: 'text2highlight', title: 'Highlight Color', config: {
					compact: true,
					showPalette: true,
					localStorageKey: 'globalcolor',
					previewChange: function() {
						gCurrentWidget.text.ApplyStyle('highlight', this.value, true, false);
					},
					opacity: true,
					transparentButton: true
				}, change: function() {
					if (gCurrentWidget) {
						gCurrentWidget.text.ApplyStyle('highlight', this.value, true, true);
					}
				} }),
				$.label({ title: 'Insert metadata text reference at cursor', class: 'mdi mdi-code-brackets', hidden: !permissions['Metadata'], css: {
					fontSize: '22px',
					cursor: 'pointer',
					width: '32px',
					height: '32px',
					display: 'flex',
					alignItems: 'center',
					justifyContent: 'center'
				}, click: () => {
					const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: { color: 'white', width: '750px' } },
						$.header({ text: 'Insert Inline Text' }),
						$.div({ class: 'formgrid' },
							$.label({ for: 'textInsertMetadata-text', text: 'Metadata Name' }),
							$.input({ type: 'text', id: 'textInsertMetadata-text' }),
							$.label({ for: 'textInsertMetadata-default', text: 'Default Text (Optional)' }),
							$.input({ type: 'text', id: 'textInsertMetadata-default' })
						),
						$.div({ class: 'panel-buttons' },
							$.button({ text: 'Cancel', click: function() {
								$dialog.remove();
							} }),
							$.button({ text: 'OK', click: function() {
								const metadataName = document.getElementById('textInsertMetadata-text').value;
								const defaultValue = document.getElementById('textInsertMetadata-default').value;
								if (metadataName == '') {
									toastError('Enter a metadata property name.');
								} else {
									const textToAdd = `[metadata="${metadataName}"${defaultValue == '' ? '' : ` default="${defaultValue}"`}]`;
									if (gCurrentWidget) {
										gCurrentWidget.text.InsertStringAtSelection(textToAdd);
									}
									$dialog.remove();
								}
							} })
						)
					));
					centerOnScreen($dialog);
				} })
			),
			$.div({ id: 'buttongrouptoolbar', hidden : true },
				$.div({ id: 'dragButton', title: 'Drag and drop this inside the button group to add a button' },
					$.i({ class: 'mdi mdi-card-plus-outline' }),
					$.span({ text: 'Button' })
				),
				$.span()
			),
			$.button({ id: 'exitwidgetmode', class: 'mdi mdi-close', title: 'Exit widget edit mode', click: ExitWidgetMode })
		),
		$editorCanvasContainer = $.div({ id: 'editor-canvascontainer', scroll: function() {
			LineLayerClear();
			LineLayerHighlightZone(gSelectedZone, true);
		} },
			$.span({ id: 'editor-canvastip', text: 'Create a new playlist or open an existing playlist to begin using the menu on the top right.' }),
			$editorCanvasShadow = $.div({ id: 'editor-canvasshadow' },
				$editorCanvas = $.div({ id: 'editor-canvas', class: 'noselect' })
			),
			$lineLayer = $.canvas({ id: 'lineLayer' }),
			$.iframe({ id: 'previewIframe', sandbox: 'allow-scripts allow-same-origin', hidden: true })
		),
		$editorPreviewControls = $.div({ id: 'editor-preview-controls' },
			$.span({ class: 'noselect', text: 'Preview' }),
			$playStop = $.i({ id: 'playStop', class: 'mdi mdi-play', title: 'Play preview', click: playPreviewPressed }),
			$.i({ id: 'changeCursor', class: 'tooltip-button mdi mdi-cursor-default-outline', click: changeCursor, title: 'Show or hide a cursor over the preview.' })
		),
		$editorTimelineZonesToolbar = $.div({ id: 'editor-timelinezones-toolbar' },
			$.ul({ css: { display: 'flex', flexFlow: 'row nowrap' } },
				$.li({ id: 'editor-view-tabs-timeline', text: 'Timeline', click: () => setTimeout(UpdateTimeBar, 0) }),
				$.li({ text: 'Zones' })
			),
			$.i({ id: 'swapZoneOrder', class: 'tooltip-button mdi mdi-swap-vertical', click: swapZoneOrder, title: 'Toggle zone ordering. This is a visual preference putting the top most zone in the canvas at the top or the bottom. This also changes the zone order in the zone view.' }),
			$.span({ id: 'editor-removeexpiredmedia', class: 'tooltip-button', title: 'The current playlist has scheduled media that has expired. Clicking this will remove the expired media.', text: 'Remove Expired Media' })
		),
		$.div({ id: 'editor-scale-canvas' },
			$canvasSlider = $.uislider({ id: 'scaleCanvas', value: 100, title: 'Zoom slider', config: {
				min: 10,
				max: 800,
				step: 10
			},
			change: function(e) {
				if (!gExpand) {
					toggleExpand(this.value);
				} else {
					resizeCanvas(this.value);
				}
			} }),
			$.i({ id: 'expandEditorCanvas', class: 'tooltip-button mdi mdi-arrow-expand-all', click: () => toggleExpand(), title: 'Expand the canvas area to the size of the playlist.' })
		),
		$timelineView = $.div({ id: 'timeline-view' },
			$timelineOptions = $.span({ id: 'timeline-options' },
				$.i({ id: 'shrinkEditorTimeline', class: 'tooltip-button mdi mdi-arrow-expand-horizontal', click: shrinkEditorTimeline, title: 'Shrink the timeline media items to see more. This shrunken view does not allow resizing via dragging.' }),
				$.i({ id: 'toggleallzones', class: 'tooltip-button mdi mdi-eye-outline', click: toggleAllZonesPressed }),
				$.i({ id: 'deletezone', class: 'tooltip-button mdi mdi-trash-can-outline', click: deleteCurrentZonePressed, title: 'Delete selected zone' }),
				$.i({ id: 'addzone', class: 'tooltip-button mdi mdi-plus-circle-outline', click: addNewZone, title: 'Add new zone' })
			),
			$.canvas({ id: 'timeBar', width: '100', height: '33', title: 'Click to view what the playlist looks like at a certain time.', click: selectTimeBarLocation }),
			$timelineNameResize = $.div({ id: 'timelinename-resize' }),
			$.div({ id: 'timeBarLine' }),
			$timeBarButtonZoneHighlight = $.div({ id: 'timeBarButtonZoneHighlight' })
		),
		$zoneView = $.div({ id: 'zone-view' },
			$.div({},
				$.div({ id: 'zone-view-addzone', class: 'tooltip-button mdi mdi-plus-circle-outline',  click: addNewZone, title: 'Add new zone' })
			)
		)
	);
	
	function updateLayout() {
		const timelineResizeStorageName = 'editor-t';
		const isLandscape = localStorage.editorlayout == 'landscape';
		document.querySelector('.timelinezones-resize')?.remove();
		if (isLandscape) {
			const $timelineOrZone = document.getElementById('editor-view-tabs-timeline').classList.contains('selected') ? $timelineView : $zoneView;
			if (!localStorage.hasOwnProperty(timelineResizeStorageName)) {
				localStorage[timelineResizeStorageName] = $timelineOrZone.offsetHeight;
			} else {
				for (const $e of [$timelineView, $zoneView]) $e.style.height = `${parseFloat(localStorage[timelineResizeStorageName])}px`;
			}
			$editorPreviewControls.insertAdjacentElement('beforebegin', $.div({ class: 'vertical-resize timelinezones-resize', draggable: {
				scrollParent: 'body',
				start: () => {
					document.body.style.cursor = 'ns-resize';
				},
				drag: function(e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY) {
					const $timelineOrZone = document.getElementById('editor-view-tabs-timeline').classList.contains('selected') ? $timelineView : $zoneView;
					const timelineViewOffset = $.offset($timelineOrZone);
					const oldHeight = $timelineOrZone.offsetHeight;
					const height = Math.ceil(timelineViewOffset.top - $editorTimelineZonesToolbar.offsetHeight + oldHeight - e.pageY - startOffsetY);
					for (const $e of [$timelineView, $zoneView]) $e.style.height = `${height}px`;
					const newHeight = $timelineOrZone.offsetHeight;
					document.scrollingElement.scrollTop += newHeight - oldHeight;
					localStorage[timelineResizeStorageName] = $timelineOrZone.offsetHeight;
					UpdateTimeBar();
					resizeCanvas();
				},
				stop: () => {
					document.body.style.cursor = '';
				}
			} }));
		} else {
			const $timelineOrZone = document.getElementById('editor-view-tabs-timeline').classList.contains('selected') ? $timelineView : $zoneView;
			if (!localStorage.hasOwnProperty(timelineResizeStorageName)) {
				localStorage[timelineResizeStorageName] = $timelineOrZone.offsetWidth;
			} else {
				for (const $e of [$timelineView, $zoneView]) $e.style.width = `${parseFloat(localStorage[timelineResizeStorageName])}px`;
			}
			$editorPreviewControls.insertAdjacentElement('beforebegin', $.div({ class: 'horizontal-resize timelinezones-resize', draggable: {
				scrollParent: 'body',
				start: () => {
					document.body.style.cursor = 'ns-resize';
				},
				drag: function(e, startX, startY, startOffsetX, startOffsetY, deltaX, deltaY) {
					const $timelineOrZone = document.getElementById('editor-view-tabs-timeline').classList.contains('selected') ? $timelineView : $zoneView;
					const timelineViewOffset = $.offset($timelineOrZone);
					const oldWidth = $timelineOrZone.offsetWidth;
					const right = timelineViewOffset.left + $timelineOrZone.offsetWidth;
					const left = e.pageX - startOffsetX + this.offsetWidth;
					const width = Math.ceil(right - left);
					for (const $e of [$timelineView, $zoneView]) $e.style.width = `${width}px`;
					const newWidth = $timelineOrZone.offsetWidth;
					document.scrollingElement.scrollLeft += newWidth - oldWidth;
					localStorage[timelineResizeStorageName] = $timelineOrZone.offsetWidth;
					UpdateTimeBar();
					resizeCanvas();
				},
				stop: () => {
					document.body.style.cursor = '';
				}
			} }));
		}
		
		document.body.classList.toggle('landscape', isLandscape);
		document.body.classList.toggle('portrait', !isLandscape);
		$editorLayout.classList.toggle('mdi-phone-rotate-portrait', isLandscape);
		$editorLayout.classList.toggle('mdi-phone-rotate-landscape', !isLandscape);
		$editorLayout.title = isLandscape ? 'Switch to portrait layout' : 'Switch to landscape layout';
	}
	$editorLayout.addEventListener('click', () => {
		localStorage.editorlayout = localStorage.editorlayout == 'landscape' ? 'portrait' : 'landscape';
		if (localStorage.editorlayout == 'landscape') {
			for (const $e of [$timelineView, $zoneView]) Object.assign($e.style, { width: '', height: '300px' });
		} else {
			for (const $e of [$timelineView, $zoneView]) Object.assign($e.style, { width: '500px', height: '' });
		}
		updateLayout();
		UpdateTimeBar();
		UpdateLineLayer();
		resizeCanvas();
	});
	updateLayout();
	
	document.getElementById('navigation-toggle').addEventListener('click', () => {
		resizeCanvas();
	});
	
	document.body.classList.remove('loading');
	InitializeFonts();
	window.addEventListener('keydown', function(e) {
		if (e.ctrlKey) {
			if (e.key == 's') {
				e.preventDefault();
				// The catch is for when the user hits cancel in the dialog
				SavePlaylist().catch(() => {});
				return false;
			} else if (e.key == 'o') {
				e.preventDefault();
				openPlaylistPressed();
				return false;
			}
		}
	});
	
	// Force the line layer to redraw when the ctrl key is toggled since only mouse move and other actions redraw it. Basically if you hover over something and press ctrl pointermove isn't called so this forces it
	const keyDownKeyUp = e => {
		if (e.key == 'Control') {
			CanvasPointerMove({ pageX: canvasLastPageX, pageY: canvasLastPageY, ctrlKey: e.ctrlKey });
		}
	};
	window.addEventListener('keydown', keyDownKeyUp);
	window.addEventListener('keyup', keyDownKeyUp);
	
	for (const $element of document.querySelectorAll('input[name="text2horizontal"]')) {
		$element.addEventListener('change', function() {
			gCurrentWidget.text.ApplyStyle('align', this.value);
		});
	}
	
	InitializeWidgets();
	
	for (const $a of document.querySelectorAll('#navigation a')) {
		$a.addEventListener('click', async function(e) {
			const url = this.href;
			if (url != 'https://link.keywesttechnology.com/mgn59' &&
				url != 'https://link.keywesttechnology.com/o9nl8' &&
				url != 'https://link.keywesttechnology.com/d49o5' &&
				url != 'https://link.keywesttechnology.com/portal') {
				if (gSaveFlag) {
					e.preventDefault();
					try {
						switch (await ConfirmationBox('Save playlist?', 'Do you wish to save your playlist before navigating away from the editor?', 'Yes', 'No', 'Cancel')) {
							case 0:
								await SavePlaylist();
								UnbindOnBeforeUnload();
								window.location.replace(url);
								break;
							case 1:
								UnbindOnBeforeUnload();
								window.location.replace(url);
								break;
						}
					} catch (exception) {
						// SavePlaylist was cancelled by the user
					}
				} else {
					UnbindOnBeforeUnload();
					window.location.replace(url);
				}
			}
		});
	}
	
	$collapseButtonGroups.addEventListener('click', e => {
		e.stopPropagation();
		for (const [media] of GetAllButtonsGrouped()) {
			media.collapsed = true;
		}
		UpdateButtonList();
	}, true);
	
	gTimeBarSelection = 0;
	setTimeout(UpdateTimeBar, 0);
	
	customTabs(document.querySelectorAll('#mediatab,#widgettab,#buttontab'), document.querySelectorAll('#medias,#widgets,#interactivebuttons'));
	customTabs(document.querySelectorAll('#editor-timelinezones-toolbar > ul:first-child > li'), [$timelineView, $zoneView]);
	
	document.getElementById('playlist-editlayout').addEventListener('click', e => {
		e.stopPropagation();
		const fixScreenOptions = () => {
			let width = parseInt($width.value, 10);
			let height = parseInt($height.value, 10);
			
			if ($landscape.checked && width < height) {
				$portrait.checked = true;
			} else if ($portrait.checked && width > height) {
				$landscape.checked = true;
			}
			
			if ($landscape.checked && Math.abs(width / height - 16 / 9) < 0.1 || $portrait.checked && $16x9.checked && Math.abs(width / height - 9 / 16) < 0.1) {
				$16x9.checked = true;
			} else if ($landscape.checked && Math.abs(width / height - 4 / 3) < 0.1 || $portrait.checked && $4x3.checked && Math.abs(width / height - 3 / 4) < 0.1) {
				$4x3.checked = true;
			} else {
				$customRatio.checked = true;
			}
		};
		
		const onScreenOptionChange = () => {
			let width = parseInt($width.value, 10);
			let height = parseInt($height.value, 10);
			
			// Swap the width and height if they don't match landscape or portrait expectations
			if ($landscape.checked && width < height || $portrait.checked && width > height) {
				[width, height, $width.value, $height.value] = [height, width, $height.value, $width.value];
			}
			
			if ($16x9.checked) {
				$height.value = $landscape.checked ? Math.round(width / (16 / 9)) : Math.round(width * (16 / 9));
			} else if ($4x3.checked) {
				$height.value = $landscape.checked ? Math.round(width / (4 / 3)) : Math.round(width * (4 / 3));
			} else {
				// Custom ratio
				// Do nothing
			}
			fixScreenOptions();
		}
		
		const $dialog = document.body.appendChild($.uidialog({ config: { modal: true }, css: {
			width: '400px'
		} },
			$.header({ text: 'Edit Dimensions' }),
			$.div({ class: 'formgrid' },
				$.label({ for: 'aspectRatio', text: 'Aspect Ratio' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'radio', id: 'editplaylistlayout-16x9', name: 'aspectRatio', value: '16x9', checked: true, change: onScreenOptionChange }),
					$.label({ for: 'editplaylistlayout-16x9', text: '16x9', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'editplaylistlayout-4x3', name: 'aspectRatio', value: '4x3', change: onScreenOptionChange }),
					$.label({ for: 'editplaylistlayout-4x3', text: '4x3', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'editplaylistlayout-customratio', name: 'aspectRatio', value: 'custom', change: onScreenOptionChange }),
					$.label({ for: 'editplaylistlayout-customratio', text: 'Custom' })
				),
				$.label({ for: 'orientation', text: 'Orientation' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'radio', id: 'editplaylistlayout-landscape', name: 'orientation', value: 'landscape', checked: true, change: onScreenOptionChange }),
					$.label({ for: 'editplaylistlayout-landscape', text: 'Landscape', css: { marginRight: '10px' } }),
					$.input({ type: 'radio', id: 'editplaylistlayout-portrait', name: 'orientation', value: 'portrait', change: onScreenOptionChange }),
					$.label({ for: 'editplaylistlayout-portrait', text: 'Portrait' }),
				),
				$.label({ text: 'Screen Units' }),
				$.div({ css: { padding: '5px 0' } },
					$.input({ type: 'text', id: 'editplaylistlayout-width', size: 4, maxlength: 6, class: 'sampleInput', css: { width: '100px' }, value: gCurrentPlaylist.pagewidth, input: fixScreenOptions, change: fixScreenOptions }),
					$.span({ text: 'x', css: { margin: '0 5px' } }),
					$.input({ type: 'text', id: 'editplaylistlayout-height', size: 4, maxlength: 6, class: 'sampleInput', css: { width: '100px' }, value: gCurrentPlaylist.pageheight, input: fixScreenOptions, change: fixScreenOptions })
				)
			),
			$.div({ class: 'panel-buttons' },
				$.button({ text: 'Cancel', click: function() {
					$dialog.remove();
				} }),
				$.button({ text: 'OK', click: function() {
					const width = parseInt($width.value, 10);
					const height = parseInt($height.value, 10);
					if (isNaN(width) || isNaN(height) || width <= 0 || width > 46080 || height <= 0 || height > 46080) {
						toastError('Invalid width or height entered. Dimensions must be less than 46080 pixels.');
						$width.value = gCurrentPlaylist.pagewidth;
						$height.value = gCurrentPlaylist.pageheight;
					} else {
						const oldWidth = gCurrentPlaylist.pagewidth;
						const oldHeight = gCurrentPlaylist.pageheight;
						gCurrentPlaylist.pagewidth = width;
						gCurrentPlaylist.pageheight = height;
						ResizePlaylist(oldWidth, oldHeight);
						reloadAllZoneViews();
						resizeCanvas();
						UpdateProperties();
						SetSaveFlag(true);
						$dialog.remove();
					}
				} })
			)
		));
		centerOnScreen($dialog);

		const $width = document.getElementById('editplaylistlayout-width');
		const $height = document.getElementById('editplaylistlayout-height');
		const $16x9 = document.getElementById('editplaylistlayout-16x9');
		const $4x3 = document.getElementById('editplaylistlayout-4x3');
		const $customRatio = document.getElementById('editplaylistlayout-customratio');
		const $landscape = document.getElementById('editplaylistlayout-landscape');
		const $portrait = document.getElementById('editplaylistlayout-portrait');
		
		fixScreenOptions();
	}, true);
	
	for (const $e of document.querySelectorAll('#properties > header')) {
		$e.addEventListener('click', function(e) {
			if ($e.classList.contains('collapsed') || e.target.classList.contains('mdi')) {
				$e.classList.toggle('collapsed');
				localStorage[$e.id] = $e.classList.contains('collapsed');
			}
		});
		if (!localStorage.hasOwnProperty($e.id)) {
			localStorage[$e.id] = true;
		}
		$e.classList.toggle('collapsed', localStorage[$e.id] == 'true');
	}
	
	for (const $e of document.querySelectorAll('#zone-x, #zone-y, #media-x, #media-y, #media-button-x, #media-button-y')) {
		$e.addEventListener('keydown', onlyInteger);
	}
	for (const $e of document.querySelectorAll('#zone-width, #zone-height, #media-width, #media-height, #media-button-width, #media-button-height, #media-runlength')) {
		$e.addEventListener('keydown', onlyPositiveIntegers);
	}
	for (const $e of document.querySelectorAll('#contentTree-addgroup, #contentTree-delete, #contentTree-upload')) {
		$e.hidden = true;
	}
	$contentTree.addEventListener('change', () => {
		if (permissions['Metadata']) {
			UpdateProperties('metadata');
		}
	});
	$contentTree.addEventListener('nochange', () => {
		if (permissions['Metadata']) {
			UpdateProperties('metadata');
		}
	});
	$contentTree.addEventListener('dragover', handleDragOver, false);
	$contentTree.addEventListener('drop', handleFileSelect, false);
	
	setInterval(fileCheckLoop, 2000);
	
	const $dragButton = document.getElementById('dragButton');
	$dragButton.classList.add('widgetbutton');
	draggable($dragButton, {
		thresholdX: 3,
		thresholdY: 3,
		start: e => {
			if (DialogWindowOpen() || InPreviewMode() || gCurrentPlaylist == null) {
				BlinkDialog();
				return false;
			}
			return $dragButton;
		},
		drag: e => {
			var zone = getZoneAt(e);
			LineLayerClear();
			LineLayerHighlightZone(zone, true);
		}
	});
	
	var startZone = null;
	var startMedia = null;
	var startMediaComponent = null;
	var startZoneDimensions = null;
	var startMediaDimensions = null;
	var startMediaComponentDimensions = null;
	var doubleClickTime = 0;
	var buttonClickTimeout = null;
	draggable($editorCanvas, {
		scrollParent: '#editor-canvascontainer',
		start: (e, startX, startY, startOffsetX, startOffsetY) => {
			// Force a blur before any other actions
			document.activeElement.blur();
			
			if (DialogWindowOpen() || InPreviewMode()) {
				return false;
			}
			
			const canvasX = Math.round(toCanvasScale(startOffsetX));
			const canvasY = Math.round(toCanvasScale(startOffsetY));
			const doubleClick = Date.now() - doubleClickTime < 200;
			
			({ zone: startZone, media: startMedia, component: startMediaComponent, type: startMoveType } = getMoveAt(canvasX, canvasY, e.ctrlKey));
			
			if (startZone != null && gSelectedZone === startZone && gCurrentWidget && (gCurrentWidget.widgettype == 'buttongroup' || gCurrentWidget.widgettype == 'text2') &&
				(isPointInRect(canvasX, canvasY, gSelectedZone.x + gCurrentWidget.x + gCurrentWidget.width - 30, gSelectedZone.y + gCurrentWidget.y, 30, 30) ||
				startMediaComponent ||
				startMoveType & MOVETYPE_MOVE_X || startMoveType & MOVETYPE_MOVE_Y)) {
				if (startMediaComponent) {
					gCurrentWidgetComponent = startMediaComponent;
					gSelectMediaLocation = 'canvas';
					
					startMediaComponentDimensions =
					{
						x: startMediaComponent.x,
						y: startMediaComponent.y,
						width: startMediaComponent.width,
						height: startMediaComponent.height
					};
					
					setTimeout(() => {
						if (gCurrentWidget) {
							if (document.getElementById('editor-view-tabs-timeline').classList.contains('selected')) {
								gCurrentWidget.$timelineView.focus();
							} else {
								gCurrentWidget.$zoneView.focus();
							}
						}
					}, 0);
					UpdateButtonList();
					UpdateProperties('media');
					
					if (doubleClick) {
						startMoveType = MOVETYPE_NONE;
						EditWidget(gSelectedZone, {
							button: startMediaComponent,
							widgettype: 'button',
							get attributes() { return startMediaComponent; },
							set attributes(attributes) { Object.assign(startMediaComponent, attributes); },
							ResizeMedia: () => {}
						}, gSelectedZone.x + gCurrentWidget.x + startMediaComponent.x, gSelectedZone.y + gCurrentWidget.y + startMediaComponent.y, startMediaComponent.width, startMediaComponent.height);
						return false;
					}
				} else if (startMoveType & MOVETYPE_MOVE_X || startMoveType & MOVETYPE_MOVE_Y) {
					startMoveType = MOVETYPE_NONE;
					return false;
				}
			} else if (startZone == null && startMedia == null && startMediaComponent == null) {
				SelectZone(null, true, false);
			} else if (startZone) {
				gSelectMediaLocation = 'canvas';
				
				var media = startZone.GetSelectedMedia();
				if (media && isPointInRect(canvasX, canvasY, startZone.x + media.x, startZone.y + media.y, media.width, media.height)) {
					if (media.type == MEDIATYPE_WIDGET) {
						clearTimeout(buttonClickTimeout);
						buttonClickTimeout = null;
						if (doubleClick) {
							startMoveType = MOVETYPE_NONE;
							EditWidget(startZone, media);
							return false;
						} else if (media.widgettype == 'buttongroup') {
							var selectedButton = media.attributes.buttons.map(button => button).reverse().find(button => button.clickable && isPointInRect(canvasX, canvasY, startZone.x + media.x + button.x, startZone.y + media.y + button.y, button.width, button.height));
							buttonClickTimeout = setTimeout(() => {
								if (selectedButton)
								{
									ClickButton(selectedButton);
									UpdateButtonList();
									return false;
								}
							}, 200);
						}
					} else {
						if (doubleClick) {
							UpdateProperties('media');
							return false;
						}
					}
				}
				
				// A locked zone is still selectable even though it has no returned move type
				if (startMoveType != MOVETYPE_NONE || startZone.locked) {
					if (e.ctrlKey) {
						SelectMedia(startZone, [startZone.media.indexOf(startMedia)], false, false, false, 'canvas', true, true);
						if (startMedia) {
							// The user can't drag or resize blank widgets
							if (startMedia.type == MEDIATYPE_WIDGET && startMedia.widgettype == 'blank') {
								startMoveType = MOVETYPE_NONE;
								return false;
							}
							startMediaDimensions =
							{
								x: startMedia.x,
								y: startMedia.y,
								width: startMedia.width,
								height: startMedia.height
							};
						}
					} else {
						SelectZone(startZone, true, false);
						startZoneDimensions =
						{
							x: startZone.x,
							y: startZone.y,
							width: startZone.width,
							height: startZone.height
						};
						startMediaDimensions = startZone.media.map(media =>
						({
							x: media.x,
							y: media.y,
							width: media.width,
							height: media.height
						}));
					}
				}
			}
			doubleClickTime = Date.now();
			if (startZone == null && startMedia == null && startMediaComponent == null) {
				return false;
			}
		},
		drag: (e, startX, startY, startOffsetX, startOffsetY) => {
			//var canvasX = Math.round(toCanvasScale(startOffsetX));
			//var canvasY = Math.round(toCanvasScale(startOffsetY));
			
			var thresholdDistanceSquared = 10 * 10;
			
			const [snapX, snapY] = GetPlaylistSnapLines(startZoneDimensions == null && startMediaDimensions ? true : false, gCurrentWidget != null && startMediaComponent != null);
			var dx = Math.round(toCanvasScale(e.pageX - startX));
			var dy = Math.round(toCanvasScale(e.pageY - startY));
			
			var snapValue = (values, delta, snapValues, min, max) => {
				min = min == undefined ? Number.NEGATIVE_INFINITY : min;
				max = max == undefined ? Number.POSITIVE_INFINITY : max;
				var closestSnapDelta = null;
				var closestSnapDistanceSquared = Number.POSITIVE_INFINITY;
				var minDelta = Number.NEGATIVE_INFINITY;
				var maxDelta = Number.POSITIVE_INFINITY;
				for (const value of values) {
					var newValue = value + delta;
					if (newValue < min) {
						minDelta = Math.max(minDelta, min - value);
					}
					if (newValue > max) {
						maxDelta = Math.min(maxDelta, max - value);
					}
					for (const snap of snapValues) {
						if (snap.value >= min && snap.value < max) {
							var distanceSquared = (newValue - snap.value) * (newValue - snap.value);
							if (distanceSquared < thresholdDistanceSquared && distanceSquared < closestSnapDistanceSquared) {
								closestSnapDistanceSquared = distanceSquared;
								closestSnapDelta = snap.value - value;
								for (const resetSnapHighlight of snapValues) {
									resetSnapHighlight.highlight = false;
								}
								snap.highlight = true;
							} else if (distanceSquared == closestSnapDistanceSquared) {
								snap.highlight = true;
							}
						}
					}
				}
				return Math.max(minDelta, Math.min(maxDelta, closestSnapDelta != null ? closestSnapDelta : delta));
			};
			
			if (startZone && startZone.locked) {
				return;
			}
			
			LineLayerClear();
			if (startMediaComponentDimensions) {
				if (startMoveType & MOVETYPE_MOVE_X || startMoveType & MOVETYPE_MOVE_Y) {
					dx = snapValue([startZone.x + startMedia.x + startMediaComponentDimensions.x, startZone.x + startMedia.x + startMediaComponentDimensions.x + startMediaComponentDimensions.width, Math.round(startZone.x + startMedia.x + startMediaComponentDimensions.x + startMediaComponentDimensions.width / 2)], dx, snapX);
					startMediaComponent.x = startMediaComponentDimensions.x + dx;
					dy = snapValue([startZone.y + startMedia.y + startMediaComponentDimensions.y, startZone.y + startMedia.y + startMediaComponentDimensions.y + startMediaComponentDimensions.height, Math.round(startZone.y + startMedia.y + startMediaComponentDimensions.y + startMediaComponentDimensions.height / 2)], dy, snapY);
					startMediaComponent.y = startMediaComponentDimensions.y + dy;
					PopulateButtonLayoutProperties(startZone);
					startZone.UpdateView();
					SetSaveFlag(true);
				} else {
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP) {
						if (startMoveType & MOVETYPE_RESIZE_LEFT) {
							dx = snapValue([startZone.x + startMedia.x + startMediaComponentDimensions.x], dx, snapX, Number.NEGATIVE_INFINITY, startZone.x + startMedia.x + startMediaComponentDimensions.x + startMediaComponentDimensions.width - 5);
							startMediaComponent.x = startMediaComponentDimensions.x + dx;
							startMediaComponent.width = startMediaComponentDimensions.width - dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_TOP) {
							dy = snapValue([startZone.y + startMedia.y + startMediaComponentDimensions.y], dy, snapY, Number.NEGATIVE_INFINITY, startZone.y + startMedia.y + startMediaComponentDimensions.y + startMediaComponentDimensions.height - 5);
							startMediaComponent.y = startMediaComponentDimensions.y + dy;
							startMediaComponent.height = startMediaComponentDimensions.height - dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						if (startMoveType & MOVETYPE_RESIZE_RIGHT) {
							dx = snapValue([startZone.x + startMedia.x + startMediaComponentDimensions.x + startMediaComponentDimensions.width], dx, snapX, startZone.x + startMedia.x + startMediaComponentDimensions.x + 5);
							startMediaComponent.width = startMediaComponentDimensions.width + dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_BOTTOM) {
							dy = snapValue([startZone.y + startMedia.y + startMediaComponentDimensions.y + startMediaComponentDimensions.height], dy, snapY, startZone.y + startMedia.y + startMediaComponentDimensions.y + 5);
							startMediaComponent.height = startMediaComponentDimensions.height + dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP || startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						PopulateButtonLayoutProperties(startZone);
						startZone.UpdateView();
						SetSaveFlag(true);
					}
				}
				LineLayerResizeBoxMediaComponent(startZone, startMedia, startMediaComponent, startMoveType);
			} else if (startZoneDimensions) {
				if (startMoveType & MOVETYPE_MOVE_X || startMoveType & MOVETYPE_MOVE_Y) {
					// Zones are constrained by the playlist boundary
					dx = Math.max(dx, 0 - startZoneDimensions.x);
					dx = Math.min(dx, gCurrentPlaylist.pagewidth - (startZoneDimensions.x + startZoneDimensions.width));
					dy = Math.max(dy, dy = 0 - startZoneDimensions.y);
					dy = Math.min(dy, gCurrentPlaylist.pageheight - (startZoneDimensions.y + startZoneDimensions.height));
					
					dx = snapValue([startZoneDimensions.x, startZoneDimensions.x + startZoneDimensions.width, Math.round(startZoneDimensions.x + startZoneDimensions.width / 2)], dx, snapX);
					dy = snapValue([startZoneDimensions.y, startZoneDimensions.y + startZoneDimensions.height, Math.round(startZoneDimensions.y + startZoneDimensions.height / 2)], dy, snapY);
					startZone.x = startZoneDimensions.x + dx;
					startZone.y = startZoneDimensions.y + dy;
					PopulateZoneLayoutProperties(startZone);
					Object.assign(startZone.$canvasView.style, {
						left: `${startZone.x}px`,
						top: `${startZone.y}px`
					});
					SetSaveFlag(true);
				} else {
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP) {
						if (startMoveType & MOVETYPE_RESIZE_LEFT) {
							dx = Math.max(dx, 0 - startZoneDimensions.x);
							dx = snapValue([startZoneDimensions.x], dx, snapX, Number.NEGATIVE_INFINITY, startZoneDimensions.x + startZoneDimensions.width - 5);
							startZone.x = startZoneDimensions.x + dx;
							startZone.width = startZoneDimensions.width - dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_TOP) {
							dy = Math.max(dy, dy = 0 - startZoneDimensions.y);
							dy = snapValue([startZoneDimensions.y], dy, snapY, Number.NEGATIVE_INFINITY, startZoneDimensions.y + startZoneDimensions.height - 5);
							startZone.y = startZoneDimensions.y + dy;
							startZone.height = startZoneDimensions.height - dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						if (startMoveType & MOVETYPE_RESIZE_RIGHT) {
							dx = Math.min(dx, gCurrentPlaylist.pagewidth - (startZoneDimensions.x + startZoneDimensions.width));
							dx = snapValue([startZoneDimensions.x + startZoneDimensions.width], dx, snapX, startZoneDimensions.x + 5);
							startZone.width = startZoneDimensions.width + dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_BOTTOM) {
							dy = Math.min(dy, gCurrentPlaylist.pageheight - (startZoneDimensions.y + startZoneDimensions.height));
							dy = snapValue([startZoneDimensions.y + startZoneDimensions.height], dy, snapY, startZoneDimensions.y + 5);
							startZone.height = startZoneDimensions.height + dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP || startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						startZone.media.forEach((media, mediaIndex) => {
							const startMediaDimension = startMediaDimensions[mediaIndex];
							media.x = startMediaDimension.x;
							media.y = startMediaDimension.y;
							media.width = startMediaDimension.width;
							media.height = startMediaDimension.height;
							media.ResizeMedia(startZoneDimensions.width, startZoneDimensions.height);
						});
						PopulateZoneLayoutProperties(startZone);
						if (startZone.GetSelectedMedia()) {
							PopulateMediaLayoutProperties(startZone);
						}
						startZone.UpdateView();
						SetSaveFlag(true);
					}
				}
				LineLayerResizeBoxZone(startZone, startMoveType);
			} else if (startMediaDimensions) {
				if (startMoveType & MOVETYPE_MOVE_X || startMoveType & MOVETYPE_MOVE_Y) {
					if (startMoveType & MOVETYPE_MOVE_X) {
						dx = snapValue([startZone.x + startMediaDimensions.x, startZone.x + startMediaDimensions.x + startMediaDimensions.width, Math.round(startZone.x + startMediaDimensions.x + startMediaDimensions.width / 2)], dx, snapX);
						startMedia.x = startMediaDimensions.x + dx;
					}
					if (startMoveType & MOVETYPE_MOVE_Y) {
						dy = snapValue([startZone.y + startMediaDimensions.y, startZone.y + startMediaDimensions.y + startMediaDimensions.height, Math.round(startZone.y + startMediaDimensions.y + startMediaDimensions.height / 2)], dy, snapY);
						startMedia.y = startMediaDimensions.y + dy;
					}
					PopulateMediaLayoutProperties(startZone);
					var mediaContainer = startZone.$canvasView.querySelector('.media');
					Object.assign(mediaContainer.style, {
						left: `${startMedia.x}px`,
						top: `${startMedia.y}px`
					});
					SetSaveFlag(true);
				} else {
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP) {
						if (startMoveType & MOVETYPE_RESIZE_LEFT) {
							dx = snapValue([startZone.x + startMediaDimensions.x], dx, snapX, Number.NEGATIVE_INFINITY, startZone.x + startMediaDimensions.x + startMediaDimensions.width - 5);
							startMedia.x = startMediaDimensions.x + dx;
							startMedia.width = startMediaDimensions.width - dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_TOP) {
							dy = snapValue([startZone.y + startMediaDimensions.y], dy, snapY, Number.NEGATIVE_INFINITY, startZone.y + startMediaDimensions.y + startMediaDimensions.height - 5);
							startMedia.y = startMediaDimensions.y + dy;
							startMedia.height = startMediaDimensions.height - dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						if (startMoveType & MOVETYPE_RESIZE_RIGHT) {
							dx = snapValue([startZone.x + startMediaDimensions.x + startMediaDimensions.width], dx, snapX, startZone.x + startMediaDimensions.x + 5);
							startMedia.width = startMediaDimensions.width + dx;
						}
						if (startMoveType & MOVETYPE_RESIZE_BOTTOM) {
							dy = snapValue([startZone.y + startMediaDimensions.y + startMediaDimensions.height], dy, snapY, startZone.y + startMediaDimensions.y + 5);
							startMedia.height = startMediaDimensions.height + dy;
						}
					}
					if (startMoveType & MOVETYPE_RESIZE_LEFT || startMoveType & MOVETYPE_RESIZE_TOP || startMoveType & MOVETYPE_RESIZE_RIGHT || startMoveType & MOVETYPE_RESIZE_BOTTOM) {
						PopulateMediaLayoutProperties(startZone);
						startZone.UpdateView();
						SetSaveFlag(true);
					}
				}
				LineLayerResizeBoxMedia(startZone, startMedia, startMoveType);
			}
			
			if (startMoveType != MOVETYPE_NONE) {
				LineLayerDrawSnapLines(snapX, snapY);
			}
		},
		stop: () => {
			LineLayerClear();
			if (startMediaComponent) {
				LineLayerResizeBoxMediaComponent(gSelectedZone, startMedia, startMediaComponent, startMoveType);
			} else if (startMedia) {
				LineLayerResizeBoxMedia(gSelectedZone, startMedia, startMoveType);
			} else {
				if (gSelectedZone.locked) {
					LineLayerHighlightZone(gSelectedZone, true);
				} else {
					LineLayerResizeBoxZone(gSelectedZone, startMoveType);
				}
			}
			startMoveType = MOVETYPE_NONE;
			startZone = null;
			startMedia = null;
			startMediaComponent = null;
			startZoneDimensions = null;
			startMediaDimensions = null;
			startMediaComponentDimensions = null;
		}
	});
	$editorCanvas.addEventListener('pointermove', CanvasPointerMove);
	
	droppable($editorCanvasShadow, {
		accept: ['content-group', 'content-leaf', 'widget', 'widgetbutton', 'moved-media'],
		drop: ContentDrop,
		region: 'content'
	});
	for (const $drop of [$timelineView, $zoneView]) {
		droppable($drop, {
			accept: ['content-group', 'content-leaf', 'widget', 'moved-media'],
			drop: ContentDrop,
			region: 'content'
		});
	}
	
	document.getElementById('media-entrytransitiontype').append(
		...gTransitionTypes.map(transitionType => $.option({ text: transitionType, value: transitionType }))
	);
	document.getElementById('media-exittransitiontype').append(
		...gTransitionTypes.map(transitionType => $.option({ text: transitionType, value: transitionType }))
	);

	document.getElementById('text2highlight').addEventListener('showpicker', function() {
		if (this.value == 'rgba(255, 255, 255, 0)') {
			this.value = 'rgba(255, 255, 255, 1)';
		}
	});
	document.getElementById('text2highlight').addEventListener('hidepicker', function() {
		if (gCurrentWidget) {
			// Force an update to rerender the highlight
			gCurrentWidget.text.Update((selection, characterStyles) => {});
		}
	});
	
	clearPlaylist();
	ExitWidgetMode(true);
	
	window.addEventListener('message', e => {
		if (e.data.event == 'time' && InPreviewMode()) {
			for (const $e of document.querySelectorAll('.timelineBar .marker')) $e.remove();
			if (localStorage.swapzoneorder == 'true') {
				if (gCurrentPlaylist.zones.length != 0 && gCurrentPlaylist.zones[0].name == 'Audio') {
					e.data.times = e.data.times.slice(0, 1).concat(e.data.times.slice(1).reverse());
				} else {
					e.data.times.reverse();
				}
			}
			
			const $timelineBars = document.getElementsByClassName('timelineBar');
			e.data.times.forEach((time, itr) => {
				if (itr < $timelineBars.length) {
					$timelineBars[itr].appendChild(
						$.div({ class: 'marker', css: {
							display: 'inline-block',
							position: 'absolute',
							top: 0,
							bottom: 0,
							left: `${Math.floor(time * gTimelineScale)}px`,
							width: '1px',
							backgroundColor: 'red',
							boxShadow: '0 0 5px white'
						} })
					);
				}
			});
		} else if (e.data.event == 'updatezoneviews' && InPreviewMode()) {
			e.data.zones.forEach((zone, zoneItr) => {
				var playlistZone = gCurrentPlaylist.zones[zoneItr];
				playlistZone.$timelineName.hidden = !zone.some(showMedia => showMedia);
				playlistZone.$timelineBar.hidden = !zone.some(showMedia => showMedia);
				for (const [mediaItr, showMedia] of zone.entries()) {
					playlistZone.media[mediaItr].$timelineView.hidden = !showMedia;
				}
			});
		}
	}, false);
	
	BindOnBeforeUnload();
	
	// Stop scrolling behavior
	$timelineView.addEventListener('keydown', e => {
		if (['PageDown', 'PageUp', 'Home', 'End', 'ArrowLeft', 'ArrowRight', 'ArrowUp', 'ArrowDown'].includes(e.key) != -1) {
			e.preventDefault();
		}
	});
	/*
	// TODO: Is this necessary anymore? Forgot what it stopped. Dragging images in the timeline?
	$timelineView.addEventListener('dragstart', e => {
		e.preventDefault();
		return false;
	});
	*/
	document.getElementById('editor-removeexpiredmedia').addEventListener('click', () => {
		for (const zone of gCurrentPlaylist.zones) {
			var expiredMedia = [];
			for (const [index, media] of zone.media.entries()) {
				const scheduledState = media.GetScheduledState();
				if (scheduledState.expired) {
					expiredMedia.push(index);
				}
			}
			if (expiredMedia.length != 0) {
				SelectMedia(zone, expiredMedia, false, false, false, '', false, true);
				deleteCurrentMedia(false);
			}
		}
	});
	document.getElementById('editor-removeexpiredmedia').hidden = true;
	
	document.getElementById('editor-new').addEventListener('click', newPlaylistPressed);
	document.getElementById('editor-open').addEventListener('click', openPlaylistPressed);
	document.getElementById('editor-save').addEventListener('click', () => SavePlaylist().catch(() => {}));
	
	document.getElementById('editor-menu').addEventListener('click', () => {
		const $editorMainHeader = document.getElementById('editor-main-header');
		const editorMainHeaderOffset = $.offset($editorMainHeader);
		const $dialog = document.body.appendChild($.ul({ id: 'editor-menu-dialog' },
			$.li({ process: function() {
				if (permissions['Editor Content Tree']) {
					this.appendChild($.button({ text: 'Import Playlist(s)', title: 'Import playlists that were exported into a zip archive', class: 'editor-requires-playlist-permission', css: { marginRight: '20px' }, click: () => { removeDialog(); importPlaylistPressed(); } }));
				}
				if (permissions['Players Admin'] || permissions['Players Operator']) {
					this.append(
						$.div({ class: 'editor-menu-dialog-separator' }),
						$.button({ text: 'Update Scheduled Players', title: 'Update all players that include the currently opened playlist in their current schedule.', click: () => { removeDialog(); UpdateScheduledPlayers(); } })
					);
					
					if (!localStorage.hasOwnProperty('publishplayer')) {
						localStorage['publishplayer'] = null;
					}
					this.appendChild(
						$.div({},
							$.uitreeselect({ id: 'playerPublishTree', title: 'Select a player to publish the current playlist to.', config: {
								tree: 'player',
								showRoot: true,
								leafIcon: node => node.lastcheckin && moment.tz(node.lastcheckin, 'YYYY-MM-DD HH:mm:ss', 'UTC').isAfter(moment.utc(node.utctime * 1000).subtract(1, 'minute')) ? 'playeron' : 'playeroff',
								filter: node => node.parentNode != null
							}, change: async function() {
								const selectedNode = (await this.getSelected())[0];
								const selectedId = selectedNode ? selectedNode.id : null;
								localStorage['publishplayer'] = selectedId;
							}, process: async $treeSelect => {
								await $treeSelect.setSelected(node => {
									if (localStorage['publishplayer'] == null) {
										return false;
									}
									const treeId = parseInt(localStorage['publishplayer'], 10);
									return node.id == treeId;
								});
							} }),
							$.button({ id: 'publishPlaylist', text: 'Publish', title: 'Send only the current playlist to a player. This will generate a schedule containing only this playlist. Note: You will need to use the Players page to set widget data where required.', click: () => { PublishPlaylist(); removeDialog(); } })
						)
					);
				}
			} },
				$.button({ text: 'New Playlist', title: 'Create a new playlist', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); newPlaylistPressed(); } }),
				$.button({ text: 'New Playlist From Template', title: 'Create a new playlist from template', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); newPlaylistPressed(); document.getElementById('template-tabs-customtemplates').click(); } }),
				$.button({ text: 'Open Playlist', title: 'Open an existing playlist', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); openPlaylistPressed(); } }, $.span({ text: 'Ctrl+O', class: 'hotkey' })),
				$.button({ text: 'Save Playlist', title: 'Save this playlist', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); SavePlaylist().catch(() => {}) } }, $.span({ text: 'Ctrl+S', class: 'hotkey' })),
				$.button({ text: 'Save As Playlist', title: 'Save this playlist as a new playlist', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); SaveAsPlaylist().catch(() => {}); } }),
				$.button({ text: 'Save As Template', title: 'Save this playlist as a new template', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); SaveAsTemplate().catch(() => {}); } }),
				$.div({ class: 'editor-menu-dialog-separator' }),
				$.button({ text: 'Export Playlist(s)', title: 'Export playlists and all their content as a zip archive. This archive can then be imported to another server', class: 'editor-requires-playlist-permission', click: () => { removeDialog(); exportPlaylistPressed() } })
			)
		));
		const removeDialog = () => {
			document.removeEventListener('click', documentClickHandler, true);
			$dialog.remove();
		};
		const documentClickHandler = e => {
			if (!$dialog.contains(e.target) && !document.getElementById('playerPublishTree')?.$tree?.contains(e.target)) {
				e.stopImmediatePropagation();
				removeDialog();
			}
		};
		document.addEventListener('click', documentClickHandler, true);
		Object.assign($dialog.style, {
			left: `${editorMainHeaderOffset.left + $editorMainHeader.offsetWidth - $dialog.querySelector('li').offsetWidth}px`,
			top: `${editorMainHeaderOffset.top + $editorMainHeader.offsetHeight}px`
		});
		if (permissions['playlisttree'].length == 0) {
			for (const $requiresPermission of document.getElementsByClassName('editor-requires-playlist-permission')) {
				$requiresPermission.style.cursor = 'not-allowed';
				$requiresPermission.title = 'There are no playlist permissions configured for this user.';
				$requiresPermission.addEventListener('click', function() {
					toastError('There are no playlist permissions configured for this user.');
				});
			}
		}
	});
	
	$contentTree.addEventListener('focus', hideAccordion);
	$contentTree.addEventListener('scroll', hideAccordion);
	
	for (const $widgetDialog of document.getElementsByClassName('widget-dialog')) {
		for (const $accordionHeader of $widgetDialog.getElementsByClassName('widget-accordion-header')) {
			var headerText = $accordionHeader.innerText;
			$accordionHeader.innerText = '';
			$accordionHeader.append(
				$.span({ text: headerText }),
				$.i({ class: 'mdi mdi-play mdi-rotate-90' })
			);
			$accordionHeader.addEventListener('click', function() {
				for (const $e of this.parentNode.querySelectorAll('.widget-accordion-header')) $e.classList.add('collapsed');
				this.classList.remove('collapsed');
			});
		}
		$widgetDialog.querySelector('.widget-accordion-header:first-child')?.click();
	}
	
	if (localStorage.hasOwnProperty('lastplaylistid')) {
		(async () => {
			var lastPlaylistId = parseInt(localStorage.lastplaylistid, 10);
			var data = await LoadPlaylistById(lastPlaylistId);
			if (!data.success && data.hasOwnProperty('removelastplaylist')) {
				delete localStorage.lastplaylistid;
				delete localStorage.zonestate;
				delete localStorage.zoneresize;
				toastError(`Problem opening previous playlist: ${data.error}`);
			}
		})();
	}
	
	window.addEventListener('resize', resizeCanvas);
	
	DragLeftResize('.horizontal-resize', 'editor-lr', () => {
		UpdateTimeBar();
		resizeCanvas();
	});
	
	const timelineNameResizeStorageName = 'editor-tn';
	draggable($timelineNameResize, {
		scrollParent: 'body',
		drag: function(e, startX, startY, startOffsetX, startOffsetY) {
			let width = Math.ceil(e.pageX - startOffsetX - $.offset($timelineOptions).left);
			width = Math.max(130, Math.min(250, width));
			$timelineOptions.style.width = `${width}px`;
			localStorage[timelineNameResizeStorageName] = $timelineOptions.offsetWidth;
			UpdateTimeBar();
		}
	});
	if (!localStorage.hasOwnProperty(timelineNameResizeStorageName)) {
		localStorage[timelineNameResizeStorageName] = parseFloat($timelineOptions.offsetWidth);
	} else {
		let width = parseFloat(localStorage[timelineNameResizeStorageName]);
		width = Math.max(130, Math.min(250, width));
		$timelineOptions.style.width = `${width}px`;
	}
	
	$timelineView.addEventListener('scroll', UpdateTimeBar);
	document.addEventListener('scroll', () => {
		UpdateTimeBar();
		UpdateLineLayer();
	});
	UpdateTimeBar();
	
	if (permissions['playlisttree'].length == 0) {
		for (const $requiresPermission of document.querySelectorAll('.editor-header-requires-playlist-permission')) {
			$requiresPermission.style.cursor = 'not-allowed';
			$requiresPermission.title = 'There are no playlist permissions configured for this user.';
			$requiresPermission.addEventListener('click', function() {
				toastError('There are no playlist permissions configured for this user.');
			});
		}
	}
});